/* $Id: imaging_detector.c,v 1.2 2001/11/26 12:49:41 csabet Exp $ */
/*----------------------------------------------------------------------- */
/* Descriptor table for raw data table for optical/IR interferometer      */
/*----------------------------------------------------------------------- */
#include "imaging_detector.h"
#include <stdlib.h>
#include <string.h>

/* internal prototypes */
/* create basic structure */
imaging_detector_table_def* 
create_imaging_detector_table_def(int nregion, int ndetect, int maxtel, 
				  int num_dim, int max_coef, 
				  char *instrument, char *date_obs);
/* find table columns (global values of results) */
void find_imaging_detector_cols(imaging_detector_table_def *tabledef);

/* file global variables for table type*/
static char *tabletype = "IMAGING_DETECTOR";

/* create a new imaging_detector_table */
imaging_detector_table_def* 
make_imaging_detector_table_def(int nregion, int ndetect, int maxtel, 
				int num_dim, int max_coef,
				char *instrument, char *date_obs,
				int *region, char **regname, int *detector, 
				int **ports, int *correlation, 
				int **corner, double *gain,
				int **naxes,
				double **crval, float **crpix, char ***ctype,
				double **cd, double ***dmc, int ****dmp)
/*----------------------------------------------------------------------- */
/*   Make a new imaging_detector_table_def.                               */
/*   Inputs:                                                              */
/*      nregion  the number of regions (rows) in the table                */
/*      ndetect  the number of detectors used                             */
/*      maxtel   the max number of ports contributing to a region         */
/*      num_dim  number of dimensions in data array                       */
/*      max_coef maximum number of entries in dmp, dmc arrays             */
/*      instrument  the name of the instrument                            */
/*      date_obs    Date (yyyy-mm-dd) observations began                  */
/*      region      array of region numbers                               */
/*      regname     array of region names, NULL=>empty string             */
/*      detector    array of detector numbers, NULL => default (1)        */
/*      ports  array of sets of ports in region, NULL=default(2)          */
/*      correlation array of correlation types for regions                */
/*             0 =background, 1=autocorrelation, 2=crosscorrelation       */
/*      corner      corner for each region.                               */
/*                  as, low x, low y, NULL => default (-1,-1).            */
/*      gain        array detector gain factors                           */
/*      naxes       array of x and y dimensions of the regions            */
/*      crval       array of x and y coordinate values at reference pixels*/
/*                  NULL => default (0)                                   */
/*      crpix       array of x and y reference pixel numbers              */
/*                  NULL => default (1)                                   */
/*      ctype       array oc x an y coordinate type strings               */
/*                  NULL => default ("X", "Y")                            */
/*      cd          array of cd (transformation) matrices as              */
/*                  CD_1_1, CD_1_2, CD_2_1, CD_2_2                        */
/*                  NULL => default (1,0,0,1)                             */
/*      dmc         dedistortion polynomial coefficients[reg][dim][term]  */
/*      dmp         dedistortion polynomial powers [reg][dim][term][dim]  */
/*   Returns:                                                             */
/*      An imaging_detector_table_def pointer; NULL indicates failure     */
/*----------------------------------------------------------------------- */
{
  imaging_detector_table_def *out;
  int                         i, j, k, l;

  /* some sanity checks */
  if (maxtel<=0) maxtel = 2;
  if (maxtel>100) maxtel = 2;

  out = create_imaging_detector_table_def(nregion, ndetect, maxtel, 
					  num_dim,  max_coef,
					  instrument, date_obs);
  if (out==NULL) return NULL; /* allocation failure */
  
  /* copy data with defaults on null */
  for (i=0; i<nregion; i++) {
    out->region[i] = region[i]; /* region number */

    if (regname!=NULL) {       /* region name */
      if (out->regname[i]!=NULL) free (out->regname[i]); /* free old memory */
      out->regname[i] = (char*)malloc(strlen(regname[i])+1); /* allocate */
      strcpy (out->regname[i], regname[i]);
    }

    if (detector!=NULL) {       /* detector number */
      out->detector[i] = detector[i];
    } else { /* default */
      out->detector[i]= 1;
    }

    if (ports!=NULL) {       /* ports */
      for (j=0; j<maxtel; j++) out->ports[i][j] = ports[i][j];
    } else { /* default */
      for (j=0; j<maxtel; j++) out->ports[i][j]= j+1;
    }

    out->correlation[i] = correlation[i]; /* correlation type */

    /* corner */
    if (corner!=NULL) {
      for (j=0; j<2; j++) out->corner[i][j] = corner[i][j];
    } else {  /* default */
      for (j=0; j<2; j++) out->corner[i][j] = -1;
    } 

    /* gain */
    if (gain!=NULL) {
      out->gain[i] = gain[i];
    } else {  /* default */
      out->gain[i] = 1.0;
    }

    /* size - required */
    for (j=0; j<2; j++) out->naxes[i][j] = naxes[i][j];

    if (crval!=NULL) {       /* crval */
       for (j=0; j<2; j++) out->crval[i][j] = crval[i][j];
    } else { /* default */
      for (j=0; j<1; j++) out->crval[i][j] = 0.;
    }

    if (crpix!=NULL) {       /* crpix */
      for (j=0; j<2; j++) out->crpix[i][j] = crpix[i][j];
    } else { /* default */
      for (j=0; j<2; j++) out->crpix[i][j]= 1;
    }

    if (ctype!=NULL) {       /* ctype */
      for (j=0; j<2; j++) strcpy(out->ctype[i][j], ctype[i][j]);
    } else { /* default */
      strcpy(out->ctype[i][0],"X");
      strcpy(out->ctype[i][1],"Y");
    }

    if (cd!=NULL) {       /* cd */
      for (j=0; j<4; j++) out->cd[i][j] = cd[i][j];
    } else { /* default */
      out->cd[i][0]= 1.0;
      out->cd[i][1]= 0.0;
      out->cd[i][2]= 0.0;
      out->cd[i][3]= 1.0;
    }

    /* loop over dimensions */
    for (l=0; l<max_coef; l++) {
      if (dmc!=NULL) {       /*dmc */
	for (j=0; j<num_dim; j++) {
	  out->dmc[i][l][j] = dmc[i][l][j]; 
	  for (k=0; k<num_dim; k++) {
	    out->dmp[i][l][j][k] = dmp[i][l][j][k];
	  }
	}
      } else { /* default */
	for (j=0; j<num_dim; j++) {
	  out->dmc[i][l][j] = 0.0;
	  for (k=0; k<num_dim; k++) {
	    out->dmp[i][l][j][k] = 0;
	  }
	} 
      }
    } /* end loop over dimensions */

  } /* end of loop copying data */
  return out; /* done */
} /* end make_imaging_detector_table_def */

imaging_detector_table_def* 
make_testing_detector_table_def(int nx, int ny, int nreg)
/*----------------------------------------------------------------------- */
/* create a, mostly dummy imaging detector table suitable for algorithm   */
/*  testing;  makes one cross correlation array nx x ny pixels            */
/*   Inputs:                                                              */
/*      nx     number of pixel in "x" direction                           */
/*      ny     number of pixel in "y" direction                           */
/*      nreg      Number of regions                                       */
/*   Returns:                                                             */
/*      An imaging_detector_table_def pointer; NULL indicates failure     */
/*----------------------------------------------------------------------- */
{
  imaging_detector_table_def* out;
  int ndetect=1, maxtel=2, num_dim=2, max_coef=1, i;
  char *instrument="test_data";
  char *date_obs="unknown";
  int *region, *detector, **telescope, *correlation, **naxes;
  int **corner=NULL;
  double *gain=NULL, **crval=NULL, **cd=NULL, ***dmc=NULL;
  int    ****dmp=NULL;
  float  **crpix=NULL;
  char   **regname=NULL, ***ctype=NULL;

  if (nreg<=0) return NULL;  /* sanity check */

  /* make arrays */
  region      = (int*) malloc(nreg*sizeof(int));
  detector    = (int*) malloc(nreg*sizeof(int));
  telescope   = (int**) malloc(nreg*sizeof(int*));
  correlation = (int*) malloc(nreg*sizeof(int));
  correlation = (int*) malloc(nreg*sizeof(int));
  naxes       = (int**) malloc(nreg*sizeof(int*));
  gain        = (double*) malloc(nreg*sizeof(double));
  for (i=0; i<nreg; i++) {
    telescope[i] = (int*)malloc(2*sizeof(int));
    naxes[i] = (int*)malloc(2*sizeof(int));

    /* set values */
    region[i] = i+1;
    detector[i] = 1;
    telescope[i][0] = 1;
    telescope[i][1] = 2;
    correlation[i] = 2;
    gain[i] = 1.0;
    naxes[i][0] = nx;
    naxes[i][1] = ny;
  }

  /* create it */
  out = 
    make_imaging_detector_table_def(nreg, ndetect, maxtel, 
				    num_dim, max_coef,
				    instrument, date_obs,
				    region, regname, detector, telescope, 
				    correlation, corner, gain, 
				    naxes, crval, crpix, 
				    ctype, cd, dmc, dmp);

  /* clean up */
  for (i=0; i<nreg; i++) {
    if (telescope[i]!=NULL) free(telescope[i]);
    if (naxes[i]!=NULL) free(naxes[i]);
  }
  if (region!=NULL) free(region);
  if (detector!=NULL) free(detector);
  if (telescope!=NULL) free(telescope);
  if (correlation!=NULL) free(correlation);
  if (gain!=NULL) free(gain);
  if (naxes!=NULL) free(naxes);

  return out;
} /* end make_testing_detector_table_def */


imaging_detector_table_def* 
read_imaging_detector_table(fitsfile *fptr, int *ierr)
/*----------------------------------------------------------------------- */
/*   Read an (entire) imaging_detector_table from an active FITS file     */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to active FITS file                     */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An imaging_detector_table_def pointer; NULL indicates failure     */
/*----------------------------------------------------------------------- */
{
  imaging_detector_table_def *out;
  table_def *tabledef;
  table_row *row;
  int i, j, k, l, indx;
  int nregion, ndetect, maxtel, num_dim, max_coef, ikey, ireg;
  char *instrument="unknown";
  char *date_obs="unknown";

  if (*ierr>0) return NULL; /* check for existing error condition */
  if (fptr==NULL) return NULL; /* sanity check */

  /* position cfitsio to IMAGING_DETECTOR table */
  if (fits_movnam_hdu(fptr, BINARY_TBL, tabletype, 0, ierr)) {
    fprintf(stderr,"ERROR %d finding %s table \n", *ierr, tabletype);
    return NULL;
  }

  /* get table descriptor from file */
  tabledef = make_table_def_from_table (fptr);
  if (tabledef==NULL) { /* failed? */
    *ierr = 10000;
    return NULL;
  }

  /* how big is this sucker */
  nregion = 1;  /* NREGION */
  ikey = table_find_key(tabledef, "NREGION");
  if (ikey>0) table_get_keyword(tabledef, ikey, (void*)&nregion, ierr);
  ndetect = 1;  /* NDETECT */
  ikey = table_find_key(tabledef, "NDETECT");
  if (ikey>0) table_get_keyword(tabledef, ikey, (void*)&ndetect, ierr);
  maxtel = 1;  /* MAXTEL */
  ikey = table_find_key(tabledef, "MAXTEL");
  if (ikey>0) table_get_keyword(tabledef, ikey, (void*)&maxtel, ierr);
  ikey = table_find_key(tabledef, "INSTRUME");
  if (ikey>0) instrument = (char*)tabledef->keydata[ikey-1];
  ikey = table_find_key(tabledef, "DATE_OBS");
  if (ikey>0) date_obs = (char*)tabledef->keydata[ikey-1];
  max_coef = 0;        /* size of dedistortion arrays */
  ikey = table_find_key(tabledef, "MAX_COEF");
  if (ikey>0) table_get_keyword(tabledef, ikey, (void*)&max_coef, ierr);
  num_dim = 0;        /* size of arrays */
  ikey = table_find_key(tabledef, "NUM_DIM");
  if (ikey>0) table_get_keyword(tabledef, ikey, (void*)&num_dim, ierr);
  if (*ierr>0) {  /* failure? */
    kill_table_def (tabledef); /* clean up */
    return NULL;
  }

  /* build basic structure */
  out = create_imaging_detector_table_def(nregion, ndetect, maxtel, 
					  num_dim, max_coef, 
					  instrument, date_obs);
  if (out==NULL) {  /* allocation failure? */
    *ierr=10000;
    kill_table_def (tabledef); /* clean up */
    return out;
  }
  
  /* get table_row to read file */
  row =  make_table_row_from_def (tabledef);
  if (row==NULL) {  /* allocation failure? */
    *ierr=10000;
    kill_table_def (tabledef); /* clean up */
    kill_imaging_detector_table_def (out); /* clean up */
    return NULL;
  }

  /* find column numbers */
  out->def = tabledef;
  find_imaging_detector_cols(out);

  /* loop over table */
  for (i=0; i<nregion; i++) {
    row->trow = i+1; /* set row number */
    table_read_row(fptr, tabledef, row, ierr);  /* read row */
    if (*ierr>0) {
      fprintf(stderr,"ERROR %d reading %s table \n", *ierr, tabletype);
      kill_table_def (tabledef); /* clean up */
      kill_imaging_detector_table_def (out); /* clean up */
      kill_table_row (row);
      return NULL;
    }

    /* save in arrays as region -1 */
    ireg = row->data[out->region_col]->tdata.tshort[0]-1;
    out->region[ireg] = ireg+1;
    out->detector[ireg] = row->data[out->detector_col]->tdata.tshort[0];
    for (j=0; j<maxtel; j++)
      out->ports[ireg][j] = row->data[out->ports_col]->tdata.tshort[j];
    out->correlation[ireg] = row->data[out->correlation_col]->tdata.tshort[0];
    /* region - take care of memory allocation */
    if (out->regname[i]!=NULL) free (out->regname[i]); /* free old memory */
    out->regname[i] = 
      (char*)malloc(strlen(row->data[out->regname_col]->tdata.tchar)+1);
    strcpy (out->regname[i], row->data[out->regname_col]->tdata.tchar);
    for (j=0; j<2; j++)
      out->corner[ireg][j] = row->data[out->corner_col]->tdata.tshort[j];
    out->gain[ireg] = row->data[out->gain_col]->tdata.tdouble[0];
    for (j=0; j<2; j++) {
      out->naxes[ireg][j] = row->data[out->naxes_col]->tdata.tshort[j];
      out->crval[ireg][j] = row->data[out->crval_col]->tdata.tdouble[j];
      out->crpix[ireg][j] = row->data[out->crpix_col]->tdata.tfloat[j];
      /* Strings kinda funky, we know this is 8x2 */
      padcopy (out->ctype[ireg][j], &row->data[out->ctype_col]->tdata.tchar[j*8], 8);
    }
    for (j=0; j<4; j++)
      out->cd[ireg][j] = row->data[out->cd_col]->tdata.tdouble[j];

    /* loops over dedistortion polynomials */
    indx = 0;
    for (j=0; j<max_coef; j++) {
      for (l=0; l<num_dim; l++) {
	out->dmc[ireg][j][l] = 
	  row->data[out->dmc_col]->tdata.tdouble[indx++];
      }
    }
    indx = 0;
    for (j=0; j<max_coef; j++) {
      for (k=0; k<num_dim; k++) {
	for (l=0; l<num_dim; l++) {
	  out->dmp[ireg][j][k][l] = 
	    row->data[out->dmp_col]->tdata.tshort[indx++];
	}
      }
    }
  } /* end of loop reading table */

  /* clean up */
  /* will be done later kill_table_def (tabledef);*/
  kill_table_row (row);

  return out; /* done */
} /* end  read_imaging_detector_table */


void write_imaging_detector_table(fitsfile *fptr, 
				  imaging_detector_table_def* table, 
				  int *ierr)
/*----------------------------------------------------------------------- */
/*   Write an (entire) imaging_detector_table to an active FITS file      */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to active FITS file                     */
/*      table     imaging_detector_table structure to write               */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  int  version = 1; /* table version number */
  table_def *tabledef = NULL;
  table_row *row = NULL;
  int i, j, k, l, indx;
  int nregion, ndetect, maxtel, num_dim, max_coef, ikey, icol, ireg;
  char err_msg[FLEN_ERRMSG], tstring[50];
  double others[2];

  if (*ierr>0) return;     /* check for existing error condition */
  if (fptr==NULL) return;  /* sanity check */
  if (table==NULL) return; /* sanity check */


  /* how big is this sucker */
  nregion = table->nregion;   /* NREGION */
  ndetect = table->ndetect;   /* NDETECT */
  maxtel  = table->maxtel;    /* MAXTEL */
  num_dim   = table->num_dim; /* number of dimensions in data */
  max_coef  = table->max_coef;/* number of coeffieicnts */
    
  /* make table_def if one doesn't exist */
  if (table->def==NULL) {
    /* create table_def */
    tabledef = make_table_def(IMAGING_DETECTOR_TABLE, 'n', table->num_dim);
    if (tabledef==NULL) { /* failed? */
      *ierr = 10000;
      return;
    }
    table->def = tabledef; /* attach structure */
    
    /* fill in table_def - first keywords */
    ikey = table_find_key(tabledef, "REVISION");
    if (ikey>0) table_put_keyword (tabledef, ikey, (void*)&version, ierr);
    
    ikey = table_find_key(tabledef, "NREGION");
    if (ikey>0) table_put_keyword(tabledef, ikey, (void*)&nregion, ierr);
    
    ikey = table_find_key(tabledef, "NDETECT");
    if (ikey>0) table_put_keyword(tabledef, ikey, (void*)&ndetect, ierr);
    
    ikey = table_find_key(tabledef, "MAXTEL");
    if (ikey>0) table_put_keyword(tabledef, ikey, (void*)&maxtel, ierr);
    
    ikey = table_find_key(tabledef, "INSTRUME");
    if (ikey>0) table_put_keyword(tabledef, ikey, (void*)table->instrument, ierr);
    
    ikey = table_find_key(tabledef, "DATE_OBS");
    if (ikey>0) table_put_keyword(tabledef, ikey, (void*)table->date_obs, ierr);
    
    ikey = table_find_key(tabledef, "MAX_COEF");
    if (ikey>0) table_put_keyword(tabledef, ikey, (void*)&table->max_coef, ierr);
    
    ikey = table_find_key(tabledef, "NUM_DIM");
    if (ikey>0) table_put_keyword(tabledef, ikey, (void*)&table->num_dim, ierr);
    
    /* funky "others" keyword that really shouldn't be there */
    others[0]= 0.5; others[1]= 3.1415 ;
    ikey = table_find_key(tabledef, "others");
    if (ikey>0) table_put_keyword(tabledef, ikey, (void*)others, ierr);

    if (*ierr>0) {  /* failure? */
      return;
    }

    /* find column numbers */
    find_imaging_detector_cols(table);
    
    /* set size of ports member */
    if (table->ports_col>=0) 
      tabledef->colcount[table->ports_col] = maxtel;
    
    
    /* size of polynomials */
    if (table->dmc_col>=0) 
      tabledef->colcount[table->dmc_col] = max_coef*num_dim;
    if (table->dmp_col>=0) 
      tabledef->colcount[table->dmp_col] = max_coef*num_dim*num_dim;
    /* dimensionality string for DMC */
    sprintf (tstring,"DMC");
    icol = table_find_col(tabledef, tstring)-1;
    if (tabledef->coldim[icol]) 
      free (tabledef->coldim[icol]);   /* free if it exists */
    tabledef->coldim[icol] = 
      (char*)malloc(FLEN_VALUE+1);     /* allocate */
    /* set dimensionality */
      sprintf (tabledef->coldim[icol],"(%d,%d)",num_dim,max_coef); 
    /* dimensionality string for DMP */
    sprintf (tstring,"DMP");
    icol = table_find_col(tabledef, tstring)-1;
    if (tabledef->coldim[icol]) 
      free (tabledef->coldim[icol]);   /* free if it exists */
    tabledef->coldim[icol] = 
      (char*)malloc(FLEN_VALUE+1);     /* allocate */
    /* set dimensionality */
      sprintf (tabledef->coldim[icol],"(%d,%d,%d)",
	       num_dim,num_dim,max_coef); 
  } /* end build new table def */

  /* create IMAGING_DETECTOR table */
  make_table_from_def (fptr, table->def, ierr);
  if (*ierr>0) {
    fprintf(stderr,"ERROR %d creating %s table \n", *ierr, tabletype);
    return;
  }
 
  /* get table_row to write file */
  row =  make_table_row_from_def (table->def);

  /* loop over table */
  for (i=0; i<nregion; i++) {
    /* save in arrays as region -1 */
    ireg = i+1;   /* region number */
    row->data[table->region_col]->tdata.tshort[0] = table->region[i];
    row->data[table->detector_col]->tdata.tshort[0] = table->detector[i];
    for (j=0; j<maxtel; j++)
      row->data[table->ports_col]->tdata.tshort[j] = table->ports[i][j];
    row->data[table->correlation_col]->tdata.tshort[0] = table->correlation[i];
    strcpy (row->data[table->regname_col]->tdata.tchar, table->regname[i]);
    for (j=0; j<2; j++)
           row->data[table->corner_col]->tdata.tshort[j] = table->corner[i][j];
    row->data[table->gain_col]->tdata.tdouble[0] = table->gain[i];
    for (j=0; j<2; j++) {
      row->data[table->naxes_col]->tdata.tshort[j] = table->naxes[i][j];
      row->data[table->crval_col]->tdata.tdouble[j] = table->crval[i][j];
      row->data[table->crpix_col]->tdata.tfloat[j] = table->crpix[i][j];
      /* Strings kinda funky, we know this is 8x2 */
      padcopy (&row->data[table->ctype_col]->tdata.tchar[j*8], 
	       table->ctype[i][j], 8);
    }
    for (j=0; j<4; j++)
      row->data[table->cd_col]->tdata.tdouble[j] = table->cd[i][j];
    
    /* loop over dedistortion polynomials */
    indx = 0;
    for (j=0; j<max_coef; j++) {
      for (l=0; l<num_dim; l++) {
	row->data[table->dmc_col]->tdata.tdouble[indx++] = 
	  table->dmc[i][j][l];
      }
    }
    indx = 0;
    for (j=0; j<max_coef; j++) {
      for (k=0; k<num_dim; k++) {
	for (l=0; l<num_dim; l++) {
	  row->data[table->dmp_col]->tdata.tshort[indx++] = 
	    table->dmp[i][j][k][l];
	}
      }
    }
    
    /* write row */
    row->trow = i+1; /* set row number */
    table_write_row(fptr, table->def, row, ierr);
    if (*ierr>0) { /* error test */
      fprintf(stderr,"ERROR %d writing %s table \n", *ierr, tabletype);
      while (fits_read_errmsg (err_msg)) {
	fprintf (stderr,"%s\n",err_msg);
      }
      kill_table_row (row);
      return;
    }
  } /* end of loop writing table */

  /* clean up */
  kill_table_row (row);
  kill_table_def (tabledef);

} /* end write_imaging_detector_table */


void print_imaging_detector_table(imaging_detector_table_def* table)
/*----------------------------------------------------------------------- */
/*   Print an imaging_detector_table to stdout                            */
/*   Inputs:                                                              */
/*      table  pointer to structure to be printed                         */
/*----------------------------------------------------------------------- */
{
  int ireg, i, j, k, l;
  char *cortype[] = {
    "Background",
    "Autocorrelation",
    "Crosscorrelation"
  };

  fprintf (stdout, "\n\n Imaging Detector Table\n");
  fprintf (stdout, " Number of regions %d\n",table->nregion);
  fprintf (stdout, " Number of detectors %d\n",table->ndetect);
  fprintf (stdout, " Maximum number of telescope per region %d\n",
	   table->maxtel);
  fprintf (stdout, " Instrument %s\n",table->instrument);
  fprintf (stdout, " Observing date %s\n",table->date_obs);
  fprintf (stdout, " Max number of polynomial terms %d\n",
	   table->max_coef);
  fprintf (stdout, " Number of data dimensions %d\n",
	   table->num_dim);
  
  for (ireg=0; ireg<table->nregion; ireg++) { /* loop over regions */
    fprintf (stdout, "\n Description of region %d\n",table->region[ireg]); 
    fprintf (stdout, " Detector is %d\n",table->detector[ireg]); 
    fprintf (stdout, " Telescopes involved are "); 
    for (i=0; i<table->maxtel; i++)  fprintf (stdout, " %d", 
					      table->ports[ireg][i]);
    fprintf (stdout, "\n"); 
    fprintf (stdout, " Region name is %s\n",table->regname[ireg]);
    fprintf (stdout, " Correlation type is %s\n",
	     cortype[table->correlation[ireg]]);
    fprintf (stdout, " Detector ADU/electron is %f\n",
	     table->gain[ireg]);
    fprintf (stdout, " Lower corner is (%d,%d)\n",
	     table->corner[ireg][0],table->corner[ireg][1]);
    /* loop over two dimensions */
    for (j=0; j<2; j++) {
      fprintf (stdout, " Dimension of axis %d is %d\n",
	       j+1,table->naxes[ireg][j]);
      fprintf (stdout, " Type of axis %d is %s\n",
	       j+1,table->ctype[ireg][j]);
      fprintf (stdout, " Reference value of axis %d is %f\n",
	       j+1,table->crval[ireg][j]);
      fprintf (stdout, " Reference pixel of axis %d is %f\n",
	       j+1,table->crpix[ireg][j]);
    } /* end of loop over dimension */
    fprintf (stdout, " CD matrix elements are %f %f %f %f\n",
	     table->cd[ireg][0],table->cd[ireg][1],
	     table->cd[ireg][2],table->cd[ireg][3]);
    if (table->max_coef>0) { /* dedistortion polynomial */
      fprintf (stdout, " Dedistortion polynomial coefficients are \n"); 
      for (l=0; l<table->num_dim; l++) { /* loop over output dimensions */
	fprintf (stdout, "   Output Dimension %d \n", l+1);
	for (i=0; i<table->max_coef; i++)  {
	  for (k=0; k<table->num_dim; k++) /* powers */
	    fprintf (stdout, " %d ", table->dmp[ireg][i][l][k]);
	  fprintf (stdout, " %f\n", table->dmc[ireg][i][l]); /* coeff */
	}
      }
    } /* end loop over dedistortion polynomial */
  } /* end loop over regions */
      fprintf (stdout, "\n\n"); 

} /* end  print_imaging_detector_table*/

void kill_imaging_detector_table_def (imaging_detector_table_def* table)
/*----------------------------------------------------------------------- */
/*   Destroy an imaging_detector_table_def                                */
/*   Inputs:                                                              */
/*      table  pointer to structure to be deleted                         */
/*----------------------------------------------------------------------- */
{
  int i, j, k;

  if (table==NULL) return; /* sanity check */

  /* keyword arrays freed as part of table->def */

  /* free inner arrays */
  for (i=0; i<table->nregion; i++) {
    if ((table->regname!=NULL)&&(table->regname[i]!=NULL))  free (table->regname[i]);
    if ((table->ports!=NULL)&&(table->ports[i]!=NULL))      free (table->ports[i]);
    if ((table->corner!=NULL)&&(table->corner[i]!=NULL))    free (table->corner[i]);
    if ((table->crval!=NULL)&&(table->crval[i]!=NULL))      free (table->crval[i]);
    if ((table->crpix!=NULL)&&(table->crpix[i]!=NULL))      free (table->crpix[i]);
    if ((table->ctype!=NULL)&&(table->ctype[i][0]!=NULL))   free (table->ctype[i][0]);
    if ((table->ctype!=NULL)&&(table->ctype[i][1]!=NULL))   free (table->ctype[i][1]);
    if ((table->ctype!=NULL)&&(table->ctype[i]!=NULL))      free (table->ctype[i]);
    if ((table->cd!=NULL)&&(table->cd[i]!=NULL))            free (table->cd[i]);
    for (j=0; j<table->max_coef; j++) {
      for (k=0; k<table->num_dim; k++) {
	  if ((table->dmp[i]!=NULL)&&(table->dmp[i][j]!=NULL)&&(table->dmp[i][j][k]!=NULL))
	    free (table->dmp[i][j][k]);
      }
      if ((table->dmc[i]!=NULL)&&(table->dmc[i][j]!=NULL)) free (table->dmc[i][j]);
      if ((table->dmc[i]!=NULL)&&(table->dmp[i][j]!=NULL)) free (table->dmp[i][j]);
    }
    if (table->dmc[i]!=NULL) free (table->dmc[i]);
    if (table->dmp[i]!=NULL) free (table->dmp[i]);
  } /* end of loop destroying inner arrays */
  
  /* free outer arrays */
  if (table->region!=NULL)      free (table->region);
  if (table->regname!=NULL)     free (table->regname);
  if (table->detector!=NULL)    free (table->detector);
  if (table->ports!=NULL)       free (table->ports);
  if (table->correlation!=NULL) free (table->correlation);
  if (table->corner!=NULL)      free (table->corner);
  if (table->gain!=NULL)        free (table->gain);
  if (table->crval!=NULL)       free (table->crval);
  if (table->crpix!=NULL)       free (table->crpix);
  if (table->ctype!=NULL)       free (table->ctype);
  if (table->cd!=NULL)          free (table->cd);
  if (table->dmp!=NULL)         free (table->dmp);
  if (table->dmc!=NULL)         free (table->dmc);

  kill_table_def(table->def); /* basic table def */

  free (table); /* main structure */
} /* end kill_imaging_detector_table_def */

/********************* Private routines ********************************/
imaging_detector_table_def* 
create_imaging_detector_table_def(int nregion, int ndetect, int maxtel, 
				  int num_dim,  int max_coef, 
				  char *instrument, char *date_obs)
/*----------------------------------------------------------------------- */
/*   Create basic structure for imaging_detector_table_def                */
/*   Inputs:                                                              */
/*      nregion  the number of regions (rows) in the table                */
/*      ndetect  the number of detectors used                             */
/*      maxtel   the max number of ports contributing to a region         */
/*      num_dim  number of dimensions in data array                       */
/*      max_coef maximum number of entries in dmp, dmc arrays             */
/*      instrument the name of the instrument                             */
/*      date_obs Date (yyyy-mm-dd) observations began                     */
/*   Returns:                                                             */
/*      An imaging_detector_table_def pointer; NULL indicates failure     */
/*----------------------------------------------------------------------- */
{
  imaging_detector_table_def *out;
  int                         i, j, k;

  /* create output object */
  out = (imaging_detector_table_def*)malloc(sizeof(imaging_detector_table_def));
  if (out==NULL)     {  /* check for allocation failure */
    fprintf(stderr,"ERROR creating %s table_def \n", tabletype);
    return out;
  }
  memset (out, 0, sizeof(imaging_detector_table_def)); /* zero fill */

  /* column index arrays */

  /* fill in global values */
  out->nregion = nregion;
  out->maxtel = maxtel;
  out->ndetect = ndetect;
  out->num_dim = num_dim;
  out->max_coef = max_coef;
  out->instrument = (char*)malloc(strlen(instrument)+1);
  strcpy (out->instrument, instrument);
  out->date_obs = (char*)malloc(strlen(date_obs)+1);
  strcpy (out->date_obs, date_obs);

  /* create arrays */
  out->region      = (int*)malloc(nregion*sizeof(int));
  out->regname     = (char**)malloc(nregion*sizeof(char*));
  out->detector    = (int*)malloc(nregion*sizeof(int));
  out->ports       = (int**)malloc(nregion*sizeof(int*));
  out->correlation = (int*)malloc(nregion*sizeof(int));
  out->corner      = (int**)malloc(nregion*sizeof(int*));
  out->gain        = (double*)malloc(nregion*sizeof(double));
  out->naxes       = (int**)malloc(nregion*sizeof(int*));
  out->crval       = (double**)malloc(nregion*sizeof(double*));
  out->crpix       = (float**)malloc(nregion*sizeof(float*));
  out->ctype       = (char***)malloc(nregion*sizeof(char**));
  out->cd          = (double**)malloc(nregion*sizeof(double*));
  out->dmc         = (double***)malloc(nregion*sizeof(double*));
  out->dmp         = (int****)malloc(nregion*sizeof(int**));

  /* some of these are arrays... */
  for (i=0; i<nregion; i++) {
    out->regname[i] = (char*)malloc(sizeof(char)); /* dummy */
    out->regname[i][0] = 0;
    out->ports[i]  = (int*)malloc(maxtel*sizeof(int));
    out->corner[i] = (int*)malloc(4*sizeof(int));
    out->naxes[i] = (int*)malloc(2*sizeof(int));
    out->crval[i] = (double*)malloc(2*sizeof(double));
    out->crpix[i] = (float*)malloc(2*sizeof(float));
    out->ctype[i] = (char**)malloc(2*sizeof(double));
    out->ctype[i][0] = (char*)malloc(9*sizeof(char)); /* max WCS string size */
    out->ctype[i][1] = (char*)malloc(9*sizeof(char)); /* max WCS string size */
    out->cd[i] = (double*)malloc(4*sizeof(double));
    out->dmc[i] = (double**)malloc(max_coef*sizeof(double*));
    out->dmp[i] = (int***)malloc(max_coef*sizeof(int**));
    /* yet more levels of array */
    for (j=0; j<max_coef; j++) {
      out->dmc[i][j] = (double*)malloc(num_dim*sizeof(double));
      out->dmp[i][j] = (int**)malloc(num_dim*sizeof(int*));
      for (k=0; k<num_dim; k++) {
	out->dmp[i][j][k] = (int*)malloc(num_dim*sizeof(int));
      }
    }
  } /* end of loop creating inner arrays */

  return out; /* done */
} /* end create_imaging_detector_table_def */


void find_imaging_detector_cols(imaging_detector_table_def *table)
/*----------------------------------------------------------------------- */
/*   Find table columns, results saved in file global variables           */
/*   values returned as array indices, i.e. 0-rel                         */
/*   Inputs:                                                              */
/*      table table_def structure to search                               */
/*----------------------------------------------------------------------- */
{
  table->region_col      = table_find_col(table->def, "REGION") - 1;
  table->detector_col    = table_find_col(table->def, "DETECTOR") - 1;
  table->ports_col       = table_find_col(table->def, "PORTS") - 1;
  table->correlation_col = table_find_col(table->def, "CORRELATION") - 1;
  table->regname_col     = table_find_col(table->def, "REGNAME") - 1;
  table->corner_col      = table_find_col(table->def, "CORNER") - 1;
  table->gain_col        = table_find_col(table->def, "GAIN") - 1;
  table->naxes_col       = table_find_col(table->def, "NAXES") - 1;
  table->crval_col       = table_find_col(table->def, "CRVAL") - 1;
  table->crpix_col       = table_find_col(table->def, "CRPIX") - 1;
  table->ctype_col       = table_find_col(table->def, "CTYPE") - 1;
  table->cd_col          = table_find_col(table->def, "CD") - 1;
  table->dmc_col         = table_find_col(table->def, "DMC") - 1;
  table->dmp_col         = table_find_col(table->def, "DMP") - 1;
} /* end find_imaging_detector_cols */
