/* $Id: table_gen.c,v 1.2 2001/11/26 12:49:28 csabet Exp $ */
/*----------------------------------------------------------------------- */
/* general FITS binary tables routines for the VLTI interferometer        */
/*----------------------------------------------------------------------- */
/*#include <malloc.h>*/
#include <string.h>
#include "VLTItable.h"

void make_table (fitsfile *fptr, long naxis2, int tfields, 
		      char* ttype[], char* tform[], char* tunit[], 
		      char* extname, 
		      int nkey, int keytype[], char* keyword[], 
		      void* keyvalue[], char* keycomment[], 
		      int *ierr)
/*----------------------------------------------------------------------- */
/*   Function to create and initialize a FITS binary table                */
/*   Inputs:                                                              */
/*      fptr       Pointer to open FITSIO file structure                  */
/*      naxis2     Minimum number of rows, generally use 0                */
/*      tfields    Number of columns in table                             */
/*      ttype      Column labels (tfield values).                         */
/*      tform      Column formats                                         */
/*      tunit      Column units                                           */
/*      extname    Name of table type                                     */
/*      nkey       Number of additional keywords in header                */
/*      keytype    Keyword data types as defined in fitsio.h              */
/*      keyword    Keywords.                                              */
/*      keyvalue   Pointers to values cast to void                        */
/*      keycomment Comment explaining keyword                             */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, returns if initially positive      */
/*----------------------------------------------------------------------- */
{
  int i;

  /* Are we already in an error state? */
  if (*ierr>0) return;
  
  /* Open/create output table */
  fits_create_tbl(fptr, BINARY_TBL, 
		  naxis2, tfields, ttype, tform, tunit, extname, 
		  ierr);

  /* add date created */
  fits_write_date(fptr, ierr);

  /* add keywords */
  for (i=0; i<nkey; i++) {
    fits_write_key(fptr, keytype[i], keyword[i], keyvalue[i], 
		   keycomment[i],
		   ierr);
  }
} /* end make_table */

void table_print_error_stack ()
/*----------------------------------------------------------------------- */
/*   print cfitsio error stack to stderr                                  */
/*----------------------------------------------------------------------- */
{
  char err_msg[FLEN_ERRMSG];
  while (fits_read_errmsg (err_msg)) {
    fprintf (stderr,"%s\n",err_msg);
  }
} /* end  table_print_error_stack*/

char FITS_type_to_code(int type)
/*----------------------------------------------------------------------- */
/*    Convert integer type to character format code                       */
/*   Input:                                                               */
/*      Data type code as defined in fitsio.h                             */
/*   Returns:                                                             */
/*      The TFORMn character code for the type                           .*/
/*----------------------------------------------------------------------- */
{
  char out;
  /* find datatype code */
  switch (type) {
  case TSHORT:
    out = 'I';
    break;
  case TINT:
    out = 'I';
    break;
  case TLONG:
    out = 'J';
    break;
  case TSTRING:
    out = 'A';
    break;
  case TFLOAT:
    out = 'E';
    break;
  case TDOUBLE:
    out = 'D';
    break;
  case TCOMPLEX:
    out = 'C';
    break;
  case TDBLCOMPLEX:
    out = 'M';
    break;
  case TLOGICAL:
    out = 'L';
    break;
  case TBIT:
    out = 'X';
    break;
  case TBYTE:
    out = 'B';
    break;
  default: /* unknown */
    out = '?';
  } /* end switch finding datatype */
  return out;
} /* end FITS_type_to_code */

int FITS_code_to_type(char code)
/*----------------------------------------------------------------------- */
/*    Convert character format code to integer type                       */
/*   Input:                                                               */
/*      Data type code as defined in fitsio.h.                            */
/*   Returns:                                                             */
/*      The TFORMn character code for the type.                           */
/*----------------------------------------------------------------------- */
{
  int out;
  /* find datatype code */
  switch (code) {
  case 'I':
    out = TSHORT;
    break;
  case 'J':
    out = TLONG;
    break;
  case  'A':
    out = TSTRING;
    break;
  case 'E':
    out = TFLOAT;
    break;
  case 'D':
    out = TDOUBLE;
    break;
  case 'C':
    out = TCOMPLEX;
    break;
  case 'M':
    out = TDBLCOMPLEX;
    break;
  case 'L':
    out = TLOGICAL;
    break;
  case 'X':
    out = TBIT;
    break;
  case 'B':
    out = TBYTE;
    break;
  default: /* unknown */
    out = -999;
  } /* end switch finding datatype */
  return out;
} /* end FITS_code_to_type */

/* Get size of FITS data types */
int FITS_type_to_size(int type)
/*----------------------------------------------------------------------- */
/*   Get size of FITS data types                                          */
/*   Input:                                                               */
/*      Data type code as defined in fitsio.h.                            */
/*   Returns:                                                             */
/*      The size in bytes.                                                */
/*----------------------------------------------------------------------- */
{
  int out;
  /* find datatype code */
  switch (type) {
  case TSHORT:
    out = sizeof(short);
    break;
  case TINT:
    out = sizeof(int);
    break;
  case TLONG:
    out = sizeof(long);
    break;
  case TSTRING:
    out = sizeof(char);
    break;
  case TFLOAT:
    out = sizeof(float);
    break;
  case TDOUBLE:
    out = sizeof(double);
    break;
  case TCOMPLEX:
    out = 2*sizeof(float);
    break;
  case TDBLCOMPLEX:
    out = 2*sizeof(double);
    break;
  case TLOGICAL:
    out = sizeof(char);
    break;
  case TBIT:
    out = 1;
    break;
  case TBYTE:
    out = sizeof(char);
    break;
  default: /* unknown */
    out = 0;
  } /* end switch finding datatype */
  return out;
} /* end FITS_type_to_ */

int table_find_col (table_def *tabledef, char* name)
/*----------------------------------------------------------------------- */
/*   Find the column number of a given label                              */
/*   Inputs:                                                              */
/*      tabledef     Pointer to the table defintion structure             */
/*      name         Desired column label                                 */
/*   Returns:                                                             */
/*      1 - rel column number, -1 => failed                               */
/*----------------------------------------------------------------------- */
{
  int i;

  if (tabledef->keywords==NULL) return -1; /* trap bad pointer */
  for (i=0; i<tabledef->ncol; i++) {
    if (!strcmp(name, tabledef->coltype[i])) return i+1;
  } /* end loop looking for column label */
  return -1; /* if you got here it wasn't found */
} /* end table_find_col */

int table_find_key (table_def *tabledef, char* name)
/*----------------------------------------------------------------------- */
/*   Find the keyword number of a given label                             */
/*   Inputs:                                                              */
/*      tabledef     Pointer to the table defintion structure             */
/*      name         Desired column label                                 */
/*   Returns:                                                             */
/*      1 - rel column number, -1 => failed                               */
/*----------------------------------------------------------------------- */
{
  int i;

  if (tabledef->keywords==NULL) return -1; /* trap bad pointer */
  for (i=0; i<tabledef->nkey; i++) {
    if (!strcmp(name, tabledef->keywords[i])) return i+1;
  } /* end loop looking for keyword */
  return -1; /* if you got here it wasn't found */
} /* end table_find_key */

