/* $Id: batch.c,v 1.2 2001/11/26 12:52:49 csabet Exp $ */
/*----------------------------------------------------------------------- */
/* Fringe amplitude data FITS binary tables routines                      */
/* Partially/fully calibrated data table for optical/IR interferometer    */
/*----------------------------------------------------------------------- */
/*#include <malloc.h>*/			/* csabet 26.Jun.01 */
#include <string.h> 		/* csabet 26.Jun.01 */
#include "batch.h"
/* #include "vndrs_OPD.h"  */		/* csabet 26.Jun.01 */

#define MAGICBLANK (-999.99)   /* Magic blanking value for bad values */
/* internal prototypes */
/* get data from table_def specific to batch table */
void get_batch_info (batch_table_def *table);

/* file global variables for table type */
static char *tabletype = "BATCH";

/* batch_row functions */
batch_row* make_batch_row(batch_table_def* table)
/*----------------------------------------------------------------------- */
/*   Make an batch_row structure                                          */
/*   Inputs:                                                              */
/*      table     pointer to batch table structure                        */
/*   Returns:                                                             */
/*      An batch_row structure pointer ; NULL indicates failure           */
/*      Data pointers set to generic structure values                     */
/*----------------------------------------------------------------------- */
{
  batch_row *out;
  int i;

  /* create basic structure */
  out = (batch_row*)malloc(sizeof(batch_row));
  if (out==NULL) return out; /* allocation failure? */
  
  /* zero memory */
  memset (out, 0, sizeof(batch_row));
  
  /* Make generic structure */
  out->raw = make_table_row_from_def(table->def);

  /* attach data pointers to those in generic structure*/
  /* When the generic row is read/written the generic array is */
  /* automatically updated */
  out->file_on  = (char*)malloc(65*sizeof(char));
  out->file_off = (char*)malloc(65*sizeof(char));
  out->file_a   = (char*)malloc(65*sizeof(char));
  out->file_b   = (char*)malloc(65*sizeof(char));
  out->reason   = (char*)malloc(65*sizeof(char));

  /* initialize */
  strcpy (out->file_on,  "");
  strcpy (out->file_off, "");
  strcpy (out->file_a,   "");
  strcpy (out->file_b,   "");
  strcpy (out->reason,   "");
  /* fill appropriate values */
  out->mu12  =  (float) MAGICBLANK;	/* csabet 26.Jun.01 */
  out->mu22  =  (float) MAGICBLANK;	/* csabet 26.Jun.01 */
  out->emu12 =  (float) MAGICBLANK;	/* csabet 26.Jun.01 */
  out->emu22 =  (float) MAGICBLANK;	/* csabet 26.Jun.01 */
  out->v12   =  (float) MAGICBLANK;	/* csabet 26.Jun.01 */
  out->v22   =  (float) MAGICBLANK;	/* csabet 26.Jun.01 */
  out->ev12  =  (float) MAGICBLANK;	/* csabet 26.Jun.01 */
  out->ev22  =  (float) MAGICBLANK;	/* csabet 26.Jun.01 */
  out->t12   =  (float) MAGICBLANK;	/* csabet 26.Jun.01 */
  out->t22   =  (float) MAGICBLANK;	/* csabet 26.Jun.01 */
  out->et12  =  (float) MAGICBLANK;	/* csabet 26.Jun.01 */
  out->et22  =  (float) MAGICBLANK;	/* csabet 26.Jun.01 */
  out->sdopd =  (float) MAGICBLANK;	/* csabet 26.Jun.01 */
  out->sdopd1=  (float) MAGICBLANK;	/* csabet 26.Jun.01 */
  out->config = 1;
  for (i=0; i<4; i++) out->kappa[i]  = 0.0;
  for (i=0; i<2; i++) out->photom[i]  = 0.0;
  for (i=0; i<2; i++) out->photomsd[i]  = 0.0;
  for (i=0; i<3; i++) out->corpower[i]  = 0.0;
  for (i=0; i<3; i++) out->corpowsd[i]  = 0.0;

  return out; /* done */
} /* end  make_batch_row */

void kill_batch_row (batch_row *row)
/*----------------------------------------------------------------------- */
/*   Delete an batch_row structure                                        */
/*   Inputs:                                                              */
/*      row    pointer to the batch_row structure.                        */
/*----------------------------------------------------------------------- */
{
  if (row==NULL) return; /* sanity check */

  free (row->raw); /* generic structure */
  /* character arrays */
  if (row->file_on) free(row->file_on);
  if (row->file_off) free(row->file_off);
  if (row->file_a) free(row->file_a);
  if (row->file_b) free(row->file_b);
  if (row->reason) free(row->reason);
  free (row);      /* basic structure */
} /* end  kill_batch_row */


/* batch_table functions */
batch_table_def*
make_batch_table_def(fitsfile *fptr, char *date_obs, int calibr,
			float framrate, float fringspd, float maxwave,
			float  minwave, char *filter, int n_valid, 
                        int *ierr)
/*----------------------------------------------------------------------- */
/*   Make a new batch_table in an open FITS file                          */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      date_obs  Date (yyyy-mm-dd) observations began                    */
/*      calibr    if not =0 then at least some data is calibrated.        */
/*      framrate  frame rate (Hz)                                         */
/*      fringspd  fringe speed (m/s)                                      */
/*      maxwave   max wave length (m)                                     */
/*      minwave   min wave length (m)                                     */
/*      filter    Filter being used                                       */
/*      n_valid   Number of valid entries                                 */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An batch_table_def pointer; NULL indicates failure                */
/*----------------------------------------------------------------------- */
{
  batch_table_def *out = NULL;
  int    version = 2; /* table version number */
  int    ikey;

  if (*ierr>0) return NULL;    /* existing error condition? */

  if (fptr==NULL) return NULL; /* sanity check */

  /* create output object */
  out = (batch_table_def*)malloc(sizeof(batch_table_def));
  if (out==NULL)     {  /* check for allocation failure */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating %s table_def \n", tabletype);
    return out;
  }
  memset (out, 0, sizeof(batch_table_def)); /* zero fill */

  /* create table_def */
  out->def = make_table_def(BATCH_TABLE, 0, 0);
  if (out->def==NULL) { /* check for failure */
    kill_batch_table_def(out); /* clean up */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating BATCH_TABLE for %s table_def \n", 
	    tabletype);
    return NULL;
  }

  /* add keywords */
  ikey = table_find_key(out->def, "REVISION");
  if ((ikey>0) && (*ierr==0)) table_put_keyword (out->def, ikey, 
						(void*)&version, ierr);
  ikey = table_find_key(out->def, "DATE_OBS");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)date_obs, ierr);

  ikey = table_find_key(out->def, "CALIBR");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)&calibr, ierr);

  ikey = table_find_key(out->def, "FRAMRATE");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)&framrate, ierr);

  ikey = table_find_key(out->def, "FRINGSPD");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)&fringspd, ierr);

  ikey = table_find_key(out->def, "MAXWAVE");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)&maxwave, ierr);

  ikey = table_find_key(out->def, "MINWAVE");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)&minwave, ierr);

  ikey = table_find_key(out->def, "FILTER");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)filter, ierr);

  ikey = table_find_key(out->def, "N_VALID");
  if ((ikey>0) && (*ierr==0)) table_put_keyword (out->def, ikey, 
						(void*)&n_valid, ierr);

  /* get info from table descriptor */
  get_batch_info (out);

  /* open  table */
  make_table_from_def (fptr, out->def, ierr);

  return out; /* done */
} /* end make_batch_table */

batch_table_def* open_batch_table(fitsfile *fptr, int *ierr)
/*----------------------------------------------------------------------- */
/*   Open an existing batch_table in the active FITS file                 */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An batch_table_def pointer; NULL indicates failure                */
/*----------------------------------------------------------------------- */
{
  batch_table_def *out = NULL;

  if (*ierr>0) return NULL;    /* existing error condition? */
  if (fptr==NULL) return NULL; /* sanity check */

  /* create output object */
  out = (batch_table_def*)malloc(sizeof(batch_table_def));
  if (out==NULL)     {  /* check for allocation failure */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating %s table_def \n", tabletype);
    return out;
  }
  memset (out, 0, sizeof(batch_table_def)); /* zero fill */

  /* position FITS file at  table */
  if (fits_movnam_hdu(fptr, BINARY_TBL, tabletype, 0, ierr)) {
    fprintf(stderr,"ERROR %d finding %s table \n", *ierr, tabletype);
    kill_batch_table_def(out); /* clean up */
    *ierr = 10000;
    return NULL;
  }

  /* get table descriptor */
  out->def = make_table_def_from_table (fptr);
  if (out->def==NULL) { /* error check */
    fprintf(stderr,"ERROR making table def for %s \n", tabletype);
    kill_batch_table_def(out); /* clean up */
    *ierr = 10000;
    return NULL;
  }

  /* copy info from table descriptor */
  get_batch_info (out);

  return out; /* done */
} /* end open_batch_table */

int read_batch_table(fitsfile *fptr, batch_table_def* table, 
			    batch_row *row, long row_number,
			    int *ierr)
/*----------------------------------------------------------------------- */
/*   Read the next row from an batch_table                                */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row_number which row (1-rel) in the table                         */   
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      row       pointer to an extant row structure                      */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  table_row *trow;
  int        i;

  if (*ierr>0) return 1;     /* existing error condition? */
  if (fptr==NULL) return 1;  /* sanity check */
  if (table==NULL) return 1; /* sanity check */
  if (row==NULL) return 1;   /* sanity check */

  /* set row number */
  if (row_number>0) { /* explicit row number */
    row->raw->trow = row_number;
    table->row = row_number;
  } else { /* next row */
    row->raw->trow = ++table->row;
  }

  /* read row */
  table_read_row(fptr, table->def, row->raw, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR read_batch_table: reading table %s \n", 
	    tabletype);
    return 1;
  }

  /* copy relevant data from generic table_row */
  /* no need to copy data since pointers are OK */
  trow = row->raw;
  row->time        = trow->data[table->time_col]->tdata.tdouble[0];
  row->timei       = trow->data[table->timei_col]->tdata.tfloat[0];
  row->uvw[0]      = trow->data[table->uvw_col]->tdata.tdouble[0];
  row->uvw[1]      = trow->data[table->uvw_col]->tdata.tdouble[1];
  row->uvw[2]      = trow->data[table->uvw_col]->tdata.tdouble[2];
  row->mu12        = trow->data[table->mu12_col]->tdata.tfloat[0];
  row->mu22        = trow->data[table->mu22_col]->tdata.tfloat[0];
  row->emu12       = trow->data[table->emu12_col]->tdata.tfloat[0];
  row->emu22       = trow->data[table->emu22_col]->tdata.tfloat[0];
  row->v12         = trow->data[table->v12_col]->tdata.tfloat[0];
  row->v22         = trow->data[table->v22_col]->tdata.tfloat[0];
  row->ev12        = trow->data[table->ev12_col]->tdata.tfloat[0];
  row->ev22        = trow->data[table->ev22_col]->tdata.tfloat[0];
  row->t12         = trow->data[table->t12_col]->tdata.tfloat[0];
  row->t22         = trow->data[table->t22_col]->tdata.tfloat[0];
  row->et12        = trow->data[table->et12_col]->tdata.tfloat[0];
  row->et22        = trow->data[table->et22_col]->tdata.tfloat[0];
  row->sdopd       = trow->data[table->sdopd_col]->tdata.tfloat[0];
  row->sdopd1      = trow->data[table->sdopd1_col]->tdata.tfloat[0];
  /* some values were not in version 1, use -1.0 */
  if (table->kappa_col >= 0) {
    for (i=0; i<4; i++)
      row->kappa[i]  = trow->data[table->kappa_col]->tdata.tfloat[i];
  } else { /* not in input */
    for (i=0; i<4; i++) row->kappa[i]  = -1.0;
  }
  if (table->photom_col >= 0) {
    for (i=0; i<2; i++)
      row->photom[i]  = trow->data[table->photom_col]->tdata.tfloat[i];
  } else { /* not in input */
    for (i=0; i<2; i++) row->photom[i]  = -1.0;
  }
  if (table->photomsd_col >= 0) {
    for (i=0; i<2; i++)
      row->photomsd[i]  = trow->data[table->photomsd_col]->tdata.tfloat[i];
  } else { /* not in input */
    for (i=0; i<2; i++) row->photomsd[i]  = -1.0;
  }
  if (table->corpower_col >= 0) {
    for (i=0; i<3; i++)
      row->corpower[i]  = trow->data[table->corpower_col]->tdata.tfloat[i];
  } else { /* not in input */
    for (i=0; i<3; i++) row->corpower[i]  = -1.0;
  }
  if (table->corpowsd_col >= 0) {
    for (i=0; i<3; i++)
      row->corpowsd[i]  = trow->data[table->corpowsd_col]->tdata.tfloat[i];
  } else { /* not in input */
    for (i=0; i<3; i++) row->corpowsd[i]  = -1.0;
  }
  if (table->config_col >= 0) 
    row->config = trow->data[table->config_col]->tdata.tshort[0];
  else
    row->config = 1;
  row->shape       = trow->data[table->shape_col]->tdata.tfloat[0];
  row->batch       = trow->data[table->batch_col]->tdata.tshort[0];
  row->source      = trow->data[table->source_col]->tdata.tshort[0];
  row->nscan       = trow->data[table->nscan_col]->tdata.tshort[0];
  row->ot[0]       = trow->data[table->ot_col]->tdata.tshort[0];
  row->ot[1]       = trow->data[table->ot_col]->tdata.tshort[1];
  row->flag        = trow->data[table->flag_col]->tdata.tshort[0];
  strncpy (row->file_on, trow->data[table->file_on_col]->tdata.tchar, 64);
  strncpy (row->file_off, trow->data[table->file_off_col]->tdata.tchar, 64);
  strncpy (row->file_a, trow->data[table->file_a_col]->tdata.tchar, 64);
  strncpy (row->file_b, trow->data[table->file_b_col]->tdata.tchar, 64);
  strncpy (row->reason, trow->data[table->reason_col]->tdata.tchar, 64);

  return 0; /* OK - done */
} /* end read_batch_table */

int write_batch_table(fitsfile *fptr, batch_table_def* table, 
			     batch_row *row, long row_number, 
			     int *ierr)
/*----------------------------------------------------------------------- */
/*   Write  the next row to an batch_table                                */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row       pointer to an extant row structure                      */
/*      row_number which row (1-rel) in the table                         */     
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  table_row *trow;
  int        i;

  if (*ierr>0) return 1;     /* existing error condition? */
  if (fptr==NULL) return 1;  /* sanity check */
  if (table==NULL) return 1; /* sanity check */
  if (row==NULL) return 1;   /* sanity check */

  /* set row number */
  if (row_number>0) { /* explicit row number */
    row->raw->trow = row_number;
    table->row = row_number;
  } else { /* next row */
    row->raw->trow = ++table->row;
  }

  /* copy relevant data to generic table_row */
  /* no need to copy data since pointers are OK */
  trow = row->raw;
  if (table->time_col >= 0) 
    trow->data[table->time_col]->tdata.tdouble[0]    = row->time;
  if (table->timei_col >= 0) 
    trow->data[table->timei_col]->tdata.tfloat[0]    = row->timei;
  if (table->uvw_col >= 0) 
    trow->data[table->uvw_col]->tdata.tdouble[0]     = row->uvw[0];
  if (table->uvw_col >= 0) 
    trow->data[table->uvw_col]->tdata.tdouble[1]     = row->uvw[1];
  if (table->uvw_col >= 0) 
    trow->data[table->uvw_col]->tdata.tdouble[2]     = row->uvw[2];
  if (table->mu12_col >= 0) 
    trow->data[table->mu12_col]->tdata.tfloat[0]     = row->mu12;
  if (table->mu22_col >= 0) 
    trow->data[table->mu22_col]->tdata.tfloat[0]     = row->mu22;
  if (table->emu12_col >= 0) 
    trow->data[table->emu12_col]->tdata.tfloat[0]    = row->emu12;
  if (table->emu22_col >= 0) 
    trow->data[table->emu22_col]->tdata.tfloat[0]    = row->emu22;
  if (table->v12_col >= 0) 
    trow->data[table->v12_col]->tdata.tfloat[0]     = row->v12;
  if (table->v22_col >= 0) 
    trow->data[table->v22_col]->tdata.tfloat[0]     = row->v22;
  if (table->ev12_col >= 0) 
    trow->data[table->ev12_col]->tdata.tfloat[0]    = row->ev12;
  if (table->ev22_col >= 0) 
    trow->data[table->ev22_col]->tdata.tfloat[0]    = row->ev22;
  if (table->t12_col >= 0) 
    trow->data[table->t12_col]->tdata.tfloat[0]     = row->t12;
  if (table->t22_col >= 0) 
    trow->data[table->t22_col]->tdata.tfloat[0]     = row->t22;
  if (table->et12_col >= 0) 
    trow->data[table->et12_col]->tdata.tfloat[0]    = row->et12;
  if (table->et22_col >= 0) 
    trow->data[table->et22_col]->tdata.tfloat[0]    = row->et22;
  if (table->sdopd_col >= 0) 
    trow->data[table->sdopd_col]->tdata.tfloat[0]    = row->sdopd;
  if (table->sdopd1_col >= 0) 
    trow->data[table->sdopd1_col]->tdata.tfloat[0]   = row->sdopd1;
  if (table->kappa_col >= 0) 
    /* some values were not in version 1, ignore */
    if (table->kappa_col >= 0) {
      for (i=0; i<4; i++)
	trow->data[table->kappa_col]->tdata.tfloat[i] = row->kappa[i];
    }
  if (table->photom_col >= 0) {
    for (i=0; i<2; i++)
      trow->data[table->photom_col]->tdata.tfloat[i] = row->photom[i];
  } 
  if (table->photomsd_col >= 0) {
    for (i=0; i<2; i++)
      trow->data[table->photomsd_col]->tdata.tfloat[i] = row->photomsd[i];
  } 
  if (table->corpower_col >= 0) {
    for (i=0; i<3; i++)
      trow->data[table->corpower_col]->tdata.tfloat[i] = row->corpower[i];
  } 
  if (table->corpowsd_col >= 0) {
    for (i=0; i<3; i++)
      trow->data[table->corpowsd_col]->tdata.tfloat[i] = row->corpowsd[i];
  } 
  if (table->config_col >= 0) 
    trow->data[table->config_col]->tdata.tshort[0] = row->config;
  if (table->shape_col >= 0) 
    trow->data[table->shape_col]->tdata.tfloat[0]   = row->shape;
  if (table->batch_col >= 0) 
    trow->data[table->batch_col]->tdata.tshort[0]   = row->batch;
  if (table->source_col >= 0) 
    trow->data[table->source_col]->tdata.tshort[0]  = row->source;
  if (table->nscan_col >= 0) 
    trow->data[table->nscan_col]->tdata.tshort[0]   = row->nscan;
  if (table->ot_col >= 0) 
    trow->data[table->ot_col]->tdata.tshort[0]      = row->ot[0];
  if (table->ot_col >= 0) 
    trow->data[table->ot_col]->tdata.tshort[1]      = row->ot[1];
  if (table->flag_col >= 0) 
    trow->data[table->flag_col]->tdata.tshort[0]    = row->flag;
  if (table->file_on_col >= 0) 
    strncpy (trow->data[table->file_on_col]->tdata.tchar, row->file_on, 64);
  if (table->file_off_col >= 0) 
    strncpy (trow->data[table->file_off_col]->tdata.tchar, row->file_off, 64);
  if (table->file_a_col >= 0) 
    strncpy (trow->data[table->file_a_col]->tdata.tchar, row->file_a, 64);
  if (table->file_b_col >= 0) 
    strncpy (trow->data[table->file_b_col]->tdata.tchar, row->file_b, 64);
  if (table->reason_col >= 0) 
    strncpy (trow->data[table->reason_col]->tdata.tchar, row->reason, 64);
  
  /* write row */
  table_write_row(fptr, table->def, row->raw, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR  write_batch_table: writing table %s \n", 
	    tabletype);
    return 1;
  }

  return 0; /* OK - done */
} /* end write_batch_table */

int close_batch_table(fitsfile *fptr, batch_table_def* table, 
			 int *ierr)
/*----------------------------------------------------------------------- */
/*   Close an open batch_table, associated structures are destroyed       */
/*   Note: this closes the full FITS file                                 */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  if (*ierr>0) return 1;    /* existing error condition? */
  if (fptr==NULL) return 1; /* sanity check */

  /* do actual file close */
  fits_close_file (fptr, ierr);

  /* delete structures */
  kill_batch_table_def(table);

  return 0; /* OK - done */
} /* end close_batch_table */

void kill_batch_table_def (batch_table_def* table)
/*----------------------------------------------------------------------- */
/*   Delete an batch_table structure                                      */
/*   Inputs:                                                              */
/*      table     table definitions structure from open or create         */
/*----------------------------------------------------------------------- */
{
  if (table==NULL) return; /* sanity check */

  /* table definition structure */
  kill_table_def(table->def);

  /* free basic structure */
  free (table);
} /* end kill_batch_table_def */

/*************************private functions ******************************/

void get_batch_info (batch_table_def *table)
/*----------------------------------------------------------------------- */
/*   Get data from table_def specific to batch table                      */
/*   Inputs:                                                              */
/*      table     table definitions structure                             */
/*----------------------------------------------------------------------- */
{
  int ikey, ierr=0;

  if (table==NULL) return;         /* sanity check */

  /* keyword values */
  ikey = table_find_key(table->def, "DATE_OBS");
  if (ikey>0) table->date_obs = (char*)table->def->keydata[ikey-1];
  ikey = table_find_key(table->def, "FILTER");
  if (ikey>0) table->filter = (char*)table->def->keydata[ikey-1];
  ikey = table_find_key(table->def, "CALIBR");
  if (ikey>0) 
    table_get_keyword(table->def, ikey, (void*)&table->calibr, &ierr);
  ikey = table_find_key(table->def, "FRAMRATE");
  if (ikey>0) 
    table_get_keyword(table->def, ikey, (void*)&table->framrate, &ierr);
  ikey = table_find_key(table->def, "FRINGSPD");
  if (ikey>0) 
    table_get_keyword(table->def, ikey, (void*)&table->fringspd, &ierr);
  ikey = table_find_key(table->def, "MAXWAVE");
  if (ikey>0) 
    table_get_keyword(table->def, ikey, (void*)&table->maxwave, &ierr);
  ikey = table_find_key(table->def, "MINWAVE");
  if (ikey>0) 
    table_get_keyword(table->def, ikey, (void*)&table->minwave, &ierr);
  ikey = table_find_key(table->def, "N_VALID");
  if (ikey>0) 
    table_get_keyword(table->def, ikey, (void*)&table->n_valid, &ierr);

  /* lookup column numbers (0-rel) */
  table->batch_col     = table_find_col (table->def, "BATCH_ID")-1;
  table->time_col      = table_find_col (table->def, "TIME")-1;
  table->timei_col     = table_find_col (table->def, "EXP_TIME")-1;
  table->source_col    = table_find_col (table->def, "SOURCE")-1;
  table->nscan_col     = table_find_col (table->def, "NSCAN")-1;
  table->ot_col        = table_find_col (table->def, "OPT_TRAIN")-1;
  /* use "TELESCOPE" if "OPT_TRAIN" not there */
  if (table->ot_col<0)table->ot_col  = table_find_col (table->def, "TELESCOPE")-1;
  table->config_col    = table_find_col (table->def, "CONFIG")-1;
  table->uvw_col       = table_find_col (table->def, "UVW")-1;
  table->mu12_col      = table_find_col (table->def, "MU12")-1;
  table->mu22_col      = table_find_col (table->def, "MU22")-1;
  table->emu12_col     = table_find_col (table->def, "EMU12")-1;
  table->emu22_col     = table_find_col (table->def, "EMU22")-1;
  table->v12_col       = table_find_col (table->def, "V12")-1;
  table->v22_col       = table_find_col (table->def, "V22")-1;
  table->ev12_col      = table_find_col (table->def, "EV12")-1;
  table->ev22_col      = table_find_col (table->def, "EV22")-1;
  table->t12_col       = table_find_col (table->def, "T12")-1;
  table->t22_col       = table_find_col (table->def, "T22")-1;
  table->et12_col      = table_find_col (table->def, "ET12")-1;
  table->et22_col      = table_find_col (table->def, "ET22")-1;
  table->sdopd_col     = table_find_col (table->def, "SDOPD")-1;
  table->sdopd1_col    = table_find_col (table->def, "SDOPD1")-1;
  table->kappa_col     = table_find_col (table->def, "KAPPA")-1;
  table->photom_col    = table_find_col (table->def, "PHOTOM")-1;
  table->photomsd_col  = table_find_col (table->def, "PHOTOMSD")-1;
  table->corpower_col  = table_find_col (table->def, "CORPOWER")-1;
  table->corpowsd_col  = table_find_col (table->def, "CORPOWSD")-1;
  table->file_on_col   = table_find_col (table->def, "FILE_ON")-1;
  table->file_off_col  = table_find_col (table->def, "FILE_OFF")-1;
  table->file_a_col    = table_find_col (table->def, "FILE_A")-1;
  table->file_b_col    = table_find_col (table->def, "FILE_B")-1;
  table->shape_col     = table_find_col (table->def, "SHAPE")-1;
  table->flag_col      = table_find_col (table->def, "FLAG")-1;  
  table->reason_col    = table_find_col (table->def, "REASON")-1;
} /* end  get_batch_info */

