/* $Id: batch.h,v 1.2 2001/11/26 12:52:50 csabet Exp $ */
#ifndef _BATCH_H
#define _BATCH_H
/*----------------------------------------------------------------------- */
/* Batch data FITS binary tables routines                                 */
/* Partially/fully calibrated data table for VLTI/VINCI interferometer    */
/*----------------------------------------------------------------------- */
#include "VLTItable.h"

typedef struct      /* table definition for batch table */
{
  table_def *def;      /* basic table information */
  long   row;          /* last row (0-rel) read or written, -1=> none. */
  char   *date_obs;    /* observing date */
  int    calibr;       /* data calibrated flag */
  double  framrate;    /* Frame rate (Hz) */
  double  fringspd;    /* Fringe speed (m/s) */
  double  maxwave;     /* maximum wavelength used (m) */
  double  minwave;     /* minimum wavelength used (m) */
  char   *filter;      /* filter name */
  int    n_valid;      /* number of valid entries in table */
  int    batch_col;    /* batch index column number */
  int    time_col;     /* time column number (0-rel)*/
  int    timei_col;    /* averaging time column number */
  int    source_col;   /* source number column number */
  int    uvw_col;      /* uvw column number */
  int    nscan_col;    /* number of scans averaged column number */
  int    ot_col;       /* Optical train numbers column number */
  int    config_col;   /* Configuration numbers column number */
  int    mu12_col;     /* first average coherence factor column number */
  int    mu22_col;     /* second average coherence factor column number */
  int    emu12_col;    /* first average coherence factor error column number */
  int    emu22_col;    /* 2nd average coherence factor error column number */
  int    v12_col;      /* first average vis sq. column number */
  int    v22_col;      /* second average vis sq. column number */
  int    ev12_col;     /* first average vis sq. error column number */
  int    ev22_col;     /* 2nd average  vis sq. error column number */
  int    t12_col;      /* first transfer fn. column number */
  int    t22_col;      /* second transfer fn.column number */
  int    et12_col;     /* first transfer fn. error column number */
  int    et22_col;     /* 2nd transfer fn. error column number */
  int    sdopd_col;    /* standard deviation of OPD residual column number */
  int    sdopd1_col;   /* standard deviation of 1st derivative of OPD column number */
  int    kappa_col;    /* coupling coefficient matrix column number */
  int    photom_col;   /* average photometric values column number */
  int    photomsd_col; /* std. dev. of photometric values column number */
  int    corpower_col; /* Moments of the correlated power column number */
  int    corpowsd_col; /* Standard deviations of the moments column number */
  int    file_on_col;  /* on-source file name column number */
  int    file_off_col; /* off-source file name column number */
  int    file_a_col;   /* port A only file name column number */
  int    file_b_col;   /* port B only file name column number */
  int    shape_col;    /* Shape factor column number */
  int    flag_col;     /* Flag column number */
  int    reason_col;   /* Flag reason column number */
} batch_table_def;

typedef struct      /* structure used for batch row */
{
  double time;       /* central time of exposure */
  double uvw[3];     /* Spatial frequency coordinated for observations */
  float  timei;      /* data averaging time */
  float  mu12;       /* first average coherence factor */
  float  mu22;       /* second average coherence factor */
  float  emu12;      /* first average coherence factor error */
  float  emu22;      /* 2nd average coherence factor error */
  float  v12;        /* first average vis sq. */
  float  v22;        /* second average vis sq. */
  float  ev12;       /* first average vis sq. error */
  float  ev22;       /* 2nd average  vis sq. error */
  float  t12;        /* first transfer fn. */
  float  t22;        /* second transfer fn. */
  float  et12;       /* first transfer fn. error */
  float  et22;       /* 2nd transfer fn. error  */
  float  sdopd;      /* standard deviation of OPD residual */
  float  sdopd1;     /* standard deviation of 1st derivative of OPD */
  float  kappa[4];   /* coupling coefficient matrix */
  float  photom[2];  /* Batch average photometric values */
  float  photomsd[2];/* Batch std. dev. of photometric values */
  float  corpower[3];/* Moments of the correlated power (filtered I1-I2). */
  float  corpowsd[3];/* Standard deviations of the moments of the correlated power */
  float  shape;      /* shape factor */
  int    batch;      /* batch number (index) */
  int    nscan;      /* number of scans averaged */
  int    source;     /* source id */
  int    ot[2];      /* optical train numbers of array */
  int    config;     /* configuration number */
  int    flag;       /* Good/bad flag */
  char   *file_on;   /* on-source file name */
  char   *file_off;  /* off-source file name  */
  char   *file_a;    /* port A only file */
  char   *file_b;    /* port B only file name */
  char   *reason;    /* Flag reason  */
  table_row *raw;    /* generic table row */
} batch_row;

/* public functions */

/* batch_row functions */
batch_row* make_batch_row(batch_table_def* table);
/*----------------------------------------------------------------------- */
/*   Make an batch_row structure                                          */
/*   Inputs:                                                              */
/*      table     pointer to batch table structure                        */
/*   Returns:                                                             */
/*      An batch_row structure pointer ; NULL indicates failure           */
/*      Data pointers set to generic structure values                     */
/*----------------------------------------------------------------------- */

void kill_batch_row (batch_row *row);
/*----------------------------------------------------------------------- */
/*   Delete an batch_row structure                                        */
/*   Inputs:                                                              */
/*      row    pointer to the batch_row structure.                        */
/*----------------------------------------------------------------------- */

/* batch_table functions */
batch_table_def* 
make_batch_table_def(fitsfile *fptr, char *date_obs, int calibr,
			float framrate, float fringspd, float maxwave,
			float  minwave, char *filter, int n_valid, 
                        int *ierr);
/*----------------------------------------------------------------------- */
/*   Make a new batch_table in an open FITS file                          */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      date_obs  Date (yyyy-mm-dd) observations began                    */
/*      calibr    if not =0 then at least some data is calibrated.        */
/*      framrate  frame rate (Hz)                                         */
/*      fringspd  fringe speed (m/s)                                      */
/*      maxwave   max wave length (m)                                     */
/*      minwave   min wave length (m)                                     */
/*      filter    Filter being used                                       */
/*      n_valid   Number of valid entries                                 */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An batch_table_def pointer; NULL indicates failure                */
/*----------------------------------------------------------------------- */

batch_table_def* 
open_batch_table(fitsfile *fptr, int *ierr);
/*----------------------------------------------------------------------- */
/*   Open an existing batch_table in the active FITS file                 */
/*   Fills batch_geom and saves in batch_table_def                        */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An batch_table_def pointer; NULL indicates failure                */
/*----------------------------------------------------------------------- */

int read_batch_table(fitsfile *fptr,  batch_table_def* table, 
			    batch_row *row, long row_number,
			    int *ierr);
/*----------------------------------------------------------------------- */
/*   Read the next row from an batch_table                                */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row_number which row (1-rel) in the table                         */   
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      row       pointer to an extant row structure                      */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */

int write_batch_table(fitsfile *fptr, batch_table_def* table, 
			     batch_row *row,  long row_number,
			     int *ierr);
/*----------------------------------------------------------------------- */
/*   Write  the next row to an batch_table                                */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row       pointer to an extant row structure                      */
/*      row_number which row (1-rel) in the table                         */ 
/*   Input/output:                                                        */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */

int close_batch_table(fitsfile *fptr, batch_table_def* table, 
			     int *ierr);
/*----------------------------------------------------------------------- */
/*   Close an open batch_table, associated structures are destroyed       */
/*   Note: this closes the full FITS file                                 */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */

void kill_batch_table_def (batch_table_def* table);
/*----------------------------------------------------------------------- */
/*   Delete an batch_table structure                                      */
/*   Inputs:                                                              */
/*      table     table definitions structure from open or create         */
/*----------------------------------------------------------------------- */

#endif /* end BATCH_H */
