/* $Id: fringe_a.c,v 1.2 2001/11/26 12:52:49 csabet Exp $ */
/*----------------------------------------------------------------------- */
/* Fringe amplitude data FITS binary tables routines                      */
/* Partially/fully calibrated data table for optical/IR interferometer    */
/*----------------------------------------------------------------------- */
/*#include <malloc.h>*/			/* csabet 26.Jun.01 */
#include <string.h> 		/* csabet 26.Jun.01 */
#include "fringe_a.h"
/* #include "vndrs_OPD.h" */		/* csabet 26.Jun.01 */

#define MAGICBLANK (-999.99)   /* Magic blanking value for bad values */

/* internal prototypes */
/* get data from table_def specific to fringe_a table */
void get_fringe_a_info (fringe_a_table_def *table);

/* file global variables for table type */
static char *tabletype = "FRINGE_A";

/* fringe_a_row functions */
fringe_a_row* make_fringe_a_row(fringe_a_table_def* table)
/*----------------------------------------------------------------------- */
/*   Make an fringe_a_row structure                                       */
/*   Inputs:                                                              */
/*      table     pointer to fringe_a table structure                     */
/*   Returns:                                                             */
/*      An fringe_a_row structure pointer ; NULL indicates failure        */
/*      Data pointers set to generic structure values                     */
/*----------------------------------------------------------------------- */
{
  fringe_a_row *out;
  /* int i, icol, nregion; */
  int i; /* csabet 28.Jun.01 */

  /* create basic structure */
  out = (fringe_a_row*)malloc(sizeof(fringe_a_row));
  if (out==NULL) return out; /* allocation failure? */
  
  /* zero memory */
  memset (out, 0, sizeof(fringe_a_row));
  
  /* Make generic structure */
  out->raw = make_table_row_from_def(table->def);

  /* fill appropriate values */
  /* out->mu12  =  -999.99; */
	out->mu12  =  (float) MAGICBLANK;	/* csabet 28.Jun.01 */
  /* out->mu22  =  -999.99; */
	out->mu22  =  (float) MAGICBLANK;	/* csabet 28.Jun.01 */
  out->config = 1;
  for (i=0; i<2; i++) out->photom[i]  = 0.0;
  for (i=0; i<3; i++) out->corpower[i]  = 0.0;

  /* attach data pointers to those in generic structure*/
  /* When the generic row is read/written the generic array is */
  /* automatically updated */

  return out; /* done */
} /* end  make_fringe_a_row */

void kill_fringe_a_row (fringe_a_row *row)
/*----------------------------------------------------------------------- */
/*   Delete an fringe_a_row structure                                     */
/*   Inputs:                                                              */
/*      row    pointer to the fringe_a_row structure.                     */
/*----------------------------------------------------------------------- */
{
  /* int i, nregion; */ /* csabet 28.Jun.01 */

  if (row==NULL) return; /* sanity check */

  free (row->raw); /* generic structure */
  free (row);      /* basic structure */
} /* end  kill_fringe_a_row */


/* fringe_a_table functions */
fringe_a_table_def*
make_fringe_a_table_def(fitsfile *fptr, char *date_obs, 
			float framrate, float fringspd, float maxwave,
			float  minwave, char *filter, int *ierr)
/*----------------------------------------------------------------------- */
/*   Make a new fringe_a_table in an open FITS file                       */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      date_obs  Date (yyyy-mm-dd) observations began                    */
/*      framrate  frame rate (Hz)                                         */
/*      fringspd  fringe speed (m/s)                                      */
/*      maxwave   max wave length (m)                                     */
/*      minwave   min wave length (m)                                     */
/*      filter    Filter being used                                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An fringe_a_table_def pointer; NULL indicates failure             */
/*----------------------------------------------------------------------- */
{
  fringe_a_table_def *out = NULL;
  int    version = 2; /* table version number */
  /* int    ikey, icol, i; */
	int	ikey;	/* csabet 28.Jun.01 */

  if (*ierr>0) return NULL;    /* existing error condition? */

  if (fptr==NULL) return NULL; /* sanity check */

  /* create output object */
  out = (fringe_a_table_def*)malloc(sizeof(fringe_a_table_def));
  if (out==NULL)     {  /* check for allocation failure */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating %s table_def \n", tabletype);
    return out;
  }
  memset (out, 0, sizeof(fringe_a_table_def)); /* zero fill */

  /* create table_def */
  out->def = make_table_def(FRINGE_A_TABLE, 0, 0);
  if (out->def==NULL) { /* check for failure */
    kill_fringe_a_table_def(out); /* clean up */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating FRINGE_A_TABLE for %s table_def \n", 
	    tabletype);
    return NULL;
  }

  /* add keywords */
  ikey = table_find_key(out->def, "REVISION");
  if ((ikey>0) && (*ierr==0)) table_put_keyword (out->def, ikey, 
						(void*)&version, ierr);
  ikey = table_find_key(out->def, "DATE_OBS");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)date_obs, ierr);

  ikey = table_find_key(out->def, "FRAMRATE");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)&framrate, ierr);

  ikey = table_find_key(out->def, "FRINGSPD");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)&fringspd, ierr);

  ikey = table_find_key(out->def, "MAXWAVE");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)&maxwave, ierr);

  ikey = table_find_key(out->def, "MINWAVE");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)&minwave, ierr);

  ikey = table_find_key(out->def, "FILTER");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)filter, ierr);


  /* get info from table descriptor */
  get_fringe_a_info (out);

  /* open  table */
  make_table_from_def (fptr, out->def, ierr);

  return out; /* done */
} /* end make_fringe_a_table */

fringe_a_table_def* open_fringe_a_table(fitsfile *fptr, int *ierr)
/*----------------------------------------------------------------------- */
/*   Open an existing fringe_a_table in the active FITS file              */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An fringe_a_table_def pointer; NULL indicates failure             */
/*----------------------------------------------------------------------- */
{
  fringe_a_table_def *out = NULL;

  if (*ierr>0) return NULL;    /* existing error condition? */
  if (fptr==NULL) return NULL; /* sanity check */

  /* create output object */
  out = (fringe_a_table_def*)malloc(sizeof(fringe_a_table_def));
  if (out==NULL)     {  /* check for allocation failure */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating %s table_def \n", tabletype);
    return out;
  }
  memset (out, 0, sizeof(fringe_a_table_def)); /* zero fill */

  /* position FITS file at  table */
  if (fits_movnam_hdu(fptr, BINARY_TBL, tabletype, 0, ierr)) {
    fprintf(stderr,"ERROR %d finding %s table \n", *ierr, tabletype);
    kill_fringe_a_table_def(out); /* clean up */
    *ierr = 10000;
    return NULL;
  }

  /* get table descriptor */
  out->def = make_table_def_from_table (fptr);
  if (out->def==NULL) { /* error check */
    fprintf(stderr,"ERROR making table def for %s \n", tabletype);
    kill_fringe_a_table_def(out); /* clean up */
    *ierr = 10000;
    return NULL;
  }

  /* copy info from table descriptor */
  get_fringe_a_info (out);

  return out; /* done */
} /* end open_fringe_a_table */

int read_fringe_a_table(fitsfile *fptr, fringe_a_table_def* table, 
			    fringe_a_row *row, long row_number,
			    int *ierr)
/*----------------------------------------------------------------------- */
/*   Read the next row from an fringe_a_table                             */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row_number which row (1-rel) in the table                         */   
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      row       pointer to an extant row structure                      */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  table_row *trow;
  int        i;

  if (*ierr>0) return 1;     /* existing error condition? */
  if (fptr==NULL) return 1;  /* sanity check */
  if (table==NULL) return 1; /* sanity check */
  if (row==NULL) return 1;   /* sanity check */

  /* set row number */
  if (row_number>0) { /* explicit row number */
    row->raw->trow = row_number;
    table->row = row_number;
  } else { /* next row */
    row->raw->trow = ++table->row;
  }

  /* read row */
  table_read_row(fptr, table->def, row->raw, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR  read_fringe_a_table: reading table %s \n", 
	    tabletype);
    return 1;
  }

  /* copy relevant data from generic table_row */
  /* no need to copy data since pointers are OK */
  trow = row->raw;
  row->time        = trow->data[table->time_col]->tdata.tdouble[0];
  row->timei       = trow->data[table->timei_col]->tdata.tfloat[0];
  row->uvw[0]      = trow->data[table->uvw_col]->tdata.tdouble[0];
  row->uvw[1]      = trow->data[table->uvw_col]->tdata.tdouble[1];
  row->uvw[2]      = trow->data[table->uvw_col]->tdata.tdouble[2];
  row->mu12        = trow->data[table->mu12_col]->tdata.tfloat[0];
  row->mu22        = trow->data[table->mu22_col]->tdata.tfloat[0];
  row->opd         = trow->data[table->opd_col]->tdata.tfloat[0];
  row->opd1        = trow->data[table->opd1_col]->tdata.tfloat[0];
  row->opd2        = trow->data[table->opd2_col]->tdata.tfloat[0];
  /* some values were not in version 1, use -1.0 */
  if (table->photom_col >= 0) {
    for (i=0; i<2; i++)
      row->photom[i]  = trow->data[table->photom_col]->tdata.tfloat[i];
  } else { /* not in input */
    for (i=0; i<2; i++) row->photom[i]  = -1.0;
  }
  if (table->corpower_col >= 0) {
    for (i=0; i<3; i++)
      row->corpower[i]  = trow->data[table->corpower_col]->tdata.tfloat[i];
  } else { /* not in input */
    for (i=0; i<3; i++) row->corpower[i]  = -1.0;
  }
  if (table->config_col >= 0) 
    row->config = trow->data[table->config_col]->tdata.tshort[0];
  else
    row->config = 1;
  row->snrp1    = trow->data[table->snrp1_col]->tdata.tfloat[0];
  row->snrp2    = trow->data[table->snrp2_col]->tdata.tfloat[0];
  row->weight   = trow->data[table->weight_col]->tdata.tfloat[0];
  row->source   = trow->data[table->source_col]->tdata.tshort[0];
  row->ot[0]    = trow->data[table->ot_col]->tdata.tshort[0];
  row->ot[1]    = trow->data[table->ot_col]->tdata.tshort[1];
  row->batch    = trow->data[table->batch_col]->tdata.tshort[0];

  return 0; /* OK - done */
} /* end read_fringe_a_table */

int write_fringe_a_table(fitsfile *fptr, fringe_a_table_def* table, 
			     fringe_a_row *row, long row_number, 
			     int *ierr)
/*----------------------------------------------------------------------- */
/*   Write  the next row to an fringe_a_table                             */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row       pointer to an extant row structure                      */
/*      row_number which row (1-rel) in the table                         */ 
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  table_row *trow;
  int        i;

  if (*ierr>0) return 1;     /* existing error condition? */
  if (fptr==NULL) return 1;  /* sanity check */
  if (table==NULL) return 1; /* sanity check */
  if (row==NULL) return 1;   /* sanity check */

  /* set row number */
  if (row_number>0) { /* explicit row number */
    row->raw->trow = row_number;
    table->row = row_number;
  } else { /* next row */
    row->raw->trow = ++table->row;
  }

  /* copy relevant data to generic table_row */
  /* no need to copy data since pointers are OK */
  trow = row->raw;
  if (table->time_col >=0)
     trow->data[table->time_col]->tdata.tdouble[0] = row->time;
  if (table->timei_col >=0)
     trow->data[table->timei_col]->tdata.tfloat[0] = row->timei;
  if (table->uvw_col >=0)
     trow->data[table->uvw_col]->tdata.tdouble[0]  = row->uvw[0];
  if (table->uvw_col >=0)
     trow->data[table->uvw_col]->tdata.tdouble[1]  = row->uvw[1];
  if (table->uvw_col >=0)
     trow->data[table->uvw_col]->tdata.tdouble[2]  = row->uvw[2];
  if (table->mu12_col >=0)
     trow->data[table->mu12_col]->tdata.tfloat[0]  = row->mu12;
  if (table->mu22_col >=0)
     trow->data[table->mu22_col]->tdata.tfloat[0]  = row->mu22;
  if (table->opd_col >=0)
     trow->data[table->opd_col]->tdata.tfloat[0]   = row->opd;
  if (table->opd2_col >=0)
     trow->data[table->opd1_col]->tdata.tfloat[0]  = row->opd1;
  if (table->opd2_col >=0)
     trow->data[table->opd2_col]->tdata.tfloat[0]  = row->opd2;
     /* some values were not in version 1, ignore */
  if (table->photom_col >= 0) {
    for (i=0; i<2; i++)
      trow->data[table->photom_col]->tdata.tfloat[i] = row->photom[i];
  } 
  if (table->corpower_col >= 0) {
    for (i=0; i<3; i++)
      trow->data[table->corpower_col]->tdata.tfloat[i] = row->corpower[i];
  } 
  if (table->config_col >= 0) 
    trow->data[table->config_col]->tdata.tshort[0] = row->config;
  if (table->snrp1_col >=0)
    trow->data[table->snrp1_col]->tdata.tfloat[0]  = row->snrp1;
  if (table->snrp2_col >=0)
    trow->data[table->snrp2_col]->tdata.tfloat[0]  = row->snrp2;
  if (table->weight_col >=0)
    trow->data[table->weight_col]->tdata.tfloat[0] = row->weight;
  if (table->source_col >=0)
    trow->data[table->source_col]->tdata.tshort[0] = row->source;
  if (table->ot_col >=0)
    trow->data[table->ot_col]->tdata.tshort[0]     = row->ot[0];
  if (table->ot_col >=0)
    trow->data[table->ot_col]->tdata.tshort[1]     = row->ot[1];
  if (table->batch_col >=0)
    trow->data[table->batch_col]->tdata.tshort[0]  = row->batch;
    
  /* write row */
  table_write_row(fptr, table->def, row->raw, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR  write_fringe_a_table: writing table %s \n", 
	    tabletype);
    return 1;
  }

  return 0; /* OK - done */
} /* end write_fringe_a_table */

int close_fringe_a_table(fitsfile *fptr, fringe_a_table_def* table, 
			 int *ierr)
/*----------------------------------------------------------------------- */
/*   Close an open fringe_a_table, associated structures are destroyed    */
/*   Note: this closes the full FITS file                                 */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  if (*ierr>0) return 1;    /* existing error condition? */
  if (fptr==NULL) return 1; /* sanity check */

  /* do actual file close */
  fits_close_file (fptr, ierr);

  /* delete structures */
  kill_fringe_a_table_def(table);

  return 0; /* OK - done */
} /* end close_fringe_a_table */

void kill_fringe_a_table_def (fringe_a_table_def* table)
/*----------------------------------------------------------------------- */
/*   Delete an fringe_a_table structure                                   */
/*   Inputs:                                                              */
/*      table     table definitions structure from open or create         */
/*----------------------------------------------------------------------- */
{
  if (table==NULL) return; /* sanity check */

  /* table definition structure */
  kill_table_def(table->def);

  /* free basic structure */
  free (table);
} /* end kill_fringe_a_table_def */

/*************************private functions ******************************/

void get_fringe_a_info (fringe_a_table_def *table)
/*----------------------------------------------------------------------- */
/*   Get data from table_def specific to fringe_a table                   */
/*   Inputs:                                                              */
/*      table     table definitions structure                             */
/*----------------------------------------------------------------------- */
{
  int ikey, ierr=0;

  if (table==NULL) return;         /* sanity check */

  /* keyword values */
  ikey = table_find_key(table->def, "DATE_OBS");
  if (ikey>0) table->date_obs = (char*)table->def->keydata[ikey-1];
  ikey = table_find_key(table->def, "FILTER");
  if (ikey>0) table->filter = (char*)table->def->keydata[ikey-1];
  ikey = table_find_key(table->def, "FRAMRATE");
  if (ikey>0) 
    table_get_keyword(table->def, ikey, (void*)&table->framrate, &ierr);
  ikey = table_find_key(table->def, "FRINGSPD");
  if (ikey>0) 
    table_get_keyword(table->def, ikey, (void*)&table->fringspd, &ierr);
  ikey = table_find_key(table->def, "MAXWAVE");
  if (ikey>0) 
    table_get_keyword(table->def, ikey, (void*)&table->maxwave, &ierr);
  ikey = table_find_key(table->def, "MINWAVE");
  if (ikey>0) 
    table_get_keyword(table->def, ikey, (void*)&table->minwave, &ierr);

  /* lookup column numbers (0-rel) */
  table->time_col      = table_find_col (table->def, "TIME")-1;
  table->timei_col     = table_find_col (table->def, "EXP_TIME")-1;
  table->source_col    = table_find_col (table->def, "SOURCE")-1;
  table->uvw_col       = table_find_col (table->def, "UVW")-1;
  table->ot_col        = table_find_col (table->def, "OPT_TRAIN")-1;
  /* use "TELESCOPE" if "OPT_TRAIN" not there */
  if (table->ot_col<0)table->ot_col  = table_find_col (table->def, "TELESCOPE")-1;
  table->config_col    = table_find_col (table->def, "CONFIG")-1;
  table->batch_col     = table_find_col (table->def, "BATCH")-1;
  table->mu12_col      = table_find_col (table->def, "MU12")-1;
  table->mu22_col      = table_find_col (table->def, "MU22")-1;
  table->opd_col       = table_find_col (table->def, "OPD")-1;
  table->opd1_col      = table_find_col (table->def, "OPD1")-1;
  table->opd2_col      = table_find_col (table->def, "OPD2")-1;
  table->photom_col    = table_find_col (table->def, "PHOTOM")-1;
  table->corpower_col  = table_find_col (table->def, "CORPOWER")-1;
  table->snrp1_col     = table_find_col (table->def, "SNRP1")-1;
  table->snrp2_col     = table_find_col (table->def, "SNRP2")-1;
  table->weight_col    = table_find_col (table->def, "WEIGHT")-1;
} /* end  get_fringe_a_info */

