/* $Id: fringe_a.h,v 1.2 2001/11/26 12:52:49 csabet Exp $ */
#ifndef _FRINGE_A_H
#define _FRINGE_A_H
/*----------------------------------------------------------------------- */
/* Fringe_A data FITS binary tables routines                              */
/* Partially/fully calibrated data table for optical/IR interferometer    */
/*----------------------------------------------------------------------- */
#include "VLTItable.h"

typedef struct      /* table definition for fringe_a table */
{
  table_def *def;      /* basic table information */
  long   row;          /* last row (0-rel) read or written, -1=> none. */
  char   *date_obs;    /* observing date */
  double  framrate;    /* Frame rate (Hz) */
  double  fringspd;    /* Fringe speed (m/s) */
  double  maxwave;     /* maximum wavelength used (m) */
  double  minwave;     /* minimum wavelength used (m) */
  char   *filter;      /* filter name */
  int    time_col;     /* time column number (0-rel)*/
  int    timei_col;    /* averaging time column number */
  int    source_col;   /* source number column number */
  int    uvw_col;      /* uvw column number */
  int    ot_col;       /* Optical train numbers column number */
  int    config_col;   /* Configuration numbers column number */
  int    batch_col;    /* Batch index column number */
  int    mu12_col;     /* Vis1 column number */
  int    mu22_col;     /* Vis2 column number */
  int    opd_col;      /* opd offset column number */
  int    opd1_col;     /* opd derivative offset column number */
  int    opd2_col;     /* opd second derivative offset column number */
  int    photom_col;   /* average photometric values column number */
  int    corpower_col; /* Moments of the correlated power column number */
  int    snrp1_col;    /* Minimum SNR photometry 1 column number */
  int    snrp2_col;    /* Minimum SNR photometry 2 column number */
  int    weight_col;   /* data weight column numbers */
} fringe_a_table_def;

typedef struct      /* structure used for fringe_a row */
{
  double time;       /* central time of exposure */
  double uvw[3];     /* Spatial frequency coordinated for observations */
  float  timei;      /* data averaging time */
  float  mu12;       /* First interferometric visibility amplitude sq */
  float  mu22;       /* Second interferometric visibility amplitude sq */
  float  snrp1;      /* Minimum SNR photometry 1 */
  float  snrp2;      /* Minimum SNR photometry 2 */
  float  weight;     /* A reliability weight of this datum. */
  float  opd;        /* OPD offset (m)  */
  float  opd1;       /* OPD derivative offset (m/s)  */
  float  opd2;       /* OPD second derivative offset (m/s/s)  */
  float  photom[2];  /* Scan average photometric values */
  float  corpower[3];/* Moments of the correlated power (filtered I1-I2). */
  int    source;     /* source id */
  int    ot[2];      /* optical train numbers of array */
  int    config;     /* configuration number */
  int    batch;      /* batch number */
  table_row *raw;    /* generic table row */
} fringe_a_row;

/* public functions */


/* fringe_a_row functions */
fringe_a_row* make_fringe_a_row(fringe_a_table_def* table);
/*----------------------------------------------------------------------- */
/*   Make an fringe_a_row structure                                       */
/*   Inputs:                                                              */
/*      table     pointer to fringe_a table structure                     */
/*   Returns:                                                             */
/*      An fringe_a_row structure pointer ; NULL indicates failure        */
/*      Data pointers set to generic structure values                     */
/*----------------------------------------------------------------------- */

void kill_fringe_a_row (fringe_a_row *row);
/*----------------------------------------------------------------------- */
/*   Delete an fringe_a_row structure                                     */
/*   Inputs:                                                              */
/*      row    pointer to the fringe_a_row structure.                     */
/*----------------------------------------------------------------------- */

/* fringe_a_table functions */
fringe_a_table_def* 
make_fringe_a_table_def(fitsfile *fptr, char *date_obs, 
			float framrate, float fringspd, float maxwave,
			float  minwave, char *filter, int *ierr);
/*----------------------------------------------------------------------- */
/*   Make a new fringe_a_table in an open FITS file                       */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      date_obs  Date (yyyy-mm-dd) observations began                    */
/*      framrate  frame rate (Hz)                                         */
/*      fringspd  fringe speed (m/s)                                      */
/*      maxwave   max wave length (m)                                     */
/*      minwave   min wave length (m)                                     */
/*      filter    Filter being used                                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An fringe_a_table_def pointer; NULL indicates failure             */
/*----------------------------------------------------------------------- */

fringe_a_table_def* 
open_fringe_a_table(fitsfile *fptr, int *ierr);
/*----------------------------------------------------------------------- */
/*   Open an existing fringe_a_table in the active FITS file              */
/*   Fills fringe_a_geom and saves in fringe_a_table_def                  */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An fringe_a_table_def pointer; NULL indicates failure             */
/*----------------------------------------------------------------------- */

int read_fringe_a_table(fitsfile *fptr,  fringe_a_table_def* table, 
			    fringe_a_row *row, long row_number,
			    int *ierr);
/*----------------------------------------------------------------------- */
/*   Read the next row from an fringe_a_table                             */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row_number which row (1-rel) in the table                         */   
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      row       pointer to an extant row structure                      */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */

int write_fringe_a_table(fitsfile *fptr, fringe_a_table_def* table, 
			     fringe_a_row *row,  long row_number,
			     int *ierr);
/*----------------------------------------------------------------------- */
/*   Write  the next row to an fringe_a_table                             */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row       pointer to an extant row structure                      */
/*      row_number which row (1-rel) in the table                         */     
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */

int close_fringe_a_table(fitsfile *fptr, fringe_a_table_def* table, 
			     int *ierr);
/*----------------------------------------------------------------------- */
/*   Close an open fringe_a_table, associated structures are destroyed    */
/*   Note: this closes the full FITS file                                 */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */

void kill_fringe_a_table_def (fringe_a_table_def* table);
/*----------------------------------------------------------------------- */
/*   Delete an fringe_a_table structure                                   */
/*   Inputs:                                                              */
/*      table     table definitions structure from open or create         */
/*----------------------------------------------------------------------- */

#endif /* end FRINGE_A_H */
