/*******************************************************************/
/* VLTI VINCI DATA REDUCTION SOFTWARE (VNDRS)                      */
/*                                                                 */
/* FILE: vndrs.h                                                   */
/*                                                                 */
/* Contains definitions of constant parameters, types and          */
/* prototypes for all ANSI C functions.                            */
/*                                                                 */
/* Written by Markus Schoeller & Rainer Wilhelm, ESO, 2000         */
/* Modified by Pierre Kervella, ESO, Feb 2002                      */
/*******************************************************************/ 

/*
   Applicable documents:
   =====================

   [1]: LdV Data Processing algorithms
        VLT-SPE-ESO-15810-2112

   [2]: V. Coude du Foresto, S. Ridgway, J. M. Mariotti,
        'Deriving object visibilities from interferograms obtained
         with a fiber stellar interferometer'
        Astron. & Astrophys., Suppl. Ser. 121, pp. 379-392 (1997)

   [3]: D. Segransan, Ph.D. Thesis, 2001 (Universite de Grenoble)
        Application of wavelets to the analysis of interferometric data

   we changed heavily the interfaces from the document [1], especially
   the main data has now its own data structure, and all data inside
   is in double, not in long, as described in [1];
 */



/* --------------------------------------------------------------- */
/* COMPILER DIRECTIVES                                             */
/* --------------------------------------------------------------- */
#ifndef _VNDRS_H
#define _VNDRS_H 

/*
#define SECURITY_MODE
*/

/* Compiler directive. If it is set, the program will run slower 
   but more safe (dimension checks etc. are made.) Keep this 
   flag set as long as you developing the code. */

/* --------------------------------------------------------------- */
/* INCLUDES                                                        */
/* --------------------------------------------------------------- */

#include <math.h>
#include <time.h>
#include "fftw.h" /* for FFTW library */

/* --------------------------------------------------------------- */
/* MATHEMATICAL CONSTANTS                                          */
/* --------------------------------------------------------------- */
#define PI 3.141592653589793238e+00
/* value taken from MATHEMATICA, (c) Wolfram Research 
   Why this high precision?
   Example: Phase k*z with k=2*pi/lambda, lambda = 1e-06 m and z = 1km
   We should get a Phase of zero, as a multiple of 1e-06 fits exactly inside 1km. 
   (1) if we take pi = 3.141592654 --> Phase = 6.283185308000000e+09
              --> (cos(Phase),sin(Phase)) = (6.819186241461076e-01, 7.314280484385868e-01)
   (2) if we take pi = 3.141592653589793 --> Phase = 6.283185307179587e+09
              --> (cos(Phase),sin(Phase)) = (9.999999999993591e-01, 1.132200850406420e-06)
   The correct (cos,sin) pair would be (1,0) --> (2) is much better than (1) */
#define TWO_PI 6.283185307179586476e+00

/* --------------------------------------------------------------- */
/* VINCI DRS CONSTANTS                                             */
/* --------------------------------------------------------------- */

/* These are values which are defined as global variables in the FLUOR code
   and are in fact constants */

#define EPSILON_APODISATION_WINDOW 1.e-09
/* the small number that will be added to the apodisation window in order to
   avoid having zeros  in the window (introduced Jan 16, 01 since the
   division of P by the apodisation window lead to problems.) */
#define RADIUS 8
#define GENERAL_FILTER_ORDER 20
#define COMPUTE_MATRIX_BLOCKOUT 0.10
#define COMPUTE_MATRIX_TRANSITION 0.
#define COMPUTE_MATRIX_CENTRAL_FREQ 0.
#define COMPUTE_MATRIX_HALF_BANDWIDTH 80.

/* --------------------------------------------------------------- */
/* OTHER CONSTANTS ...                                             */
/* --------------------------------------------------------------- */
#define MEMORY_PROTOCOL_THRESHOLD 1000
/* when writing an ASCII memory protocol, only increments whose absolute
   value is larger than this threshold (in units of BYTES) are written
   to the protocol */

#define MAX_STRING_LENGTH 255
/* the maximum length of a string used in the code */

enum
{
	PASSED = 0,
	FAILED = 1
};

enum
{
	TRUE	= 0,
	FALSE	= 1
};

enum
{
	FULL_CORRECTION		= 0,
	PSEUDO_CORRECTION	= 1
};

typedef unsigned int VisCorrectType;	/*	Set to one of the following:
											FULL_CORRECTION, PSEUDO_CORRECTION */

typedef unsigned int StatusType;		/*	Set to one of the following:
											PASSED, FAILED */

/* --------------------------------------------------------------- */
/* TYPEDEFS                                                        */
/* --------------------------------------------------------------- */

/* Double number */
/* ------------- */
/* Within the VNDRS we use the 'double' type for floating-point numbers. */

/* Complex number */
/* -------------- */
/* We use the FFTW type definition of a complex number, hence:
   typedef struct { fftw_real re, im; } fftw_complex;
   while the type 'fftw_real' is 'double' by default.
   (see include file 'fftw.h') */
typedef fftw_complex complex;
 
/* Double vector */
/* ------------- */
typedef struct {
                 long dimension; /* the dimension of the 1-D vector,
                                    indices run from 0 to Dimension-1 */
                 double *data;   /* the 1-D field holding the components
                                    of the vector */
               } vector_double;

/* Complex vector */
/* -------------- */
typedef struct { 
                 long dimension; /* the dimension of the 1-D vector,
                                    indices run from 0 to Dimension-1 */
                 complex *data;  /* the 1-D field holding the components
                                    of the vector */
               } vector_complex;

/* Double matrix */
/* ------------- */
typedef struct {
                 long dimension_1; /* number of lines of the matrix */ 
                 long dimension_2; /* number of columns of the matrix */ 
                 double *data;     /* the matrix elements */  
               } matrix_double; 
 
/* Complex matrix */ 
/* -------------- */ 
typedef struct { 
                 long dimension_1; /* number of lines of the matrix */ 
                 long dimension_2; /* number of columns of the matrix */ 
                 complex *data;    /* the matrix elements */  
               } matrix_complex; 
 
/* Vector of double matrices */ 
/* ------------------------- */ 
typedef struct { 
                 long dimension; /* the number of matrices, 
                                    indices run from 0 to Dimension-1 */ 
                 matrix_double *data;   /* the 1-D field holding the matrices */ 
               } vector_matrices_double; 
 
/* Boolean */ 
/* ------- */ 
typedef long bool; 
 
/* Individual scan, Fourier spectrum and power spectrum */ 
/* ---------------------------------------------------- */ 
/* 
  please note that 'vector_double' and 'vector_complex' are structs that keeps the 
  vector's length and points to a double array 
*/ 
typedef vector_double         individual_scan; 

typedef vector_complex        fourier_spectrum; 
typedef vector_double         power_spectrum;

/*added by DS 04/07/2002*/ 
typedef matrix_complex        wavelet_complex_spectrum; 
typedef matrix_double         wavelet_power_spectrum; 

 
 
/* Full scan & Batch of full scans */ 
/* ------------------------------- */ 
/* 
   to organize the structure as below might not give the most efficient 
   code, but very likely the most organized 
 */ 
 
typedef struct 
{ 
    individual_scan *I1; 
    individual_scan *I2; 
    individual_scan *PA; 
    individual_scan *PB; 
    individual_scan *QL; 
} full_scan; 
/* A 'full scan' consists of 4 individual scans (the 2 interferometric signals I1 and I2, 
   and the 2 photometric signals PA and PB) */ 
 
 
typedef struct 
{ 
  double time; 
  double dl_offset;  
  double mu1_squared;  
  double mu2_squared;  
} scan_data; 
/* A 'scan_data' stores the informations related to 1 scan */ 
 
typedef struct 
{ 
    size_t number_of_scans; 
    full_scan **fullscans; 
    scan_data *scan_data; 
} batch; 
/* A 'batch' is an array of full scans */ 
 
/* Full powerspectrum */ 
/* ------------------ */ 
/* 
   this structure will hold the powerspectra computed by the DRS 
 */ 
 
typedef struct 
{ 
    power_spectrum *I1_pow; 
    power_spectrum *I2_pow; 
} full_pow; 
 
typedef struct 
{ 
    size_t number_of_pows; 
    full_pow **fullpows; 
} powbatch; 
 
 
/* Memory struct (contains all information on current memory usage) */ 
/* ---------------------------------------------------------------- */ 
typedef struct 
{ 
    long currently_allocated; 
    /* currently used memory is stored in the global variable 'vndrs_memory.currently allocated' */ 
 
    long delta; 
    /* when writing an ASCII memory usage protocol, not every allocation or free process 
       is written to the protocol. Changes in total memory are added up in the 
       variable "vndrs_memory.delta". Only if this value is larger than the treshold 
       (defined above as a constant) the protocol is updated. */ 
 
    FILE *file_pointer; 
    /* file pointer to the memory protocol file */ 
 
    /* COUNTERS THAT KEEP TRACK OF HOW MANY TIMES SPECIFIC ALLOCATION AND 
       FREE ROUTINES HAVE BEEN CALLED */ 
    long calloc_vector_double_calls; 
    long calloc_vector_complex_calls; 
    long calloc_matrix_double_calls; 
    long calloc_matrix_complex_calls; 
    long calloc_full_scan_calls; 
    long calloc_full_pow_calls; 
    long calloc_batch_calls; 
    long calloc_powbatch_calls; 
} memory_usage; 
 
/* Protocol struct (contains information on the runtime protocol) */ 
/* -------------------------------------------------------------- */ 
typedef struct 
{ 
    char mode[MAX_STRING_LENGTH]; 
    /* can be either 'NONE', 'SCREEN' or 'DISK' */ 
 
    FILE *file_pointer; 
    /* file pointer to the runtime protocol */ 
 
    char text[MAX_STRING_LENGTH]; 
    /* the string that shall be written to the runtime protocol */  
} protocol; 
 
  
/* &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& */ 
/* &&                                                           && */ 
/* &&         FUNCTION PROTOTYPES (REQUIRED IN ANSI C)          && */ 
/* &&                                                           && */ 
/* &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& */ 
 
 
/* -------------------------------------------------------------- */ 
/* MAIN FUNCTIONS FOR FRINGE PROCESSING                           */ 
/* -------------------------------------------------------------- */ 

/*
    Complete data processing report. pkervell 22-Oct-02.
*/

void vndrs_complete_report(double mu1_fft_avg, double mu1_wl_avg,
                           double mu1_fft_avg_bs, double mu1_wl_avg_bs,
                           double mu1_fft_avg_noncentered, double mu1_wl_avg_noncentered,
                           double mu1_fft_avg_mil, double mu1_wl_avg_mil,
                           double unweighted_mu1_fft, double unweighted_mu1_wl,
                           double avg_background_fourier, double avg_background_wavelets,
                           double background_fourier, double background_wavelets,
                           double slope_background_fft, double slope_background_wl,
                           double Fringe_speed, double Frame_rate, long N_frames_cut, long N_frames_precut,
                           double Ci2_mu1_fft_avg_min, double Ci2_mu1_fft_avg_max,
                           double Ci2_mu1_wl_avg_min, double Ci2_mu1_wl_avg_max,
                           double Ci3_mu1_fft_avg_min, double Ci3_mu1_fft_avg_max,
                           double Ci3_mu1_wl_avg_min, double Ci3_mu1_wl_avg_max,
                           double jack_fft, double jack_wl,
                           double simple_error_fft, double simple_error_wl,
                           double Wavenum_interval,
                           vector_double *Wavenum_limits_fft, vector_double *Wavenum_limits_wl,
                           double Wavenum_min_lowband_fft, double Wavenum_max_lowband_fft,
                           double Wavenum_min_lowband_wl, double Wavenum_max_lowband_wl,
                           double Wavenum_min_hi_band_fft, double Wavenum_max_hi_band_fft,
                           double Wavenum_min_hi_band_wl, double Wavenum_max_hi_band_wl,
                           double wavenum_fft_min, double wavenum_fft_max,
                           double wavenum_wl_min, double wavenum_wl_max,
                           double skew_mean_fft, double skew_mean_wl,
                           double skew_median_fft, double skew_median_wl,
                           double kurtosis_mean_fft, double kurtosis_mean_wl,
                           double kurtosis_median_fft, double kurtosis_median_wl,
                           double Wavenum_theo, double weighted_peak_pos, double weighted_peak_pos_wl,
                           double weighted_peak_pos_fft_centered, double weighted_peak_pos_wl_centered,
                           vector_double *Off_offsets_Mean, vector_double *Off_offsets_Stddev,
                           vector_double *A_Mean, vector_double *B_Mean, vector_double *On_source_Mean,
                           double on_src_bg_I1, double on_src_bg_I2, double on_src_bg_PA, double on_src_bg_PB,
                           double SNR_I1, double SNR_I2, double SNR_PA, double SNR_PB,
                           double avg_photom, double average_unbalance,
                           double kappa1A, double kappa1B, double kappa2A, double kappa2B,
                           double kappa1A_err, double kappa1B_err, double kappa2A_err, double kappa2B_err,
                           double relative_kappa_norm_var,
                           vector_double *splitting_ratio_A, vector_double *splitting_ratio_B,
                           vector_double *off_source_spectrum_slopes, vector_double *number_of_pickups,
                           vector_double *gains, vector_double *RONs, long flag_remove,
                           long N_scans_processed, long N_scans_on,
                           vector_double *removed_indices, long numberOfScansRejected, long n_sig_reject,
                           long scans_rejected_0, long scans_rejected_1, long scans_rejected_2,
                           long scans_rejected_3, long scans_rejected_4,
                           long scans_rejected_5, long scans_rejected_6,
                           long scans_rejected_7, long scans_rejected_8,
                           double mu1_fft_stdev_avg, double Ci_mu1_fft_avg_min, double Ci_mu1_fft_avg_max,
                           double mu1_wl_stdev_avg, double Ci_mu1_wl_avg_min, double Ci_mu1_wl_avg_max,
                           double mu1_fft_med, double Ci_mu1_fft_med_min, double Ci_mu1_fft_med_max,
                           double mu1_wl_med, double Ci_mu1_wl_med_min, double Ci_mu1_wl_med_max,
                           double total_weight_batch, double time_batch, double baseline_batch,
                           double azimuth_batch, double u_batch, double v_batch, int Autotest_case,
                           vector_double *histogram_scale_fft, vector_double *histogram_fft,
                           vector_double *histogram_scale_wl, vector_double *histogram_wl,
                           double peak_pos_mean, double peak_pos_sigma, double peak_width_avg, double peak_width_sigma,
                           double altitude_batch_obj, double azimuth_batch_obj,
                           double wavenum_min_halfmax_fft, double wavenum_max_halfmax_fft,
                           double wavenum_min_halfmax_wl, double wavenum_max_halfmax_wl,
                           int ionic_in, int select_scans,
                           int *warnings_flag, int *reject_flag /*out*/);
                           
/*
    Display and saving of the photometric values of the batch.
    pkervell 22-Oct-02
*/

void vndrs_photometry_info(vector_double *Off_offsets_Stddev,
                           vector_double *Off_offsets_Mean,
                           vector_double *A_Mean,
                           vector_double *B_Mean,
                           vector_double *On_source_Mean,
                           double SNR_I1, double SNR_I2,
                           double SNR_PA, double SNR_PB);

/*
    Verification of the processed data, emission of warnings - pkervell 20-Oct-02
*/

void vndrs_warnings(long N_scans_processed, long N_scans_on, double SNR_PA, double SNR_PB, double avg_photom,
    double mu1_fft_avg_mil, double mu1_fft_avg, double mu1_wl_avg_mil, double mu1_wl_avg,
    double skew_mean_fft, double skew_mean_wl, double weighted_peak_pos, double Wavenum_theo, double average_unbalance,
    vector_double *number_of_pickups, vector_double *off_source_spectrum_slopes, double Frame_rate, double N_frames_precut,
    double slope_background_fft, double slope_background_wl, double N_frames_cut,
    double mu1_fft_med, double mu1_fft_stdev_avg, double mu1_wl_med, double mu1_wl_stdev_avg,
    vector_double *gains, vector_double *RONs, double kappa1A, double kappa1B, double kappa2A, double kappa2B,
    int Autotest_case, vector_double *splitting_ratio_A, vector_double *splitting_ratio_B, int ionic_in,
    int *warnings_flag, int *reject_flag); /* out */
    
/*
    Computation of the gains of the LISA camera pixels - pkervell 19-Oct-02
*/
void vndrs_compute_camera_gains(full_scan *Avg_power_spectra_off,
                                full_scan *Avg_power_spectra_on,
                                vector_double *gains, /* out */
                                vector_double *RONs); /* out */

/*
    Automated verification of the readout noise spectrum - pkervell 18-Oct-02.
*/
void vndrs_check_readout_noise_spectrum(full_scan *Avg_power_spectra_off,
                                        vector_double *off_source_spectrum_slopes,
                                        vector_double *number_of_pickups);

/*
    This procedure recenters the PSD peaks using the Pixel_offset provided by the user
    (originally computed from the wavelets cetroiding in the PSD). 
    pkervell - 15-Oct-02
*/
void vndrs_center_PSD_peak(matrix_double *PSDs,
                           vector_double *Flag_processed,
                           long min_pixel_lowband, /* minimum pixel in the PSD to consider in the centering */
                           double Theoretical_position, /* in PSD pixels */
                           vector_double *Measured_positions); /* in PSD pixels */

/*
    Integration of the background corrected PSD between specified limits. This is quite similar to the 
    procedure vndrs_integrate_fringe_power, but does not ask for a background and accepts normalized
    interferogram spectrum.
    pkervell - 15-Oct-02.
*/
double vndrs_integrate_PSD(vector_double *PSD,
                           double wavenum_interval,
                           double wavenum_min,
                           double wavenum_max);

/*
    Estimation of the peak limits from the corrected power spectrum - pkervell 15-Oct-02.
*/
void vndrs_estimate_peak_limits(vector_double *Pow_spec_avg,
                                double Wavenum_theo,
                                double Wavenum_interval,
                                double *wavenum_min, /* output */
                                double *wavenum_max, /* output */
                                double *wavenum_min_halfmax, /* output */
                                double *wavenum_max_halfmax); /* output */


/*
    Removal of the background offsets scan by scan using the vectors of computed backgrounds.
    Only PA and PB backgrounds are removed scan by scan (they are the only important in the
    calibration), while the I1 and I2 backgrounds are removed using the off-source values.
    - pkervell 10-Oct-02.
    The batch input is replaced.
*/
void vndrs_remove_backgrounds_per_scan(batch *Batch_in_out,
                          vector_double *Background_I1,
                          vector_double *Background_I2,
                          vector_double *Background_PA,
                          vector_double *Background_PB);

/*
    Removal of the background offsets (globally...) - pkervell 04-Oct-73.
    The batch input is replaced.
*/
void vndrs_remove_offsets(batch *Batch_in_out,
                          vector_double *Offsets);


/* 
    Computes the frequency scale (in Hertz, or whatever unit is Frame_rate) of the PSD -
    the resulting vector contains only the first half of the frequency scale, up to the Nyquist freq,
    that is supposed to be the size of the input vector Freq_scale.
    pkervell - 03-Oct-02
*/
void vndrs_compute_freq_scale(double Frame_rate,
                              vector_double *Freq_scale);


/* Computes the average power spectra of the five signals (incl. QL) over one batch and
stores it in a fullscan structure. pkervell 03-Oct-02 */

void vndrs_compute_average_power_spectra(batch *Batch_in,
                                         full_scan *Power_spectra);

/*
    The function below selects the scans based on their visibility. 
    
    This is mandatory, because we are facing something like a Cauchy distribution,
    one of the statistical worst nightmares. To go back to something like a gaussian
    distribution, we have to remove the extreme outliers (3 or 4 sigmas for example).
*/

double vndrs_eliminate_visibility_outliers(vector_double *mu2_raw,      /* input */
                                        double N_sigmas,
                                        vector_double *Flag_processed, /* output */
                                        vector_double *removed_indices); /* output */


/*
    The function below selects the scans based on their photometry. The very high
    photometry scans are removed from the final sample, in order to avoid biasing problems.
    The limit is computed inside in terms of number of times the median distance to the
    median (the distribution is very skewed).
*/

long vndrs_eliminate_extreme_scans(vector_double *signal_p,       /* input */
                                   double N_med_dist,
                                   vector_double *Flag_processed, /* output */
                                   vector_double *removed_indices); /* output */

/*
    Compute and removes the computed offsets from the on-source batch - pkervell 10-Oct-02
*/
void vndrs_remove_computed_offsets(batch *Batch_off, /* input/output */
                               batch *Batch_A, /* input/output */
                               batch *Batch_B, /* input/output */
                               batch *Batch_on, /* input */
                               matrix_double *Matrix, /*input*/
                               vector_double *Off_offsets); /*out*/

/*
    Independant estimation of the background offset level (not directly from off sourcce batch)
    pkervell - 24-Sep-02
*/
void vndrs_evaluate_off_source_background(batch *Batch_on,
                                          batch *Batch_A,
                                          batch *Batch_B,
                                          batch *Batch_off,
                                          matrix_double *Matrix,
                                          vector_double *Background_PA_off, /* fluctuations of the off-src */
                                          vector_double *Background_PB_off,
                                          vector_double *Background_I1, /*out*/
                                          vector_double *Background_I2, /*out*/
                                          vector_double *Background_PA, /* fluctuations of the on-src */
                                          vector_double *Background_PB);
/*
     Computation of the synthetic photometric signals from the
     interferometric outputs.
     pkervell 17-Sep-02
*/

void vndrs_prepare_synthetic_photom(vector_double *I1_cal,
                                    vector_double *I2_cal,
                                    matrix_double *Matrix_coefs,
                                    /* outputs */
                                    vector_double *PA_synth,
                                    vector_double *PB_synth);

/*
     Selection of type and computation of the matrix.
     Introduced to make the main procedure code more compact.
     pkervell 13-Sep-02
*/
void vndrs_select_matrix(batch *Batch_A,
                         batch *Batch_B,
                         batch *Batch_on,
                         vector_double *Off_offsets,
                         vector_double *WA_mean,
                         vector_double *WB_mean,
                         int ionic_in,
                         matrix_double *Average_Matrix,
                         matrix_double *Standard_dev_Matrix,
                         vector_double *Relative_MSE,
                         int verbose_on);

/* just to display and write the header on screen and in the log file */

void vndrs_display_and_write_header(void);

/* Writes the power spectra and Wiener filters in log file - pkervell - 24-Sep-02 */
void vndrs_write_power_spectra_to_file(long dim_pow,
                                       vector_double *Freq_scale_avg_spectra,
                                       full_scan *Avg_power_spectra_on,
                                       full_scan *Avg_power_spectra_off,
                                       vector_double *WA_mean,
                                       vector_double *WB_mean,
                                       vector_double *Freq_scale_fringes,
                                       vector_double *Pow_spec_avg_I1,
                                       vector_double *Phase_avg_I1,
                                       vector_double *Pow_spec_avg_I1_no_centering,
                                       vector_double *Phase_avg_I1_no_centering,
                                       vector_double *Pow_spec_avg_I1_off,
                                       vector_double *Background_PSD,
                                       vector_double *Pow_spec_avg_I1_corrected,
                                       vector_double *WL_spec_avg_I1_vector_corrected,
                                       vector_double *WL_spec_avg_I1_vector_no_centering,
                                       vector_double *WL_spectrum_vector_off,
                                       vector_double *Background_PSD_wl_vector);



/*
     Removal of the background from the processed combined PSD
     pkervell 09-Sep-02
*/

void vndrs_remove_psd_background(long N_frames,
                                vector_double *Pow_spec_avg,
                                double Wavenum_min_lowband,
                                double Wavenum_max_lowband,
                                double Wavenum_min_hi_band,
                                double Wavenum_max_hi_band,
                                int debias,
                                double slope_limit,
                                double Frame_rate,
                                double Fringe_speed,
                                /* outputs */
                                double *slope_background,
                                vector_double *Background_PSD,
                                vector_double *Pow_spec_avg_corrected);

/*
    Procedure to check if we have really fringes in our scan. Multiple checks are done to
    ensure that the signal has all the characteristics of a fringe packet. It should return a
    code corresponding to the reason for rejection.
    added pkervell 07-Sep-02
*/
double vndrs_check_fringes(/* inputs */
                           matrix_double *norm_t_wavelets,
                           long N_frames_cut,
                           double Frame_rate,
                           double Fringe_speed,
                           double fringes_length,
                           double Wavenum_theo,
                           double Wavenum_min,
                           double Wavenum_max,
                           double Wavenum_min_out,
                           double Wavenum_max_out,                                
                           double Wavenum_interval_cut,
                           /* output */
                           vector_double *fringes_centering_outputs);                            
/*
    Detection of fringes and centering
    added pkervell 01-Sep-02
    This function detects and centers the fringes on one scan, and outputs the
    cut scan and the detection flag.
*/

double vndrs_detect_center_fringes(/* inputs */
                                full_scan *Scan_on_precut,
                                double Apod_blockout,
                                double Apod_transition,
                                long N_frames_cut,
                                double param_wl,  /* usually 17.0 fringes */
                                double fringes_length,
                                int ionic_in,
                                double Wavenum_min,
                                double Wavenum_max,
                                double Wavenum_min_out,
                                double Wavenum_max_out,                                
                                double Sample_interval_precut,
                                double Wavenum_interval_precut,
                                /* outputs */
                                full_scan *Scan_on_cut,
                                double *fringes_off_center
                                );

/*
    Removes the slope of a vector. The N_frames_avg values of the
    vector are considered to evaluate the start and end points.
*/

void vndrs_remove_vector_slope(vector_double *input_vector, /* input vector, REPLACED */
                          long N_frames_avg,
                          vector_double *subtracted_vector); /* the subtracted vector (linear slope) */
                                
/*
    added pkervell 01-Sep-02
    Computation of the Quicklook signal used for fringe detection.
    It is slightly different from the quicklook used at the telescope in that there is no
    division by the photometry.
*/
    
void vndrs_quicklook(full_scan *Scan_on_precut, /* input */
                     double Apod_blockout,
                     double Apod_transition,
                     int ionic_in,
                     double Wavenum_min,
                     double Wavenum_max,
                     double Wavenum_min_out,
                     double Wavenum_max_out,
                     double Sample_interval_precut,
                      /* outputs */
                     individual_scan *quicklook_signal,
                     individual_scan *quicklook_noise);

/* 
    Core function to reduce the LdV non dispersed data. 
    Described in [1], section 2.1 
 */
 
void vndrs_compute_visibilities ( 
                                 /* --------------- */ 
                                 /* input variables */ 
                                 /* --------------- */  
                                 batch *Batch_on, 
                                 batch *Batch_off, 
                                 batch *Batch_A, 
                                 batch *Batch_B, 
                                 int Choice_matrix, 
                                 double Frame_rate, 
                                 double Fringe_speed, 
                                 double Shape_factor,
				 long N_frames_cut,
				 long fringes_length,
                                 double Wavenum_theo,
                                 vector_double *Wavenum_limits_fft,  /* vector of 2 */ 
                                 vector_double *Wavenum_limits_wl,  /* vector of 2 */ 
                                 vector_double *Wavenum_limits_QL,  /* vector of 2 */ 
				 bool process_undetected,
				 int use_mean_wiener,
				 int verbose_on,
				 int debias,
				 int depiston,
				 int median_output,
				 int background_per_scan,
				 float center_wl_threshold,
				 int select_scans,
				 float max_jump,
                                 long Max_average, 
                                 bool two_channels_only,  /* introduced to distinguish between full 
                                                             and pseudo correction */ 
                                 float Rejection_threshold, 
                                 double Apod_blockout, 
                                 double Apod_transition, 
                                 double Mult_corr, 
                                 long B,
				 double lat,  /* coordinates of Paranal, added pkervell 15-apr-02 */
				 double longi,
				 double RA_object, /* coordinates of the object */
				 double Dec_object,
				 double t1x, double t1y, double t1z, /* coordinates of the telescopes (in meters) */
				 double t2x, double t2y, double t2z,
				 int ionic_in,
				 /* --------------------------*/ 
                                 /* input or output variables */ 
                                 /* --------------------------*/ 
                                 matrix_double *Matrix,             /* 2x2 matrix */ 
                                 /* ---------------- */ 
                                 /* output variables */ 
                                 /* ---------------- */ 
                                 vector_double *Relative_MSE,       /* vector of 2 doubles */ 
                                 batch *Batch_on_processed, 
                                 batch *Batch_off_processed, 
                                 long *N_scans_processed, 
                                 long *numberOfScansRejected, 
                                 vector_double *Mu1_squared, 
                                 vector_double *Mu2_squared, 
				 double *time_batch,
				 double *baseline_batch,
				 double *azimuth_batch,
				 double *u_batch,
				 double *v_batch,
                                 double *altitude_src,
                                 double *azimuth_src,
				 double *Mu1_batch_wl, double *ErrMu1_batch_min_wl, double *ErrMu1_batch_max_wl,
                                 double *ErrMu1_batch_wl,
				 double *Mu2_batch_wl, double *ErrMu2_batch_min_wl, double *ErrMu2_batch_max_wl,
				 double *Mu1_batch_fft, double *ErrMu1_batch_min_fft, double *ErrMu1_batch_max_fft,
                                 double *ErrMu1_batch_fft,
				 double *Mu2_batch_fft, double *ErrMu2_batch_min_fft, double *ErrMu2_batch_max_fft,
                                 double *background_fourier, double *background_wavelets,
                                 double *background_fourier_med, double *background_wavelets_med,
                                 double *effective_wavelength, double *effective_wavelength_wl, double *effective_wavelength_error,
                                 double *Mu1_skew_fft, double *Mu1_skew_wl,
                                 double *Photom_power, double *total_weight_batch,
                                 int *warnings_flag, int *reject_flag); 
 
/*
  Cuts a series of scans according to the delay line offsets that have
  been saved together with the scans. pkervell 13-mar-2002
*/

void vndrs_trim_batch (
                           /* --------------- */
                           /* input variables */
                           /* --------------- */
                           batch *Batch_in,
			   /* Input batch to trim */
			   double Fringe_speed,
			   double Frame_rate,
			   int center_fringes,
			   /* boolean to center or not the fringes */
			   long N_frames_cut,
			   /* Number of frames to keep around the fringes */
                           /* series of scans */
                           /* ---------------- */
                           /* output variables */
                           /* ---------------- */
                           batch *Batch_out
			   );

/*
  Cuts an individual scan according to the position of the fringes, given in pixels from
  the left side of the scan. pkervell 18-mar-2002
*/

void vndrs_trim_scan (
                           /* --------------- */
                           /* input variables */
                           /* --------------- */
                           individual_scan *scan,
			   long N_frames_cut,
			   /* Number of frames to keep around the fringes */
			   long fringes_position,
			   /* Fringes position relatively to the center, in PIXELS */
                           /* ---------------- */
                           /* output variables */
                           /* ---------------- */
                           individual_scan *scan_out
			   );

/*
   Cuts a vector_double from index_min to index_max. The output variable scan_out is NOT
   allocated inside the function.
   pkervell 09-Sep-02
*/

void vndrs_trim_scan_segment (
                           /* --------------- */
                           /* input variables */
                           /* --------------- */
                           individual_scan *scan,
			   long index_init,
			   long index_end,
			   /* index of begin and end of the trim, in PIXELS */
                           /* ---------------- */
                           /* output variables */
                           /* ---------------- */
                           individual_scan *scan_out
			   );


/* The following routine sets the coefficients of the transfer matrix --either by computing 
   them (--> fitting procedure) or by setting them to the user-preset values */ 
void vndrs_get_transfer_matrix(/* --------------- */ 
                               /* input variables */ 
                               /* --------------- */  
                               batch *Batch_A, 
                               /* series of scans obtained with Beam_A only */ 
                               batch *Batch_B, 
                               /* series of scans obtained with Beam_B only */ 
                               vector_double *Off_offsets,
                               vector_double *WA_mean,
                               vector_double *WB_mean,
 			       int ionic_in,
			       /* when the IONIC beam combiner is used */
                               /* -------------------------- */ 
                               /* input and output variables */ 
                               /* -------------------------- */  
                               matrix_double *Matrix, 
                               /* The 2 x 2 transfer matrix between the photometric and interferometric signals 
                                  (both measured at photometers (see [2], Eq. (19) on p. 383): 
 
                                          [ I1 ]   [ kappa_1A  kappa_1B ]   [ PA ] 
                                          |    | = |                    | * |    | 
                                          [ I2 ]   [ kappa_2A  kappa_2B ]   [ PB ] 
			       */ 
                               matrix_double *Standard_dev,  /* added pkervell 06-Sep-02 */
                               vector_double *Relative_MSE,        /* 2 doubles vector */ 
			       int verbose_on
			       ); 
 
/* 
    Computes the transfer function between the photometric and 
    interferometric signals of LdV. 
    Described in [1], section 2.6 
 */ 
void vndrs_compute_matrix ( 
                            
                           /* --------------- */ 
                           /* input variables */ 
                           /* --------------- */ 
                           batch *Batch_A, 
                           batch *Batch_B, 
                           /* series of scans obtained with beam B only (PA = 0) */
                           vector_double *Off_offsets,
                           vector_double *WA_mean,
                           vector_double *WB_mean,
			   int ionic_in,
                           /* ---------------- */ 
                           /* output variables */ 
                           /* ---------------- */ 
                           /* the next two entries are changed from the original concept, 
                              where they were defined as vectors of matrices */ 
                           matrix_double *Variances,    /* 2x2 matrix */ 
                           matrix_double *Matrix,       /* 2x2 matrix */ 
                           matrix_double *Standard_dev, /* 2x2 matrix */
                           vector_double *Relative_MSE,  /* vector of 2 */ 
			   int verbose_on
                          ); 

                          /* Following declaration added 19-apr-2002 by pkervell */
void vndrs_compute_matrix_least_squares(
                           /* --------------- */
                           /* input variables */
                           /* --------------- */
                           batch *Batch_off,
                           /* series of scans obtained with beam A and beam B off */
                           batch *Batch_A,
                           /* series of scans obtained with beam A only */
                           batch *Batch_B,
                           /* series of scans obtained with beam B only */
                           batch *Batch_on,
                           /* series of scans obtained on source */
			   int ionic_in,
                           /* ---------------- */
                           /* output variables */
                           /* ---------------- */
                           matrix_double *Matrix,
                           /* The transfer matrix that is computed in this routine (see comments below) */
                           matrix_double *Standard_errors,
                           /* The standard errors of the kappas (2x2 matrix) */
                           int *Autotest_case,
                           /* Flag to indicate that the matrix was computed using the Autotest case */
			   int verbose_on);
                           

                          /* Following declaration added 05-mar-2002 by pkervell */
void vndrs_compute_matrix_per_scan (
                           /* --------------- */
                           /* input variables */
                           /* --------------- */
                           full_scan *Scan_On,
                           /* scans obtained on source */
                           double Frame_rate,
                           /* fringe aquisition frequency (unit Hz) */
			   matrix_double *Average_Matrix,   /* added pkervell 08-mar-2002 */
			   /* The average matrix of on-source scans used in case the linfit2 does not converge */
                           /* ---------------- */
                           /* output variables */
                           /* ---------------- */
                           /* the next two entries are changed from the original concept,
                              where they were defined as vectors of matrices */
                           matrix_double *Matrix,
                           /* The transfer matrix that is computed in this routine (see comments below) */
			   int verbose_on
                          );

void vndrs_get_transfer_matrix_coefficients(/* ----------------------------- */ 
                                            /* input variables               */ 
                                            /* ----------------------------- */  
                                            const matrix_double *TransferMatrix, 
                                            /* ----------------------------- */ 
                                            /* output variables              */ 
                                            /* ----------------------------- */ 
                                            double *kappa_1A, 
                                            double *kappa_1B, 
                                            double *kappa_2A, 
                                            double *kappa_2B 
                                           ); 
 
/* 
    Computes the average matrix and the standard deviation from 
    an array of 2x2 matrices. 
    Described in [1], section 2.3 
 
    Unlike described in [1], this routine will only average one 
    matrix, and thus has to be called twice. We also change the 
    name accordingly. 
 */ 
matrix_double *vndrs_average_matrix ( /* returns 2x2 matrix */ 
                                      /* --------------- */ 
                                      /* input variables */ 
                                      /* --------------- */ 
                                      vector_matrices_double *Matrices  /* vector of 2x2 matrices */ 
                                    ); 
 
/* 
    Performs a full photometric correction of the data. 
    Described in [1], section 2.14 
 */ 
void vndrs_full_correction (/* --------------- */ 
                            /* input variables */ 
                            /* --------------- */ 
                            batch *Batch_off, 
                            full_scan *Master_spectrum_off,
                            /* Avg power spectra of the detector noise only - pkervell 11-Oct-02 */
                            full_scan *Scan_on, 
			    power_spectrum *WA_mean,   /* added pkervell 14-mar-02 */
			    power_spectrum *WB_mean,   /* added pkervell 14-mar-02 */
                            vector_double *Background_PA, /* updated if too low fluxes are found */
                            vector_double *Background_PB,
                            matrix_double *Matrix_coefs,  /* 2x2 matrix */ 
                            /* 2 x 1 vector: lower Wavenum_limit, higher Wavenum_limit */
                            double Frame_rate, 
                            double Fringe_speed, 
			    double fringes_length,
                            double Rejection_threshold, 
                            double Apod_blockout, 
                            double Apod_transition, 
                            double Mult_corr,
			    int debias,
			    int ionic_in,
			    int verbose_on,
                            /* ---------------- */ 
                            /* output variables */ 
                            /* ---------------- */ 
                            double *Min_SNR_P,
                            double *unbalance_coef,
                            full_scan *Scan_on_processed, 
                            full_scan *Scan_off_processed, 
                            double *Processed_flag 
                           ); 
 
/* 
    Described in [1], section 2.16 
 */ 
 
void vndrs_process_off ( 
                        /* --------------- */ 
                        /* input variables */ 
                        /* --------------- */ 
                        full_scan *Master_scan_off, 
                        individual_scan *Photom_I1,
                        individual_scan *PP1_apod,
                        individual_scan *Photom_I2,
                        individual_scan *PP2_apod,
                        vector_double *Apodisation_filter,
                        /* ---------------- */ 
                        /* output variables */ 
                        /* ---------------- */ 
                        individual_scan *I1_off_processed,
                        individual_scan *I2_off_processed,
			int verbose_on
                       ); 

/*
    Creation of the master off-source scan from the master offsrc spectrum
*/
void vndrs_create_master_off_source_scan(full_scan *Master_spectrum_off, /* in*/
                                         full_scan *Master_scan_off,/* out*/
                                         int verbose_on); 

double vndrs_integrate_fringe_power ( 
                                     /* --------------- */ 
                                     /* input variables */ 
                                     /* --------------- */ 
                                     individual_scan *Scan_on, 
                                     individual_scan *Scan_off, 
                                     double Wavenumber_interval, 
                                     double Sigma_min, 
                                     double Sigma_max, 
                                     int verbose_on,
                                     power_spectrum *Sp_power_cal, /* added for checking calib - pkervell 02-Sep-02 */
                                     vector_double *Real_part,
                                     vector_double *Imaginary_part,
                                     vector_double *Phase,
                                     power_spectrum *Sp_power_off_cal); 
 

/*
  Wavelet transform based computation of mu^2:
  Integrates the power spectral density of the fringe 
  signal in the processed scan, and normalizes it to 
  return the squared coherence factors Mu_squared for 
  each interferometric channel. 
 */ 
double vndrs_integrate_wavelets_power(matrix_double *norm_t_wavelets,
				      double Wavenum_interval,
				      double Wavenum_limits_low,
				      double Wavenum_limits_high,
				      double fringes_length,
				      int verbose_on);


/* 
    Described in [1], section 2.9 
 */ 
double vndrs_area_power_dark ( 
                              /* --------------- */ 
                              /* input variables */ 
                              /* --------------- */ 
                              power_spectrum *Spectral_power, 
                              power_spectrum *Dark_spectral_power, 
                              double Freq_interval, 
                              double Fmin, 
                              double Fmax,
			      int verbose_on
                             ); 
 
 double vndrs_area_power_dark_no_background (
                              /* --------------- */
                              /* input variables */
                              /* --------------- */
                              power_spectrum *Spectral_power,
                              double Freq_interval,
                              double Fmin,
                              double Fmax,
			      int verbose_on
                             );

/*
    Special version of the apodisation window generator, designed to accept
    parameters more adequate for operation on the fringes themselves and not
    on the whole scan. It computes an apodisation window centered on the
    scan and letting the 'width' central pixels completely through. The transition
    is outside of the 'width' pixels - pkervell 04-Oct-02
*/
individual_scan *vndrs_apodisation_center(
                                           /* --------------- */
                                           /* input variables */
                                           /* --------------- */
                                           long Signal_size, /* in number of frames */
                                           long Width,     /* in number of frames */
                                           long Transition /* in number of frames */
                                          );
                             
/* 
    Described in [1], section 2.20 
 */ 
individual_scan *vndrs_apodisation_window ( 
                                           /* --------------- */ 
                                           /* input variables */ 
                                           /* --------------- */ 
                                           long Signal_size, 
                                           double Blockout, 
                                           double Transmission 
                                          ); 
 
/* 
    Computes the area of an array x(T) between Tmin and Tmax. 
    The area is computed by the trapeze method between the 
    fractional indices corresponding to the boundaries. The 
    value of x(t) is linearly interpolated between Tmin and Tmax. 
    The abscissa origin is T0=0. By default, dt=1. This function 
    works even when Tmax-Tmin<1. 
    Described in [1], section 2.10 
 */ 
double vndrs_area ( 
                   /* --------------- */ 
                   /* input variables */ 
                   /* --------------- */ 
                   vector_double *Array, 
                   double Dt, 
                   double Tmin, 
                   double Tmax 
                  ); 
 
/* 
    Described in [1], section 2.11 
 */ 
double vndrs_linear_interp ( 
                            /* --------------- */ 
                            /* input variables */ 
                            /* --------------- */ 
                            vector_double *Array, 
                            double Tinterp 
                           ); 
 
/*
    Outputs the linear least squares fit regression coefficients a and b so that
    we have the model y = a.x + b. It also returns the correlation coefficient on the fit.
    This last parameter varies from 0 to 1 and a value of 1 indicates a perfect fit.
    pkervell 28-Sep-02.
*/

void vndrs_computes_least_squares(vector_double *x,
                                  vector_double *y,
                                  double *slope,
                                  double *zero_point,
                                  double *corr_coef);

/* 
    This function computes the proportionality factor between 
    two sets of data. 
    Described in [1], section 2.18 
 */ 
double vndrs_linfit ( 
                     /* --------------- */ 
                     /* input variables */ 
                     /* --------------- */ 
                     vector_double *Signal_1, 
                     vector_double *Signal_2 
                    ); 
 
/* 
    This funtion computes the coefficients a1 and a2 which give the 
    best fit between Signal and a1*Signal1+a2*Signal2 (least square). 
    Described in [1], section 2.19 
 */ 
void vndrs_linfit_2 ( 
                     /* --------------- */ 
                     /* input variables */ 
                     /* --------------- */ 
                     individual_scan *Signal, 
                     individual_scan *Signal1, 
                     individual_scan *Signal2, 
                     /* ---------------- */ 
                     /* output variables */ 
                     /* ---------------- */ 
                     double *a1, 
                     double *a2, 
                     matrix_double *Covariances,    /* 2x2 matrix */ 
                     individual_scan *Signal_fit, 
                     double *MSE_of_fit 
                    ); 
 
 
/* ----------------------------------------------------------------- */ 
/* FAST FOURIER AND WAVELETS TRANSFORMS (BASED ON FFTW LIBRARY)      */ 
/* ----------------------------------------------------------------- */ 
 
void  vndrs_wavelet(long n,
                    double param,
                    long index_freq_min,
                    long index_freq_max,
                    vector_double *coi,
                    vector_double *y,
                    matrix_complex *wave);

void vndrs_where2_max(matrix_double *arr,
                      vector_double *posmax);

matrix_double *vndrs_wavelet_phase(matrix_complex *wave);

void vndrs_wavelet_power_spectrum_time_average(long n, 
                                               matrix_double *wave_power_spectrum,
                                               vector_double *wave_freq);
                                               
void vndrs_wavelet_power_spectrum_frequency_average(long n,
                                                    matrix_double *wave_power_spectrum,
                                                    vector_double *wave_time);

matrix_double *vndrs_wavelet_power_spectrum(matrix_complex *wave);
/* changed - pkervell 04-Aug-02 memory is now allocated inside function */

void vndrs_wave_function(long nk,
                        double param,
                        double scale1,
                        vector_double *kwave, vector_complex *daughter_wave);
                        
double FACTORIAL(double m);

void vndrs_fft_real_2_complex( 
                              /* --------------- */ 
                              /* input variables */ 
                              /* --------------- */ 
                              const vector_double *Signal_Real, 
                              /* ---------------- */ 
                              /* output variables */ 
                              /* ---------------- */ 
                              vector_complex *Spectrum 
                             ); 
 
void vndrs_ifft_complex_2_real( 
                               /* --------------- */ 
                               /* input variables */ 
                               /* --------------- */                               
                               const vector_complex *Spectrum, 
                               /* ---------------- */ 
                               /* output variables */ 
                               /* ---------------- */ 
                               vector_double *Signal_Real 
                              ); 
 
void vndrs_fft( 
               /* --------------- */ 
               /* input variables */ 
               /* --------------- */   	        
               const vector_complex *Signal, 
               /* ---------------- */ 
               /* output variables */ 
               /* ---------------- */                  
	       vector_complex *Spectrum 
              ); 
 
void vndrs_ifft( 
                /* --------------- */ 
                /* input variables */ 
                /* --------------- */                 
                const vector_complex *Spectrum, 
                /* ---------------- */ 
                /* output variables */ 
                /* ---------------- */                    
                vector_complex *Signal 
               ); 
 
 
/* ----------------------------------------------------------------- */ 
/* OTHER SIGNAL PROCESSING ROUTINES (POWER SPECTRUM, FILTERING,      */ 
/* CONVOLUTION, NOISE)                                               */ 
/* ----------------------------------------------------------------- */ 
 
/* 
    Computes the power spectrum of the input signal. 
    Described in [1], section 2.13 
 */ 
power_spectrum *vndrs_power_spectrum ( 
                                      /* --------------- */ 
                                      /* input variables */ 
                                      /* --------------- */ 
                                      individual_scan *Signal    
                                     ); 

/*
    This procedure puts the center of the fringe packet at the zero position of the vector,
    assuming it is previously at the center. This is necessary to compute the phase properly.
    Dimension has to be a multiple of 2.
    pkervell - 25-Oct-02.
*/
vector_double *vndrs_zero_pos_scan_center(vector_double *Scan);

/* 
    Routine which copies one side of the powerspectrum to 
    the full size. Input spectrum shall have half the size of 
    the output spectrum. The full size is twice the half size 
    plus the dimension modulo 
 */ 
power_spectrum *vndrs_mirror_half_pow( 
                                      /* --------------- */ 
                                      /* input variables */ 
                                      /* --------------- */ 
                                      power_spectrum *half_pow, 
                                      long           dimension_modulo 
                                     ); 
 
/* 
    Computes the power spectrum of the input signal and 
    takes half of it: 0<f<1/2nt (Nyquist frequency). 
    Described in [1], section 2.12 
 */ 
power_spectrum *vndrs_one_sided_pow ( 
                                     /* --------------- */ 
                                     /* input variables */ 
                                     /* --------------- */                            
                                     individual_scan *Signal 
                                    ); 
 
 
/*  pkervell - 03-Sep-02
    Smoothing of a vector_double by a moving average of a given half_width:
    if half_width = 2, the average is over I-2, I-1, I, I+1, I+2
    MEMORY ALLOCATED INSIDE FUNCTION FOR OUTPUT
*/
vector_double *vndrs_vector_smooth(vector_double *input_vector,
                                   long half_width);

/* 
  This routine does only a Fourier convolution, i.e. implicitely 
  assumes that the Kernel is large enough so no aliasing occurs. 
 */ 
individual_scan *vndrs_convolution( 
                                   /* --------------- */ 
                                   /* input variables */ 
                                   /* --------------- */  
                                   individual_scan *Signal, 
                                   individual_scan *Kernel 
                                  ); 
 
/* 
    Described in [1], section 2.17 
 */ 
individual_scan *vndrs_filter ( 
                               /* --------------- */ 
                               /* input variables */ 
                               /* --------------- */  
                               individual_scan *Input_signal, 
                               power_spectrum *Freq_filter 
                              ); 
 
/* 
    Computes a bandpass or a low-pass filter. The filter response 
    is defined by 1/(1+abs(f-f0)/delta_f)^n). 
    Described in [1], section 2.5 
 */ 
power_spectrum *vndrs_compute_filter ( 
                                      /* --------------- */ 
                                      /* input variables */ 
                                      /* --------------- */  
                                      long N_frames, 
                                      double Sample_interval, 
                                      double Central_freq, 
                                      double Half_bandwidth, 
                                      long Filter_order 
                                     ); 
 
/*
  added pkervell 15-mar-02: computation of the Wiener filters from the average
  PSD of PA and PB
*/
power_spectrum *vndrs_compute_wiener_mean(
			       const batch *Batch_on,
			       const int verbose_on,
                               const int channel
			       );

/* 
    Described in [1], section 2.22 
 */ 
power_spectrum *vndrs_wiener ( 
                              /* --------------- */ 
                              /* input variables */ 
                              /* --------------- */  
                              individual_scan *Signal_OSPS, 
                              individual_scan *Noise_OSPS, 
			      int verbose_on
                             ); 
 
/* 
    Described in [1], section 2.23 
 */ 
power_spectrum *vndrs_noise_model ( 
                                   /* --------------- */ 
                                   /* input variables */ 
                                   /* --------------- */  
                                   individual_scan *Dark_current, 
                                   long Radius 
                                  ); 
 
/* ------------------------------------------------------------------- */ 
/* STATISTICS                                                          */ 
/* ------------------------------------------------------------------- */ 

/*
    Computation of the histogram of a distribution.
    pkervell - 22-Oct-02
*/
void vndrs_histogram(vector_double *values, /* input values */
                     double range, /* range, in number of sigmas, from M-sig to M+sig */
                     vector_double *histogram_scale, /* out */
                     vector_double *histogram); /* out */


/*
    Kurtosis of a distribution.
    pkervell - 21-Oct-02
    Kurtosis is based on the size of a distribution's tails.
    Distributions with relatively large tails are called "leptokurtic";
    those with small tails are called "platykurtic." A distribution with
    the same kurtosis as the normal distribution is called "mesokurtic."
*/
double vndrs_kurtosis(vector_double *values);

/* 
    Computes the statistics: mean, standard deviation, minimum and maximum 
    data values for each channel in the batch. 
    Described in [1], section 2.4 
 */ 
void vndrs_batch_statistics ( 
                             /* --------------- */ 
                             /* input variables */ 
                             /* --------------- */  
                             batch *Batch, 
                             /* ---------------- */ 
                             /* output variables */ 
                             /* ---------------- */  
                             vector_double *Mean_val,           /* vector of 4 */ 
                             vector_double *Standard_deviation, /* vector of 4 */ 
                             vector_double *Minimum,            /* vector of 4 */ 
                             vector_double *Maximum             /* vector of 4 */ 
                            ); 
 

/*
    Computes the photometric averages of each scan of a batch and puts the result in a
    vector  - pkervell 10-Oct-02
*/
void vndrs_compute_average_vector_from_batch(batch *Batch_in, /* input */
                                  vector_double *I1_mean_vector,
                                  vector_double *I2_mean_vector,
                                  vector_double *PA_mean_vector,
                                  vector_double *PB_mean_vector); /*out*/

/*
    Computes the simple error bar given by stddev/sqrt(N)
*/
double vndrs_compute_simple_error(vector_double *mu2);

/*
    Computes the average of the variable (kappa1 + kappa2)/(kappa1.kappa2)
    with kappa1=kappa1A.kappa1B and kappa2=kappa2A.kappa2B, knowing the kappa matrix and errors.
    reminder: variance of the sum is the sum of variances.
*/
double vndrs_mean_kappa_norm(matrix_double *Matrix,
                                 matrix_double *Stdev_matrix);

/*
    Computes the variance of the variable (kappa1 + kappa2)/(kappa1.kappa2)
    with kappa1=kappa1A.kappa1B and kappa2=kappa2A.kappa2B, knowing the kappa matrix and errors.
    reminder: variance of the sum is the sum of variances.
*/
double vndrs_variance_kappa_norm(matrix_double *Matrix,
                                 matrix_double *Stdev_matrix);

/*
    Computes the variance of a product of normal variables.
    formula from http://www.ses.swin.edu.au/homes/browne/
*/
double vndrs_variance_product(double mean_A,
                              double var_A,
                              double mean_B,
                              double var_B);

/*
    Computes the mean of the quotient of two normal variables.
    formula from http://www.ses.swin.edu.au/homes/browne/
*/
double vndrs_mean_quotient(double mean_A,
                              double var_A,
                              double mean_B,
                              double var_B);

/*
    Computes the variance of the quotient of two normal variables.
    formula from http://www.ses.swin.edu.au/homes/browne/
*/
double vndrs_variance_quotient(double mean_A,
                              double var_A,
                              double mean_B,
                              double var_B);
 
 /*
    Jackknife estimation of the error bars - pkervell 26-Sep-02
    The principle is that the stddev (sigma) of the original sample is equal to
    the stddev of the averages of the N samples built by removing
    one element from the original one.
*/
double vndrs_compute_jacknife_error(vector_double *y);
                                       

/* added pkervell 29-Aug-02: computation of the average value of sqrt(PA.PB) over the fringes */ 
 double vndrs_power_photom(full_scan *Scan_on,
                          double fringes_length);

/* 
    Described in [1], section 2.24 
    This procedure becomes superfluous and should be exchanged 
    by calls to vndrs_mean_scan and vndrs_stddev_scan. 
 */ 
void vndrs_stats ( 
                  /* --------------- */ 
                  /* input variables */ 
                  /* --------------- */  
                  individual_scan *Signal, 
                  /* ---------------- */ 
                  /* output variables */ 
                  /* ---------------- */                    
                  double *Mean, 
                  double *Stddev 
                 ); 
 
/* 
    This routine computes the mean over one scan. 
 */ 
double vndrs_mean_scan( 
                       /* --------------- */ 
                       /* input variables */ 
                       /* --------------- */  
                       individual_scan *Signal 
                      ); 

/* added pkervell 14-Oct-02 :
computation of the arithmetic-geometric average of two doubles (convergence 1% or 100 iter.) */
double vndrs_arith_geom_mean(double x, double y);

/* added pkervell 14-Oct-02 :
computation of the arithmetic-geometric average of two vectors over the central 'width'
pixels of a scan */
double vndrs_arith_geom_mean_scan_center( 
                       /* --------------- */ 
                       /* input variables */ 
                       /* --------------- */  
                       individual_scan *Signal_1,
                       individual_scan *Signal_2,
                       long width 
                      );
 
/* added pkervell 28-Sep-02 :
computation of the average of the logarithms of the intensities over the central
'width' pixels of a scan, to take into account the intrinsic log-normal statistics
of the photometry. The logarithm of the flux has a normal distribution. (TBC) */
double vndrs_mean_log_scan_center( 
                       /* --------------- */ 
                       /* input variables */ 
                       /* --------------- */  
                       individual_scan *Signal,
                       long width 
                      );

/* added pkervell 28-Sep-02 :
computation of the average of the logarithms of the intensities over a vector_double,
to take into account for example the intrinsic log-normal statistics of the photometry.
The logarithm of the flux has in principle a normal statistics. (TBC) */
double vndrs_mean_log_scan( 
                       /* --------------- */ 
                       /* input variables */ 
                       /* --------------- */  
                       individual_scan *Signal
                      );

/* 
    This routine computes the mean over one scan, weighted by the vector 'weights'.
    'weights' must not be of zero sum. 
 */ 
double vndrs_mean_scan_weighted( 
                       /* --------------- */ 
                       /* input variables */ 
                       /* --------------- */  
                       individual_scan *Signal,
                       vector_double *weights 
                      ); 

/*
    Computes the average of a vector between two indices. pkervell - 28-Oct-02
*/
double vndrs_mean_scan_limits( 
                       /* --------------- */ 
                       /* input variables */ 
                       /* --------------- */  
                       individual_scan *Signal,
                       long index_min,
                       long index_max 
                      );

/* 
    This routine computes the mean over the central 'width' pixels of one scan. 
 */ 
double vndrs_mean_scan_center( 
                       /* --------------- */ 
                       /* input variables */ 
                       /* --------------- */  
                       individual_scan *Signal,
                       long width 
                      ); 

/* added pkervell 26-Sep-02 :
computation of the average over the central 'width' pixels of a scan,
weighted by a linear weighting as a function of the distance to the center.*/
double vndrs_mean_scan_center_weighted( 
                       /* --------------- */ 
                       /* input variables */ 
                       /* --------------- */  
                       individual_scan *Signal,
                       long width 
                      );

/* 
    This routine computes the sum over one scan. 
 */ 
double vndrs_sum_scan( 
                      /* --------------- */ 
                      /* input variables */ 
                      /* --------------- */  
                      individual_scan *Signal 
                     ); 
 
/* 
    This routine computes the mean squared over one scan. 
 */ 
double vndrs_mean_squared_scan( 
                               /* --------------- */ 
                               /* input variables */ 
                               /* --------------- */  
                               individual_scan *Signal 
                              ); 
 
/* 
    This routine computes the stddev over one scan. 
 */ 
double vndrs_stddev_scan( 
                         /* --------------- */ 
                         /* input variables */ 
                         /* --------------- */  
                         individual_scan *Signal 
                        ); 
 
/* 
    This routine computes the minimum over one scan. 
 */ 
double vndrs_minimum_scan( 
                          /* --------------- */ 
                          /* input variables */ 
                          /* --------------- */  
                          individual_scan *Signal 
                         ); 
 
/* 
    This routine computes the maximum over one scan. 
 */ 
double vndrs_maximum_scan( 
                          /* --------------- */ 
                          /* input variables */ 
                          /* --------------- */  
                          individual_scan *Signal 
                         ); 

/*
    Computes the skewness of a sample of data points put in a double vector.
*/
double vndrs_skewness(vector_double *x);

 
/* 
  This routine returns the median of an individual scan 
*/ 
double vndrs_median_scan( 
                         /* --------------- */ 
                         /* input variables */ 
                         /* --------------- */  
                         const individual_scan *Signal 
                        ); 

/* 
  This routine returns the median of an individual scan over the central "width" length. 
*/ 
double vndrs_median_scan_center( 
                       /* --------------- */ 
                       /* input variables */ 
                       /* --------------- */  
                       individual_scan *Signal,
                       long width 
                        ); 
 
 
/* Routine that averages over a batch each of the four 
   scan types (I1,I2,PA,PB) and returns one fullscan */ 
full_scan *vndrs_average_batch_to_fullscan( 
                                           /* --------------- */ 
                                           /* input variables */ 
                                           /* --------------- */  
                                           batch *Batch 
                                          ); 
 
 
 
/* Routine that concatenates two batches producing a single batch (scan direction) */ 
batch *vndrs_concatenate_batches_scan_direction( 
                                                /* --------------- */ 
                                                /* input variables */ 
                                                /* --------------- */  
                                                const batch *Batch_A, 
                                                const batch *Batch_B 
                                               ); 
 
/* Routine that concatenates two batches producing a single batch (scan direction) */ 
batch *vndrs_concatenate_batches_frame_direction( 
                                                 /* --------------- */ 
                                                 /* input variables */ 
                                                 /* --------------- */  
                                                 const batch *Batch_A, 
                                                 const batch *Batch_B 
                                                ); 
 
/* Routines that subtracts the mean from a scan */ 
void vndrs_subtract_mean_from_scan(/* ---------------------------------------- */ 
                                   /* input and output variable (is replaced!) */ 
                                   /* ---------------------------------------- */ 
                                   individual_scan *dest 
                                  ); 

 
/* Computes the median of a double matrix along one direction, to produce the median
vector along the other direction. Only the records corresonding to Flag_vector = 1
are taken into account - MEMORY TO BE ALLOCATED OUTSIDE OF THIS FUNCTION
pkervell - 24-Sep-02 */

void vndrs_median_matrix_to_vector_double(vector_double *Pow_spec_med, /* output */
                                          matrix_double *Pow_spec_matrix,
                                          vector_double *Flag_vector); /* input */


/* ------------------------------------------------------------------- */ 
/* PROGRAM FLOW / RUNTIME PROTOCOL                                     */ 
/* ------------------------------------------------------------------- */ 
void vndrs_initialization(char *protocol_mode, char *file_name); 
void vndrs_exit(void); 
void vndrs_start_runtime_protocol(char *protocol_mode, char *file_name); 
void vndrs_runtime_protocol(void); 
void vndrs_end_runtime_protocol(void); 
void vndrs_error_alert(void); 
 
/* ------------------------------------------------------------------- */ 
/* MEMORY HANDLING: LEAKAGE CONTROL                                    */ 
/* ------------------------------------------------------------------- */ 
 
void vndrs_start_memory_usage_protocol(void); 
void vndrs_protocol_memory_usage(long MemoryIncrement); 
void vndrs_end_memory_usage_protocol(void); 
 
/* ------------------------------------------------------------------- */ 
/* MEMORY HANDLING: (ALLOC & FREE) ROUTINES                            */ 
/* ------------------------------------------------------------------- */ 
 
/* Double vector */ 
/* ------------- */ 
vector_double *vndrs_calloc_vector_double( 
                                          /* --------------- */ 
                                          /* input variables */ 
                                          /* --------------- */ 
                                          long dimension); 
void vndrs_free_vector_double(  
                              /* --------------- */ 
                              /* input variables */ 
                              /* --------------- */ 
                              vector_double *vector 
                             ); 
 
/* Complex vector */ 
/* -------------- */ 
vector_complex *vndrs_calloc_vector_complex( 
                                            /* --------------- */ 
                                            /* input variables */ 
                                            /* --------------- */ 
                                            long dimension 
                                           ); 
void vndrs_free_vector_complex( 
                               /* --------------- */ 
                               /* input variables */ 
                               /* --------------- */ 
                               vector_complex *vector 
                              ); 
 
matrix_double *vndrs_calloc_matrix_double( 
	long dimension_1, 
	long dimension_2); 
 
void vndrs_free_matrix_double( 
                              /* --------------- */ 
                              /* input variables */ 
                              /* --------------- */ 
                              matrix_double *matrix 
                             ); 
 
/* Complex matrix */ 
/* -------------- */ 
matrix_complex *vndrs_calloc_matrix_complex( 
                                            /* --------------- */ 
                                            /* input variables */ 
                                            /* --------------- */ 
                                            long dimension_1, long dimension_2 
                                           ); 
void vndrs_free_matrix_complex( 
                               /* --------------- */ 
                               /* input variables */ 
                               /* --------------- */ 
                               matrix_complex *matrix 
                              ); 
 
 
/* Full scan (consists of 4 individual scans I1, I2, PA, PB) */ 
/* --------------------------------------------------------- */ 
full_scan *vndrs_calloc_full_scan( 
                                  /* --------------- */ 
                                  /* input variables */ 
                                  /* --------------- */ 
                                  const long dimension_individual_scan 
                                 ); 
void vndrs_free_full_scan( 
                          /* --------------- */ 
                          /* input variables */ 
                          /* --------------- */ 
                          full_scan *FullScan 
                         ); 
 
/* Full powerspectrum (consists of 2 individual powerspectra I1_pow, I2_pow) */ 
/* ------------------------------------------------------------------------- */ 
full_pow *vndrs_calloc_full_pow( 
                                 /* --------------- */ 
                                 /* input variables */ 
                                 /* --------------- */ 
                                 const long dimension_individual_spectrum 
                                ); 
void vndrs_free_full_pow( 
                         /* --------------- */ 
                         /* input variables */ 
                         /* --------------- */ 
                         full_pow *FullPowerSpectrum 
                        ); 
 
batch *vndrsCallocBatchPerBlock( 
	long	scanBlock, 
	long	numberOfFullScans, 
	long	numberOfFrames); 
 
powbatch *vndrsCallocPowBatchPerBlock( 
	int		scanBlock, 
	long	number_of_full_spectra, 
	long	dimension_individual_spectrum); 
 
/* Batch of full scans */ 
/* ------------------- */ 
batch *vndrs_calloc_batch( 
                          /* --------------- */ 
                          /* input variables */ 
                          /* --------------- */ 
                          const long number_of_full_scans, const long dimension_individual_scan 
                         ); 
void vndrs_free_batch( 
                      /* --------------- */ 
                      /* input variables */ 
                      /* --------------- */ 
                      batch *Batch 
                     ); 
 
void vndrsFreePowBatchPerBlock( 
	unsigned int	dimension, 
	powbatch		*Batch); 
 
void vndrsFreeBatchPerBlock( 
	long	scanBlock, 
	batch	*Batch); 
 
/* Batch of full powerspectra (a 1-D field of full powerspectra) */ 
/* ------------------------------------------------------------- */ 
powbatch *vndrs_calloc_powbatch( 
                                /* --------------- */ 
                                /* input variables */ 
                                /* --------------- */ 
                                const long number_of_full_spectra, 
                                const long dimension_individual_spectrum 
                               ); 
void vndrs_free_powbatch( 
                         /* --------------- */ 
                         /* input variables */ 
                         /* --------------- */ 
                         powbatch *Batch 
                        ); 
 
/* &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& */ 
/*                                                                                  */ 
/* MATHEMATICAL LIBRARY                                                             */ 
/*                                                                                  */ 
/* &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& */ 
 
/* -------------------------------------------------------------------------------- */ 
/* ROUTINES TO ACCESS AND TO SET MATRIX ELEMENTS                                    */ 
/* -------------------------------------------------------------------------------- */ 
 
double  get_element_matrix_double( 
                                  /* --------------- */ 
                                  /* input variables */ 
                                  /* --------------- */ 
                                  const matrix_double *matrix, 
                                  const long i, const long j 
                                 ); 
 
 
void set_element_matrix_double( 
	matrix_double	*matrix, 
	const long		i, 
	const long		j, 
	const double	value); 
 
 
complex get_element_matrix_complex( 
                                   /* --------------- */ 
                                   /* input variables */ 
                                   /* --------------- */ 
                                   const matrix_complex *matrix, 
                                   const long i, const long j 
                                  ); 
void set_element_matrix_complex( 
				/* --------------- */ 
				/* input variables */ 
				/* --------------- */   
				matrix_complex *matrix, 
				const long i, const long j, 
				const complex value 
				); 
 
void vndrs_matrix_sub(
                      /* -------------------------------------------- */
                      /* input and output variable (will be replaced) */
                      /* -------------------------------------------- */  
                      matrix_double *dest,
                      /* -------------- */
                      /* input variable */
                      /* -------------- */  
                      const matrix_double *op
                     );

void vndrs_matrix_add(
                      /* -------------------------------------------- */
                      /* input and output variable (will be replaced) */
                      /* -------------------------------------------- */  
                      matrix_double *dest,
                      /* -------------- */
                      /* input variable */
                      /* -------------- */  
                      const matrix_double *op
                     );


void vndrs_matrix_scale(
                        /* -------------------------------------------- */
                        /* input and output variable (will be replaced) */
                        /* -------------------------------------------- */
                        matrix_double *dest,
                        /* -------------- */
                        /* input variable */
                        /* -------------- */                         
                        const double val
                       );

/* Square root of matrix dest (IN-PLACE TRANSFORM!) Check before that there are no
negative values ! */
void vndrs_matrix_sqrt(
                        /* -------------------------------------------- */
                        /* input and output variable (will be replaced) */
                        /* -------------------------------------------- */
                        matrix_double *dest
                       );

void vndrs_matrix_add_constant(
                      /* -------------------------------------------- */
                      /* input and output variable (will be replaced) */
                      /* -------------------------------------------- */  
                      matrix_double *dest,
                      /* -------------- */
                      /* input variable */
                      /* -------------- */  
                      const double val
                     );

/* Finding the minimum and the maximum of a matrix of doubles */ 
double vndrs_matrix_min( 
                        /* -------------- */ 
                        /* input variable */ 
                        /* -------------- */ 
                        const matrix_double *src 
                       ); 

double vndrs_matrix_max( 
                        /* -------------- */ 
                        /* input variable */ 
                        /* -------------- */ 
                        const matrix_double *src 
                       ); 


/*
    Copy of one matrix to another - pkervell 17-Oct-02
*/
void vndrs_matrix_memcpy(
                      /* ----------------------------------- */
                      /*  output variable (will be replaced) */
                      /* ----------------------------------- */  
                      matrix_double *dest,
                      /* -------------- */
                      /* input variable */
                      /* -------------- */  
                      const matrix_double *op
                     );
                     
                     
/*
    Returns the local maximum of a matrix - pkervell 21-Jan-04
*/
double vndrs_matrix_local_max(
                        /* -------------- */
                        /* input variable */
                        /* -------------- */
                        const matrix_double *matrix,
                        const long I_min, const long I_max,
                        const long J_min, const long J_max
                       );
                     

/*
    Procedure to cancel the edges of the wavelets power spectrum over a specified
    width on each (time) side. Matrix is replaced. No security check, verify the
    matrix and width values. - pkervell 17-Oct-02.
*/
void vndrs_cancel_wl_edges(matrix_double *norm_t_wavelets, long width);

/* -------------------------------------------------------------------------------- */ 
/* ROUTINES ACTING ON VECTORS                                                       */ 
/* -------------------------------------------------------------------------------- */ 
 
/* Sorting vectors */ 
void vndrs_sort_shellsmethod( 
                             /* -------------------------------------------- */ 
                             /* input and output variable (will be replaced) */ 
                             /* -------------------------------------------- */   
                             vector_double *Vector 
                            ); 
 
/* Filling a vector with constant values */ 
void vndrs_vector_fill( 
                        /* -------------------------------------------- */ 
                        /* input and output variable (will be replaced) */ 
                        /* -------------------------------------------- */ 
                        vector_double *dest, 
                        /* -------------- */ 
                        /* input variable */ 
                        /* -------------- */                          
                        const double val 
                       ); 
 
/* Adding, subtracting, multiplying and dividing vectors (double & complex) */ 
void vndrs_vector_add( 
                      /* -------------------------------------------- */ 
                      /* input and output variable (will be replaced) */ 
                      /* -------------------------------------------- */   
                      vector_double *dest, 
                      /* -------------- */ 
                      /* input variable */ 
                      /* -------------- */   
                      const vector_double *op 
                     ); 
void vndrs_vector_complex_add( 
                              /* -------------------------------------------- */ 
                              /* input and output variable (will be replaced) */ 
                              /* -------------------------------------------- */ 
                              vector_complex *dest, 
                              /* -------------- */ 
                              /* input variable */ 
                              /* -------------- */  
                              const vector_complex *op 
                             ); 
void vndrs_vector_sub( 
                      /* -------------------------------------------- */ 
                      /* input and output variable (will be replaced) */ 
                      /* -------------------------------------------- */ 
                      vector_double *dest, 
                      /* -------------- */ 
                      /* input variable */ 
                      /* -------------- */   
                      const vector_double *op 
                     ); 
void vndrs_vector_complex_sub( 
                              /* -------------------------------------------- */ 
                              /* input and output variable (will be replaced) */ 
                              /* -------------------------------------------- */ 
                              vector_complex *dest, 
                              /* -------------- */ 
                              /* input variable */ 
                              /* -------------- */ 
                              const vector_complex *op 
                             ); 
void vndrs_vector_mul( 
                      /* -------------------------------------------- */ 
                      /* input and output variable (will be replaced) */ 
                      /* -------------------------------------------- */ 
                      vector_double *dest, 
                      /* -------------- */ 
                      /* input variable */ 
                      /* -------------- */   
                      const vector_double *op 
                     ); 
void vndrs_vector_complex_mul( 
                              /* -------------------------------------------- */ 
                              /* input and output variable (will be replaced) */ 
                              /* -------------------------------------------- */ 
                              vector_complex *dest, 
                              /* -------------- */ 
                              /* input variable */ 
                              /* -------------- */ 
                              const vector_complex *op 
                             ); 
void vndrs_vector_complex_double_mul( 
                                     /* -------------------------------------------- */ 
                                     /* input and output variable (will be replaced) */ 
                                     /* -------------------------------------------- */ 
                                     vector_complex *dest, 
                                     /* -------------- */ 
                                     /* input variable */ 
                                     /* -------------- */ 
                                     const vector_double *op 
                                    ); 
void vndrs_vector_div( 
                      /* -------------------------------------------- */ 
                      /* input and output variable (will be replaced) */ 
                      /* -------------------------------------------- */ 
                      vector_double *dest, 
                      /* -------------- */ 
                      /* input variable */ 
                      /* -------------- */   
                      const vector_double *op 
                     ); 
void vndrs_vector_complex_div( 
                              /* -------------------------------------------- */ 
                              /* input and output variable (will be replaced) */ 
                              /* -------------------------------------------- */ 
                              vector_complex *dest, 
                              /* -------------- */ 
                              /* input variable */ 
                              /* -------------- */               
                              const vector_complex *op 
                             ); 
 
/* Adding a real-valued constant to each component of a vector */ 
void vndrs_vector_add_constant( 
                               /* -------------------------------------------- */ 
                               /* input and output variable (will be replaced) */ 
                               /* -------------------------------------------- */ 
                               vector_double *dest, 
                               /* -------------- */ 
                               /* input variable */ 
                               /* -------------- */     
                               const double val 
                              ); 

void vndrs_vector_square(
                        /* -------------------------------------------- */
                        /* input and output variable (will be replaced) */
                        /* -------------------------------------------- */
                        vector_double *dest
                       );
/* The real-valued vector 'dest' is multiplied by itself (IN-PLACE TRANSFORM!) */

/* Scaling vectors (multiplying their components with a real-valued constant) */ 
void vndrs_vector_scale( 
                        /* -------------------------------------------- */ 
                        /* input and output variable (will be replaced) */ 
                        /* -------------------------------------------- */ 
                        vector_double *dest, 
                        /* -------------- */ 
                        /* input variable */ 
                        /* -------------- */                          
                        const double val 
                       ); 
void vndrs_vector_complex_scale( 
                                /* -------------------------------------------- */ 
                                /* input and output variable (will be replaced) */ 
                                /* -------------------------------------------- */ 
                                vector_complex *dest, 
                                /* -------------- */ 
                                /* input variable */ 
                                /* -------------- */ 
                                const double val 
                               ); 
 
/* Taking the sqrt (component-wise) */ 
void vndrs_vector_sqrt( 
                       /* -------------------------------------------- */ 
                       /* input and output variable (will be replaced) */ 
                       /* -------------------------------------------- */ 
                       vector_double *dest 
                      ); 
 
void vndrs_vector_trim_positive( 
                                /* -------------------------------------------- */ 
                                /* input and output variable (will be replaced) */ 
                                /* -------------------------------------------- */ 
                                vector_double *dest 
                               ); 

/* Copying the content of a batch to a 2nd batch */ 
void vndrs_batch_memcpy( 
                         /* -------------------------------------------- */ 
                         /* input and output variable (will be replaced) */ 
                         /* -------------------------------------------- */ 
                         batch *dest, 
                         /* -------------- */ 
                         /* input variable */ 
                         /* -------------- */ 
                         const batch *src 
                        ); 
 
/* Copying the content of a vector to a 2nd vector */ 
void vndrs_vector_memcpy( 
                         /* -------------------------------------------- */ 
                         /* input and output variable (will be replaced) */ 
                         /* -------------------------------------------- */ 
                         vector_double *dest, 
                         /* -------------- */ 
                         /* input variable */ 
                         /* -------------- */ 
                         const vector_double *src 
                        ); 
void vndrs_vector_complex_memcpy( 
                                 /* -------------------------------------------- */ 
                                 /* input and output variable (will be replaced) */ 
                                 /* -------------------------------------------- */ 
                                 vector_complex *dest, 
                                 /* -------------- */ 
                                 /* input variable */ 
                                 /* -------------- */ 
                                 const vector_complex *src 
                                ); 
 
/* Finding the minimum and the maximum of a real-valued vector */ 
double vndrs_vector_min( 
                        /* -------------- */ 
                        /* input variable */ 
                        /* -------------- */ 
                        const vector_double *src 
                       ); 
double vndrs_vector_max( 
                        /* -------------- */ 
                        /* input variable */ 
                        /* -------------- */ 
                        const vector_double *src 
                       ); 

/* Finding the minimum and the maximum of a real-valued vector over "width" pixels 
around the center (-width/2, +width/2) */ 
double vndrs_vector_min_center( 
                        /* -------------- */ 
                        /* input variable */ 
                        /* -------------- */ 
                        const vector_double *src,
                        long width 
                       ); 
double vndrs_vector_max_center( 
                        /* -------------- */ 
                        /* input variable */ 
                        /* -------------- */ 
                        const vector_double *src ,
                        long width
                       ); 
 
/* Generating a complex vector from its modulus and phase vectors */ 
void vndrs_make_complex_vector_modulus_phase( 
                                             /* --------------- */ 
                                             /* input variables */ 
                                             /* --------------- */ 
                                             const vector_double *modulus, 
                                             const vector_double *phase, 
                                             /* ---------------- */ 
                                             /* output variables */ 
                                             /* ---------------- */ 
                                             vector_complex *dest 
                                            ); 
 
/* Extract real/imaginary parts of a complex vector */ 
void  vndrs_retrieve_re_im( 
                             /* --------------- */ 
                             /* input variables */ 
                             /* --------------- */ 
                             const vector_complex *src, 
                             /* ---------------- */ 
                             /* output variables */ 
                             /* ---------------- */ 
                             vector_double *real, 
                             vector_double *imaginary 
                            ); 

/* Extract modulus / phase of a complex vector */ 
void  vndrs_retrieve_modulus( 
                             /* --------------- */ 
                             /* input variables */ 
                             /* --------------- */ 
                             const vector_complex *src, 
                             /* ---------------- */ 
                             /* output variables */ 
                             /* ---------------- */ 
                             vector_double *modulus 
                            ); 
void  vndrs_retrieve_phase( 
                           /* --------------- */ 
                           /* input variables */ 
                           /* --------------- */ 
                           const vector_complex *src, 
                           /* ---------------- */ 
                           /* output variables */ 
                           /* ---------------- */ 
                           vector_double *phase 
                          ); 
 
 
 
/* ------------------------------------------------------------- */ 
/* ALL KINDS OF CONVERSIONS                                      */ 
/* ------------------------------------------------------------- */ 
 
void vndrs_vector_double_2_vector_complex( 
                                          /* --------------- */ 
                                          /* input variables */ 
                                          /* --------------- */ 
                                          const vector_double *RealVector, 
                                          /* ---------------- */ 
                                          /* output variables */ 
                                          /* ---------------- */ 
                                          vector_complex *ComplexVector 
                                         ); 
void vndrs_vector_complex_2_vector_double( 
                                          /* --------------- */ 
                                          /* input variables */ 
                                          /* --------------- */ 
                                          const vector_complex *ComplexVector, 
                                          /* ---------------- */ 
                                          /* output variables */ 
                                          /* ---------------- */                                           
                                          vector_double *RealVector 
                                         ); 
void vndrs_vector_complex_2_vector_double_real( 
                                          /* --------------- */ 
                                          /* input variables */ 
                                          /* --------------- */ 
                                          const vector_complex *ComplexVector, 
                                          /* ---------------- */ 
                                          /* output variables */ 
                                          /* ---------------- */   
 
                                          vector_double *RealVector 
                                         ); 
 
/* -------------------------------------------------------------------------------- */ 
/* ROUTINES FOR HANDLING COMPLEX NUMBERS                                            */ 
/* -------------------------------------------------------------------------------- */ 
 
/* Generating complex numbers */ 
complex complex_of_modulus_phase( 
                                 /* --------------- */ 
                                 /* input variables */ 
                                 /* --------------- */ 
                                 double modulus, double phase 
                                ); 
 
/* Modulus and phase of a complex number */ 
double  vndrs_abs_complex( 
                          /* --------------- */ 
                          /* input variables */ 
                          /* --------------- */ 
                          const complex z 
                         ); 
double  vndrs_arg_complex( 
                          /* --------------- */ 
                          /* input variables */ 
                          /* --------------- */ 
                          complex z 
                         ); 
 
/* Multiplying a complex number by a real factor */ 
complex vndrs_real_complex_mul( 
                               /* --------------- */ 
                               /* input variables */ 
                               /* --------------- */ 
                               const double real_value, 
                               const complex complex_value 
                              ); 
 
/* Adding, subtracting, multiplying, dividing complex numbers */ 
complex vndrs_add_complex( 
                          /* --------------- */ 
                          /* input variables */ 
                          /* --------------- */ 
                          const complex a, 
                          const complex b 
                         ); 
complex vndrs_sub_complex( 
                          /* --------------- */ 
                          /* input variables */ 
                          /* --------------- */ 
                          const complex a, 
                          const complex b 
                         ); 
complex vndrs_mul_complex( 
                          /* --------------- */ 
                          /* input variables */ 
                          /* --------------- */ 
                          const complex a, 
                          const complex b 
                         ); 
complex vndrs_div_complex( 
                          /* --------------- */ 
                          /* input variables */ 
                          /* --------------- */ 
                          const complex a, 
                          const complex b 
                         ); 

/* Bessel function J1(x), inspired from Numerical Recipes - pkervell 04-Nov-02 */

double vndrs_bessel_j1(double x);
 
/* -------------------------------------------------------------------------------- */ 
/* ROUTINES FOR DATA IO HANDLING                                                    */ 
/* -------------------------------------------------------------------------------- */ 
 
void vndrs_generate_full_filename(const char *filename, 
                                  char *Full_Filename 
                                 ); 
 
/* Routine to dump double vectors in ASCII format into a file */ 
void vndrs_fprintf_vector_double( 
                                 /* --------------- */ 
                                 /* input variables */ 
                                 /* --------------- */ 
                                 const vector_double *vector, 
                                 const char *filename 
                                ); 
 
/* Routine to dump complex vectors in ASCII format into a file */ 
void vndrs_fprintf_vector_complex( 
                                  /* --------------- */ 
                                  /* input variables */ 
                                  /* --------------- */ 
                                  const vector_complex *vector, 
                                  const char *filename 
                                 ); 
 
/* Routine to dump double matrices in ASCII format into a file */ 
void vndrs_fprintf_matrix_double( 
                                 /* --------------- */ 
                                 /* input variables */ 
                                 /* --------------- */ 
                                 const matrix_double *matrix, 
                                 const char *filename 
                                ); 
 
/* Routine to dump complex matrices in ASCII format into a file */ 
void vndrs_fprintf_matrix_complex( 
                                  /* --------------- */ 
                                  /* input variables */ 
                                  /* --------------- */ 
                                  const matrix_complex *matrix, 
                                  const char *filename 
                                 ); 
 
batch *vndrs_load_fluor_test_data( 
	long which); 




/* --------------------------------------------------------------- */ 
/* BOOTSTRAPPING FUNCTIONS (D. Segransan/pkervell)                 */ 
/* --------------------------------------------------------------- */ 


void vndrs_compute_bootstrap(vector_double *y,
                             vector_double *weights, /* added pkervell - 09-Sep-02 */
			     long B,
			     /* output */
			     double *mean_val,
			     double *mean_val_bootstrap,
			     double *stddev_mean_bootstrap,
			     double *bias_mean,
			     double *full_stddev_mean,
			     double *Ci_mean_min,
			     double *Ci_mean_max,
			     double *Ci2_mean_min,
			     double *Ci2_mean_max,
			     double *Ci3_mean_min,
			     double *Ci3_mean_max,
			     double *median_val,
			     double *median_val_bootstrap,
			     double *stddev_median_bootstrap,
			     double *bias_median,
			     double *full_stddev_median,
			     double *Ci_median_min,
			     double *Ci_median_max,
			     double *Ci2_median_min,
			     double *Ci2_median_max,
			     double *Ci3_median_min,
			     double *Ci3_median_max,
			     double *skew_mean,
			     double *skew_med,
                             double *kurtosis_mean,
                             double *kurtosis_med);

vector_double *vndrs_extract_nonzero(
					 vector_double *mu2_vector,
                                         vector_double *flag_processed
					 );

void vndrs_bootstrap_generator(
			       vector_double *y,
			       vector_double *w,
			       long B,
			       vector_double *yb,
                               vector_double *wb
			       );

void vndrs_bootstrap(/* input */
		     long select,
		     vector_double *y,
		     vector_double *w,
		     vector_double *yb,
		     vector_double *wb,
		     long B,
		     /* output */
		     double *theta_c,
		     double *theta_B_p,
		     double *se_B,
		     double *Bias,
		     double *Rms,
		     double *Ci_min_1,
		     double *Ci_max_1,
		     double *Ci_min_2,
		     double *Ci_max_2,
		     double *Ci_min_3,
		     double *Ci_max_3, /* added 2 and 3 sigmas - pkervell - 18-Oct-02 */
                     double *skew,
                     double *kurtosis);

int vndrs_sort_bootstrap(vector_double *y);

long vndrs_round(double x);

/* --------------------------------------------------------------- */ 
/* BASELINE RELATED                                                */ 
/* --------------------------------------------------------------- */ 

/*
    Conversion of HH MM SS.SS/DD MM SS.SS right ascension and Dec
    to decimal degrees values - pkervell - 18-Oct-02
*/

void vndrs_convert_RA_Dec_to_decimal_degs(double RA_h, double RA_m, double RA_s,
                                          double Dec_d, double Dec_m, double Dec_s,
                                          double *RA, double *Dec); /* out */

/*
    Conversion of decimal degrees right ascension and Dec
    to HH MM SS.SS/DD MM SS.SS values - pkervell - 18-Oct-02
*/

void vndrs_convert_decimal_degs_to_RA_Dec(double RA, double Dec, /* in */
                                          double *RA_h, double *RA_m, double *RA_s, /* out */
                                          double *Dec_d, double *Dec_m, double *Dec_s);


double vndrs_juliandate_to_local_sidereal(
    double julian_date, double longitude);

void vndrs_convert_HA_Dec_to_Alt_Az(double hour_angle,
				    double Dec_object,
				    double latitude,
				    double *altitude,
				    double *azimuth);

void vndrs_project_baseline(double hour_angle,
			    double Dec_object,
			    double latitude,
			    double t1x, double t1y, double t1z,
			    double t2x, double t2y, double t2z,
                            double *altitude_object, double *azimuth_object,
			    double *projected_baseline,
			    double *azimuth_baseline, 
			    double *u,
			    double *v);


/* --------------------------------------------------------------- */ 
/* PISTON RELATED                                                  */ 
/* --------------------------------------------------------------- */ 

void covsrt(matrix_double *covar,
	    long ma,
	    vector_double *ia,
	    long mfit);

void mrqmin2D(vector_double *x,
	      vector_double *y,
	      vector_double *sig,
	      long ndata,
	      vector_double *a,
	      vector_double *ia,
	      matrix_double *covar,
	      matrix_double *alpha,
	      double *chisq,
	      double *lambda,
	      double gain,
	      vector_double *naxes);

double vndrs_lmfit_gauss2D(vector_double *array2D,
			   vector_double *naxes,
			   vector_double *gues,
			   vector_double *a);

void vndrs_gauss2D(double x,
		   vector_double *a,
		   double *y,
		   vector_double *dyda,
		   long na,
		   vector_double *naxes);

void vndrs_swap_lines(matrix_double *Matrix, double i, double j);

void vndrs_gaussj(matrix_double *a,
	    long n,
	    matrix_double *b,
	    long m);

void mrqcof2D(vector_double *x,
	      vector_double *y,
	      vector_double *sig,
	      long ndata,
	      vector_double *a,
	      vector_double *ia,
	      matrix_double *alpha,
	      vector_double *beta,
	      double *chisq,
	      vector_double *naxes);

void vndrs_wave_select(matrix_double *wave_power_spectrum, vector_double *res);

void where2(matrix_double *arr,
	    vector_double *naxes,
	    double *mxx,
	    long *xx,
	    long *yy,
	    long minx,
	    long maxx,
	    long miny,
	    long maxy);


/* --------------------------------------------------------------- */ 
/* VINCI DRS INCLUDES                                              */ 
/* --------------------------------------------------------------- */ 
 
#include "vndrs_IO.h" 
 
#endif
