/**************************************************************
 *                                                            *
 *         Statistical library                                *
 *                                                            *
 *     Bootstrap Resampling methods                           *
 *       - Standard error                                     *
 *       - confidence interval                                *
 *       - bias estimation                                    *
 *                                                            *
 *     Written by Damien Segransan,                           *
 *     Observatoire de Geneve, march 2002                     *
 *     Revised by Pierre Kervella, 2002-2003
 *                                                            *
 *    Reference : "An Introduction to Bootstrap", B. Efron    *
 *                                                            *
 *                                                            *
 * The bootstrap involves repeated re-estimation of a         *
 * parameter using random samples with replacement            *
 * from the original data. Because the sampling is with       *
 * replacement, some items in the data set are selected two or*
 * more times and other are not selected at all. When this    *
 * is repeated a hundred or a thousand times, we get          *
 * pseudo-samples that behave similarly to the underlying     *
 * distribution of the data.                                  *
 *                                                            *
 * You can use these pseudo-samples in several ways. First,   *
 * you can estimate the mean of these pseudo-samples          *
 * should be close to the estimate itself. If there is a      *
 * discrepancy, you have a quantitative estimate of bias.     *
 *                                                            * 
 * Second, you can look at the standard deviation of these    *
 * pseudo-samples. This gives you a bootstrap standard        *
 * error of your estimate. This standard error is not reliant *
 * on any distributional assumptions (like normality).        *
 *                                                            *
 * Third, you can compute the 2.5 percentile and the 97.5     *
 * percentile of these pseudo-samples. This gives you a       *
 * bootstrap confidence interval. You could also use the      *
 * classic formula for the confidence interval (plus or minus *
 * 1.96 standard errors).                                     *
 **************************************************************/

/************************************************************ 
 * Y = {y1, y2,.., yn}  -> [n]     // y is the original data vector
 * We are interested in computing the T(Y) and its standard error.
 * F(y) can be the mean of Y or the median or something else...
 * 
 * Yb = -> [n x B] // yb is the bootstrap replicate of y builded by randomly
 *                  throwing (with replacements) datapoints from y.
 * teta_c      = T(Y)           // -> [1] F : average, median or other... 
 * teta_b      = T(Yb,1)        // -> [B] F : is applied to the 1st dimension of yb
 * teta_b_p    = teta_b(sum)/B  // -> [1] Bootstrap estimate of the average, the median or other
 * 
 * se_B[Y]     = sqrt( (teta_b - teta_b_p)^2 )(sum)/(B-1))   // -> [1] standard error
 * Bias        = teta_b_p - teta_c                          // Bias
 * RMS         = SEB_b * sqrt(  1 + (Bias / SEB_b)^2)       // RMS including the Bias
 * 
 * Confidence interval for a 68.4% probability
 * one need at least 500..2000 bootstraps (Efron, p158)
 *  Z_b = (teta_b - teta_c) /  sqrt( (teta_b  - teta_b_p)^2 )
 *  sort(Z_b)(int(.684*B))
 *  sort(Z_b)(int(.316*B))
 *  Ci_min = Se_B * sort(Z_b)(int(.316*B))
 *  Ci_max = Se_B * sort(Z_b)(int(.684*B))
 ************************************************************/


#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<vndrs.h>

/*
    Jackknife estimation of the error bars - pkervell 26-Sep-02
    The principle is that the stddev (sigma) of the original sample is equal to
    the stddev of the averages of the N samples built by removing
    one element from the original one.
*/
double vndrs_compute_jacknife_error(vector_double *y)
{
    long N_samples, I, J, ind;
    vector_double *averages_vector, *current_sample;
    double variance_jack, average_sample, average_jack, sigma_jack;
    
    N_samples = y->dimension;
    if(N_samples <= 1) return 0.0; /* This method does not work for one element only... */
    
    current_sample = vndrs_calloc_vector_double(N_samples-1);
    averages_vector = vndrs_calloc_vector_double(N_samples);
    
    
    average_sample = vndrs_mean_scan(y);
    
    for(I=0;I<N_samples;I++)
    {
        ind = 0;
        /* Construction of the current_vector (= original without element I) */
        for(J=0;J<I;J++)
        {
            current_sample->data[ind] = y->data[J];
            ind++;
        }
        for(J=I+1;J<N_samples;J++)
        {
            current_sample->data[ind] = y->data[J];
            ind++;
        }
        /* Average values vector */
        averages_vector->data[I] = vndrs_mean_scan(current_sample);
    }
    
    /* Now we compute the actual Jackknife variance */ 
    variance_jack = 0.0;
    average_jack = vndrs_mean_scan(averages_vector);
    for(I=0;I<N_samples;I++)
        variance_jack += (averages_vector->data[I]-average_jack)*(averages_vector->data[I]-average_jack);
        
    variance_jack = variance_jack * 1.0 * (1.0*N_samples - 1.0)/(1.0*N_samples);
    
    sigma_jack = sqrt(variance_jack);
    
    vndrs_free_vector_double(current_sample);
    vndrs_free_vector_double(averages_vector);
    
    return sigma_jack;
}


/*
This is the main routine to call the computation of the bootstrapping
and output the relevant parameters. Added 95 and 99% confidence - pkervell 18-Oct-02
*/

/************************************************************/
void vndrs_compute_bootstrap(
			     vector_double *y,
                             vector_double *weights, /* added pkervell - 09-Sep-02 */
			     long B,
			     /* output */
			     double *mean_val,
			     double *mean_val_bootstrap,
			     double *stddev_mean_bootstrap,
			     double *bias_mean,
			     double *full_stddev_mean,
			     double *Ci_mean_min,
			     double *Ci_mean_max,
			     double *Ci2_mean_min,
			     double *Ci2_mean_max,
			     double *Ci3_mean_min,
			     double *Ci3_mean_max,
			     double *median_val,
			     double *median_val_bootstrap,
			     double *stddev_median_bootstrap,
			     double *bias_median,
			     double *full_stddev_median,
			     double *Ci_median_min,
			     double *Ci_median_max,
			     double *Ci2_median_min,
			     double *Ci2_median_max,
			     double *Ci3_median_min,
			     double *Ci3_median_max,
			     double *skew_mean,
			     double *skew_med,
			     double *kurtosis_mean,
			     double *kurtosis_med
			     )
{
  long N_samples;
  vector_double *yb, *wb;

  N_samples = y->dimension;

  yb = vndrs_calloc_vector_double(B*N_samples);
  wb = vndrs_calloc_vector_double(B*N_samples);


  vndrs_bootstrap_generator(y, weights, B, yb, wb);

  vndrs_bootstrap(
		  0,
		  y,
                  weights,
		  yb,
                  wb,
		  B,
		  mean_val,
		  mean_val_bootstrap,
		  stddev_mean_bootstrap,
		  bias_mean,
		  full_stddev_mean,
		  Ci_mean_min,
		  Ci_mean_max,
		  Ci2_mean_min,
		  Ci2_mean_max,
		  Ci3_mean_min,
		  Ci3_mean_max,
                  skew_mean,
                  kurtosis_mean
		  );

  vndrs_bootstrap(
		  1,
		  y,
                  weights,
		  yb,
                  wb,
		  B,
		  median_val,
		  median_val_bootstrap,
		  stddev_median_bootstrap,
		  bias_median,
		  full_stddev_median,
		  Ci_median_min,
		  Ci_median_max,
		  Ci2_median_min,
		  Ci2_median_max,
		  Ci3_median_min,
		  Ci3_median_max,
                  skew_med,
                  kurtosis_med
		  );


  vndrs_free_vector_double(yb);
  vndrs_free_vector_double(wb);

}


/************************************************************/
void vndrs_bootstrap_generator(
			       vector_double *y,
			       vector_double *w,
			       long B,
			       vector_double *yb,
                               vector_double *wb)
{
  vector_double *rdm;
  long i,j,n;
  double temp=0.0;

  srand(time(0)); 
  
  n = y->dimension;

  rdm = vndrs_calloc_vector_double(n*B);
  
  for  (i=0;i<B;i++)
    {
    for (j=0;j<n;j++)
      {
	temp = vndrs_round((n-1)*((double)rand()/RAND_MAX));
	rdm->data[j*B+i] = temp;
	yb->data[j*B+i] = y->data[(long)rdm->data[j*B+i]];
	wb->data[j*B+i] = w->data[(long)rdm->data[j*B+i]];
	/* This procedure is ok */
     }
    }
  vndrs_free_vector_double(rdm);
}

/************************************************************/
void vndrs_bootstrap(
		     /* input */
		     long select,
		     vector_double *y,
		     vector_double *w,
		     vector_double *yb,
		     vector_double *wb,
		     long B,
		     /* output */
		     double *theta_c,
		     double *theta_B_p,
		     double *se_B,
		     double *Bias,
		     double *Rms,
		     double *Ci_min_1,
		     double *Ci_max_1,
		     double *Ci_min_2,
		     double *Ci_max_2,
		     double *Ci_min_3,
		     double *Ci_max_3, /* added 2 and 3 sigmas - pkervell - 18-Oct-02 */
                     double *skew,
                     double *kurtosis
		     )
{
  long i,j,n,buffer;
  vector_double *tmp_arr, *tmp_arr_w;
  vector_double *Z_B;
  vector_double *theta_B;
  int dummy;

  n = y->dimension;

  
    Z_B     = vndrs_calloc_vector_double(B);
    theta_B = vndrs_calloc_vector_double(B);

    /*    printf(" B = %d , n = %d \n",B,n); */

  
  /* median or average */
  if(select == 1)
    *theta_c = vndrs_median_scan(y); /* the median is not weighted */
  else if(select == 0)
    *theta_c = vndrs_mean_scan_weighted(y,w);
  else
    *theta_c = 0.0;

 
  /*  printf(" theta_c = %f \n", *theta_c);*/
 

  /*Bootstrapped median or average*/
  *theta_B_p =0.0;

  tmp_arr     = vndrs_calloc_vector_double(n);
  tmp_arr_w   = vndrs_calloc_vector_double(n);
  
  for  (i=0;i<=B-1;i++)
    {
    for (j=0;j<=n-1;j++)
      {
	tmp_arr->data[j]  = yb->data[j*B+i];
	tmp_arr_w->data[j]= wb->data[j*B+i];
      }
    
    /* median or average */
    if(select == 0)
      theta_B->data[i] = vndrs_mean_scan_weighted(tmp_arr, tmp_arr_w);
    else if(select == 1)
      theta_B->data[i] = vndrs_median_scan(tmp_arr);
    else
      theta_B->data[i] = 0.0;

   *theta_B_p += theta_B->data[i];
    }
  *theta_B_p /= (double)(B);
  vndrs_free_vector_double(tmp_arr);
  vndrs_free_vector_double(tmp_arr_w);
  
  /* Skewness */
  *skew = vndrs_skewness(theta_B);
  
  /* Kurtosis */
  *kurtosis = vndrs_kurtosis(theta_B);
  
  /*Bootstrapped standard error */
  *se_B = 0.0;
  for  (i=0;i<=B-1;i++){
    *se_B += pow(theta_B->data[i] - *theta_B_p ,2);
  }

  *se_B = pow( *se_B/(double)(B-1) ,0.5);

  /*Bias estimate : "An Introduction to Bootstrap", B. Efron p125*/
  /*uniform probability distribution P0*/
  *Bias = *theta_B_p - *theta_c;

  /* Bootstrap rms including bias*/
  *Rms = *se_B * pow(1+ pow(*Bias / *se_B,2),0.5);

  /* Confidence intervals */

  dummy = vndrs_sort_bootstrap(&theta_B[0]);

  /* Here we derive the error bars at 1 sigma (68.4% confidence), 95% and 99% confidence */
  /* AAAARGHH !! There was a bug !! The confidence interval for 1 sigma is from 0.158
  to 0.842 percentiles !! This is MANDATORY, otherwise you get a tiny error bar...
  pkervell - 19-Oct-02 */
  buffer = vndrs_round(0.158*B);
  *Ci_min_1 = *theta_c - theta_B->data[buffer];

  buffer = vndrs_round(0.842*B);
  *Ci_max_1 = theta_B->data[buffer] - *theta_c;

  buffer = vndrs_round(0.025*B);
  *Ci_min_2 = *theta_c - theta_B->data[buffer];

  buffer = vndrs_round(0.975*B);
  *Ci_max_2 = theta_B->data[buffer] - *theta_c;

  buffer = vndrs_round(0.005*B);
  *Ci_min_3 = *theta_c - theta_B->data[buffer];

  buffer = vndrs_round(0.995*B);
  *Ci_max_3 = theta_B->data[buffer] - *theta_c;
  
  /* printf(" Ci_min = %f, Ci_max = %f \n", *Ci_min, *Ci_max); */

  vndrs_free_vector_double(Z_B);
  vndrs_free_vector_double(theta_B);
}


/************************************************************/
int vndrs_sort_bootstrap(vector_double *y)
{
  /* tested by D. Segransan. It works ok, march 24th 2002 */

  long i,j,inc,n;
  double v;

  n = y->dimension;

  inc=1;
  do{
     inc *=3;
     inc++;
  } while (inc <=n);

  do {
    inc /=3;
    for (i=inc+1;i<=n;i++){
      v=y->data[i-1];
      j=i;
      while (y->data[j-inc-1]>v){
        y->data[j-1]=y->data[j-inc-1];
        j -= inc;
        if (j<=inc) break;
      }
      y->data[j-1]=v;
    }
  } while (inc>1);
  return(0);
}



/*-----------------------------------------------------*/
long vndrs_round(double x)
{
  double a,rest;
  a = x;
  rest = (a-floor(a));
  if  (rest <  0.5)  {a = floor(x);}
  if  (rest >= 0.5)  {a = ceil(x);}
  return(a);
}




