/******************************************************************************
*******************************************************************************
*				European Southern Observatory
*		 Vlti ViNci Data Reduction Software (VNDRS)
*
* Module name:	vndrs_DataReduction.c
* Description:	Contains the main program for data reduction
*
*
* History:		
* 01-May-01		(csabet) Revised
* 13-Mar-02             (pkervell) Deep revision, also in other .c files
* 18-Mar-02             (pkervell & dsegransan) Wavelets implemented
* 25-Apr-02             (pkervell) Computation of baseline and revision
* 21-Jun-02             (pkervell) Modification of the names of export files
* xx-Jul-02             (segransan) Introduction of piston estimation
* 29-Jul-02             (pkervell) Adapted to IONIC data (option -ionic)
* 04-Aug-02             (pkervell) Revision and memory management changes
* 05-Sep-02		(pkervell) Lots of changes in Wiener + QL + ...
* 28-Sep-02		(pkervell) Complete revision of algorithm
* 04-Oct-02		(pkervell) Full debugging and introduction of off-src calib.
* 15-May-03		(pkervell) Alt. and Az. added to the rvf file
* 16-Dec-06     (pkervell) Reinstallation of depistonization of visibilities
*
*******************************************************************************
******************************************************************************/

/******************************************************************************
*	Compiler directives
******************************************************************************/
#define	VNDRS_DATAREDUCTION_DEBUGx

/******************************************************************************
*	Include files
******************************************************************************/
#include "vndrs.h"
#include "vndrs_FITS.h"
#include "vndrs_DataReduction.h"
#include "history.h"	/* csabet 28-Jun-01 */
#include <unistd.h>     /* pkervell 19-mar-02 */
#include <time.h>     /* pkervell 16-apr-02 */
#include "vndrs_Globalvar.h"

/*****************************************************************************
*	Global Variables. Possibly read from command line 
*****************************************************************************/


char	files_list[MAX_STRING_LENGTH];
/* added pkervell 13-mar-2002:
   Name of the ASCII file containing the names of the files to process 
   one complete name per line, with order off, a, b, on
*/

char	infile[MAX_STRING_LENGTH];		/*	Root input FITS file name */
int	source_id = 1;        			/*	Source identifier */  
char	file_off[MAX_STRING_LENGTH]; 
char	file_on[MAX_STRING_LENGTH]; 
char	file_a[MAX_STRING_LENGTH]; 
char	file_b[MAX_STRING_LENGTH];		/*	File names for off-source, on-source, */
						/*	beam A, beam B signals */ 
char    protocol_file_name[MAX_STRING_LENGTH];  /*      File name of the output protocol */
char	extfile[50];				/*	File extension */ 
int	matrix_flag = 1;		        /*	Matrix computation method (0=least squares, 1=standard FLUOR method) */
int	export_flag = 0;		        /*	Export only flag  */ 
int	use_mean_wiener = 1;		        /*	Use the Wiener filters computed on the mean spectra (default yes) */ 
int	verbose_on = 0;		                /*	Outputs files of the intermediate steps of processing (default no) */ 
int     debias = 0;                             /*      Remove power found by off-source processing - modified pkervell 11-Oct-02 */
int     depiston = 0;						/*      Piston factor on visibilities - modified pkervell 16-Dec-06 */
long    fringes_length = 100;                   /*      Length of the main peak of the fringe packet (used for wavelets) */
int     median_output = 0;                      /*      Computes the batch visibilities as the median of the obtained values */
int     background_per_scan = 0;                /*      Computes the background for each scan onsrc if set to 1 */
float   center_wl = 5;                          /*      Center the fringes before reduction using wavelets, defines the threshold */
int     select_scans = 1;	                /*      Activate or not the selection of the scans (peak width,...) 21-Jan-04 */
float   max_jump = 20.0;                        /*      Maximum acceptable fringes OPD jump between scans (microns) */
double  apod = 0.1;                             /*      Default value for apodization length (fraction of the scan length) */
double  blockout = 0.0;                         /*      Default value for blockout parameter (fraction of the scan length) */
int     ionic_in = 0;                           /*      Set to 1 when the IONIC beam combiner is used (one interf output) */
double  min_wave    = -1;          	        /*      Minimum wavenumber default (not specified by user) */
double  max_wave    = -1;             	        /*      Maximum wavenumber default (not specified by user) */
double  min_wave_QL = 350000;                   /*      Minimum wavenumber for QL (in m^-1) for MONA */
double  max_wave_QL = 600000;                   /*      Maximum wavenumber for QL (in m^-1) for MONA */
long    bootstrap_number = 3000;                /*      Bootstrap number for the evaluation of biases */
char	directory_process[MAX_STRING_LENGTH];	/*	Name of directory to process as a whole */
int     directory_process_on = 1;               /*      Flag to process a whole directory, default is on to have dir processing by default */
char	outfile[MAX_STRING_LENGTH];		/*	output FITS/ascii file name */
char	instrument[MAX_STRING_LENGTH];		/*	Instrument mode */
char	protocol_mode[MAX_STRING_LENGTH];	/*	Protocol mode */
char	report_file[MAX_STRING_LENGTH];		/*	report file */
int	ierr;				        /*	Error status */ 
long	numberOfScansRequested = -1;		/*	User requested on-source scan number */
float	threshold = 5.0;		        /*	Photometric rejection threshold in sigmas - default value 5.0 pkervell 06-Nov-02
                                                        This is "legal" as we select also by piston and other factors. Scans below
                                                        one and the threshold are processed using the mean of P over the fringes.*/
float	shape = 0.000011;		        /*	User requested shape factor pkervell 12-mar-2002 */
bool    process_all = 0;                        /*      User request to process scans with QL undetected fringes if true
							added pkervell 12-mar-2002 default is NOT process them */
int     output_ascii = 0;                       /*      Output the processed fringes and pows to ASCII files
                                                        superseded by in_out_ascii option - pkervell 15-Oct-03 */
long	N_frames_cut = 256;                     /*      Size of the cut scans for processing */
int	list_targets = 0;			/*	Output the list of files in the chosen directory (default: no) */
int	fourier_out = 0;			/*	Outputs the fourier values to the rvf file if set to 1 (default: 0) */

double  lat = -0.429792176;                     /*      Coordinates of Paranal, for baseline related computations (radians) */
double  longi = 1.228750578;
double  Pi = 3.14159265358979311599796346854418516;

/******************************************************************************
*
* Name:			main		
*
* Input/Output:	See function arguments to avoid duplication
*
* Description:	Main program for Data Reduction task
*
*
*
* History:
* Mar-02 and Apr-02: heavily updated (pkervell)
******************************************************************************/
int main(
	int	argc, 
	char	**argv)
{

  /*  Local Declarations
      --------------------*/
  /*	batch			*On_source, *Off_source, *Beam_A, *Beam_B; */
  batch	*On_source = NULL, *Off_source = NULL, *Beam_A= NULL, *Beam_B = NULL; /* csabet 09-Aug-2001 */
  batch	*Batch_on_processed, *Batch_off_processed;
  double Wavenum_theo; /* added pkervell 11-Oct-02 */
  vector_double	*Wavenum_limits_fft, *Wavenum_limits_wl, *Wavenum_limits_QL, *Relative_MSE;
  matrix_double	*Matrix;
  matrix_double *Matrix_buffer, *Standard_dev_Matrix_buffer;
  matrix_double *matrix_variances;

  long	N_scans_processed, e, i, j;
  long	numberOfScansRejected;			
  
  vector_double	*Mu1_squared, *Mu2_squared;
  int         CalculateMatrix;
  int         loop_process;
/*  bool	createFitsFile=TRUE; removed pkervell - 04-Aug-02 */
    
  char	command_unix[MAX_STRING_LENGTH];  /* added pkervell 19-mar-02 */
  
  char	infile_on[MAX_STRING_LENGTH];
  char	infile_off[MAX_STRING_LENGTH];
  char	infile_a[MAX_STRING_LENGTH];
  char	infile_b[MAX_STRING_LENGTH];
  
  char	outfile_header[MAX_STRING_LENGTH];
  char	outfile_data[MAX_STRING_LENGTH];
/*  char	outfile_proc[MAX_STRING_LENGTH]; removed - pkervell - 04-Aug-02 */
  char	outfile_FITS[MAX_STRING_LENGTH];
  
  char filename_date[MAX_STRING_LENGTH], filename_time_hr[MAX_STRING_LENGTH],
      filename_time_min[MAX_STRING_LENGTH], filename_time_sec[MAX_STRING_LENGTH],
      filename_time_thou[MAX_STRING_LENGTH];

  char	arcfile[MAX_STRING_LENGTH];
  
  /*	Variables for reading FITS files header */
  int
      nb_sources,
      status = 0,
      obsmode=0, nscan=0, nfram=0, insmode=0, acqmode=0, 
      obsmode_a=0, nfram_a=0, insmode_a=0,
      obsmode_b=0, nfram_b=0, insmode_b=0,
      obsmode_off=0, nfram_off=0, insmode_off=0; 
  
  double
      fringespeed = 0.0,   /* default values are a dummy */
      framerate = 0.0, mjdobs = 0.,
      fringespeed_off = 0.0,   /* default values are a dummy */
      framerate_off = 0.0,
      fringespeed_a = 0.0,   /* default values are a dummy */
      framerate_a = 0.0,
      fringespeed_b = 0.0,   /* default values are a dummy */
      framerate_b = 0.0;
  
  /* double  fringes_wavelength = 2.128e-6;    */      /* Modified pkervell 06-Nov-02 */
  double  fringes_wavelength = 2.195e-6;          /*      Fringes effective wavelength (meters)
							  default is K band */
  /* fitsfile	*outfptr; removed pkervell 04-Aug-02 */ 
  FILE *f_list;

  matrix_double *Processed_I1, *Processed_I2; /* added pkervell 14-apr-02 */

  /*	Variables for reading FITS files to put them in log file - pkervell 14-Apr-02 */
  double time_step, delay_step, freq, minfreq, maxfreq, speed, rate, RA, Dec,
      t1x, t1y, t1z, t2x, t2y, t2z, t3x, t3y, t3z;
  int scan_mode, ierr;
  char name[MAX_STRING_LENGTH], Date_Obs[MAX_STRING_LENGTH], Observer[MAX_STRING_LENGTH],
      band[MAX_STRING_LENGTH], station1[MAX_STRING_LENGTH], station2[MAX_STRING_LENGTH],
      telescope[MAX_STRING_LENGTH], OBname[MAX_STRING_LENGTH], equinox[MAX_STRING_LENGTH],
      Catg[MAX_STRING_LENGTH], ReadMode[MAX_STRING_LENGTH];

  /* Batch values */
  char	batches_file_name[MAX_STRING_LENGTH];
  FILE *f_batches_values;
  double time_batch, baseline_batch, azimuth_batch, u_batch, v_batch,
      Mu1_batch_wl, ErrMu1_batch_min_wl, ErrMu1_batch_max_wl, ErrMu1_batch_wl,
      Mu2_batch_wl, ErrMu2_batch_min_wl, ErrMu2_batch_max_wl,
      Mu1_batch_fft, ErrMu1_batch_min_fft, ErrMu1_batch_max_fft, ErrMu1_batch_fft,
      Mu2_batch_fft, ErrMu2_batch_min_fft, ErrMu2_batch_max_fft,
      background_fourier, background_wavelets,
      background_fourier_med, background_wavelets_med,
      effective_wavelength, effective_wavelength_wl, effective_wavelength_error,
      Mu1_skew_fft, Mu1_skew_wl, Photom_power,
      altitude_src, azimuth_src;
  int warnings_flag, reject_flag, no_coordinates = 0;
  double total_weight_batch;

  /* Time variables */
  time_t CurrentTime;
  struct tm *CurrentTimeStruct;
  char CurrentTimeString[MAX_STRING_LENGTH]; 

  /* Ground baseline length and azimuth */
  double baseline_length, ground_azim;
  
  /* Wavelength limits */
  double min_wave_fft, min_wave_wl, max_wave_fft, max_wave_wl;
  
  /* Standard format object coordinates */
  double RA_h, RA_m, RA_s, Dec_d, Dec_m, Dec_s;

    int autotest_mode = 0; /* equals one in case of Autotest */

  /*	fringe_a_table_def	*datadef; */
  /* fringe_a_table_def	*datadef = NULL; removed pkervell - 04-Aug-02 */
  /* opd_data	*opd_rate;  removed pkervell - 04-Aug-02 */
  
  /* Initialize default values */
  sprintf(directory_process, "./"); /* modified pkervell 19-mar-2002 */
  sprintf(protocol_mode, "Disk"); /* modified pkervell 13-mar-2002 */
  sprintf(infile, "VINCI.");
  sprintf(instrument,"VINCI");
  sprintf(extfile,".fits");
  sprintf(report_file,"dummy");
  strcpy(arcfile," ");
  sprintf(directory_process,".");
  
  /* Creates the processed data file for batch values with the current time as name */
  CurrentTime = time(NULL);
  CurrentTimeStruct = localtime(&CurrentTime);
  sprintf(CurrentTimeString, "%i_%02i_%02iT%02i_%02i_%02i",
	  CurrentTimeStruct->tm_year+1900,
	  CurrentTimeStruct->tm_mon+1,
	  CurrentTimeStruct->tm_mday,
	  CurrentTimeStruct->tm_hour,
	  CurrentTimeStruct->tm_min,
	  CurrentTimeStruct->tm_sec
	  );


  /*	Read User Commands */
  vndrs_DataReductionInput (argc, argv, &ierr); 
  CalculateMatrix = matrix_flag;

  /*	Allocate memory for the transfer matrix */
  Matrix = vndrs_calloc_matrix_double(2, 2);
  Matrix_buffer = vndrs_calloc_matrix_double(2, 2);
  Standard_dev_Matrix_buffer = vndrs_calloc_matrix_double(2, 2);
  Relative_MSE = vndrs_calloc_vector_double(2);
  matrix_variances = vndrs_calloc_matrix_double(2, 2);
  Wavenum_limits_fft = vndrs_calloc_vector_double(2);
  Wavenum_limits_wl = vndrs_calloc_vector_double(2);
  Wavenum_limits_QL = vndrs_calloc_vector_double(2);
      

  sprintf(batches_file_name, "%s%s%s", "Pro_", CurrentTimeString, ".rvf");
  /* 'rvf' stands for raw visibilities file */

  /* Writing header of the batches log file */

  if((f_batches_values = fopen(batches_file_name,"a")) == NULL)
      {
      fprintf(stderr, "  *** Problem encountered in opening the file for batches values output.\n");
      vndrs_error_alert();
      }
  fprintf(f_batches_values, "# VINCI Data Reduction Software \n");
  fprintf(f_batches_values, "# Error bars are 1 sigma (68%% confidence), computed using bootstrapping.\n");
  fprintf(f_batches_values, "# Photometric rejection threshold set to %.1f sigmas.\n", threshold);
  if(debias == 0)
      fprintf(f_batches_values, "# Mu^2 values are NOT debiased by off-sources. \n"); 
  if(depiston == 0)
      fprintf(f_batches_values, "# Mu^2 values are NOT corrected for the piston factor. \n"); 
  if(depiston == 1)
      fprintf(f_batches_values, "# Mu^2 values ARE corrected for the piston factor. \n"); 
  if(median_output == 1)
      fprintf(f_batches_values, "# Mu^2 values below are MEDIAN values over the batch. \n"); 
  else if(median_output == 0)
      fprintf(f_batches_values, "# Mu^2 values below are AVERAGE values over the batch. \n"); 

  fprintf(f_batches_values, "# infile  time_batch  name  catg  freq_Hz  N_scans_processed  alt_object  az_object  baseline azimuth  Mu^2_fft  ErrMu^2_fft  Mu^2_wl  ErrMu^2_wl  total_weight  warnings  reject_flags\n");

  fclose(f_batches_values);

  loop_process = 1.0; /* default value */
  
  /* PROCESSING OF A WHOLE DIRECTORY */
  
  if(directory_process_on == 1)   /* the user has specified the reduction of a whole directory */
      {
      /* Selects directory */ 
      if(chdir(directory_process)==0)
	  {
	  fprintf(stderr, "\n Directory selected for processing : %s",directory_process);
	  /* The system returns to the previous directory afterwards. pkervell 19-mar-02 */
	  }
      else
	  {
	  fprintf(stderr, "  *** Failed trying to go into the specified directory: %s \n", directory_process);
	  vndrs_error_alert();
	  }
      
      /* Build list of files and save it to the current directory. This command will not work outside
	 of UNIX (and maybe not in all UNIX versions) */
      sprintf(command_unix,"ls -1 *%s > list_VINCI_files.txt",extfile);
      system(command_unix);
      }
  
  /* ---------------------------------------------------------------------------------- */
  /* Main loop over which the pipeline is run if we have more than one file to process  */
  /* the user specifies an ASCII file containing the series of files to process through */
  /* the option '-list' in the command line call to vndrs_DataReduction.                */
  /* The loop is a bit complicated to stay compatible with single file operation.       */
  /* ---------------------------------------------------------------------------------- */
  
  /* CHECK IF WE HAVE SEVERAL FILES TO PROCESS */
  
  if((strlen(files_list) == 0) && (directory_process_on == 0))
    /* Case where we have only ONE file to process, as no file list has been defined */
    loop_process = 0;
  else if((strlen(files_list) != 0) | (directory_process_on != 0))
    {
      loop_process = 1;
      /* open file containing the list of names to process */
      if(strlen(files_list) != 0)
	{
	  if ((f_list = fopen(files_list,"r")) == NULL)
	    {
	      fprintf(stderr, "  *** Failed trying to open the specified file containing the names of the batches. %s\n", files_list);
	      vndrs_error_alert();
	    }
	}
      else if ((f_list = fopen("list_VINCI_files.txt","r")) == NULL)
	{
	  fprintf(stderr, "  *** Failed trying to open the automatically created file list 'list_VINCI_files.txt'\n");
	  vndrs_error_alert();
	}
    }






  /* Produces a list of the files for the night */
  /* ------------------------------------------ */
  /* added pkervell 08-Oct-02 */
  
    if(list_targets == 1)
	  {
	  /* Initialize runtime and memory usage protocol */
	  sprintf(protocol_file_name, "Observed_objects_list.log");
	  vndrs_initialization(protocol_mode, protocol_file_name);

	  /* Read the names of the first files to process */
	  if(fscanf(f_list,"%s\n%s\n%s\n%s\n", infile_off, infile_a, infile_b, infile_on) == EOF)
	      {
	      fprintf(stderr, "  *** No set of files in directory.\n");
	      vndrs_error_alert();
	      }
              
	  printf("\n List of sources in the selected directory:\n\n");
          printf("           On Source File                         Name    Freq      Scans    Frames\n");
          printf(" ----------------------------------------------------------------------------------\n");
	      
	  sprintf(vndrs_protocol.text, "# List of VINCI observed targets:\n"); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "# Directory selected : %s \n",directory_process); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "#           On Source File                 Name            Freq    Scans    Frames\n"); 
	  vndrs_runtime_protocol();

	  do
	      {
	      /* Read the necessary data files */

	      Off_source = vndrs_open_vinci_data(infile_off, -1);
	      Beam_A     = vndrs_open_vinci_data(infile_a, -1);
	      Beam_B     = vndrs_open_vinci_data(infile_b, -1);
	      On_source  = vndrs_open_vinci_data(infile_on, -1);
     
	      /* Reads the header for the acquisition parameters */
	      vndrs_OPD_get_info (&infile_on, &time_step, &delay_step,
				  &freq, &minfreq,
				  &maxfreq, &scan_mode, 
				  &fringespeed, &framerate, 
				  &name, &Date_Obs, &Observer,
				  &band, &station1, &station2, 
				  &telescope, &OBname, 
				  &RA, &Dec, 
				  &equinox, &Catg, &ReadMode,
				  &t1x, &t1y, &t1z,
				  &t2x, &t2y, &t2z,
				  &t3x, &t3y, &t3z,
				  &ierr);

	      vndrs_read_vinci_header(infile_on, &obsmode, &nscan, &nfram, &insmode, &acqmode, &fringespeed, 
	     &framerate, &mjdobs);

             /* IMPORTANT: Correction of the incorrect fringe speed !! pkervell 06-Nov-02 */
             fringespeed = 1.0316 * fringespeed;
             
             /* Chooses the effective wavelength corresponding to the filter used */
             if(strncmp(band,"K",1)==0) fringes_wavelength = 2.195e-6;
             else if(strncmp(band,"H",1)==0)
             {
                fringes_wavelength = 1.670e-6;
                ionic_in = 1; /* If H band, then we are using IONIC - pkervell - 13-Feb-03 */
                min_wave_QL = 470000;  /*      Minimum wavenumber for QL (in m^-1) for IONIC */
                max_wave_QL = 850000;  /*      Maximum wavenumber for QL (in m^-1) for IONIC */
             }
             else fringes_wavelength = 2.195e-6;  /* default K band */

             printf(" %s    %15s    %4i      %4ld      %4ld \n", infile_on, name, (int)(fringespeed/fringes_wavelength),
                                (long)nscan, (long)nfram);	     
             sprintf(vndrs_protocol.text, " %s    %15s    %4i      %4ld      %4ld \n", infile_on, name,
                                (int)(fringespeed/fringes_wavelength), (long)nscan, (long)nfram); 
             vndrs_runtime_protocol();


	      /* finds the names of the next fits files to read */
	      if(fscanf(f_list,"%s\n%s\n%s\n%s\n", infile_off, infile_a, infile_b, infile_on) == EOF)
		  loop_process = -1;

	      vndrs_free_batch(Off_source);
	      vndrs_free_batch(Beam_A);
	      vndrs_free_batch(Beam_B);
	      vndrs_free_batch(On_source);
	      }
	  while(loop_process>0);

          /* Ending the protocol */
          vndrs_end_runtime_protocol();
	  
          fclose(f_list);
	  }

      /* Now we reopen the file containing the list, to position the file handle at the beginning */
      loop_process = 1;
      /* open file containing the list of names to process */
      if(strlen(files_list) != 0)
	{
	  if ((f_list = fopen(files_list,"r")) == NULL)
	    {
	      fprintf(stderr, "Failed trying to open the specified file containing the names of the batches. %s\n", files_list);
	      vndrs_error_alert();
	    }
	}
      else if ((f_list = fopen("list_VINCI_files.txt","r")) == NULL)
	{
	  fprintf(stderr, "Failed trying to open the automatically created file list 'list_VINCI_files.txt'\n");
	  vndrs_error_alert();
	}


  /* ------------------------------- */

  do
    {
      if(loop_process == 0)
	{
	  /*	Initialize runtime and memory usage protocol */
	  sprintf(protocol_file_name, "%s%s%s%s", infile, file_on, extfile, ".log");
	  vndrs_initialization(protocol_mode, protocol_file_name);
	  
	  /*    sprintf(vndrs_protocol.text,"Protocol file : %s\n", protocol_file_name); 
		vndrs_runtime_protocol();*/
	  
	  /* Name of the output file for processed data */
	  sprintf(outfile, "%s%s%s%s", infile, file_on, extfile, ".pro");
	  
	  /* Name of the report file */
	  sprintf(report_file, "%s%s%s%s" , infile, file_on, extfile, ".rpt");
	}
      else
	{
	    if(fscanf(f_list,"%s\n%s\n%s\n%s\n", file_off, file_a, file_b, file_on) == EOF)
		{
		loop_process = -1;
		break;
		}

	  /*	Initialize runtime and memory usage protocol */
	  sprintf(protocol_file_name, "%s%s", file_on, ".log");
	  vndrs_initialization(protocol_mode, protocol_file_name);
	  
	  /* Name of the output file for processed data */
	  sprintf(outfile, "%s%s", file_on, ".pro");
	  
	  /* Name of the report file */
	  sprintf(report_file, "%s%s" , file_on, ".rpt");
	}
      
      
      if(matrix_flag == 3)
	  {
	  /*	Set the (default) matrix elements */
	  /* pkervell 13-mar-02: updated values from October 2001 data */
	  /* set_element_matrix_double (Matrix, 0, 0, 1.5959);
	     set_element_matrix_double (Matrix, 0, 1, 0.3068);
	     set_element_matrix_double (Matrix, 1, 0, 0.6006);
	     set_element_matrix_double (Matrix, 1, 1, 0.7178); */
	  
	  /* pkervell 20-apr-02: Autotest data from october 2001 */
	  /*      Kappa1A        Kappa2A        Kappa1B        Kappa2B */
	  /*   1.026967e+00   5.970827e-01   1.170558e+00   1.064070e+01 */
	  /* For reference, we have :
	     Matrix(0, 0) = Kappa_1A
	     Matrix(0, 1) = Kappa_1B
	     Matrix(1, 0) = Kappa_2A
	     Matrix(1, 1) = Kappa_2B */
	  
	  /* pkervell: Autotest data from early 2002 */
	  set_element_matrix_double (Matrix, 0, 0, 1.026967);
	  set_element_matrix_double (Matrix, 0, 1, 1.170558);
	  set_element_matrix_double (Matrix, 1, 0, 0.5970827);
	  set_element_matrix_double (Matrix, 1, 1, 10.64070);
	  }
      
      
      /*==============================================================*/
      /*	READ INPUT DATA FROM FILE                                   */
      /*==============================================================*/ 
      if (strcmp(instrument,"FLUOR_ASCII") == 0) 
	{ 
	  /*  Read the data from ASCII file
	      -------------------------------*/
	  sprintf(vndrs_protocol.text, "# --------------------------------------------------\n"); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "#  FLUOR test data (from ASCII file) ...  |\n"); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "# --------------------------------------------------\n"); 
	  vndrs_runtime_protocol();
	  
	  On_source  = vndrs_load_fluor_test_data(0);
	  Off_source = vndrs_load_fluor_test_data(1);
	  Beam_A     = vndrs_load_fluor_test_data(2);
	  Beam_B     = vndrs_load_fluor_test_data(3);
	}
      else if (strcmp(instrument,"FLUOR_FITS") == 0) 
	{ 
	  /*	Read the data from FLUOR FITS file
		------------------------------------*/
	  sprintf(vndrs_protocol.text, "# --------------------------------------------------\n"); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "#  FLUOR FITS data file ...               |\n"); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "# --------------------------------------------------\n"); 
	  vndrs_runtime_protocol();
	  
	  nb_sources = vndrs_open_fluor_data(infile);
	  Off_source = vndrs_select_fluor_data(0, source_id);
	  On_source  = vndrs_select_fluor_data(1, source_id);
	  Beam_A     = vndrs_select_fluor_data(2, source_id);
	  Beam_B     = vndrs_select_fluor_data(3, source_id);
	  vndrs_close_fluor_data(); 
	}
      else if (strcmp(instrument, "VINCI") == 0) 
	{

	  /*	Read the data from VINCI FITS file
		------------------------------------*/ 	  
	  if(loop_process == 0) /* CASE OF SINGLE-FILE PROCESSING */
	    {
	      
	      printf("-----------------------------------------------------\n"); 
	      printf("         VINCI FITS data files : \n\n"); 
	      printf("     Off:  %s%s%s \n", infile, file_off, extfile); 
	      printf("  A only:  %s%s%s \n", infile, file_a, extfile); 
	      printf("  B only:  %s%s%s \n", infile, file_b, extfile); 
	      printf(" Fringes:  %s%s%s \n", infile, file_on, extfile); 
	      printf("-----------------------------------------------------\n"); 
	      
	      sprintf(vndrs_protocol.text, "#\n"); 
	      vndrs_runtime_protocol();
	      sprintf(vndrs_protocol.text, "# VINCI FITS data files : \n"); 
	      vndrs_runtime_protocol();
	      sprintf(vndrs_protocol.text, "# Off:      %s%s%s \n", infile, file_off, extfile); 
	      vndrs_runtime_protocol();
	      sprintf(vndrs_protocol.text, "# A only:   %s%s%s \n", infile, file_a, extfile); 
	      vndrs_runtime_protocol();
	      sprintf(vndrs_protocol.text, "# B only:   %s%s%s \n", infile, file_b, extfile); 
	      vndrs_runtime_protocol();
	      sprintf(vndrs_protocol.text, "# Fringes:  %s%s%s \n", infile, file_on, extfile); 
	      vndrs_runtime_protocol();
	      sprintf(vndrs_protocol.text, "#\n"); 
	      vndrs_runtime_protocol();
	      
	      sprintf(infile_on, "%s%s%s", infile, file_on, extfile);
	      sprintf(infile_off, "%s%s%s", infile, file_off, extfile);
	      sprintf(infile_a, "%s%s%s", infile, file_a, extfile);
	      sprintf(infile_b, "%s%s%s", infile, file_b, extfile);
	      strcpy(arcfile,"VINCI.");
	      strcat(arcfile,file_on); 
	      strcat(arcfile,extfile); 
	      
	      nb_sources = 1; 

	 
	      
	      /*	1. Export FITS header to ASCII file
			-------------------------------------*/ 
	      if (export_flag != 0)
		{
		/* sprintf(outfile_header, "%s%s%s", infile, file_on, extfile);
		   sprintf(outfile_data, "%s%s%s", infile, file_on, extfile);
		   strcat(outfile_header, "_head.txt");
		   strcat(outfile_data, "_data.txt"); */

		/* We don't want ':' characters in the output file name for export format */
		sscanf(file_on,"%10sT%2s:%2s:%2s.%3s", filename_date, filename_time_hr,
		       filename_time_min, filename_time_sec, filename_time_thou);
		sprintf(outfile_header, "%10sT%s-%s-%s-%s", filename_date, filename_time_hr,
		       filename_time_min, filename_time_sec, filename_time_thou);
		sprintf(outfile_data, "%sT%s-%s-%s-%s", filename_date, filename_time_hr,
			filename_time_min, filename_time_sec, filename_time_thou);
                        
		strcat(outfile_header, ".hdr");   /* modified pkervell 21-Jun-02 */
		strcat(outfile_data, ".dat");
		
                /* strcpy(outfile_header, infile_on);
                strcpy(outfile_data, infile_on);
                strcat(outfile_header, ".hdr");   modified pkervell 13-Jan-03 for Pascal Bord
		strcat(outfile_data, ".dat");
		vndrs_export_header(outfile_header, infile_on); */
		}
	      
	      /*	Initialize FITS output unless only export requested
			-----------------------------------------------------*/
	      strcpy(outfile_FITS, outfile);
	      status = 0;
	    }
	  else    /* CASE OF MULTI-FILE PROCESSING */
	    {
	      /* in the case of multiple files, we have the full name already */
	      
	      printf("-----------------------------------------------------\n"); 
	      printf("         VINCI FITS data files : \n\n"); 
	      printf("     Off:  %s \n", file_off); 
	      printf("  A only:  %s \n", file_a); 
	      printf("  B only:  %s \n", file_b); 
	      printf(" Fringes:  %s \n", file_on); 
	      printf("-----------------------------------------------------\n"); 
	      
	      sprintf(vndrs_protocol.text, "#\n"); 
	      vndrs_runtime_protocol();
	      sprintf(vndrs_protocol.text, "# VINCI FITS data files : \n"); 
	      vndrs_runtime_protocol();
	      sprintf(vndrs_protocol.text, "# Off:      %s \n", file_off); 
	      vndrs_runtime_protocol();
	      sprintf(vndrs_protocol.text, "# A only:   %s \n", file_a); 
	      vndrs_runtime_protocol();
	      sprintf(vndrs_protocol.text, "# B only:   %s \n", file_b); 
	      vndrs_runtime_protocol();
	      sprintf(vndrs_protocol.text, "# Fringes:  %s \n", file_on); 
	      vndrs_runtime_protocol();
	      sprintf(vndrs_protocol.text, "#\n"); 
	      vndrs_runtime_protocol();
	      
	      sprintf(infile_on, "%s", file_on);
	      sprintf(infile_off, "%s", file_off);
	      sprintf(infile_a, "%s", file_a);
	      sprintf(infile_b, "%s", file_b);
	      strcpy(arcfile,file_on); 
	      


	      /*	1. Export FITS header to ASCII file
			-------------------------------------*/ 
	      if (export_flag != 0)
		{
		/* sprintf(outfile_header, "%s", file_on);
		  sprintf(outfile_data, "%s", file_on);
		  strcat(outfile_header, "_head.txt");
		  strcat(outfile_data, "_data.txt"); */

		/* modified pkerell 21-Jun-02 */
		/* We don't want ':' characters in the output file name for export format */
		sscanf(file_on,"VINCI.%10sT%2s:%2s:%2s.%3s.fits", filename_date, filename_time_hr,
		       filename_time_min, filename_time_sec, filename_time_thou);
		sprintf(outfile_header, "%sT%s-%s-%s-%s", filename_date, filename_time_hr,
		       filename_time_min, filename_time_sec, filename_time_thou);
		sprintf(outfile_data, "%sT%s-%s-%s-%s", filename_date, filename_time_hr,
		       filename_time_min, filename_time_sec, filename_time_thou);
		strcat(outfile_header, ".hdr");
		strcat(outfile_data, ".dat");

                /* strcpy(outfile_header, infile_on);
                strcpy(outfile_data, infile_on);
                strcat(outfile_header, ".hdr"); 
		strcat(outfile_data, ".dat");
		vndrs_export_header(outfile_header, infile_on); */

		}
	      
	      /*	Initialize FITS output unless only export requested
			-----------------------------------------------------*/
	      strcpy(outfile_FITS, outfile);
	      status = 0;
	    }
	  
            sprintf(vndrs_protocol.text, "# WIENER FILTER COMPUTED FROM AVERAGE PSDs \n"); 
            vndrs_runtime_protocol();
	  
            if(background_per_scan == 1)
	    {
	      sprintf(vndrs_protocol.text, "# BACKGROUNDS REMOVED FOR EACH ONSRC SCAN INDIVIDUALLY \n"); 
	      vndrs_runtime_protocol();
	    }
            else
	    {
	      sprintf(vndrs_protocol.text, "# BACKGROUNDS COMPUTED FROM OFFSRC BATCH \n"); 
	      vndrs_runtime_protocol();
	    }
	  
            if(debias == 1)
	    {
	      sprintf(vndrs_protocol.text, "# OFF-SOURCE SCANS PROCESSED\n"); 
	      vndrs_runtime_protocol();
	    }
            else
	    {
	      sprintf(vndrs_protocol.text, "# OFF-SOURCE SCANS NOT PROCESSED\n"); 
	      vndrs_runtime_protocol();
	    }
	  
            if (export_flag == 2)
	    {
	      sprintf(vndrs_protocol.text, "# FILE PROCESSED IN EXPORT ONLY MODE, NO PROCESSING OF DATA WAS DONE \n"); 
	      vndrs_runtime_protocol();
	    }

            if(median_output == 1)
	    {
	      sprintf(vndrs_protocol.text, "# BATCH VALUES ARE MEDIAN VALUES \n"); 
	      vndrs_runtime_protocol();
	    }
            else if(median_output == 0)
            {
              sprintf(vndrs_protocol.text, "# BATCH VALUES ARE AVERAGE VALUES \n"); 
	      vndrs_runtime_protocol();
            }
	  
	  sprintf(vndrs_protocol.text, "# PHOTOMETRIC REJECTION THRESHOLD = %f sigmas \n", threshold); 
	  vndrs_runtime_protocol();
	  
	  /*	2.	Reading header information for fringe processing 
		Header information is read from all file types to cross-check validity of parameters. (pkervell 18-apr-02)
		-------------------------------------------------------*/
	  
	  vndrs_read_vinci_header(infile_off, &obsmode_off, &nscan, &nfram_off, &insmode_off, &acqmode, &fringespeed_off, 
				  &framerate_off, &mjdobs);
	  vndrs_read_vinci_header(infile_a, &obsmode_a, &nscan, &nfram_a, &insmode_a, &acqmode, &fringespeed_a, 
				  &framerate_a, &mjdobs);
	  vndrs_read_vinci_header(infile_b, &obsmode_b, &nscan, &nfram_b, &insmode_b, &acqmode, &fringespeed_b, 
				  &framerate_b, &mjdobs);
	  vndrs_read_vinci_header(infile_on, &obsmode, &nscan, &nfram, &insmode, &acqmode, &fringespeed, 
				  &framerate, &mjdobs);

	  /*
	    insmode contains the instrument mode: 0:UNDEFINED 1:AUTOTEST 2:AUTOCOLL 3:STARINTF
	    obsmode contains the current batch type: 0:UNDEFINED 1:OFFSOURCE 2:BEAMA 3:BEAMB 4:ONSOURCE
	    acqmode contains the scanning type: 0:No wave 1:Rising + Falling slopes 2:Rising
	  */

	  /* Test to check that we are really processing a correct series of files: OFF, A, B, ON */
	  if(obsmode_off != 1)
	      {
	      printf("\n  *** The file %s is not of the expected OFF SOURCE type.\n", infile_off);
	      printf("      Please check the directory or file list content.\n");
	      vndrs_error_alert();
	      }
	  if(obsmode_a != 2)
	      {
	      printf("\n  *** The file %s is not of the expected BEAM A type.\n", infile_a);
	      printf("      Please check the directory or file list content.\n");
	      vndrs_error_alert();
	      }
	  if(obsmode_b != 3)
	      {
	      printf("\n  *** The file %s is not of the expected BEAM B type.\n", infile_b);
	      printf("      Please check the directory or file list content.\n");
	      vndrs_error_alert();
	      }
	  if(obsmode != 4)
	      {
	      printf("\n  *** The file %s is not of the expected ON SOURCE type.\n", infile_on);
	      printf("      Please check the directory or file list content.\n");
	      vndrs_error_alert();
	      }

	  if((nfram_off != nfram) | (nfram_a != nfram) | (nfram_b != nfram))
	      {
	      printf("\n  *** Incoherent number of frames between files:\n");
	      printf(" Off source file %s frame number = %i \n", infile_off, nfram_off);
	      printf("     Beam A file %s frame number = %i \n", infile_a, nfram_a);
	      printf("     Beam B file %s frame number = %i \n", infile_b, nfram_b);
	      printf("  On source file %s frame number = %i \n\n", infile_on, nfram);
	      vndrs_error_alert();
	      }	      

	  if((framerate_a != framerate) | (framerate_b != framerate) | (framerate_off != framerate))
	      {
	      printf("\n  *** Incoherent camera frequency (frame rate) between files:\n");
	      printf(" Off source file %s camera frequency = %f \n", infile_off, framerate_off);
	      printf("     Beam A file %s camera frequency = %f \n", infile_a, framerate_a);
	      printf("     Beam B file %s camera frequency = %f \n", infile_b, framerate_b);
	      printf("  On source file %s camera frequency = %f \n\n", infile_on, framerate);
	      /* vndrs_error_alert(); removed pkervell 28-Jul-02 for older data that have a problem in header */
	      }	      

	  /* Fringe speed is not checked, at it is not properly initialized by the acquisition software, and
	     is kept at the value of the star just before, therefore it causes problems. If this is fixed in the
	     VINCI software, then the test could be reactivated - pkervell 18-apr-02
	  if((fringespeed_a != fringespeed) | (fringespeed_b != fringespeed) | (fringespeed_off != fringespeed))
	      {
	      printf(" Error: Incoherent fringe speed between files:\n");
	      printf(" Off source file %s shows a fringe speed of %f \n", infile_off, fringespeed_off);
	      printf("     Beam A file %s shows a fringe speed of %f \n", infile_a, fringespeed_a);
	      printf("     Beam B file %s shows a fringe speed of %f \n", infile_b, fringespeed_b);
	      printf("  On source file %s shows a fringe speed of %f \n", infile_on, fringespeed);
	      vndrs_error_alert();
	      }	      
	  */
	
	  if(N_frames_cut > nfram) N_frames_cut = nfram;
	  /* In case the number of frames selected for the extraction of the fringes is too large,
	     take the width of the scan. pkervell 13-mar-2002 */
	  
	  /*	3.	Loading data files in batch structures
		Only Off_source can use the user requested scan unmber
		----------------------------------------------------*/

	  Off_source = vndrs_open_vinci_data(infile_off, -1);  /* changed pkervell 21-apr-02 */ 
	  Beam_A     = vndrs_open_vinci_data(infile_a, -1);
	  Beam_B     = vndrs_open_vinci_data(infile_b, -1);
	  On_source  = vndrs_open_vinci_data(infile_on, numberOfScansRequested);
	  
	  /*	4.	Exporting batch structures to a single ASCII file
		-------------------------------------------------------*/
	  if (export_flag != 0)
	    vndrs_export_data(outfile_data, Off_source, On_source, Beam_A, Beam_B); 



	  /*	5.	Reading the auxiliary data from on-source file header
		to be saved in the log file.
		-------------------------------------------------------*/
	  vndrs_OPD_get_info (&infile_on, &time_step, &delay_step,
			      &freq, &minfreq,
			      &maxfreq, &scan_mode, 
			      &speed, &rate, 
			      &name, &Date_Obs, &Observer,
			      &band, &station1, &station2, 
			      &telescope, &OBname, 
			      &RA, &Dec, 
			      &equinox, &Catg, &ReadMode,
			      &t1x, &t1y, &t1z,
			      &t2x, &t2y, &t2z,
			      &t3x, &t3y, &t3z,
			      &ierr);
	
        /* Patch to introduce the missing coordinates of the early commissioning objects that have non-standard
            names. This list will have to be updated to be able to process all objects of early 2001. */
        /* Proper motions are neglected for all stars (negligible for baseline computation) */
        if(strcmp(name,"AUTOTEST") == 0) /* Autotest */
        {
            autotest_mode = 1;
            RA = 0.0; Dec = 0.0;
            RA_h = 0.0; RA_m = 0.0; RA_s = 0.0; Dec_d = 0.0; Dec_m = 0.0; Dec_s = 0.0;
        } else if(RA == 0.0 && Dec == 0.0)
        {   /* Case unknown star... */
            autotest_mode = 0;
            no_coordinates = 1;
            RA_h = 0.0; RA_m = 0.0; RA_s = 0.0; Dec_d = 0.0; Dec_m = 0.0; Dec_s = 0.0;
        } else autotest_mode = 0;
        
        if(strcmp(name,"tet-Cen") == 0) /* 14 06 40.9485 -36 22 11.836 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(14, 06, 40.9485, -36, 22, 11.836, &RA, &Dec);
        }
        else if(strcmp(name,"R-Crt") == 0) /* 11 00 33.8539 -18 19 29.581 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(11, 00, 33.8539, -18, 19, 29.581, &RA, &Dec);
        }
        else if(strcmp(name,"Delta-Ophiuci") == 0) /* 16 14 20.7395 -03 41 39.563 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(16, 14, 20.7395, -03, 41, 39.563, &RA, &Dec);
        }
        else if(strcmp(name,"SWVir") == 0)  /* 13 14 04.3850 -02 48 25.149 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(13, 14, 04.3850, -02, 48, 25.149, &RA, &Dec);
        }
        else if(strcmp(name,"Sigma-Librae") == 0) /* 15 04 04.2155 -25 16 55.073 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(15, 04, 04.2155, -25, 16, 55.073, &RA, &Dec);
        }
        else if(strcmp(name,"Sigma-Lib") == 0) /* 15 04 04.2155 -25 16 55.073 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(15, 04, 04.2155, -25, 16, 55.073, &RA, &Dec);
        }
        else if(strcmp(name,"Sigma_Librae") == 0) /* 15 04 04.2155 -25 16 55.073 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(15, 04, 04.2155, -25, 16, 55.073, &RA, &Dec);
        }
        else if(strcmp(name,"Alpha-Cen-A") == 0) /* 14 39 36.50  -60 50 02.3 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(14, 39, 36.50,  -60, 50, 02.3, &RA, &Dec);
        }
        else if(strcmp(name,"Alpha-Cen-B") == 0) /* 14 39 35.08  -60 50 13.8 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(14, 39, 35.08,  -60, 50, 13.8, &RA, &Dec);
        }
        else if(strcmp(name,"Alpha-Centauri") == 0) /* 14 39 36.50  -60 50 02.3 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(14, 39, 36.50,  -60, 50, 02.3, &RA, &Dec);
        }
        else if(strcmp(name,"Alpha-Centauri-A") == 0) /* 14 39 36.50  -60 50 02.3 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(14, 39, 36.50,  -60, 50, 02.3, &RA, &Dec);
        }
        else if(strcmp(name,"Alpha-Cent") == 0) /* 14 39 36.50  -60 50 02.3 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(14, 39, 36.50,  -60, 50, 02.3, &RA, &Dec);
        }
        else if(strcmp(name,"Alpha-Centauri-B") == 0) /* 14 39 35.08  -60 50 13.8 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(14, 39, 35.08,  -60, 50, 13.8, &RA, &Dec);
        }
        else if(strcmp(name,"Alpha-Canis-Majoris-A") == 0) /* 06 45 08.9173 -16 42 58.017 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(06, 45, 08.9173, -16, 42, 58.017, &RA, &Dec);
        }
        else if(strcmp(name,"Eta-Sagitarii") == 0) /* 18 17 37.6351 -36 45 42.070 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(18, 17, 37.6351, -36, 45, 42.070, &RA, &Dec);
        }
        else if(strcmp(name,"Delta-Virginis") == 0) /* 12 55 36.2078 +03 23 50.893 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(12, 55, 36.2078, +03, 23, 50.893, &RA, &Dec);
        }
        else if(strcmp(name,"Alpha-Hydrae") == 0) /* 09 27 35.3   -08 39 31.0 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(9, 27, 35.3, -8, 39, 31.0, &RA, &Dec);
        }
        else if(strcmp(name,"Epsilon-Scorpii") == 0) /* 16 50 09.8   -34 17 36.0 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(16, 50, 09.8, -34, 17, 36.0, &RA, &Dec);
        }
        else if(strcmp(name,"Alfa-Cmi") == 0) /* 07 39 18.1 +05 13 30.0 */
        {
            no_coordinates = 1;
            vndrs_convert_RA_Dec_to_decimal_degs(07, 39, 18.1, +05, 13, 30.0, &RA, &Dec);
        }
        else if(strcmp(name,"Sirius") == 0) /* 06 45 08.9173 -16 42 58.017 */
        {
            no_coordinates = 0; /* The coordinates are sometimes present */
            vndrs_convert_RA_Dec_to_decimal_degs(06, 45, 08.9, -16, 42, 58.0, &RA, &Dec);
        }
        else if(strcmp(name,"hd41511") == 0) /* 06 04 59.13 -16 29 04.0 */
        {
            no_coordinates = 0; /* The coordinates are sometimes present */
            vndrs_convert_RA_Dec_to_decimal_degs(06, 04, 59.13, -16, 29, 04.0, &RA, &Dec);
        }
        
        if(no_coordinates == 1)
        {
            printf("  *** Source coordinates taken from internal list. \n\n");
            sprintf(vndrs_protocol.text, "# Source coordinates taken from internal list. \n");
            vndrs_runtime_protocol();
        }
        
        /* Patch to be able to process the UT1-UT3 early batches where the base coordinates
        are wrong - pkervell 20-Jan-03
        # Station 1 name:   U1
        # Station 1 coordinates X, Y, Z (meters):
        +9.9249   +20.3346   +0.0000 
        # Station 2 name:   U3
        # Station 2 coordinates X, Y, Z (meters):
        -44.9146   -66.1831   +0.0094 */

        if(strcmp(station1,"U1") == 0 && strcmp(station2,"U3") == 0)
        {
                t1x =  +9.9249; t1y = +20.3346; t1z = +0.0000; /* U1 */
                t2x = -44.9146; t2y = -66.1831; t2z = +0.0094; /* U3 */
        }
        
        /* Patch to be able to process the UT1-UT4 early batches where the base coordinates
        are wrong - pkervell 03-Feb-03 */
        
        if(strcmp(station1,"U4") == 0 && strcmp(station2,"U1") == 0)
        {
                t1x = -103.3058; t1y = -43.9989; t1z = +0.0000; /* U4 */
                t2x = +9.9249; t2y = +20.3346; t2z = +0.0000; /* U1 */
        }

        /* Patch to be able to process the batches where the base coordinates
        are zero. This happened mainly for the E0-G0 baseline during the early
        commissioning of the VLTI - pkervell 18-Oct-02 */
        
        if((t1x==0.0 && t1y==0.0 && t1z==0.0 && t2x==0.0 && t2y==0.0 && t2z==0.0) &&
            strcmp(station1,"T") == 0 && strcmp(station2,"T") == 0)
        {
                t1x = -30.7599; t1y = +40.1963; t1z = -0.0016; /* E0 */
                t2x = -45.8958; t2y = +34.9903; t2z = +0.0000; /* G0 */
                printf("  *** Baseline coordinates not present in header. Using E0-G0. \n\n");
                sprintf(vndrs_protocol.text, "# Baseline coordinates not present in header. Using E0-G0. \n");
                vndrs_runtime_protocol();
        }

        /* Patch to be able to process the batches where the base coordinates
        are zero. - pkervell 09-Feb-03 */
        
        if((t1x==0.0 && t1y==0.0 && t1z==0.0 && t2x==0.0 && t2y==0.0 && t2z==0.0) &&
            strcmp(station1,"E0") == 0 && strcmp(station2,"G1") == 0)
        {
                t1x = -30.7774; t1y = +40.1977; t1z = -0.0091; /* E0 */
                t2x = -66.7157; t2y = +95.5015; t2z = +0.0000; /* G1 */
                printf("  *** Baseline coordinates not present in header. Using E0-G1. \n\n");
                sprintf(vndrs_protocol.text, "# Baseline coordinates not present in header. Using E0-G1. \n");
                vndrs_runtime_protocol();
        }

        /* Patch to be able to process the 8m baseline batches. - pkervell 12-Feb-03 */
        
        if(strcmp(station1,"B3") == 0 && strcmp(station2,"C3") == 0)
        {
                t1x = -3.3476; t1y = +83.4805; t1z = +0.0000; /* B3 */
                t2x = -10.8939; t2y = +80.8678; t2z = +0.0006; /* C3 */
                printf("  *** Baseline coordinates B3-C3 taken from internal list. \n\n");
                sprintf(vndrs_protocol.text, "# Baseline coordinates B3-C3 taken from internal list. \n");
                vndrs_runtime_protocol();
        }

        /* Computation of standard format coordinates */
        if(autotest_mode == 0)
            vndrs_convert_decimal_degs_to_RA_Dec(RA, Dec,
                                             &RA_h, &RA_m, &RA_s, /* out */
                                             &Dec_d, &Dec_m, &Dec_s);
        
        
/*----------------------------------------------------------------------- */
/*   Inputs:                                                              */
/*      infile    input file with header keywords                         */
/*   Output:                                                              */
/*      time_step   time interval in seconds between frames in s          */
/*      delay_step  delay step between frames in meters                   */
/*      freq        "center" frequency of bandpass in Hz                  */
/*      minfreq     lowest frequency of bandpass in Hz                    */
/*      maxfreq     highest frequency of bandpass in Hz                   */
/*      scan_mode   Scaning mode, 0=none, 1=up+down, 2=up, 3=down only    */
/*      name        target name                                           */
/*      Date_Obs    Observing date                                        */
/*      Observer    Observer's name                                       */
/*      band        Observing band                                        */
/*      station1    first station name                                    */
/*      station2    second station name                                   */
/*      telescope   telescope name                                        */
/*      OBname      Bob blob name                                         */
/*      RA          Target RA (deg)                                       */
/*      Dec         Target Dec (deg)                                      */
/*      equinox     Coordinate equinox                                    */
/*      Catg        Observation category                                  */
/*      ReadMode    Readout Mode                                          */
/*      ierr        Error code: 0 => ok, checked on input                 */
/*----------------------------------------------------------------------- */
     
          /* IMPORTANT: Correction of the incorrect fringe speed of the piezo !! pkervell 06-Nov-02 */
          fringespeed = 1.0316 * fringespeed;

	  /* Chooses the effective wavelength corresponding to the filter used */
	  if(strncmp(band,"K",1)==0) fringes_wavelength = 2.195e-6;
	  else if(strncmp(band,"H",1)==0)
          {
            fringes_wavelength = 1.670e-6;
            /* Wavelength read on peak average position on sky of alf ind. TBC !! */
            ionic_in = 1; /* If H band, then we are using IONIC - pkervell 13-feb-03 */
            min_wave_QL = 470000;  /*      Minimum wavenumber for QL (in m^-1) for IONIC */
            max_wave_QL = 850000;  /*      Maximum wavenumber for QL (in m^-1) for IONIC */
          }
          else fringes_wavelength = 2.195e-6;  /* default K band */
	  
          Wavenum_theo = 1.0/(fringes_wavelength);
          
	  printf("        Target name:  %s \n", name); 
	  printf("            OB name:  %s \n", OBname); 
          printf("               Date:  %s \n", Date_Obs);
	  printf("    Right Ascension:   %02d:%02d:%05.2f =  %+011f deg\n", (int)RA_h, (int)RA_m, RA_s, RA);
	  printf("        Declination:  %+03d:%02d:%05.2f =  %+011f deg\n", (int)Dec_d, (int)Dec_m, Dec_s, Dec);
	  printf("            Equinox:  %s \n", equinox);
	  printf("    Telescopes used:  %s \n", telescope);
	  printf("     Station 1 name:  %s \n", station1);
	  printf("  Station 1 (X,Y,Z):  %+03.4f   %+03.4f   %+03.4f meters\n", t1x, t1y, t1z);
	  printf("     Station 2 name:  %s \n", station2);
	  printf("  Station 2 (X,Y,Z):  %+03.4f   %+03.4f   %+03.4f meters\n", t2x, t2y, t2z);
	  printf("    Rejection limit:  %.1f sigmas\n", threshold);
	  printf("             Filter:  %s (lambda = %1.3f microns) \n", band, fringes_wavelength*1e6);
	  printf("    Number of scans:  %i \n", nscan);
	  printf("   Number of frames:  %i \n", nfram);
          printf("     Scanning speed:  %i microns/s\n", (int)vndrs_round(speed*1e6));   
          printf("         Frame rate:  %5.2f Hz\n", rate);   
	  printf("  Fringes frequency:  %5.2f Hz\n", (fringespeed/fringes_wavelength) );

	  sprintf(vndrs_protocol.text, "#\n"); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "# Target name:      %s\n", name); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "# OB name:          %s\n", OBname); 
	  vndrs_runtime_protocol();
          sprintf(vndrs_protocol.text, "# Date:  %s \n", Date_Obs);
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "# Right Ascension (degrees): ( = %02d:%02d:%05.2f )\n", (int)RA_h, (int)RA_m, RA_s); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "  %+11f \n", RA); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "# Declination (degrees): ( = %+03d:%02d:%05.2f )\n", (int)Dec_d, (int)Dec_m, Dec_s); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "  %+11f \n", Dec); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "# Equinox:          %s\n", equinox); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "# Telescopes used:  %s\n", telescope); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "# Station 1 name:   %s\n", station1); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "# Station 1 coordinates X, Y, Z (meters):\n"); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "  %+03.4f   %+03.4f   %+03.4f \n", t1x, t1y, t1z); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "# Station 2 name:   %s\n", station2); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "# Station 2 coordinates X, Y, Z (meters):\n"); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "  %+03.4f   %+03.4f   %+03.4f \n", t2x, t2y, t2z); 
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "# Rejection limit:   %.1f \n", threshold);
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "# Maximum fringes jump:  %.1f \n", max_jump);          
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "# Filter:   %s   (lambda = %1.3f microns) \n", band, fringes_wavelength*1e6);
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "# Number of scans:   %i \n", nscan);
	  vndrs_runtime_protocol();
	  sprintf(vndrs_protocol.text, "# Number of frames:  %i \n", nfram);          
	  vndrs_runtime_protocol();
          sprintf(vndrs_protocol.text, "# Scanning speed:  %i microns/s\n", (int)vndrs_round(speed*1e6));   
	  vndrs_runtime_protocol();  
          sprintf(vndrs_protocol.text, "# Frame rate:  %5.2f Hz\n", rate);   
	  vndrs_runtime_protocol();  
	  sprintf(vndrs_protocol.text, "# Fringes frequency:  %5.2f Hz\n", (fringespeed/fringes_wavelength) );
	  vndrs_runtime_protocol();  
	  sprintf(vndrs_protocol.text, "#\n"); 
	  vndrs_runtime_protocol();  
	}
      else
	{ 
	  printf("  *** Unknown instrument mode %s (option -instr)\n",instrument);
	  vndrs_DataReductionUsage(); 
	  vndrs_error_alert();
	}
      
      baseline_length = sqrt((t2x-t1x)*(t2x-t1x) + (t2y-t1y)*(t2y-t1y) + (t2z-t1z)*(t2z-t1z)); /* in meters */
      if((t2x-t1x) != 0.0) ground_azim = atan2((t2y-t1y),(t2x-t1x))*180/Pi; else ground_azim = 0.0;
      if(ground_azim < 0.0) ground_azim += 360.0;
      
      /* The relationship linking baseline and integration limits is more or less a linear
      one, established from 3 baselines: 16 (sids, 165 and 296Hz), 66 (sids, 296 Hz),
      102.5m (UTs, 704 Hz). UTs peaks are a bit wider than SIDs ones, even extrapolating the relation
      to the longer baseline. It is MUCH better to err on the wide side for the integration band...
      The noise is pretty flat and of zero mean, while if some power of the peak is lost, it
      can create big biases. - pkervell 16-Oct-02 */

      if(ionic_in == 0)
      { /* MONA case */
        min_wave_fft = -359.87 * baseline_length + 294070;
        max_wave_fft = +317.01 * baseline_length + 656870;
        min_wave_wl  = -551.39 * baseline_length + 273980;
        max_wave_wl  = +707.19 * baseline_length + 696440;
        /* wavenumbers in m^-1 */
        
        if(strcmp(station1,"U1") == 0 ||
            strcmp(station1,"U2") == 0 ||
            strcmp(station1,"U3") == 0 ||
            strcmp(station1,"U4") == 0) 
        {/* UTs produce highly "vibrated" data - specific integration range - pkervell 27-Oct-02 */
                min_wave_fft = 200000.0;
                max_wave_fft = 800000.0;
                min_wave_wl  = 200000.0;
                max_wave_wl  = 800000.0;
        }
      }
      else
      { /* IONIC case */
        min_wave_fft = (2.19/1.58) * (-359.87 * baseline_length + 294070);
        max_wave_fft = (2.19/1.58) * (+317.01 * baseline_length + 656870);
        min_wave_wl  = (2.19/1.58) * (-551.39 * baseline_length + 273980);
        max_wave_wl  = (2.19/1.58) * (+707.19 * baseline_length + 696440);
        /* wavenumbers in m^-1 */
        
        if(strcmp(station1,"U1") == 0 ||
            strcmp(station1,"U2") == 0 ||
            strcmp(station1,"U3") == 0 ||
            strcmp(station1,"U4") == 0) 
        {/* UTs produce highly "vibrated" data - specific integration range - pkervell 27-Oct-02 */
                min_wave_fft = 300000.0;
                max_wave_fft = 1200000.0;
                min_wave_wl  = 300000.0;
                max_wave_wl  = 1200000.0;
        }
      }
      
      Wavenum_limits_QL->data[0] = min_wave_QL;
      Wavenum_limits_QL->data[1] = max_wave_QL;

      printf("    Ground baseline:  %.5f meters\n", baseline_length);
      printf("  Ground BL azimuth:  %5.2f degs (S=0, cw)\n", ground_azim);
      sprintf(vndrs_protocol.text, "#    Ground baseline: %.5f meters\n", baseline_length);
      vndrs_runtime_protocol();  
      sprintf(vndrs_protocol.text, "#  Ground BL azimuth:  %5.2f degs (S=0, cw)\n", ground_azim);
      vndrs_runtime_protocol();  
                                
      if(min_wave > 0.0 && max_wave > min_wave) /* if the user has specified integration limits */
      {
            Wavenum_limits_fft->data[0] = min_wave;
            Wavenum_limits_fft->data[1] = max_wave;
            
            Wavenum_limits_wl->data[0] = min_wave;
            Wavenum_limits_wl->data[1] = max_wave; 
                       
            sprintf(vndrs_protocol.text, "# User-specified wavenumber integration limits.\n");
            vndrs_runtime_protocol();  
      }
      else
      {
            Wavenum_limits_fft->data[0] = min_wave_fft;
            Wavenum_limits_fft->data[1] = max_wave_fft;
            
            Wavenum_limits_wl->data[0] = min_wave_wl;
            Wavenum_limits_wl->data[1] = max_wave_wl;            

            sprintf(vndrs_protocol.text, "# Wavenumber integration limits computed from baseline length.\n");
            vndrs_runtime_protocol();  
      }

        printf(" Wavenum limits FFT:  %4ld    %4ld cm^-1\n",
                                (long)(Wavenum_limits_fft->data[0]/100.0), (long)(Wavenum_limits_fft->data[1]/100.0));
        printf("  Wavenum limits WL:  %4ld    %4ld cm^-1\n",
                                (long)(Wavenum_limits_wl->data[0]/100.0), (long)(Wavenum_limits_wl->data[1]/100.0));
        printf("  Wavenum limits QL:  %4ld    %4ld cm^-1\n\n",
                                (long)(Wavenum_limits_QL->data[0]/100.0), (long)(Wavenum_limits_QL->data[1]/100.0));
        sprintf(vndrs_protocol.text, "# FFT (cm^-1): %6.3e  %6.3e\n",
                                Wavenum_limits_fft->data[0]/100.0, Wavenum_limits_fft->data[1]/100.0);
        vndrs_runtime_protocol();  
        sprintf(vndrs_protocol.text, "# WL  (cm^-1): %6.3e  %6.3e\n",
                                Wavenum_limits_wl->data[0]/100.0, Wavenum_limits_wl->data[1]/100.0);
        vndrs_runtime_protocol();  
        sprintf(vndrs_protocol.text, "# QL  (cm^-1): %6.3e  %6.3e\n",
                                Wavenum_limits_QL->data[0]/100.0, Wavenum_limits_QL->data[1]/100.0);
        vndrs_runtime_protocol();  
        
        

      /*	Allocate memory for on- and off-source batches, 
		on- and off-source power spectra and squared coherence factors
		------------------------------------------------------------*/
      /* modified 13-mar-2002 pkervell to allow for the truncation of the scans */
      
      Batch_on_processed = vndrs_calloc_batch(On_source->number_of_scans, 
					      N_frames_cut);

      
      Batch_off_processed = vndrs_calloc_batch(On_source->number_of_scans,
					       N_frames_cut);
      
            
      Mu1_squared = vndrs_calloc_vector_double(On_source->number_of_scans);
      Mu2_squared = vndrs_calloc_vector_double(On_source->number_of_scans);
      


      
      /*	Compute visibilities
		----------------------*/
      /* modified pkervell 12-mar-2002 */
      
      if (export_flag != 2)    /* If we are not in export only mode (without processing), go for data reduction ! :-) */
	{
	/* printf("\n Used Kappas:\n   %10e    %10e    %10e    %10e\n",
	   get_element_matrix_double(Matrix,0,0),
	   get_element_matrix_double(Matrix,0,1),
	   get_element_matrix_double(Matrix,1,0),
	   get_element_matrix_double(Matrix,1,1));*/

	  vndrs_compute_visibilities (/* Inputs */
	                              On_source, Off_source, Beam_A, Beam_B, CalculateMatrix,
				      framerate, fringespeed, shape, N_frames_cut, fringes_length, Wavenum_theo,
                                      Wavenum_limits_fft, Wavenum_limits_wl, Wavenum_limits_QL,
                                      process_all, use_mean_wiener, verbose_on, debias, depiston, median_output, 
                                      background_per_scan, center_wl,
                                      select_scans, max_jump, DUMMY_VALUE, DUMMY_VALUE,
                                      threshold, blockout, apod,
                                      MULT_CORR, bootstrap_number,
				      lat, longi, RA, Dec, t1x, t1y, t1z, t2x, t2y, t2z, ionic_in,
				      /* Inputs/Outputs */
				      Matrix,
				      /* Outputs */
				      Relative_MSE, Batch_on_processed, Batch_off_processed, 
                                      &N_scans_processed, &numberOfScansRejected,
				      Mu1_squared, Mu2_squared,
				      &time_batch,  &baseline_batch, &azimuth_batch, &u_batch, &v_batch,
                                      &altitude_src, &azimuth_src,
				      &Mu1_batch_wl, &ErrMu1_batch_min_wl, &ErrMu1_batch_max_wl, &ErrMu1_batch_wl,
				      &Mu2_batch_wl, &ErrMu2_batch_min_wl, &ErrMu2_batch_max_wl,
				      &Mu1_batch_fft, &ErrMu1_batch_min_fft, &ErrMu1_batch_max_fft, &ErrMu1_batch_fft,
				      &Mu2_batch_fft, &ErrMu2_batch_min_fft, &ErrMu2_batch_max_fft,
                                      &background_fourier, &background_wavelets,
                                      &background_fourier_med, &background_wavelets_med,
                                      &effective_wavelength, &effective_wavelength_wl, &effective_wavelength_error,
                                      &Mu1_skew_fft, &Mu1_skew_wl,
                                      &Photom_power, &total_weight_batch,
                                      &warnings_flag, &reject_flag);


        vndrs_free_batch(On_source);
        vndrs_free_batch(Off_source);
        vndrs_free_batch(Beam_A);
        vndrs_free_batch(Beam_B);


        /* Saving the results of the last processed batch to the observation log file of the night
            pkervell 15-apr-02 */
        if(N_scans_processed > 0)
        {
            if((f_batches_values = fopen(batches_file_name,"a")) == NULL)
            {
                fprintf(stderr, "  *** Problem encountered in opening the file for batches values output.\n");
                vndrs_error_alert();
            }
            
            fprintf(f_batches_values,
    " %20s  %7.8f  %20s  %10s  %3ld  %5ld  %7.3f %7.3f  % 010.7e  % 07.4e  % 07.4e  % 07.4e  % 07.4e  % 07.4e  %7ld  %1d  %1d\n",
                        infile_on, time_batch, name, Catg,
                        (long)(fringespeed/fringes_wavelength), N_scans_processed,
                        altitude_src, azimuth_src,
                        baseline_batch, azimuth_batch, /* effective_wavelength, effective_wavelength_error,*/
                        Mu1_batch_fft, ErrMu1_batch_fft, Mu1_batch_wl, ErrMu1_batch_wl,
                        (long)total_weight_batch, warnings_flag, reject_flag);
            
            if(median_output == 0)
                printf(" Average values were written in batch log file : %s \n",batches_file_name);
            else if(median_output == 1)
                printf(" Median values were written in batch log file : %s \n",batches_file_name);
            
            fclose(f_batches_values);
        }


	  /* Save the processed scans to an ASCII file - pkervell 14-apr-02 */
          /* This is not exactly the same as the verbose_on = 2 option that outputs the
          raw and processed scans directly from the MainRoutines, but I remove it from
          the list of options in the command line help - pkervell 15-Oct-03 */
          
	  Processed_I1 = vndrs_calloc_matrix_double(N_scans_processed,N_frames_cut);
	  Processed_I2 = vndrs_calloc_matrix_double(N_scans_processed,N_frames_cut);
	  
	  if(output_ascii == 1)
	      {
	      /* on source processed scans */
	      e=0;
	      for(i=0;i<nscan;i++)
		  if(Batch_on_processed->fullscans[i]->I1->data[0] > 0.5)
                      /* only the processed scans, for which index 0 value is 1 */
		      {
		      for(j=0;j<N_frames_cut;j++)
			  {
			  set_element_matrix_double(Processed_I1,e,j,Batch_on_processed->fullscans[i]->I1->data[j]);
			  set_element_matrix_double(Processed_I2,e,j,Batch_on_processed->fullscans[i]->I2->data[j]);
			  }
		      e++;
		      }
	      
	      sprintf(outfile_data, "%s", file_on);
	      strcat(outfile_data, "_I1_on_.dat");
	      vndrs_fprintf_matrix_double(Processed_I1, outfile_data);
	      
	      sprintf(outfile_data, "%s", file_on);
	      strcat(outfile_data, "_I2_on_.dat");
	      vndrs_fprintf_matrix_double(Processed_I2, outfile_data);
	      
	      /* off source scans */
	      e=0;
	      for(i=0;i<nscan;i++)
		  if(Batch_on_processed->fullscans[i]->I1->data[0] > 0.5)
                  /* only the processed scans, for which index 0 value is 1 */
		      {
		      for(j=0;j<N_frames_cut;j++)
			  {
			  set_element_matrix_double(Processed_I1,e,j,Batch_off_processed->fullscans[i]->I1->data[j]);
			  set_element_matrix_double(Processed_I2,e,j,Batch_off_processed->fullscans[i]->I2->data[j]);
			  }
		      e++;
		      }
	      
	      sprintf(outfile_data, "%s", file_on);
	      strcat(outfile_data, "_I1_off.dat");
	      vndrs_fprintf_matrix_double(Processed_I1, outfile_data);
	      
	      sprintf(outfile_data, "%s", file_on);
	      strcat(outfile_data, "_I2_off.dat");
	      vndrs_fprintf_matrix_double(Processed_I2, outfile_data);
	      	      
	      sprintf(outfile_data, "%s", file_on);
	      strcat(outfile_data, "_I1_pow.dat");
	      vndrs_fprintf_matrix_double(Processed_I1, outfile_data);
	      
	      sprintf(outfile_data, "%s", file_on);
	      strcat(outfile_data, "_I2_pow.dat");
	      vndrs_fprintf_matrix_double(Processed_I2, outfile_data);
	      
	      /* ----------------------------------- */
	      }


	} /*if (export_flag !=2) l 1045 */
      
      
      if(loop_process == 0) /* CASE OF SINGLE-FILE PROCESSING */
	{
	  printf("\n-----------------------------------------------------\n"); 
	  printf("             VINCI FITS Fringes File :\n"); 
	  printf("        %s%s%s \n",infile, file_on, extfile); 
	  printf("              PROCESSED successfully \n"); 
	  printf("-----------------------------------------------------\n"); 
	  
	  
	  /*	Export processed data to ASCII file
		-------------------------------------*/
	  /*  removed pkervell 21-Jun-2002
	  sprintf(outfile_proc, "%s%s", infile, file_on);
	  strcat(outfile_proc, "_proc.txt");
	  if(export_flag != 0)
	    vndrs_export_proc (outfile_proc, Batch_on_processed); 
	  */
	}
      else      /* CASE OF MULTI-FILE PROCESSING */
	{
	  printf("\n-----------------------------------------------------\n"); 
	  printf("             VINCI FITS Fringes File :\n"); 
	  printf("        %s \n", file_on); 
	  printf("              PROCESSED successfully \n"); 
	  printf("-----------------------------------------------------\n"); 
	  
	  
	  /*	Export processed data to ASCII file
		-------------------------------------*/
	  /*  removed pkervell 21-Jun-2002
	  sprintf(outfile_proc, "%s", file_on);
	  strcat(outfile_proc, "_proc.txt");
	  if(export_flag != 0)
	    vndrs_export_proc (outfile_proc, Batch_on_processed);
	  */
	}
      
      if(verbose_on == 1)
	{
	  vndrs_fprintf_vector_double(Mu1_squared, "Mu1Squared.dat");
	  vndrs_fprintf_vector_double(Mu2_squared, "Mu2Squared.dat");
	  vndrs_fprintf_matrix_double(Matrix, "TransferMatrix.dat");
	  vndrs_fprintf_vector_double(Relative_MSE, "RelativeMSE_Matrix.dat");
	}
      
      
     if(export_flag != 2) 
	{
	  
	  /*	Refine OPD estimates and get rates
		------------------------------------*/
/*	  status= 0;
 	  vndrs_OPD_fit(infile_on, On_source, Batch_on_processed, Off_source, &opd_rate, &status);
	  if (status != 0) 
	    { 
	      table_print_error_stack();
	      fprintf(stderr, "Error determining OPD/rates\n");
	      vndrs_error_alert();
	    }*/
	  
	  
	  /*	Write report if requested
		---------------------------*/
/*	  if (strlen(report_file) > 0) 
	    {
	      vndrs_PipelineReport (report_file, file_on, opd_rate); 
              vndrs_QC1_Report (report_file, 
				arcfile, 
				opd_rate, 
				N_scans_processed, 
				numberOfScansRejected,
				Matrix); 
	    }*/
	  
        
	  /*	Create FITS file
		------------------*/ 
/*	  if (createFitsFile) 
	    {
	      status = 0;
	      datadef = vndrs_make_data_table (infile_on, outfile_FITS, &outfptr, &tel1, &tel2, &source, 
					       &batch_id, &status, shape, threshold, apod, blockout);
	      if ((!datadef) || (status != 0)) 
		{ 
		  table_print_error_stack();  
		  fprintf(stderr, "Error %d initializing output FITS %s\n", 
			  status,outfile_FITS);
		  vndrs_error_alert();
		}
	    }*/
	  
        
	  /*	Write processed data to FITS file
		-----------------------------------*/
/*	  if (createFitsFile)
            { 
	      status = 0;
	      vndrs_write_data_table (outfptr, datadef, Batch_on_processed, opd_rate, tel1, tel2, source, 
				      batch_id, infile_on, infile_off, infile_a, infile_b, Matrix, shape, &status);
	      if (status != 0) 
		{
		  table_print_error_stack();  
		  fprintf(stderr, "Error writing output FITS %s\n", outfile_FITS);
		  vndrs_error_alert();
		}
	      fits_close_file(outfptr, &status);
	   }*/
	}

      
      /*	Free FITS pointers 
		--------------------*/ 
/*       if (createFitsFile) kill_fringe_a_table_def(datadef);
      datadef = NULL;*/
      
  
      /* vndrs_free_opd_data(opd_rate); removed pkervell 04-Aug-02 */
      vndrs_free_vector_double(Mu1_squared);
      vndrs_free_vector_double(Mu2_squared);
      vndrs_free_batch(Batch_on_processed);
      vndrs_free_batch(Batch_off_processed);
      vndrs_free_matrix_double(Processed_I1);
      vndrs_free_matrix_double(Processed_I2);
    }
  while(loop_process == 1);


  vndrs_free_vector_double(Wavenum_limits_fft);
  vndrs_free_vector_double(Wavenum_limits_wl);
  vndrs_free_vector_double(Wavenum_limits_QL);
  vndrs_free_matrix_double(Matrix);
  vndrs_free_matrix_double(Matrix_buffer);
  vndrs_free_matrix_double(Standard_dev_Matrix_buffer);
  vndrs_free_matrix_double(matrix_variances);
  vndrs_free_vector_double(Relative_MSE);

  /* close the file containing the names of the batches */
  if(loop_process = -1)
      fclose(f_list);
  
  /* removes the file list */
  system("rm list_VINCI_files.txt");

  vndrs_exit();
  return ierr;
}






void vndrs_DataReductionUsage(void)
     /*-----------------------------------------------------------------------*/
     /*   Tells about usage of program and bails out                          */
     /*-----------------------------------------------------------------------*/
{
  fprintf(stderr, " VINCI DATA PROCESSING SOFTWARE \n");
  fprintf(stderr, " Usage: vndrs_DataReduction [option] filenames ...\n\n");
  fprintf(stderr, " This command reduces the data from the VINCI instrument of the VLT Interferometer\n\n");
  fprintf(stderr, " Arguments:\n");
  fprintf(stderr, "  -list ...     Name of an ASCII file containing the names of the files to process\n");
  fprintf(stderr, "                (one per line, order: Off, A, B, On).\n");
  fprintf(stderr, "  -dir ...      Using this option, all files in the given directory are processed (default: '.').\n");
  fprintf(stderr, "                * Do not use this option in conjunction with '-list'\n");
  fprintf(stderr, "                * If you want to process the current directory, specify './'\n");
  fprintf(stderr, "                * Be careful to leave ONLY (off, a, b, on) files with the specified extension (.fits by default)\n");
  fprintf(stderr, "                * NO 'abandoned' mixed beam file, or it will produce incoherent results...\n");
  fprintf(stderr, "                * If neither list, nor dir, nor -file_... are used, then the current directory is processed.\n");
  fprintf(stderr, "  -protocol ... Specifies where the runtime protocol will be written (default: Disk).\n");
  fprintf(stderr, "                Can be either 'Nowhere' or 'Disk'.\n");
  fprintf(stderr, "  -in ...       Root name of FITS file to contain the data (default: VINCI.) \n");
  fprintf(stderr, "  -source ...   Source identifier (integer identifying the source, default 1)\n");
  fprintf(stderr, "  -ext ...      File extension (default: .fits) \n");
  fprintf(stderr, "  -export n     Export flag (0=no export, 1=export+process, 2=export only, default: 0) \n");
  fprintf(stderr, "  -file_off ... Off_source file name \n");
  fprintf(stderr, "  -file_on ...  On_source file name \n");
  fprintf(stderr, "  -file_a ...   Beam A file name \n");
  fprintf(stderr, "  -file_b ...   Beam B file name \n");
  fprintf(stderr, "  -instr ...    Instrument: VINCI, FLUOR_ASCII, FLUOR_FITS (default: VINCI) \n");
  /* fprintf(stderr, "  -report ...   Report file for index of processing (default: 'on_source_name'.rpt) \n");*/
  /* removed pkervell 13-mar-2002 */
  fprintf(stderr, "  -nscans nnn   Number of scans requested. Applies to on source fringes scans only. (default: -1 = all) \n");
  /* fprintf(stderr, "  -apod nnn     Apodization transition factor (default 0.1) \n");
  fprintf(stderr, "  -blockout nnn Apodization blockout factor (default 0.0) \n");*/
  fprintf(stderr, "  -threshold nn Photometric rejection threshold (number of sigmas), (default: 5) \n");
  fprintf(stderr, "  -noselect     De-activate the wavelet-based selection of the scans \n");
  fprintf(stderr, "  -min_wave nnn Minimum wavenumber (floating point, in cm-1), default: computed for baseline \n");
  fprintf(stderr, "  -max_wave nnn Maximum wavenumber (floating point, in cm-1), default: computed for baseline \n");
  fprintf(stderr, "  -shape nnn    Shape factor expressed in meters (default: 0.000011, measured in Autotest) \n");
  fprintf(stderr, "  -process_all  Process scans that have not shown fringes on QL (default: not activated) \n");
  fprintf(stderr, "  -cut nnn      Size in pixels of the cut and recentered scans (default: min(256, scan length)) \n");
  fprintf(stderr, "  -fr_length nn Specifies the length of the fringes for integration by wavelets (in pixels, default : 100)\n");
  fprintf(stderr, "  -max_jump nn  Specifies the maximum acceptable jump of fringes from scan to scan (in microns, default : 30)\n");
  fprintf(stderr, "  -median_out   Outputs the median values instead of averages as batch values (default : no) \n");
  fprintf(stderr, "  -verbose_on   Outputs all the intermediate processed data in ASCII format for each scan (default : no) \n");
  fprintf(stderr, "  -debias n     Remove bg (n=1) or not (n=0) using the off-source avg spectrum. (default : 0) \n");    
  fprintf(stderr, "  -depiston n   Apply (n=1) or not (n=0) the piston correction factor on mu2. (default : 0) \n");    
  fprintf(stderr, "  -bg n         Computes the backgrounds from the offsrc (= 0) or from the onsrc (default = 1).\n");
  fprintf(stderr, "  -center_wl nn Does fringes centering with wavelets with nn sigmas treshold (5 if no value specified, 0 to disable) \n");   
  fprintf(stderr, "  -bootstrap nn Specifies the number of bootstrapping for the estimation of biases (default: 3000) \n");   
  fprintf(stderr, "  -in_out_ascii Outputs the raw and processed scans to ASCII format files (1 obs. only, default: no)\n");   
/*  fprintf(stderr, "  -output_ascii    Outputs the processed scans and power spectra to ASCII format files (default: no)\n");
superseded by the -in_out_ascii option - pkervell 15-Oct-03 */
/*  fprintf(stderr, "  -ionic        Option to be used when IONIC beam combiner (one interf output) is used (default: no)\n");*/
  fprintf(stderr, "  -piston nn    Maximum residual after fringe peak 2D gaussian fit: 1.0 accepts all, 0.0 none (default: 0.3)\n");
  fprintf(stderr, "  -list_targets Outputs the list of all observed targets for the selected directory\n");
  fprintf(stderr, "  -matrix       Least squares fit if = 0, FLUOR method if = 1 (default 0)\n");
  fprintf(stderr, "\n");
  
  vndrs_error_alert();
}



/******************************************************************************
 *				European Southern Observatory
 *		 Vlti ViNci Data Reduction Software (VNDRS)
 *
 * Module name:	
 * Input/Output:	See function arguments to avoid duplication
 * Global values:
 *	infile	input FITS file name
 *	SI		Solution interval in scans
 *
 * Description:	Parse control info from command line
 *
 * History:		
 * 01-May-01		(csabet) Revised
 ******************************************************************************/
void vndrs_DataReductionInput (
			       int	  argc,	/*	In:	Number of arguments from command line */
			       char	**argv,	/*	In:	Array of strings from command line */
			       int	 *ierr)	/*	Ou:	Error code: 0 => ok */
{
  
  /*  Local Declarations
      --------------------*/
  int	ax;	
  char	*arg;
  
  
  /*  Algorithm
      -----------*/
  
#ifdef VNDRS_DATAREDUCTION_DEBUG
  fprintf(stderr, "number of arguments = %d\n",argc);
  if (argc<=1) vndrs_DataReductionUsage(); /* must have arguments */
#endif
  

  
  /*	Parse command line
	--------------------*/
  for (ax=1; ax<argc; ax++) 
    {
      arg = argv[ax];
      
#ifdef VNDRS_DATAREDUCTION_DEBUG
      fprintf(stderr, "arg = %s\n",arg);
#endif

      if (strcmp(arg, "-in") == 0) 
	strcpy(infile,argv[++ax]);
      else if (strcmp(arg, "-list") == 0) 
	{
	sscanf(argv[++ax],"%s", files_list);
	directory_process_on = 0; /* do not use with -dir option */
	}
      else if (strcmp(arg, "-ext") == 0) 
	strcpy(extfile,argv[++ax]);
      else if (strcmp(arg, "-out") == 0) 
	strcpy(outfile,argv[++ax]);
      else if (strcmp(arg, "-protocol") == 0) 
	strcpy(protocol_mode, argv[++ax]);
      else if (strcmp(arg, "-verbose_on") == 0) 
	verbose_on=1;    /* added pkervell 15-mar-2002 */
      else if (strcmp(arg, "-in_out_ascii") == 0)
	verbose_on=2;    /* added pkervell 15-oct-2003 */
      else if (strcmp(arg, "-median_out") == 0) 
	median_output=1;    /* added pkervell 20-apr-2002 */
      else if (strcmp(arg, "-debias") == 0)
      { 
	sscanf(argv[++ax],"%d",&debias);
        if (debias <= 0) debias = 0; else debias = 1;
      }
      else if (strcmp(arg, "-depiston") == 0)
      { 
	sscanf(argv[++ax],"%d",&depiston);
        if (depiston <= 0) depiston = 0; else depiston = 1;
      }
      else if (strcmp(arg, "-matrix") == 0) 
	sscanf(argv[++ax],"%d",&matrix_flag);
      else if (strcmp(arg, "-bootstrap") == 0)
	{
	  sscanf(argv[++ax],"%ld",&bootstrap_number);
	  if (bootstrap_number <= 0)
	     vndrs_DataReductionUsage();
	}      
      else if (strcmp(arg, "-nscans") == 0)
	{
	  sscanf(argv[++ax],"%ld",&numberOfScansRequested);
	  if (numberOfScansRequested <= 0)
	     vndrs_DataReductionUsage();
	}
      else if (strcmp(arg, "-max_jump") == 0)  /* modified pkervell 18-apr-2002 */
	  sscanf(argv[++ax],"%f",&max_jump);
      else if (strcmp(arg, "-export") == 0) 
	sscanf(argv[++ax],"%d",&export_flag);    /* added pkervell 13-mar-2002 */
      else if (strcmp(arg, "-source") == 0) 
	sscanf(argv[++ax],"%d",&source_id);
      else if (strcmp(arg, "-file_off") == 0) 
	strcpy(file_off,argv[++ax]);
      else if (strcmp(arg, "-file_on") == 0) 
	  {
	  directory_process_on = 0; /* do not use with -dir option */
	  strcpy(file_on,argv[++ax]);
	  }
      else if (strcmp(arg, "-file_a") == 0) 
	strcpy(file_a,argv[++ax]);
      else if (strcmp(arg, "-file_b") == 0)
	strcpy(file_b,argv[++ax]);
      else if (strcmp(arg, "-apod") == 0)        /* added pkervell 12-mar-2002 */
	sscanf(argv[++ax],"%f",&apod);
      else if (strcmp(arg, "-min_wave") == 0)        /* added pkervell 18-mar-2002 */
	sscanf(argv[++ax],"%f",&min_wave);
      else if (strcmp(arg, "-max_wave") == 0)        /* added pkervell 18-mar-2002 */
	sscanf(argv[++ax],"%f",&max_wave);
      else if (strcmp(arg, "-blockout") == 0)        /* added pkervell 12-mar-2002 */
	sscanf(argv[++ax],"%f",&blockout);
      else if (strcmp(arg, "-instr") == 0) 
	strcpy(instrument,argv[++ax]);
      /* else if (strcmp(arg, "-report") == 0)    modified pkervell 12-mar-2002 
        strcpy(report_file, argv[++ax]);*/
      else if (strcmp(arg, "-threshold") == 0)    /* added pkervell 12-mar-2002 */
	sscanf(argv[++ax],"%f",&threshold);
      else if (strcmp(arg, "-noselect") == 0)    /* added pkervell 21-Jan-04 */
	select_scans = 0;
      else if (strcmp(arg, "-shape") == 0)        /* added pkervell 12-mar-2002 */
	sscanf(argv[++ax],"%f",&shape);
      else if (strcmp(arg, "-process_all") == 0)  /* added pkervell 12-mar-2002 */
	process_all = 1;
      else if (strcmp(arg, "-bg") == 0)  /* added pkervell 11-Oct-2002 */
	sscanf(argv[++ax],"%d",&background_per_scan);
      else if (strcmp(arg, "-cut") == 0)          /* modified pkervell 15-apr-2002 */
	sscanf(argv[++ax],"%ld",&N_frames_cut);
      else if (strcmp(arg, "-fr_length") == 0)          /* added pkervell 18-mar-2002 */
	sscanf(argv[++ax],"%ld",&fringes_length);
      else if (strcmp(arg, "-output_ascii") == 0)          /* added pkervell 14-apr-2002 */
	output_ascii = 1; /* superseded by the -in_out_ascii option - pkervell - 15-Oct-03 */
/*      else if (strcmp(arg, "-ionic") == 0)
	  ionic_in = 1;*/
      else if (strcmp(arg, "-dir") == 0) 
	{
	  sscanf(argv[++ax],"%s", directory_process);
	  if(strlen(files_list) != 0) vndrs_DataReductionUsage();
	  /* do not use this option with -list */
	  directory_process_on = 1;
	}
      else if (strcmp(arg, "-list_targets") == 0)
	  list_targets = 1;  /* added pkervell - 08-Oct-02 */
      else /* unknown argument  */ 
	vndrs_DataReductionUsage();
    }
  
  /*	Return status
	---------------*/
  *ierr = 0;
  
}



void vndrs_PipelineReport (char *report_file, char *file_on, opd_data *opd)
     /*----------------------------------------------------------------------- */
     /*  Write processing report                                               */
     /*   Input:                                                               */
     /*     report_file  name of the report file to write                      */
     /*     file_on      name of input on-source data file                     */
     /*     opd          structure with results and descriptive info           */
     /*----------------------------------------------------------------------- */
{
  FILE *stream;
  char Full_Filename[255];
  char ttime[FLEN_CARD];
  
  /* Generate full filename */
  vndrs_generate_full_filename(report_file, Full_Filename);
  
  history_time(ttime); /* get current time/date */
  /* open output file to append */
  if ((stream = fopen(Full_Filename,"a")) == NULL)
    {
      fprintf(stderr, "Failed trying to open file %s\n", Full_Filename);
      vndrs_error_alert();
    }
  
  /* Write entry */
  
  fprintf (stream,"\n**********************\n");
  fprintf (stream,"   Instrument = VINCI\n");
  fprintf (stream,"   Instrument_mode(1) = undispersed\n");
  fprintf (stream,"   Instrument_mode(2) = fringe rate = %6.1f Hz\n",opd->fringerate);
  /*	fprintf (stream,"   Instrument_mode(3) = %d scans\n",opd->nscans); */
  fprintf (stream,"   Instrument_mode(3) = %ld scans\n",opd->nscans); /* csabet 09-Aug-2001 */
  fprintf (stream,"   Instrument_mode(4) = band = %s\n",opd->band);
  fprintf (stream,"   Telescope = %s\n",opd->telescope);
  fprintf (stream,"   Baseline = %s - %s\n",opd->station1, opd->station2);
  fprintf (stream,"   Observer = %s\n",opd->Observer);
  fprintf (stream,"   OBname = %s\n",opd->OBname);
  fprintf (stream,"   Date_Obs = %s\n",opd->Date_Obs);
  fprintf (stream,"   File_Obs = %s\n",file_on);
  fprintf (stream,"   Date_pipeline = %s\n", ttime);
  fprintf (stream,"   Target_Name = %s\n", opd->name);
  fprintf (stream,"   Data mu12 = %6.3f+/-%6.3f \n", opd->mu12,opd->emu12);
  fprintf (stream,"   Data mu22 = %6.3f+/-%6.3f \n", opd->mu22,opd->emu22);
  /*	fprintf (stream,"   Data sigma opd rate = %6.1lf u/s\n",1.0e6*opd->sigma_OPDdot); */
  fprintf (stream,"   Data sigma opd rate = %6.1f u/s\n",1.0e6*opd->sigma_OPDdot); /* csabet 09-Aug-2001 */
  /* close output */
  fclose(stream);
  
} /* end vndrs_PipelineReport */

void vndrs_QC1_Report (
		       char		*report_file, 
		       char		*file_on, 
		       opd_data	*opd,
		       long		N_scans_processed,
		       long		numberOfScansRejected,
		       matrix_double   *Matrix)
     /*----------------------------------------------------------------------- */
     /*  Write processing report                                               */
     /*   Input:                                                               */
     /*     report_file  name of the report file to write                      */
     /*     file_on      name of input on-source data file                     */
     /*     opd          structure with results and descriptive info           */
     /*----------------------------------------------------------------------- */
{
  FILE	*stream;
  char	Full_Filename[255], QC1comm[255];
  char	ttime[FLEN_CARD];
  float	SNRPA, SNRPB;
  
  /* photometric values */
  if (opd->sigma_ph[0]<=0.0) opd->sigma_ph[0] = 1.0;
  if (opd->sigma_ph[1]<=0.0) opd->sigma_ph[1] = 1.0;
  /* SNRPA = opd->mean_ph[0] / opd->sigma_ph[0]; */
  SNRPA = (float) (opd->mean_ph[0] / opd->sigma_ph[0]);	/* csabet 28.Jun.01 */
  /* SNRPB = opd->mean_ph[1] / opd->sigma_ph[1]; */
  SNRPB = (float) (opd->mean_ph[1] / opd->sigma_ph[1]);	/* csabet 28.Jun.01 */
  
  
  /* Generate full filename */
  vndrs_generate_full_filename(report_file, Full_Filename);
  
  
  /* Generate QC1 command */ 
  strcpy(QC1comm,"QC1LogWriter -s VINCI -f ");
  strcat(QC1comm,file_on);
  strcat(QC1comm," -k ");
  history_time(ttime); /* get current time/date */
  
  /* open output file to append */
  if ((stream = fopen(Full_Filename,"a")) == NULL)
    {
      fprintf(stderr, "Failed trying to open file %s\n", Full_Filename);
      vndrs_error_alert();
    }
  
  
  /* Write entry */
  fprintf (stream,"%s INSTRUMENT.MODE=undispersed\n",QC1comm);
  fprintf (stream,"%s FRINGE.RATE=%.1f\n",QC1comm,opd->fringerate);
  /*	fprintf (stream,"%s NSCANS=%d\n",QC1comm,opd->nscans); */
  fprintf (stream,"%s NSCANS=%ld\n",QC1comm,opd->nscans); /* csabet 09-Aug-2001 */
  fprintf (stream,"%s BAND=\"%s\"\n",QC1comm,opd->band);
  fprintf (stream,"%s TELESCOPE=\"%s\"\n",QC1comm,opd->telescope);
  fprintf (stream,"%s STATION1=\"%s\"\n",QC1comm,opd->station1);
  fprintf (stream,"%s STATION2=\"%s\"\n",QC1comm,opd->station2);
  fprintf (stream,"%s OBSERVER=\"%s\"\n",QC1comm,opd->Observer);
  fprintf (stream,"%s OB.NAME=\"%s\"\n",QC1comm,opd->OBname);
  fprintf (stream,"%s DATE.OBS=\"%s\"\n",QC1comm,opd->Date_Obs);
  fprintf (stream,"%s DATE.PIPE=\"%s\"\n",QC1comm,ttime);
  fprintf (stream,"%s TARGET.NAME=\"%s\"\n",QC1comm,opd->name);
  fprintf (stream,"%s MU12=%.6f\n",  QC1comm,opd->mu12);
  fprintf (stream,"%s DMU12=%.6f\n", QC1comm,opd->emu12);
  fprintf (stream,"%s MU22=%.6f\n",  QC1comm,opd->mu22);
  fprintf (stream,"%s DMU22=%.6f\n", QC1comm,opd->emu22);
  /*	fprintf (stream,"%s OPD.RATE=%.1lf\n",QC1comm,1.0e6*opd->sigma_OPDdot); */ /* csabet 09-Aug-2001 */
  fprintf (stream,"%s OPD.RATE=%.1f\n",QC1comm,1.0e6*opd->sigma_OPDdot); /* csabet 09-Aug-2001 */
  fprintf (stream,"%s RA=%.6f\n",QC1comm,opd->RA);
  fprintf (stream,"%s DEC=%.6f\n",QC1comm,opd->Dec);
  fprintf (stream,"%s EQUINOX=\"%s\" \n",QC1comm,opd->equinox);
  fprintf (stream,"%s SNRPA=%.1f\n",QC1comm,SNRPA);
  fprintf (stream,"%s SNRPB=%.1f\n",QC1comm,SNRPB);
  fprintf (stream,"%s CATG=\"%s\" \n",QC1comm,opd->Catg);
  fprintf (stream,"%s READMODE=\"%s\" \n",QC1comm,opd->ReadMode);
  fprintf (stream,"%s SCANS.PROCESSED=%ld\n",QC1comm, N_scans_processed);	
  fprintf (stream,"%s SCANS.REJECTED=%ld\n",QC1comm, numberOfScansRejected);
  fprintf (stream,"%s TRANSFER.MATRIX11=%.9f\n",QC1comm, get_element_matrix_double(Matrix,0,0));
  fprintf (stream,"%s TRANSFER.MATRIX12=%.9f\n",QC1comm, get_element_matrix_double(Matrix,0,1));
  fprintf (stream,"%s TRANSFER.MATRIX21=%.9f\n",QC1comm, get_element_matrix_double(Matrix,1,0));
  fprintf (stream,"%s TRANSFER.MATRIX22=%.9f\n",QC1comm, get_element_matrix_double(Matrix,1,1));
  /* close output */
  fclose(stream);
 
  
} /* end vndrs_PipelineReport */
