/* $Id: vndrs_FITS.c,v 1.10 2001/10/30 12:54:33 csabet Exp $ */
/*----------------------------------------------------------------------- */
/* Partially processed visibility data FITS binary tables routines        */
/* W. D. Cotton  March, 2001                                              */ 
/*----------------------------------------------------------------------- */
#include "vndrs_DataReduction.h"
#include "fitsio.h"
#include "vndrs_FITS.h"
#include "VLTIfileUtil.h"	/* csabet 28.Jun.01 */
#include "history.h"	/* csabet 28.Jun.01 */
#include <sys/stat.h>
#include <sys/types.h>

/* table of sources in source table */
#define MAX_SOURCE 10 /* maximum number of sources allowed */
#define MAX_BATCH 10  /* maximum number of batches allowed */
#define	AT_OVER_SID_DIAMETER_INIT	(1.8)	/*	AT/Sid diameter */
											/*	csabet 28.Jun.01 */
#define	AT_OVER_SID_DIAMETER_SMALL	(0.4)	/*	AT/Sid diameter */
											/*	csabet 28.Jun.01 */
#define	AT_OVER_SID_DIAMETER_LARGE	(8.0)	/*	AT/Sid diameter */
											/*	csabet 28.Jun.01 */

/* internal prototypes */
void vndrs_update_batch_table (fitsfile *outfptr, opd_data *opd,
			       int tel1, int tel2, int source, int batch_id,
			       char *file_on, char *file_off, 
			       char *file_a, char *file_b, 
			       matrix_double *Matrix, double shape, 
			       int *ierr);

float vndrs_FITS_mean_SNR(individual_scan *scan, double RON);

int vndrs_make_data_file (char *infile,
			  char *outfile,
			  fitsfile **infptr, 
			  fitsfile **outfptr, 
			  int *ierr,
			  double shape,
			  double threshold,
			  double apod,
			  double blockout
			  )
/*----------------------------------------------------------------------- */
/*   Create/init FITS file for VINCI processed data                       */
/*   If the output file doesn't exist it is created and initialized.      */
/*   Input file is assumed an ESO VLTI/VINCI file.                        */
/*   Inputs:                                                              */
/*      infile    input (on-source) file name                             */
/*      outfile   output file name                                        */
/*   Output:                                                              */
/*      infptr    cfitsio pointer to open input FITS file                 */
/*      outfptr   cfitsio pointer to open output FITS file                */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0=> file previously existed, 1=> was just created                 */
/*----------------------------------------------------------------------- */
{
  int doCrea = 1; /* was file created? */
  struct stat junk;
  long naxes[2] = {0,0};
  int  nkeys = 11;
  char *keys[] = {"HISTORY", "DATE-OBS", "MJD-OBS","ARCFILE ","TELESCOP",
		  "INSTRUME","ORIGIN", "EQUINOX", "RADECSYS", "UTC", "LST"};
  char hiline[FLEN_CARD], ttime[FLEN_CARD];
  float temp;
  int   itemp;

  /* existing error? */
  /* if (*ierr>0) return; */
	if (*ierr>0) return (*ierr);	/* csabet 28.Jun.01 */

  /* try opening input */
  fits_open_file (infptr, infile, READONLY, ierr);
  if (*ierr>0) {
      fprintf (stderr,"vndrs_make_data_file: error %d opening input %s\n",
	       *ierr, infile);
      return doCrea;
  }
  /* does output exist? */
  if (stat(outfile, &junk)) { /* no, create */
    fits_create_file (outfptr, outfile, ierr);
    fits_write_imghdr(*outfptr, 8, 2, naxes, ierr);
    /* need PRO TYPE for the archiving to recognize */
    fits_write_key_str (*outfptr, "HIERARCH ESO PRO TYPE ","RED",
			"VINCI pipeline product", ierr);
    /* copy keywords */
    copy_header_keywords (*infptr, *outfptr, nkeys, keys, ierr);
    if (*ierr>0) return doCrea;

    /* add history */
    sprintf(hiline,"*** VLTI/VINCI pipeline vndrs_DataReduction ***");
    fits_write_history (*outfptr, hiline, ierr);
    history_time(ttime); /* get current time/date */
    sprintf(hiline,"vndrs_DataReduction run %s", ttime);
    fits_write_history (*outfptr, hiline, ierr);
    /* temp = SHAPE_FACTOR; */
	temp = shape;	/* csabet 28.Jun.01 */
    sprintf(hiline,"vndrs_DataReduction SHAPE_FACTOR = %f",temp );
    fits_write_history (*outfptr, hiline, ierr);
    itemp = threshold;
    sprintf(hiline,"vndrs_DataReduction REJECTION_THRESHOLD = %d",itemp );
    fits_write_history (*outfptr, hiline, ierr);
    temp = blockout;
    sprintf(hiline,"vndrs_DataReduction APOD_BLOCKOUT = %f",temp );
    fits_write_history (*outfptr, hiline, ierr);
    /* temp = APOD_TRANSITION; */
	temp = (float) apod;	/* csabet 28.Jun.01 */
    sprintf(hiline,"vndrs_DataReduction APOD_TRANSITION = %f",temp );
    fits_write_history (*outfptr, hiline, ierr);
    itemp = MULT_CORR;
    sprintf(hiline,"vndrs_DataReduction MULT_CORR = %d",itemp );
    fits_write_history (*outfptr, hiline, ierr);
    /* temp = EPSILON_APODISATION_WINDOW; */
	temp = (float) EPSILON_APODISATION_WINDOW;	/* csabet 28.Jun.01 */
    sprintf(hiline,"vndrs_DataReduction EPSILON_APODISATION_WINDOW = %f",temp );
    fits_write_history (*outfptr, hiline, ierr);
    itemp = RADIUS;
    sprintf(hiline,"vndrs_DataReduction RADIUS = %d",itemp );
    fits_write_history (*outfptr, hiline, ierr);
    itemp = GENERAL_FILTER_ORDER;
    sprintf(hiline,"vndrs_DataReduction GENERAL_FILTER_ORDER = %d",itemp );
    fits_write_history (*outfptr, hiline, ierr);
    /* temp = COMPUTE_MATRIX_BLOCKOUT; */
	temp = (float) COMPUTE_MATRIX_BLOCKOUT;	/* csabet 28.Jun.01 */
    sprintf(hiline,"vndrs_DataReduction COMPUTE_MATRIX_BLOCKOUT = %f",temp );
    fits_write_history (*outfptr, hiline, ierr);
    temp = COMPUTE_MATRIX_TRANSITION;
    sprintf(hiline,"vndrs_DataReduction COMPUTE_MATRIX_TRANSITION = %f",temp );
    fits_write_history (*outfptr, hiline, ierr);
    temp = COMPUTE_MATRIX_CENTRAL_FREQ;
    sprintf(hiline,"vndrs_DataReduction COMPUTE_MATRIX_CENTRAL_FREQ = %f",temp );
    fits_write_history (*outfptr, hiline, ierr);
    temp = COMPUTE_MATRIX_HALF_BANDWIDTH;
    sprintf(hiline,"vndrs_DataReduction COMPUTE_MATRIX_HALF_BANDWIDTH = %f",temp );
    fits_write_history (*outfptr, hiline, ierr);

  } else { /* yes, open */
    fits_open_file (outfptr, outfile, READWRITE, ierr);
    if (*ierr>0) {
      fprintf (stderr,"vndrs_make_data_file: error %d create/open output %s\n",
	       *ierr, outfile);
      return doCrea;
    }
    doCrea = 0; /* previously exists */
  }

  return doCrea;
} /* end vndrs_make_data_file */



fringe_a_table_def* vndrs_make_data_table (char *infile,
					   char *outfile,
					   fitsfile **outfptr, 
					   int *tel1, int *tel2,
					   int *source, int *batch_id,
					   int *ierr,
					   double shape,  /* added pkervell 17-mar-02 */
					   double threshold,
					   double apod,
					   double blockout
					   )

/*----------------------------------------------------------------------- */
/*   Create/init FITS fringe_a table                                      */
/*   If the output file doesn't exist it is created and initialized with  */
/*   a source and array geometry table and the fringe_a table created     */
/*   Input file is assumed an ESO VINCI archive file                      */
/*   Inputs:                                                              */
/*      infile    input (on-source) file name                             */
/*      outfile   output file name                                        */
/*   Output:                                                              */
/*      outfptr   cfitsio pointer to open output FITS file                */
/*      tel1,2    telescope numbers                                       */
/*      source    source number                                           */
/*      batch_id  batch number in the batch table                         */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*       fringe_a_table_def structure to use to write table               */
/*       NULL on failure                                                  */
/*----------------------------------------------------------------------- */
{
  int mount = 3; /* siderostats for now */
  fringe_a_table_def *tabledef;
  fitsfile *infptr;
  float framrate, fringspd, maxwave, minwave;
  double mjdobs;
  /* char date_obs[30], filter[60], comment[60]; */
	char date_obs[30], filter[60];	/* csabet 28.Jun.01 */

  char commObs[60], commFilt[60], commRate[60], commSpd[60];
  char commMxwave[60],commMnwave[60], commMjdobs[60];
  char *default_str="        ";
  float default_flt = 0.0;
  double default_dbl = 0.0;
/*	long  default_lng = 0; */	/* csabet 28.Jun.01 */

  /* existing error? */
  if (*ierr>0) return NULL;

  /* create/init output FITS file */
  vndrs_make_data_file (infile, outfile, &infptr, outfptr, ierr, shape, threshold, apod, blockout);
  if (*ierr>0) return NULL;

  /* get observing date from input header */
  default_fits_read_key_str(infptr, "DATE-OBS", date_obs, commObs,  
			    (void*)default_str, ierr);
  if (*ierr>0) return NULL;
  default_fits_read_key_dbl(infptr, "MJD-OBS", &mjdobs, commMjdobs,  
			    (void*)&default_dbl, ierr);
  if (*ierr>0) return NULL;
  default_fits_read_key_str(infptr, "HIERARCH ESO INS FILT1 NAME", filter, 
		    commFilt,  (void*)default_str, ierr);
  if (*ierr>0) return NULL;

  /* frame rate in Hz */
  default_fits_read_key(infptr, TFLOAT, "HIERARCH ESO DET QL FRAMERATE", 
		&framrate, commRate,  (void*)&default_flt, ierr); 
  if (*ierr>0) return NULL;
  /* fringe speed (m/s)  */
  default_fits_read_key(infptr, TFLOAT, "HIERARCH ESO DET QL FRINGESPD", 
		&fringspd, commSpd,  (void*)&default_flt, ierr); 
  if (*ierr>0) return NULL;
  /* max wave length */
  default_fits_read_key(infptr, TFLOAT, "HIERARCH ESO DET QL MAXWAVEK", 
		&maxwave, commMxwave,  (void*)&default_flt, ierr); 
  if (*ierr>0) return NULL;
  /*  min wave length  */
  default_fits_read_key(infptr, TFLOAT, "HIERARCH ESO DET QL MINWAVEK", 
		&minwave, commMnwave,  (void*)&default_flt, ierr); 
  if (*ierr>0) return NULL;

  /* does the table already exist, try opening to see */
  fits_movnam_hdu (*outfptr, BINARY_TBL, "FRINGE_A", 1, ierr);
  if (*ierr== BAD_HDU_NUM) { /* doesn't exist, create */

    *ierr = 0;
    fits_clear_errmsg(); /* grumble */
    /* but first array geometry, optical train and source tables */
    vndrs_vlti_geometry (infptr, *outfptr, date_obs, mount, tel1, tel2, 
			 ierr);
    if (*ierr>0) return NULL; /* failed - bomb out */

    /* optical train table */
    vndrs_vlti_optical_train (infptr, *outfptr, *tel1, *tel2, ierr);
    if (*ierr>0) return NULL; /* failed - bomb out */

    /* source table */
    *source = vndrs_vlti_source (infptr, *outfptr, ierr);
    if (*ierr>0) return NULL; /* failed - bomb out */

    /* batch table */
    *batch_id = vndrs_vlti_batch (infptr, *outfptr, ierr);
    if (*ierr>0) return NULL; /* failed - bomb out */

    /* finally make data table */
    tabledef = 
      make_fringe_a_table_def(*outfptr, date_obs, 
			      framrate, fringspd, maxwave, minwave, 
			      filter, ierr);
    if ((tabledef==NULL) || (*ierr>0)) { /* error test */
      fprintf (stderr,"error %d after fringe table create\n",
	       *ierr);
      return NULL;  /* bomb out */
    }
  } else { /* exists, open */

    /* but first array geometry and source tables */
    vndrs_vlti_geometry (infptr, *outfptr, date_obs, mount, tel1, tel2, 
			 ierr);
    if (*ierr>0) return NULL; /* failed - bomb out */

    /* and then the source info */
    *source = vndrs_vlti_source (infptr, *outfptr, ierr);
    if (*ierr>0) return NULL; /* failed - bomb out */

    /* batch table */
    *batch_id = vndrs_vlti_batch (infptr, *outfptr, ierr);
    if (*ierr>0) return NULL; /* failed - bomb out */

    /* finally the data table */
    tabledef = open_fringe_a_table(*outfptr, ierr);
    if ((tabledef==NULL) || (*ierr>0)) return NULL; /* failed - bomb out */
  }

  return tabledef; /* done */

}  /* end VINCI_make_data_table */


void vndrs_write_data_table (fitsfile *outfptr, fringe_a_table_def *datadef, 
			     batch *batch_on, opd_data *opd,
			     int tel1, int tel2, int source, int batch_id,
			     char *file_on, char *file_off, 
			     char *file_a, char *file_b, 
                             matrix_double *Matrix, double shape,
			     int *ierr)
/*----------------------------------------------------------------------- */
/*   Write data to FITS fringe_a table                                    */
/*   Inputs:                                                              */
/*      outfptr   cfitsio pointer to open output FITS file                */
/*      datadef   output table definition structure                       */
/*      batch_on  processed on source batch,  scan_data member is used    */
/*      opd       batch and scan values of OPD and rate                   */
/*      tel1,2    telescope numbers                                       */
/*      source    source number                                           */
/*      batch_id  batch number in the batch table                         */
/*      file_on   on source raw FITS file name                            */
/*      file_off  off source raw FITS file name                           */
/*      file_a    Port A only raw FITS file name                          */
/*      file_b    Port B only raw FITS file name                          */
/*      Matrix    Kappa matrix.                                           */
/*      shape     shape factor (e.g. 0.000013)                            */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  fringe_a_row *row;
  /* long i, irow; */
	long			irow;	/* csabet 28.Jun.01 */
	unsigned int	i;		/* csabet 28.Jun.01 */
  char comment[80];

  if (*ierr>0) return; /* existing error conditions */

  /* update batch table */
  vndrs_update_batch_table (outfptr, opd, tel1, tel2, source, 
			    batch_id, file_on, file_off, file_a, file_b, 
			    Matrix, shape, ierr);
  if (*ierr>0) return;

  /* create table row structure */
  row = make_fringe_a_row (datadef);

  /* position in fringe_a table */
  fits_movnam_hdu (outfptr, BINARY_TBL, "FRINGE_A", 1, ierr);
  if (*ierr>0) return;

  /* common values */
  row->source = source;
  row->batch = batch_id;
  row->ot[0] = tel1;
  row->ot[1] = tel2;
  row->config = 1;

  /* as yet undetermined values */
  row->opd  = 0.0;
  row->opd1 = 0.0;
  row->opd2 = 0.0;
  row->mu12 = 0.0;
  row->mu22 = 0.0;
  row->snrp1 = 0.0;
  row->snrp2 = 0.0;
  row->uvw[0] = 0.0;
  row->uvw[1] = 0.0;
  row->uvw[2] = 0.0;
  row->timei = 0.0;

  /* add to end */
  fits_read_key_lng(outfptr, "NAXIS2", &irow, comment, ierr);

  /* loop over batch - writing data */
  for (i=0; i<batch_on->number_of_scans; i++) {
    row->time = batch_on->scan_data[i].time;
    if (opd->OPDresid[i]!=MAGICBLANK) {
      /* row->weight = opd->weight[i]; */
		row->weight = (float) (opd->weight[i]); /* csabet 28.Jun.01 */
      /* row->opd  = opd->OPDresid[i]; */
		row->opd  = (float) (opd->OPDresid[i]); /* csabet 28.Jun.01 */
      /* row->opd1  = opd->OPDdot[i]; */
		row->opd1  = (float) (opd->OPDdot[i]); /* csabet 28.Jun.01 */
      /* row->opd2  = opd->OPDdotdot[i]; */
		row->opd2  = (float) (opd->OPDdotdot[i]); /* csabet 28.Jun.01 */
    } else { /* bad solution */
      row->weight = 0.0;
      row->opd  =  0.0;
      row->opd1 =  0.0;
      row->opd2 =  0.0;
    }
    /* interferometric values */
    /* row->mu12 = batch_on->scan_data[i].mu1_squared; */
	row->mu12 = (float) (batch_on->scan_data[i].mu1_squared);	/* csabet 28.Jun.01 */
    /* row->mu22 = batch_on->scan_data[i].mu2_squared; */
	row->mu22 = (float) (batch_on->scan_data[i].mu2_squared);	/* csabet 28.Jun.01 */
    /* correlated power moments */
    /* row->corpower[0] = opd->corpower[3*i]; */
	row->corpower[0] = (float) (opd->corpower[3*i]);	/* csabet 28.Jun.01 */
    /* row->corpower[1] = opd->corpower[3*i+1]; */
	row->corpower[1] = (float) (opd->corpower[3*i+1]);	/* csabet 28.Jun.01 */
    /* row->corpower[2] = opd->corpower[3*i+2]; */
	row->corpower[2] = (float) (opd->corpower[3*i+2]);	/* csabet 28.Jun.01 */
    /* photometric SNR */
	/* modified pkervell 13-mar-2002: the mean SNR is now computed, using the readout noise
	   of the LISA pixels, set to their value as of now. This is important for the selection
	   in IDL-based software tools. See log book 001-10-29.pke for RON values.
	*/
	row->snrp1 = vndrs_FITS_mean_SNR(batch_on->fullscans[i]->PA, 1.97);
	row->snrp2 = vndrs_FITS_mean_SNR(batch_on->fullscans[i]->PB, 1.61);
    /* photometric values */
    /* row->photom[0] = opd->photom[2*i]; */
	row->photom[0] = (float) (opd->photom[2*i]);	/* csabet 28.Jun.01 */
    /* row->photom[1] = opd->photom[2*i+1]; */
	row->photom[1] = (float) (opd->photom[2*i+1]);	/* csabet 28.Jun.01 */
    /* write table unless it's bad */
    if (row->weight>0.0)
      write_fringe_a_table(outfptr, datadef, row, ++irow, ierr);
    if (*ierr>0) {
      fprintf (stderr,"vndrs_write_data_table: error %d writing output FITS\n",
	       *ierr);
      return;  /* bomb out */
    }
  } /* end loop writing data */
} /* end vndrs_write_data_table */


void vndrs_vlti_geometry(fitsfile *infptr, fitsfile *outfptr, 
			 char *date_obs, int mount,
			 int *tel1, int *tel2, int *ierr) 
/*----------------------------------------------------------------------- */
/*   Save Array Geometry for all of VLTI array                            */
/*   Copies data from input file.                                         */
/*   If an ARRAY_GEOMETRY table exists, the current telescopes are        */
/*   determined.                                                          */
/*   Values needed from primary header: (not yet available)               */
/*  HIERARCH ESO ISS CONF NTEL    long   number of telescopes             */
/*  HIERARCH ESO ISS CONF STATION1 string  pad name                       */
/*  HIERARCH ESO ISS CONF T1X     double X(m) coordinate                  */
/*  HIERARCH ESO ISS CONF T1Y     double Y(m) coordinate                  */
/*  HIERARCH ESO ISS CONF T1Z     double Z(m) coordinate                  */
/*  HIERARCH ESO ISS CONF STATION2 string pad name                        */
/*  HIERARCH ESO ISS CONF T2X     double X(m) coordinate                  */
/*  HIERARCH ESO ISS CONF T2Y     double Y(m) coordinate                  */
/*  HIERARCH ESO ISS CONF T2Z     double Z(m) coordinate                  */
/*  HIERARCH ESO ISS CONF STATION3 string  pad name                       */
/*  HIERARCH ESO ISS CONF T3X     double X(m) coordinate                  */
/*  HIERARCH ESO ISS CONF T3Y     double Y(m) coordinate                  */
/*  HIERARCH ESO ISS CONF T3Z     double Z(m) coordinate                  */
/*  HIERARCH ESO ISS GEOELEV      double VLTI site elevation              */
/*  HIERARCH ESO ISS GEOLAT       double VLTI site latitude               */
/*  HIERARCH ESO ISS GEOLONG      double VLTI site longitude              */
/*   Inputs:                                                              */
/*      infptr    cfitsio pointer to open FITS file for header info       */
/*      outfptr   cfitsio pointer to open FITS file for table             */
/*      date_obs  Date (yyyy-mm-dd) observations began                    */
/*      mount     mount type 0 = alt-az, 3=siderostat                     */
/*   Output:                                                              */
/*      tel1,2    telescope numbers                                       */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
  array_geometry_table_row *row;
  array_geometry_table_def *geomtabledef;
  /* int  ikey, i, isnew=0; */
	int		i, isnew=0;	/* csabet 28.Jun.01 */
  double xxx = 0.0, yyy = 0.0, zzz = 0.0, mjd;	/* csabet 09-Aug.2001. Compiler may not initialise */
  char t1name[11], t2name[11], telescop[21], comment[60];
  char name1[20], name2[20];
  double t1xyz[3], t2xyz[3], geolat, geolong, geoelev;
  double VLTIxyz[] ={0.0, 0.0, 0.0};
  long  ntel;
  char *default_str="        ";
/*	float default_flt = 0.0; */		/* csabet 28.Jun.01 */
  double default_dbl = 0.0;
  long  default_lng = 0;
  /* float diameter = 1.8;  */ /* AT/Sid diameter */
  double diameter = AT_OVER_SID_DIAMETER_INIT; /* AT/Sid diameter */	/* csabet 26.Jun.01 */
  
  if (*ierr>0) return; /* existing error condition? */
  if (!infptr) return; /* sanity check */
  if (!outfptr) return; /* sanity check */

  /* get info from input file header */

  default_fits_read_key(infptr, TSTRING, "TELESCOP", 
		&telescop, comment,  (void*)default_str, ierr); 
  if (*ierr>0) return;

  /* if this uses the siderostats change the diameter */
  /* if (!strncmp (telescop, "SID", 3)) diameter = 0.4; */
  if (!strncmp (telescop, "SID", 3)) diameter = (float) AT_OVER_SID_DIAMETER_SMALL;	/* csabet 28.Jun.01 */
  default_fits_read_key_dbl(infptr, "MJD-OBS", &mjd, comment,  
			    (void*)&default_dbl, ierr);
  if (*ierr>0) return;

  default_fits_read_key_lng(infptr, "HIERARCH ESO ISS CONF NTEL", &ntel, comment,  
			    (void*)&default_lng, ierr);
  if (*ierr>0) return;
  if (ntel < 2) ntel = 2;
  default_fits_read_key_str(infptr, "HIERARCH ESO ISS CONF STATION1", t1name, 
		    comment,  (void*)default_str, ierr);
  if (*ierr>0) return;
  default_fits_read_key_dbl(infptr, "HIERARCH ESO ISS CONF T1X", &t1xyz[0], comment,  
			    (void*)&t1xyz[0], ierr);
  if (*ierr>0) return;
  default_fits_read_key_dbl(infptr, "HIERARCH ESO ISS CONF T1Y", &t1xyz[1], comment,  
			    (void*)&t1xyz[1], ierr);
  if (*ierr>0) return;
  default_fits_read_key_dbl(infptr, "HIERARCH ESO ISS CONF T1Z", &t1xyz[2], comment,  
			    (void*)&t1xyz[2], ierr);
  if (*ierr>0) return;
  default_fits_read_key_str(infptr, "HIERARCH ESO ISS CONF STATION2", t2name,
		    comment,  (void*)default_str, ierr);
  if (*ierr>0) return;
  default_fits_read_key_dbl(infptr, "HIERARCH ESO ISS CONF T2X", &t2xyz[0], comment,  
			    (void*)&t2xyz[0], ierr);
  if (*ierr>0) return;
  default_fits_read_key_dbl(infptr, "HIERARCH ESO ISS CONF T2Y", &t2xyz[1], comment,  
			    (void*)&t2xyz[1], ierr);
  if (*ierr>0) return;
  default_fits_read_key_dbl(infptr, "HIERARCH ESO ISS CONF T2Z", &t2xyz[2], comment,  
			    (void*)&t2xyz[2], ierr);
  if (*ierr>0) return;
  default_fits_read_key_dbl(infptr, "HIERARCH ESO ISS GEOELEV", &geoelev, comment,  
			    (void*)&default_dbl, ierr);
  if (*ierr>0) return;
  default_fits_read_key_dbl(infptr, "HIERARCH ESO ISS GEOLAT", &geolat, comment,  
			    (void*)&default_dbl, ierr);
  if (*ierr>0) return;
  default_fits_read_key_dbl(infptr, "HIERARCH ESO ISS GEOLONG ", &geolong, comment,  
			    (void*)&default_dbl, ierr);
  if (*ierr>0) return;

  /* only the siderostats at E0 and G0 are available in the beginning */
  if (mjd<52100.0) {
    /* diameter = 0.4; */
    diameter = (float) AT_OVER_SID_DIAMETER_SMALL;	/* csabet 28.Jun.01 */
    t1xyz[0] = -30.741; t1xyz[1] = 40.186; t1xyz[2] = 0.0; /* flip signs */
    t2xyz[0] = -45.871; t2xyz[1] = 34.982; t2xyz[2] = 0.0;
    strcpy (t1name, "E0");
    strcpy (t2name, "G0");
  }
  /* only siderostats */
  strcpy (name1, "SID1    ");
  strcpy (name2, "SID2    ");

  *tel1 = 1;
  *tel2 = 2;

  /* see if table exists, if not create */
  fits_movnam_hdu (outfptr, BINARY_TBL, "ARRAY_GEOMETRY", 1, ierr);
  if (*ierr) { /* doesn't exist, create */

    *ierr = 0;
    fits_clear_errmsg(); /* grumble */
    isnew = 1;
    
    /* create array geometry table */
    geomtabledef = make_array_geometry_table_def(outfptr, "VLTI", date_obs,
						 VLTIxyz, ierr);
    if ((geomtabledef==NULL) || (*ierr>0)) { /* error test */
      fprintf (stderr,"error %d after array geometry table create\n",
	       *ierr);
      return;
    } /* end error test */ 
  } else { /* open existing one */
    geomtabledef = open_array_geometry_table(outfptr, ierr);
    isnew = 0;
    if ((geomtabledef==NULL) || (*ierr>0)) { /* error test */
      fprintf (stderr,"error %d after array geometry table open \n",
	       *ierr);
      return;
    } /* end error test */ 
  }
  
  /* make row structure */
  row = make_array_geometry_table_row (geomtabledef);
  
  /* init row structures */
  /* row->diameter = diameter;  */ /* diameter */
	row->diameter = (float) diameter;	/* Is this precise enough? csabet 28.Jun.01 */  /* diameter */
  row->mntsta = mount;  /* mount type = siderostat */
  
  /* loop over telescopes */
  if (isnew) {
    for (i=0; i<ntel; i++) {
      /* UT are different */
      if (i==0) {
	/* if (!strncmp(t1name, "UT", 2)) row->diameter = 8.0; */
	if (!strncmp(t1name, "UT", 2)) row->diameter = AT_OVER_SID_DIAMETER_LARGE; /* csabet 28.Jun.01 */
	
	strncpy (row->sta_name, t1name, 8); /* sta_name */
	strncpy (row->tel_name, name1,  8); /* tel_name */
      }
      if (i==1) {
	/* if (!strncmp(t2name, "UT", 2)) row->diameter = 8.0; */
	if (!strncmp(t2name, "UT", 2)) row->diameter = AT_OVER_SID_DIAMETER_LARGE; /* csabet 28.Jun.01 */
	strncpy (row->sta_name, t2name, 8); /* sta_name */
	strncpy (row->tel_name, name2,  8); /* tel_name */
      }
      row->index = i+1;  /* telescope index */

      /* use rotated positions from input header if given */
      /* flip sign */
      if (i==0) {
	xxx = -t1xyz[0]; yyy = -t1xyz[1]; zzz = -t1xyz[2]; 
      }
      if (i==1) {
	xxx = -t2xyz[0]; yyy = -t2xyz[1]; zzz = -t2xyz[2]; 
      }

      row->staxyz[0] = xxx;
      row->staxyz[1] = yyy;
      row->staxyz[2] = zzz;
      
      /* write row */
      write_array_geometry_table(outfptr, geomtabledef, row, -1, ierr);
      if (*ierr>0) { /* error test */
	fprintf(stderr,"ERROR %d writing ARRAY_GEOMETRY_TABLE\n", 
		*ierr);
	return;
      }
    } /* end loop over rows */
  } /* end of writing new table */

  /* clean up */
  kill_array_geometry_table_def(geomtabledef); 
  kill_array_geometry_table_row(row);
  
} /* end vndrs_vlti_geometry */

void vndrs_vlti_optical_train(fitsfile *infptr, fitsfile *outfptr,  
			      int tel1, int tel2, int *ierr)
/*----------------------------------------------------------------------- */
/*   Save Optical train for current VLTI configuration if it doesn't exist*/
/*   Copies data from input file.                                         */
/*   If an OPTICAL_TRAN table exists nothing is done.                     */
/*   Inputs:                                                              */
/*      infptr    cfitsio pointer to open FITS file for header info       */
/*      outfptr   cfitsio pointer to open FITS file for table             */
/*      tel1,2    telescope numbers                                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/* HIERARCH ESO ISS CONF DL1    = 'DL1     '   / Delay line used          */
/* HIERARCH ESO ISS CONF DL2    = 'DL2     '   / Delay line used          */
/* HIERARCH ESO ISS CONF INPUT1 =            7 / Input channel in lab     */
/* HIERARCH ESO ISS CONF INPUT2 =            6 / Input channel in lab     */
/* HIERARCH ESO ISS CONF SWYARD1= 'direct  '   / Switchyard configuration */
/* HIERARCH ESO ISS CONF SWYARD2= 'direct  '   / Switchyard configuration */
/*----------------------------------------------------------------------- */
{
  char date_obs[30], sw1[30], sw2[30], DL1[20], DL2[20];
  char comment[60];
  long i, dl1, dl2, input1, input2;
  int  isnew;
  optical_train_table_row *row;
  optical_train_table_def *tabledef;
  
  if (*ierr>0) return; /* existing error condition? */
  if (!infptr) return; /* sanity check */
  if (!outfptr) return; /* sanity check */

  /* get info from input file header */
  strcpy (date_obs,"unknown ");
  default_fits_read_key_str(infptr, "DATE-OBS", date_obs, comment,  
			    (void*)&date_obs, ierr);
  if (*ierr>0) return;
  strcpy (DL1,"DL1 ");
  default_fits_read_key(infptr, TSTRING, "HIERARCH ESO ISS CONF DL1", 
		DL1, comment,  (void*)&DL1, ierr); 
  if (*ierr>0) return;
  sscanf (DL1,"DL%ld",&dl1);
  strcpy (DL2,"DL2 ");
  default_fits_read_key(infptr, TSTRING, "HIERARCH ESO ISS CONF DL2", 
		DL2, comment,  (void*)&DL2, ierr); 
  if (*ierr>0) return;
  sscanf (DL2,"DL%ld",&dl2);

  input1 = 7;
  default_fits_read_key(infptr, TLONG, "HIERARCH ESO ISS CONF INPUT1", 
		&input1, comment,  (void*)&input1, ierr); 
  if (*ierr>0) return;
  input2 = 6;
  default_fits_read_key(infptr, TLONG, "HIERARCH ESO ISS CONF INPUT2", 
		&input2, comment,  (void*)&input2, ierr); 
  if (*ierr>0) return;

  strcpy (sw1,"direct ");
  default_fits_read_key(infptr, TSTRING, "HIERARCH ESO ISS CONF SWYARD1", 
		sw1, comment,  (void*)&sw1, ierr); 
  if (*ierr>0) return;
  strcpy (sw2,"direct ");
  default_fits_read_key(infptr, TSTRING, "HIERARCH ESO ISS CONF SWYARD2", 
		sw2, comment,  (void*)&sw2, ierr); 
  if (*ierr>0) return;

  /* see if table exists, if not create */
  fits_movnam_hdu (outfptr, BINARY_TBL, "OPTICAL_TRAIN", 1, ierr);
  if (*ierr) { /* doesn't exist, create */

    *ierr = 0;
    fits_clear_errmsg(); /* grumble */
    isnew = 1;
    
    /* create array geometry table */
    tabledef = 
      make_optical_train_table_def(outfptr, "VLTI", date_obs, ierr);
    if ((tabledef==NULL) || (*ierr>0)) { /* error test */
      fprintf (stderr,"error %d after optical train table create\n",
	       *ierr);
      return;
    } /* end error test */ 
  } else { /* open existing one */
    tabledef = open_optical_train_table(outfptr, ierr);
    isnew = 0;
    if ((tabledef==NULL) || (*ierr>0)) { /* error test */
      fprintf (stderr,"error %d after optical train  table open \n",
	       *ierr);
      return;
    } /* end error test */ 
  }
  
  /* make row structure */
  row = make_optical_train_table_row (tabledef);
  
  /* loop over trains */
  if (isnew) {
    for (i=0; i<2; i++) {
      row->index = i+1;
      if (i==0) {
	row->telescope = tel1;
	row->dly_line = dl1;
	row->lab_beam = input1;
	strncpy (row->switchyard, sw1, 16);
      } else {
	row->telescope = tel2;
	row->dly_line = dl2;
	row->lab_beam = input2;
	strncpy (row->switchyard, sw2, 16);
      }
      /* write row */
      write_optical_train_table(outfptr, tabledef, row, -1, ierr);
      if (*ierr>0) { /* error test */
	fprintf(stderr,"ERROR %d writing OPTICAL_TRAIN_TABLE\n", 
		*ierr);
	return;
      }
    } /* end loop over rows */
  } /* end of writing new table */

  /* clean up */
  kill_optical_train_table_def(tabledef); 
  kill_optical_train_table_row(row);
  
} /* end vndrs_vlti_opt_tran */

int vndrs_vlti_source(fitsfile *infptr, fitsfile *outfptr, int *ierr) 
/*----------------------------------------------------------------------- */
/*   Save source info in Source table.                                    */
/*   Uses global array to keep track of which sources are already         */
/*   encountered.                                                         */
/*   Copies data from input file.                                         */
/*   Values needed from primary header:                                   */
/*  RA    double    / 00:00:00.0 RA (J2000) pointing (deg)                */
/*  DEC   double    / 00:00:00.0 DEC (J2000) pointing (deg)               */
/*  EPOCH double    / Standard FK5 (years)                                */
/*  HIERARCH ESO OBS TARG NAME  string = source name                      */
/*  HIERARCH ESO INS FILT1 NAME  string = filter name                     */
/*   Inputs:                                                              */
/*      infptr      cfitsio pointer to open FITS file for header info     */
/*      outfptr     cfitsio pointer to open FITS file for source table    */
/*                  if null on entry open file                            */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Return:                                                              */
/*      Source number for the source                                      */
/*----------------------------------------------------------------------- */
{
  source_table_def *srctabledef;
  source_table_row *row;
  /* char temp[100], source[18], target[60], comment[80], filter[20], mode[40]; */
	char	target[60], comment[80], filter[20], mode[40]; /* csabet 28.Jun.01 */
  int i, isnew, isour, last, source_id=0;
  double ra, dec, freq, equinox;
  long ir, nrow, n_valid;
  double default_dbl = 0.0;
  char *default_str="Calibration";

  /* if table doesn't exist create and add current source
     make large enough for MAX_SOURCE entries */
  /* see if table exists, if not create */
  fits_movnam_hdu (outfptr, BINARY_TBL, "SOURCE", 1, ierr);
  if (*ierr) { /* doesn't exist, create */

    *ierr = 0;
    fits_clear_errmsg(); /* grumble */
    isnew = 1;
    
    /* create source table */
    srctabledef = make_source_table_def(outfptr, 1, ierr);
    if ((srctabledef==NULL) || (*ierr>0)) { /* error test */
      fprintf (stderr,"error %d after source table create\n",
	       *ierr);
      /* return; */
		return (*ierr);	/* csabet 28.Jun.01 */
    } /* end error test */ 
  } else { /* open existing one */
    srctabledef = open_source_table(outfptr, ierr);
    isnew = 0;
    if ((srctabledef==NULL) || (*ierr>0)) { /* error test */
      fprintf (stderr,"error %d after source table open \n",
	       *ierr);
      /* return; */
		return (*ierr);	/* csabet 28.Jun.01 */
    } /* end error test */ 
  }

  /* make row structure */
  row = make_source_table_row (srctabledef);

  /* if new write dummy entries */
  if (isnew) {
    /* init row structures */
    /* source name */
    strcpy (row->source, "not set");
    row->source_id = 0;
    row->raepp = 0.0;
    row->decepp = 0.0;
    strncpy (row->equinox, "J2000", 8);
    row->diameter = 0.0;
    row->diam_err = 0.0;
    for (isour=0; isour<MAX_SOURCE; isour++) {
      /* write row */
      ir = isour+1;
      write_source_table(outfptr, srctabledef, row, ir, ierr);
      if (*ierr>0) { /* error test */
 	fprintf(stderr,"ERROR %d writing SOURCE_TABLE\n", *ierr);
	return -1;
      }
    } /* end loop over table */
  } /* end of fill */
  
  /* get source information from file header */
  default_fits_read_key_dbl(infptr, "RA", &ra, comment,  
			    (void*)&default_dbl, ierr);
  if (*ierr>0) return -1; /* failed - bomb out */
  default_fits_read_key_dbl(infptr, "DEC", &dec, comment,  
			    (void*)&default_dbl, ierr);
  if (*ierr>0) return -1; /* failed - bomb out */
  /* debug */
  /*fprintf(stderr," RA = %lf dec = %lf\n",ra, dec);*/
  default_fits_read_key_dbl(infptr, "EQUINOX", &equinox, comment,  
			    (void*)&default_dbl, ierr);
  if (*ierr>0) return -1; /* failed - bomb out */
  /* Filter name */
  default_fits_read_key(infptr, TSTRING, "HIERARCH ESO INS FILT1 NAME", 
		filter, comment,  (void*)default_str, ierr); 
  if (*ierr>0) return -1; /* failed - bomb out */
  /* set frequency */
  if (!strncmp (filter,"K       ",8)) {
    freq = 2.997924562e8 / 2.128e-6;
  } else {
    freq = 2.997924562e8 / 2.128e-6;
  }
  /* source name - use INS MODE as the default for autotest/col */
  default_fits_read_key_str(infptr, "HIERARCH ESO INS MODE", 
     mode, comment, (void*)default_str, ierr);
  /* if (*ierr>0) return; */
	if (*ierr>0) return (*ierr);	/* csabet 28.Jun.01 */
  for (i=0;i<40;i++) target[i] = 0;
  default_fits_read_key_str(infptr, "HIERARCH ESO OBS TARG NAME", 
     target, comment, (void*)mode, ierr);
  /* if (*ierr>0) return; */
	if (*ierr>0) return (*ierr);	/* csabet 28.Jun.01 */
   /*fprintf(stderr," INS MODE %s\n",mode); debug */
   /*fprintf(stderr," TARG NAME %s\n",target);  debug */

  /* read through table to see if it's there */
  source_id = -1;
  last = -1;
  n_valid = 0;
  /* number of rows */
  fits_read_key_lng(outfptr, "NAXIS2", &nrow, comment, ierr);
  if (nrow<=0) nrow = MAX_SOURCE;
  for (isour=0; isour<nrow; isour++) {
    /* read row */
    ir = isour+1;
    read_source_table(outfptr, srctabledef, row, ir, ierr);
    if (*ierr>0) { /* error test */
      fprintf(stderr,"ERROR %d reading SOURCE_TABLE\n", *ierr);
      return -1;
    }
    /* is this the current source ? */
    if (!strncmp (row->source, target, 16)) {
      source_id = row->source_id;
    }
    if (row->source_id>last) last = row->source_id;
    if (row->source_id>0) n_valid = ir;
  } /* end loop over table */
  
  /* debug */
  /*fprintf(stderr,"after lookup source %d %d\n",last,source_id);*/

  /* if not, write next */
  if (source_id<=0) {
    source_id = last + 1;
    row->source_id = (int)source_id;
    strncpy (row->source, target, 16);
    row->raepp = ra;
    row->decepp = dec;
    row->freq[0] = freq;
    if (fabs (equinox-2000.0)<0.1) strncpy (row->equinox, "J2000", 8);
    if (fabs (equinox-1950.0)<0.1) strncpy (row->equinox, "B1950", 8);
    /* write row */
    ir = source_id;
    if (ir>n_valid) n_valid = ir;
    write_source_table(outfptr, srctabledef, row, ir, ierr);
    if (*ierr>0) { /* error test */
      fprintf(stderr,"ERROR %d writing SOURCE_TABLE\n", *ierr);
      return -1;
    }
  }
  
  /* update n_valid keyword */
  fits_update_key_lng (outfptr, "N_VALID", n_valid,
		       "Number of valid entries", ierr);

  /* clean up */
  kill_source_table_def(srctabledef);
  kill_source_table_row(row); 
  
  return source_id;
} /* end vndrs_vlti_source */

int vndrs_vlti_batch(fitsfile *infptr, fitsfile *outfptr, int *ierr)
/*----------------------------------------------------------------------- */
/*   Save batch info in Batch table.                                      */
/*   Copies data from input file.                                         */
/*   Only creates table if necessary and allocates a batch ID.            */
/*   no other information is included                                     */
/*   Values needed from primary header:                                   */
/*   DATE-OBS= 2001-03-24T00:45:39.2477 / Obs Start (UTC)                 */
/*   HIERARCH ESO DET QL FRAMERATE= 802.89 / frame rate                   */
/*   HIERARCH ESO DET QL FRINGESPD= 0.000351 / fringe speed               */
/*   HIERARCH ESO DET QL MAXWAVEK = 2.5E-06 / max wave length             */
/*   HIERARCH ESO DET QL MINWAVEK = 1.95E-06 / min wave length            */
/*   HIERARCH ESO INS FILT1 NAME  = 'K       '   / Filter common name.    */
/*   Inputs:                                                              */
/*      infptr      cfitsio pointer to open FITS file for header info     */
/*      outfptr     cfitsio pointer to open FITS file for batch table     */
/*                  if null on entry open file                            */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Return:                                                              */
/*      Batch number                                                      */
/*----------------------------------------------------------------------- */
{
  batch_table_def *tabledef;
  batch_row *row;
  int calibr=0; /* not calibrated */
  float framrate, fringspd, maxwave, minwave;
  char date_obs[30], filter[60];
  /* char temp[100], batch[18], target[60], comment[80]; */
	char comment[80];	/* csabet 28.Jun.01 */
  int isnew, batch_id=0;
  long ir, ibat, nrow;
  char *default_str="        ";
  float default_flt = 0.0;
/*	double default_dbl = 0.0;
  long  default_lng = 0;	*/  /* csabet 28.Jun.01. Unused variables */

  /* get info from FITS header */
  default_fits_read_key_str(infptr, "DATE-OBS", date_obs, comment,  
			    (void*)default_str, ierr);
  if (*ierr>0) return batch_id; /* failed - bomb out */
  default_fits_read_key_str(infptr, "HIERARCH ESO INS FILT1 NAME", filter, 
		    comment,  (void*)default_str, ierr);
  if (*ierr>0) return batch_id; /* failed - bomb out */
  /* frame rate in Hz */
  default_fits_read_key(infptr, TFLOAT, "HIERARCH ESO DET QL FRAMERATE", 
		&framrate, comment,  (void*)&default_flt, ierr); 
  if (*ierr>0) return batch_id;
  /* fringe speed (m/s)  */
  default_fits_read_key(infptr, TFLOAT, "HIERARCH ESO DET QL FRINGESPD", 
		&fringspd, comment,  (void*)&default_flt, ierr); 
  if (*ierr>0) return batch_id;
  /* max wave length */
  default_fits_read_key(infptr, TFLOAT, "HIERARCH ESO DET QL MAXWAVEK", 
		&maxwave, comment,  (void*)&default_flt, ierr); 
  if (*ierr>0) return batch_id;
  /*  min wave length  */
  default_fits_read_key(infptr, TFLOAT, "HIERARCH ESO DET QL MINWAVEK", 
		&minwave, comment,  (void*)&default_flt, ierr); 
  if (*ierr>0) return batch_id;

  /* if table doesn't exist create and add current batch
     make large enough for MAX_BATCH entries */
  /* see if table exists, if not create */
  fits_movnam_hdu (outfptr, BINARY_TBL, "BATCH", 1, ierr);
  if (*ierr) { /* doesn't exist, create */

    *ierr = 0;
    fits_clear_errmsg(); /* grumble */
    isnew = 1;
    
    /* create batch table */
    tabledef = 
      make_batch_table_def(outfptr, date_obs, calibr,
			   framrate, fringspd, maxwave, minwave, 
			   filter, 1, ierr);
    if ((tabledef==NULL) || (*ierr>0)) 
	{ /* error test */
      fprintf (stderr,"error %d after batch table create\n",
	       *ierr);
      /* return; */
		return (*ierr);	/* csabet 28.Jun.01 */
    } /* end error test */ 
  } else { /* open existing one */
    tabledef = open_batch_table(outfptr, ierr);
    isnew = 0;
    if ((tabledef==NULL) || (*ierr>0)) { /* error test */
      fprintf (stderr,"error %d after batch table open \n",
	       *ierr);
      /* return; */
		return (*ierr);	/* csabet 28.Jun.01 */
    } /* end error test */ 
  }

  /* make row structure */
  row = make_batch_row (tabledef);

  /* if new write dummy entries */
  if (isnew) {
        for (ibat=0; ibat<MAX_BATCH; ibat++) {
      /* init row structures */
      row->batch = -1;
      /* write row */
      ir = ibat+1;
      write_batch_table(outfptr, tabledef, row, ir, ierr);
      if (*ierr>0) { /* error test */
 	fprintf(stderr,"ERROR %d writing BATCH_TABLE\n", *ierr);
	return -1;
      }
    } /* end loop over table */
  } /* end of fill */
  
  /* read through table to find last */
  batch_id = 1;
  /* number of rows */
  fits_read_key_lng(outfptr, "NAXIS2", &nrow, comment, ierr);
  if (nrow<=0) nrow = MAX_BATCH;
  for (ibat=0; ibat<nrow; ibat++) {
    /* read row */
    ir = ibat+1;
    read_batch_table(outfptr, tabledef, row, ir, ierr);
    if (*ierr>0) { /* error test */
      fprintf(stderr,"ERROR %d reading BATCH_TABLE\n", *ierr);
      return -1;
    }
    /* is this one used? */
    if (row->batch<=0) { /* no, use it */
      batch_id = ir;
      break;
    }
  } /* end loop over table */

  /* write row */
  ir = batch_id;
  row->batch = (int)batch_id;
  write_batch_table(outfptr, tabledef, row, ir, ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d writing BATCH_TABLE\n", *ierr);
    return -1;
  }
  
  /* clean up */
  kill_batch_table_def(tabledef);
  kill_batch_row(row); 
  
  return batch_id;
} /* end vndrs_vlti_batch  */

void vndrs_update_batch_table (fitsfile *outfptr, opd_data *opd,
			       int tel1, int tel2, int source, int batch_id,
			       char *file_on, char *file_off, 
			       char *file_a, char *file_b, 
			       matrix_double *Matrix, double shape, 
			       int *ierr)
/*----------------------------------------------------------------------- */
/*   Update FITS batch table                                              */
/*   Inputs:                                                              */
/*      outfptr   cfitsio pointer to open output FITS file                */
/*      batch_on  processed on source batch,  scan_data member is used    */
/*      tel1,2    telescope numbers                                       */
/*      source    source number                                           */
/*      batch_id  batch number in the batch table                         */
/*      file_on   on source raw FITS file name                            */
/*      file_off  off source raw FITS file name                           */
/*      file_a    Port A only raw FITS file name                          */
/*      file_b    Port B only raw FITS file name                          */
/*      Matrix    Kappa matrix.                                           */
/*      shape     shape factor (e.g. 0.000013)                            */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*----------------------------------------------------------------------- */
{
	batch_table_def	*tabledef;
	batch_row *row;
  long ir;

  shape  = shape * 1.0e6; /* convert to units used in analysis */

  if (*ierr>0) return; /* existing error condition */

  /* position in batch table */
  fits_movnam_hdu (outfptr, BINARY_TBL, "BATCH", 1, ierr);
  if (*ierr>0) return;

  /* access to table */
  tabledef = open_batch_table(outfptr, ierr);
  if ((tabledef==NULL) || (*ierr>0)) { /* error test */
    fprintf (stderr,"error %d after batch table open \n",
	     *ierr);
    return;
  } /* end error test */ 

  /* make row structure */
  row = make_batch_row (tabledef);

  /* read previous entry */
  ir = batch_id;
  read_batch_table(outfptr, tabledef, row, ir, ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d reading BATCH_TABLE\n", *ierr);
    return;
  }

  /* set batch information */
  strncpy (row->file_on, file_on, 64);
  strncpy (row->file_off, file_off, 64);
  strncpy (row->file_a, file_a, 64);
  strncpy (row->file_b, file_b, 64);
  row->source = source;
  row->ot[0] = tel1;
  row->ot[1] = tel2;
  row->config = 1;
  /* row->kappa[0] = get_element_matrix_double(Matrix, 0,0); */
	row->kappa[0] = (float) (get_element_matrix_double(Matrix, 0,0));	/* csabet 28.Jun.01 */
  /* row->kappa[1] = get_element_matrix_double(Matrix, 0,1); */
	row->kappa[1] = (float) (get_element_matrix_double(Matrix, 0,1));	/* csabet 28.Jun.01 */
  /* row->kappa[2] = get_element_matrix_double(Matrix, 1,0); */
	row->kappa[2] = (float) (get_element_matrix_double(Matrix, 1,0));	/* csabet 28.Jun.01 */
  /* row->kappa[3] = get_element_matrix_double(Matrix, 1,1); */
	row->kappa[3] = (float) (get_element_matrix_double(Matrix, 1,1));	/* csabet 28.Jun.01 */
  row->nscan  = opd->nscans;
  row->time   = opd->avg_time;
  /* row->timei  = opd->time_interv * 86400.0; */
	row->timei  = (float) (opd->time_interv * 86400.0);	/* csabet 28.Jun.01 */
  /* row->sdopd  = opd->sigma_OPD; */
	row->sdopd  = (float) (opd->sigma_OPD);	/* csabet 28.Jun.01 */
  /* row->sdopd1 = opd->sigma_OPDdot; */
	row->sdopd1 = (float) (opd->sigma_OPDdot);	/* csabet 28.Jun.01 */
  /* row->photom[0]  = opd->mean_ph[0]; */
	row->photom[0]  = (float) (opd->mean_ph[0]);	/* csabet 28.Jun.01 */
  /* row->photom[1]  = opd->mean_ph[1]; */
	row->photom[1]  = (float) (opd->mean_ph[1]);	/* csabet 28.Jun.01 */
  /* row->corpower[0]  = opd->mean_cp[0]; */
	row->corpower[0]  = (float) (opd->mean_cp[0]);	/* csabet 28.Jun.01 */
  /* row->corpower[1]  = opd->mean_cp[1]; */
	row->corpower[1]  = (float) (opd->mean_cp[1]);	/* csabet 28.Jun.01 */
  /* row->corpower[2]  = opd->mean_cp[2]; */
	row->corpower[2]  = (float) (opd->mean_cp[2]);	/* csabet 28.Jun.01 */
  /* row->photomsd[0]  = opd->sigma_ph[0]; */
	row->photomsd[0]  = (float) (opd->sigma_ph[0]);	/* csabet 28.Jun.01 */
  /* row->photomsd[1]  = opd->sigma_ph[1]; */
	row->photomsd[1]  = (float) (opd->sigma_ph[1]);	/* csabet 28.Jun.01 */
  /* row->corpowsd[0]  = opd->sigma_cp[0]; */
	row->corpowsd[0]  = (float) (opd->sigma_cp[0]);	/* csabet 28.Jun.01 */
  /* row->corpowsd[1]  = opd->sigma_cp[1]; */
	row->corpowsd[1]  = (float) (opd->sigma_cp[1]);	/* csabet 28.Jun.01 */
  /* row->corpowsd[2]  = opd->sigma_cp[2]; */
	row->corpowsd[2]  = (float) (opd->sigma_cp[2]);	/* csabet 28.Jun.01 */
  row->mu12   = opd->mu12;
  row->mu22   = opd->mu22;
  row->emu12  = opd->emu12;
  row->emu22  = opd->emu22;
  /* row->shape  = shape; */
	row->shape  = (float) shape;	/* csabet 28.Jun.01 */

  /* rewrite it */
  write_batch_table(outfptr, tabledef, row, ir, ierr);
  if (*ierr>0) { /* error test */
    fprintf(stderr,"ERROR %d writing BATCH_TABLE\n", *ierr);
    return;
  }

  /* update n_valid keyword */
    fits_update_key_lng (outfptr, "N_VALID", ir,
			"Number of valid entries", ierr);
			
			
	/* Record Visibility values (csabet 22.Oct.2001)
	----------------------------------------------*/
    fits_write_key_fixflt(
		outfptr, 
		"HIERARCH ESO PRO MU12", 
		opd->mu12,
		6,
		"Average first visibility",
		ierr);
	if (*ierr>0)
	{
		fprintf(stderr,"ERROR %d writing HIERARCH ESO PRO MU12\n", *ierr);
		return;
	}

	/* Record Visibility values (csabet 22.Oct.2001)
	----------------------------------------------*/
    fits_write_key_fixflt(
		outfptr, 
		"HIERARCH ESO PRO EMU12", 
		opd->emu12,
		6,
		"Variance of Average first visibility",
		ierr);
	if (*ierr>0)
	{
		fprintf(stderr,"ERROR %d writing HIERARCH ESO PRO EMU12\n", *ierr);
		return;
	}
	
	/* Record Visibility values (csabet 22.Oct.2001)
	----------------------------------------------*/
    fits_write_key_fixflt(
		outfptr, 
		"HIERARCH ESO PRO MU22", 
		opd->mu22,
		6,
		"Average second visibility",
		ierr);
	if (*ierr>0)
	{
		fprintf(stderr,"ERROR %d writing HIERARCH ESO PRO MU22\n", *ierr);
		return;
	}
	
	/* Record Visibility values (csabet 22.Oct.2001)
	----------------------------------------------*/
    fits_write_key_fixflt(
		outfptr,
		"HIERARCH ESO PRO EMU22", 
		opd->emu22,
		6,
		"Variance of Average second visibility",
		ierr);
	if (*ierr>0)
	{
		fprintf(stderr,"ERROR %d writing HIERARCH ESO PRO EMU22\n", *ierr);
		return;
	}
	

} /* end vndrs_update_batch_table */

float vndrs_FITS_mean_SNR(individual_scan *scan, double RON)
/*----------------------------------------------------------------------- */
/*   Compute the mean SNR for a scan                                      */
/*   Inputs:                                                              */
/*      scan      scan to be analysed.                                    */
/*      RON       the readout noise of the current pixel                  */
/*   Returns:                                                             */
/*      the mean SNR                                                      */
/*----------------------------------------------------------------------- */
{
  /* RON is the readout noise of the pixel of LISA corresponding to the current beam */
  long i;
  double sum;
  float meansnr;

  /* modified pkervell 13-mar-2002, as the previous definition (min_SNR) was
     irrelevant for the selection of processed data: the photometric fluctuations are not the
     problem here, but rather the ratio average(scan)/readout_noise(pixel) */

  sum = 0.0;

  for (i=0; i<scan->dimension; i++)
      sum  += scan->data[i];

  if(abs(RON*scan->dimension) >= 1e-12)
    meansnr = sum/(RON * scan->dimension);
  else
    meansnr = 0.0;
  return meansnr;

} /* end vndrs_FITS_SNR */
