/******************************************************************************
*				European Southern Observatory
*		 Vlti ViNci Data Reduction Software (VNDRS)
*
* File:			vndrs_MainRoutines.c
* Description:	Contains the core routines for fringe processing.
*				Described in [1], section 2.1
*
* History:
* ??-???-00		(Markus Schoeller & Rainer Wilhelm, ESO, 2000) Initial design		
* 01-May-01		(csabet) Revised
* 08-Mar-02             (pkervell) Modified matrix computation at each scan
* 10-Mar-02             (pkervell) Debias of mu^2s by removal of background
* xx-Jul-02             (segransan) Modified to deal with pistonned scans
* 29-Jul-02             (pkervell) Modified to accept data from IONIC (one interf output only)
* 04-Aug-02             (pkervell) Simplification of vndrs_Mainroutines.c + memory alloc check
* 09-Aug-02		(pkervell) Changed background removal to average over batch
* 29-Aug-02		(pkervell) Computation of the photometric signal sqrt(PA.PB) level
*			on the fringes, and removal of the "sanity checks" on the visibilities.
* 03-Sep-02		(pkervell) Simultaneous fringe detection and reduction (same loop),
* 			code cleanup and background removal modifications
*			Wiener filter computation was completely wrong !
* 08-Sep-02		(pkervell) Photometry is corrected better, innovative estimation of mu2
* 13-Sep-02		(pkervell) Clean version with piston bias correction
* 29-Sep-02		(pkervell) First science grade version
* 07-Nov-02		(pkervell) Corrected wavelength (2.195 microns)
* 03-Jan-04		(pkervell) Modified computation of histogram after referee comment
* 16-Dec-06     (pkervell) Reinstallation of depistonization of visibilities
*
******************************************************************************/
#define Pi       3.14159265358979311599796346854418516

/******************************************************************************
*	Compiler directives
******************************************************************************/

/******************************************************************************
*	Include files
******************************************************************************/
#include "vndrs.h"
#include "vndrs_Globalvar.h"
#include<stdio.h>
#include<stdlib.h>
#include<math.h>

void vndrs_compute_visibilities (
                                 /* --------------- */
                                 /* input variables */
                                 /* --------------- */ 
                                 batch *Batch_on,
                                 /* series of scans obtained on source */
                                 batch *Batch_off,
                                 /* series of scans obtained Off_source */
                                 batch *Batch_A,
                                 /* series of scans obtained with beam A only */
                                 batch *Batch_B,
                                 /* series of scans obtained with beam B only */
                                 int Choice_matrix,
                                 /* Flag that is used to decide whether the transfer matrix
                                    will be computed from the mixed beams (0), the average
				    of on-source batch (1), or for each scan (2), or fixed (3) */
                                 double Frame_rate,
                                 /* Frame aquistition frequency in units of Hz */
                                 double Fringe_speed,
                                 /* Speed of the fringes in units of m/sec*/
                                 double Shape_factor,
                                 /* shape factor for the fringe power integration */
				 long N_frames_cut,
				 /* number of frames to be trimmed for processing */
				 long fringes_length,
				 /* number of points to keep for the fringes in the wavelets transform */
                                 double Wavenum_theo,
                                 /* Theoretical central wavenumber of the filter */
                                 vector_double *Wavenum_limits_fft, 
                                 /* 2 x 1 vector: lower Wavenum_limit, higher Wavenum_limit */
                                 vector_double *Wavenum_limits_wl, 
                                 /* 2 x 1 vector: lower Wavenum_limit, higher Wavenum_limit */
                                 vector_double *Wavenum_limits_QL, 
                                 /* 2 x 1 vector: lower Wavenum_limit, higher Wavenum_limit FOR QUICKLOOK */
				 bool process_undetected,
				 /* added pkervell 12-mar-2001 true if user wants to process scans where no
				    fringes were found by the QL during acquisition */
				 int use_mean_wiener,    /* added pkervell 15-mar-02 */
				 /* flag to tell the processing sw to use or not the Wiener filters of
				  averaged PSDs of PA and PB */
				 int verbose_on,    /* added pkervell 15-mar-02 */
				 /* flag to output the intermediate processed data to ASCII files */
				 int debias,
				 /* flag to activate or not the removal of the off-source scans power */
				 int depiston,
				 /* flag to activate or not the correction factor of the piston on mu2 */
				 int median_output,
				 /* flag to choose the batch values to be the median values */
				 int background_per_scan,
				 /* flag to activate or not the computation of the wavelet derived visibilities */
				 float center_wl_threshold,
				 /* activates the centering of the fringes through WL if set > 0, defines the threshold */
				 int select_scans,   /* added - pkervell - 29-Jul-02 */
				 /* threshold on the piston level acceptable 0.25 to 0.3 is a good starting value */
				 float max_jump,
				 /* maximum acceptable fringes OPD jump between scans (in microns) */
                                 long Max_average,
                                 /* Maximum number of Off_source scans to be reduced */
                                 bool two_channels_only,  /* introduced to distinguish between full
                                                             and pseudo correction */
                                 float Rejection_threshold,
                                 double Apod_blockout,
                                 double Apod_transition,
                                 double Mult_corr,
				 long B,   /* bootstrap number  added pkervell dsegransan 20-mar-02 */
				 double lat,  /* coordinates of Paranal, in radians - added pkervell 15-apr-02 */
				 double longi,
				 double RA_object, /* coordinates of the object */
				 double Dec_object,
				 double t1x, double t1y, double t1z, /* coordinates of the telescopes (in meters) */
				 double t2x, double t2y, double t2z,
				 int ionic_in,
                                 /* --------------------------*/
                                 /* input or output variables */
                                 /* --------------------------*/
                                 matrix_double *Average_Matrix,
                                 /* The 2 x 2 transfer matrix between the photometric and interferometric signals
                                    (both measured at photometers (see [2], Eq. (19) on p. 383):

                                            [ I1 ]   [ kappa_1A  kappa_1B ]   [ PA ]
                                            |    | = |                    | * |    |
                                            [ I2 ]   [ kappa_2A  kappa_2B ]   [ PB ]
				 */
                                 /* ---------------- */
                                 /* output variables */
                                 /* ---------------- */
                                 vector_double *Relative_MSE,       /* 2x double */
                                 batch *Batch_on_processed,
                                 /* Processed series of On_source scans */
                                 batch *Batch_off_processed,
                                 long *N_scans_processed,
                                 /* Number of processed (=reduced) scans */
                                 long *numberOfScansRejected,	/* Number of scans not processed */
                                 vector_double *Mu1_squared,
                                 /* Average squared coherence factor */
                                 vector_double *Mu2_squared,
                                 /* Average residual subtracted background */
				 double *time_batch,
				 double *baseline_batch,
				 double *azimuth_batch,
				 double *u_batch,
				 double *v_batch,
                                 double *altitude_src, /* added pkervell 15-may-03 */
                                 double *azimuth_src,
				 double *Mu1_batch_wl, double *ErrMu1_batch_min_wl, double *ErrMu1_batch_max_wl,
                                 double *ErrMu1_batch_wl,
				 double *Mu2_batch_wl, double *ErrMu2_batch_min_wl, double *ErrMu2_batch_max_wl,
				 double *Mu1_batch_fft, double *ErrMu1_batch_min_fft, double *ErrMu1_batch_max_fft,
                                 double *ErrMu1_batch_fft,
				 double *Mu2_batch_fft, double *ErrMu2_batch_min_fft, double *ErrMu2_batch_max_fft,
                                 double *background_fourier, double *background_wavelets,
                                 double *background_fourier_med, double *background_wavelets_med,
                                 double *effective_wavelength, double *effective_wavelength_wl, double *effective_wavelength_error,
                                 double *Mu1_skew_fft, double *Mu1_skew_wl,
                                 double *Photom_power, double *total_weight_batch,
                                 int *warnings_flag, int *reject_flag)
                                  
{
    /* LOCAL VARIABLES */ 
    /* --------------- */

    long I, J, K, N_scans_on,  N_scans_off, N_frames_on, N_frames_off, N_mixed, ii;
    double Wavenum_interval = 0.0;	/* default value csabet 26.Jun.01 */
    double Wavenum_interval_precut = 0.0;	/* added - pkervell 20-apr-02 */
    double Sample_interval = 0.0;	/* added - pkervell 04-Sep-02 */
    double Wavenum_bandwidth_fft, Wavenum_min_hi_band_fft, Wavenum_max_hi_band_fft;
    double Wavenum_min_lowband_fft, Wavenum_max_lowband_fft;
    double Wavenum_bandwidth_wl, Wavenum_min_hi_band_wl, Wavenum_max_hi_band_wl;
    double Wavenum_min_lowband_wl, Wavenum_max_lowband_wl;
    double Wavenum_min_out, Wavenum_max_out;     /* added pkervell 15-mar-2002 */
    double Wavenum_min, Wavenum_max;     /* added pkervell 25-apr-2002 */
    double surface_in, surface_out_wl, surface_out_fft;
    long index_freq_low_fft, index_freq_high_fft, index_freq_low_fft_cut,
        index_freq_high_fft_cut, index_freq_low_wl, index_freq_high_wl,
        index_freq_low_wl_cut, index_freq_high_wl_cut,
        index_freq_low_QL_cut, index_freq_high_QL_cut;
    double kappa1A, kappa1B, kappa2A, kappa2B;
    double kappa1A_err, kappa1B_err, kappa2A_err, kappa2B_err;
    matrix_double *norm_t_wavelets_I1, *norm_t_wavelets_I1_check, *norm_t_wavelets_I1_off;
    matrix_double *Standard_dev_Matrix;
    /* Squared coherence factors variables, as computed with the wavelets transform */
    vector_double *Mu1_squared_wl;
    /* Power spectrum background, as computed with the wavelets transform */
    vector_double *Mu2_squared_wl;
  
    /* BEWARE: Standard_dev_Matrix is in fact the matrix of the standard ERRORS (stddev/sqrt(N)) */
    power_spectrum *WA_mean, *WB_mean;
    /* WA_mean and WB_mean are the Wiener filters computed from the average PSDs of PA and PB */
    vector_double *Off_offsets_Mean, *Off_offsets_Stddev, *Off_offsets_Min, *Off_offsets_Max;
    /* 4 x 1 vectors holding the statistics of the Off_source batch. For example,
     'Off_offsets_Mean' holds the mean values of the four signals I1, I2,
     PA and PB averaged over all frames and over all full scans. */
    vector_double *On_source_Mean, *On_source_Stddev, *On_source_Min, *On_source_Max;
    vector_double *A_Mean, *A_Stddev, *A_Min, *A_Max;
    vector_double *B_Mean, *B_Stddev, *B_Min, *B_Max;
    /* 4 x 1 vectors holding the statistics of the On_source batch.  */
    vector_double *Flag_processed;
    batch *Batch_A_cut, *Batch_B_cut;
    batch *Batch_on_cut;
    /* Series of On_source scans cut to center the fringes added pkervell 13-mar-2002 */
    batch *Batch_off_cut;
    /* Series of Off_source scans cut to a smaller size added pkervell 13-mar-2002 */
    char command[MAX_STRING_LENGTH];
    /* Command passed to the operating system */
    batch *Batch_on_precut, *Batch_off_precut;   /* Series of scans cut to min(512 points,N_frames_on) to find the fringes */
    matrix_complex *t_wavelets_I1, *t_wavelets_I1_off;   /* Wavelets transform of the PROCESSED interferogram I1,
				      allocated OUTSIDE the vndrs_wavelets function - segransan - xx-Jul-02 */
    long N_frames_precut;
    long wavelets_fringes_center = 0;
    double jump = 0.0;
    vector_double *fringes_found_wavelets;
    vector_double *peak_QL;  /* added pkervell - 09-Aug-02 */
    /* Added variables for fringe detection and piston estimation - pkervell 04-Sep-02 */
    vector_double *coi_I1,   /* coi = cone of influence */
                *in_to_outband_ratio,
                *residus_QL,
                *fringes_centering_outputs,
                *fwhm_peak_time,
                *peak_position_freq,
                *fwhm_peak_freq,
                *background_fit_QL,
                *weights_visibilities,
                *fringes_off_center;
    double buffer_fringes_off_center;
    vector_double *signal_p; /* sqrt(PA.PB) on the fringes
                                 pkervell 04-Sep-02 */
    double total_weighting_PSD = 0.0; /* added pkervell 09-Sep-02 */
    double signal_buffer, buffer_flag;
    /* Average power spectra - added - pkervell 09-Aug-02 */
    power_spectrum *Pow_spec_avg_I1,
                 *Pow_spec_avg_I1_no_centering,
                 *Pow_spec_avg_I1_no_centering_corrected,
                 *Pow_spec_avg_buffer_I1,
                 *Pow_spec_avg_I1_off,
                 *Pow_spec_avg_buffer_I1_off,
                 *Pow_spec_avg_I1_corrected;
    vector_double *Background_PSD_wl_vector,
                  *WL_spec_avg_I1_vector,
                  *WL_spec_avg_I1_vector_no_centering,
                  *WL_spec_avg_I1_vector_no_centering_corrected,
                  *WL_spec_avg_I1_vector_off,
                  *WL_spec_avg_I1_vector_corrected,
                  *Phase_avg_I1_no_centering,
                  *Phase_I1;
    /* Removed indices - added pkervell 27-Sep-02 */
    vector_double *removed_indices;
    double flag_remove = 0.0;
    
    /* All the variables for the BOOTSTRAPPING */
    vector_double *non_zero_weights, *non_zero_mu1_wl, *non_zero_mu2_wl, *non_zero_mu1_fft,
                  *non_zero_signal_p, *non_zero_peak_pos, *non_zero_peak_width;
    /* background removal - pkervell 09-Sep-02 */
    vector_double *Background_PSD;
    double Normalization_Fourier, Normalization_WL;
    /* peak position */
    double weight_psd, weight_psd_wl, running_max;
    /* Photometric control variables */
    full_scan *Avg_power_spectra_on, *Avg_power_spectra_off;
    vector_double *Freq_scale_avg_spectra;
    vector_double *Freq_scale_fringes;
    /* double PA_mean, PB_mean, I1_mean, I2_mean; /* average values */

    double mu1_fft_avg_bs = 0.0, mu1_fft_avg_mil = 0.0,  mu1_fft_avg_noncentered = 0.0, mu1_fft_avg = 0.0,
            mu1_fft_boot_avg = 0.0, mu1_fft_stdev_avg = 0.0, Bias_mu1_fft_avg = 0.0,
            Rms_mu1_fft_avg = 0.0, Ci_mu1_fft_avg_min= 0.0, Ci_mu1_fft_avg_max= 0.0,
            Ci2_mu1_fft_avg_min= 0.0, Ci2_mu1_fft_avg_max= 0.0, Ci3_mu1_fft_avg_min= 0.0, Ci3_mu1_fft_avg_max= 0.0;
    double mu1_fft_med = 0.0, mu1_fft_boot_med = 0.0, mu1_fft_stdev_med = 0.0, Bias_mu1_fft_med = 0.0,
            Rms_mu1_fft_med = 0.0, Ci_mu1_fft_med_min= 0.0, Ci_mu1_fft_med_max= 0.0,
            Ci2_mu1_fft_med_min= 0.0, Ci2_mu1_fft_med_max= 0.0, Ci3_mu1_fft_med_min= 0.0, Ci3_mu1_fft_med_max= 0.0;
    double mu1_wl_avg_bs = 0.0, mu1_wl_avg_mil = 0.0, mu1_wl_avg_noncentered = 0.0, mu1_wl_avg = 0.0,
            mu1_wl_boot_avg = 0.0, mu1_wl_stdev_avg = 0.0, Bias_mu1_wl_avg = 0.0,
            Rms_mu1_wl_avg = 0.0, Ci_mu1_wl_avg_min= 0.0, Ci_mu1_wl_avg_max= 0.0,
            Ci2_mu1_wl_avg_min= 0.0, Ci2_mu1_wl_avg_max= 0.0,  Ci3_mu1_wl_avg_min= 0.0, Ci3_mu1_wl_avg_max= 0.0;
    double mu1_wl_med = 0.0, mu1_wl_boot_med = 0.0, mu1_wl_stdev_med = 0.0, Bias_mu1_wl_med = 0.0,
            Rms_mu1_wl_med = 0.0, Ci_mu1_wl_med_min= 0.0, Ci_mu1_wl_med_max= 0.0,
            Ci2_mu1_wl_med_min= 0.0, Ci2_mu1_wl_med_max= 0.0, Ci3_mu1_wl_med_min= 0.0, Ci3_mu1_wl_med_max= 0.0;
    double skew_mean_fft = 0.0,  skew_mean_wl =0.0, skew_median_fft = 0.0,  skew_median_wl =0.0,
            kurtosis_mean_fft = 0.0,  kurtosis_mean_wl =0.0, kurtosis_median_fft = 0.0,  kurtosis_median_wl =0.0;

    double jack_fft, jack_wl;
    double local_sidereal, hour_angle;
    vector_double *time_scan, *projected_baseline, *azimuth_baseline, *u_sampled, *v_sampled;
    double buffer_projected_baseline, buffer_azimuth_baseline, buffer_u_sampled, buffer_v_sampled;
    
    double avg_photom;
    double kappa_norm, kappa_norm_var, relative_kappa_norm_var;
    double simple_error_fft, simple_error_wl;
    double n_sig_reject;
    vector_double *Apodisation_fringes;
    
    vector_double *Background_PA, *Background_PB,
        *Background_I1, *Background_I2, *Background_PA_off, *Background_PB_off;
    vector_double *Photometry_I1, *Photometry_I2, *Photometry_PA, *Photometry_PB; /* vector of averages over each scan */
    
    batch *Batch_on_raw; /* variable used to save the original Batch_on - pkervell 10-Oct-02 */
    
    full_scan *Master_spectrum_off; /* added pkervell 11-Oct-02 */
    vector_double *Backgrounds;
    double SNR_PA, SNR_PB, SNR_I1, SNR_I2;
    long index_time_min, index_time_max;
    double wavenum_wl_min, wavenum_wl_max, wavenum_fft_min, wavenum_fft_max;
    long theoretical_peak_pos;
    matrix_double *PSDs_matrix, *Phase_matrix, *PSDs_matrix_wl, *Phase_matrix_wl, *PSDs_matrix_off, *PSDs_matrix_wl_off;
    double weighted_peak_pos_fft_centered, weighted_peak_pos_wl_centered;
    
    double unbalance_buffer, average_unbalance;
    vector_double *Unbalance_coef, *non_zero_unbalance;
    double unweighted_mu1_fft, unweighted_mu1_wl;
    /* off-source power spectrum verification variables */
    vector_double *off_source_spectrum_slopes, *number_of_pickups;
    double slope_background_fft, slope_background_wl;
    int Autotest_case;
    vector_double *gains, *RONs;
    long scans_rejected_0 = 0, scans_rejected_1 = 0, scans_rejected_2 = 0, scans_rejected_3 = 0,
         scans_rejected_4 = 0, scans_rejected_5 = 0, scans_rejected_6 = 0, scans_rejected_7 = 0,
         scans_rejected_8 = 0;
    double total_flux_A_only, total_flux_B_only;
    vector_double *splitting_ratio_A, *splitting_ratio_B;
    vector_double *histogram_scale_wl, *histogram_wl, *histogram_scale_fft, *histogram_fft;
    double weighted_peak_pos, weighted_peak_pos_wl;
    double peak_pos_mean, peak_pos_sigma, peak_width_sigma, peak_width_avg;
    double buffer_altitude_object, buffer_azimuth_object;
    vector_double *altitude_object, *azimuth_object;
    double altitude_batch_obj =0.0, azimuth_batch_obj =0.0;
    double wavenum_min_halfmax_fft, wavenum_max_halfmax_fft,
           wavenum_min_halfmax_wl, wavenum_max_halfmax_wl;
    double on_src_bg_I1 = 0.0, on_src_bg_I2 = 0.0, on_src_bg_PA = 0.0, on_src_bg_PB = 0.0;
    
    vector_double *Real_part_fft_buffer, *Imaginary_part_fft_buffer;
    matrix_double *Real_part_fft, *Imaginary_part_fft;
    vector_double *Real_part_fft_no_centering, *Imaginary_part_fft_no_centering;
    vector_double *Amplitude_fft, *Phase_fft;
    vector_double *Real_fft, *Imaginary_fft;
    vector_double *OPD_phase, *Total_OPD;
    
    vector_double *high_freq_soft;
    double threshold_soft;
    complex complex_null;
    matrix_double *norm_t_wavelets_avg; /* pkervell - 03-Feb-03, to illustrate VINCI paper */

    /* pkervell - 15-Oct-03 */
    FILE *ascii_values1;
    FILE *ascii_values2;
    FILE *ascii_values3;
    FILE *ascii_values4;

    double avg_background_fourier=0.0, avg_background_wavelets=0.0;
    /* per scan BG , Fourier not used for the moment - pkervell - 04-Jan-04 */
    vector_double *fringe_peak_strehl; /* added - pkervell - 21-Jan-04 */
    /* END OF VARIABLE DECLARATIONS */


    /* VARIABLES INITIALIZATIONS */
    /* For wavelets localization of the fringes */
    if(Batch_on->fullscans[0]->I1->dimension < 512)
    {
        ii=0;
	while ( pow(2.0,++ii)< Batch_on->fullscans[0]->I1->dimension ); 
        N_frames_precut =  pow(2.0,ii);
    }
    else
      N_frames_precut   = 512;  /* value set by default if the dimension of the input scan is large enough */

    /* Set number of scans, frames etc. */
    N_scans_off       = Batch_off->number_of_scans;
    N_scans_on        = Batch_on->number_of_scans;
    N_mixed           = Batch_A->number_of_scans;
    N_frames_on       = Batch_on->fullscans[0]->I1->dimension;
    *N_scans_processed = 0;
    *numberOfScansRejected = 0;
    *time_batch = 0.0;
    *baseline_batch = 0.0;
    *azimuth_batch = 0.0;
    *u_batch = 0.0;
    *v_batch = 0.0;
    *Mu1_batch_wl = 0.0;
    *ErrMu1_batch_min_wl = 0.0;
    *ErrMu1_batch_max_wl = 0.0;
    *Mu1_batch_fft = 0.0;
    *ErrMu1_batch_min_fft = 0.0;
    *ErrMu1_batch_max_fft = 0.0;
    /* END OF VARIABLES INITIALIZATIONS */


    /*===================*/
    /* MEMORY ALLOCATION */
    /*===================*/
    /* Allocate memory for the per scan transfer matrix */
    Standard_dev_Matrix = vndrs_calloc_matrix_double(2,2); /* pkervell 06-Sep-02 */  
    Off_offsets_Mean   = vndrs_calloc_vector_double(4);
    Off_offsets_Stddev = vndrs_calloc_vector_double(4);
    Off_offsets_Min    = vndrs_calloc_vector_double(4);
    Off_offsets_Max    = vndrs_calloc_vector_double(4);
    On_source_Mean   = vndrs_calloc_vector_double(4);
    On_source_Stddev = vndrs_calloc_vector_double(4);
    On_source_Min    = vndrs_calloc_vector_double(4);
    On_source_Max    = vndrs_calloc_vector_double(4);
    A_Mean   = vndrs_calloc_vector_double(4);
    A_Stddev = vndrs_calloc_vector_double(4);
    A_Min    = vndrs_calloc_vector_double(4);
    A_Max    = vndrs_calloc_vector_double(4);
    B_Mean   = vndrs_calloc_vector_double(4);
    B_Stddev = vndrs_calloc_vector_double(4);
    B_Min    = vndrs_calloc_vector_double(4);
    B_Max    = vndrs_calloc_vector_double(4);
    Batch_A_cut = vndrs_calloc_batch(N_mixed, N_frames_cut);
    Batch_B_cut = vndrs_calloc_batch(N_mixed, N_frames_cut);
    Batch_on_cut = vndrs_calloc_batch(N_scans_on, N_frames_cut);
    Batch_off_cut = vndrs_calloc_batch(N_scans_off, N_frames_cut);
    Batch_on_precut = vndrs_calloc_batch(N_scans_on, N_frames_precut);
    Batch_off_precut = vndrs_calloc_batch(N_scans_off, N_frames_precut);
    fringes_found_wavelets = vndrs_calloc_vector_double(N_scans_on);
    Mu1_squared_wl = vndrs_calloc_vector_double(N_scans_on);
    Mu2_squared_wl = vndrs_calloc_vector_double(N_scans_on);
    residus_QL =   vndrs_calloc_vector_double(N_scans_on);
    peak_QL =   vndrs_calloc_vector_double(N_scans_on);
    /* added pkervell - 01-Sep-02 variables for scans quality control */
    in_to_outband_ratio = vndrs_calloc_vector_double(N_scans_on);
    fwhm_peak_time = vndrs_calloc_vector_double(N_scans_on);
    peak_position_freq = vndrs_calloc_vector_double(N_scans_on);
    fwhm_peak_freq = vndrs_calloc_vector_double(N_scans_on);
    background_fit_QL = vndrs_calloc_vector_double(N_scans_on);
    fringes_off_center = vndrs_calloc_vector_double(N_scans_on);
    signal_p = vndrs_calloc_vector_double(N_scans_on);
    Flag_processed = vndrs_calloc_vector_double(N_scans_on);
    fringe_peak_strehl = vndrs_calloc_vector_double(N_scans_on); /* pkervell - 21-Jan-04 */
    /* Average power spectra - added pkervell - 09-Aug-02 */
    Pow_spec_avg_I1 = vndrs_calloc_vector_double(N_frames_cut);
    Phase_I1 = vndrs_calloc_vector_double(N_frames_cut);
    Phase_avg_I1_no_centering = vndrs_calloc_vector_double(N_frames_cut);
    Pow_spec_avg_I1_no_centering = vndrs_calloc_vector_double(N_frames_cut);
    Pow_spec_avg_I1_no_centering_corrected = vndrs_calloc_vector_double(N_frames_cut);
    Pow_spec_avg_buffer_I1 = vndrs_calloc_vector_double(N_frames_cut);
    Pow_spec_avg_I1_off = vndrs_calloc_vector_double(N_frames_cut);
    Pow_spec_avg_buffer_I1_off = vndrs_calloc_vector_double(N_frames_cut);
    Background_PSD = vndrs_calloc_vector_double(N_frames_cut);
    Pow_spec_avg_I1_corrected = vndrs_calloc_vector_double(N_frames_cut);
    
    Background_PSD_wl_vector = vndrs_calloc_vector_double(N_frames_cut/2);
    WL_spec_avg_I1_vector = vndrs_calloc_vector_double(N_frames_cut/2);
    WL_spec_avg_I1_vector_no_centering = vndrs_calloc_vector_double(N_frames_cut/2);
    WL_spec_avg_I1_vector_no_centering_corrected = vndrs_calloc_vector_double(N_frames_cut/2);
    WL_spec_avg_I1_vector_corrected = vndrs_calloc_vector_double(N_frames_cut/2);
    WL_spec_avg_I1_vector_off = vndrs_calloc_vector_double(N_frames_cut/2);
    
    weights_visibilities = vndrs_calloc_vector_double(N_scans_on);
    
    removed_indices = vndrs_calloc_vector_double(N_scans_on);
    
    Avg_power_spectra_on = vndrs_calloc_full_scan(N_frames_precut);
    Avg_power_spectra_off = vndrs_calloc_full_scan(N_frames_precut);
    Freq_scale_avg_spectra = vndrs_calloc_vector_double(N_frames_precut/2); /* only up to Nyquist */
    
    Freq_scale_fringes = vndrs_calloc_vector_double(N_frames_cut/2); /* only up to Nyquist */
    Background_PA = vndrs_calloc_vector_double(N_scans_on);
    Background_PB = vndrs_calloc_vector_double(N_scans_on);
    Background_I1 = vndrs_calloc_vector_double(N_scans_on);
    Background_I2 = vndrs_calloc_vector_double(N_scans_on);
    Background_PA_off = vndrs_calloc_vector_double(N_scans_off);
    Background_PB_off = vndrs_calloc_vector_double(N_scans_off);
    Photometry_I1 = vndrs_calloc_vector_double(N_scans_on);
    Photometry_I2 = vndrs_calloc_vector_double(N_scans_on);
    Photometry_PA = vndrs_calloc_vector_double(N_scans_on);
    Photometry_PB = vndrs_calloc_vector_double(N_scans_on);
    Batch_on_raw = vndrs_calloc_batch(N_scans_on, N_frames_on);
    Master_spectrum_off = vndrs_calloc_full_scan(N_frames_cut);
    Backgrounds = vndrs_calloc_vector_double(6);
    
    PSDs_matrix = vndrs_calloc_matrix_double(N_scans_on,N_frames_cut/2);
    Phase_matrix = vndrs_calloc_matrix_double(N_scans_on,N_frames_cut/2);
    Phase_matrix_wl = vndrs_calloc_matrix_double(N_scans_on,N_frames_cut/2);
    PSDs_matrix_wl = vndrs_calloc_matrix_double(N_scans_on,N_frames_cut/2);
    PSDs_matrix_off = vndrs_calloc_matrix_double(N_scans_on,N_frames_cut/2);
    PSDs_matrix_wl_off = vndrs_calloc_matrix_double(N_scans_on,N_frames_cut/2);
    
    time_scan = vndrs_calloc_vector_double(N_scans_on);
    projected_baseline = vndrs_calloc_vector_double(N_scans_on);
    azimuth_baseline = vndrs_calloc_vector_double(N_scans_on);
    u_sampled = vndrs_calloc_vector_double(N_scans_on);
    v_sampled = vndrs_calloc_vector_double(N_scans_on);
    altitude_object = vndrs_calloc_vector_double(N_scans_on);
    azimuth_object = vndrs_calloc_vector_double(N_scans_on);
    
    norm_t_wavelets_I1_check = vndrs_calloc_matrix_double(N_frames_cut, N_frames_cut/2);
    Unbalance_coef = vndrs_calloc_vector_double(N_scans_on);
    off_source_spectrum_slopes = vndrs_calloc_vector_double(4);
    number_of_pickups = vndrs_calloc_vector_double(4);
    gains = vndrs_calloc_vector_double(4);
    RONs = vndrs_calloc_vector_double(4);
    splitting_ratio_A = vndrs_calloc_vector_double(4);
    splitting_ratio_B = vndrs_calloc_vector_double(4);
    
    Amplitude_fft = vndrs_calloc_vector_double(N_scans_on);
    Phase_fft = vndrs_calloc_vector_double(N_scans_on);
    Real_fft = vndrs_calloc_vector_double(N_scans_on);
    Imaginary_fft = vndrs_calloc_vector_double(N_scans_on);
    Real_part_fft_buffer = vndrs_calloc_vector_double(N_frames_cut);
    Imaginary_part_fft_buffer = vndrs_calloc_vector_double(N_frames_cut);
    Real_part_fft = vndrs_calloc_matrix_double(N_scans_on, N_frames_cut/2);
    Imaginary_part_fft = vndrs_calloc_matrix_double(N_scans_on, N_frames_cut/2);
    Real_part_fft_no_centering = vndrs_calloc_vector_double(N_frames_cut/2);
    Imaginary_part_fft_no_centering = vndrs_calloc_vector_double(N_frames_cut/2);
    OPD_phase = vndrs_calloc_vector_double(N_scans_on);
    Total_OPD = vndrs_calloc_vector_double(N_scans_on);
    high_freq_soft = vndrs_calloc_vector_double(N_frames_cut);
    
    norm_t_wavelets_avg = vndrs_calloc_matrix_double(N_frames_cut, N_frames_cut/2);
    /* END OF MEMORY ALLOCATION */
    
    /* Saving the original batch_on */
    vndrs_batch_memcpy(Batch_on_raw, Batch_on);
      
    /* INITIALIZE BATCH OUTPUT DIRECTORY FOR VERBOSE */
    if(verbose_on == 1)
    {
        sprintf(DataPath,"./Batch");
        sprintf(command,"mkdir -p %s",DataPath);
        system(command);
        strcat(DataPath,"/");
    }

    /* Perform statistical operations on the Off_source batch */

    vndrs_batch_statistics(Batch_off, Off_offsets_Mean, Off_offsets_Stddev, Off_offsets_Min, Off_offsets_Max);
    vndrs_batch_statistics(Batch_A, A_Mean, A_Stddev, A_Min, A_Max);
    vndrs_batch_statistics(Batch_B, B_Mean, B_Stddev, B_Min, B_Max);
    vndrs_batch_statistics(Batch_on, On_source_Mean, On_source_Stddev, On_source_Min, On_source_Max);

    /* Computation of splitting ratios of the beam combiner for monitoring - pkervell - 22-Oct-02 */
    total_flux_A_only = vndrs_sum_scan(A_Mean) - vndrs_sum_scan(Off_offsets_Mean);
    total_flux_B_only = vndrs_sum_scan(B_Mean) - vndrs_sum_scan(Off_offsets_Mean);
    for(I=0;I<4;I++)
    {
        splitting_ratio_A->data[I] = (A_Mean->data[I]-Off_offsets_Mean->data[I])/total_flux_A_only;
        splitting_ratio_B->data[I] = (B_Mean->data[I]-Off_offsets_Mean->data[I])/total_flux_B_only;
    }

    SNR_I1 = (On_source_Mean->data[0]-Off_offsets_Mean->data[0])/Off_offsets_Stddev->data[0];
    SNR_I2 = (On_source_Mean->data[1]-Off_offsets_Mean->data[1])/Off_offsets_Stddev->data[1];
    SNR_PA = (On_source_Mean->data[2]-Off_offsets_Mean->data[2])/Off_offsets_Stddev->data[2];
    SNR_PB = (On_source_Mean->data[3]-Off_offsets_Mean->data[3])/Off_offsets_Stddev->data[3];
    
    vndrs_photometry_info(Off_offsets_Stddev, Off_offsets_Mean, A_Mean, B_Mean, On_source_Mean,
                          SNR_I1, SNR_I2, SNR_PA, SNR_PB);
                          
    /* ------------ */

    /* Cutting of the scans in order to center the fringes and to accelerate the processing */
    vndrs_trim_batch(Batch_A, Fringe_speed, Frame_rate, 0, N_frames_cut, Batch_A_cut);
    vndrs_trim_batch(Batch_B, Fringe_speed, Frame_rate, 0, N_frames_cut, Batch_B_cut);
    vndrs_trim_batch(Batch_on, Fringe_speed, Frame_rate, 0, N_frames_cut, Batch_on_cut);
    vndrs_trim_batch(Batch_off, Fringe_speed, Frame_rate, 0, N_frames_cut, Batch_off_cut);
    /* pkervell - 03-Sep-02 : necessary for the Wiener filter computation the cut is superseded
       after the fringes are really centered */

                     
    /* Compute the sample interval in wavenumber space (unit 1/m) */
    /* ---------------------------------------------------------- */
    /* Explanation:
        The sample interval in wavenumber space is the reciprocal
        value of the extension of the scan in OPD space:
        
        Extension in OPD space = N_frames_on * Delta_OPD
        
        with Delta_OPD being the sample interval in OPD space (unit m).
        
        Delta_OPD = Fringe_speed * Delta_t = Fringe_speed / Frame_rate
        
        with the OPD modulation speed (unit m/sec) 'Fringe_speed' and
        the sample interval in time (units sec) 'Delta_t' = 1/Frame_rate.
        Hence, the extension in OPD space is
        
        Extension in OPD space = N_frames_on * Fringe_speed / Frame_rate
        
        Its reciprocal value is the the sample interval in wavenumber space:
        
        Wavenum_interval = Frame_rate / Fringe_speed / N_frames_on
    */
    
    Wavenum_interval = fabs(Frame_rate / Fringe_speed / N_frames_cut); /* unit: 1/m */
    Wavenum_interval_precut = fabs(Frame_rate / Fringe_speed / N_frames_precut); /* unit: 1/m */

    /* Extreme limits of the PSD from FFT */
    Wavenum_min = 0.0;
    Wavenum_max = fabs(Frame_rate / (2 * Fringe_speed));  /* Nyquist freq. */

    /* Computation of the integration bandwidth */
    Wavenum_bandwidth_fft = Wavenum_limits_fft->data[1] - Wavenum_limits_fft->data[0];
    Wavenum_bandwidth_wl = Wavenum_limits_wl->data[1] - Wavenum_limits_wl->data[0];

    /* This is the width of the frequency integration ranges for the mu^2 computation */
    /* The lateral bands are TO BE INVESTIGATED ! */
    Wavenum_min_lowband_fft = 0.7*Wavenum_limits_fft->data[0];
    Wavenum_max_lowband_fft = Wavenum_limits_fft->data[0]; 
    Wavenum_min_hi_band_fft = Wavenum_limits_fft->data[1]; 
    Wavenum_max_hi_band_fft = 0.9*Wavenum_max;

    Wavenum_min_lowband_wl = 0.7*Wavenum_limits_wl->data[0];
    Wavenum_max_lowband_wl = Wavenum_limits_wl->data[0]; 
    Wavenum_min_hi_band_wl = Wavenum_limits_wl->data[1]; 
    Wavenum_max_hi_band_wl = 0.9*Wavenum_max /* Wavenum_min_hi_band_wl + 0.6*Wavenum_limits_fft->data[0]*/;
    /* we do not want the extreme edges of the spectrum */
        
    Wavenum_min_out = 0.0;
    Wavenum_max_out = Wavenum_max;
    
    theoretical_peak_pos = (long)(Wavenum_theo/Wavenum_interval);
        
    index_freq_low_fft  = (long)(Wavenum_limits_fft->data[0]/fabs(Frame_rate / Fringe_speed / N_frames_precut));
    index_freq_high_fft = (long)(Wavenum_limits_fft->data[1]/fabs(Frame_rate / Fringe_speed / N_frames_precut));
    
    index_freq_low_wl  = (long)(Wavenum_limits_wl->data[0]/fabs(Frame_rate / Fringe_speed / N_frames_precut));
    index_freq_high_wl = (long)(Wavenum_limits_wl->data[1]/fabs(Frame_rate / Fringe_speed / N_frames_precut));

    index_freq_low_fft_cut  = (long)(Wavenum_limits_fft->data[0]/fabs(Frame_rate / Fringe_speed / N_frames_cut));
    index_freq_high_fft_cut = (long)(Wavenum_limits_fft->data[1]/fabs(Frame_rate / Fringe_speed / N_frames_cut));

    index_freq_low_wl_cut  = (long)(Wavenum_limits_wl->data[0]/fabs(Frame_rate / Fringe_speed / N_frames_cut));
    index_freq_high_wl_cut = (long)(Wavenum_limits_wl->data[1]/fabs(Frame_rate / Fringe_speed / N_frames_cut));

    index_freq_low_QL_cut  = (long)(Wavenum_limits_QL->data[0]/fabs(Frame_rate / Fringe_speed / N_frames_cut));
    index_freq_high_QL_cut = (long)(Wavenum_limits_QL->data[1]/fabs(Frame_rate / Fringe_speed / N_frames_cut));

    Sample_interval = fabs(1.0*Fringe_speed / Frame_rate);

    /* Normalization factors for the power spectra */
    Normalization_Fourier = 4.0 / (Shape_factor * Wavenum_interval* Wavenum_interval);
    Normalization_WL = 8.0 / (Shape_factor * Wavenum_interval * Wavenum_interval * sqrt(N_frames_cut));;

    /* Relative surfaces of the different integration regions */
    surface_in = 1.0;
    surface_out_fft = (Wavenum_max_hi_band_fft-Wavenum_min_hi_band_fft)/(Wavenum_limits_fft->data[1]-Wavenum_limits_fft->data[0]);
    surface_out_wl = (Wavenum_max_hi_band_wl-Wavenum_min_hi_band_wl)/(Wavenum_limits_wl->data[1]-Wavenum_limits_wl->data[0]);

    /* Computes the Wiener filters from the average PSDs of the photometric channels */
    WA_mean = vndrs_compute_wiener_mean(Batch_on_cut, verbose_on, 0);
    WB_mean = vndrs_compute_wiener_mean(Batch_on_cut, verbose_on, 1);
    WA_mean->data[0] = 1.0;
    WB_mean->data[0] = 1.0; /* to be sure... */
    WA_mean->data[WA_mean->dimension-1] = 1.0;
    WB_mean->data[WB_mean->dimension-1] = 1.0; /* to be sure... */
    /* WA_mean and WB_mean are allocated inside the procedure */

    

    if(Choice_matrix == 1 || ionic_in == 1) /* only method when IONIC is used */
    {
        printf("\n  Kappa matrix computed using standard method.\n");
        sprintf(vndrs_protocol.text,"# Kappa matrix values computed using standard method. \n");
        vndrs_runtime_protocol();
        /* Kappa matrix computation, using the regular off-source offsets and the standard FLUOR method */
        /* We use the cut versions here to be able to use the Wiener filters.*/
        vndrs_select_matrix(Batch_A_cut, 
                        Batch_B_cut,
                        Batch_on_cut,
                        Off_offsets_Mean,
                        WA_mean,
                        WB_mean,
                        ionic_in,
                        Average_Matrix,
                        Standard_dev_Matrix,
                        Relative_MSE,
                        verbose_on);
    }
    else
    {
        sprintf(vndrs_protocol.text,"# Kappa matrix values computed using least squares method.\n");
        vndrs_runtime_protocol();
        /* Least squares computation (background independent) - pkervell - 13-Oct-02 */
        vndrs_compute_matrix_least_squares(Batch_off, Batch_A, Batch_B, Batch_on, ionic_in,
                                        Average_Matrix, Standard_dev_Matrix, &Autotest_case, verbose_on);
    }
    
    /* Save the matrix values in easier to read variables */
    vndrs_get_transfer_matrix_coefficients(Average_Matrix, &kappa1A, &kappa1B, &kappa2A, &kappa2B);
    vndrs_get_transfer_matrix_coefficients(Standard_dev_Matrix, &kappa1A_err, &kappa1B_err, &kappa2A_err, &kappa2B_err);

    if(ionic_in == 0)
    { /* MONA case */
        printf("  Kappa matrix :                    K1A = %6.4f  K1B = %6.4f  K2A = %6.4f  K2B = %6.4f \n",
            kappa1A, kappa1B, kappa2A, kappa2B);
        printf("  Kappa matrix errors :             K1A = %6.4f  K1B = %6.4f  K2A = %6.4f  K2B = %6.4f \n",
            kappa1A_err, kappa1B_err, kappa2A_err, kappa2B_err);

        sprintf(vndrs_protocol.text,"# Kappa matrix values : Kappa1A, Kappa1B, Kappa2A, Kappa2B \n");
        vndrs_runtime_protocol();
        sprintf(vndrs_protocol.text,"  %12e   %12e   %12e   %12e\n", kappa1A, kappa1B, kappa2A, kappa2B);
        vndrs_runtime_protocol();
        sprintf(vndrs_protocol.text,"# Kappa matrix standard errors :  Kap1A_err, Kap1B_err, Kap2A_err, Kap2B_err\n");
        vndrs_runtime_protocol();
        sprintf(vndrs_protocol.text,"  %12e   %12e   %12e   %12e\n", kappa1A_err, kappa1B_err, kappa2A_err, kappa2B_err);
        vndrs_runtime_protocol();
    }
    else
    {	/* IONIC case */
        printf("  Kappa matrix for IONIC:           K1A = %6.4f  K1B = %6.4f \n",
            kappa1A, kappa1B);
        printf("  Kappa matrix errors :             K1A = %6.4f  K1B = %6.4f \n",
            kappa1A_err, kappa1B_err);    

        sprintf(vndrs_protocol.text,"# Kappa matrix values for IONIC : Kappa1A, Kappa1B \n");
        vndrs_runtime_protocol();
        sprintf(vndrs_protocol.text,"  %12e   %12e\n", kappa1A, kappa1B);
        vndrs_runtime_protocol();
        sprintf(vndrs_protocol.text,"# Kappa matrix standard errors :  Kap1A_err, Kap1B_err\n");
        vndrs_runtime_protocol();
        sprintf(vndrs_protocol.text,"  %12e   %12e \n", kappa1A_err, kappa1B_err);
        vndrs_runtime_protocol();
    }
    

    if(kappa1A < 0.0 || kappa1B < 0.0 || kappa2A < 0.0 || kappa2B < 0.0)
    { /* error case: some of the coefficients are negative - pkervell - 13-Nov-02 */
        kappa1A = 0.0; kappa1B = 0.0; kappa2A = 0.0; kappa2B = 0.0;
        printf("\n  *** Error: Kappa matrix shows negative values.\n");
        sprintf(vndrs_protocol.text,"#  *** Error: Kappa matrix shows negative values. \n");
    }

    /* --------- */

    if(ionic_in == 0)
    {
        /*
            Computes the mean and variance of the variable (kappa1 + kappa2)/(kappa1.kappa2)
            with kappa1=kappa1A.kappa1B and kappa2=kappa2A.kappa2B, knowing the kappa matrix and errors.
        */
        
        kappa_norm = vndrs_mean_kappa_norm(Average_Matrix,
                                        Standard_dev_Matrix);
        kappa_norm_var = vndrs_variance_kappa_norm(Average_Matrix,
                                        Standard_dev_Matrix);
        relative_kappa_norm_var = kappa_norm_var / (kappa_norm*kappa_norm);
        printf("\n  Kappa norm introduced relative error = %4.2f %%\n",  100.0*sqrt(relative_kappa_norm_var));
    }
    else
    {
        relative_kappa_norm_var =  (kappa1A_err*kappa1A_err + kappa1B_err*kappa1B_err)/(kappa1A*kappa1A + kappa1B*kappa1B);
        /* in the case of IONIC, we output this error (correct ?) - pkervell 13-Feb-03 */
        printf("\n  Kappa norm introduced relative error = %4.2f %%\n",  100.0*sqrt(relative_kappa_norm_var));
    }
    
    /* ------------- */

    /* Independant estimation of the background offset level (not directly from off source batch) */
    if(background_per_scan == 1)
    {
        vndrs_evaluate_off_source_background(Batch_on_raw,
                                                Batch_A,
                                                Batch_B,
                                                Batch_off,
                                                Average_Matrix,
                                                Background_PA_off,
                                                Background_PB_off,
                                                Background_I1,
                                                Background_I2,
                                                Background_PA,
                                                Background_PB);
            
        on_src_bg_I1 = vndrs_mean_scan(Background_I1);
        on_src_bg_I2 = vndrs_mean_scan(Background_I2);
        on_src_bg_PA = vndrs_mean_scan(Background_PA);
        on_src_bg_PB = vndrs_mean_scan(Background_PB);
            
        printf("\n  Avg ONSRC background (ADUs):      I1 = %7.2f  I2 = %7.2f  PA = %7.2f  PB = %7.2f \n",
        on_src_bg_I1, on_src_bg_I2, on_src_bg_PA, on_src_bg_PB);
            
        sprintf(vndrs_protocol.text,"# Avg bg on-source (ADUs) I1, I2, PA, PB:\n");
        vndrs_runtime_protocol();
        sprintf(vndrs_protocol.text,"  %7.4f   %7.4f   %7.4f   %7.4f\n",
                on_src_bg_I1, on_src_bg_I2, on_src_bg_PA, on_src_bg_PB);
        vndrs_runtime_protocol();

        /* Actual removal of the backgrounds */
        vndrs_remove_backgrounds_per_scan(Batch_on, /* in/out */
                                        Background_I1,
                                        Background_I2,
                                        Background_PA,
                                        Background_PB);
        vndrs_remove_backgrounds_per_scan(Batch_on_precut, /* in/out */
                                        Background_I1,
                                        Background_I2,
                                        Background_PA,
                                        Background_PB);
        vndrs_remove_backgrounds_per_scan(Batch_off, /* in/out */
                                        Background_I1,
                                        Background_I2,
                                        Background_PA,
                                        Background_PB);
        vndrs_remove_backgrounds_per_scan(Batch_off_precut, /* in/out */
                                        Background_I1,
                                        Background_I2,
                                        Background_PA,
                                        Background_PB);
        vndrs_remove_backgrounds_per_scan(Batch_off_cut, /* in/out */
                                        Background_I1,
                                        Background_I2,
                                        Background_PA,
                                        Background_PB);

    }
    else
    {	/* Removal of the regular (constant) offsets computed from the offsrc */
        vndrs_remove_offsets(Batch_on, Off_offsets_Mean);
        vndrs_remove_offsets(Batch_off, Off_offsets_Mean);
        vndrs_remove_offsets(Batch_on_precut, Off_offsets_Mean);
        vndrs_remove_offsets(Batch_off_precut, Off_offsets_Mean);
        vndrs_remove_offsets(Batch_on_cut, Off_offsets_Mean);
        vndrs_remove_offsets(Batch_off_cut, Off_offsets_Mean);
    }
    
    /* Computation of the average of the signals over each scan of the batch - pkervell 10-Oct-02 */
    vndrs_compute_average_vector_from_batch(Batch_on, Photometry_I1, Photometry_I2, Photometry_PA, Photometry_PB);

    /* Precutting of the background corrected scans */
    vndrs_trim_batch(Batch_on, Fringe_speed, Frame_rate, 0, N_frames_precut, Batch_on_precut);
    vndrs_trim_batch(Batch_off, Fringe_speed, Frame_rate, 0, N_frames_precut, Batch_off_precut);

    /* Computation of the average signals power spectra over the batch */
    vndrs_compute_average_power_spectra(Batch_on_precut,
                                        Avg_power_spectra_on);
    vndrs_compute_average_power_spectra(Batch_off_precut,
                                        Avg_power_spectra_off);
    vndrs_compute_freq_scale(Frame_rate, Freq_scale_avg_spectra);
    /* We build also a frequency scale (in WAVENUMBER, cm^-1) for the power spectra of the processed fringes */
    vndrs_compute_freq_scale(2.0*Wavenum_max/100.0, Freq_scale_fringes);
    
    /* Computation of the LISA camera gains and RONs - pkervell 19-Oct-02 */
    vndrs_compute_camera_gains(Avg_power_spectra_off, Avg_power_spectra_on, gains, RONs);
    
    /* Here we check that the off-source power spectrum (readout noise) is really flat for all pixels
       and that no pickup is present.  The procedure returns the number of pickups above
       5 sigmas of the spectrum noise. Results are displayed and saved at the end of the processing,
       in the log. pkervell - 18-Oct-02 */
    vndrs_check_readout_noise_spectrum(Avg_power_spectra_off, off_source_spectrum_slopes, number_of_pickups);

    /* --------- */
    
    /* Computation of the average spectrum of the detector noise, used to remove the off-source processing - pkervell 11-Oct-02 */
    
    if(debias == 1)
    {
        vndrs_compute_average_power_spectra(Batch_off_cut, Master_spectrum_off);
        
        /* Setting average values to zero */
        Master_spectrum_off->I1->data[0] = 0.0; Master_spectrum_off->I1->data[N_frames_cut-1] = 0.0;
        Master_spectrum_off->I2->data[0] = 0.0; Master_spectrum_off->I2->data[N_frames_cut-1] = 0.0;
        Master_spectrum_off->PA->data[0] = 0.0; Master_spectrum_off->PA->data[N_frames_cut-1] = 0.0;
        Master_spectrum_off->PB->data[0] = 0.0; Master_spectrum_off->PB->data[N_frames_cut-1] = 0.0;

    }
        
    /* vndrs_fprintf_vector_double(Master_spectrum_off->I1, "Master_spectrum_off_I1.dat");
    vndrs_fprintf_vector_double(Master_spectrum_off->I2, "Master_spectrum_off_I2.dat");*/
    
    /* --------- */
    
    vndrs_display_and_write_header();

    /* Creation of the filter for the spatial apodisation of the interferogram around the fringes (Fourier only) */
    Apodisation_fringes =  vndrs_apodisation_center(N_frames_cut, fringes_length, (long)fringes_length/10);

    /* ================================================================= */
    /*                           MAIN LOOP                               */
    /*                  over all scans of the batch                      */
    /* ================================================================= */ 
    
    
    /*  MEMORY ALLOCATION FOR I1  WAVELETS */
    coi_I1             = vndrs_calloc_vector_double(N_frames_cut);
    t_wavelets_I1      = vndrs_calloc_matrix_complex(N_frames_cut,N_frames_cut/2);
    t_wavelets_I1_off  = vndrs_calloc_matrix_complex(N_frames_cut,N_frames_cut/2);
    N_frames_off       = Batch_off->fullscans[0]->I1->dimension;
    
    if((kappa1A > 0.05 && kappa1B > 0.05 && kappa2A > 0.05 && kappa2B > 0.05) || (ionic_in == 1))
    {
        for (I=0; I<N_scans_on; I++)
        {
            
            if ( verbose_on == 1)
            {
                /* Creates temporary directories */ 
                sprintf(DataPath,"./Scan_%05d",(int)I);
                sprintf(command,"mkdir -p %s",DataPath);
                system(command);
                strcat(DataPath,"/");
                /* Dump the non-processed (raw) On_source scans to disk */
                vndrs_fprintf_vector_double(Batch_on->fullscans[I]->I1, "Raw_On_I1.dat");
                vndrs_fprintf_vector_double(Batch_on->fullscans[I]->I2, "Raw_On_I2.dat");
                vndrs_fprintf_vector_double(Batch_on->fullscans[I]->PA, "Raw_On_PA.dat");
                vndrs_fprintf_vector_double(Batch_on->fullscans[I]->PB, "Raw_On_PB.dat");
            }
                
            /* Saving time stamp and Delay Line offset for the scan */ 
            Batch_on_processed->scan_data[I].time      = Batch_on->scan_data[I].time;
            Batch_on_processed->scan_data[I].dl_offset = Batch_on->scan_data[I].dl_offset;
    
            /* Computation of baseline, time,... for the current scan */
            /*  The julian date is computed from the current scan time, which is expressed in days since JJ = 2400000.5
                This convention is called modified julian date. It is not kept in the final data file, where the
                regular Julian date is written. */
            time_scan->data[I] = Batch_on_processed->scan_data[I].time + 2400000.5;
            local_sidereal = vndrs_juliandate_to_local_sidereal(time_scan->data[I], longi);
            hour_angle = (local_sidereal - RA_object*12.0/180.0);  /* beware: RA and Dec are in degrees ! */
            
            vndrs_project_baseline(hour_angle, 
                                    Dec_object,
                                    lat,
                                    t1x,t1y,t1z,
                                    t2x,t2y,t2z,
                                    &buffer_altitude_object,
                                    &buffer_azimuth_object,
                                    &buffer_projected_baseline,
                                    &buffer_azimuth_baseline,
                                    &buffer_u_sampled, &buffer_v_sampled);
                                    
            projected_baseline->data[I] = buffer_projected_baseline;
            azimuth_baseline->data[I] = buffer_azimuth_baseline;
            altitude_object->data[I] = buffer_altitude_object;
            azimuth_object->data[I] = buffer_azimuth_object;
            u_sampled->data[I] = buffer_u_sampled;
            v_sampled->data[I] = buffer_v_sampled;
            
            wavelets_fringes_center = vndrs_detect_center_fringes(
                                    Batch_on_precut->fullscans[I], /* input, output for QL */
                                    Apod_blockout,
                                    Apod_transition,
                                    N_frames_cut,
                                    17.0,  /* usually 17.0 fringes, good for MONA but maybe not for IONIC */
                                    fringes_length,
                                    ionic_in,
                                    Wavenum_limits_QL->data[0],
                                    Wavenum_limits_QL->data[1],
                                    Wavenum_min_hi_band_fft,
                                    Wavenum_max_hi_band_fft,
                                    Sample_interval,
                                    Wavenum_interval_precut,
                                    /* outputs */
                                    Batch_on_cut->fullscans[I],
                                    /* flag: 1 if not centered properly */ 
                                    &buffer_fringes_off_center);
            
            fringes_off_center->data[I] = buffer_fringes_off_center;
    
            if ( verbose_on == 1)
            {
                vndrs_fprintf_vector_double(Batch_on_cut->fullscans[I]->I1, "I1_On_raw_cut.dat");
                vndrs_fprintf_vector_double(Batch_on_cut->fullscans[I]->I2, "I2_On_raw_cut.dat");
                vndrs_fprintf_vector_double(Batch_on_cut->fullscans[I]->PA, "PA_On_raw_cut.dat");
                vndrs_fprintf_vector_double(Batch_on_cut->fullscans[I]->PB, "PB_On_raw_cut.dat");
            }
                        
            /* Additional parameters saved into output batch structure */
            Batch_on_cut->scan_data[I].time = Batch_on_precut->scan_data[I].time;
            Batch_on_cut->scan_data[I].dl_offset = 1000000.0 * (wavelets_fringes_center) * (Fringe_speed / Frame_rate);
    
            /* pkervell 13-mar-2002 Process only the scans that have been detected
                if the user has not selected the "-process_undetected" flag */
    
            /* Rejection of the scan following a strong jump in OPD (user specified limit) */
            jump = 0.0;
            if(I>0) jump = Batch_on_cut->scan_data[I].dl_offset - Batch_on_cut->scan_data[I-1].dl_offset;
            
            if(jump > max_jump && I>0)
            {
                Flag_processed->data[I] = 0;
                fringes_found_wavelets->data[I] = -7;     /* Code -7 = big jump */
            }
            else if(fringes_off_center->data[I] == 1.0)   /* Fringes have not been centered properly (too much off) */
            {
                Flag_processed->data[I] = 0;
                fringes_found_wavelets->data[I] = -8;     /* Code -8 = fringes not centered */
            }
            else if(fringes_off_center->data[I] == 0.0)   /* Fringes centered ok */
            {
                buffer_flag = 0.0;
                /* Now we do the correction on all properly centered scans and check the
                processed fringes after - pkervell 25-Sep-02 */
                vndrs_full_correction(Batch_off_cut,
                                Master_spectrum_off,
                                Batch_on_cut->fullscans[I],
                                WA_mean,
                                WB_mean,
                                Background_PA, /* updated if negative fluxes are found */
                                Background_PB,
                                Average_Matrix,
                                Frame_rate,
                                Fringe_speed,
                                fringes_length,
                                Rejection_threshold,
                                Apod_blockout,
                                Apod_transition,
                                Mult_corr,
                                debias,
                                ionic_in,
                                verbose_on,
                                &signal_buffer,
                                &unbalance_buffer,
                                Batch_on_processed->fullscans[I], /* subtracted interferogram in I1, added in I2 */
                                Batch_off_processed->fullscans[I],
                                &buffer_flag);
    
                Flag_processed->data[I] = buffer_flag;        
                Unbalance_coef->data[I] = unbalance_buffer;
                signal_p->data[I] = signal_buffer;   /* this variable holds the geom average of PA and PB (sqrt(PA.PB)) SNR */
                signal_buffer = 0.0;
            }
            
             
            /* Added pkervell - 15-Oct-03 : ASCII output of the processed scans */
            if (verbose_on == 2)
            {
                ascii_values1 = fopen("I_calib.txt","a");
                for(J=0;J<N_frames_cut;J++)
                    fprintf(ascii_values1, "%f ", Batch_on_processed->fullscans[I]->I1->data[J]);
                fprintf(ascii_values1, "\n");
                fclose(ascii_values1);
            }

            if(Flag_processed->data[I] > 0)
            {    
                /* Wavelets */
                vndrs_wavelet(N_frames_cut, 17.0,
                            (long)(Wavenum_min/Wavenum_interval),
                            (long)(Wavenum_max/Wavenum_interval),
                            coi_I1,
                            Batch_on_processed->fullscans[I]->I1,
                            t_wavelets_I1);
                    
                /* Soft filtering */
                /* the values the complex wavelets transform are set to zero when they are below the threshold
                 seuil=abs(median(abs(wave(N/2,)))/0.6745 * sqrt(2*log(N))).
                This is mathematically correct - dsegransan 16-Nov-02 - implemented by pkervell */

                /* for(J=0;J<N_frames_cut;J++)
                    high_freq_soft->data[J] = vndrs_abs_complex(get_element_matrix_complex(t_wavelets_I1,J,N_frames_cut/2));
                threshold_soft = fabs(vndrs_median_scan(high_freq_soft)/0.6745 * sqrt(2*log(N_frames_cut)));
                complex_null = complex_of_modulus_phase(0,0);

                for(J=0;J<N_frames_cut;J++)
                    for(K=0;K<(long)(N_frames_cut/2);K++)
                    {
                        if(vndrs_abs_complex(get_element_matrix_complex(t_wavelets_I1,J,K)) <= threshold_soft)
                            set_element_matrix_complex(t_wavelets_I1, J, K, complex_null);
                    }  removed pkervell 17-Nov-02 - gives very bad results as the background is not removed properly
                       from the fringe peak */
                    
                /*N x N/2 array*/
                norm_t_wavelets_I1 = vndrs_wavelet_power_spectrum(t_wavelets_I1);
                /* Memory allocated inside functions - pkervell 04-Aug-02 */
            
                if (verbose_on == 1)
                    vndrs_fprintf_matrix_double(norm_t_wavelets_I1, "T_wavelets_subtracted_I.dat");
        
                /* "On the fly" verification of the processed scans - pkervell 25-Sep-02 */
                /* in the following, we use the "flag_processed" variable to check which
                scans have actually been processed and should be checked. This variable is updated
                after the check to reject the incorrect scans at the time of the bootstrap. */
                        
                fringes_centering_outputs = vndrs_calloc_vector_double(10); /* only 8 used now */
    
                /* Duplication of the matrix of the wl transform as it will be modified in check_fringes */
                vndrs_matrix_memcpy(norm_t_wavelets_I1_check, norm_t_wavelets_I1);
                
                /* From whatever reason, from time to time the Flag_processed value is 1 and the wavelets
                power spectrum contains only zeros ! */
                if(vndrs_matrix_max(norm_t_wavelets_I1) != vndrs_matrix_min(norm_t_wavelets_I1))
                        fringes_found_wavelets->data[I] = vndrs_check_fringes(norm_t_wavelets_I1_check,
                                                                    N_frames_cut,
                                                                    Frame_rate,
                                                                    Fringe_speed,
                                                                    fringes_length,
                                                                    Wavenum_theo,
                                                                    Wavenum_limits_wl->data[0],
                                                                    Wavenum_limits_wl->data[1],
                                                                    Wavenum_min_hi_band_wl,
                                                                    Wavenum_max_hi_band_wl,
                                                                    Wavenum_interval,                                
                                                                    fringes_centering_outputs);
                else Flag_processed->data[I] = -2; /* case when the noise is too large */
            
                in_to_outband_ratio->data[I] = fringes_centering_outputs->data[0];
                wavelets_fringes_center = fringes_centering_outputs->data[1];
                residus_QL->data[I] = fringes_centering_outputs->data[2];
                peak_QL->data[I] = fringes_centering_outputs->data[3];
                fwhm_peak_time->data[I] = fringes_centering_outputs->data[4] * 1000000.0 * (Fringe_speed / Frame_rate);
                peak_position_freq->data[I] = fringes_centering_outputs->data[5]; /* position in frequency domain (pixels) */
                fwhm_peak_freq->data[I] = fringes_centering_outputs->data[6]; /* fwhm in frequency domain (Hz) */
                background_fit_QL->data[I] = fringes_centering_outputs->data[7]; /* background of the 2d fit */
    
                /* update of the Flag_processed variable */
                if((fringes_found_wavelets->data[I] != 1) & (select_scans == 1)) Flag_processed->data[I] = -1;
                if(select_scans == 0) Flag_processed->data[I] = 1;
                /* if the selection is de-activated, no update of the flag - pkervell 21-Jan-04 */
                
                vndrs_free_vector_double(fringes_centering_outputs);
                
                if(Flag_processed->data[I] > 0)
                {
                
                    /* NEW ! Correction of the first order piston effect : the fringe peak is displaced
                    by atmospheric piston during the batch. The relationship between the fringe peak
                    position in frequency and the visibility is a very nice line (to be adressed
                    in paper to be written...). Actually, it is not a line, but a dependence in square
                    root: it is easily seen by analysing the dispersion of the visibilities when applying
                    a variable exponent on the correction. The square root shows clearly the minimum dispersion.
                    We therefore correct directly this effect here using the result
                    of the fringe peak centroiding by wavelets computed above.
                    There is also a detectable dependence of the visibility with the power 0.20 of the fringe
                    peak (1/5 precisely), but it is not incorporated here, as the errors on the peak positions
                    is pretty large with the sampling we have on the fringes (5 points/fringe). The benefit
                    is in any case almost nothing, with less than 5% decrease of the error bar sigma. We can increase
                    the acceptance range in vndrs_check_fringes to 40%.
                    
                    The dependence of the visibility as a function of photometry is almost inexistent, with
                    only a power 0.063 very weak dependence, probably coming from the kappa errors: encouraging !
                    Tests on the Proxima/UTs calibrator.

                    pkervell 04-Oct-02
                    
                    removed functionnality waiting for further tests - pkervell */
    
                    /* Added off-source scan processing for wavelets  - pkervell 11-Oct-02 */
                    if(debias == 1)
                    {
                        vndrs_wavelet(N_frames_cut, 17.0,
                                (long)(Wavenum_min/Wavenum_interval), /* We process the whole freq range to recenter scans afterwards */
                                (long)(Wavenum_max/Wavenum_interval), /* pkervell 19-Oct-02 */
                                coi_I1,
                                Batch_off_processed->fullscans[I]->I1,
                                t_wavelets_I1_off);
                        /* memory allocated inside power_spectrum */
                        norm_t_wavelets_I1_off = vndrs_wavelet_power_spectrum(t_wavelets_I1_off);
                        vndrs_matrix_sub(norm_t_wavelets_I1, norm_t_wavelets_I1_off);
                        /* noise removed here instead of at power integration to compute properly the average wl spectrum */
                    }
                
                    Mu1_squared_wl->data[I] =
                            vndrs_integrate_wavelets_power(norm_t_wavelets_I1,
                                                        Wavenum_interval,
                                                        Wavenum_limits_wl->data[0],
                                                        Wavenum_limits_wl->data[1],
                                                        fringes_length,
                                                        verbose_on)
                                                        * Normalization_WL;
                                                        /* * sqrt(peak_position_freq->data[I]/theoretical_peak_pos);*/

                    /* Computation of the background of each wavelet spectrum - pkervell - 05-Jan-04 */
                    
                    Mu2_squared_wl->data[I] =
                            vndrs_integrate_wavelets_power(norm_t_wavelets_I1,
                                                        Wavenum_interval,
                                                        Wavenum_min_hi_band_wl,
                                                        Wavenum_max_hi_band_wl,
                                                        fringes_length,
                                                        verbose_on)
                                                        * Normalization_WL
                                                        /* * sqrt(peak_position_freq->data[I]/theoretical_peak_pos)*/
                                                        / surface_out_wl;

                    /* Correction of the background for each scan - pkervell - 05-Jan-04 */
                    Mu1_squared_wl->data[I] -= Mu2_squared_wl->data[I];

                    /* Correction of the wavelet spectrum for the measured background - pkervell - 05-Jan-04 */
                    vndrs_matrix_add_constant(norm_t_wavelets_I1,
                    -0.5*Mu2_squared_wl->data[I]/((Wavenum_max_hi_band_wl-Wavenum_min_hi_band_wl)/Wavenum_interval)
                    / fringes_length / Normalization_WL);
                    /* / (sqrt(peak_position_freq->data[I]/theoretical_peak_pos)));*/
                    
                    /* Computation of the maximum value of the fringe peak in the wavelet PSD - pkervell - 21-Jan-04 */
                    fringe_peak_strehl->data[I] = vndrs_matrix_local_max(norm_t_wavelets_I1,
                        (norm_t_wavelets_I1->dimension_1)/2 - fringes_length/2,
                        (norm_t_wavelets_I1->dimension_1)/2 + fringes_length/2,
                        Wavenum_limits_wl->data[0]/Wavenum_interval,
                        Wavenum_limits_wl->data[1]/Wavenum_interval)/Mu1_squared_wl->data[I];
                     
                    /* Apodisation : pkervell 04-Oct-02 */
                    /* Center apodisation filter - pkervell 04-Oct-02 : it is intended to cancel the residual noise around the
                    fringes by softening the rest of the scan, while still allowing the wavelets transform to work. Memory
                    allocated inside. We do it for the FOURIER MEASUREMENT only. For the WAVELETS, this is not a good idea
                    at all, as it tends to add artificially some visibility to the fringes: as the background is not uniform
                    any more, it diminishes the background, that is not removed correctly at the end of the processing.
                    We are therefore not using exactly the same signal for both processings. */
        
                    vndrs_vector_mul(Batch_on_processed->fullscans[I]->I1, Apodisation_fringes);
                    vndrs_vector_mul(Batch_off_processed->fullscans[I]->I1, Apodisation_fringes);
            
                    Mu1_squared->data[I]
                        = vndrs_integrate_fringe_power(Batch_on_processed->fullscans[I]->I1,
                                                Batch_off_processed->fullscans[I]->I1,
                                                Wavenum_interval,
                                                Wavenum_limits_fft->data[0],
                                                Wavenum_limits_fft->data[1],
                                                verbose_on,
                                                Pow_spec_avg_buffer_I1,
                                                Real_part_fft_buffer,
                                                Imaginary_part_fft_buffer,
                                                Phase_I1,
                                                Pow_spec_avg_buffer_I1_off)
                                                * Normalization_Fourier;
                                                /* * sqrt(peak_position_freq->data[I]/theoretical_peak_pos);*/
                    
                    /* Weighting of the visibility values */
                    /* ---------------------------------- */
                    /* Added weighting by the unbalance coefficient, to try to recover the true visibility of the
                    extended objects observed with the UTs, in particular Eta Car. Reference: O. Guyon, A&A 387, 366 (2002).
                    This may also help for the Sids observations, to filter the "semi-zero" crossings. - pkervell 17-Oct-02
                    */
                    
                    if(signal_p->data[I] > 0.0)
                        weights_visibilities->data[I] = signal_p->data[I] * Unbalance_coef->data[I];
                    else
                        weights_visibilities->data[I] = 0.0;
                    
                    /* Normalization of power and amplitude spectra */
                    /* -------------------------------------------- */
                    
                    vndrs_vector_scale(Real_part_fft_buffer, 2.0 / (sqrt(Shape_factor) * Wavenum_interval));
                    vndrs_vector_scale(Imaginary_part_fft_buffer, 2.0 / (sqrt(Shape_factor) * Wavenum_interval));
                    vndrs_vector_scale(Pow_spec_avg_buffer_I1, Normalization_Fourier);
                    vndrs_matrix_scale(norm_t_wavelets_I1, Normalization_WL);
                    if(debias == 1)
                    {
                        vndrs_vector_scale(Pow_spec_avg_buffer_I1_off, Normalization_Fourier);
                        vndrs_matrix_scale(norm_t_wavelets_I1_off, Normalization_WL);
                    }
                    
                    /* Computation of amplitude and phase */
                    /* ---------------------------------- */
                    
                    Real_fft->data[I] = (index_freq_high_QL_cut - index_freq_low_QL_cut)*
                    vndrs_mean_scan_limits(Real_part_fft_buffer, index_freq_low_QL_cut, index_freq_high_QL_cut);
                    
                    Imaginary_fft->data[I] = (index_freq_high_QL_cut - index_freq_low_QL_cut)*
                    vndrs_mean_scan_limits(Imaginary_part_fft_buffer, index_freq_low_QL_cut, index_freq_high_QL_cut);
                    
                    /* For the moment, the amplitude is not correct due to a problem of normalization */
                    Amplitude_fft->data[I] = sqrt(Real_fft->data[I]*Real_fft->data[I]
                        + Imaginary_fft->data[I]*Imaginary_fft->data[I]);
                        
                    Phase_fft->data[I] = 180.0*atan2(Imaginary_fft->data[I], Real_fft->data[I])/Pi;

                    /* Update of the total OPD of the scan */
                    /* ----------------------------------- */
                    OPD_phase->data[I] = Phase_fft->data[I]/Wavenum_theo/360.0;
                    Total_OPD->data[I] = Batch_on_cut->scan_data[I].dl_offset - 1e6*OPD_phase->data[I];
                    /* in microns - sign corrected - pkervell - 19-Nov-03 */

                    /* Computation of the vector spectrum from the wl power spectrum array */
                    /* ------------------------------------------------------------------- */
                    index_time_min = (long)(N_frames_cut/2 - (fringes_length/2));
                    index_time_max = (long)(N_frames_cut/2 + (fringes_length/2));
                    for(J=0; J<(N_frames_cut/2); J++)
                    {
                        WL_spec_avg_I1_vector->data[J] = 0.0;
                        WL_spec_avg_I1_vector_off->data[J] = 0.0;
                    }
                    for(K=0; K<(N_frames_cut/2); K++)
                        for(J=index_time_min; J<index_time_max; J++)
                        {
                            WL_spec_avg_I1_vector->data[K] += get_element_matrix_double(norm_t_wavelets_I1,J,K);
                            if(debias == 1)
                                WL_spec_avg_I1_vector_off->data[K] += get_element_matrix_double(norm_t_wavelets_I1_off,J,K);
                        }

                    /* Correction of the effect of peak position on PSD */
                    /* ------------------------------------------------ */
                    /* added pkervell 08-Nov-02 */
                    /* Removed pkervell - 15-Nov-02 incorrect when visibility is low, but may add stability TBC */
					/* Reinstalled pkervell - 16-Dec-06 for further tests */
					/* Important: this changes the absolute value of the mu2, as the target peak freauency is not */
					/* known with an excellent accuracy, but this is compensated at the calibration level by the */
					/* division by the calibrator mu2. */
					/* A) Version with square root - 16-Dec-06 */
                    if(depiston == 1)
                    {
						vndrs_vector_scale(Pow_spec_avg_buffer_I1, sqrt(peak_position_freq->data[I]/theoretical_peak_pos));
						vndrs_vector_scale(WL_spec_avg_I1_vector, sqrt(peak_position_freq->data[I]/theoretical_peak_pos));
						if(debias == 1)
						{
							vndrs_vector_scale(Pow_spec_avg_buffer_I1_off, sqrt(peak_position_freq->data[I]/theoretical_peak_pos));
							vndrs_vector_scale(WL_spec_avg_I1_vector_off, sqrt(peak_position_freq->data[I]/theoretical_peak_pos));
						}
					}
					/* B) Version with linear dependance - 16-Dec-06 */
                    /* if(depiston == 1)
                    {
						vndrs_vector_scale(Pow_spec_avg_buffer_I1, (peak_position_freq->data[I]/theoretical_peak_pos));
						vndrs_vector_scale(WL_spec_avg_I1_vector, (peak_position_freq->data[I]/theoretical_peak_pos));
						if(debias == 1)
						{
							vndrs_vector_scale(Pow_spec_avg_buffer_I1_off, (peak_position_freq->data[I]/theoretical_peak_pos));
							vndrs_vector_scale(WL_spec_avg_I1_vector_off, (peak_position_freq->data[I]/theoretical_peak_pos));
						}
					} */
                    
                    /* Filling of the matrix of the PSDs of all processed scans of the batch */
                    /* --------------------------------------------------------------------- */
                    for(J=0;J<N_frames_cut/2;J++)
                    {
                        set_element_matrix_double(Real_part_fft, I, J, Real_part_fft_buffer->data[J]);
                        set_element_matrix_double(Imaginary_part_fft, I, J, Imaginary_part_fft_buffer->data[J]);
                        set_element_matrix_double(PSDs_matrix, I, J, Pow_spec_avg_buffer_I1->data[J]);
                        set_element_matrix_double(Phase_matrix, I, J, Phase_I1->data[J]);
                        set_element_matrix_double(PSDs_matrix_wl, I, J, WL_spec_avg_I1_vector->data[J]);
                        set_element_matrix_double(PSDs_matrix_off, I, J, Pow_spec_avg_buffer_I1_off->data[J]);
                        set_element_matrix_double(PSDs_matrix_wl_off, I, J, WL_spec_avg_I1_vector_off->data[J]);
                    }
                                    
                    if(debias == 1)
                        vndrs_free_matrix_double(norm_t_wavelets_I1_off);
                    
                    /* Update of the average wavelets PSD - pkervell 03-Feb-03 */
                    /* ------------------------------------------------------- */
                    vndrs_matrix_scale(norm_t_wavelets_I1, weights_visibilities->data[I]);
                    vndrs_matrix_add(norm_t_wavelets_avg, norm_t_wavelets_I1);

                    
                } /* end of SECOND if Flag_processed->data[I] > 0 */
        
                    
                /* The memory for the wavelets power spectra was allocated in the loop */
                vndrs_free_matrix_double(norm_t_wavelets_I1);
        
            } /* end of FIRST if Flag_processed->data[I] > 0 */
    
            /* display of the mu^2 values */
            printf(" %4ld | %+7.2f | %+7.2f | %+7.2f | %+7.2f | %6.2f |%8.2f |%8.2f | % 2d | % 2d |%8.2f | %4.2f |%8.2f | % 8.5f % 8.5f |\n",
                I,
                Batch_on_precut->scan_data[I].dl_offset, /* original OPD offset from raw file */
                Batch_on_cut->scan_data[I].dl_offset, /* computed OPD from wavelets */                          
                Phase_fft->data[I],
                Total_OPD->data[I],
                fwhm_peak_time->data[I],
                peak_position_freq->data[I]*Frame_rate/N_frames_cut,
                fwhm_peak_freq->data[I]*Frame_rate/N_frames_cut,
                /* fringe_peak_strehl->data[I],*/
                (int)Flag_processed->data[I],
                (int)fringes_found_wavelets->data[I],
                signal_p->data[I],
                Unbalance_coef->data[I],
                weights_visibilities->data[I],
                /* Mu1_squared->data[I],*/
                Mu1_squared_wl->data[I],
                Mu2_squared_wl->data[I]); /* background */
    
        }   /*  for (I=0; I<N_scans_on; I++)*/
        printf("-------------------------------------------------------------------------------------------------------------------------------------\n");
        /* ================================================================= */
        /* End loop; over all scans of the batch                             */
        /* ================================================================= */ 
    }
    else
    {
        printf("\n  *** Kappa coefficient(s) below 5%%. Batch ignored.\n");
        sprintf(vndrs_protocol.text,"# *** Kappa coefficient(s) below 5 percents. Batch ignored.\n");
        vndrs_runtime_protocol();
    }

    /* Added pkervell - 15-Oct-03 : ASCII output of the raw and processed scans */
    if(verbose_on == 2)
    {
        ascii_values1 = fopen("I1_raw.txt","w");
        ascii_values2 = fopen("I2_raw.txt","w");
        ascii_values3 = fopen("PA_raw.txt","w");
        ascii_values4 = fopen("PB_raw.txt","w");
        for(I=0;I<N_scans_on;I++)
        {
            for(J=0;J<N_frames_cut;J++)
            {
                fprintf(ascii_values1, "%f ", Batch_on_cut->fullscans[I]->I1->data[J]);
                fprintf(ascii_values2, "%f ", Batch_on_cut->fullscans[I]->I2->data[J]);
                fprintf(ascii_values3, "%f ", Batch_on_cut->fullscans[I]->PA->data[J]);
                fprintf(ascii_values4, "%f ", Batch_on_cut->fullscans[I]->PB->data[J]);
            }
            fprintf(ascii_values1, "\n");
            fprintf(ascii_values2, "\n");
            fprintf(ascii_values3, "\n");
            fprintf(ascii_values4, "\n");
        }
        fclose(ascii_values1);
        fclose(ascii_values2);
        fclose(ascii_values3);
        fclose(ascii_values4);
    }
    
    
    vndrs_free_vector_double(coi_I1);
    vndrs_free_matrix_complex(t_wavelets_I1);
    vndrs_free_matrix_complex(t_wavelets_I1_off);

    /* Computation of the number of reduced scans and number of scans rejected by category */
    *N_scans_processed = 0;
    for(I=0;I<N_scans_on;I++)
    {
        if(Flag_processed->data[I] > 0) *N_scans_processed = *N_scans_processed + 1;
        
        if(fringes_found_wavelets->data[I] == -7) scans_rejected_7++; /* big jump */
        else if(fringes_found_wavelets->data[I] == -8) scans_rejected_8++; /* fringes could not be centered */
        else if(Flag_processed->data[I] == 0) scans_rejected_0++; 	      /* photometry too low */
        if(fringes_found_wavelets->data[I] == -1) scans_rejected_1++; /* not used - reserved */
        /* if(fringes_found_wavelets->data[I] == -2) scans_rejected_2++; not used */
        if(fringes_found_wavelets->data[I] == -3) scans_rejected_3++; /* fringe packet too wide or narrow in time space */
        if(fringes_found_wavelets->data[I] == -4) scans_rejected_4++; /* fringe peak not at correct freq position in wl transorm */
        if(fringes_found_wavelets->data[I] == -5) scans_rejected_5++; /* fringe peak too wide or narrow in wl transorm */
        if(fringes_found_wavelets->data[I] == -6) scans_rejected_6++; /* not used - reserved */
    }
    *numberOfScansRejected = N_scans_on - *N_scans_processed;
    
    /* Elimination of the visibility outliers - pkervell 28-Sep-02 */
    /* ----------------------------------------------------------- */
    /* We remove the scans that are very improbable.
      The effect is positive on the quality of the resulting fit and the skewness. */
    
    n_sig_reject = 3.0;
    if(*N_scans_processed > 10) /* We do not call the filtering for less than 10 scans */
        flag_remove = vndrs_eliminate_visibility_outliers(Mu1_squared, n_sig_reject , Flag_processed, removed_indices);
    
    /* Recomputation of the number of reduced scans */
    *N_scans_processed = 0;
    for(I=0;I<N_scans_on;I++)
        if(Flag_processed->data[I] > 0)  *N_scans_processed = *N_scans_processed + 1;
    *numberOfScansRejected = N_scans_on - *N_scans_processed;

    /* ----- */
    
    if(*N_scans_processed > 0)
    {
       /* Computation of the NON-RECENTERED average power spectrum and phase */
       /* ------------------------------------------------------------------ */
    
        for(I=0;I<N_frames_cut/2;I++)
        {
            Pow_spec_avg_I1_no_centering->data[I] = 0.0;
            WL_spec_avg_I1_vector_no_centering->data[I] = 0.0;
            Phase_avg_I1_no_centering->data[I] = 0.0;
            Real_part_fft_no_centering->data[I] = 0.0;
            Imaginary_part_fft_no_centering->data[I] = 0.0;
        }
        
        total_weighting_PSD = 0.0;
        
        for(I=0;I<N_scans_on;I++)
            if(Flag_processed->data[I] > 0)
            {
                for(J=0;J<N_frames_cut/2;J++)
                {
                    Pow_spec_avg_I1_no_centering->data[J] += get_element_matrix_double(PSDs_matrix, I, J) *
                        weights_visibilities->data[I];
                    Phase_avg_I1_no_centering->data[J] += get_element_matrix_double(Phase_matrix, I, J) *
                        weights_visibilities->data[I];
                    WL_spec_avg_I1_vector_no_centering->data[J] += get_element_matrix_double(PSDs_matrix_wl, I, J) * 
                        weights_visibilities->data[I];
                    Real_part_fft_no_centering->data[J] += get_element_matrix_double(Real_part_fft, I, J) * 
                        weights_visibilities->data[I];
                    Imaginary_part_fft_no_centering->data[J] += get_element_matrix_double(Imaginary_part_fft, I, J) * 
                        weights_visibilities->data[I];
                }
                total_weighting_PSD += weights_visibilities->data[I];
            }

        /* added normalization of the average power spectra - pkervell - 09-Aug-02 */
        vndrs_vector_scale(Pow_spec_avg_I1_no_centering, 1./total_weighting_PSD);
        vndrs_vector_scale(WL_spec_avg_I1_vector_no_centering, 1./total_weighting_PSD);
        vndrs_vector_scale(Phase_avg_I1_no_centering, 1./total_weighting_PSD);
        vndrs_vector_scale(Real_part_fft_no_centering, 1./total_weighting_PSD);
        vndrs_vector_scale(Imaginary_part_fft_no_centering, 1./total_weighting_PSD);
        
        /* Normalization of the average wavelets transform - pkervell - 03-Feb-03 */
        vndrs_matrix_scale(norm_t_wavelets_avg, 1./total_weighting_PSD);
        
        /*
        vndrs_fprintf_vector_double(Real_part_fft_no_centering, "Real_part_fft_nc.dat");
        vndrs_fprintf_vector_double(Imaginary_part_fft_no_centering, "Imaginary_part_fft_nc.dat");
        */
                
        /* ---- */
        
        /* POWER SPECTRA RECENTERING */
        /* ------------------------- */
        /* NEW ! Recentering of the power spectra according to the position computed by the wavelets
        transformation. This is important to reduce the energy spread in the final computation of the
        mu^2 from the average power spectrum. PSDs are replaced in the matrices. pkervell 15-Oct-02
        The effect is clearly visible on the processed PSD.
        The average position of the fringe peak is brought to the theoretical wavenumber.*/
        
        vndrs_center_PSD_peak(Phase_matrix, Flag_processed, Wavenum_min_lowband_fft/Wavenum_interval,
                            theoretical_peak_pos, peak_position_freq);
        vndrs_center_PSD_peak(Phase_matrix_wl, Flag_processed, Wavenum_min_lowband_fft/Wavenum_interval,
                            theoretical_peak_pos, peak_position_freq);
        vndrs_center_PSD_peak(PSDs_matrix, Flag_processed, Wavenum_min_lowband_fft/Wavenum_interval,
                            theoretical_peak_pos, peak_position_freq);
        vndrs_center_PSD_peak(PSDs_matrix_wl, Flag_processed, Wavenum_min_lowband_wl/Wavenum_interval,
                            theoretical_peak_pos, peak_position_freq);

        /* Computation of the average RECENTERED power spectrum */
        /* ---------------------------------------------------- */
    
        for(I=0;I<N_frames_cut/2;I++)
        {
            Pow_spec_avg_I1->data[I] = 0.0;
            WL_spec_avg_I1_vector->data[I] = 0.0;
            WL_spec_avg_I1_vector_off->data[I] = 0.0;
            Phase_I1->data[I] = 0.0;
        }
        
        total_weighting_PSD = 0.0;
        
        for(I=0;I<N_scans_on;I++)
            if(Flag_processed->data[I] > 0)
            {
                for(J=0;J<N_frames_cut/2;J++)
                {
                    Pow_spec_avg_I1->data[J] += get_element_matrix_double(PSDs_matrix, I, J) *
                    weights_visibilities->data[I];
                    Pow_spec_avg_I1_off->data[J] += get_element_matrix_double(PSDs_matrix_off, I, J) *
                    weights_visibilities->data[I];
                    WL_spec_avg_I1_vector->data[J] += get_element_matrix_double(PSDs_matrix_wl, I, J) * 
                    weights_visibilities->data[I];
                    WL_spec_avg_I1_vector_off->data[J] += get_element_matrix_double(PSDs_matrix_wl_off, I, J) * 
                    weights_visibilities->data[I];
                    /* We re-use the following variables */
                    Phase_I1->data[J] += get_element_matrix_double(Phase_matrix, I, J) *
                        weights_visibilities->data[I];
                }
                total_weighting_PSD += weights_visibilities->data[I];
            }
            
        /* added normalization of the average power spectra - pkervell - 09-Aug-02 */
        vndrs_vector_scale(Pow_spec_avg_I1, 1./total_weighting_PSD);
        vndrs_vector_scale(WL_spec_avg_I1_vector, 1./total_weighting_PSD);
        vndrs_vector_scale(Pow_spec_avg_I1_off, 1./total_weighting_PSD);
        vndrs_vector_scale(WL_spec_avg_I1_vector_off, 1./total_weighting_PSD);
        vndrs_vector_scale(Phase_I1, 1./total_weighting_PSD);
  
        *time_batch = 0.0; *baseline_batch = 0.0; *azimuth_batch = 0.0; *u_batch = 0.0; *v_batch = 0.0;

        for(I=0;I<N_scans_on;I++)
            if(Flag_processed->data[I] > 0)
            {
                /* average time of the batch (only processed and selected scans) */
                *time_batch += time_scan->data[I]*weights_visibilities->data[I]/total_weighting_PSD;
                /* average baseline length of the batch (only processed scans) */
                *baseline_batch += projected_baseline->data[I]*weights_visibilities->data[I]/total_weighting_PSD;
                /* average azimuth of the batch (only processed scans) */ 
                *azimuth_batch += azimuth_baseline->data[I]*weights_visibilities->data[I]/total_weighting_PSD; 
                *u_batch += u_sampled->data[I]*weights_visibilities->data[I]/total_weighting_PSD;
                *v_batch += v_sampled->data[I]*weights_visibilities->data[I]/total_weighting_PSD;
                /* average altitude and azimuth of object */
                altitude_batch_obj += altitude_object->data[I]*weights_visibilities->data[I]/total_weighting_PSD;
                azimuth_batch_obj += azimuth_object->data[I]*weights_visibilities->data[I]/total_weighting_PSD;
            }

        *altitude_src = altitude_batch_obj; /* added - pkervell 15-may-03 */
        *azimuth_src = azimuth_batch_obj;

        /* Computation of the Fourier combined spectrum residual background */
        /* ---------------------------------------------------------------- */
        /* added pkervell 09-Sep-02 - we use a linear model */
        
        vndrs_remove_psd_background(N_frames_cut, Pow_spec_avg_I1,
                                    Wavenum_min_lowband_fft,
                                    Wavenum_max_lowband_fft,
                                    Wavenum_min_hi_band_fft,
                                    Wavenum_max_hi_band_fft,
                                    debias,
                                    5e-8,  /* limit acceptable slope (ADU^2/Hz^2) */
                                    /* Changed from 2e-7 to 5e-8 on 21-Jan-03 - pkervell */
                                    Frame_rate,
                                    Fringe_speed,
                                    &slope_background_fft,
                                    Background_PSD,
                                    Pow_spec_avg_I1_corrected);
        
        *background_fourier = vndrs_area(Background_PSD, Wavenum_interval,
                                            Wavenum_limits_fft->data[0], Wavenum_limits_fft->data[1])/Wavenum_interval;
        
        
        /* Computation of the Wavelets combined spectrum background */
        /* -------------------------------------------------------- */
        /* added pkervell 13-Sep-02 - we use a linear model
            now removes only the second order residual of the correction - pkervell - 25-Sep-02 */
        
        vndrs_remove_psd_background(N_frames_cut,
                                    WL_spec_avg_I1_vector,
                                    Wavenum_min_lowband_wl,
                                    Wavenum_max_lowband_wl,
                                    Wavenum_min_hi_band_wl,
                                    Wavenum_max_hi_band_wl,
                                    debias,
                                    5e-8,  /* limit acceptable slope (ADU^2/Hz^2) */
                                    /* Changed from 2e-7 to 5e-8 on 21-Jan-03 - pkervell */
                                    Frame_rate,
                                    Fringe_speed,
                                    &slope_background_wl,
                                    Background_PSD_wl_vector,
                                    WL_spec_avg_I1_vector_corrected);
        
        
        
        /* Background computation */
        /* ---------------------- */
        
        *background_wavelets = vndrs_area(Background_PSD_wl_vector, Wavenum_interval,
                                            Wavenum_limits_wl->data[0], Wavenum_limits_wl->data[1])/Wavenum_interval;
        
        /* Correction of the background of the non-recentered PSDs */
        /* ------------------------------------------------------- */
        /* The background is supposed identical to recentered PSDs.*/
        
        vndrs_vector_memcpy(Pow_spec_avg_I1_no_centering_corrected, Pow_spec_avg_I1_no_centering);
        vndrs_vector_sub(Pow_spec_avg_I1_no_centering_corrected, Background_PSD);
        vndrs_vector_memcpy(WL_spec_avg_I1_vector_no_centering_corrected, WL_spec_avg_I1_vector_no_centering);
        vndrs_vector_sub(WL_spec_avg_I1_vector_no_centering_corrected, Background_PSD_wl_vector);
       
        /* Computation of the effective wavelength of non-centered PSDs */
        /* ------------------------------------------------------------ */

        weighted_peak_pos = 0.0; weight_psd = 0.0; running_max = 0.0;
        weighted_peak_pos_wl = 0.0; weight_psd_wl = 0.0;
        for(I=index_freq_low_wl_cut;I<index_freq_high_wl_cut;I++)
        {
            weighted_peak_pos_wl += I * WL_spec_avg_I1_vector_no_centering_corrected->data[I];
            weight_psd_wl += WL_spec_avg_I1_vector_no_centering_corrected->data[I];
        }

        for(I=index_freq_low_fft_cut;I<index_freq_high_fft_cut;I++)
        {
            weighted_peak_pos += I * Pow_spec_avg_I1_no_centering_corrected->data[I];
            weight_psd += Pow_spec_avg_I1_no_centering_corrected->data[I];
        }
        
        if(weight_psd != 0.0 && weight_psd_wl != 0.0)
        {
            weighted_peak_pos = weighted_peak_pos/(theoretical_peak_pos * weight_psd);
            weighted_peak_pos_wl = weighted_peak_pos_wl/(theoretical_peak_pos * weight_psd_wl);
        }
        else
        {
            weighted_peak_pos = 1.0;
            weighted_peak_pos_wl = 1.0;
        }

        /* Computation of the effective wavelength of RECENTERED PSDs */
        /* ---------------------------------------------------------- */
        /* Just to check that it is really 1.0... */
        
        weighted_peak_pos_fft_centered = 0.0; weight_psd = 0.0; running_max = 0.0;
        weighted_peak_pos_wl_centered = 0.0; weight_psd_wl = 0.0;
        for(I=index_freq_low_wl_cut;I<index_freq_high_wl_cut;I++)
        {
            weighted_peak_pos_wl_centered += I * WL_spec_avg_I1_vector_corrected->data[I];
            weight_psd_wl += WL_spec_avg_I1_vector_corrected->data[I];
        }
        
        for(I=index_freq_low_fft_cut;I<index_freq_high_fft_cut;I++)
        {
            weighted_peak_pos_fft_centered += I * Pow_spec_avg_I1_corrected->data[I];
            weight_psd += Pow_spec_avg_I1_corrected->data[I];
        }
        
        if(weight_psd != 0.0 && weight_psd_wl != 0.0)
        {
            weighted_peak_pos_fft_centered = weighted_peak_pos_fft_centered/(theoretical_peak_pos * weight_psd);
            weighted_peak_pos_wl_centered = weighted_peak_pos_wl_centered/(theoretical_peak_pos * weight_psd_wl);
        }
        else
        {
            weighted_peak_pos_fft_centered = 1.0;
            weighted_peak_pos_wl_centered = 1.0;
        }
        
        /* ----------- */
        
        
        /* Actual correction of the individual scans visibility values for the background */
        /* ------------------------------------------------------------------------------ */
        /* This step is only for the second order BG residual now (v3.1) - pkervell - 04-Jan-04 */
        
        for(I=0;I<N_scans_on;I++)
            if(Flag_processed->data[I] > 0)
            {
                Mu1_squared->data[I] -= *background_fourier;
                Mu1_squared_wl->data[I] -= *background_wavelets;
            }
        
        
        /* Output of the final selected and corrected visibility values to the log file */
        /* ---------------------------------------------------------------------------- */
        
        for(I=0;I<N_scans_on;I++)
        {
            if(Flag_processed->data[I] != 0)
                {
                sprintf(vndrs_protocol.text,"% 4ld % 8.10f % 6.2f % 6.2f % 7.4f % 6.2f % 6.2f %+7.2f %+7.2f %+8.2f %+8.2f % 4.2f % 6.2f % 4.2f % 8.2f %+d %+d % 5.2f % 5.2f % 7.2f % 7.2f % 7.2f % 7.2f % 6.2f % 6.4f % 6.2f % 8.6f % 8.6f % 8.6f\n",
                    I,
                    time_scan->data[I],
                    altitude_object->data[I],
                    azimuth_object->data[I],
                    projected_baseline->data[I],
                    azimuth_baseline->data[I],
                    peak_QL->data[I],
                    Batch_on_precut->scan_data[I].dl_offset, /* original OPD offset from raw file */
                    Batch_on_cut->scan_data[I].dl_offset, /* computed OPD from wavelets */                          
                    Phase_fft->data[I],
                    Total_OPD->data[I],
                    fwhm_peak_time->data[I],
                    peak_position_freq->data[I]*Frame_rate/N_frames_cut,
                    fwhm_peak_freq->data[I]*Frame_rate/N_frames_cut,
                    fringe_peak_strehl->data[I],
                    (int)Flag_processed->data[I],
                    (int)fringes_found_wavelets->data[I],
                    Background_PA->data[I],
                    Background_PB->data[I],
                    Photometry_I1->data[I],
                    Photometry_I2->data[I],
                    Photometry_PA->data[I],
                    Photometry_PB->data[I],
                    signal_p->data[I],
                    Unbalance_coef->data[I],
                    weights_visibilities->data[I],
                    Mu1_squared->data[I],
                    Mu1_squared_wl->data[I],
                    Mu2_squared_wl->data[I]);
                vndrs_runtime_protocol();
                }
        }
        
        /* BOOTSTRAPPING OF THE RESULTING MU^2 */
        /* ----------------------------------- */
        /* MODIFICATION : THE VISIBILITY IS NOW INTEGRATED DIRECTLY FROM THE
           AVERAGE SPECTRUM - see below - pkervell 15-Oct-02 */
        /* The bootstrapping is used only to determine the error bars. */
        /* Memory is allocated inside function vndrs_extract_nonzero */
        /* Extraction of the vectors to be used for the bootstrap - pkervell - 09-Sep-02 */
        non_zero_weights = vndrs_extract_nonzero(weights_visibilities, Flag_processed);
        non_zero_mu1_wl = vndrs_extract_nonzero(Mu1_squared_wl, Flag_processed);
        non_zero_mu2_wl = vndrs_extract_nonzero(Mu2_squared_wl, Flag_processed);
        non_zero_mu1_fft = vndrs_extract_nonzero(Mu1_squared, Flag_processed);
        non_zero_signal_p = vndrs_extract_nonzero(signal_p, Flag_processed);
        non_zero_unbalance = vndrs_extract_nonzero(Unbalance_coef, Flag_processed);
        non_zero_peak_pos = vndrs_extract_nonzero(peak_position_freq, Flag_processed);
        non_zero_peak_width = vndrs_extract_nonzero(fwhm_peak_freq, Flag_processed);
        
        /* For user information */
        /* -------------------- */
        avg_photom = vndrs_mean_scan(non_zero_signal_p); /* we compute the average of the photom on the fringes over the batch */
        average_unbalance = vndrs_mean_scan(non_zero_unbalance);
        *total_weight_batch = vndrs_sum_scan(non_zero_weights);
        peak_pos_mean = vndrs_mean_scan(non_zero_peak_pos); /* mean of the power peak pos in the PSD (in pixels) */
        peak_pos_sigma = vndrs_stddev_scan(non_zero_peak_pos); /* stddev of the power peak jitter in the PSD (in pixels) */
        peak_width_avg = vndrs_mean_scan(non_zero_peak_width);
        peak_width_sigma = vndrs_stddev_scan(non_zero_peak_width);
        
        /* Computation of the histograms - 22-Oct-02 */
        /* ----------------------------------------- */
        histogram_scale_fft = vndrs_calloc_vector_double(11);
        histogram_fft = vndrs_calloc_vector_double(11);
        histogram_scale_wl = vndrs_calloc_vector_double(11);
        histogram_wl = vndrs_calloc_vector_double(11);
        vndrs_histogram(non_zero_mu1_wl, 3.0, histogram_scale_wl, histogram_wl);
        vndrs_histogram(non_zero_mu1_fft, 3.0, histogram_scale_fft, histogram_fft);
 
        /* Effective wavelength and errors returned to main program */
        /* -------------------------------------------------------- */
        /* pkervell 23-Oct-02 */
        /* Corrected bug to return the wavelength for wl, but it is not used afterwards
           (only Fourier is valid) - pkervell 13-Nov-02 */
        *effective_wavelength = 1000000.0 / (weighted_peak_pos * Wavenum_theo);
        *effective_wavelength_wl = 1000000.0 / (weighted_peak_pos_wl * Wavenum_theo);
        *effective_wavelength_error = fabs(1000000.0 / (weighted_peak_pos * Wavenum_theo -
                peak_pos_sigma * Wavenum_interval / sqrt(*N_scans_processed))
                    - 1000000.0 / (weighted_peak_pos * Wavenum_theo));

                           
        /* Computations in the case of the wavelets */
        /* ---------------------------------------- */
        vndrs_compute_bootstrap(  non_zero_mu1_wl,
                                    non_zero_weights,
                                    B,
                                    /* output */
                                    &mu1_wl_avg_bs, &mu1_wl_boot_avg, &mu1_wl_stdev_avg, &Bias_mu1_wl_avg, &Rms_mu1_wl_avg,
                                    &Ci_mu1_wl_avg_min, &Ci_mu1_wl_avg_max,
                                    &Ci2_mu1_wl_avg_min, &Ci2_mu1_wl_avg_max, &Ci3_mu1_wl_avg_min, &Ci3_mu1_wl_avg_max,
                                    &mu1_wl_med, &mu1_wl_boot_med, &mu1_wl_stdev_med, &Bias_mu1_wl_med, &Rms_mu1_wl_med, 
                                    &Ci_mu1_wl_med_min, &Ci_mu1_wl_med_max,
                                    &Ci2_mu1_wl_med_min, &Ci2_mu1_wl_med_max, &Ci3_mu1_wl_med_min, &Ci3_mu1_wl_med_max,
                                    &skew_mean_wl, &skew_median_wl, &kurtosis_mean_wl, &kurtosis_median_wl);
        
        avg_background_wavelets = vndrs_mean_scan(non_zero_mu2_wl);
        
        /* Computations and displays in the case of the Fourier processing */
        /* --------------------------------------------------------------- */
        
        vndrs_compute_bootstrap(  non_zero_mu1_fft,
                                    non_zero_weights,
                                    B,
                                    /* output */
                                    &mu1_fft_avg_bs, &mu1_fft_boot_avg, &mu1_fft_stdev_avg, &Bias_mu1_fft_avg, &Rms_mu1_fft_avg,
                                    &Ci_mu1_fft_avg_min, &Ci_mu1_fft_avg_max,
                                    &Ci2_mu1_fft_avg_min, &Ci2_mu1_fft_avg_max, &Ci3_mu1_fft_avg_min, &Ci3_mu1_fft_avg_max,
                                    &mu1_fft_med, &mu1_fft_boot_med, &mu1_fft_stdev_med, &Bias_mu1_fft_med, &Rms_mu1_fft_med, 
                                    &Ci_mu1_fft_med_min, &Ci_mu1_fft_med_max,
                                    &Ci2_mu1_fft_med_min, &Ci2_mu1_fft_med_max, &Ci3_mu1_fft_med_min, &Ci3_mu1_fft_med_max,
                                    &skew_mean_fft, &skew_median_fft, &kurtosis_mean_fft, &kurtosis_median_fft);
        
                                        
        /* Jackknife errors */
        /* ---------------- */
        if(*N_scans_processed > 2)
        {
            jack_fft = vndrs_compute_jacknife_error(non_zero_mu1_fft);
            jack_wl  = vndrs_compute_jacknife_error(non_zero_mu1_wl);
        }
        else
        {
            jack_fft = 0.0;
            jack_wl =0.0;
        }
        
        /* Simple error bars */
        /* ----------------- */
        simple_error_fft = vndrs_compute_simple_error(non_zero_mu1_fft);
        simple_error_wl = vndrs_compute_simple_error(non_zero_mu1_wl);
        
        /* FINAL ERROR BARS */
        /* ---------------- */
        /* Computation of the final error bars values taking into account the errors on the kappa coefficients
            pkervell - 27-sep-02 */
        mu1_fft_stdev_avg = sqrt(vndrs_variance_product(mu1_fft_avg, mu1_fft_stdev_avg*mu1_fft_stdev_avg, 1.0, relative_kappa_norm_var));;
        mu1_wl_stdev_avg = sqrt(vndrs_variance_product(mu1_wl_avg, mu1_wl_stdev_avg*mu1_wl_stdev_avg, 1.0, relative_kappa_norm_var));;
        mu1_fft_stdev_med = sqrt(vndrs_variance_product(mu1_fft_med, mu1_fft_stdev_med*mu1_fft_stdev_med, 1.0, relative_kappa_norm_var));;
        mu1_wl_stdev_med = sqrt(vndrs_variance_product(mu1_wl_med, mu1_wl_stdev_med*mu1_wl_stdev_med, 1.0, relative_kappa_norm_var));;
        
        Ci_mu1_wl_avg_min = sqrt(vndrs_variance_product(mu1_wl_avg, Ci_mu1_wl_avg_min*Ci_mu1_wl_avg_min, 1.0, relative_kappa_norm_var));
        Ci_mu1_wl_avg_max = sqrt(vndrs_variance_product(mu1_wl_avg, Ci_mu1_wl_avg_max*Ci_mu1_wl_avg_max, 1.0, relative_kappa_norm_var));
        Ci_mu1_fft_avg_min = sqrt(vndrs_variance_product(mu1_fft_avg, Ci_mu1_fft_avg_min*Ci_mu1_fft_avg_min, 1.0, relative_kappa_norm_var));
        Ci_mu1_fft_avg_max = sqrt(vndrs_variance_product(mu1_fft_avg, Ci_mu1_fft_avg_max*Ci_mu1_fft_avg_max, 1.0, relative_kappa_norm_var));
        
        
        /* Verification of the frequency limits up to which there is power in the spectrum */
        /* ------------------------------------------------------------------------------- */
        /* the visibility outliers are not removed from the spectrum at the moment */
        /* pkervell 15-Oct-02 */
        
        vndrs_estimate_peak_limits(Pow_spec_avg_I1_corrected, Wavenum_theo, Wavenum_interval,
                                    &wavenum_fft_min, &wavenum_fft_max,
                                    &wavenum_min_halfmax_fft, &wavenum_max_halfmax_fft); /* outputs */
        
        vndrs_estimate_peak_limits(WL_spec_avg_I1_vector_corrected, Wavenum_theo, Wavenum_interval,
                                    &wavenum_wl_min, &wavenum_wl_max, /* outputs */
                                    &wavenum_min_halfmax_wl, &wavenum_max_halfmax_wl); /* outputs */
            
        /* Integration of the power in the fringe peak */
        /* ------------------------------------------- */
        /* BEWARE : median values are still computed by bootstrap with the specified wavenumber limits */
        
        /* "mil" = measured integration limits */
        mu1_fft_avg_mil = vndrs_integrate_PSD(Pow_spec_avg_I1_corrected,
            Wavenum_interval, wavenum_fft_min,  wavenum_fft_max);
        mu1_wl_avg_mil = vndrs_integrate_PSD(WL_spec_avg_I1_vector_corrected,
            Wavenum_interval, wavenum_wl_min,  wavenum_wl_max);
        
        /* Specified integration limits values - these results are returned to the user */
        mu1_fft_avg = vndrs_integrate_PSD(Pow_spec_avg_I1_corrected,
            Wavenum_interval, Wavenum_limits_fft->data[0], Wavenum_limits_fft->data[1]);
        mu1_wl_avg = vndrs_integrate_PSD(WL_spec_avg_I1_vector_corrected,
            Wavenum_interval, Wavenum_limits_wl->data[0], Wavenum_limits_wl->data[1]);

        mu1_fft_avg_noncentered = vndrs_integrate_PSD(Pow_spec_avg_I1_no_centering_corrected,
            Wavenum_interval, Wavenum_limits_fft->data[0], Wavenum_limits_fft->data[1]);
        mu1_wl_avg_noncentered = vndrs_integrate_PSD(WL_spec_avg_I1_vector_no_centering_corrected,
            Wavenum_interval, Wavenum_limits_wl->data[0], Wavenum_limits_wl->data[1]);

        /* Unweighted values */
        /* ----------------- */
        unweighted_mu1_fft = vndrs_mean_scan(non_zero_mu1_fft);
        unweighted_mu1_wl = vndrs_mean_scan(non_zero_mu1_wl);
        
    }
    else
    {
        *time_batch = 0.0;
        *baseline_batch = 0.0;
        *azimuth_batch = 0.0;
        *u_batch = 0.0;
        *v_batch = 0.0;
    }

     
    /* ----------------------------------- */
    
    /* Batch values (average or median, set by the user) pkervell 15-apr-02 */
    /* BEWARE OF MEDIAN VALUES ! - pkervell 26-Sep-02 */
    if(median_output == 0)
        {
        if(*N_scans_processed > 1)
            {
            *Mu1_batch_wl = mu1_wl_avg;
            *ErrMu1_batch_wl = mu1_wl_stdev_avg;
            *ErrMu1_batch_min_wl = Ci_mu1_wl_avg_min;
            *ErrMu1_batch_max_wl = Ci_mu1_wl_avg_max;
            *Mu1_batch_fft = mu1_fft_avg;
            *ErrMu1_batch_fft = mu1_fft_stdev_avg;
            *ErrMu1_batch_min_fft = Ci_mu1_fft_avg_min;
            *ErrMu1_batch_max_fft = Ci_mu1_fft_avg_max;
            *Mu1_skew_fft = skew_mean_fft;
            *Mu1_skew_wl = skew_mean_wl;
            *Photom_power = avg_photom;
            }
        else if(*N_scans_processed == 1)  /* case when we have only one scan processed */
            {
            *Mu1_batch_wl = mu1_wl_avg;
            mu1_wl_stdev_avg = 0.0;
            *ErrMu1_batch_wl = 0.0;
            *ErrMu1_batch_min_wl = 0.0;
            *ErrMu1_batch_max_wl = 0.0;
            *Mu1_batch_fft = mu1_fft_avg;
            mu1_fft_stdev_avg = 0.0;
            *ErrMu1_batch_fft = 0.0;
            *ErrMu1_batch_min_fft = 0.0;
            *ErrMu1_batch_max_fft = 0.0;
            *Mu1_skew_fft = 0.0;
            *Mu1_skew_wl = 0.0;
            *Photom_power = avg_photom;
        }
    }
    else
    {
        if(*N_scans_processed > 1)
        {
            *Mu1_batch_wl = mu1_wl_med;
            *ErrMu1_batch_wl = mu1_wl_stdev_med;
            *ErrMu1_batch_min_wl = Ci_mu1_wl_med_min;
            *ErrMu1_batch_max_wl = Ci_mu1_wl_med_max;
            *Mu1_batch_fft = mu1_fft_med;
            *ErrMu1_batch_fft = mu1_fft_stdev_med;
            *ErrMu1_batch_min_fft = Ci_mu1_fft_med_min;
            *ErrMu1_batch_max_fft = Ci_mu1_fft_med_max;
            *Mu1_skew_fft = skew_median_fft;
            *Mu1_skew_wl = skew_median_wl;
            *Photom_power = avg_photom;
        }
        else if(*N_scans_processed == 1)  /* case when we have only one scan processed */
        {
            *Mu1_batch_wl = mu1_wl_med;
            mu1_wl_stdev_med = 0.0;
            *ErrMu1_batch_wl = 0.0;
            *ErrMu1_batch_min_wl = 0.0;
            *ErrMu1_batch_max_wl = 0.0;
            *Mu1_batch_fft = mu1_fft_med;
            mu1_fft_stdev_med = 0.0;
            *ErrMu1_batch_fft = 0.0;
            *ErrMu1_batch_min_fft = 0.0;
            *ErrMu1_batch_max_fft = 0.0;
            *Mu1_skew_fft = 0.0;
            *Mu1_skew_wl = 0.0;
            *Photom_power = avg_photom;
        }
    }


    if(*N_scans_processed == 0)
    {       
	printf("\n Sorry... No interferogram processed in the batch. \n");
	sprintf(vndrs_protocol.text, "# NO INERFEROGRAM PROCESSED IN THE BATCH \n");  vndrs_runtime_protocol();
    }
    

    if(*N_scans_processed > 0)
    {

        vndrs_complete_report(mu1_fft_avg, mu1_wl_avg,
                        mu1_fft_avg_bs, mu1_wl_avg_bs,
                        mu1_fft_avg_noncentered, mu1_wl_avg_noncentered,
                        mu1_fft_avg_mil, mu1_wl_avg_mil,
                        unweighted_mu1_fft, unweighted_mu1_wl,
                        avg_background_fourier, avg_background_wavelets/*10*/,
                        *background_fourier, *background_wavelets/*11*/,
                        slope_background_fft, slope_background_wl,
                        Fringe_speed, Frame_rate, N_frames_cut, N_frames_precut,
                        Ci2_mu1_fft_avg_min, Ci2_mu1_fft_avg_max,
                        Ci2_mu1_wl_avg_min, Ci2_mu1_wl_avg_max,
                        Ci3_mu1_fft_avg_min/*21*/, Ci3_mu1_fft_avg_max,
                        Ci3_mu1_wl_avg_min, Ci3_mu1_wl_avg_max,
                        jack_fft, jack_wl,
                        simple_error_fft, simple_error_wl, Wavenum_interval,
                        Wavenum_limits_fft, Wavenum_limits_wl,
                        Wavenum_min_lowband_fft, Wavenum_max_lowband_fft,
                        Wavenum_min_lowband_wl, Wavenum_max_lowband_wl,
                        Wavenum_min_hi_band_fft, Wavenum_max_hi_band_fft,
                        Wavenum_min_hi_band_wl, Wavenum_max_hi_band_wl,
                        wavenum_fft_min/*39*/, wavenum_fft_max,
                        wavenum_wl_min, wavenum_wl_max,
                        skew_mean_fft, skew_mean_wl,
                        skew_median_fft, skew_median_wl,
                        kurtosis_mean_fft, kurtosis_mean_wl,
                        kurtosis_median_fft/*49*/, kurtosis_median_wl,
                        Wavenum_theo, weighted_peak_pos, weighted_peak_pos_wl,
                        weighted_peak_pos_fft_centered, weighted_peak_pos_wl_centered,
                        Off_offsets_Mean, Off_offsets_Stddev,
                        A_Mean, B_Mean,
                        On_source_Mean,
                        on_src_bg_I1, on_src_bg_I2, on_src_bg_PA, on_src_bg_PB,
                        SNR_I1, SNR_I2, SNR_PA, SNR_PB,
                        avg_photom, average_unbalance,
                        kappa1A, kappa1B, kappa2A, kappa2B,
                        kappa1A_err, kappa1B_err, kappa2A_err, kappa2B_err,
                        relative_kappa_norm_var,
                        splitting_ratio_A, splitting_ratio_B,
                        off_source_spectrum_slopes, number_of_pickups,
                        gains, RONs, flag_remove,
                        *N_scans_processed, N_scans_on,
                        removed_indices, *numberOfScansRejected /*84*/, n_sig_reject,
                        scans_rejected_0, scans_rejected_1, scans_rejected_2,
                        scans_rejected_3, scans_rejected_4,
                        scans_rejected_5, scans_rejected_6,
                        scans_rejected_7, scans_rejected_8,
                        mu1_fft_stdev_avg, Ci_mu1_fft_avg_min, Ci_mu1_fft_avg_max,
                        mu1_wl_stdev_avg, Ci_mu1_wl_avg_min, Ci_mu1_wl_avg_max,
                        mu1_fft_med, Ci_mu1_fft_med_min, Ci_mu1_fft_med_max,
                        mu1_wl_med, Ci_mu1_wl_med_min, Ci_mu1_wl_med_max,
                        *total_weight_batch, *time_batch, *baseline_batch,
                        *azimuth_batch, *u_batch, *v_batch, Autotest_case,
                        histogram_scale_fft, histogram_fft,
                        histogram_scale_wl, histogram_wl,
                        peak_pos_mean, peak_pos_sigma, peak_width_avg, peak_width_sigma,
                        altitude_batch_obj, azimuth_batch_obj,
                        wavenum_min_halfmax_fft, wavenum_max_halfmax_fft,
                        wavenum_min_halfmax_wl, wavenum_max_halfmax_wl,
                        ionic_in, select_scans,
                        warnings_flag, reject_flag /*out*/);

    }
        /* ---- */

    vndrs_write_power_spectra_to_file(N_frames_cut/2,
                                        Freq_scale_avg_spectra,
                                        Avg_power_spectra_on,
                                        Avg_power_spectra_off,
                                        WA_mean,
                                        WB_mean,
                                        Freq_scale_fringes,
                                        Pow_spec_avg_I1,
                                        Phase_I1,
                                        Pow_spec_avg_I1_no_centering,
                                        Phase_avg_I1_no_centering,
                                        Pow_spec_avg_I1_off,
                                        Background_PSD,
                                        Pow_spec_avg_I1_corrected,
                                        WL_spec_avg_I1_vector_corrected,
                                        WL_spec_avg_I1_vector_no_centering,
                                        WL_spec_avg_I1_vector_off,
                                        Background_PSD_wl_vector);
 
    for(I=0;I<N_scans_on;I++)
        if(Flag_processed->data[I] > 0)
        {
            /* SELECTION OF THE KIND OF DATA TO BE SAVED IN THE FITS OUTPUT FILE */
                /* Updating coherence factors if the user has selected to ouput raw visibilities */ 
                Batch_on_processed->scan_data[I].mu1_squared = Mu1_squared_wl->data[I]-*background_wavelets;
                Batch_on_processed->scan_data[I].mu2_squared = 0.0;

                /* Batch_on_processed->scan_data[I].mu1_squared = Mu1_squared->data[I]-*background_fourier;
                Batch_on_processed->scan_data[I].mu2_squared = 0.0;*/

        }

    if(verbose_on == 1)
    {
        vndrs_fprintf_vector_double(Background_PA, "Background_PA.dat");
        vndrs_fprintf_vector_double(Background_PB, "Background_PB.dat");
        vndrs_fprintf_vector_double(Background_PA_off, "Background_PA_off.dat");
        vndrs_fprintf_vector_double(Background_PB_off, "Background_PB_off.dat");
        vndrs_fprintf_vector_double(Photometry_I1, "Photometry_I1.dat");
        vndrs_fprintf_vector_double(Photometry_I2, "Photometry_I2.dat");
        vndrs_fprintf_vector_double(Photometry_PA, "Photometry_PA.dat");
        vndrs_fprintf_vector_double(Photometry_PB, "Photometry_PB.dat");
        /* added pkervell 03-Feb-03 - to illustrate VINCI processing paper */
        vndrs_fprintf_matrix_double(norm_t_wavelets_avg, "T_wavelets_average.dat");
    }


    vndrs_free_matrix_double(norm_t_wavelets_avg);
    vndrs_free_matrix_double(Standard_dev_Matrix);
    vndrs_free_batch(Batch_on_precut); 
    vndrs_free_batch(Batch_off_precut); 
    vndrs_free_batch(Batch_A_cut);
    vndrs_free_batch(Batch_B_cut);
    vndrs_free_batch(Batch_on_cut);
    vndrs_free_batch(Batch_off_cut);
    vndrs_free_vector_double(WA_mean);
    vndrs_free_vector_double(WB_mean);
    vndrs_free_vector_double(Off_offsets_Mean);
    vndrs_free_vector_double(Off_offsets_Stddev);
    vndrs_free_vector_double(Off_offsets_Min);
    vndrs_free_vector_double(Off_offsets_Max);
    vndrs_free_vector_double(On_source_Mean);
    vndrs_free_vector_double(On_source_Stddev);
    vndrs_free_vector_double(On_source_Min);
    vndrs_free_vector_double(On_source_Max);
    vndrs_free_vector_double(A_Mean);
    vndrs_free_vector_double(A_Stddev);
    vndrs_free_vector_double(A_Min);
    vndrs_free_vector_double(A_Max);
    vndrs_free_vector_double(B_Mean);
    vndrs_free_vector_double(B_Stddev);
    vndrs_free_vector_double(B_Min);
    vndrs_free_vector_double(B_Max);
    vndrs_free_vector_double(fringes_found_wavelets);
    vndrs_free_vector_double(Flag_processed);
    vndrs_free_vector_double(Mu1_squared_wl);
    vndrs_free_vector_double(Mu2_squared_wl);
    
    vndrs_free_vector_double(in_to_outband_ratio);
    vndrs_free_vector_double(peak_QL);
    vndrs_free_vector_double(residus_QL);
    vndrs_free_vector_double(fwhm_peak_time);
    vndrs_free_vector_double(peak_position_freq);
    vndrs_free_vector_double(fwhm_peak_freq);
    vndrs_free_vector_double(background_fit_QL);
    vndrs_free_vector_double(signal_p);    
    vndrs_free_vector_double(fringe_peak_strehl); /* pkervell - 21-Jan-04 */

    vndrs_free_vector_double(Pow_spec_avg_I1);
    vndrs_free_vector_double(Pow_spec_avg_I1_no_centering);
    vndrs_free_vector_double(Pow_spec_avg_I1_no_centering_corrected);
    vndrs_free_vector_double(Phase_avg_I1_no_centering);
    vndrs_free_vector_double(Pow_spec_avg_I1_off);
    vndrs_free_vector_double(Pow_spec_avg_buffer_I1);
    vndrs_free_vector_double(Pow_spec_avg_buffer_I1_off);
    vndrs_free_vector_double(Background_PSD);
    vndrs_free_vector_double(Pow_spec_avg_I1_corrected);
    vndrs_free_vector_double(Background_PSD_wl_vector);
    vndrs_free_vector_double(WL_spec_avg_I1_vector);
    vndrs_free_vector_double(WL_spec_avg_I1_vector_no_centering);
    vndrs_free_vector_double(WL_spec_avg_I1_vector_no_centering_corrected);
    vndrs_free_vector_double(WL_spec_avg_I1_vector_off);
    vndrs_free_vector_double(WL_spec_avg_I1_vector_corrected);
    vndrs_free_vector_double(Phase_I1);
    
    vndrs_free_vector_double(weights_visibilities);
    vndrs_free_vector_double(fringes_off_center);
    vndrs_free_vector_double(removed_indices);/**/
    vndrs_free_full_scan(Avg_power_spectra_on);
    vndrs_free_full_scan(Avg_power_spectra_off);
    vndrs_free_vector_double(Freq_scale_avg_spectra);
    vndrs_free_vector_double(Freq_scale_fringes);
    vndrs_free_vector_double(Apodisation_fringes);
    vndrs_free_vector_double(Background_PA);
    vndrs_free_vector_double(Background_PB);
    vndrs_free_vector_double(Background_I1);
    vndrs_free_vector_double(Background_I2);
    vndrs_free_vector_double(Background_PA_off);
    vndrs_free_vector_double(Background_PB_off);
    vndrs_free_vector_double(Photometry_I1);
    vndrs_free_vector_double(Photometry_I2);
    vndrs_free_vector_double(Photometry_PA);
    vndrs_free_vector_double(Photometry_PB);
    vndrs_free_batch(Batch_on_raw);
    vndrs_free_full_scan(Master_spectrum_off);
    vndrs_free_vector_double(Backgrounds);
    
    vndrs_free_matrix_double(PSDs_matrix);
    vndrs_free_matrix_double(Phase_matrix);
    vndrs_free_matrix_double(Phase_matrix_wl);
    vndrs_free_matrix_double(PSDs_matrix_wl);
    vndrs_free_matrix_double(PSDs_matrix_off);
    vndrs_free_matrix_double(PSDs_matrix_wl_off);

    vndrs_free_vector_double(time_scan);
    vndrs_free_vector_double(projected_baseline);
    vndrs_free_vector_double(azimuth_baseline);
    vndrs_free_vector_double(u_sampled);
    vndrs_free_vector_double(v_sampled);
    vndrs_free_matrix_double(norm_t_wavelets_I1_check);
    
    vndrs_free_vector_double(Unbalance_coef);
    vndrs_free_vector_double(off_source_spectrum_slopes);
    vndrs_free_vector_double(number_of_pickups);
    vndrs_free_vector_double(gains);
    vndrs_free_vector_double(RONs);
    vndrs_free_vector_double(splitting_ratio_A);
    vndrs_free_vector_double(splitting_ratio_B);
    vndrs_free_vector_double(altitude_object);
    vndrs_free_vector_double(azimuth_object);

    vndrs_free_vector_double(Amplitude_fft);
    vndrs_free_vector_double(Phase_fft);
    vndrs_free_vector_double(Real_fft);
    vndrs_free_vector_double(Imaginary_fft);
    vndrs_free_vector_double(Real_part_fft_buffer);
    vndrs_free_vector_double(Imaginary_part_fft_buffer);
    vndrs_free_matrix_double(Real_part_fft);
    vndrs_free_matrix_double(Imaginary_part_fft);
    vndrs_free_vector_double(Real_part_fft_no_centering);
    vndrs_free_vector_double(Imaginary_part_fft_no_centering);
    vndrs_free_vector_double(OPD_phase);
    vndrs_free_vector_double(Total_OPD);
    
    vndrs_free_vector_double(high_freq_soft);
    
    if(*N_scans_processed > 0)
    {
        vndrs_free_vector_double(histogram_scale_wl);
        vndrs_free_vector_double(histogram_wl);
        vndrs_free_vector_double(histogram_scale_fft);
        vndrs_free_vector_double(histogram_fft);
        vndrs_free_vector_double(non_zero_weights);
        vndrs_free_vector_double(non_zero_mu1_wl);
        vndrs_free_vector_double(non_zero_mu2_wl);
        vndrs_free_vector_double(non_zero_mu1_fft);
        vndrs_free_vector_double(non_zero_signal_p);
        vndrs_free_vector_double(non_zero_unbalance);
        vndrs_free_vector_double(non_zero_peak_pos);
        vndrs_free_vector_double(non_zero_peak_width);
    }
}






/*
    Performs a full photometric correction of the data.
    Described in [1], section 2.14
 */

void vndrs_full_correction (/* --------------- */
                            /* input variables */
                            /* --------------- */
                            batch *Batch_off,
                            /* Off_source batch */
                            full_scan *Master_spectrum_off,
                            /* Master avg power spectrum of the detector noise only - pkervell 11-Oct-02 */
                            full_scan *Scan_on,
                            /* the full scan to be processed */
			    power_spectrum *WA_mean,   /* added pkervell 14-mar-02 */
			    /* The Wiener filter computed from the average of PA PSDs */
			    power_spectrum *WB_mean,   /* added pkervell 14-mar-02 */
			    /* The Wiener filter computed from the average of PB PSDs */
                            vector_double *Background_PA, /* updated if too low fluxes are found */
                            vector_double *Background_PB,
                            matrix_double *Matrix_coefs,
                            /* The 2 x 2 transfer matrix between the photometric and interferometric signals
                               (both measured at photometers (see [2], Eq. (19) on p. 383):

                                    [ I1 ]   [ kappa_1A  kappa_1B ]   [ PA ]
                                    |    | = |                    | * |    |
                                    [ I2 ]   [ kappa_2A  kappa_2B ]   [ PB ]
			    */
                            double Frame_rate,
                            /* Fringe aquisition frequency (in units of Hz) */
                            double Fringe_speed,
                            /* OPD modulation speed (in units of m/sec) */
			    double fringes_length,
			    /* length of the fringe packet, in pixels */
                            double Rejection_threshold,
                            double Apod_blockout,
                            double Apod_transition,
                            double Mult_corr,
			    int debias,
			    int ionic_in,
			    int verbose_on,
                            /* ---------------- */
                            /* output variables */
                            /* ---------------- */
                            double *Min_P,
                            double *unbalance_coef,
                            full_scan *Scan_on_processed,
                            full_scan *Scan_off_processed,
			    double *Processed_flag)
{
    /* Local variables */
    /* --------------- */
    long N_frames, N_scans_off, N_scans_sig, I;
    int process_fringes;
    double Sample_interval_OPD;
    /* sample interval in units of m(eters), i.e. OPD,  between two adjacent samples in a scan */
    double kappa_1A, kappa_1B, kappa_2A, kappa_2B;
    /* the coefficients of the transfer matrix between the 2 photometric signals PA, PB and
       the 2 interferometric signals I1, I2:

               [ I1 ]   [ kappa_1A  kappa_1B ]   [ PA ]
               |    | = |                    | * |    |
               [ I2 ]   [ kappa_2A  kappa_2B ]   [ PB ]
    */
    double Kappa1, Kappa2;
    /* Kappa1 = sqrt(kappa_1A * kappa_1B);
       Kappa2 = sqrt(kappa_2A * kappa_2B);
    */
    double Stddev_PA_off, Stddev_PB_off;

    individual_scan *Photom_I1, *Photom_I2;
    individual_scan *PA_filtered, *PB_filtered;
    individual_scan *PA_off_filtered, *PB_off_filtered;
    individual_scan *P, *PP1, *PP2;
    individual_scan *Apodisation_filter;
    individual_scan *I1_cal_photom, *I2_cal_photom;
    individual_scan *PP1_apod, *PP2_apod;
    individual_scan *I1_processed, *I2_processed;
    individual_scan *I1_off_processed, *I2_off_processed;
    individual_scan *buffer;
    full_scan *Master_scan_off;
    
    double avg_photom, avg_PA, avg_PB, minimum_PA, minimum_PB, minimum_P, minimum_P_on_fringes;
    
    vector_double *subtracted_vector;


    /* Get the number of frames and compute the OPD interval (unit m) between two frames */
    N_scans_off         = Batch_off->number_of_scans;
    N_frames            = Scan_on->I1->dimension;
    Sample_interval_OPD = fabs(Fringe_speed / Frame_rate); /* This time it's the sample interval
                                                              in units of m(eters) ==> OPD (!!!) */
    /*===================*/
    /* MEMORY ALLOCATION */
    /*===================*/
    
    buffer     = vndrs_calloc_vector_double(N_frames);
    Photom_I1  = vndrs_calloc_vector_double(N_frames);
    Photom_I2  = vndrs_calloc_vector_double(N_frames);
    P             = vndrs_calloc_vector_double(N_frames);
    PP1           = vndrs_calloc_vector_double(N_frames);
    PP2           = vndrs_calloc_vector_double(N_frames);
    I1_cal_photom = vndrs_calloc_vector_double(N_frames);
    I2_cal_photom = vndrs_calloc_vector_double(N_frames);
    PP1_apod         = vndrs_calloc_vector_double(N_frames);
    PP2_apod         = vndrs_calloc_vector_double(N_frames);
    I1_processed     = vndrs_calloc_vector_double(N_frames);
    I2_processed     = vndrs_calloc_vector_double(N_frames);
    I1_off_processed = vndrs_calloc_vector_double(N_frames);
    I2_off_processed = vndrs_calloc_vector_double(N_frames);
    Master_scan_off  = vndrs_calloc_full_scan(N_frames);
    subtracted_vector = vndrs_calloc_vector_double(N_frames);
   
    /*==========================*/
    /* END OF MEMORY ALLOCATION */
    /*==========================*/

    
    /* WHY ONLY THE FIRST ONE ? - good question... let's take a few - modified - pkervell 09-Sep-02 */

    Stddev_PA_off = 0.0;
    Stddev_PB_off = 0.0;
    
    if(N_scans_off < 3) N_scans_sig = N_scans_off;
    else N_scans_sig = 3;
    
    for(I=0;I<N_scans_sig;I++)
    {
        PA_off_filtered = vndrs_filter(Batch_off->fullscans[I]->PA, WA_mean);
        PB_off_filtered = vndrs_filter(Batch_off->fullscans[I]->PB, WB_mean);
        /* memory allocated in the vndrs_filter procedure */
        Stddev_PA_off += vndrs_stddev_scan(PA_off_filtered)/N_scans_sig;
        Stddev_PB_off += vndrs_stddev_scan(PB_off_filtered)/N_scans_sig;
        
        vndrs_free_vector_double(PA_off_filtered);
        vndrs_free_vector_double(PB_off_filtered);
    }
	
    /* Get the matrix coefficients */
    vndrs_get_transfer_matrix_coefficients(Matrix_coefs, &kappa_1A, &kappa_1B, &kappa_2A, &kappa_2B);

    /* Frequency filtering of the photometric signals, both on and Off_source;
       the on-source are 'trimmed positive' afterwards (i.e. their negative components
       are replaced by zeros.) */
    /* We remove the general slope of the signals to avoid edge effects - pkervell 14-Oct-02 */
    vndrs_remove_vector_slope(Scan_on->PA, 3, subtracted_vector);    
    PA_filtered = vndrs_filter(Scan_on->PA, WA_mean);
    vndrs_vector_add(PA_filtered, subtracted_vector);
    vndrs_vector_add(Scan_on->PA, subtracted_vector);
    
    vndrs_remove_vector_slope(Scan_on->PB, 3, subtracted_vector);    
    PB_filtered = vndrs_filter(Scan_on->PB, WB_mean);
    vndrs_vector_add(PB_filtered, subtracted_vector);
    vndrs_vector_add(Scan_on->PB, subtracted_vector);
    
    /* There is often a problem in the estimation of the background of the interferograms,
    that is often overestimated for the low SNR scans. This is really bad in the sense that
    it creates artificially an unlimited ammount of visibility at the division by the geometrical
    mean of PA and PB. In the following, I ASSUME that the minimum light reached on the scan
    is ABOVE the zero flux. This means that if it is below, I add the corresponding
    difference. - pkervell 14-Oct-02 */
    
    minimum_PA = vndrs_vector_min(PA_filtered);
    minimum_PB = vndrs_vector_min(PB_filtered);
    
    if(minimum_PA < 0.0)
    {
        /* printf("PA corrected by %f ADU.\n", -minimum_PA);*/
        vndrs_vector_add_constant(PA_filtered, -minimum_PA);
        Background_PA->data[I] = Background_PA->data[I] + minimum_PA;
    }
    if(minimum_PB < 0.0)
    {
        /* printf("PB corrected by %f ADU.\n", -minimum_PB);*/
        vndrs_vector_add_constant(PB_filtered, -minimum_PB);
        Background_PB->data[I] = Background_PB->data[I] + minimum_PB;
    }
    
    vndrs_vector_trim_positive(PA_filtered); /* In principle, this is not necessary any more */
    vndrs_vector_trim_positive(PB_filtered);

    /* Recomputing minima and averages, on the fringes only */
    minimum_PA = vndrs_vector_min_center(PA_filtered, fringes_length);
    minimum_PB = vndrs_vector_min_center(PB_filtered, fringes_length);    
    avg_PA = vndrs_mean_scan_center_weighted(Scan_on->PA, fringes_length);
    avg_PB = vndrs_mean_scan_center_weighted(Scan_on->PB, fringes_length);    
    
    /* Computation of the photometric calibration for each interferometric channel */

    /* Photom_I1_unfiltered = kappa_1A * PA_cal + kappa_1B * PB_cal */
    vndrs_vector_memcpy(Photom_I1, Scan_on->PA);
    vndrs_vector_memcpy(buffer, Scan_on->PB);
    vndrs_vector_scale(Photom_I1, kappa_1A);
    vndrs_vector_scale(buffer, kappa_1B);
    vndrs_vector_add(Photom_I1, buffer);

    if(ionic_in == 0)
    {
        /* Photom_I2_unfiltered = kappa_2A * PA_cal + kappa_2B * PB_cal */
        vndrs_vector_memcpy(Photom_I2, Scan_on->PA);
        vndrs_vector_memcpy(buffer, Scan_on->PB);
        vndrs_vector_scale(Photom_I2, kappa_2A);
        vndrs_vector_scale(buffer, kappa_2B);
        vndrs_vector_add(Photom_I2, buffer);
    }
    
    /* Computation of the Kappa1 and Kappa2 coefficients */
    Kappa1 = sqrt(kappa_1A * kappa_1B);
    if(ionic_in == 0)
	Kappa2 = sqrt(kappa_2A * kappa_2B);
    else
	Kappa2 = 1.0; /* necessary for the tests below - pkervell 29-Jul-02 */

    /* Computation of the final photometric signals */
    /* P = sqrt(PA_filtered * PB_filtered) (compare p. 384, right column of [2])*/   
    vndrs_vector_memcpy(P, PA_filtered);
    vndrs_vector_mul(P, PB_filtered);
    
    minimum_P = vndrs_vector_min(P);
    minimum_P_on_fringes = vndrs_vector_min_center(P, fringes_length);

    vndrs_vector_trim_positive(P); /* To be sure... */
    vndrs_vector_sqrt(P);

    avg_photom = vndrs_mean_scan_center_weighted(P, fringes_length);
    /* avg_photom = vndrs_mean_scan_center(P, fringes_length); */

    /* the average photometry SNR is returned to the main program */
    *Min_P = avg_photom / ((Stddev_PA_off + Stddev_PB_off)/2.0);
    
    /* unbalance coefficient, returned to the main program - pkervell 17-Oct-02 */
    if((avg_PA > avg_PB) && (avg_PA > 0.0) && (avg_PB > 0.0))
        *unbalance_coef = avg_PB/avg_PA;
    else if((avg_PA < avg_PB) && (avg_PA > 0.0) && (avg_PB > 0.0))
        *unbalance_coef = avg_PA/avg_PB;
    else *unbalance_coef = 0.0;
    
    process_fringes = 0;
    
    if((minimum_P > Rejection_threshold * Stddev_PA_off) &&
        (minimum_P > Rejection_threshold * Stddev_PB_off))
    {
        /* Regular calibration using the sqrt(PA.PB) signal - no modifications needed on P */        
        if(debias == 1)
        {
            /* Creation of the master off-source scan */
            vndrs_create_master_off_source_scan(Master_spectrum_off, Master_scan_off, verbose_on);
            if(verbose_on == 1)
            {
                vndrs_fprintf_vector_double(Master_scan_off->I1, "Master_scan_off_I1_bis.dat");
                vndrs_fprintf_vector_double(Master_scan_off->I2, "Master_scan_off_I2_bis.dat");
            }

        }
        process_fringes = 1;
    }
    else if((minimum_P_on_fringes > (Rejection_threshold * Stddev_PA_off))
        && (minimum_P_on_fringes > (Rejection_threshold * Stddev_PB_off)))
    /* else if((minimum_P_on_fringes > (Stddev_PA_off))
        && (minimum_P_on_fringes > (Stddev_PB_off))) tested and rejected pkervell 06-Nov-02 */
        
    {	/* Very faint case:  - pkervell - 04-Oct-02 */
    
        vndrs_vector_fill(P, avg_photom);
        if(debias == 1)
            /* Creation of the master off-source scan. */
            vndrs_create_master_off_source_scan(Master_spectrum_off, Master_scan_off, verbose_on);
        
        process_fringes = 2;
    }
    else process_fringes = 0;

    /* Construction of the apodisation filter (MEMORY FOR Apodisation_filter IS ALLOCATED INSIDE THE ROUTINE) */
    Apodisation_filter = vndrs_apodisation_window(N_frames,
                                                Apod_blockout, 
                                                Apod_transition);
    
    /* -------------------------------------------------------------- */
    /* CHECK IF THE CURRENT SCAN WILL BE PROCESSED OR NOT             */
    /* -------------------------------------------------------------- */
    /*if((minimum_PA > Stddev_PA_off) &&
        (minimum_PB > Stddev_PB_off))*/
    if(process_fringes > 0)
    {
        /* CURRENT SCAN WILL BE PROCESSED ...                         */
        /* :::::::::::::::::::::::::::::::::::::::::::::::::::::::::: */

        /* PP1 = 2*Kappa1*P */ 
        vndrs_vector_memcpy(PP1, P);
        vndrs_vector_scale(PP1, 2.0 * Kappa1);

        if(ionic_in == 0)
        {
            /* PP2 = 2*Kappa2*P */
            vndrs_vector_memcpy(PP2, P);
            vndrs_vector_scale(PP2, 2.0 * Kappa2);
        }
        
        /* Photometric Calibration of the interferometric signals */
        /* I1_cal_photom = I1_cal - Photom_I1 */
        vndrs_vector_memcpy(I1_cal_photom, Scan_on->I1);
        vndrs_vector_sub(I1_cal_photom, Photom_I1);
        /* mean on fringes set to zero (pkervell - 11-Oct-02) */
        vndrs_vector_add_constant(I1_cal_photom, - vndrs_mean_scan_center(I1_cal_photom, fringes_length));
        /* Removal of general slope */
        vndrs_remove_vector_slope(I1_cal_photom, 3, subtracted_vector);
        /* Apodisation of the photometry calibrated signals - pkervell 12-Oct-02 */
        vndrs_vector_mul(I1_cal_photom, Apodisation_filter);

        if(ionic_in == 0)
        {
            /* I2_cal_photom = I2_cal - Photom_I2 */
            vndrs_vector_memcpy(I2_cal_photom, Scan_on->I2);
            vndrs_vector_sub(I2_cal_photom, Photom_I2);
            vndrs_vector_add_constant(I2_cal_photom, - vndrs_mean_scan_center(I2_cal_photom, fringes_length));
            vndrs_remove_vector_slope(I2_cal_photom, 3, subtracted_vector);    
            vndrs_vector_mul(I2_cal_photom, Apodisation_filter);
        }
        
        /* Apodisation of the PP1 and PP2 signals */
        vndrs_vector_memcpy(PP1_apod, Apodisation_filter);
        vndrs_vector_div(PP1_apod, PP1);
        if(ionic_in == 0)
        {
            vndrs_vector_memcpy(PP2_apod, Apodisation_filter);
            vndrs_vector_div(PP2_apod, PP2);
        }
        
        /* Final on source signals */
        /* I1_processed = PP1_apod * I1_cal_photom */ 
        vndrs_vector_memcpy(I1_processed, PP1_apod);
        vndrs_vector_mul(I1_processed, I1_cal_photom);

        if(ionic_in == 0)
        {
            /* I2_processed = PP2_apod * I2_cal_photom */
            vndrs_vector_memcpy(I2_processed, PP2_apod);
            vndrs_vector_mul(I2_processed, I2_cal_photom);
        }
        
        /* We remove the general slope of the processed signals to avoid edge effects later - pkervell 14-Oct-02 */
        vndrs_remove_vector_slope(I1_processed, 3, subtracted_vector);    
        if(ionic_in == 0) vndrs_remove_vector_slope(I2_processed, 3, subtracted_vector);

	/* If we have activated the debias by the off-source scans - pkervell 09-Sep-02 */
        if(debias == 1 && ionic_in == 0)
        {
	    /* Final Off_source signals */
	    vndrs_process_off(Master_scan_off,
                              Photom_I1,
			      PP1_apod,
                              Photom_I2,
			      PP2_apod,
                              Apodisation_filter,
			      I1_off_processed,
			      I2_off_processed,
			      verbose_on);
        }

        if(ionic_in == 0)
        {
            /* Now we output the subtracted and added interferograms in the
            I1 and I2 signals of the processed scan. Scaling removed as it may bias the data
            between two stars for which the kappas are not identical. */
            vndrs_vector_memcpy(Scan_on_processed->I1, I1_processed);
            /* vndrs_vector_scale(Scan_on_processed->I1, Kappa1);*/
            vndrs_vector_memcpy(buffer, I2_processed);
            /* vndrs_vector_scale(buffer, Kappa2);*/
            vndrs_vector_sub(Scan_on_processed->I1, buffer);
            /* vndrs_vector_scale(Scan_on_processed->I1, 1.0/(Kappa1+Kappa2));*/
            vndrs_vector_scale(Scan_on_processed->I1, 0.5);
            /* Setting average to zero (to be sure !) */
            vndrs_vector_add_constant(Scan_on_processed->I1, - vndrs_mean_scan_center(Scan_on_processed->I1, fringes_length));
            
            vndrs_vector_memcpy(Scan_on_processed->I2, I1_processed);
            vndrs_vector_memcpy(buffer, I2_processed);
            vndrs_vector_add(Scan_on_processed->I2, buffer);
            vndrs_vector_scale(Scan_on_processed->I2, 0.5);
            vndrs_vector_add_constant(Scan_on_processed->I2, - vndrs_mean_scan_center(Scan_on_processed->I2, fringes_length));
            
            vndrs_vector_memcpy(Scan_off_processed->I1, I1_off_processed);
            vndrs_vector_memcpy(buffer, I2_off_processed);
            vndrs_vector_sub(Scan_off_processed->I1, buffer);
            vndrs_vector_scale(Scan_off_processed->I1, 0.5);
            vndrs_vector_add_constant(Scan_off_processed->I1, - vndrs_mean_scan_center(Scan_off_processed->I1, fringes_length));
            
            vndrs_vector_memcpy(Scan_off_processed->I2, I1_off_processed);
            vndrs_vector_memcpy(buffer, I2_off_processed);
            vndrs_vector_add(Scan_off_processed->I2, buffer);
            vndrs_vector_scale(Scan_off_processed->I2, 0.5);
            vndrs_vector_add_constant(Scan_off_processed->I2, - vndrs_mean_scan_center(Scan_off_processed->I2, fringes_length));
        }
        else
        {
            /* In the case of IONIC, the same processed scan is placed in the output variable */
            vndrs_vector_memcpy(Scan_on_processed->I1, I1_processed);        
            vndrs_vector_memcpy(Scan_on_processed->I2, I1_processed);            
        }
        
        /* Copying photometry into output variables */
        vndrs_vector_memcpy(Scan_on_processed->PA, PA_filtered);
        vndrs_vector_memcpy(Scan_on_processed->PB, PB_filtered);
       
        /* Set output flag to 1 or 2 (= 'processed normal', 2 = 'processed faint mode') */
        *Processed_flag = process_fringes;

    }
    else
    {
        /* CURRENT SCAN WILL NOT BE PROCESSED ...                     */
        /* :::::::::::::::::::::::::::::::::::::::::::::::::::::::::: */
        /* Output variables in case of no reduction  are already initialized to 0 */
        /* Set output flag to 0 (= 'not processed') */
        *Processed_flag = 0;
    }

    if(verbose_on == 1)
        {
        vndrs_fprintf_vector_double(P, "P.dat");
        vndrs_fprintf_vector_double(Photom_I1, "Photom_I1.dat");
        vndrs_fprintf_vector_double(Photom_I2, "Photom_I2.dat");
        vndrs_fprintf_vector_double(PA_filtered, "PA_filtered.dat");
        vndrs_fprintf_vector_double(PB_filtered, "PB_filtered.dat");
        vndrs_fprintf_vector_double(PP1, "PP1.dat");
        vndrs_fprintf_vector_double(PP2, "PP2.dat");
        vndrs_fprintf_vector_double(I1_cal_photom, "I1_cal_photom.dat");
        vndrs_fprintf_vector_double(I2_cal_photom, "I2_cal_photom.dat");
        vndrs_fprintf_vector_double(PP1_apod, "PP1_apod.dat");
        vndrs_fprintf_vector_double(PP2_apod, "PP2_apod.dat");
        vndrs_fprintf_vector_double(I1_processed, "I1_processed_raw.dat");
        vndrs_fprintf_vector_double(I2_processed, "I2_processed_raw.dat");
        vndrs_fprintf_vector_double(Scan_on_processed->I1, "I1_processed_apod.dat");
        vndrs_fprintf_vector_double(Scan_on_processed->I2, "I2_processed_apod.dat");
        vndrs_fprintf_vector_double(I1_off_processed, "I1_off_processed.dat");
        vndrs_fprintf_vector_double(I2_off_processed, "I2_off_processed.dat");
        vndrs_fprintf_vector_double(PA_filtered, "PA_filtered.dat");
        vndrs_fprintf_vector_double(PB_filtered, "PB_filtered.dat");
        }


    /*================*/
    /* MEMORY RELEASE */
    /*================*/
    
    vndrs_free_vector_double(buffer);
    vndrs_free_vector_double(Photom_I1);
    vndrs_free_vector_double(Photom_I2);
    vndrs_free_vector_double(PA_filtered);
    vndrs_free_vector_double(PB_filtered);
    vndrs_free_vector_double(P);
    vndrs_free_vector_double(PP1);
    vndrs_free_vector_double(PP2);
    vndrs_free_vector_double(I1_cal_photom);
    vndrs_free_vector_double(I2_cal_photom);
    vndrs_free_vector_double(PP1_apod);
    vndrs_free_vector_double(PP2_apod);
    vndrs_free_vector_double(I1_processed);
    vndrs_free_vector_double(I2_processed);
    vndrs_free_vector_double(I1_off_processed);
    vndrs_free_vector_double(I2_off_processed);
    vndrs_free_full_scan(Master_scan_off);
    vndrs_free_vector_double(subtracted_vector);
    vndrs_free_vector_double(Apodisation_filter); 

    return;
}



/*
    Described in [1], section 2.16
 */
void vndrs_process_off ( 
                        /* --------------- */ 
                        /* input variables */ 
                        /* --------------- */ 
                        full_scan *Master_scan_off,
                        individual_scan *Photom_I1,
                        individual_scan *PP1_apod,
                        individual_scan *Photom_I2,
                        individual_scan *PP2_apod,
                        vector_double *Apodisation_filter,
                        /* ---------------- */ 
                        /* output variables */ 
                        /* ---------------- */ 
                        individual_scan *I1_off_processed,
                        individual_scan *I2_off_processed,
			int verbose_on
                       )
{
    long N_frames;
    double Mean_cond_dark;
    vector_double *dummy;
    
    N_frames = Master_scan_off->I1->dimension;
    
    dummy = vndrs_calloc_vector_double(N_frames);
    
    /*    sprintf(vndrs_protocol.text," N_off_samples  : %d, N_scans: %d \n",
	  N_off_samples, N_scans); vndrs_runtime_protocol(); */

    /* There is a problem here... */
    vndrs_vector_memcpy(I1_off_processed, Master_scan_off->I1);
    vndrs_vector_memcpy(I2_off_processed, Master_scan_off->I2);
    
    if(verbose_on == 1)
    {    
        vndrs_fprintf_vector_double(Master_scan_off->I1, "Master_scan_off_I1_ter.dat");    
        vndrs_fprintf_vector_double(Master_scan_off->I2, "Master_scan_off_I2_ter.dat");    
        vndrs_fprintf_vector_double(I1_off_processed, "I1_off_processed_base.dat");    
        vndrs_fprintf_vector_double(I2_off_processed, "I2_off_processed_base.dat");    
    }
    /* Now we process I1 and I2_off_processed using the same calibration signals as the on-source */ 
 
    vndrs_vector_sub(I1_off_processed, Photom_I1);
    vndrs_vector_mul(I1_off_processed, PP1_apod);
    vndrs_vector_sub(I2_off_processed, Photom_I2);
    vndrs_vector_mul(I2_off_processed, PP2_apod);

    /* We remove the general slope of the signals to avoid edge effects later - pkervell 14-Oct-02 */
    vndrs_remove_vector_slope(I1_off_processed, 3, dummy);    
    vndrs_remove_vector_slope(I2_off_processed, 3, dummy);    

    /* Setting average values to zero */
    Mean_cond_dark = vndrs_mean_scan(I1_off_processed);
    vndrs_vector_add_constant(I1_off_processed, -Mean_cond_dark);
    Mean_cond_dark = vndrs_mean_scan(I2_off_processed);
    vndrs_vector_add_constant(I2_off_processed, -Mean_cond_dark);
    
    /* (Re-)apodisation of the processed off-source (better to have clean edges) */
    vndrs_vector_mul(I1_off_processed, Apodisation_filter);
    vndrs_vector_mul(I2_off_processed, Apodisation_filter);
   
    if(verbose_on == 1)
    {
        vndrs_fprintf_vector_double(I1_off_processed, "I1_off_processed_raw.dat");    
        vndrs_fprintf_vector_double(I2_off_processed, "I2_off_processed_raw.dat");    
    }
    
    vndrs_free_vector_double(dummy);
    
    return;
}

/*
    Creation of the master off-source scan from the master offsrc spectrum
*/
void vndrs_create_master_off_source_scan(full_scan *Master_spectrum_off, /* in*/
                                         full_scan *Master_scan_off, /* out*/
                                         int verbose_on) 
{
    long I;
    long N_off_samples;
    power_spectrum *Off_power_I1, *Off_power_I2;
    fourier_spectrum *complex_vector_dummy;
    vector_double *vector_dummy, *random_vector;
    vector_complex *Cond_dark_complex;

    N_off_samples = Master_spectrum_off->I1->dimension;

    complex_vector_dummy = vndrs_calloc_vector_complex(N_off_samples);
    vector_dummy = vndrs_calloc_vector_double(N_off_samples);
    Cond_dark_complex = vndrs_calloc_vector_complex(N_off_samples);
    Off_power_I1 = vndrs_calloc_vector_double(N_off_samples);
    Off_power_I2 = vndrs_calloc_vector_double(N_off_samples);
    random_vector = vndrs_calloc_vector_double(N_off_samples);
    
    /* Creation of the amplitude spectra of the master off-source for I1 and I2 */
    vndrs_vector_memcpy(Off_power_I1, Master_spectrum_off->I1);
    vndrs_vector_memcpy(Off_power_I2, Master_spectrum_off->I2);
    vndrs_vector_sqrt(Off_power_I1);
    vndrs_vector_sqrt(Off_power_I2);

    /* Generation of the master off-source I1 and I2 scans with random phase, from the master off-source PSD - pkervell 11-Oct-02 */
    srand(time(0)); /* BEWARE !! THE PHASE HAS TO BE SYMMETRIC !! Otherwise, factor 2 underestimation of noise... - pkervell 12-Oct-02 */
    for(I=0;I<N_off_samples;I++) random_vector->data[I] = 2*Pi*((double)rand()/RAND_MAX);   
    /* Generation of a full vector of random phases, half of it is used below for I2 */     
    for(I=0;I<N_off_samples/2;I++)
    {
        vector_dummy->data[I] = random_vector->data[I];
        vector_dummy->data[N_off_samples-1-I] = random_vector->data[I];
    }
    vndrs_make_complex_vector_modulus_phase(Off_power_I1, vector_dummy, complex_vector_dummy);    
    vndrs_ifft(complex_vector_dummy, Cond_dark_complex);
    vndrs_vector_complex_2_vector_double_real(Cond_dark_complex, Master_scan_off->I1);

    if(verbose_on == 1)
    {
        vndrs_fprintf_vector_double(vector_dummy, "Off_phase_I1.dat");
        vndrs_fprintf_vector_complex(complex_vector_dummy, "Off_complex_spectrum_I1.dat");
        vndrs_fprintf_vector_complex(Cond_dark_complex, "Off_cond_dark_complex_I1.dat");   
        vndrs_fprintf_vector_double(Master_scan_off->I1, "Master_scan_off_I1.dat");   
    }

    for(I=0;I<N_off_samples/2;I++)
    {
        vector_dummy->data[I] = random_vector->data[N_off_samples-1-I];
        vector_dummy->data[N_off_samples-1-I] = random_vector->data[N_off_samples-1-I];
    }
    vndrs_make_complex_vector_modulus_phase(Off_power_I2, vector_dummy, complex_vector_dummy);    
    vndrs_ifft(complex_vector_dummy, Cond_dark_complex);
    vndrs_vector_complex_2_vector_double_real(Cond_dark_complex, Master_scan_off->I2);

    if(verbose_on == 1)
    {
        vndrs_fprintf_vector_double(vector_dummy, "Off_phase_I2.dat");
        vndrs_fprintf_vector_complex(complex_vector_dummy, "Off_complex_spectrum_I2.dat");
        vndrs_fprintf_vector_complex(Cond_dark_complex, "Off_cond_dark_complex_I2.dat");   
        vndrs_fprintf_vector_double(Master_scan_off->I2, "Master_scan_off_I2.dat");   
    }
    
    vndrs_free_vector_double(Off_power_I1);
    vndrs_free_vector_double(Off_power_I2);
    vndrs_free_vector_complex(Cond_dark_complex);
    vndrs_free_vector_complex(complex_vector_dummy);
    vndrs_free_vector_double(vector_dummy);    
    vndrs_free_vector_double(random_vector);    
}














































































