/*******************************************************************/
/* VLTI VINCI DATA REDUCTION SOFTWARE (VNDRS)                      */
/*                                                                 */
/* FILE: vndrs_MathLib.c                                           */
/*                                                                 */
/* The self-made math library for VNDRS (with interface to FFTW)   */
/*                                                                 */
/* Written by Markus Schoeller & Rainer Wilhelm, ESO, 2000         */
/*                                                                 */
/*******************************************************************/ 

#include "vndrs.h"

/* -------------------------------------------------------------------------------- */
/* ROUTINES TO ACCESS AND TO SET MATRIX ELEMENTS                                    */
/* -------------------------------------------------------------------------------- */

/* Matrices are stored as a single contiguous block in ROW-MAJOR order. Basically
   this means that as you step through adjacent memory locations, the first dimension's
   index (--> line index) varies most slowly and the last dimension's index (for
   a 2-D matrix the column index) varies most quickly.

   Let's regard a 3 x 4 matrix:
                    (0,0)  (0,1)  (0,2)  (0,3)
                    (1,0)  (1,1)  (1,2)  (1,3)
                    (2,0)  (2,1)  (2,2)  (2,3)
                    
   Hence, the line index 'i' runs from 0...2 and the column index 'j' runs
   from 0...3. In the memory, the matrix is stored as:
   
   (0,0), (0,1), (0,2), (0,3), (1,0), (1,1), ..., (2,2), (2,3)

   Hence, the element (i,j) has the following 1-D index:

   (i,j) --> j + 4 * i

   Example: (1,2) --> 2 + 4*1 = 6
*/
double  get_element_matrix_double(
                                  /* --------------- */
                                  /* input variables */
                                  /* --------------- */
                                  const matrix_double *matrix,
                                  const long i, const long j
                                 ) 
{
    return matrix->data[j + matrix->dimension_2*i];
}


/******************************************************************************
*				European Southern Observatory
*		 Vlti ViNci Data Reduction Software (VNDRS)
*
* Module name:	
* Input/Output:	See function arguments to avoid duplication
* Description:	
*
*
* History:		
* 01-May-01		(csabet) Revised
******************************************************************************/
void set_element_matrix_double(
	matrix_double	*matrix,	/*	IO:	*/
	const long		i, 			/*	In:	*/
	const long		j,			/*	In:	*/
	const double	value)		/*	In:	*/
{
	matrix->data[j + matrix->dimension_2*i] = value;
	return;
}



complex get_element_matrix_complex(
                                   /* --------------- */
                                   /* input variables */
                                   /* --------------- */
                                   const matrix_complex *matrix,
                                   const long i, const long j
                                  )
{
    return matrix->data[j + matrix->dimension_2*i];
}

void    set_element_matrix_complex(
                                   /* --------------- */
                                   /* input variables */
                                   /* --------------- */  
                                   matrix_complex *matrix,
                                   const long i, const long j,
                                   const complex value
                                  )
{
    matrix->data[j + matrix->dimension_2*i] = value;
    return;
}

void vndrs_matrix_sub(
                      /* -------------------------------------------- */
                      /* input and output variable (will be replaced) */
                      /* -------------------------------------------- */  
                      matrix_double *dest,
                      /* -------------- */
                      /* input variable */
                      /* -------------- */  
                      const matrix_double *op
                     )
/* The 2 real-valued matrices 'dest' and 'op' (same dimension!) are added.
   The result is written to the matrix 'dest' (IN-PLACE ADDING!) */
{
    long i;
    
    for(i=0; i < ((dest->dimension_1)*(dest->dimension_2)); i++)
        {
            dest->data[i] -= op->data[i];
        }
    return;
}


void vndrs_matrix_add(
                      /* -------------------------------------------- */
                      /* input and output variable (will be replaced) */
                      /* -------------------------------------------- */  
                      matrix_double *dest,
                      /* -------------- */
                      /* input variable */
                      /* -------------- */  
                      const matrix_double *op
                     )
/* The 2 real-valued matrices 'dest' and 'op' (same dimension!) are added.
   The result is written to the matrix 'dest' (IN-PLACE ADDING!) */
{
    long i;
    
    for(i=0; i < ((dest->dimension_1)*(dest->dimension_2)); i++)
        {
            dest->data[i] += op->data[i];
        }
    return;
}

void vndrs_matrix_scale(
                        /* -------------------------------------------- */
                        /* input and output variable (will be replaced) */
                        /* -------------------------------------------- */
                        matrix_double *dest,
                        /* -------------- */
                        /* input variable */
                        /* -------------- */                         
                        const double val
                       )
/* The real-valued matrix 'dest' is multiplied with the real-valued constant
   'val'. The matrix 'dest' is replaced by its scaled version! (IN-PLACE TRANSFORM!) */
{
    /* local variables */
    long i;

    for(i=0; i < ((dest->dimension_1)*(dest->dimension_2)); i++)
    {
        dest->data[i] *= val;
    }
    return;
}


void vndrs_matrix_sqrt(
                        /* -------------------------------------------- */
                        /* input and output variable (will be replaced) */
                        /* -------------------------------------------- */
                        matrix_double *dest
                       )
/* Square root of matrix dest (IN-PLACE TRANSFORM!) Check before that there are no
negative values ! */
{
    /* local variables */
    long i;

    for(i=0; i < ((dest->dimension_1)*(dest->dimension_2)); i++)
        if(dest->data[i] >= 0.0)
            dest->data[i] = sqrt(dest->data[i]);
        else
            dest->data[i] = 0.0;

    return;
}


void vndrs_matrix_add_constant(
                      /* -------------------------------------------- */
                      /* input and output variable (will be replaced) */
                      /* -------------------------------------------- */  
                      matrix_double *dest,
                      /* -------------- */
                      /* input variable */
                      /* -------------- */  
                      const double val
                     )
{
    long i;
    
    for(i=0; i < ((dest->dimension_1)*(dest->dimension_2)); i++)
        {
            dest->data[i] += val;
        }
    return;
}

/*
    Copy of one matrix to another - pkervell 17-Oct-02
*/
void vndrs_matrix_memcpy(
                      /* ----------------------------------- */
                      /*  output variable (will be replaced) */
                      /* ----------------------------------- */  
                      matrix_double *dest,
                      /* -------------- */
                      /* input variable */
                      /* -------------- */  
                      const matrix_double *op
                     )
{
    long i;
    
    for(i=0; i < ((dest->dimension_1)*(dest->dimension_2)); i++)
            dest->data[i] = op->data[i];

    return;
}

/*
    Returns the local maximum of a matrix - pkervell 21-Jan-04
*/
double vndrs_matrix_local_max(
                        /* -------------- */
                        /* input variable */
                        /* -------------- */
                        const matrix_double *matrix,
                        const long I_min, const long I_max,
                        const long J_min, const long J_max
                       )
/* This routine returns the maximum component of a real-valued matrix,
between I_min < i < I_max and J_min < j < J_max */
{
    double maximum, current_val;
    long i,j;
    
    /* If there is an inversion of the limits */
    if(I_min > I_max)
    {
        fprintf(stderr, "Problem in the computation of a matrix local maximum (x direction).\n");
        vndrs_error_alert();
    }
    if(J_min > J_max)
    {
        fprintf(stderr, "Problem in the computation of a matrix local maximum (y direction).\n");
        vndrs_error_alert();
    }
    
    maximum = get_element_matrix_double(matrix, 0, 0);
  
    for(i=0; i<matrix->dimension_1; i++)
        for(j=0; j<matrix->dimension_2; j++)
        {
            current_val = get_element_matrix_double(matrix, i, j);
            if(current_val > maximum) maximum = current_val;
        }
    return maximum;
}


/* -------------------------------------------------------------------------------- */
/* ROUTINES ACTING ON VECTORS                                                       */
/* -------------------------------------------------------------------------------- */

void vndrs_vector_fill(
                        /* -------------------------------------------- */
                        /* input and output variable (will be replaced) */
                        /* -------------------------------------------- */
                        vector_double *dest,
                        /* -------------- */
                        /* input variable */
                        /* -------------- */                         
                        const double val
                       )
/* All components of the vector 'dest' are set to the constant value 'val' */
{
    long i;

    for(i=0; i<dest->dimension; i++)
    {
        dest->data[i] = val;
    }
    return;
}

void vndrs_vector_scale(
                        /* -------------------------------------------- */
                        /* input and output variable (will be replaced) */
                        /* -------------------------------------------- */
                        vector_double *dest,
                        /* -------------- */
                        /* input variable */
                        /* -------------- */                         
                        const double val
                       )
/* The real-valued vector 'dest' is multiplied with the real-valued constant
   'val'. The vector 'dest' is replaced by its scaled version! (IN-PLACE TRANSFORM!) */
{
    /* local variables */
    long i;

    for(i=0; i<dest->dimension; i++)
    {
        dest->data[i] *= val;
    }
    return;
}

void vndrs_vector_square(
                        /* -------------------------------------------- */
                        /* input and output variable (will be replaced) */
                        /* -------------------------------------------- */
                        vector_double *dest
                       )
/* The real-valued vector 'dest' is multiplied by itself (IN-PLACE TRANSFORM!) */
{
    /* local variables */
    long i;

    for(i=0; i<dest->dimension; i++)
        dest->data[i] = dest->data[i] * dest->data[i];

    return;
}

void vndrs_vector_complex_scale(
                                /* -------------------------------------------- */
                                /* input and output variable (will be replaced) */
                                /* -------------------------------------------- */
                                vector_complex *dest,
                                /* -------------- */
                                /* input variable */
                                /* -------------- */
                                const double val
                               )
/* The complex-valued vector 'dest' is multiplied with the real-valued constant
   'val'. The vector 'dest' is replaced by its scaled version! (IN-PLACE TRANSFORM!) */
{
    /* local variables */
    long i;

    for(i=0; i<dest->dimension; i++)
    {
        dest->data[i] = vndrs_real_complex_mul(val, dest->data[i]);
    }
    return;
}

void vndrs_vector_add(
                      /* -------------------------------------------- */
                      /* input and output variable (will be replaced) */
                      /* -------------------------------------------- */  
                      vector_double *dest,
                      /* -------------- */
                      /* input variable */
                      /* -------------- */  
                      const vector_double *op
                     )
/* The 2 real-valued vectors 'dest' and 'op' (same dimension!) are added.
   The result is written to the vector 'dest' (IN-PLACE ADDING!) */
{
    long i;

    #ifdef SECURITY_MODE
    /* Check if the 2 vectors share the same dimension */
    if((dest->dimension)!=(op->dimension))
    {
        fprintf(stderr, "Error in routine 'vndrs_vector_add'.\n");
        fprintf(stderr, "The 2 vectors to be added must have the\n");
        fprintf(stderr, "same dimension! You tried to add 2 vectors\n");
        fprintf(stderr, "having the dimensions %ld and %ld.\n", op->dimension, dest->dimension);
			/* csabet 07-Aug-2001. Format precision corrected */
        vndrs_error_alert();
    }
    #endif
    
    for(i=0; i < dest->dimension; i++)
    {
        dest->data[i] += op->data[i];
    }
    return;
}


void vndrs_vector_complex_add(
                              /* -------------------------------------------- */
                              /* input and output variable (will be replaced) */
                              /* -------------------------------------------- */
                              vector_complex *dest,
                              /* -------------- */
                              /* input variable */
                              /* -------------- */ 
                              const vector_complex *op
                             )
/* The 2 complex vectors 'dest' and 'op' (same dimension!) are added.
   The result is written to the vector 'dest'. (IN PLACE ADDING!) */
{
    long i;

    #ifdef SECURITY_MODE
    /* Check if the 3 vectors share the same dimension */
    if((dest->dimension)!=(op->dimension))
    {
        fprintf(stderr, "Error in routine 'vndrs_vector_complex_add'.\n");
        fprintf(stderr, "The 2 vectors to be added must have the\n");
        fprintf(stderr, "same dimension! You tried to add 2 vectors\n");
        fprintf(stderr, "having the dimensions %ld and %ld.\n", dest->dimension, op->dimension);
			/* csabet 07-Aug-2001. Format precision corrected */
        vndrs_error_alert();
    }
    #endif
    
    for(i=0; i < dest->dimension; i++)
    {
        dest->data[i] = vndrs_add_complex(dest->data[i], op->data[i]);
    }
    return;
}

void vndrs_vector_sub(
                      /* -------------------------------------------- */
                      /* input and output variable (will be replaced) */
                      /* -------------------------------------------- */
                      vector_double *dest,
                      /* -------------- */
                      /* input variable */
                      /* -------------- */  
                      const vector_double *op
                     )
/* The 2 real-valued vectors 'dest' and 'op' (same dimension!) are subtracted.
   The result 'dest-op' is written to the vector 'dest'. (IN-PLACE PROCESS!) */
{
    long i;

    #ifdef SECURITY_MODE
    /* Check if the 2 vectors share the same dimension */
    if((dest->dimension)!=(op->dimension))
    {
        fprintf(stderr, "Error in routine 'vndrs_vector_sub'.\n");
        fprintf(stderr, "The 2 vectors to be subtracted must have the\n");
        fprintf(stderr, "same dimension! You tried to subtract 2 vectors\n");
        fprintf(stderr, "having the dimensions %ld and %ld.\n", dest->dimension, op->dimension);
			/* csabet 07-Aug-2001. Format precision corrected */
        vndrs_error_alert();
    }
    #endif
    
    for(i=0; i < dest->dimension; i++)
    {
        dest->data[i] -= op->data[i];
    }
    return;
}


void vndrs_vector_complex_sub(
                              /* -------------------------------------------- */
                              /* input and output variable (will be replaced) */
                              /* -------------------------------------------- */
                              vector_complex *dest,
                              /* -------------- */
                              /* input variable */
                              /* -------------- */
                              const vector_complex *op
                             )
/* The 2 complex vectors 'dest' and 'op' (same dimension!) are subtracted ('dest-op').
   The result 'dest-op' is written to the vector 'dest'. */
{
    long i;

    #ifdef SECURITY_MODE
    /* Check if the 2 vectors share the same dimension */
    if((dest->dimension)!=(op->dimension))
    {
        fprintf(stderr, "Error in routine 'vndrs_vector_complex_sub'.\n");
        fprintf(stderr, "The 2 vectors to be subtracted must have the\n");
        fprintf(stderr, "same dimension! You tried to add 2 vectors\n");
        fprintf(stderr, "having the dimensions %ld and %ld.\n", dest->dimension, op->dimension);
			/* csabet 07-Aug-2001. Format precision corrected */
        vndrs_error_alert();
    }
    #endif
    
    for(i=0; i < dest->dimension; i++)
    {
        dest->data[i] = vndrs_sub_complex(dest->data[i], op->data[i]);
    }
    return;
}

void vndrs_vector_mul(
                      /* -------------------------------------------- */
                      /* input and output variable (will be replaced) */
                      /* -------------------------------------------- */
                      vector_double *dest,
                      /* -------------- */
                      /* input variable */
                      /* -------------- */  
                      const vector_double *op
                     )
/* The 2 vectors 'dest' and 'op' are multiplied COMPONENT-WISE (!!).
   This is NOT a scalar product and NOT a vector product!
   Make sure that the 2 vectors share the same dimension!
   The result is written to the vector 'dest' (IN-PLACE!) */
{
    long i;

    #ifdef SECURITY_MODE
    /* Check if the 2 vectors share the same dimension */
    if((dest->dimension)!=(op->dimension))
    {
        fprintf(stderr, "Error in routine 'vndrs_vector_mul'.\n");
        fprintf(stderr, "The 2 vectors to be multiplied must have the\n");
        fprintf(stderr, "same dimension! You tried to multiply 2 vectors\n");
        fprintf(stderr, "having the dimensions %ld and %ld.\n", dest->dimension, op->dimension);
			/* csabet 07-Aug-2001. Format precision corrected */
        vndrs_error_alert();
    }
    #endif
    
    for(i=0; i < dest->dimension; i++)
    {
        dest->data[i] *= op->data[i];
    }
    return;
}

void vndrs_vector_complex_mul(
                              /* -------------------------------------------- */
                              /* input and output variable (will be replaced) */
                              /* -------------------------------------------- */
                              vector_complex *dest,
                              /* -------------- */
                              /* input variable */
                              /* -------------- */
                              const vector_complex *op
                             )
/* The 2 complex vectors 'dest' and 'op' are multiplied COMPONENT-WISE (!!).
   This is NOT a scalar product and NOT a vector product!
   Make sure that the 2 vectors share the same dimension!
   The result is written to the vector 'dest' (IN-PLACE!) */
{
    long i;

    #ifdef SECURITY_MODE
    /* Check if the 2 vectors share the same dimension */
    if((dest->dimension)!=(op->dimension))
    {
        fprintf(stderr, "Error in routine 'vndrs_vector_complex_mul'.\n");
        fprintf(stderr, "The 2 vectors to be multiplied must have the\n");
        fprintf(stderr, "same dimension! You tried to multiply 2 vectors\n");
        fprintf(stderr, "having the dimensions %ld and %ld.\n", dest->dimension, op->dimension);
			/* csabet 07-Aug-2001. Format precision corrected */
        vndrs_error_alert();
    }
    #endif
    
    for(i=0; i < dest->dimension; i++)
    {
        dest->data[i] = vndrs_mul_complex(dest->data[i], op->data[i]);
    }
    return;
}

void vndrs_vector_complex_double_mul(
                                     /* -------------------------------------------- */
                                     /* input and output variable (will be replaced) */
                                     /* -------------------------------------------- */
                                     vector_complex *dest,
                                     /* -------------- */
                                     /* input variable */
                                     /* -------------- */
                                     const vector_double *op
                                    )
/* The 2 vectors 'dest' (complex) and 'op' are multiplied COMPONENT-WISE (!!).
   Make sure that both vectors share the same dimension!
   The result is written to the vector 'dest' (IN-PLACE!) */
{
    long i;

    /* Check if the 2 vectors share the same dimension */
    if((dest->dimension)!=(op->dimension))
    {
        fprintf(stderr, "Error in routine 'vndrs_vector_complex_double_mul '.\n");
        fprintf(stderr, "The 2 vectors to be multiplied must have the\n");
        fprintf(stderr, "same dimension! You tried to multiply 2 vectors\n");
        fprintf(stderr, "having the dimensions %ld and %ld.\n", dest->dimension, op->dimension);
			/* csabet 07-Aug-2001. Format precision corrected */
        vndrs_error_alert();
    }
    
    for(i=0; i < dest->dimension; i++)
    {
        dest->data[i] = vndrs_real_complex_mul(op->data[i], dest->data[i]);
    }

    return;
}

void vndrs_vector_div(
                      /* -------------------------------------------- */
                      /* input and output variable (will be replaced) */
                      /* -------------------------------------------- */
                      vector_double *dest,
                      /* -------------- */
                      /* input variable */
                      /* -------------- */  
                      const vector_double *op
                     )
/* The 2 vectors 'dest' and 'op' are divided COMPONENT-WISE (!!).
   Make sure that the 2 vectors share the same dimension!
   The result is written to the vector 'dest' (IN-PLACE!) */
{
    long i;

    #ifdef SECURITY_MODE
    /* Check if the 2 vectors share the same dimension */
    if((dest->dimension)!=(op->dimension))
    {
        fprintf(stderr, "Error in routine 'vndrs_vector_div'.\n");
        fprintf(stderr, "The 2 vectors to be divided must have the\n");
        fprintf(stderr, "same dimension! You tried to divide 2 vectors\n");
        fprintf(stderr, "having the dimensions %ld and %ld.\n", dest->dimension, op->dimension);
			/* csabet 07-Aug-2001. Format precision corrected */
        vndrs_error_alert();
    }
    #endif
    
    for(i=0; i < dest->dimension; i++)
    {
    /* And what happens when there is a division by zero ?? pkervell - 23-apr-02 */
    /* if(op->data[i] < 0.000001) printf(" Problem of division by zero !! \n");*/
        dest->data[i] = dest->data[i] / op->data[i];
    }
    return;
}

void vndrs_vector_complex_div(
                              /* -------------------------------------------- */
                              /* input and output variable (will be replaced) */
                              /* -------------------------------------------- */
                              vector_complex *dest,
                              /* -------------- */
                              /* input variable */
                              /* -------------- */              
                              const vector_complex *op
                             )
/* The 2 complex vectors 'dest' and 'op' are divided COMPONENT-WISE (!!).
   Make sure that the 2 vectors share the same dimension!
   The result is written to the vector 'dest' (IN-PLACE!) */
{
    long i;

    #ifdef SECURITY_MODE
    /* Check if the 2 vectors share the same dimension */
    if((dest->dimension)!=(op->dimension))
    {
        fprintf(stderr, "Error in routine 'vndrs_vector_complex_div'.\n");
        fprintf(stderr, "The 2 vectors to be divided must have the\n");
        fprintf(stderr, "same dimension! You tried to divide 2 vectors\n");
        fprintf(stderr, "having the dimensions %ld and %ld.\n", dest->dimension, op->dimension);
			/* csabet 07-Aug-2001. Format precision corrected */
        vndrs_error_alert();
    }
    #endif
    
    for(i=0; i < dest->dimension; i++)
    {
        dest->data[i] = vndrs_div_complex(dest->data[i], op->data[i]);
    }
    return;
}

void vndrs_vector_add_constant(
                               /* -------------------------------------------- */
                               /* input and output variable (will be replaced) */
                               /* -------------------------------------------- */
                               vector_double *dest,
                               /* -------------- */
                               /* input variable */
                               /* -------------- */    
                               const double val
                              )
/* The constant (real-valued) 'val' is added to each component of the
   vector 'dest'. Within the routine, the vector 'dest' is replaced
   by the new vector 'dest'. */
{
    long i;

    for(i=0; i<dest->dimension; i++)
    {
        dest->data[i] += val;
    }
    return;
}

void vndrs_vector_sqrt(
                       /* -------------------------------------------- */
                       /* input and output variable (will be replaced) */
                       /* -------------------------------------------- */
                       vector_double *dest
                      )
/* The routine replaces the vector 'dest' by a vector whose components
   are obtained by taking the sqrt of the components of 'dest'.
   (IN-PLACE TRANSFORM!!) */
{
    long i;

    for(i=0; i<dest->dimension; i++)
    {
        if(dest->data[i] >= 0.0)
            dest->data[i] = sqrt(dest->data[i]);
        else
            dest->data[i] = 0.0; /* added pkervell - 01-Oct-02 to avoid the crashes with negative values */
            /* it is important to check the vector before taking the sqrt */           
    }
    return;
}

void vndrs_vector_trim_positive(
                                /* -------------------------------------------- */
                                /* input and output variable (will be replaced) */
                                /* -------------------------------------------- */
                                vector_double *dest
                               )
/* Negative components of 'dest' are replaced by zeros. The vector is replaced
   by its 'trimmed' version within this routine! */
{
    long i;

    for(i=0; i<dest->dimension; i++)
    {
        if(dest->data[i]<0)
	{
            dest->data[i] = 0.;
	}
    }
    return;
} 

double vndrs_vector_min(
                        /* -------------- */
                        /* input variable */
                        /* -------------- */
                        const vector_double *src
                       )
/* This routine returns the minimum component of a real-valued vector. */
{
    double minimum;
    long i;

    minimum = src->data[0];
  
    for(i=1; i<src->dimension; i++) /* start at 2nd index! */
    {
        if(src->data[i] < minimum) minimum = src->data[i];
    }
    return minimum;
}

double vndrs_vector_max(
                        /* -------------- */
                        /* input variable */
                        /* -------------- */
                        const vector_double *src
                       )
/* This routine returns the maximum component of a real-valued vector */
{
    double maximum;
    long i;
    
    maximum = src->data[0];
  
    for(i=1; i<src->dimension; i++) /* start at 2nd index! */
    {
        if(src->data[i] > maximum) maximum = src->data[i];
    }
    return maximum;
}

double vndrs_vector_min_center(
                        /* -------------- */
                        /* input variable */
                        /* -------------- */
                        const vector_double *src,
                        long width
                       )
/* This routine returns the minimum component of a real-valued vector over a length of "width". */
{
    double minimum;
    long i, index_min, index_max;
    
    if(width<src->dimension)
    {
        index_min = src->dimension/2 - width/2;
        index_max = src->dimension/2 + width/2;
    }
    else
    {
        index_min = 0;
        index_max = src->dimension;
    }
    
    minimum = src->data[index_min];
  
    for(i=index_min; i<index_max; i++)
    {
        if(src->data[i] < minimum) minimum = src->data[i];
    }
    return minimum;
}

double vndrs_vector_max_center(
                        /* -------------- */
                        /* input variable */
                        /* -------------- */
                        const vector_double *src,
                        long width
                       )
/* This routine returns the maximum component of a real-valued vector */
{
    double maximum;
    long i, index_min, index_max;
    
    if(width<src->dimension)
    {
        index_min = src->dimension/2 - width/2;
        index_max = src->dimension/2 + width/2;
    }
    else
    {
        index_min = 0;
        index_max = src->dimension;
    }
    
    maximum = src->data[index_min];

    for(i=index_min; i<index_max; i++)
        {
        if(src->data[i] > maximum) maximum = src->data[i];
    }
    return maximum;
}

/* Copying the content of a batch to a 2nd batch */ 
void vndrs_batch_memcpy( 
                         /* -------------------------------------------- */ 
                         /* input and output variable (will be replaced) */ 
                         /* -------------------------------------------- */ 
                         batch *dest, 
                         /* -------------- */ 
                         /* input variable */ 
                         /* -------------- */ 
                         const batch *src 
                        )
{
    long N_scans, I;
    
    N_scans = src->number_of_scans;
    
    for(I=0;I<N_scans;I++)
    {
        vndrs_vector_memcpy(dest->fullscans[I]->I1, src->fullscans[I]->I1);
        vndrs_vector_memcpy(dest->fullscans[I]->I2, src->fullscans[I]->I2);
        vndrs_vector_memcpy(dest->fullscans[I]->PA, src->fullscans[I]->PA);
        vndrs_vector_memcpy(dest->fullscans[I]->PB, src->fullscans[I]->PB);
        vndrs_vector_memcpy(dest->fullscans[I]->QL, src->fullscans[I]->QL);
    }
    
}


void vndrs_vector_memcpy(
                         /* -------------------------------------------- */
                         /* input and output variable (will be replaced) */
                         /* -------------------------------------------- */
                         vector_double *dest,
                         /* -------------- */
                         /* input variable */
                         /* -------------- */
                         const vector_double *src
                        )
/* This vector copies the contents of the vector 'src' to the vector 'dest'.
   Make sure that memory for BOTH vectors has been allocated before calling
   this function! */
{
    long i;
    
    #ifdef SECURITY_MODE
    /* Check if the 2 vectors share the same dimension */
    if((dest->dimension)!=(src->dimension))
    {
        fprintf(stderr, "Error in routine 'vndrs_vector_memcpy'.\n");
        fprintf(stderr, "The 2 vectors must have the same dimension!\n");
        fprintf(stderr, "You tried to copy 2 vectors\n");
        fprintf(stderr, "having the dimensions %ld and %ld.\n", dest->dimension, src->dimension);
			/* csabet 07-Aug-2001. Format precision corrected */
        vndrs_error_alert();
    }
    #endif
    
    for(i=0; i<src->dimension; i++)
    {
        dest->data[i] = src->data[i];
    }    
    return;
}

void vndrs_vector_complex_memcpy(
                                 /* -------------------------------------------- */
                                 /* input and output variable (will be replaced) */
                                 /* -------------------------------------------- */
                                 vector_complex *dest,
                                 /* -------------- */
                                 /* input variable */
                                 /* -------------- */
                                 const vector_complex *src
                                )
/* This vector copies the contents of the complex vector 'src' to the complex vector 'dest'.
   Make sure that memory for BOTH vectors has been allocated before calling
   this function! */
{
    long i;

    #ifdef SECURITY_MODE
    /* Check if the 2 vectors share the same dimension */
    if((dest->dimension)!=(src->dimension))
    {
        fprintf(stderr, "Error in routine 'vndrs_vector_complex_memcpy'.\n");
        fprintf(stderr, "The 2 vectors must have the same dimension!\n");
        fprintf(stderr, "You tried to copy 2 vectors\n");
        fprintf(stderr, "having the dimensions %ld and %ld.\n", dest->dimension, src->dimension); 
                /* csabet 07-Aug-2001. Format precision corrected */
        vndrs_error_alert();
    }
    #endif

    for(i=0; i<src->dimension; i++)
    {
        dest->data[i] = src->data[i];
    }    
    return;
}

void vndrs_make_complex_vector_modulus_phase(
                                             /* --------------- */
                                             /* input variables */
                                             /* --------------- */
                                             const vector_double *modulus,
                                             const vector_double *phase,
                                             /* ---------------- */
                                             /* output variables */
                                             /* ---------------- */
                                             vector_complex *dest
                                            )
/* For 2 given vectors holding the modulus and phase of complex numbers, this routine fills a
   third vector ('dest') with complex numbers whose modulus and phase correspond to the components
   of the 2 input vectors.

   Make sure that memory for all vectors has been allocated before calling this function! */
{
    long i;

    for(i=0; i < dest->dimension; i++)
        dest->data[i] = complex_of_modulus_phase(modulus->data[i], phase->data[i]);

    return;
}

/* Extract real/imaginary parts of a complex vector */ 
void  vndrs_retrieve_re_im( 
                             /* --------------- */ 
                             /* input variables */ 
                             /* --------------- */ 
                             const vector_complex *src, 
                             /* ---------------- */ 
                             /* output variables */ 
                             /* ---------------- */ 
                             vector_double *real, 
                             vector_double *imaginary 
                            )
{
    long i;
   
    for(i=0; i < src->dimension; i++)
    {
        real->data[i] = src->data[i].re;
        imaginary->data[i] = src->data[i].im;
    }
    
    return;
}

void  vndrs_retrieve_modulus(
                             /* --------------- */
                             /* input variables */
                             /* --------------- */
                             const vector_complex *src,
                             /* ---------------- */
                             /* output variables */
                             /* ---------------- */
                             vector_double *modulus
                            )
/* For a given complex vector 'src', this function extracts the modulus of the
   vector components and stores them in a new vector 'modulus' (real-valued),
   which is returned as output parameter.

   Make sure that memory for all vectors has been allocated before calling
   this function! */
{
    long i;
   
    for(i=0; i < src->dimension; i++)
        modulus->data[i] = vndrs_abs_complex(src->data[i]);
    
    return;
}

void  vndrs_retrieve_phase(
                           /* --------------- */
                           /* input variables */
                           /* --------------- */
                           const vector_complex *src,
                           /* ---------------- */
                           /* output variables */
                           /* ---------------- */
                           vector_double *phase
                          )
/* For a given complex vector 'src', this function extracts the phase of the
   vector components and stores them in a new vector 'phase' (real-valued),
   which is returned as output parameter.

   Make sure that memory for all vectors has been allocated before calling
   this function! */
{
    long i;
    
    for(i=0; i < src->dimension; i++)
        phase->data[i] = vndrs_arg_complex(src->data[i]);
    
    return;
}


/* -------------------------------------------------------------------------------- */
/* ROUTINES FOR SORTING VECTORS                                                     */
/* -------------------------------------------------------------------------------- */

void vndrs_sort_shellsmethod(
                             /* -------------------------------------------- */
                             /* input and output variable (will be replaced) */
                             /* -------------------------------------------- */  
                             vector_double *Vector
                            )
/* The vector is sorted in ascending order. The vector is replaced by its
   sorted counterpart, --> IN-PLACE transform! Make sure that memory has
   been allocated for the vector before calling this function! 
   A 'Straight Insertion Method (Shell's Method)' is used. This method requires
   small list dimensions to be effective (< 50). */
{
    /* Local variables */
    long i, j;
    double Help;    

    /* Main part */
    for(i=1; i<Vector->dimension; i++) /* Loop starts at 2nd position in vector */
    {
        /* Pick vector element @ position 'i': */
        Help = Vector->data[i];

        /* Set index 'j' to the position left of 'i' */
        j = i-1;
        while((j>=0)&&(Vector->data[j]>Help))
	/* look for the place to insert element 'i' (between 0 ... i-1) */
        {
            Vector->data[j+1] = Vector->data[j];
            j--;
        }
        Vector->data[j+1] = Help;
    }    

    return;
}



/* -------------------------------------------------------------------------------- */
/* ROUTINES FOR HANDLING COMPLEX NUMBERS                                            */
/* -------------------------------------------------------------------------------- */

complex complex_of_modulus_phase(
                                 /* --------------- */
                                 /* input variables */
                                 /* --------------- */
                                 double modulus, double phase
                                )
/* returns a complex number with the specified modulus and phase.
   The phase is specified in units of radian! */
{
    complex c;
    
    c.re = modulus * cos(phase);
    c.im = modulus * sin(phase);
    
    return c;
}

double  vndrs_abs_complex(
                          /* --------------- */
                          /* input variables */
                          /* --------------- */
                          const complex z
                         )
/* returns the absolute value of a complex number z */
{
    double x,y,ans /*,temp*/;

    /* x = fabs(z.re);
    y = fabs(z.im);*/
    
    /* Why all these complications ?? This does not look very clear to me...
    changed pkervell 28-Oct-02 */
    x = z.re;
    y = z.im;
    ans = sqrt(x*x + y*y);
    
    
    /* if(x==0.0)
    {
        ans = y;
    }
    else if(y==0.0)
    {
        ans = x;
    }
    else if(x>y)
    {
        temp = y/x;
        ans = x*sqrt(1.0+temp*temp);
    }
    else
    {
        temp = x/y;
        ans = y*sqrt(1.0+temp*temp);
    }*/
    
    return ans;
}


double  vndrs_arg_complex(
                          /* --------------- */
                          /* input variables */
                          /* --------------- */
                          complex z
                         )
/* calculates the phase (in rad) of the complex number z.
   It uses the ANSI C-function "atan2" which uses the signs of
   its arguments to compute the quadrant of the return value. */
/* The result is between 0 and 2Pi !! */
{
    double Phase;
    /* result of the function */
    
    /* modification for the SUN SOLARIS cc compiler */
    if(fabs(z.re) < 1.E-300)
    {
        z.re = 1.E-300;
    }
    
    Phase = atan2(z.im, z.re);

    /* The atan2 function returns values that are between -Pi and +Pi.
       In the following, it is made sure that the result is between
       0 and 2*Pi.
       Removed: -pi -> pi phase is better - pkervell 25-Oct-02 */
    
    /*if(Phase<0.)
    {
        Phase=Phase+TWO_PI;
    }*/

    return Phase;
}

complex vndrs_real_complex_mul(
                               /* --------------- */
                               /* input variables */
                               /* --------------- */
                               const double real_value,
                               const complex complex_value
                              )
/* calculates the product of the real number 'real_value' and the complex
   number 'complex_value' */
{
    /* local variables */
    complex x;

    x.re = real_value * complex_value.re;
    x.im = real_value * complex_value.im;
    
    return x;
}

complex vndrs_add_complex(
                          /* --------------- */
                          /* input variables */
                          /* --------------- */
                          const complex a,
                          const complex b
                         )
/* This function adds the 2 complex numbers 'a' and 'b' and
   returns the sum as a complex number. */
{
    complex c;

    c.re = a.re + b.re;
    c.im = a.im + b.im;

    return c;
}

complex vndrs_sub_complex(
                          /* --------------- */
                          /* input variables */
                          /* --------------- */
                          const complex a,
                          const complex b
                         )
/* This function subtracts the 2 complex numbers 'a' and 'b'
   and returns the difference 'a-b' as a complex number. */
{
    complex c;

    c.re = a.re - b.re;
    c.im = a.im - b.im;

    return c;
}

complex vndrs_mul_complex(
                          /* --------------- */
                          /* input variables */
                          /* --------------- */
                          const complex a,
                          const complex b
                         )
/* Returns the complex product a*b */
{
    complex c;
    
    c.re = a.re*b.re - a.im*b.im;
    c.im = a.im*b.re + a.re*b.im;
    
    return c;
}

complex vndrs_div_complex(
                          /* --------------- */
                          /* input variables */
                          /* --------------- */
                          const complex a,
                          const complex b
                         )
/* Returns the complex quotient a/b */
{
    complex c;
    double r, den;

    if(fabs(b.re) >= fabs(b.im))
    {
        r = b.im/b.re;
        den = b.re + r*b.im;
        c.re = (a.re + r*a.im)/den;
        c.im = (a.im - r*a.re)/den;
    }
    else
    {
        r = b.re/b.im;
        den = b.im + r*b.re;
        c.re = (a.re*r + a.im)/den;
        c.im = (a.im*r - a.re)/den;
    }
    return c;
}


/* Bessel function J1(x), inspired from Numerical Recipes - pkervell 04-Nov-02 */

double vndrs_bessel_j1(double x)
{
	double ax,z;
	double xx,y,ans,ans1,ans2;

	if ((ax=fabs(x)) < 8.0) {
		y=x*x;
		ans1=x*(72362614232.0+y*(-7895059235.0+y*(242396853.1
			+y*(-2972611.439+y*(15704.48260+y*(-30.16036606))))));
		ans2=144725228442.0+y*(2300535178.0+y*(18583304.74
			+y*(99447.43394+y*(376.9991397+y*1.0))));
		ans=ans1/ans2;
	} else {
		z=8.0/ax;
		y=z*z;
		xx=ax-2.356194491;
		ans1=1.0+y*(0.183105e-2+y*(-0.3516396496e-4
			+y*(0.2457520174e-5+y*(-0.240337019e-6))));
		ans2=0.04687499995+y*(-0.2002690873e-3
			+y*(0.8449199096e-5+y*(-0.88228987e-6
			+y*0.105787412e-6)));
		ans=sqrt(0.636619772/ax)*(cos(xx)*ans1-z*sin(xx)*ans2);
		if (x < 0.0) ans = -ans;
	}
        
	return ans;
}

