/**********************************************************
***********************************************************
*	      	European Southern Observatory
*	  VLTI ViNci Data Reduction Software (VNDRS)
*
* Module name:  vndrs_Piston.c	
* Description:	Select centered and non-pistonned data
*
* History:		
* 18-Jul-2002  Damien Segransan
* 27-Jul-2002  pkervell - comments and compilation at Paranal
* 04-Aug-2002  pkervell - revision and memory check
* 29-Aug-2002  pkervell - modified to send back the width of
*              the peak in the wavelets transform.
               THERE SEEMS TO BE A MEMORY LEAK (hidden...)
* 25-Sep-2002  pkervell - removed nrerror call
* 27-Sep-2002  pkervell - complete rewriting (very bad original code!)
***********************************************************
**********************************************************/

/**********************************************************
*	Include files
**********************************************************/
#include "vndrs.h"
#include "vndrs_Globalvar.h"
#include <math.h>
#include "lmfit.h"
#include <stdio.h>

/**********************************************************
*	Constant definitions
**********************************************************/
#define MA 3 /* Number of parameters to fit. */
#define MB 6 /* Number of parameters to fit. */

/**********************************************************
*	Functions
**********************************************************/

/*
  Expand in storage the covariance matrix covar so as to take 
  into account parameters that are being held fixed 
  (For the latter, return 0 covariance)
*/

void covsrt(matrix_double *covar,
	    long ma_pk,
	    vector_double *ia,
	    long mfit)
{
    long i,j,k;
    double buffer;
    
    for (i=mfit;i<ma_pk;i++)
	for (j=0;j<i;j++)
        {
            set_element_matrix_double(covar, i, j, 0.0);
            set_element_matrix_double(covar, j, i, 0.0);
        }
        
    k=mfit;
    for (j=(ma_pk-1);j>=0;j--)
	{
	if (ia->data[j])
	    {
            k--;
	    for (i=0;i<ma_pk;i++)
                {
                buffer = get_element_matrix_double(covar, i, k);
                set_element_matrix_double(covar, i, k, get_element_matrix_double(covar, i, j));
		set_element_matrix_double(covar, i, j, buffer);
                }
            for (i=0;i<ma_pk;i++)
                {
                buffer = get_element_matrix_double(covar, k, i);
                set_element_matrix_double(covar, k, i, get_element_matrix_double(covar, j, i));
		set_element_matrix_double(covar, j, i, buffer);
                }
            }
	}
}

/*----------------------------------------------------------------------*/

/* Swaps the lines i and j of the input matrix */
void vndrs_swap_lines(matrix_double *Matrix, double i, double j)
{
    long J, N_columns;
    double buffer_1, buffer_2;
    
    N_columns = Matrix->dimension_2;
    
    for(J=0;J<N_columns;J++)
    {
        buffer_1 = get_element_matrix_double(Matrix, i, J);
        buffer_2 = get_element_matrix_double(Matrix, j, J);
        set_element_matrix_double(Matrix, i, J, buffer_2);
        set_element_matrix_double(Matrix, j, J, buffer_1);
    }
}

/* Gauss-Jordan matrix inversion */

void vndrs_gaussj(matrix_double *a,
	    long n,
	    matrix_double *b,
	    long m)
    
/*
  Linear equation solution by Gauss-Jordan elimination,
  nr eq 2.1.1. 
  
  INPUT : 
  a[1..n][1..n] is the input matrix. 
  b[1..n][1..m] is input containing the m right-hand side vector. 
  
  OUTPUT :
  a is replaced by its matrix inverse
  b is replaced by the corresponding set of solution vectors
*/
    
{
    vector_double *ipiv, *buffer_b;
    long i,icol,irow,j,k,l,ll, I;
    double pivinv, buffer, buffer_2, big;
    matrix_double *index_track;
    
    index_track = vndrs_calloc_matrix_double(n,1);
    ipiv = vndrs_calloc_vector_double(n);
    buffer_b = vndrs_calloc_vector_double(n);

    for(I=0;I<n;I++) 	index_track->data[I]=I;
    
    for (j=0;j<n;j++) ipiv->data[j]= j;

    for (i=0;i<n;i++)
    {

        /* printf("\n Matrix a before swap:\n");
        for(I=0;I<8;I++)
        {
        printf(" %+5f, %+5f, %+5f, %+5f, %+5f, %+5f, %+5f, %+5f \n",
        get_element_matrix_double(a, I, 0),
        get_element_matrix_double(a, I, 1),
        get_element_matrix_double(a, I, 2),
        get_element_matrix_double(a, I, 3),
        get_element_matrix_double(a, I, 4),
        get_element_matrix_double(a, I, 5),
        get_element_matrix_double(a, I, 6),
        get_element_matrix_double(a, I, 7));
        }*/
        
        
        big = 0.0;
        /*---*/
        for (j=i;j<n;j++)
            {
                if (fabs(get_element_matrix_double(a, j, i)) > fabs(get_element_matrix_double(a, i, i)))
                ipiv->data[i] = j;
            }
                
        /* for(I=0;I<8;I++) printf(" I: %i, ipiv : %i\n",I,(int)ipiv->data[I]);*/
        /*---*/

        vndrs_swap_lines(a, i, (int)ipiv->data[i]);
        vndrs_swap_lines(b, i, (int)ipiv->data[i]);
        vndrs_swap_lines(index_track, i, (int)ipiv->data[i]);
        
        /* printf("\n Matrix a after swap:\n");
        for(I=0;I<6;I++)
        {
        printf(" %+5f, %+5f, %+5f, %+5f, %+5f, %+5f \n",
        get_element_matrix_double(a, I, 0),
        get_element_matrix_double(a, I, 1),
        get_element_matrix_double(a, I, 2),
        get_element_matrix_double(a, I, 3),
        get_element_matrix_double(a, I, 4),
        get_element_matrix_double(a, I, 5));
        }*/

        /*printf("\n Vector b:\n");
        printf(" %+5f, %+5f, %+5f, %+5f, %+5f, %+5f, %+5f, %+5f \n",
        get_element_matrix_double(b, 0, 0),
        get_element_matrix_double(b, 1, 0),
        get_element_matrix_double(b, 2, 0),
        get_element_matrix_double(b, 3, 0),
        get_element_matrix_double(b, 4, 0),
        get_element_matrix_double(b, 5, 0),
        get_element_matrix_double(b, 6, 0),
        get_element_matrix_double(b, 7, 0));*/


        if (get_element_matrix_double(a, i, i) == 0.0)
            {printf(" Error: Singular Matrix-2\n");
            vndrs_error_alert(); }
        
        pivinv=1.0/get_element_matrix_double(a, i, i);

	for (l=i;l<n;l++)
        {
            buffer = get_element_matrix_double(a, i, l);
            set_element_matrix_double(a, i, l, buffer * pivinv);
        }
	for (l=i;l<m;l++)
        {
            buffer = get_element_matrix_double(b, i, l);
            set_element_matrix_double(b, i, l, buffer * pivinv);
        }

	set_element_matrix_double(a, i, i, 1.0); /* to be sure... */
	

	for (ll=0;ll<n;ll++)
        if(ll != i)
        {
            buffer_2 = get_element_matrix_double(a, ll, i);
            for (l=0;l<n;l++)
            {
                buffer = get_element_matrix_double(a, ll, l)-
                        get_element_matrix_double(a, i, l) * buffer_2;
                set_element_matrix_double(a, ll, l, buffer);
            }
            for (l=0;l<m;l++)
            {
                buffer = get_element_matrix_double(b, ll, l)-
                        get_element_matrix_double(b, i, l)* buffer_2;
                set_element_matrix_double(b, ll, l, buffer);
            }                
        }
        
        /* printf("\n Matrix a after sub:\n");
        for(I=0;I<8;I++)
        {
        printf(" %+5f, %+5f, %+5f, %+5f, %+5f, %+5f, %+5f, %+5f \n",
        get_element_matrix_double(a, I, 0),
        get_element_matrix_double(a, I, 1),
        get_element_matrix_double(a, I, 2),
        get_element_matrix_double(a, I, 3),
        get_element_matrix_double(a, I, 4),
        get_element_matrix_double(a, I, 5),
        get_element_matrix_double(a, I, 6),
        get_element_matrix_double(a, I, 7));
        }*/
   
    }
            
    vndrs_free_vector_double(ipiv);
    vndrs_free_vector_double(buffer_b);
    vndrs_free_matrix_double(index_track);
}



/*----------------------------------------------------------------------*/

void where2(matrix_double *arr,
	    vector_double *naxes,
	    double *mxx,
	    long *xx,
	    long *yy,
	    long minx,
	    long maxx,
	    long miny,
	    long maxy)
{
    long ii,jj;
    minx = (minx>=0 & minx<maxx)*minx;
    maxx =  maxx*(maxx>=0 & maxx<naxes->data[0]) +(!((maxx>=0 & maxx<naxes->data[0])))*naxes->data[0];
    miny = (miny>=0 & miny<maxy)*miny;
    maxy =  maxy*(maxy>=0 & maxy<naxes->data[1]) +(!((maxy>=0 & maxy<naxes->data[1])))*naxes->data[1];
    
    for (ii = minx; ii < maxx; ii++)
	{
	for (jj = miny; jj < maxy; jj++)
	    {
	    if (get_element_matrix_double(arr, ii, jj) > *mxx)
		{
		*mxx=get_element_matrix_double(arr, ii, jj);
		*xx=ii;
		*yy=jj;
		} 
	    }
	}
    
}


long FWHM_gauss1D(vector_double *arr,
		  long n,
		  double mxx,
		  long minx,
		  long maxx)
{
    
    long ii, fwhm;
    
    minx = (minx>=0 & minx<(maxx-1))*minx  ;
    maxx =  maxx*(maxx>=(minx+5) & maxx<n) +(!((maxx>=(minx+5) & maxx<n)))*n;
    fwhm =0;

    if(maxx>(arr->dimension)) maxx = arr->dimension;
    
    for (ii = minx; ii < maxx; ii++)
	if (arr->data[ii] > mxx/2.) fwhm++;
    
    return fwhm;
}

/*----------------------------------------------------------------------- */

    /*
      INPUTS : 
      wsdI
      n
      gues
      
      OUTPUT :
      a : parametres du vecteur fite
    */

double vndrs_lmfit_gauss2D(vector_double *array2D,
			   vector_double *naxes,
			   vector_double *gues,
			   vector_double *a)
{
    
    /* Local variables */
    long i, j, k, mfit=6, nfree, itmax, niter, ntots, index;
    vector_double *la, *naxes_s;
    long originX,originY;
    double lambda,chisq,ochisq;
    vector_double *tmp2, *x, *y, *sig;
    matrix_double *covar, *alpha;
    double conv, tol, sigma, chi2_first;
    double norm;
    
    /* Memory allocation */
    tmp2 = vndrs_calloc_vector_double(6);
    la = vndrs_calloc_vector_double(6); /*= 1 ou 0 en fonction des parametres a fitter ou a garder constant*/
    naxes_s = vndrs_calloc_vector_double(2);

    naxes_s->data[0]=(long)(naxes->data[0]/4);
    /*naxes_s->data[1]=(long)(128);  /* changed pkervell 25-Sep-02 */
    naxes_s->data[1]=(long)(naxes->data[1]/2);
    ntots = naxes_s->data[0]*naxes_s->data[1]; /*fit 2D sur une petite partie de la surface*/
    originX = (long)(a->data[1]-naxes_s->data[0]/2);
    originY = (long)(a->data[3]-naxes_s->data[1]/2);

    /* if(originX<0) originX = 0;
    if(originY<0) originY = 0;

    if((naxes_s->data[0]+originX)>naxes->data[0]) originX = naxes->data[0]-naxes_s->data[0];
    if((naxes_s->data[1]+originY)>naxes->data[1]) originY = naxes->data[1]-naxes_s->data[1];*/

    x = vndrs_calloc_vector_double(ntots);
    y = vndrs_calloc_vector_double(ntots);
    sig = vndrs_calloc_vector_double(ntots);
    covar = vndrs_calloc_matrix_double(6, 6);/* covariance matrix */
    alpha = vndrs_calloc_matrix_double(6, 6);/* curvature matrix */
        
    tol=1e-4; /* TEST - changed from 1e-7 to accelerate the processing - pkervell - 09-Sep-02 */
    itmax=10; /* changed from 15 - pkervell */
    /* fprintf(stderr,"guess a[1]=%8.1f, a[2]=%8.1f;   a[3]=%6.2f, a[4]=%6.2f;   a[5]=%6.2f\n",a->data[1],a->data[2],a->data[3],a->data[4],a->data[5]);
    fprintf(stderr,"naxes_s[0]=%i, naxes_s[1]=%i\n", (int)naxes_s->data[0], (int)naxes_s->data[1]);
    fprintf(stderr,"origin =%i,%i \n", (int)originX, (int)originY);*/
    

    k=0;
    norm=0.0;
    for (i=0;i<naxes_s->data[0];i++)
	{ 
	for (j=0;j<naxes_s->data[1];j++)
	    {
	    index =(i+originX) + naxes->data[0]*(j+originY);
	    x->data[k]=(double)index;
	    y->data[k]=(double)array2D->data[index];
	    sig->data[k]=(double)(sqrt(fabs(y->data[k]))+1.0);
	    norm += y->data[k]*y->data[k]/(sig->data[k]*sig->data[k]);
	    k++;
	    }
	}

    /* The norm is ok */
    /*printf(" k = %d\n",k);
    for(i=0;i<20;i++)
    {
        printf("x[%d] = %f\n", i, x->data[i]);
        printf("y[%d] = %f\n", i, y->data[i]);    
        printf("sig[%d] = %f\n", i, sig->data[i]);
    }*/
    for (i=0;i<mfit;i++) la->data[i]=1;
    for (i=0;i<6;i++)
    {
        a->data[i]=gues->data[i];
        /*printf(" a->data[%d]=%f\n", i, a->data[i]);*/
    }
    lambda = -1;
    mrqmin2D(x, y, sig, ntots, a, la, covar, alpha, &chisq, &lambda, 10., naxes);
    niter=0;
    chi2_first = chisq;


    /* Removed to speed up the processing, and also because there is a SEGMENTATION FAULT and a MEMORY LEAK
    inside this piece of code !! */

#if 0    
    while(1) /* This part of the code apparently does NOT change anything in the results... */
    {
	ochisq=chisq;
	mrqmin2D(x, y, sig, ntots, a, la, covar, alpha, &chisq, &lambda, 10., naxes);
        printf(" Crashes here before 1... \n");
        printf(" Crashes here after 2... \n");
	niter++;
	conv = 2.0 * fabs(chisq-ochisq) / (chisq+ochisq);
	if (conv <= tol) {  fprintf(stderr,"convergence ok \n",conv); break;}
	if (lambda > 1e10) {  fprintf(stderr,"Lambda too high !!!\n",itmax);  break;}
	if (niter >= itmax) { fprintf(stderr,"reached maximum number of iterations (%i).\n \n",itmax);  break;}
    }
#endif
    
    /* dernier appel de mrqmin2D avec alambda = 0  pour 
    obtenir les bonnes valeurs de la matrice de covariance et de la matrice de courbure */
    lambda=0.0;
    mrqmin2D(x, y, sig, ntots, a, la, covar, alpha, &chisq, &lambda, 10., naxes);
    
    nfree = ntots - 6; /*attention, n est + petit lorsque weight=0*/
    if((chisq > 0) && (nfree >= 0)) sigma= sqrt((double)(nfree)/chisq); else sigma = 0.0;

    vndrs_free_vector_double(tmp2);
    vndrs_free_vector_double(la);
    vndrs_free_vector_double(naxes_s);
    vndrs_free_vector_double(x);
    vndrs_free_vector_double(y);
    vndrs_free_vector_double(sig);
    vndrs_free_matrix_double(covar);/* covariance matrix */
    vndrs_free_matrix_double(alpha);/* curvature matrix */
    
    if (conv <= tol)   /* changed from 1e-9 - pkervell - 09-Sep-02 */
	return chisq/norm;
    else
	return 1.0;
    
}

/*-------------------------------------------------------------------------------------------*/


/*----------------------------------------------------------------------- */

void vndrs_gauss2D(double x,
		   vector_double *a,
		   double *y,
		   vector_double *dyda,
		   long na,
		   vector_double *naxes)
{
    /*REM : 
      x=1..512  et a[2]=256 => a[3]=5..100. raisonnable
      x=1..1024 et a[2]=512 => a[3]=5..200. raisonnable
      FHWM = 2*a->data[2]*sqrt(log(2))
      a->data[2] =  FHWM/(2*sqrt(log(2)))
    */
    
    /* Local variables */
    double ex,arg_a, arg_b, arg_a2, arg_b2, arg2, fac_a, fac_b;
    long xx,yy,i;
    
    xx = (long)((long)x % (long)naxes->data[0]);
    yy = (long)(x / (long)naxes->data[0]);
    
    arg_a  = (xx - a->data[1]) / a->data[2]  ;
    arg_a2 = -arg_a*arg_a;
    arg_b  = (yy - a->data[3]) / a->data[4]  ;
    arg_b2 = -arg_b*arg_b;
    arg2 = arg_a2+arg_b2;
    ex   =  exp(arg2);
    fac_a  =  a->data[0] * ex*2*arg_a;
    fac_b  =  a->data[0] * ex*2*arg_b;
    
    *y   =  a->data[0] * ex + a->data[5];
    
    
    /* fprintf(stderr,"index=%3i=%3 i +naxes[0] %3i; expo(%3i,%3i)=%12.6e \n",(long)x,xx,yy,xx,yy,*y);*/
    
    dyda->data[0] = ex;
    dyda->data[1] = fac_a / a->data[2];
    dyda->data[2] = fac_a*arg_a / a->data[2];
    dyda->data[3] = fac_b / a->data[4];
    dyda->data[4] = fac_b*arg_b / a->data[4];
    dyda->data[5] = 1.0;
    
}


/*----------------------------------------------------------------------- */

/*----------------------------------------------------------------------*/

void mrqcof2D(vector_double *x,
	      vector_double *y,
	      vector_double *sig,
	      long ndata,
	      vector_double *a,
	      vector_double *ia,
	      matrix_double *alpha,
	      vector_double *beta,
	      double *chisq,
	      vector_double *naxes)

/*
  Evaluate the linearized fitting matrix alpha and vector beta (nr 15.5.8)
  and calculate chi2
  
  beta_k  =  -.5 d(chi2)/da_k
  alpha_kl =  -.5 d2(chi2)/da_k/da_l
  sum_l alpha_kl deltaa_l = beta_k

  lm algo. on remplace les termes diag de alpha par alpha*(1+lambda)
  et on calcule chi2
*/

{
    long i, j, k, l, m, ma_pk = 6, mfit = 0;
    double ymod, wt, sig2i, dy, buffer, buffer_2 = 0.0;
    vector_double *dyda;
    
    dyda = vndrs_calloc_vector_double(ma_pk);
    for (j=0;j<ma_pk;j++)
	if (ia->data[j]) mfit++;
    /* printf(" alpha->dim = %d, %d\n",alpha->dimension_1, alpha->dimension_2);
    printf(" beta->dim = %d\n",beta->dimension);
    printf(" m_fit = %d\n", mfit);*/
    for (j=0;j<mfit;j++)
    {/*initialize symmetric alpha and beta*/
        /* for(k=0;k<j;k++) set_element_matrix_double(alpha, j, k, 0.0);*/
        beta->data[j]=0.0;
    }

    *chisq=0.0;
    /* *prob=1.0;*/
    
    /*printf(" ndata = %d\n", ndata);*/
    for (i=0;i<ndata;i++)
	{/*summation loop over all data*/
        vndrs_gauss2D(x->data[i], a, &ymod, dyda, ma_pk, naxes);
        /* printf(" ymod = %f\n",ymod); /* ymod est ok la premiere fois */
        if(sig->data[i] != 0.0)
            sig2i = 1.0/(sig->data[i]*sig->data[i]);
	dy = y->data[i]-ymod;
        /* printf(" dy, sig2i = %f, %f\n",dy,sig2i); dy et sig2i sont ok */
	for (j=0,l=1;l<=ma_pk;l++)
	    {
            wt=dyda->data[l-1]*sig2i;
            /*printf(" wt= %f\n",wt); /* wt est ok */
            for (j++,k=0,m=1;m<=l;m++)
                {
                buffer = get_element_matrix_double(alpha, j-1, ++k-1);
                set_element_matrix_double(alpha, j-1, k-1, buffer + wt*dyda->data[m-1]);
                }
            /*printf(" beta->data[%d]= %f\n",j,beta->data[j]); /* beta est ok */
            beta->data[j-1] += dy*wt;
	    }
	 buffer_2 += dy*dy*sig2i; /*find chi2*/
	}
    *chisq = buffer;
    for (j=1;j<mfit;j++) /*fill in the symmetric side*/
	for (k=0;k<(j-1);k++) set_element_matrix_double(alpha, k, j, get_element_matrix_double(alpha, j, k));
        
    /*for(i=0;i<mfit;i++)
            printf("  sig[%d] = %f\n", i, sig->data[i]);
    for(i=0;i<mfit;i++)
            printf("  dyda[%d] = %f\n", i, dyda->data[i]);
    for(i=0;i<mfit;i++)
        for(j=0;j<mfit;j++)
            printf(" alpha[%d][%d] = %f\n", i, j, get_element_matrix_double(alpha,i,j));*/
            
    vndrs_free_vector_double(dyda);
}


/*----------------------------------------------------------------------*/
/* Memory management is VERY DUBIOUS HERE !! */
void mrqmin2D(vector_double *x,
	      vector_double *y,
	      vector_double *sig,
	      long ndata,
	      vector_double *a,
	      vector_double *ia,
	      matrix_double *covar,
	      matrix_double *alpha,
	      double *chisq,
	      double *lambda,
	      double gain,
	      vector_double *naxes)

{
    long j, k, l, m, ma_pk = 6;
    long mfit;
    double ochisq;
    vector_double *a_try, *beta, *da, *gamma;
    matrix_double *oneda;
    double buffer;
    
    /* Initialisation and memory allocation */
    a_try = vndrs_calloc_vector_double(ma_pk); /*new guess for a*/
    beta  = vndrs_calloc_vector_double(ma_pk); /*   */
    da    = vndrs_calloc_vector_double(ma_pk); /*increment for a*/
    gamma = vndrs_calloc_vector_double(ma_pk); 
    /* mfit, nombre de parametres a fitter parmi les ma_pk parametres
    de la fonction */
    for (mfit=0,j=0;j<ma_pk;j++) if(ia->data[j]) mfit++;
    /* oneda = vndrs_calloc_matrix_double(mfit,1);/*  */
    oneda = vndrs_calloc_matrix_double(mfit,1);/*  */
    
    if (*lambda < 0.0)
    { 
	/* Initialization de a_try */
	for(k=0;k<ma_pk;k++)
	    {
	    a_try->data[k] = a->data[k];
	    da->data[k]=0.0;
	    }
	/* initialisation de lambda */
	*lambda=0.001;       
         
	/* compute alpha, beta et chi2. a n'est pas modifie */
        /* This call to mrqcof2D crashes the program... The problem is there...*/
	mrqcof2D(x, y, sig, ndata, a, ia, alpha, beta, chisq, naxes);
	ochisq = (*chisq);
    }
    
    /* Alter linearized fitting matrix alpha  by augmenting diagonal elements */
    for (j=0,l=0;l<ma_pk;l++)
    {
	    for(k=0, m=0;m<ma_pk;m++)
            {
                set_element_matrix_double(covar, j, k, get_element_matrix_double(alpha, j, k));
                k++;
            }
            if(get_element_matrix_double(alpha, j, j) > 0.0)
                gamma->data[j]=1./sqrt(get_element_matrix_double(alpha, j, j));
            else
                gamma->data[j]=0.;
	    set_element_matrix_double(covar, j, j, get_element_matrix_double(alpha, j, j)*(1.0+(*lambda)));
	    set_element_matrix_double(oneda, j, 1, beta->data[j]);
            j++;
    }
    
    
    for (j=0, l=0;l<ma_pk;l++)
    {
            for (k=0, m=0;m<ma_pk;m++)
            {
                    buffer = gamma->data[j]*gamma->data[k]*get_element_matrix_double(alpha, j, k);
                    set_element_matrix_double(covar, j, k, buffer);
                    k++;
            }
            
            buffer = gamma->data[j]*gamma->data[j]*get_element_matrix_double(alpha, j, j)*(1.0+(*lambda));
            set_element_matrix_double(covar, j, j, buffer);
            set_element_matrix_double(oneda, j, 1, gamma->data[j]*beta->data[j]);
            j++;
    }

    vndrs_gaussj(covar, mfit, oneda, 1);
    /*output covar = alpha^-1 and X = ondeda*/
	
    for (j=0;j<mfit;j++) da->data[j]=gamma->data[j]*get_element_matrix_double(oneda, j, 1);
	
    /* Once converged, evaluate covariance matrix */
    if (*lambda == 0.0)
    {
        covsrt(covar, ma_pk, ia, mfit);
        covsrt(alpha, ma_pk, ia, mfit);
        for (j=0;j<mfit;j++)
        {
            buffer = get_element_matrix_double(covar, j, j) * gamma->data[j]*gamma->data[j];
            set_element_matrix_double(covar, j, j, buffer);
        }

        vndrs_free_matrix_double(oneda);
        vndrs_free_vector_double(da);
        vndrs_free_vector_double(beta);
        vndrs_free_vector_double(a_try);
        vndrs_free_vector_double(gamma);  /* added pkervell 29-Jul-02 */
        return;
    }
 	
    /*Did the try succeed??*/
    for (j=0,l=0;l<ma_pk;l++)
        a_try->data[l] = a->data[l] + da->data[j++];

    mrqcof2D(x, y, sig, ndata, a_try, ia, covar, da, chisq, naxes);	
	
    if (*chisq < ochisq)
    {/*Success, accept the new solution*/
        /*                fprintf(stderr,"success\n");*/
        *lambda /= gain;
        ochisq=(*chisq);
        for (j=0, l=0;l<ma_pk;l++)
        {
            for (k=0, m=0;m<ma_pk;m++)
            {
                set_element_matrix_double(alpha, j, k, get_element_matrix_double(covar, j, k));
                k++;
            }

            beta->data[j]=da->data[j];
            a->data[l]=a_try->data[l];
            j++;
        }
    } else
    {/*Failure, increase alambda and return*/
     /*                fprintf(stderr,"\n\nfailure for lambda = %e",*lambda);*/
        *lambda *= gain;
        *chisq = ochisq;
    }
    
    vndrs_free_matrix_double(oneda);
    vndrs_free_vector_double(da);
    vndrs_free_vector_double(beta);
    vndrs_free_vector_double(a_try);
    vndrs_free_vector_double(gamma);  /* added pkervell 29-Jul-02 */	
}




/*----------------------------------------------------------------------- */

void vndrs_wave_select(matrix_double *wave_power_spectrum, vector_double *res)

/* This function is intended to select the scans that suffer from piston effect.
   It computes the residual of the fit of a gaussian on the peak in the wavelets spectrum.
   This value can then be used as a criteria for the selection of the scans.
   written segransan xx-Jul-02, revised pkervell 09-Aug-02
   Now the function returns a vector to pass the results */
{
    /* Declaration of variables */
    vector_double *wsdIx, *wsdIy;
    vector_double *array2D;
    vector_double *sigx, *sigy;
    vector_double *x,*y;
    vector_double *gues1, *a, *gues2, *b,*gues3, *c, *naxes;

    double mxx, buffer;
    long xx,yy;
    long i,j;

    /* Memory allocation */
    naxes = vndrs_calloc_vector_double(2);
    /* Size of the power spectrum (in X and Y), and allocation of corresponding 1D vector */
    naxes->data[0] = wave_power_spectrum->dimension_1;
    naxes->data[1] = wave_power_spectrum->dimension_2;
    array2D  = vndrs_calloc_vector_double(naxes->data[0]*naxes->data[1]);

    /* Constitution of a copy of the input 2D wavelets power spectrum */
    for (i = 0; i <  naxes->data[0]; i++)
	for (j = 0; j < naxes->data[1]; j++)
        {
            array2D->data[(long)(i+j*naxes->data[0])] = 
                        get_element_matrix_double(wave_power_spectrum,i,j);
        }
    
    /* remark: the variable 3 is defined at the beginning of the code to be equal to 3 (number of
       parameters to fit. */
    gues1 = vndrs_calloc_vector_double(3);    /* First guesses */
    gues2 = vndrs_calloc_vector_double(3);    /* First guesses */
    gues3 = vndrs_calloc_vector_double(6);
    a = vndrs_calloc_vector_double(3);
    b = vndrs_calloc_vector_double(3);
    c = vndrs_calloc_vector_double(6);   
    x = vndrs_calloc_vector_double(naxes->data[0]);
    wsdIx = vndrs_calloc_vector_double(naxes->data[0]);
    sigx  = vndrs_calloc_vector_double(naxes->data[0]);
    y     = vndrs_calloc_vector_double(naxes->data[1]);
    wsdIy = vndrs_calloc_vector_double(naxes->data[1]);
    sigy  = vndrs_calloc_vector_double(naxes->data[1]);


    /* Search for the maximum of the wavelets spectrum and its position. */
    
    mxx=0.0;    /* Peak value */
    xx=0;       /* X position of max */
    yy=0;       /* Y position of max */

    where2(wave_power_spectrum, naxes, &mxx, &xx, &yy,
            10, naxes->data[0]-10, 20, naxes->data[1]-20 );
     
    for (i=0;i<naxes->data[0];i++)
	wsdIx->data[i] = get_element_matrix_double(wave_power_spectrum, i, yy);
	/* wsdIx is a cut through the WPSD (wavelet power spectral density) along the X axis
	   where the maximum has been found. */

    for (j=0;j<naxes->data[1];j++)
	wsdIy->data[j] = get_element_matrix_double(wave_power_spectrum, xx, j);
	/* wsdIx is a cut through the WPSD (wavelet power spectral density) along the Y axis
	   where the maximum has been found. */

    /* for(i=0;i<naxes->data[1];i++)
        printf(" wsdIx->data[%d] = %f \n", i, wsdIx->data[i]);
    for(i=0;i<naxes->data[1];i++)
        printf(" wsdIy->data[%d] = %f \n", i, wsdIy->data[i]); look ok */

    /* fprintf(stderr,"MAX(im[%i,%i]=%f\n",xx,yy,mxx);*/
    /* Definition of the initial values for the Gaussian fit */
    gues3->data[0] = mxx;
    gues3->data[1] = xx;
    gues3->data[2] = FWHM_gauss1D(wsdIx, naxes->data[0], mxx, 10, naxes->data[0]-10) / (2*sqrt(log(2)));
    gues3->data[3] = yy;
    gues3->data[4] = FWHM_gauss1D(wsdIy, naxes->data[1], mxx, 10, naxes->data[1]-10) / (2*sqrt(log(2)));
    gues3->data[5] = mxx/10.;
    /* for(i=0;i<6;i++)
        printf(" gues3->data[%d] = %f \n", i, gues3->data[i]);*/

    c->data[0] = mxx;
    c->data[1] = xx;
    c->data[2] = FWHM_gauss1D(wsdIx, naxes->data[0], mxx, 10, naxes->data[0]-10) / (2*sqrt(log(2)));
    c->data[3] = yy;
    c->data[4] = FWHM_gauss1D(wsdIy, naxes->data[1], mxx, 23, naxes->data[1]-20) / (2*sqrt(log(2)));
    c->data[5] = mxx/10.;

    /* fprintf(stderr,"avt fit amp = %8.1f;  X0=%8.1f;   A_X0=%6.2f;  Y0=%6.2f;   A_Y0=%6.2f; BG=%6.2f;  \n", c->data[1], c->data[2], c->data[3], c->data[4], c->data[5], c->data[6]);*/

    /* OK up to now !! :-) */

    /*
       Core of the function:
       =====================
       The residual of the Gaussian fit is computed starting
       from the initial guess values computed above.
    */
    /* This call causes segmentation faults - pkervell 25-Sep-02 */
    /* res->data[0] = vndrs_lmfit_gauss2D(array2D, naxes, gues3, c );*/
    res->data[0] = 0.0; /* for TEST - pkervell */
    res->data[1] = c->data[0]; /* amplitude of the fit */
    res->data[2] = c->data[2] * (2.0*sqrt(log(2.0))); /* fwhm in time domain */
    res->data[3] = c->data[3]; /* position of the peak in frequency domain */
    res->data[4] = c->data[4] * (2.0*sqrt(log(2.0))); /* fwhm in frequency domain */
    res->data[5] = c->data[5]; /* background */

    /* Deallocation of the memory. */
    vndrs_free_vector_double(c);
    vndrs_free_vector_double(gues3);
    vndrs_free_vector_double(gues1);
    vndrs_free_vector_double(gues2);
    vndrs_free_vector_double(a);
    vndrs_free_vector_double(b);
    vndrs_free_vector_double(x);
    vndrs_free_vector_double(wsdIx);
    vndrs_free_vector_double(sigx);
    vndrs_free_vector_double(y);
    vndrs_free_vector_double(wsdIy);
    vndrs_free_vector_double(sigy);
    vndrs_free_vector_double(array2D);
    vndrs_free_vector_double(naxes);


    /* The function returns the numerical value of the residual of the Gaussian fit.*/
}

/*----------------------------------------------------------------------- */

