/*******************************************************************/
/* VLTI VINCI DATA REDUCTION SOFTWARE (VNDRS)                      */
/*                                                                 */
/* FILE: vndrs_SignalProcessing.c                                  */
/*                                                                 */
/* Contains the routines for signal processing                     */
/*                                                                 */
/* Written by Markus Schoeller & Rainer Wilhelm, ESO, 2000         */
/* Modified by Pierre Kervella, Feb 2002                           */
/*                                                                 */
/*******************************************************************/ 

#include "vndrs.h"


/* ----------------------------------------------------------------- */
/* FAST FOURIER TRANSFORMS (BASED ON FFTW LIBRARY)                   */
/* ----------------------------------------------------------------- */

void vndrs_fft_real_2_complex(
                              /* --------------- */
                              /* input variables */
                              /* --------------- */
                              const vector_double *Signal_Real,
                              /* ---------------- */
                              /* output variables */
                              /* ---------------- */
                              vector_complex *Spectrum
                             )
/* For a given REAL-valued signal 'Signal_Real', this function computes the FFT of
   this signal. The Fourier transform is a COMPLEX-valued spectrum and it is stored
   in the complex vector 'Spectrum'.
   
   Make sure, that for both vectors memory has been allocated before calling this
   function!! */ 
{
    /* LOCAL VARIABLES */
    /* --------------- */
    long dimension; /* the dimension of all vectors */
    vector_complex *Signal; /* the 'complex version' of the REAL-valued signal */
    
    /* MAIN PART */ 
    /* --------- */
    /* Get dimension */
    dimension = Signal_Real->dimension;

    /* Allocate memory for the 'intermediate' complex signal vector */
    Signal = vndrs_calloc_vector_complex(dimension);
    
    /* Fill 'intermediate' complex signal vector */
    vndrs_vector_double_2_vector_complex(Signal_Real, Signal);

    /* Perform the complex-to-complex FFT */
    vndrs_fft(Signal, Spectrum);

    /* Free memory for the 'intermediate' complex signal vector */
    vndrs_free_vector_complex(Signal); 
    
    return;
}

void vndrs_ifft_complex_2_real(
                               /* --------------- */
                               /* input variables */
                               /* --------------- */                              
                               const vector_complex *Spectrum,
                               /* ---------------- */
                               /* output variables */
                               /* ---------------- */
                               vector_double *Signal_Real
                              )
/* For a given HERMETIAN spectrum 'Spectrum', this function computes the inverse FFT of
   this spectrum. The inverse Fourier transform is a REAL-valued spectrum and it is stored
   in the real-valued vector 'Signal_Real'.
   
   Make sure, that for both vectors memory has been allocated before calling this
   function!! */ 
{
    /* LOCAL VARIABLES */
    /* --------------- */
    long dimension; /* the dimension of all vectors */
    vector_complex *Signal; /* the 'complex version' of the REAL-valued output signal */
    
    /* MAIN PART */ 
    /* --------- */
    /* Get dimension */
    dimension = Spectrum->dimension;

    /* Allocate memory for the 'intermediate' complex signal vector
       (will be the output of the inverse FFT) */
    Signal = vndrs_calloc_vector_complex(dimension);

    /* Perform the complex-to-complex FFT */
    vndrs_ifft(Spectrum, Signal);
    
    /* Convert 'intermediate' complex signal to a REAL-valued signal */
    /* WARNING: we are using this routine, which returns the real part
                instead of the modulus; otherwise negative values are
                not represented correctly, but are multiplied by -1 */
    vndrs_vector_complex_2_vector_double_real(Signal, Signal_Real);

    /* Free memory for the 'intermediate' complex signal vector */
    vndrs_free_vector_complex(Signal); 
    
    return;
}


void vndrs_fft(
               /* --------------- */
               /* input variables */
               /* --------------- */   	       
               const vector_complex *Signal,
               /* ---------------- */
               /* output variables */
               /* ---------------- */                 
	       vector_complex *Spectrum
              )
/* This function computes the Fourier spectrum of the complex vector 'Signal'.
   The complex spectrum is returned as output vector.
   The resulting fourierspectrum has its origin (frequency zero) at pixel 0.
   Make sure that memory for both, the input and the output signal has been
   allocated before calling this function.
*/
{
    /* LOCAL VARIABLES */
    /* --------------- */
    fftw_plan Plan;
    /* plan for the FFTW (see FFTW manual, p. 22). Has to be generated BEFORE calling the
       FFTW kernel */
    long dimension;
    /* dimension of the signal to be transformed */

    /* MAIN PART */
    /* --------- */
    /* SET DIMENSION OF SIGNAL AND ITS TRANSFORM */
    dimension = Signal->dimension;
    
    /* CREATE THE PLAN FOR THE FFT */
    Plan = fftw_create_plan(dimension, FFTW_FORWARD, FFTW_ESTIMATE);
    /* Parameters (compare FFTW manual, p. 23):
       ::::::::::::::::::::::::::::::::::::::::
         FFTW_FORWARD  --> Perform the FFT in forward direction
         FFTW_ESTIMATE --> Provide a "reasonable" plan (for a RISC processor
	                   with many registers).
	                   An alternative would be 'FFTW_MEASURE' which tells
			   FFTW to find the optimal plan by actually computing
			   several FFTs and measuring their execution time.
    */ 
    
    /* PERFORM THE FOURIER TRANSFORM */
    fftw_one(Plan, Signal->data, Spectrum->data);

    /* NORMALIZE SPECTRUM (DIVIDE IT BY 'Dimension') */
    vndrs_vector_complex_scale(Spectrum, 1./((double)dimension)); 
    
    /* DESTROY THE PLAN */
    fftw_destroy_plan(Plan);   
    
    return;
}



/*
    Additional function to provide easy access to the inverse FFT.
*/
void vndrs_ifft(
                /* --------------- */
                /* input variables */
                /* --------------- */                
                const vector_complex *Spectrum,
                /* ---------------- */
                /* output variables */
                /* ---------------- */                   
                vector_complex *Signal
               )
/* This function computes the inverse Fourier transform of the spectrum 'Spectrum'.
   The input and the output vectors are both complex-valued vectors.
   The Fourier Spectrum has its origin (frequency zero) at pixel 0.
   Make sure that memory for both, input and output vector has been allocated before
   calling this function.
*/
{
    /* LOCAL VARIABLES */
    /* --------------- */
    fftw_plan Plan;
    /* plan for the FFTW (see FFTW manual, p. 22). Has to be generated BEFORE calling the
       FFTW kernel */
    
    long dimension;
    /* dimension of the spectrum to be back-transformed */
    
    /* MAIN PART */
    /* --------- */
    /* SET DIMENSION OF SIGNAL AND ITS TRANSFORM */
    dimension = (long)(Spectrum->dimension);
 
    /* CREATE THE PLAN FOR THE FFT */
    Plan = fftw_create_plan(dimension, FFTW_BACKWARD, FFTW_ESTIMATE);
    /* Parameters (compare FFTW manual, p. 23):
       ::::::::::::::::::::::::::::::::::::::::
         FFTW_BACKWARD  --> Perform the FFT in backward direction
         FFTW_ESTIMATE --> Provide a "reasonable" plan (for a RISC processor
	                   with many registers).
	                   
                           An alternative would be 'FFTW_MEASURE' which tells
			   FFTW to find the optimal plan by actually computing
			   several FFTs and measuring their execution time. */
    
    /* PERFORM THE FOURIER TRANSFORM */
    fftw_one(Plan, Spectrum->data, Signal->data);

    
    /* NORMALIZE SIGNAL (DIVIDE IT BY 'Dimension') */
    /* vndrs_vector_complex_scale(Signal, 1./((double)dimension)); */
 
    /* DESTROY THE PLAN */
    fftw_destroy_plan(Plan);

    return;
}


/* ----------------------------------------------------------------- */
/* OTHER SIGNAL PROCESSING ROUTINES (POWER SPECTRUM, FILTERING,      */
/* CONVOLUTION, NOISE)                                               */
/* ----------------------------------------------------------------- */

/*  pkervell - 03-Sep-02
    Smoothing of a vector_double by a moving average of a given half_width:
    if half_width = 2, the average is over I-2, I-1, I, I+1, I+2
    MEMORY ALLOCATED INSIDE FUNCTION FOR OUTPUT
*/
vector_double *vndrs_vector_smooth(vector_double *input_vector,
                                   long half_width)
{
    long I, J, N_samples;
    vector_double *buffer;
    
    N_samples = input_vector->dimension;
    buffer = vndrs_calloc_vector_double(N_samples);

    /* first value is kept */
    buffer->data[0] = input_vector->data[0];
    
    /* a few first values are only partially smoothed */
    for(I=1;I<half_width;I++)
        for(J=0;J<(2*I+1);J++)
            buffer->data[I] += input_vector->data[J]/(2.0*I+1);
        
    for(I=half_width; I<(N_samples-half_width); I++)
        for(J=(I-half_width);J<(I+half_width+1);J++)
            buffer->data[I] += input_vector->data[J]/(2.0*half_width+1);

    /* few last values are only partially smoothed */
    for(I=(N_samples-half_width);I<N_samples-1;I++)
        for(J=N_samples-1;J>(2*I-N_samples-2);J--)
            buffer->data[I] += input_vector->data[J]/(2.0*(N_samples-I)+1);

    /* last value is kept */
    buffer->data[N_samples-1] = input_vector->data[N_samples-1];

    return(buffer);        
}

/*
    Special version of the apodisation window generator, designed to accept
    parameters more adequate for operation on the fringes themselves and not
    on the whole scan. It computes an apodisation window centered on the
    scan and letting the 'width' central pixels completely through. The transition
    is outside of the 'width' pixels - pkervell 04-Oct-02
*/
vector_double *vndrs_apodisation_center(
                                           /* --------------- */
                                           /* input variables */
                                           /* --------------- */
                                           long Signal_size, /* in number of frames */
                                           long Width,     /* in number of frames */
                                           long Transition /* in number of frames */
                                          )
/* Memory for the output vector (apodisation window) is allocated within this routine! */
{
    long I;
    double Trans_ok, Half_width, value, Current_pos;
    individual_scan *Apodisation_signal;

    if((Width + 2*Transition)>Signal_size | Width <= 0 | Transition < 0)  /* error handling */
    {
        Trans_ok = 0.;
        Half_width = Signal_size/2;
    }
    else /* case ok */
    {
        Trans_ok = Transition;
        Half_width = Width / 2;
    }

    Apodisation_signal = vndrs_calloc_vector_double(Signal_size);

    for (I=0; I<Signal_size; I++)
    {
        Current_pos = 1.0 * fabs(I - Signal_size/2);
        if (Current_pos > (Half_width + Trans_ok))
            Apodisation_signal->data[I] = 0;
        else
        { 
            if (Current_pos > Half_width)
            { 
                value = cos( (PI * 0.5) * (Current_pos - Half_width)/Trans_ok);
                Apodisation_signal->data[I] = value * value;
            } 
            else
                Apodisation_signal->data[I] = 1.0;
        } 
    }

    return Apodisation_signal;
}

/*
    Described in [1], section 2.20
*/
individual_scan *vndrs_apodisation_window (
                                           /* --------------- */
                                           /* input variables */
                                           /* --------------- */
                                           long Signal_size,
                                           double Blockout,
                                           double Transmission
                                          )
/* Memory for the output vector (apodisation window) is allocated within this routine! */
{
    long I;
    double Current, Block, Trans, value, Inv_signal_size, Inv_transmission;
    individual_scan *Apodisation_signal;

    if (Signal_size == 0)
    {
        fprintf(stderr, "The number of Signal_size is 0.\n");
        vndrs_error_alert();
    }
    Inv_signal_size = 1. / Signal_size;

    if (Transmission == 0.)
    {
        /* it is perfectly legal that Transmission is 0,
           but then the formula below is not used anyway */
        Inv_transmission = 1.;
    }
    else
        Inv_transmission = 1. / Transmission;

    Apodisation_signal = vndrs_calloc_vector_double(Signal_size);

    Block = 0.5 - Blockout;
    Trans = Block - Transmission;

    for (I=0; I<Signal_size; I++)
    {
        Current = fabs((double)I * Inv_signal_size - 0.5);

        if (Current >= Block)
            Apodisation_signal->data[I] = 0;
        else
        { 
            /* there is a bug in [1], claiming the next line should be <= */
            if (Current >= Trans)
            { 
                value = (cos((PI * 0.5) * (Current - Trans) * Inv_transmission));
                Apodisation_signal->data[I] = value * value;
            } 
            else
                Apodisation_signal->data[I] = 1;
        } 
    }

    return Apodisation_signal;
}

/*
    Computes the power spectrum of the input signal.
    Described in [1], section 2.13
    MEMORY ALLOCATED INSIDE FUNCTION
 */
power_spectrum *vndrs_power_spectrum(
                                     /* --------------- */
                                     /* input variables */
                                     /* --------------- */
                                     individual_scan *Signal   
                                    )
{
    long N_elements;
    fourier_spectrum *Fourier_signal;
    power_spectrum *Power_signal;

    N_elements = Signal->dimension;

    Fourier_signal = vndrs_calloc_vector_complex(N_elements);
    Power_signal = vndrs_calloc_vector_double(N_elements);

    vndrs_fft_real_2_complex(Signal, Fourier_signal);
    vndrs_retrieve_modulus(Fourier_signal, Power_signal);
    vndrs_vector_mul(Power_signal, Power_signal);

    vndrs_free_vector_complex(Fourier_signal);
    
    return Power_signal;
}

/*
    This procedure puts the center of the fringe packet at the zero position of the vector,
    assuming it is previously at the center. This is necessary to compute the phase properly.
    Dimension has to be a multiple of 2.
    pkervell - 25-Oct-02.
*/
vector_double *vndrs_zero_pos_scan_center(vector_double *Scan)
{
    long N_frames, I;
    vector_double *buffer;

    N_frames = Scan->dimension;
    buffer = vndrs_calloc_vector_double(N_frames);
 
    for(I=0;I<N_frames/2;I++)
        buffer->data[I+N_frames/2] = Scan->data[I];
    for(I=N_frames/2;I<N_frames;I++)
        buffer->data[I-N_frames/2] = Scan->data[I];
        
    vndrs_vector_memcpy(Scan, buffer);
        
    vndrs_free_vector_double(buffer);
}

power_spectrum *vndrs_mirror_half_pow(
                                      /* --------------- */
                                      /* input variables */
                                      /* --------------- */
                                      power_spectrum *half_pow,
                                      long           Dimension_modulo
                                     )
/*
    Routine which copies one side of the powerspectrum to
    the full size. Input spectrum shall have half the size of
    the output spectrum.

    Update on 23.02.2001:
    The second argument Dimension_modulo has a value 0 or 1 for
    a full power spectrum of even or odd dimension. The size
    of the returned full power spectrum is then 2*Half_size+Modulo.
    The extra pixel corresponds to the highest frequencies and is
    at index Half_size in the array. It is left initialised to 0.

    Memory for the full power spectrum is allocated within this
    routine!

 */
{
    long I, J, Dimension_half_pow, Dimension_full_pow;
    power_spectrum *Full_pow;

    /* Get dimension of the 'half' power spectrum */
    Dimension_half_pow = half_pow->dimension;

    /* The dimension of the full power spectrum is twice the dimension of the half power spectrum */
    Dimension_full_pow = Dimension_half_pow*2 + Dimension_modulo;

    /* Allocate memory for the full power spectrum */
    Full_pow = vndrs_calloc_vector_double(Dimension_full_pow);

    /* Set the first value of the full power spectrum to the first value of the half power spectrum */
    Full_pow->data[0] = half_pow->data[0];

    /* Initalises to zero */ 
    for (I=1; I<Dimension_full_pow; I++)  Full_pow->data[I] = 0.; 


    for (I=1, J=Dimension_full_pow-1; I<Dimension_half_pow; I++, J--)
    {
         Full_pow->data[I] = half_pow->data[I];
         Full_pow->data[J] = half_pow->data[I];
    }

    return Full_pow;
}

/*
    Computes the power spectrum of the input signal and
    takes half of it: 0<f<1/2nt (Nyquist frequency).
    Described in [1], section 2.12
 */
power_spectrum *vndrs_one_sided_pow (
                                     /* --------------- */
                                     /* input variables */
                                     /* --------------- */                           
                                     individual_scan *Signal
                                    )
{
    long I, N_elements;
    power_spectrum *Power_signal, *Half_power_signal;

    N_elements = Signal->dimension;

    Half_power_signal = vndrs_calloc_vector_double(N_elements/2);

    Power_signal = vndrs_power_spectrum(Signal);

    for (I=0; I<N_elements/2; I++)
        Half_power_signal->data[I] = Power_signal->data[I];

    vndrs_free_vector_double(Power_signal);

    return Half_power_signal;
}

/*
    Described in [1], section 2.17
*/
individual_scan *vndrs_filter (
                               /* --------------- */
                               /* input variables */
                               /* --------------- */ 
                               individual_scan *Input_signal,
                               power_spectrum *Freq_filter
                              )
/* IMPORTANT:
   Memory for the output variable of type 'individual_scan' is
   allocated inside this routine! */
{
    /* Local variables */
    long N_elements;
    individual_scan *Filtered_signal;
    fourier_spectrum *Fourier_signal;

    /* Get dimension of input signal (=dimension of output signal) */
    N_elements = Input_signal->dimension;

    /* Allocate memory for the Fourier transform of the input signal and
       for the filtered signal */
    Fourier_signal = vndrs_calloc_vector_complex(N_elements);
    Filtered_signal = vndrs_calloc_vector_double(N_elements);
    
    /* Compute Fourier transform of the input signal */
    vndrs_fft_real_2_complex(Input_signal, Fourier_signal);
    
    
    /* Perform the filtering: ==> compute the product of the Fourier-transformed input signal
       and the frequency filter (product in Fourier space) */
    vndrs_vector_complex_double_mul(Fourier_signal, Freq_filter);
    
    /* Transform the filtered signal (Fourier domain) back to the time domain */
    vndrs_ifft_complex_2_real(Fourier_signal, Filtered_signal);
    
    /* Free memory for the Fourier-transformed signal */
    vndrs_free_vector_complex(Fourier_signal);

    return Filtered_signal;
}

/*
  This routine does only a Fourier convolution, i.e. implicitely
  assumes that the Kernel is large enough so no aliasing occurs.
 */
individual_scan *vndrs_convolution(
                                   /* --------------- */
                                   /* input variables */
                                   /* --------------- */ 
                                   individual_scan *Signal,
                                   individual_scan *Kernel
                                  )
{
    long N_elements;
    individual_scan *Convolved_signal;
    fourier_spectrum *Fourier_signal, *Fourier_kernel;

    N_elements = Signal->dimension;

    Convolved_signal = vndrs_calloc_vector_double(N_elements);
    Fourier_signal = vndrs_calloc_vector_complex(N_elements);
    Fourier_kernel = vndrs_calloc_vector_complex(N_elements);
    
    vndrs_fft_real_2_complex(Signal, Fourier_signal);
    vndrs_fft_real_2_complex(Kernel, Fourier_kernel);

    vndrs_vector_complex_mul(Fourier_signal, Fourier_kernel);
    
    vndrs_ifft_complex_2_real(Fourier_signal, Convolved_signal);

    /* since the above is not keeping the energy, we have to
       multiply by the number of pixels in the scan */
    vndrs_vector_scale(Convolved_signal, (double)N_elements);

    vndrs_free_vector_complex(Fourier_signal);
    vndrs_free_vector_complex(Fourier_kernel);

    return Convolved_signal;
}

/*
    Computes a bandpass or a low-pass filter. The filter response
    is defined by 1/(1+abs(f-f0)/delta_f)^n).
    Described in [1], section 2.5
 */
power_spectrum *vndrs_compute_filter (
                                      /* --------------- */
                                      /* input variables */
                                      /* --------------- */ 
                                      long N_frames,
                                      double Sample_interval,
                                      /* time interval between two samples of a scan; unit = sec */
                                      double Central_freq,
                                      double Half_bandwidth,
                                      long Filter_order
                                     )
{
    power_spectrum *Filter, *Full_filter;
    long Filter_size, Filter_size_modulo, I;
    double Element_freq, Ind_central_freq,
           Ind_half_width, Inv_ind_half_width, X;

    /*
    sprintf(vndrs_protocol.text, "N_frames:        %d\n", N_frames);               vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text, "Sample_interval: %22.14e\n", Sample_interval);   vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text, "Central_freq:    %f\n", Central_freq);           vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text, "Half_bandwidth:  %f\n", Half_bandwidth);         vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text, "Filter_order:    %d\n", Filter_order);           vndrs_runtime_protocol();
    */

    /*
      Filter size computation
     */
    /* Unlike described in [1], we do not add 1 to Filter_size */
    Filter_size = N_frames / 2;


	/* ============================== */
        /* Changed on Feb 23/01: Filter_size_modulo 
           takes on the value 0 or 1 for 
           even or odd spectra */
	/* ============================== */
    Filter_size_modulo = N_frames - Filter_size*2; 

    Filter = vndrs_calloc_vector_double(Filter_size);

    /*
      Computation of the element frequency
     */
    if (N_frames == 0)
    {
        fprintf(stderr, "The number of frames is 0.\n");
        vndrs_error_alert();
    }
    if (Sample_interval == 0.)
    {
        fprintf(stderr, "The size of the sample interval is 0.\n");
        vndrs_error_alert();
    }
    Element_freq = 1. / (N_frames * Sample_interval);

    /*
      Computation of the indices corresponding to the central
      frequency and the half width of the filter
     */
    Ind_central_freq = Central_freq / Element_freq;
    Ind_half_width = Half_bandwidth / Element_freq;
    Inv_ind_half_width = 1. / Ind_half_width;

    /*
    sprintf(vndrs_protocol.text, "Element_freq:  %22.14e\n", Element_freq);             vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text, "Ind_central_freq:  %22.14e\n", Ind_central_freq);     vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text, "Ind_half_width:  %22.14e\n", Ind_half_width);         vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text, "Inv_ind_half_width:  %22.14e\n", Inv_ind_half_width); vndrs_runtime_protocol();
    */

    for (I=0; I<Filter_size; I++)
    {
        X = fabs(I - Ind_central_freq) * Inv_ind_half_width;
        Filter->data[I] = 1. / (1. + pow(X, (double)Filter_order));
    }

	/* ============================== */
        /* Changed on Feb 23/01: Filter_size_modulo 
           takes on the value 0 or 1 for 
           even or odd spectra */
	/* ============================== */
    Full_filter = vndrs_mirror_half_pow(Filter, Filter_size_modulo);

    vndrs_free_vector_double(Filter);

    return Full_filter;
}


/*
  Added pkervell 15-mar-02: computation of the Wiener filters from the average
  PSD of PA and PB. The average of the Off_source power spectrum is used to estimate
  the noise distribution.
  changed - pkervell 02-Oct-02 : now we estimate the noise BG from the higher frequencies
  of the average on-source power spectrum: it is much cleaner and gives realistic Wieners,
  that do not let the photon shot noise in.
*/
power_spectrum *vndrs_compute_wiener_mean(
			       const batch *Batch_on,
			       const int verbose_on,
                               const int channel  /* stupid... but error when return both Wieners at the same time...
                                                        0 for WA, 1 for WB */
                               )
{
    long I, J, N_elements, N_scans_on, N_scans_high_signal = 0;
    long index_freq_min_high, index_freq_max_high;
    power_spectrum *WA_mean, *WB_mean;
    power_spectrum *On_source_power_A ,*On_source_power_B;
    power_spectrum *Off_source_power_A ,*Off_source_power_B;
    individual_scan *On_source_signal_A, *On_source_signal_B;
    vector_double *Buffer_A, *Buffer_B;
    vector_double *indices_freq, *values_A, *values_B;
    double slope_A, zero_point_A, slope_B, zero_point_B, dummy;
    long N_elem_fit;
    /* double Signal_level, Signal_threshold;*/
    
    N_elements = Batch_on->fullscans[0]->PA->dimension;
    N_scans_on = Batch_on->number_of_scans;
    
    /* Allocate memory for the rectified On_source signal */
    On_source_power_A = vndrs_calloc_vector_double(N_elements);
    On_source_power_B = vndrs_calloc_vector_double(N_elements);
    On_source_signal_A = vndrs_calloc_vector_double(N_elements);
    On_source_signal_B = vndrs_calloc_vector_double(N_elements);
    
    
    for(J=0;J<N_scans_on;J++)
    {
            N_scans_high_signal++;
            
            vndrs_vector_memcpy(On_source_signal_A, Batch_on->fullscans[J]->PA);
            vndrs_vector_memcpy(On_source_signal_B, Batch_on->fullscans[J]->PB);

            Buffer_A = vndrs_power_spectrum(On_source_signal_A);
            Buffer_B = vndrs_power_spectrum(On_source_signal_B);
            
            /* moving sum of the spectra */
            vndrs_vector_add(On_source_power_A, Buffer_A);
            vndrs_vector_add(On_source_power_B, Buffer_B);
            
            /* These vectors are allocated in the function vndrs_power_spectrum */
            vndrs_free_vector_double(Buffer_A);
            vndrs_free_vector_double(Buffer_B);
    }
    
    /* printf("   N_scans used for Wiener computation: %d \n", N_scans_high_signal);*/
    /* Division of the sum by the number of on-source power spectra added */
    vndrs_vector_scale(On_source_power_A, 1.0/N_scans_high_signal);
    vndrs_vector_scale(On_source_power_B, 1.0/N_scans_high_signal);
   
    
    /* Estimation of the off-source power spectrum */
    /* We linearly fit from half of the Nyquist freq (= fringes) to the Nyquist freq - TO BE ADJUSTED !!*/
    index_freq_min_high = (long)(N_elements/4.0);
    index_freq_max_high = (long)(N_elements/2.0);
    /* printf(" index_freq_min_high = %ld, index_freq_max_high = %ld\n",index_freq_min_high, index_freq_max_high);*/
    N_elem_fit = index_freq_max_high - index_freq_min_high;
    
    indices_freq = vndrs_calloc_vector_double(N_elem_fit);
    values_A = vndrs_calloc_vector_double(N_elem_fit);
    values_B = vndrs_calloc_vector_double(N_elem_fit);

    /* filling the indices vector - linear model */
    for(I=index_freq_min_high;I<index_freq_max_high;I++)
        indices_freq->data[I-index_freq_min_high] = I;
    /* filling the values vector - linear model */
    for(I=index_freq_min_high;I<index_freq_max_high;I++)
        values_A->data[I-index_freq_min_high] = On_source_power_A->data[I];
    for(I=index_freq_min_high;I<index_freq_max_high;I++)
        values_B->data[I-index_freq_min_high] = On_source_power_B->data[I];

    /* Least squares fit */
    vndrs_computes_least_squares(indices_freq,
                                 values_A,
                                 &slope_A,
                                 &zero_point_A,
                                 &dummy);
    vndrs_computes_least_squares(indices_freq,
                                 values_B,
                                 &slope_B,
                                 &zero_point_B,
                                 &dummy);
    
    /* Filling the model noise spectrum */
    Off_source_power_A = vndrs_calloc_vector_double(N_elements);
    Off_source_power_B = vndrs_calloc_vector_double(N_elements);
    for(I=0;I<N_elements;I++)
    {
        Off_source_power_A->data[I] = slope_A * I + zero_point_A;
        Off_source_power_B->data[I] = slope_B * I + zero_point_B;
    }
            
    /* Compute the Wiener filters - memory allocated inside function */
    /* No memory leak in these function calls... */
    WA_mean = vndrs_wiener(On_source_power_A, Off_source_power_A, verbose_on);
    WB_mean = vndrs_wiener(On_source_power_B, Off_source_power_B, verbose_on);
    
    WA_mean->data[0] = 1.0;
    WB_mean->data[0] = 1.0;
    
    if(verbose_on == 1)
      {
	vndrs_fprintf_vector_double(On_source_power_A, "Signal_power_A_Wiener.dat");
	vndrs_fprintf_vector_double(On_source_power_B, "Signal_power_B_Wiener.dat");
	vndrs_fprintf_vector_double(Off_source_power_A, "Noise_model_A_Wiener.dat");
	vndrs_fprintf_vector_double(Off_source_power_B, "Noise_model_B_Wiener.dat");
        vndrs_fprintf_vector_double(WA_mean, "WA_mean.dat");        
        vndrs_fprintf_vector_double(WB_mean, "WB_mean.dat");
      }
      
    vndrs_free_vector_double(On_source_power_A);
    vndrs_free_vector_double(On_source_power_B);
    vndrs_free_vector_double(On_source_signal_A);
    vndrs_free_vector_double(On_source_signal_B);
    vndrs_free_vector_double(Off_source_power_A);
    vndrs_free_vector_double(Off_source_power_B);
        
    vndrs_free_vector_double(indices_freq);
    vndrs_free_vector_double(values_A);
    vndrs_free_vector_double(values_B);
    
    if(channel == 0)
    {
        vndrs_free_vector_double(WB_mean);
        return(WA_mean);
    }
    else
    {
        vndrs_free_vector_double(WA_mean);
        return(WB_mean);
    }
}


/*
    Described in [1], section 2.22
    MEMORY FOR THE RETURNED VECTOR IS ALLOCATED INSIDE THE FUNCTION "mirror_half_pow"
 */
power_spectrum *vndrs_wiener (
                              /* --------------- */
                              /* input variables */
                              /* --------------- */ 
                              power_spectrum *Signal_OSPS, /* One-Sided Power Sprectrum = OSPS */
                              power_spectrum *Noise_OSPS,
			      int verbose_on
                             )
{
    power_spectrum *wiener_raw, *wiener_smoothed,
    *Signal_smoothed, *Noise_smoothed, *Half_wiener_filter, *Wiener_filter;
    long I, N_elements, N_modulo, Index_first_zero, Smoothing_half_width;


    N_elements = (long)((Signal_OSPS->dimension)/2); /* we consider only half of the spectrum */
    Smoothing_half_width = (long)((double)N_elements/64.0); /* THIS IS EMPIRICAL !!! pkervell 03-Sep-02 */
    if(Smoothing_half_width == 0) Smoothing_half_width = 1;
    
	/* ============================== */
        /* Changed on Feb 23/01: N_modulo 
           takes on the value 0 or 1 for 
           even or odd spectra */
	/* ============================== */
    N_modulo   = Signal_OSPS->dimension - 2*N_elements; 


    /* Allocate memory for some vectors */
    wiener_raw      = vndrs_calloc_vector_double(N_elements);
    Half_wiener_filter   = vndrs_calloc_vector_double(N_elements);

    Index_first_zero = N_elements;
   
    /* SMOOTHING OF input spectra */
    /* added - pkervell 03-Sep-02  */
    /* The input spectra are (by nature) noisy, and this causes additional noise on the Wiener that is
    not completely removed by the smoothing afterwards. It is therefore better to smooth them before
    the computation. Memory is allocated by smoothing function.
    */

    /* cancellation of the noise average value */
    Noise_OSPS->data[0] = 0.0;
    
    Signal_smoothed = vndrs_vector_smooth(Signal_OSPS, Smoothing_half_width);
    Noise_smoothed = vndrs_vector_smooth(Noise_OSPS, Smoothing_half_width);
    
    /*
       Force the filter to be equal to 1 at frequency 0,
       to ensure that the mean is conserved
     */
    wiener_raw->data[0] = 1.;

    for (I=1; I<N_elements; I++)
    {
        /* Check if the power spectrum of the measured signal is zero at the given wavenumber (index I) */ 
        if ((Signal_smoothed->data[I] >= Noise_smoothed->data[I]) && (Noise_smoothed->data[I] > 0.0))
        {
          wiener_raw->data[I] = (Signal_smoothed->data[I] - Noise_smoothed->data[I]) / Signal_smoothed->data[I];
	  /* The well known optimal (Wiener) filter formula (see for example:
	     'Numerical Recipes in C', 2nd Ed., p/ 547 ff) */
        }
        else if (Noise_smoothed->data[I] <= 0.0)
            wiener_raw->data[I] = 1.;
        else
        {
            Index_first_zero = I;
            /* The power spectrum of the signal has zero value at index 'I' */
            wiener_raw->data[I] = 0.;
            break;
        }
    }

    /* The Wiener filter has been computed and it is stored in the variable 'Signal_OSPS_cal' */

    /* SMOOTHING OF WIENER FILTER */
    /* -------------------------- */
    /* In the LabView code it is said that smoothing is an average over the neighboring
       pixels, while in fact it is a median: pkervell 15-feb-02 changed to moving average
    */

    wiener_smoothed = vndrs_vector_smooth(wiener_raw, Smoothing_half_width);

    
    /* Now we finalize the filter by normalizing it in 0 */
    wiener_smoothed->data[0] = 1.0; /* to be sure... */     
    
    /* Last check */
    for(I=1; I<N_elements; I++)
    {
        if (wiener_smoothed->data[I] > 1.)
            wiener_smoothed->data[I] = 1.;

        if (wiener_smoothed->data[I] <= 0.)
            wiener_smoothed->data[I] = 0.;
    }

    for(I=Index_first_zero; I<N_elements; I++)
            wiener_smoothed->data[I] = 0.;

    if(verbose_on == 1)
      {
	vndrs_fprintf_vector_double(wiener_smoothed, "Wiener_smoothed.dat");
	vndrs_fprintf_vector_double(wiener_raw, "Wiener_raw.dat");
      }

    /* SET THE WIENER FILTER TO THE SMOOTHED SPECTRUM 'wiener_smoothed' COMPUTED ABOVE */

    /* Up to index of the first zero we use the smoothed spectrum, --starting
       from this index of first zero, the Wiener filter is filled with zeros. */
    for (I=0; I<Index_first_zero; I++)
        Half_wiener_filter->data[I] = wiener_smoothed->data[I];
        /* Half_wiener_filter->data[I] = wiener_raw->data[I];*/
        /* Half_wiener_filter->data[I] = sqrt(wiener_smoothed->data[I]);*/
        /* ATTENTION !!! we take the sqrt as we convolve in fft afterwards pkervell 08-Sep-02 */    
    
    for (; I<N_elements; I++)
        Half_wiener_filter->data[I] = 0.0;
        
	/* ============================== */
        /* Changed on Feb 23/01: N_modulo 
           takes on the value 0 or 1 for 
           even or odd spectra */
	/* ============================== */
        
    Wiener_filter = vndrs_mirror_half_pow(Half_wiener_filter, N_modulo);
    /* memory for the Wiener filter is allocated inside  this function !! */
 
    vndrs_free_vector_double(wiener_raw);
    vndrs_free_vector_double(wiener_smoothed);
    vndrs_free_vector_double(Signal_smoothed);
    vndrs_free_vector_double(Noise_smoothed);
    vndrs_free_vector_double(Half_wiener_filter);

    /* for(I=0;I<100;I++) printf(" W[%d] = %f \n", I, Wiener_filter->data[I]);*/
    return Wiener_filter;

}

/*
    Described in [1], section 2.23
 */
power_spectrum *vndrs_noise_model (
                                   /* --------------- */
                                   /* input variables */
                                   /* --------------- */ 
                                   individual_scan *Dark_current,
                                   long Radius
                                  )
{
    long I, J, N_elements, N_zero;
    double Zero_pow;
    vector_double *Extract, *Ramp;
    power_spectrum *Dark_current_pow, *Noise_power_model, *Convolved_pow;

    N_elements = Dark_current->dimension;

    Noise_power_model = vndrs_calloc_vector_double(N_elements);
    Ramp = vndrs_calloc_vector_double(N_elements);

    Dark_current_pow = vndrs_power_spectrum(Dark_current);

    /* Computation of the zero frequency value from the median
       of the first percent in length of the power spectrum */
    /* WARNING: CHECK WITH VINCENT, WHY THIS WORKS */
    /* since we work with full powerspectra, and the LabView only with
       half powerspectra, we have to divide by an additional 2 in the
       following formula */

    N_zero = (N_elements/2/100);
    Extract = vndrs_calloc_vector_double(N_zero);
    for (I=0; I<N_zero; I++)
        Extract->data[I] = Dark_current_pow->data[I];
    Zero_pow = vndrs_median_scan(Extract);
    Dark_current_pow->data[0] = Zero_pow;

    /* Creation of the ramp pattern */
    /* Old code which does not give a centered ramp */
    /*
    for (I=0; I<Radius; I++)
        Ramp->data[I] = I;
    for (I=Radius, J=Radius-1; I<2*Radius; I++, J--)
        Ramp->data[I] = J;
    */

    Ramp->data[0] = Radius;
    for (I=1, J=N_elements-1; I<Radius; I++, J--)
    {
        Ramp->data[I] = Radius-I;
        Ramp->data[J] = Radius-I;
    }

    /* normalize the ramp to the sum */
    vndrs_vector_scale(Ramp, 1./vndrs_sum_scan(Ramp));

    /* Convolution of the two signals */
    Convolved_pow = vndrs_convolution(Dark_current_pow, Ramp);

    /* THE FOLLOWING CODE IS WRONG AND RESULTED FROM A MISINTERPRETATION OF
       THE LABVIEW CODE, WHICH GIVES TOO LARGE ARRAYS AFTER A CONVOLUTION;
       WE ONLY HAVE TO RETURN THE CONVOLVED SPECTRUM */
    /* Truncation of the resulting power spectrum */
    /*
    vndrs_vector_memcpy(Noise_power_model, Dark_current_pow);
    for (I=Radius; I<N_elements-Radius; I++)
        Noise_power_model->data[I] = Convolved_pow->data[I];
    */

    vndrs_vector_memcpy(Noise_power_model, Convolved_pow);

    vndrs_free_vector_double(Dark_current_pow);
    vndrs_free_vector_double(Convolved_pow);
    vndrs_free_vector_double(Ramp);
    vndrs_free_vector_double(Extract);

    return Noise_power_model;
}

