/*******************************************************************/
/* VLTI VINCI DATA REDUCTION SOFTWARE (VNDRS)                      */
/*                                                                 */
/* FILE: vndrs_Statistics.c                                        */
/*                                                                 */
/* Contains the routines for statistics (mean, stdev etc.)          */
/*                                                                 */
/* Written by Markus Schoeller & Rainer Wilhelm, ESO, 2000         */
/*                                                                 */
/*******************************************************************/ 

#include "vndrs.h"

/*
    Computation of the histogram of a distribution. MEMORY ALLOCATED INSIDE.
    pkervell - 22-Oct-02
*/
void vndrs_histogram(vector_double *values, /* input values */
                     double range, /* range, in number of sigmas, from M-sig to M+sig */
                     vector_double *histogram_scale, /* out */
                     vector_double *histogram) /* out */
{
    long I, J, N_values, N_bins;
    double Mean_value, size_bin, sigma_values;
    
    N_bins = histogram_scale->dimension;
    N_values = values->dimension;
    Mean_value = vndrs_mean_scan(values);
    sigma_values = vndrs_stddev_scan(values);
    size_bin = (2.0 * range * sigma_values)/N_bins;
    
    /* The central value of the histogram is the mean of the distribution */
    for(I=0;I<N_bins;I++)
        histogram_scale->data[I] = (I - N_bins/2.0)*size_bin + Mean_value;
    
    /* Creation of the histogram. Too large and small values are not taken into account. */
    for(I=0;I<N_values;I++)
        for(J=0;J<N_bins;J++)
            if(values->data[I] > (histogram_scale->data[J]-size_bin/2.0) &&
                values->data[I] < (histogram_scale->data[J]+size_bin/2.0))
                    histogram->data[J] = histogram->data[J] + 1;
    /* This algorithm is not optimal, but for a few hundred values, the execution
    time is zero anyway... */
}


/*
    Kurtosis of a distribution.
    pkervell - 21-Oct-02
    Kurtosis is based on the size of a distribution's tails. It is high
    for thin tails (peaked top), and is low for fat tails (flat top).
    Distributions with relatively large tails are called "leptokurtic";
    those with small tails are called "platykurtic." A distribution with
    the same kurtosis as the normal distribution is called "mesokurtic."
*/
double vndrs_kurtosis(vector_double *values)
{
    long I, N_elements;
    double mean_val, sigma_val;
    double Sum_4 = 0.0;
    double kurtosis_val;
    
    N_elements = values->dimension;
    mean_val = vndrs_mean_scan(values);
    sigma_val = vndrs_stddev_scan(values);
    
    for(I=0;I<N_elements;I++)
    {
        Sum_4 = Sum_4 + (values->data[I] - mean_val)*(values->data[I] - mean_val)*
                (values->data[I] - mean_val)*(values->data[I] - mean_val);
    }
    
    if(N_elements > 1 && sigma_val > 0)
        kurtosis_val = Sum_4 / ((N_elements - 1) * sigma_val * sigma_val * sigma_val * sigma_val) - 3.0;
    else kurtosis_val = 0.0;
    
    return kurtosis_val;
}

/* Finding the minimum and the maximum of a matrix of doubles */ 
double vndrs_matrix_min( 
                        /* -------------- */ 
                        /* input variable */ 
                        /* -------------- */ 
                        const matrix_double *src 
                       )
{
    long dim1, dim2, i, j;
    double min_val, buffer;
    
    dim1 = src->dimension_1;
    dim2 = src->dimension_2;
    
    min_val = get_element_matrix_double(src,0,0);
    for(i=0;i<dim1;i++)
        for(j=0;j<dim2;j++)
        {
            buffer = get_element_matrix_double(src,i,j);
            if(buffer < min_val)
                min_val = buffer;
        }
    
    return min_val;
}


double vndrs_matrix_max( 
                        /* -------------- */ 
                        /* input variable */ 
                        /* -------------- */ 
                        const matrix_double *src 
                       )
{
    long dim1, dim2, i, j;
    double max_val, buffer;
    
    dim1 = src->dimension_1;
    dim2 = src->dimension_2;
    
    max_val = get_element_matrix_double(src,0,0);
    for(i=0;i<dim1;i++)
        for(j=0;j<dim2;j++)
        {
            buffer = get_element_matrix_double(src,i,j);
            if(buffer > max_val)
                max_val = buffer;
        }
    
    return max_val;

}


/*
    Computes the simple error bar given by stddev/sqrt(N)
*/
double vndrs_compute_simple_error(vector_double *mu2)
{
    vector_double *mu2_cal;
    double Std_mu2;
    long I, N_scans;

    N_scans = mu2->dimension;

    Std_mu2 = vndrs_stddev_scan(mu2);
    
    return (Std_mu2/sqrt(N_scans));
}

/*
    Computes the skewness coefficient of a sample of data points put in a double vector.
*/
double vndrs_skewness(vector_double *x)
{
    long n,I;
    double mean_x, std_x, Sum;
    
    n = x->dimension;
    
    mean_x = vndrs_mean_scan(x);
    std_x = vndrs_stddev_scan(x);
    
    Sum = 0.0;
    
    for(I=0;I<n;I++)
        Sum += (x->data[I]-mean_x)*(x->data[I]-mean_x)*(x->data[I]-mean_x);

    return Sum/((n-1)*std_x*std_x*std_x);
}

/*
    Outputs the linear least squares fit regression coefficients a and b so that
    we have the model y = a.x + b. It also returns the correlation coefficient on the fit.
    This last parameter varies from 0 to 1 and a value of 1 indicates a perfect fit.
    pkervell 28-Sep-02.
*/

void vndrs_computes_least_squares(vector_double *x,
                                  vector_double *y,
                                  double *slope,
                                  double *zero_point,
                                  double *corr_coef)
{
    double sum_xy, sum_x, sum_y, sum_x2, sum_y2;
    vector_double *xy, *x2, *y2;
    long n;
    
    /* The two tables should be the same length. If they are not, we take the smaller as size */
    if(x->dimension >= y->dimension)
        n = y->dimension;
    else n = x->dimension;
    
    xy = vndrs_calloc_vector_double(n);
    x2 = vndrs_calloc_vector_double(n);
    y2 = vndrs_calloc_vector_double(n);
    
    sum_x = vndrs_sum_scan(x);
    sum_y = vndrs_sum_scan(y);
    
    vndrs_vector_memcpy(xy, x);
    vndrs_vector_mul(xy, y);
    sum_xy = vndrs_sum_scan(xy);

    vndrs_vector_memcpy(x2, x);
    vndrs_vector_mul(x2, x);
    sum_x2 = vndrs_sum_scan(x2);

    vndrs_vector_memcpy(y2, y);
    vndrs_vector_mul(y2, y);
    sum_y2 = vndrs_sum_scan(y2);
    
    *slope = (n * sum_xy - sum_x * sum_y)/(n * sum_x2 - sum_x * sum_x);
    *zero_point = (sum_y * sum_x2 - sum_x * sum_xy)/(n * sum_x2 - sum_x * sum_x);
    *corr_coef = (n * sum_xy - sum_x * sum_y)*(n * sum_xy - sum_x * sum_y)/
               ((n * sum_x2 - sum_x2)*(n * sum_y2 - sum_y * sum_y));
    
    vndrs_free_vector_double(xy);
    vndrs_free_vector_double(x2);
    vndrs_free_vector_double(y2);
    
}


/*
    Computes the average of the variable (kappa1 + kappa2)/(kappa1.kappa2)
    with kappa1=kappa1A.kappa1B and kappa2=kappa2A.kappa2B, knowing the kappa matrix and errors.
    reminder: variance of the sum is the sum of variances.
*/
double vndrs_mean_kappa_norm(matrix_double *Matrix,
                                 matrix_double *Stdev_matrix)
{
    double mean_kappa_norm;
    double kappa1, kappa2, kappa1_var, kappa2_var;
    double kappa1A, kappa1B, kappa2A, kappa2B;
    double kappa1A_err, kappa1B_err, kappa2A_err, kappa2B_err;
    double sum_k1k2, prod_k1k2, sum_k1k2_var, prod_k1k2_var;
    
    vndrs_get_transfer_matrix_coefficients(Matrix, &kappa1A, &kappa1B, &kappa2A, &kappa2B);
    vndrs_get_transfer_matrix_coefficients(Stdev_matrix, &kappa1A_err, &kappa1B_err, &kappa2A_err, &kappa2B_err);

    kappa1 = kappa1A * kappa1B;
    kappa2 = kappa2A * kappa2B;
    kappa1_var = vndrs_variance_product(kappa1A, kappa1A_err*kappa1A_err, kappa1B, kappa1B_err*kappa1B_err);
    kappa2_var = vndrs_variance_product(kappa2A, kappa2A_err*kappa2A_err, kappa2B, kappa2B_err*kappa2B_err);

    sum_k1k2_var = kappa1_var + kappa2_var;
    prod_k1k2_var = vndrs_variance_product(kappa1, kappa1_var, kappa2, kappa2_var);
    
    sum_k1k2 = kappa1 + kappa2;
    prod_k1k2 = kappa1 * kappa2;

    mean_kappa_norm = vndrs_mean_quotient(sum_k1k2, sum_k1k2_var, prod_k1k2, prod_k1k2_var);
    
    return(mean_kappa_norm);
}

/*
    Computes the variance of the variable (kappa1 + kappa2)/(kappa1.kappa2)
    with kappa1=kappa1A.kappa1B and kappa2=kappa2A.kappa2B, knowing the kappa matrix and errors.
    reminder: variance of the sum is the sum of variances.
*/
double vndrs_variance_kappa_norm(matrix_double *Matrix,
                                 matrix_double *Stdev_matrix)
{
    double variance_kappa_norm;
    double kappa1, kappa2, kappa1_var, kappa2_var;
    double kappa1A, kappa1B, kappa2A, kappa2B;
    double kappa1A_err, kappa1B_err, kappa2A_err, kappa2B_err;
    double sum_k1k2, prod_k1k2, sum_k1k2_var, prod_k1k2_var;
    
    vndrs_get_transfer_matrix_coefficients(Matrix, &kappa1A, &kappa1B, &kappa2A, &kappa2B);
    vndrs_get_transfer_matrix_coefficients(Stdev_matrix, &kappa1A_err, &kappa1B_err, &kappa2A_err, &kappa2B_err);

    kappa1 = kappa1A * kappa1B;
    kappa2 = kappa2A * kappa2B;
    kappa1_var = vndrs_variance_product(kappa1A, kappa1A_err*kappa1A_err, kappa1B, kappa1B_err*kappa1B_err);
    kappa2_var = vndrs_variance_product(kappa2A, kappa2A_err*kappa2A_err, kappa2B, kappa2B_err*kappa2B_err);
    
    sum_k1k2_var = kappa1_var + kappa2_var;
    prod_k1k2_var = vndrs_variance_product(kappa1, kappa1_var, kappa2, kappa2_var);
    
    sum_k1k2 = kappa1 + kappa2;
    prod_k1k2 = kappa1 * kappa2;
    
    variance_kappa_norm = vndrs_variance_quotient(sum_k1k2, sum_k1k2_var, prod_k1k2, prod_k1k2_var);
    
    return(variance_kappa_norm);
}

/*
    Computes the variance of a product of normal variables.
    formula from http://www.ses.swin.edu.au/homes/browne/
*/
double vndrs_variance_product(double mean_A,
                              double var_A,
                              double mean_B,
                              double var_B)
{
    double variance_product;
    
    variance_product = (mean_A*mean_A + var_A)*(mean_B*mean_B + var_B) - (mean_A*mean_B)*(mean_A*mean_B);
    /* To avoid crashes... */
    if(variance_product < 0.0) variance_product = fabs(mean_A*mean_A*mean_B*mean_B);
     
    return variance_product;
}

/*
    Computes the mean of the quotient of two normal variables.
    formula from http://www.ses.swin.edu.au/homes/browne/
*/
double vndrs_mean_quotient(double mean_A,
                              double var_A,
                              double mean_B,
                              double var_B)
{
    double mean_quotient, xB;
    
    xB = var_B/(mean_B*mean_B);
    mean_quotient = (mean_A/mean_B)*(1+xB);
    
    return mean_quotient;
}

/*
    Computes the variance of the quotient of two normal variables.
    formula from http://www.ses.swin.edu.au/homes/browne/
*/
double vndrs_variance_quotient(double mean_A,
                              double var_A,
                              double mean_B,
                              double var_B)
{
    double variance_quotient, xA, xB;
    
    xA = var_A/(mean_A*mean_A);
    xB = var_B/(mean_B*mean_B);
    variance_quotient = (mean_A/mean_B)*(mean_A/mean_B)*
                        (xA + xB - xB*xB);
    /* To avoid crashes... */
    if(variance_quotient < 0.0) variance_quotient = fabs((mean_A/mean_B)*(mean_A/mean_B));
    
    return variance_quotient;
}


/*
    Computes the statistics: mean, standard deviation, minimum and maximum
    data values for each channel in the batch.
    Described in [1], section 2.4
 */
void vndrs_batch_statistics (
                             /* --------------- */
                             /* input variables */
                             /* --------------- */ 
                             batch *Batch,
                             /* ---------------- */
                             /* output variables */
                             /* ---------------- */ 
                             vector_double *Mean_val,           /* vector of 4 */
                             vector_double *Standard_deviation, /* vector of 4 */
                             vector_double *Minimum,            /* vector of 4 */
                             vector_double *Maximum             /* vector of 4 */
                            )
{
    /* Local variables */
    long I, J, Number_of_scans, Number_of_frames, Number_of_points;
    double Value_I1, Value_I2, Value_PA, Value_PB;
    double Sum_I1, Sum_I2, Sum_PA, Sum_PB;
    double Sum_square_I1, Sum_square_I2, Sum_square_PA, Sum_square_PB;
    double Min_I1, Min_I2, Min_PA, Min_PB;
    double Max_I1, Max_I2, Max_PA, Max_PB;
    
    
    /* Check dimension of output vectors (should be 4) */
    #ifdef SECURITY_MODE
    if((Mean_val->dimension!=4)||(Standard_deviation->dimension!=4)||
       (Minimum->dimension!=4)||(Maximum->dimension!=4))
    {
        fprintf(stderr, "Error in routine 'vndrs_batch_statistics':\n");
        fprintf(stderr, "At least one of the four output vectors 'Mean_val',\n");
        fprintf(stderr, "'Standard_deviation', 'Minimum', 'Maximum' has a\n");
        fprintf(stderr, "a dimension that is NOT equal to 4! The dimensions are:\n");
        fprintf(stderr, "%ld, %ld, %ld, %ld\n", Mean_val->dimension, Standard_deviation->dimension,
                Minimum->dimension, Maximum->dimension);
                /* csabet 07-Aug-2001. Format precision corrected */
        vndrs_error_alert();
    }
    #endif

    Number_of_scans = Batch->number_of_scans;
    Number_of_frames = Batch->fullscans[0]->I1->dimension;
    Number_of_points = Number_of_scans * Number_of_frames;
    
    /* Protocol Info */
    /*    sprintf(vndrs_protocol.text,"| Number of full scans = %ld                |\n", Number_of_scans);
	vndrs_runtime_protocol();
	
    sprintf(vndrs_protocol.text,"| Number of frames     = %ld                |\n", Number_of_frames);	
	vndrs_runtime_protocol();
	
    sprintf(vndrs_protocol.text,"| Number of data sets (I1, I2, PA, PB)           |\n"); 
	vndrs_runtime_protocol();
	
    sprintf(vndrs_protocol.text,"| in whole batch       = %ld                |\n", Number_of_points);
    vndrs_runtime_protocol();*/

    Sum_I1 = 0;
    Sum_I2 = 0;
    Sum_PA = 0;
    Sum_PB = 0;
    Sum_square_I1 = 0;
    Sum_square_I2 = 0;
    Sum_square_PA = 0;
    Sum_square_PB = 0;
    Min_I1 = Batch->fullscans[0]->I1->data[0];
    Min_I2 = Batch->fullscans[0]->I2->data[0];
    Min_PA = Batch->fullscans[0]->PA->data[0];
    Min_PB = Batch->fullscans[0]->PB->data[0];
    Max_I1 = Batch->fullscans[0]->I1->data[0];
    Max_I2 = Batch->fullscans[0]->I2->data[0];
    Max_PA = Batch->fullscans[0]->PA->data[0];
    Max_PB = Batch->fullscans[0]->PB->data[0];

    for (I=0; I<Number_of_scans; I++)
        for (J=0; J<Number_of_frames; J++)
    {
        Value_I1 = Batch->fullscans[I]->I1->data[J];
        Value_I2 = Batch->fullscans[I]->I2->data[J];
        Value_PA = Batch->fullscans[I]->PA->data[J];
        Value_PB = Batch->fullscans[I]->PB->data[J];

        Sum_I1 += Value_I1;
        Sum_I2 += Value_I2;
        Sum_PA += Value_PA;
        Sum_PB += Value_PB;

        Sum_square_I1 += Value_I1*Value_I1;
        Sum_square_I2 += Value_I2*Value_I2;
        Sum_square_PA += Value_PA*Value_PA;
        Sum_square_PB += Value_PB*Value_PB;

        Min_I1 = Value_I1<Min_I1 ? Value_I1 : Min_I1;
        Min_I2 = Value_I2<Min_I2 ? Value_I2 : Min_I2;
        Min_PA = Value_PA<Min_PA ? Value_PA : Min_PA;
        Min_PB = Value_PB<Min_PB ? Value_PB : Min_PB;

        Max_I1 = Value_I1>Max_I1 ? Value_I1 : Max_I1;
        Max_I2 = Value_I2>Max_I2 ? Value_I2 : Max_I2;
        Max_PA = Value_PA>Max_PA ? Value_PA : Max_PA;
        Max_PB = Value_PB>Max_PB ? Value_PB : Max_PB;
    }

    Mean_val->data[0] = Sum_I1 / Number_of_points;
    Mean_val->data[1] = Sum_I2 / Number_of_points;
    Mean_val->data[2] = Sum_PA / Number_of_points;
    Mean_val->data[3] = Sum_PB / Number_of_points;

    Standard_deviation->data[0] = sqrt(Sum_square_I1 / Number_of_points - Mean_val->data[0]*Mean_val->data[0]);
    Standard_deviation->data[1] = sqrt(Sum_square_I2 / Number_of_points - Mean_val->data[1]*Mean_val->data[1]);
    Standard_deviation->data[2] = sqrt(Sum_square_PA / Number_of_points - Mean_val->data[2]*Mean_val->data[2]);
    Standard_deviation->data[3] = sqrt(Sum_square_PB / Number_of_points - Mean_val->data[3]*Mean_val->data[3]);
    /* BUG FIX, Jan 15, 01:
       'sqrt' was missing, i.e. it was computing the variance instead of the standard deviation! */
    Minimum->data[0] = Min_I1;
    Minimum->data[1] = Min_I2;
    Minimum->data[2] = Min_PA;
    Minimum->data[3] = Min_PB;

    Maximum->data[0] = Max_I1;
    Maximum->data[1] = Max_I2;
    Maximum->data[2] = Max_PA;
    Maximum->data[3] = Max_PB;

    return;
}

/*
    Described in [1], section 2.24
    This procedure becomes superfluous and should be exchanged
    by calls to vndrs_mean_scan and vndrs_stddev_scan.
 */
void vndrs_stats (
                  /* --------------- */
                  /* input variables */
                  /* --------------- */ 
                  individual_scan *Signal,
                  /* ---------------- */
                  /* output variables */
                  /* ---------------- */                   
                  double *Mean,
                  double *Stddev
                 )
{
    long I, Number_of_frames;
    double Value, Sum, Sum_square;

    Number_of_frames = Signal->dimension;
    Sum              = 0.;
    Sum_square       = 0.;

    for (I=0; I<Number_of_frames; I++)
    {
        Value = Signal->data[I];
        Sum += Value;
        Sum_square += Value * Value;
    }
    *Mean = Sum / (double)Number_of_frames;
    *Stddev = Sum_square/(double)Number_of_frames - (*Mean)*(*Mean);

    return;
}

/*
    This routine computes the mean over one scan.
 */
double vndrs_mean_scan(
                       /* --------------- */
                       /* input variables */
                       /* --------------- */ 
                       individual_scan *Signal
                      )
{
    long I, Number_of_frames;
    double Sum;

    Number_of_frames = Signal->dimension;
    Sum = 0.;

    for (I=0; I<Number_of_frames; I++)
        Sum += Signal->data[I];

    return Sum/(double)Number_of_frames;
}

/*
    This routine computes the mean over one scan, weighted by the vector 'weights'.
 */
double vndrs_mean_scan_weighted(
                       /* --------------- */
                       /* input variables */
                       /* --------------- */ 
                       individual_scan *Signal,
                       vector_double *weights 
                      )
{
    long I, Number_of_frames;
    double Sum, Sum_weights;

    Number_of_frames = Signal->dimension;
    Sum = 0.;
    Sum_weights = 0.;

    for (I=0; I<Number_of_frames; I++)
    {
        Sum += Signal->data[I]*weights->data[I];
        Sum_weights += weights->data[I];
    }

    return Sum/(double)Sum_weights;
}

/* added pkervell 14-Oct-02 :
computation of the arithmetic-geometric average of two doubles (convergence 1% or 100 iter.) */
double vndrs_arith_geom_mean(double x, double y)
{
    long I, N_max = 100;
    vector_double *a, *g;
    double arith_geom_mean;
    
    a = vndrs_calloc_vector_double(N_max);
    g = vndrs_calloc_vector_double(N_max);
    
    /* THIS FUNCTION ACCEPTS ONLY POSITIVE VALUES ! Check to be done outside */
    x=fabs(x);
    y=fabs(y);
    
    a->data[0] = (x+y)/2;
    g->data[0] = sqrt(x*y);
    
    I=1;
    do
    {
        a->data[I] = (a->data[I-1] + g->data[I-1])/2;
        g->data[I] = sqrt(a->data[I-1]*g->data[I-1]);
        
        arith_geom_mean = a->data[I];
     
        I++;
    } while((I<N_max-1) | ((g->data[I]-a->data[I])/a->data[I]) > 0.01);
    
    vndrs_free_vector_double(a);
    vndrs_free_vector_double(g);
    
    return arith_geom_mean;
}

/* added pkervell 14-Oct-02 :
computation of the arithmetic-geometric average of two vectors over the central 'width'
pixels of a scan. The computation is done in two steps: first the a-g mean for each frame,
and then a regular mean for the whole considered width. */
double vndrs_arith_geom_mean_scan_center( 
                       /* --------------- */ 
                       /* input variables */ 
                       /* --------------- */  
                       individual_scan *Signal_1,
                       individual_scan *Signal_2,
                       long width 
                      )
{
    long I, Number_of_frames, half_width;
    vector_double *vector_ag_mean;
    double ag_mean;
    
    Number_of_frames = Signal_1->dimension;
    
    if(width >= Number_of_frames) half_width = (long)(Number_of_frames/2);
    else half_width = (long)(width/2);

    vector_ag_mean = vndrs_calloc_vector_double(2*half_width);

    for ( I = (Number_of_frames/2 - half_width) ; I< (Number_of_frames/2 + half_width) ; I++)
        vector_ag_mean->data[I-(Number_of_frames/2 - half_width)] = 
                vndrs_arith_geom_mean(Signal_1->data[I], Signal_2->data[I]);

    ag_mean = vndrs_mean_scan(vector_ag_mean);
    
    vndrs_free_vector_double(vector_ag_mean);
    
    return ag_mean;
}


/* added pkervell 29-Aug-02 :
computation of the average over the central 'width' pixels of a scan */
double vndrs_mean_scan_center( 
                       /* --------------- */ 
                       /* input variables */ 
                       /* --------------- */  
                       individual_scan *Signal,
                       long width 
                      )
{
    long I, Number_of_frames, half_width;
    double Sum;

    Number_of_frames = Signal->dimension;
    Sum = 0.;

    if(width >= Number_of_frames) half_width = Number_of_frames/2;
    else half_width = width/2;

    for ( I = (Number_of_frames/2 - half_width) ; I< (Number_of_frames/2 + half_width) ; I++)
        Sum += Signal->data[I];

    return Sum/(double)(2*half_width);
}

/* added pkervell 28-Sep-02 :
computation of the average of the logarithms of the intensities over the central
'width' pixels of a scan, to take into account the intrinsic log-normal statistics
of the photometry. The logarithm of the flux has in principle a normal statistics. (TBC) */
double vndrs_mean_log_scan_center( 
                       /* --------------- */ 
                       /* input variables */ 
                       /* --------------- */  
                       individual_scan *Signal,
                       long width 
                      )
{
    long I, ind, Number_of_frames, half_width;
    double Sum;

    Number_of_frames = Signal->dimension;
    if(width >= Number_of_frames) half_width = Number_of_frames/2;
    else half_width = width/2;
    
    Sum = 0.;
    ind = 0;
    
    for ( I = (Number_of_frames/2 - half_width) ; I< (Number_of_frames/2 + half_width) ; I++)
    {
        if(Signal->data[I] > 0.0)
        {
            Sum += log(Signal->data[I]); /* natural logarithm */
            ind++;
        }
    }
    
    return exp(Sum/(double)ind);
}

/*
    Computes the average of a vector between two indices. pkervell - 28-Oct-02
*/
double vndrs_mean_scan_limits( 
                       /* --------------- */ 
                       /* input variables */ 
                       /* --------------- */  
                       individual_scan *Signal,
                       long index_min,
                       long index_max 
                      )
{
    long I, Number_of_frames, buffer;
    double Sum;

    Number_of_frames = Signal->dimension;
    Sum = 0.;

    if(index_max > (Number_of_frames-1) || index_max < 0) index_max = Number_of_frames - 1;
    if(index_min < 0 || index_min > (Number_of_frames-1)) index_min = 0;        

    for ( I = index_min ; I< index_max ; I++)
        Sum += Signal->data[I];

    if(index_max == index_min)
        return 0.0;
    else
        return Sum/(double)(index_max-index_min);
}


/* added pkervell 28-Sep-02 :
computation of the average of the logarithms of the intensities over a vector_double,
to take into account for example the intrinsic log-normal statistics of the photometry.
The logarithm of the flux has in principle a normal statistics. (TBC) */
double vndrs_mean_log_scan( 
                       /* --------------- */ 
                       /* input variables */ 
                       /* --------------- */  
                       individual_scan *Signal
                      )
{
    long I, Number_of_frames;
    double Sum;

    Number_of_frames = Signal->dimension;
    
    Sum = 0.;
    
    for ( I = 0 ; I< Number_of_frames ; I++)
    {
        if(Signal->data[I] > 0.0)
            Sum += log(Signal->data[I]); /* natural logarithm */
    }
    
    return exp(Sum/(double)Number_of_frames);
}

/*
    This routine computes the sum over one scan.
 */
double vndrs_sum_scan(
                      /* --------------- */
                      /* input variables */
                      /* --------------- */ 
                      individual_scan *Signal
                     )
{
    long I, Number_of_frames;
    double Sum;

    Number_of_frames = Signal->dimension;
    Sum = 0.;

    for (I=0; I<Number_of_frames; I++)
        Sum += Signal->data[I];

    return Sum;
}


/* added pkervell 26-Sep-02 :
computation of the average over the central 'width' pixels of a scan,
weighted by a linear weighting as a function of the distance to the center.*/
double vndrs_mean_scan_center_weighted( 
                       /* --------------- */ 
                       /* input variables */ 
                       /* --------------- */  
                       individual_scan *Signal,
                       long width 
                      )
{
    long I, ind, Number_of_frames, half_width;
    double Sum, weighting;
    vector_double *weights;

    Number_of_frames = Signal->dimension;
    if(width >= Number_of_frames) half_width = (long)(Number_of_frames/2);
    else half_width = (long)vndrs_round(width/2.0);

    /* Construction of the weighting vector */
    weights = vndrs_calloc_vector_double(width);
    for(I=0;I<half_width;I++) /* ascending ramp */
        weights->data[I] = (1.0*I+1.0)/(1.0*half_width);
    for(I=half_width;I<width;I++) /* descending ramp */
        weights->data[I] = 1.0 - (I+1.0-1.0*half_width)/(1.0*half_width);
    
    weighting = vndrs_sum_scan(weights);
    
    Sum = 0.;
    ind = 0;
    
    for ( I = (Number_of_frames/2 - half_width) ; I< (Number_of_frames/2 + half_width) ; I++)
    {
        Sum += Signal->data[I] * weights->data[ind];
        ind++;
    }
    
    vndrs_free_vector_double(weights);
    
    return Sum/(double)weighting;
}


/*
    This routine computes the mean squared over one scan.
 */
double vndrs_mean_squared_scan(
                               /* --------------- */
                               /* input variables */
                               /* --------------- */ 
                               individual_scan *Signal
                              )
{
    long I, Number_of_frames;
    double Value, Sum_square;

    Number_of_frames = Signal->dimension;
    Sum_square = 0.;

    for (I=0; I<Number_of_frames; I++)
    {
        Value = Signal->data[I];
        Sum_square += Value * Value;
    }

    return Sum_square/(double)Number_of_frames;
}

/*
    This routine computes the stddev over one scan.
 */
double vndrs_stddev_scan(
                         /* --------------- */
                         /* input variables */
                         /* --------------- */ 
                         individual_scan *Signal
                        )
{
    long I, Number_of_frames;
    double Value, Sum, Sum_square;

    Number_of_frames = Signal->dimension;
    Sum = 0.;
    Sum_square = 0.;

    for (I=0; I<Number_of_frames; I++)
    {
        Value = Signal->data[I];
        Sum += Value;
        Sum_square += Value * Value;
    }
    Sum /= (double)Number_of_frames;
    Sum_square /= (double)Number_of_frames;

    return sqrt(Sum_square-Sum*Sum);
    /* BUG FIX, Jan 04/01: sqrt was missing !! */
}

/*
    This routine computes the minimum over one scan.
 */
double vndrs_minimum_scan(
                          /* --------------- */
                          /* input variables */
                          /* --------------- */ 
                          individual_scan *Signal
                         )
{
    long I, Number_of_frames;
    double Min;

    Number_of_frames = Signal->dimension;
    Min = Signal->data[0];

    for (I=1; I<Number_of_frames; I++)
        Min = Signal->data[I]<Min ? Signal->data[I] : Min;

    return Min;
}

/*
    This routine computes the maximum over one scan.
 */
double vndrs_maximum_scan(
                          /* --------------- */
                          /* input variables */
                          /* --------------- */ 
                          individual_scan *Signal
                         )
{
    long I, Number_of_frames;
    double Max;

    Number_of_frames = Signal->dimension;
    Max = Signal->data[0];

    for (I=1; I<Number_of_frames; I++)
        Max = Signal->data[I]>Max ? Signal->data[I] : Max;

    return Max;
}


/* 
  This routine returns the median of an individual scan over the central "width" length. 
*/ 
double vndrs_median_scan_center( 
                       /* --------------- */ 
                       /* input variables */ 
                       /* --------------- */  
                       individual_scan *Signal,
                       long width 
                        )
{
    long I, Number_of_frames, half_width, ind;
    double median_value;
    vector_double *center_part;

    Number_of_frames = Signal->dimension;
    center_part = vndrs_calloc_vector_double(width);
    
    if(width >= Number_of_frames) half_width = Number_of_frames/2;
    else half_width = width/2;
    
    ind = 0;
    for ( I = (Number_of_frames/2 - half_width) ; I< (Number_of_frames/2 + half_width) ; I++)
    {
        center_part->data[ind] = Signal->data[I];
        ind++;
    }
    
    median_value = vndrs_median_scan(center_part);
    
    vndrs_free_vector_double(center_part);
    
    return median_value;
}


double vndrs_median_scan(
                         /* --------------- */
                         /* input variables */
                         /* --------------- */ 
                         const individual_scan *Signal
                        )
/* Returns the median of an individual scan */
{
    /* Local variables */
    double Median;
    vector_double *TempSignal;

    /* Allocate memory for temporary signal vector */
    TempSignal = vndrs_calloc_vector_double(Signal->dimension);

    /* Copy 'Signal' to 'TempSignal' */
    vndrs_vector_memcpy(TempSignal, Signal);
    
    /* Sort the Signal vector is ascending order */
    vndrs_sort_shellsmethod(TempSignal);
    
    /* Get the central value (=median) if the dimension is even */
    /* modified pkervell 13-mar-2002 this procedure was giving wrong value for odd lengths */
    /* Medians are tricky !! */
    if((TempSignal->dimension) == 1)
      Median = TempSignal->data[0];
    else
      {
	if(abs(((int)(TempSignal->dimension)/2.0) - (TempSignal->dimension)/2.0)<0.01)
	  Median = TempSignal->data[(int)((TempSignal->dimension)/2.0)];    
	else
	  /* Get the central value (=median) if the dimension is even */
	  Median = 0.5*(TempSignal->data[(int)((TempSignal->dimension)/2.0-0.5)]+
			TempSignal->data[(int)((TempSignal->dimension)/2.0+0.5)]);    
      }

    /* Free memory for temp. signal */
    vndrs_free_vector_double(TempSignal);

    /* Return the result */
    return Median;
}


void vndrs_subtract_mean_from_scan(/* ---------------------------------------- */
                                   /* input and output variable (is replaced!) */
                                   /* ---------------------------------------- */
                                   individual_scan *dest
                                  )
/* Computes the mean of the scan 'dest' and subtracts this mean value
   from the scan. The result is returned as vector 'dest'. This means
   that the vector 'dest' is REPLACED in this routine!
   Make sure that memory for the vector 'dest' has been allocated before
   calling this function! */
{
    /* Local variables */
    double Mean;

    /* Compute the mean of the scan */
    Mean = vndrs_mean_scan(dest);
    
    /* Subtract the mean value from the scan */ 
    vndrs_vector_add_constant(dest, -Mean);

    return;   
}

/* Computes the median of a double matrix along one direction, to produce the median
vector along the other direction - MEMORY TO BE ALLOCATED OUTSIDE OF THIS FUNCTION
pkervell - 24-Sep-02 */

void vndrs_median_matrix_to_vector_double(vector_double *Pow_spec_med, /* output */
                                          matrix_double *Pow_spec_matrix, /* input */
                                          vector_double *Flag_vector) /* input */
{
    long dim_1, dim_2, dim_3, I, J, ind;
    vector_double *single_vector;
    
    dim_1 = Pow_spec_matrix->dimension_1;
    dim_2 = Pow_spec_matrix->dimension_2;
    dim_3 = vndrs_sum_scan(Flag_vector);
    
    single_vector = vndrs_calloc_vector_double(dim_3);
    
    for(J=0;J<dim_2;J++)
    {
        ind = 0;
        for(I=0;I<dim_1;I++)
            if(Flag_vector->data[I] == 1)
            {
                single_vector->data[ind] = get_element_matrix_double(Pow_spec_matrix, I, J);
                ind++;
            }
        Pow_spec_med->data[J] = vndrs_median_scan(single_vector);
    }
    
    vndrs_free_vector_double(single_vector);
}




