/*******************************************************************/
/* VLTI VINCI DATA REDUCTION SOFTWARE (VNDRS)                      */
/*                                                                 */
/* FILE: vndrs_Support.c                                           */
/*                                                                 */
/* Contains additional support routines                            */
/*                                                                 */
/* Written by Markus Schoeller & Rainer Wilhelm, ESO, 2000         */
/* Numerous additions by pkervell, 2002                            */
/*******************************************************************/ 

#include "vndrs.h"
#include <math.h>
#include "vndrs_Globalvar.h"


/*
    Computation of the gains of the LISA camera pixels - pkervell 19-Oct-02
*/
void vndrs_compute_camera_gains(full_scan *Avg_power_spectra_off,
                                full_scan *Avg_power_spectra_on,
                                vector_double *gains, /* out */
                                vector_double *RONs) /* out */
{
    /*    
    The idea here is to estimate the sigma^2 only from the upper 1/3rd part
    of the power spectral density, which is clean from low-freq photometry, and only affected by
    readout noise and photon shot noise. We use the fact that the variance is the integral of the
    power spectral density.
    This estimation is IMPRECISE but gives a signal if one pixel is dead or dying.
    */
    long I, N_elements, pix_min, pix_max;
    vector_double *variance_on, *variance_off, *mean_on, *mean_off;
    double fraction_used = 3.0; /* inverse of the fraction of the spectrum used for computation */
    
    /* We are using only half of the spectrum */
    N_elements = (long)(Avg_power_spectra_off->I1->dimension/2);
    
    variance_on = vndrs_calloc_vector_double(4);
    variance_off = vndrs_calloc_vector_double(4);
    mean_on = vndrs_calloc_vector_double(4);
    mean_off = vndrs_calloc_vector_double(4);
    
    pix_min = (long)((fraction_used - 1.0)*N_elements/fraction_used);
    pix_max = N_elements;
    
    for(I=pix_min; I<pix_max; I++)
    {
        variance_off->data[0] = variance_off->data[0] + 2.0 * fraction_used * Avg_power_spectra_off->I1->data[I];
        variance_on->data[0] = variance_on->data[0] + 2.0 * fraction_used * Avg_power_spectra_on->I1->data[I];
        variance_off->data[1] = variance_off->data[1] + 2.0 * fraction_used * Avg_power_spectra_off->I2->data[I];
        variance_on->data[1] = variance_on->data[1] + 2.0 * fraction_used * Avg_power_spectra_on->I2->data[I];
        variance_off->data[2] = variance_off->data[2] + 2.0 * fraction_used * Avg_power_spectra_off->PA->data[I];
        variance_on->data[2] = variance_on->data[2] + 2.0 * fraction_used * Avg_power_spectra_on->PA->data[I];
        variance_off->data[3] = variance_off->data[3] + 2.0 * fraction_used * Avg_power_spectra_off->PB->data[I];
        variance_on->data[3] = variance_on->data[3] + 2.0 * fraction_used * Avg_power_spectra_on->PB->data[I];
    }

    /* Mean values (squared in the PSDs) */
    mean_on->data[0] = Avg_power_spectra_on->I1->data[0];
    mean_on->data[1] = Avg_power_spectra_on->I2->data[0];
    mean_on->data[2] = Avg_power_spectra_on->PA->data[0];
    mean_on->data[3] = Avg_power_spectra_on->PB->data[0];
    mean_off->data[0] = Avg_power_spectra_off->I1->data[0];
    mean_off->data[1] = Avg_power_spectra_off->I2->data[0];
    mean_off->data[2] = Avg_power_spectra_off->PA->data[0];
    mean_off->data[3] = Avg_power_spectra_off->PB->data[0];
    
    /* Actual computation of the gains */
    
    for(I=0;I<4;I++)
        gains->data[I] = (sqrt(mean_on->data[I]) - sqrt(mean_off->data[I]))/(variance_on->data[I] - variance_off->data[I]);
    
    for(I=0;I<4;I++)
        RONs->data[I] = gains->data[I] * sqrt(variance_off->data[I] - sqrt(mean_off->data[I])/gains->data[I]);
        
    vndrs_free_vector_double(variance_on);
    vndrs_free_vector_double(variance_off);
    vndrs_free_vector_double(mean_on);
    vndrs_free_vector_double(mean_off);
}


/*
    Automated verification of the readout noise spectrum - pkervell 18-Oct-02.
*/
void vndrs_check_readout_noise_spectrum(full_scan *Avg_power_spectra_off,
                                        vector_double *off_source_spectrum_slopes,
                                        vector_double *number_of_pickups)
{
    long I, N_elements, N_ignore;
    vector_double *indices, *values_I1, *values_I2, *values_PA, *values_PB;
    double dummy, slope_I1, slope_I2, slope_PA, slope_PB,
        zero_I1, zero_I2, zero_PA, zero_PB;
    double sigma_I1, sigma_I2, sigma_PA, sigma_PB;
    double N_picks_I1 = 0.0, N_picks_I2 = 0.0, N_picks_PA = 0.0, N_picks_PB = 0.0;
    
    /* We consider only half of the spectrum, the other half is identical */
    N_elements = (Avg_power_spectra_off->I1->dimension)/2; 
    
    /* We ignore the first 2 elements of the spectrum, if possible */
    if(N_elements > 10) N_ignore = 2; else N_ignore = 0;
    indices   = vndrs_calloc_vector_double(N_elements); /* changed size pkervell - 12-Jan-03 */
    values_I1 = vndrs_calloc_vector_double(N_elements);
    values_I2 = vndrs_calloc_vector_double(N_elements);
    values_PA = vndrs_calloc_vector_double(N_elements);
    values_PB = vndrs_calloc_vector_double(N_elements);
   
    /* Creation of the values vectors, excluding the first N_ignore values (average and very low freqs) */
    for(I=N_ignore;I<N_elements;I++)
    {
        indices->data[I] = I-N_ignore;
        values_I1->data[I-N_ignore] = Avg_power_spectra_off->I1->data[I];
        values_I2->data[I-N_ignore] = Avg_power_spectra_off->I2->data[I];
        values_PA->data[I-N_ignore] = Avg_power_spectra_off->PA->data[I];
        values_PB->data[I-N_ignore] = Avg_power_spectra_off->PB->data[I];
    }
    
    /* Estimation of the slopes of the spectra */
    vndrs_computes_least_squares(indices, values_I1, &slope_I1, &zero_I1, &dummy);
    vndrs_computes_least_squares(indices, values_I2, &slope_I2, &zero_I2, &dummy);
    vndrs_computes_least_squares(indices, values_PA, &slope_PA, &zero_PA, &dummy);
    vndrs_computes_least_squares(indices, values_PB, &slope_PB, &zero_PB, &dummy);

    /* Subtraction of the slope background to the original spectra */
    for(I=0;I<N_elements-N_ignore;I++)
    {
        values_I1->data[I] = values_I1->data[I] - (I * slope_I1 + zero_I1);
        values_I2->data[I] = values_I1->data[I] - (I * slope_I2 + zero_I2);
        values_PA->data[I] = values_I1->data[I] - (I * slope_PA + zero_PA);
        values_PB->data[I] = values_I1->data[I] - (I * slope_PB + zero_PB);
    }
    
    /* Computation of residual noise sigmas */
    sigma_I1 = vndrs_stddev_scan(values_I1);
    sigma_I2 = vndrs_stddev_scan(values_I2);
    sigma_PA = vndrs_stddev_scan(values_PA);
    sigma_PB = vndrs_stddev_scan(values_PB);
    
    /* Detection and counting of 5 sigmas outliers */
    /* We ignore again the first N_ignore samples (total ignored 2*N_ignore) to avoid
    residual photometry contamination in the search for electrical pickups */
     for(I=N_ignore;I<N_elements-N_ignore;I++)
    {
        if(values_I1->data[I] > 5 * sigma_I1) N_picks_I1++;
        if(values_I2->data[I] > 5 * sigma_I2) N_picks_I2++;
        if(values_PA->data[I] > 5 * sigma_PA) N_picks_PA++;
        if(values_PB->data[I] > 5 * sigma_PB) N_picks_PB++;
    }
    
    /* Copy to output variables */
    off_source_spectrum_slopes->data[0] = slope_I1;
    off_source_spectrum_slopes->data[1] = slope_I2;
    off_source_spectrum_slopes->data[2] = slope_PA;
    off_source_spectrum_slopes->data[3] = slope_PB;
    number_of_pickups->data[0] = N_picks_I1;
    number_of_pickups->data[1] = N_picks_I2;
    number_of_pickups->data[2] = N_picks_PA;
    number_of_pickups->data[3] = N_picks_PB;

    vndrs_free_vector_double(indices);
    vndrs_free_vector_double(values_I1);
    vndrs_free_vector_double(values_I2);
    vndrs_free_vector_double(values_PA);
    vndrs_free_vector_double(values_PB);
}

/*
    Procedure to cancel the edges of the wavelets power spectrum over a specified
    width on each (time) side. Matrix is replaced. No security check, verify the
    matrix and width values. - pkervell 17-Oct-02.
*/
void vndrs_cancel_wl_edges(matrix_double *norm_t_wavelets, long width)
{
    /* dimension 1 of the matrix is time, dimension 2 is frequency */
    long I, J;
    
    for(I=0; I<width; I++)
        for(J=0; J<norm_t_wavelets->dimension_2; J++)
        {
            set_element_matrix_double(norm_t_wavelets, I, J, 0.0);
            set_element_matrix_double(norm_t_wavelets, norm_t_wavelets->dimension_1 - I - 1, J, 0.0);
        }
}

/*
    This procedure recenters the PSD peaks using the Pixel_offset provided by the user
    (originally computed from the wavelets cetroiding in the PSD). 
    pkervell - 15-Oct-02
*/
void vndrs_center_PSD_peak(matrix_double *PSDs,
                           vector_double *Flag_processed,
                           long min_pixel_lowband, /* minimum pixel in the PSD to consider in the centering */
                           double Theoretical_position, /* in PSD pixels */
                           vector_double *Measured_positions) /* in PSD pixels */
{
    long I, J, K, N_scans, N_frames_psd;
    vector_double *buffer_psd, *Pixel_offsets;
    double low_freqs_value;
    
    N_scans = PSDs->dimension_1;
    N_frames_psd = PSDs->dimension_2;
    
    buffer_psd = vndrs_calloc_vector_double(N_frames_psd);
    Pixel_offsets = vndrs_calloc_vector_double(N_scans);
    
    for(I=0;I<N_scans;I++)
        Pixel_offsets->data[I] = Measured_positions->data[I] - Theoretical_position;
    
    for(I=0;I<N_scans;I++)
    if(Flag_processed->data[I] > 0)
    {
        /* First of all, we replace the pixels that are contaminated by photometry by the first value above
        the user-specified minimum frequency - pkervell 19-Oct-02 */
        low_freqs_value = get_element_matrix_double(PSDs, I, min_pixel_lowband);
        for(K=0;K<min_pixel_lowband;K++) set_element_matrix_double(PSDs, I, K, low_freqs_value);
        
        for(K=0;K<N_frames_psd;K++) buffer_psd->data[K] = get_element_matrix_double(PSDs, I, K);
        
        if(Pixel_offsets->data[I] > 0)
        {
            for(J=Pixel_offsets->data[I];J<N_frames_psd;J++)
                set_element_matrix_double(PSDs, I, (long)(J-Pixel_offsets->data[I]), buffer_psd->data[J]);
            /* Completion of the displaced vector with the last value of the original PSD */
            for(J=N_frames_psd-Pixel_offsets->data[I];J<N_frames_psd;J++)
                set_element_matrix_double(PSDs, I, J, buffer_psd->data[N_frames_psd-1]);                
        }

        if(Pixel_offsets->data[I] < 0)
        {
            Pixel_offsets->data[I] = fabs(Pixel_offsets->data[I]);
            for(J=Pixel_offsets->data[I];J<N_frames_psd;J++)
                set_element_matrix_double(PSDs, I, J, buffer_psd->data[(long)(J-Pixel_offsets->data[I])]);
            /* Completion of the displaced vector with the first value of the original PSD */
            for(J=0;J<Pixel_offsets->data[I];J++)
                set_element_matrix_double(PSDs, I, J, buffer_psd->data[0]);                
        }

    }
    
    vndrs_free_vector_double(buffer_psd);
    vndrs_free_vector_double(Pixel_offsets);
}


/*
    Integration of the background corrected PSD between specified limits. This is quite similar to the 
    procedure vndrs_integrate_fringe_power, but does not ask for a background and accepts normalized
    interferogram spectrum.
    pkervell - 15-Oct-02.
*/
double vndrs_integrate_PSD(vector_double *PSD,
                           double wavenum_interval,
                           double wavenum_min,
                           double wavenum_max)
{
    long N_elements, index_min, index_max;
    double sum_value = 0.0;
    
    N_elements = PSD->dimension;
    index_min = vndrs_round(wavenum_min / wavenum_interval);
    index_max = vndrs_round(wavenum_max / wavenum_interval);
    
    if(index_min < 0 || index_max < 0 || index_min >= N_elements || index_max >= N_elements)
    {
        index_min = 0;
        index_max = N_elements;
        sum_value = 0.0;
    }
    else
    {
        sum_value = vndrs_area(PSD, wavenum_interval, wavenum_min, wavenum_max)/wavenum_interval;
    }
    
    return sum_value;
}

/*
    Estimation of the peak limits from the corrected power spectrum - pkervell 15-Oct-02.
*/
void vndrs_estimate_peak_limits(vector_double *Pow_spec_avg,
                                double Wavenum_theo,
                                double Wavenum_interval,
                                double *wavenum_min, /* output */
                                double *wavenum_max, /* output */
                                double *wavenum_min_halfmax, /* output */
                                double *wavenum_max_halfmax) /* output */
{
    long I, N_frames, center_pix, pix_min = 0, pix_max = 0, pix_min_hm = 0, pix_max_hm = 0;
    double max_value;
    
    N_frames = Pow_spec_avg->dimension;
    /* We suppose by default that the central pixel is the maximum (about true after recentering) */
    center_pix = vndrs_round(Wavenum_theo/Wavenum_interval);
    max_value = Pow_spec_avg->data[center_pix];

    if(center_pix < 0 || center_pix >= N_frames)
    {
        *wavenum_min = 0.0;
        *wavenum_max = 0.0;
        *wavenum_min_halfmax = 0.0;
        *wavenum_max_halfmax = 0.0;
    }
    else
    {
        
        for(I=center_pix; I<N_frames; I++)
        {
            if(Pow_spec_avg->data[I] <= max_value/2.0)
            {
                pix_max_hm = I;
                break;
            }
        }
    
        for(I=center_pix; I<N_frames; I++)
        {
            if(Pow_spec_avg->data[I] <= 0.0)
            {
                pix_max = I;
                break;
            }
        }
    
        for(I=center_pix; I>=0; I--)
        {
            if(Pow_spec_avg->data[I] <= max_value/2.0)
            {
                pix_min_hm = I;
                break;
            }
        }
        for(I=center_pix; I>=0; I--)
        {
            if(Pow_spec_avg->data[I] <= 0.0)
            {
                pix_min = I;
                break;
            }
        }
        
        *wavenum_min = pix_min * Wavenum_interval;
        *wavenum_max = pix_max * Wavenum_interval;
        *wavenum_min_halfmax = pix_min_hm * Wavenum_interval;
        *wavenum_max_halfmax = pix_max_hm * Wavenum_interval;
    }
}


/*
    Computes the photometric averages of each scan of a batch and puts the result in a
    vector  - pkervell 10-Oct-02
*/
void vndrs_compute_average_vector_from_batch(batch *Batch_in, /* input */
                                  vector_double *I1_mean_vector,
                                  vector_double *I2_mean_vector,
                                  vector_double *PA_mean_vector,
                                  vector_double *PB_mean_vector) /*out*/
{
    long N_scans, I;
    
    N_scans = Batch_in->number_of_scans;
    
    for(I=0;I<N_scans;I++)
    {
        I1_mean_vector->data[I] = vndrs_mean_scan(Batch_in->fullscans[I]->I1);
        I2_mean_vector->data[I] = vndrs_mean_scan(Batch_in->fullscans[I]->I2);
        PA_mean_vector->data[I] = vndrs_mean_scan(Batch_in->fullscans[I]->PA);
        PB_mean_vector->data[I] = vndrs_mean_scan(Batch_in->fullscans[I]->PB);
    }
    
}


/*
    Compute and removes computed offsets from the on-source batch - pkervell 10-Oct-02
*/
void vndrs_remove_computed_offsets(batch *Batch_off, /* input/output */
                               batch *Batch_A, /* input/output */
                               batch *Batch_B, /* input/output */
                               batch *Batch_on, /* input */
                               matrix_double *Matrix, /*input*/
                               vector_double *Off_offsets) /*out*/
{
    /* TO BE CHANGED IF WE WANT TO USE THIS !! */
    /* vndrs_evaluate_off_source_background(batch *Batch_on,
                                          batch *Batch_off,
                                          matrix_double *Matrix,
                                          vector_double *Background_PA,
                                          vector_double *Background_PB);*/                                         
    /* Removal of background offsets */
    vndrs_remove_offsets(Batch_on, Off_offsets);
    vndrs_remove_offsets(Batch_off, Off_offsets);
    vndrs_remove_offsets(Batch_A, Off_offsets);
    vndrs_remove_offsets(Batch_B, Off_offsets);

}


/*
    Independant estimation of the background offset level (not directly from off sourcce batch)
    pkervell - 24-Sep-02
*/
void vndrs_evaluate_off_source_background(batch *Batch_on,
                                          batch *Batch_A,
                                          batch *Batch_B,
                                          batch *Batch_off,
                                          matrix_double *Matrix,
                                          vector_double *Background_PA_off, /* fluctuations of the original off-src */
                                          vector_double *Background_PB_off,
                                          vector_double *Background_I1, /*out*/
                                          vector_double *Background_I2, /*out*/
                                          vector_double *Background_PA, /* fluctuations of the on-src */
                                          vector_double *Background_PB) /*out*/
{
    double kappa1A, kappa1B, kappa2A, kappa2B, buffer;
    vector_double *Means, *dummy;
    double
        N_off_min, N_off_max,
        I1_onsrc, I2_onsrc, PA_onsrc, PB_onsrc,
        I1_A, I2_A, PA_A, PB_A, 
        I1_B, I2_B, PA_B, PB_B, 
        BG_PA_A1, BG_PB_B1, BG_PA_A2, BG_PB_B2,
        I1_offsrc, I2_offsrc, PA_offsrc, PB_offsrc; /* the initial off source offsets */
    long N_scans_on, N_scans_off, I;

    N_scans_on = Batch_on->number_of_scans;
    N_scans_off = Batch_off->number_of_scans;

    /* Acceptable limit of the computed background, in number of times the original offsrc bg */
    /* pkervell 18-Nov-02 */
    N_off_min = 0.5;
    N_off_max = 1.5;
    
    Means = vndrs_calloc_vector_double(4);
    dummy = vndrs_calloc_vector_double(4);
    
    vndrs_get_transfer_matrix_coefficients(Matrix, &kappa1A, &kappa1B, &kappa2A, &kappa2B);

    for(I=0;I<N_scans_off;I++)
    {
        Background_PA_off->data[I] = vndrs_mean_scan(Batch_off->fullscans[I]->PA);
        Background_PB_off->data[I] = vndrs_mean_scan(Batch_off->fullscans[I]->PB);
    }
    
    vndrs_batch_statistics(Batch_off,
                            Means,
                            dummy,
                            dummy,
                            dummy);

    I1_offsrc = Means->data[0];
    I2_offsrc = Means->data[1];
    PA_offsrc = Means->data[2];
    PB_offsrc = Means->data[3];

    vndrs_batch_statistics(Batch_A,
                            Means,
                            dummy,
                            dummy,
                            dummy);

    I1_A = Means->data[0];
    I2_A = Means->data[1];
    PA_A = Means->data[2];
    PB_A = Means->data[3];
    
    BG_PA_A1 = PA_A - ((I1_A - I1_offsrc) - kappa1B*(PB_A-PB_offsrc)) / kappa1A; 
    BG_PA_A2 = PA_A - ((I2_A - I2_offsrc) - kappa2B*(PB_A-PB_offsrc)) / kappa2A; 
    
    vndrs_batch_statistics(Batch_B,
                            Means,
                            dummy,
                            dummy,
                            dummy);

    I1_B = Means->data[0];
    I2_B = Means->data[1];
    PA_B = Means->data[2];
    PB_B = Means->data[3];
    
    BG_PB_B1 = PB_B - ((I1_B - I1_offsrc) - kappa1A*(PA_B-PA_offsrc)) / kappa1B; 
    BG_PB_B2 = PB_B - ((I2_B - I2_offsrc) - kappa2A*(PA_B-PA_offsrc)) / kappa2B; 

    /* printf(" BG_PA_A1 = %f, BG_PA_A2 = %f, BG_PB_B1 = %f, BG_PB_B2 = %f\n", BG_PA_A1, BG_PA_A2, BG_PB_B1, BG_PB_B2);*/
    for(I=0;I<N_scans_on;I++)
    {
        PA_onsrc = vndrs_mean_scan(Batch_on->fullscans[I]->PA);
        PB_onsrc = vndrs_mean_scan(Batch_on->fullscans[I]->PB);
        I1_onsrc = vndrs_mean_scan(Batch_on->fullscans[I]->I1);
        I2_onsrc = vndrs_mean_scan(Batch_on->fullscans[I]->I2);
 
        /* I1_off and I2_off background estimation */
        /* (I1-I1_off) = kappa1A * (PA-PA_off) + kappa1B * (PB-PB_off)
            (I2-I2_off) = kappa2A * (PA-PA_off) + kappa2B * (PB-PB_off)
           
            (I1 - I2 - I1_off + I2_off) = (kappa1A-kappa2A)*(PA-PA_off) + (kappa1B-kappa2B)*(PB-PB_off)
            (I2_off - I1_off) = (kappa1A-kappa2A)*(PA-PA_off) + (kappa1B-kappa2B)*(PB-PB_off) - (I1 - I2)
           
            (I1 + I2 - (I1_off + I2_off)) = (kappa1A+kappa2A)*(PA-PA_off) + (kappa1B+kappa2B)*(PB-PB_off)
            (I2_off + I1_off) = (I1 + I2) - (kappa1A+kappa2A)*(PA-PA_off) - (kappa1B+kappa2B)*(PB-PB_off)
           
            I2_off = 1/2 * ((I1 + I2) - (I1 - I2) - 2*kappa2A*(PA-PA_off) - 2*kappa2B*(PB-PB_off))
            I1_off = 1/2 * ((I1 + I2) + (I1 - I2) - 2*kappa1A*(PA-PA_off) - 2*kappa1B*(PB-PB_off))
        */
       
        /* I1_off */
        Background_I1->data[I] = 0.5 * ((I1_onsrc + I2_onsrc) + (I1_onsrc - I2_onsrc) -
                    2*kappa1A*(PA_onsrc - PA_offsrc) - 2*kappa1B*(PB_onsrc - PB_offsrc));
        if(Background_I1->data[I] > N_off_max*I1_offsrc) Background_I1->data[I] = N_off_max*I1_offsrc;
        if(Background_I1->data[I] < N_off_min*I1_offsrc) Background_I1->data[I] = N_off_min*I1_offsrc;
        
        /* I2_off */
        Background_I2->data[I] = 0.5 * ((I1_onsrc + I2_onsrc) - (I1_onsrc - I2_onsrc) -
                    2*kappa2A*(PA_onsrc - PA_offsrc) - 2*kappa2B*(PB_onsrc - PB_offsrc));
        /* if(I2_offsrc < Background_I2->data[I]) Background_I2->data[I] = I2_offsrc;*/
        if(Background_I2->data[I] > N_off_max*I2_offsrc) Background_I2->data[I] = N_off_max*I2_offsrc;
        if(Background_I2->data[I] < N_off_min*I2_offsrc) Background_I2->data[I] = N_off_min*I2_offsrc;
   
        /* We limit here the amplitude of the variations of the computed background from half of 
        the off-source value (minimum) to the off-source value (maximum). */
        
        /* printf("Scan number: %ld, Background I1 orig : %f, Computed : %f \n", I, I1_offsrc, Background_I1->data[I]);
        printf("Scan number: %ld, Background I2 orig : %f, Computed : %f \n\n", I, I2_offsrc, Background_I2->data[I]);*/
    }

    for(I=0;I<N_scans_on;I++)
    {
        PA_onsrc = vndrs_mean_scan(Batch_on->fullscans[I]->PA);
        PB_onsrc = vndrs_mean_scan(Batch_on->fullscans[I]->PB);
        I1_onsrc = vndrs_mean_scan(Batch_on->fullscans[I]->I1);
        I2_onsrc = vndrs_mean_scan(Batch_on->fullscans[I]->I2);
        
        /* PA_off */
        Background_PA->data[I] = PA_onsrc - ((I1_onsrc-I1_offsrc) - kappa1B * (PB_onsrc - PB_offsrc)) / kappa1A;
        buffer = PA_onsrc - ((I2_onsrc-I2_offsrc) - kappa2B * (PB_onsrc - PB_offsrc)) / kappa2A;
        if(buffer < Background_PA->data[I]) Background_PA->data[I] = buffer;
        /* if(PA_offsrc < Background_PA->data[I]) Background_PA->data[I] = PA_offsrc;
        if(BG_PA_A1 < Background_PA->data[I]) Background_PA->data[I] = BG_PA_A1;
        if(BG_PA_A2 < Background_PA->data[I]) Background_PA->data[I] = BG_PA_A2;*/
        if(Background_PA->data[I] > N_off_max*PA_offsrc) Background_PA->data[I] = N_off_max*PA_offsrc;
        if(Background_PA->data[I] > N_off_max*BG_PA_A1) Background_PA->data[I] = N_off_max*BG_PA_A1;
        if(Background_PA->data[I] > N_off_max*BG_PA_A2) Background_PA->data[I] = N_off_max*BG_PA_A2;
        if(Background_PA->data[I] < N_off_min*PA_offsrc) Background_PA->data[I] = N_off_min*PA_offsrc;
        
        /* We limit here the amplitude of the variations of the computed background from half of 
        the off-source value (minimum) to the off-source value (maximum). */

        /* PB_off */
        Background_PB->data[I] = PB_onsrc - ((I1_onsrc-I1_offsrc) - kappa1A * (PA_onsrc - PA_offsrc)) / kappa1B;
        buffer = PB_onsrc - ((I2_onsrc-I2_offsrc) - kappa2A * (PA_onsrc - PA_offsrc)) / kappa2B;
        if(buffer < Background_PB->data[I]) Background_PB->data[I] = buffer;
        /* if(PB_offsrc < Background_PB->data[I]) Background_PB->data[I] = PB_offsrc;
        if(BG_PB_B1 < Background_PB->data[I]) Background_PB->data[I] = BG_PB_B1;
        if(BG_PB_B2 < Background_PB->data[I]) Background_PB->data[I] = BG_PB_B2;*/
        if(N_off_max*PB_offsrc < Background_PB->data[I]) Background_PB->data[I] = N_off_max*PB_offsrc;
        if(N_off_max*BG_PB_B1 < Background_PB->data[I]) Background_PB->data[I] = N_off_max*BG_PB_B1;
        if(N_off_max*BG_PB_B2 < Background_PB->data[I]) Background_PB->data[I] = N_off_max*BG_PB_B2;
        if(Background_PB->data[I] < N_off_min*PB_offsrc) Background_PB->data[I] = N_off_min*PB_offsrc;
         
        /* Modification pkervell - 12-Oct-02 : we ABSOLUTELY do not want to overestimate the background, as
        it causes severe instability of the visibility. This is not easy to estimate the BG level with this
        method, but I could not find a better one for the moment unfortunately... Therefore, it is important
        to make all the possible validity checks (not too large, not too small,...) on the obtained values. */
    }


    vndrs_free_vector_double(Means);
    vndrs_free_vector_double(dummy);
}

/*
    Removal of the background offsets scan by scan using the vectors of computed backgrounds.
    Only PA and PB backgrounds are removed scan by scan (they are the only important in the
    calibration), while the I1 and I2 backgrounds are removed using the off-source values.
    - pkervell 10-Oct-02.
    The batch input is replaced.
*/
void vndrs_remove_backgrounds_per_scan(batch *Batch_in_out,
                          vector_double *Background_I1,
                          vector_double *Background_I2,
                          vector_double *Background_PA,
                          vector_double *Background_PB)
{
    long N_scans, I;
    
    N_scans = Batch_in_out->number_of_scans;
    
    for(I=0;I<N_scans;I++)
    {
        vndrs_vector_add_constant(Batch_in_out->fullscans[I]->I1, -Background_I1->data[I]);
        vndrs_vector_add_constant(Batch_in_out->fullscans[I]->I2, -Background_I2->data[I]);
        vndrs_vector_add_constant(Batch_in_out->fullscans[I]->PA, -Background_PA->data[I]);
        vndrs_vector_add_constant(Batch_in_out->fullscans[I]->PB, -Background_PB->data[I]);
    }

}

/*
    Removal of the background offsets (globally...) - pkervell 04-Oct-73 (my day of birth ! :-)).
    The batch input is replaced.
*/
void vndrs_remove_offsets(batch *Batch_in_out,
                          vector_double *Offsets)
{
    long N_scans, I;
    
    N_scans = Batch_in_out->number_of_scans;
    
    for(I=0;I<N_scans;I++)
    {
        vndrs_vector_add_constant(Batch_in_out->fullscans[I]->I1, -Offsets->data[0]);
        vndrs_vector_add_constant(Batch_in_out->fullscans[I]->I2, -Offsets->data[1]);
        vndrs_vector_add_constant(Batch_in_out->fullscans[I]->PA, -Offsets->data[2]);
        vndrs_vector_add_constant(Batch_in_out->fullscans[I]->PB, -Offsets->data[3]);
    }

}


/* Computes the frequency scale (in Hertz, or whatever unit is Frame_rate) of the PSD -
the resulting vector contains only the first half of the frequency scale, up to the Nyquist freq,
that is supposed to be the size of the input vector Freq_scale.
pkervell - 03-Oct-02 */
void vndrs_compute_freq_scale(double Frame_rate,
                              vector_double *Freq_scale)
{
    double N_frames;
    long I;
    
    N_frames = Freq_scale-> dimension;
    
    for(I=0;I<N_frames;I++)
        Freq_scale->data[I] = (double)I * Frame_rate * 0.5/((double)N_frames);
    
}

/* Computes the average power spectra of the five signals (incl. QL) over one batch and
stores it in a fullscan structure. pkervell 03-Oct-02 */

void vndrs_compute_average_power_spectra(batch *Batch_in,
                                         full_scan *Power_spectra)
{
    long N_frames, N_scans, I;
    vector_double *buffer_pow_I1, *buffer_pow_I2, *buffer_pow_PA, *buffer_pow_PB, *buffer_pow_QL;
    vector_double *avg_pow_I1, *avg_pow_I2, *avg_pow_PA, *avg_pow_PB, *avg_pow_QL;
    
    N_frames = Batch_in->fullscans[0]->I1->dimension;
    N_scans = Batch_in->number_of_scans;
    
    avg_pow_I1 = vndrs_calloc_vector_double(N_frames);
    avg_pow_I2 = vndrs_calloc_vector_double(N_frames);
    avg_pow_PA = vndrs_calloc_vector_double(N_frames);
    avg_pow_PB = vndrs_calloc_vector_double(N_frames);
    avg_pow_QL = vndrs_calloc_vector_double(N_frames);
    
    for(I=0;I<N_scans;I++)
    {
        buffer_pow_I1 = vndrs_power_spectrum(Batch_in->fullscans[I]->I1);
        buffer_pow_I2 = vndrs_power_spectrum(Batch_in->fullscans[I]->I2);
        buffer_pow_PA = vndrs_power_spectrum(Batch_in->fullscans[I]->PA);
        buffer_pow_PB = vndrs_power_spectrum(Batch_in->fullscans[I]->PB);
        buffer_pow_QL = vndrs_power_spectrum(Batch_in->fullscans[I]->QL);
        
        vndrs_vector_add(avg_pow_I1, buffer_pow_I1);
        vndrs_vector_add(avg_pow_I2, buffer_pow_I2);
        vndrs_vector_add(avg_pow_PA, buffer_pow_PA);
        vndrs_vector_add(avg_pow_PB, buffer_pow_PB);
        vndrs_vector_add(avg_pow_QL, buffer_pow_QL);
        
        /* Memory was allocated by vndrs_power_spectrum */
        vndrs_free_vector_double(buffer_pow_I1);
        vndrs_free_vector_double(buffer_pow_I2);
        vndrs_free_vector_double(buffer_pow_PA);
        vndrs_free_vector_double(buffer_pow_PB);
        vndrs_free_vector_double(buffer_pow_QL);
    }
    
    vndrs_vector_scale(avg_pow_I1, 1.0/N_scans);
    vndrs_vector_scale(avg_pow_I2, 1.0/N_scans);
    vndrs_vector_scale(avg_pow_PA, 1.0/N_scans);
    vndrs_vector_scale(avg_pow_PB, 1.0/N_scans);
    vndrs_vector_scale(avg_pow_QL, 1.0/N_scans);

    vndrs_vector_memcpy(Power_spectra->I1, avg_pow_I1);
    vndrs_vector_memcpy(Power_spectra->I2, avg_pow_I2);
    vndrs_vector_memcpy(Power_spectra->PA, avg_pow_PA);
    vndrs_vector_memcpy(Power_spectra->PB, avg_pow_PB);
    vndrs_vector_memcpy(Power_spectra->QL, avg_pow_QL);

    vndrs_free_vector_double(avg_pow_I1);
    vndrs_free_vector_double(avg_pow_I2);
    vndrs_free_vector_double(avg_pow_PA);
    vndrs_free_vector_double(avg_pow_PB);
    vndrs_free_vector_double(avg_pow_QL);

}


/*
    The function below selects the scans based on their visibility.
    The limit is computed inside in terms of number of times the standard deviation.
    
    This is mandatory, because we are facing something like a Cauchy distribution,
    one of the statistical worst nightmares. To go back to something like a gaussian
    distribution, we have to remove the extreme outliers (3 or 4 sigmas for example).
*/

double vndrs_eliminate_visibility_outliers(vector_double *mu2_raw,      /* input */
                                        double N_sigmas,
                                        vector_double *Flag_processed, /* output */
                                        vector_double *removed_indices) /* output */
{
    vector_double *non_zero_mu2_raw;
    double Median_mu2, Std_mu2, lim_up, lim_down;
    long I, N_scans, N_nonzero_scans, index;
    double flag_remove;
                                 
    N_scans = Flag_processed->dimension;
    
    /* Memory allocated below */
    non_zero_mu2_raw = vndrs_extract_nonzero(mu2_raw, Flag_processed);
    N_nonzero_scans = non_zero_mu2_raw->dimension;
    
    Std_mu2 = vndrs_stddev_scan(non_zero_mu2_raw);
    Median_mu2 = vndrs_median_scan(non_zero_mu2_raw);
    
    lim_up = Median_mu2 + N_sigmas * Std_mu2;
    lim_down = Median_mu2 - N_sigmas * Std_mu2;
    
    index = 0;
    flag_remove = 0;

    for(I=0;I<N_scans;I++)
    {
        if(Flag_processed->data[I] > 0)
            if(mu2_raw->data[I]>lim_up | mu2_raw->data[I]<lim_down)
            {
                flag_remove++;
                Flag_processed->data[I] = -2;
                removed_indices->data[index] = I;
                index++;
            }
    }

    vndrs_free_vector_double(non_zero_mu2_raw);

    return flag_remove;
}




/*
    The function below selects the scans based on their photometry. The very high
    photometry scans are removed from the final sample, in order to avoid biasing problems.
    The limit is computed inside in terms of number of times the median distance to the
    median (the distribution is skewed).
*/

long vndrs_eliminate_extreme_scans(vector_double *signal_p,       /* input */
                                   double N_med_dist,
                                   vector_double *Flag_processed, /* output */
                                   vector_double *removed_indices) /* output */
{
    vector_double *non_zero_signals;
    double Med_dist, Med_val, lim_up;
    long I, N_scans, index;
    double flag_remove;
    
    N_scans = Flag_processed->dimension;
    
    /* Memory allocated below */
    non_zero_signals = vndrs_extract_nonzero(signal_p, Flag_processed);

    Med_val = vndrs_median_scan(non_zero_signals); /* The reference median value */
    vndrs_vector_add_constant(non_zero_signals, -Med_val); /* distances to the median */
    vndrs_vector_mul(non_zero_signals, non_zero_signals); /* squared */
    Med_dist = sqrt(vndrs_median_scan(non_zero_signals)); /* The reference median distance */
    
    lim_up = Med_val + N_med_dist * Med_dist;
    index = 0;
    flag_remove = 0;
    
    for(I=0;I<N_scans;I++)
    {
        if(signal_p->data[I] > lim_up)
        {
            flag_remove++;
            Flag_processed->data[I] = -3;
            removed_indices->data[index] = I;
            index++;
        }
    }

    vndrs_free_vector_double(non_zero_signals);

    return flag_remove;
}


/*
     Computation of the synthetic photometric signals from the
     interferometric outputs.
     pkervell 17-Sep-02
*/

void vndrs_prepare_synthetic_photom(vector_double *I1_cal,
                                    vector_double *I2_cal,
                                    matrix_double *Matrix_coefs,
                                    /* outputs */
                                    vector_double *PA_synth,
                                    vector_double *PB_synth)
{
    individual_scan *I11, *I12, *I21, *I22;
    double coef, kappa_1A, kappa_2B, kappa_1B, kappa_2A;
    long N_frames;
    
    N_frames = I1_cal->dimension;
    
    I11  = vndrs_calloc_vector_double(N_frames);
    I12  = vndrs_calloc_vector_double(N_frames);
    I21  = vndrs_calloc_vector_double(N_frames);
    I22  = vndrs_calloc_vector_double(N_frames);

    /* Get the matrix coefficients */
    vndrs_get_transfer_matrix_coefficients(Matrix_coefs, &kappa_1A, &kappa_1B, &kappa_2A, &kappa_2B);

    coef = 1.0/(kappa_1A*kappa_2B-kappa_1B*kappa_2A);

    vndrs_vector_memcpy(I11, I1_cal);
    vndrs_vector_memcpy(I12, I2_cal);
    vndrs_vector_memcpy(I21, I2_cal);
    vndrs_vector_memcpy(I22, I1_cal);
	
    vndrs_vector_scale(I11,kappa_2B);
    vndrs_vector_scale(I12,kappa_1B);
    vndrs_vector_sub(I11,I12);
    vndrs_vector_scale(I11, coef);  /* "synthetic" PA */
    vndrs_vector_memcpy(PA_synth, I11);
    
    vndrs_vector_scale(I21,kappa_1A);
    vndrs_vector_scale(I22,kappa_2A);
    vndrs_vector_sub(I21,I22);
    vndrs_vector_scale(I21, coef);  /* "synthetic" PB */
    vndrs_vector_memcpy(PB_synth, I21);

    vndrs_free_vector_double(I11);  /* added pkervell - 23-apr-02 */
    vndrs_free_vector_double(I21);
    vndrs_free_vector_double(I12);
    vndrs_free_vector_double(I22);
}


/*
     Selection of type and computation of the matrix.
     Introduced to make the main procedure code more compact.
     pkervell 13-Sep-02
*/
void vndrs_select_matrix(batch *Batch_A,
                         batch *Batch_B,
                         batch *Batch_on,
                         vector_double *Off_offsets,
                         vector_double *WA_mean,
                         vector_double *WB_mean,
                         int ionic_in,
                         matrix_double *Average_Matrix,
                         matrix_double *Standard_dev_Matrix,
                         vector_double *Relative_MSE,
                         int verbose_on)
{

    double kappa1A, kappa1B, kappa2A, kappa2B;
    double kappa1A_err, kappa1B_err, kappa2A_err, kappa2B_err;
    
    sprintf(vndrs_protocol.text,"# KAPPA MATRIX COMPUTATION BY MIXED BEAMS AVERAGE \n");
    vndrs_runtime_protocol();
    vndrs_get_transfer_matrix(Batch_A,
                                Batch_B,
                                Off_offsets,
                                WA_mean,
                                WB_mean,
                                ionic_in,
                                Average_Matrix,
                                Standard_dev_Matrix,
                                Relative_MSE,
                                verbose_on
                                );


    /* Save the matrix values in easier to read variables */
    vndrs_get_transfer_matrix_coefficients(Average_Matrix, &kappa1A, &kappa1B, &kappa2A, &kappa2B);
    vndrs_get_transfer_matrix_coefficients(Standard_dev_Matrix, &kappa1A_err, &kappa1B_err, &kappa2A_err, &kappa2B_err);
}

/*
     Removal of the background from the processed combined Fourier PSD
     pkervell 09-Sep-02
*/

void vndrs_remove_psd_background(long N_frames,
                                vector_double *Pow_spec_avg,
                                double Wavenum_min_lowband,
                                double Wavenum_max_lowband,
                                double Wavenum_min_hi_band,
                                double Wavenum_max_hi_band,
                                int debias,
                                double slope_limit,
                                double Frame_rate,
                                double Fringe_speed,
                                /* outputs */
                                double *slope_background,
                                vector_double *Background_PSD,
                                vector_double *Pow_spec_avg_corrected)
{
    long I, index_freq_min_low, index_freq_min_high,
    index_freq_max_low, index_freq_max_high, N_indices;
    double buffer_slope, zero_point, dummy, avg_value;
    vector_double *indices_freq, *values;

    index_freq_min_low  = (long)(Wavenum_min_lowband/fabs(Frame_rate / Fringe_speed / N_frames));
    index_freq_max_low  = (long)(Wavenum_max_lowband/fabs(Frame_rate / Fringe_speed / N_frames));
    index_freq_min_high = (long)(Wavenum_min_hi_band/fabs(Frame_rate / Fringe_speed / N_frames));
    index_freq_max_high = (long)(Wavenum_max_hi_band/fabs(Frame_rate / Fringe_speed / N_frames));
    N_indices = (index_freq_max_low - index_freq_min_low) +
                (index_freq_max_high- index_freq_min_high);

    indices_freq = vndrs_calloc_vector_double(N_indices);
    values = vndrs_calloc_vector_double(N_indices);
    
    /* filling the indices vector */
    for(I=index_freq_min_low;I<index_freq_max_low;I++)
        indices_freq->data[I-index_freq_min_low] = I;
    for(I=index_freq_min_high;I<index_freq_max_high;I++)
        indices_freq->data[I-index_freq_min_high+(index_freq_max_low-index_freq_min_low)] = I;
    
    /* filling the values vector - linear model */
    for(I=index_freq_min_low;I<index_freq_max_low;I++)
            values->data[I-index_freq_min_low] = Pow_spec_avg->data[I];
    for(I=index_freq_min_high;I<index_freq_max_high;I++)
            values->data[I-index_freq_min_high+(index_freq_max_low-index_freq_min_low)] = Pow_spec_avg->data[I];


    /* Least squares fit of the BG */
    vndrs_computes_least_squares(indices_freq,
                                 values,
                                 &buffer_slope,
                                 &zero_point,
                                 &dummy);

    *slope_background = buffer_slope;
    
    /* If the slope of the computed background is too large, then we subtract the BG using only the
    higher part of the spectrum (this means that we have photometric contamination). */
    if(debias == 0 && fabs(*slope_background/(Frame_rate/N_frames)) < slope_limit)
    {
        /* linear model */
        for(I=0;I<(N_frames/2);I++)
            Background_PSD->data[I] = zero_point + buffer_slope * I;
    }
    else /* if we debias by the off-source, we may have some photometric contamination in the lower band,
            therefore we take only the average of the upper band values to remove them. pkervell 11-Oct-02 */
    {
        avg_value = 0.0;
        for(I=index_freq_min_high;I<index_freq_max_high;I++)
            avg_value += Pow_spec_avg->data[I]/(index_freq_max_high- index_freq_min_high);
        
        for(I=0;I<(N_frames/2);I++)
            Background_PSD->data[I] = avg_value;
    }
    
    /* actual subtraction */
    vndrs_vector_memcpy(Pow_spec_avg_corrected, Pow_spec_avg);
    vndrs_vector_sub(Pow_spec_avg_corrected, Background_PSD);

    vndrs_free_vector_double(indices_freq); 
    vndrs_free_vector_double(values);   
}


/*
    Procedure to check if we have really fringes in our scan. Multiple checks are done to
    ensure that the signal has all the characteristics of a fringe packet. It should return a
    code corresponding to the reason for rejection.
    added pkervell 07-Sep-02
*/
double vndrs_check_fringes(/* inputs */
                           matrix_double *norm_t_wavelets,
                           long N_frames_cut,
                           double Frame_rate,
                           double Fringe_speed,
                           double fringes_length,
                           double Wavenum_theo,
                           double Wavenum_min,
                           double Wavenum_max,
                           double Wavenum_min_out,
                           double Wavenum_max_out,                                
                           double Wavenum_interval_cut,
                           /* output */
                           vector_double *fringes_centering_outputs)
{
    /* Declaration of variables */
    int fringes_flag = 1;
    double Sample_interval, Element_freq, wavelets_fringes_center;
    vector_double *buffer_peak, *pos_max;
    
    /* Values measured on the scan */
    double 	in_to_outband_ratio,
                fringes_center,
                residual,
                peak_max,
                fwhm_time,
                pos_freq,
                fwhm_freq,
                background_fit;
    /* Theoretical values */
    double 	fwhm_time_th,
                pos_freq_th,
                fwhm_freq_th;    
    /* Threshold levels */
    double      /*threshold_ratio = 0,*/	/* minimum in to out band ratio - not used now */
                threshold_peak_max = 0.0,  	/* QL peak maximum lower limit (code 2) */
                threshold_fwhm_time = 50.0,  	/* in percents of the theoretical value (code 3) */
                threshold_pos_freq = 30.0, 	/* in percents of the theoretical value (code 4) */
                threshold_fwhm_freq = 40.0, 	/* in percents of the theoretical value (code 5) */
                threshold_piston = 2; 	/* residual as computed by the gauss2d function (code 6) not used */

    buffer_peak =   vndrs_calloc_vector_double(10); /* only 8 used for now */
    pos_max = vndrs_calloc_vector_double(3);
    
    /* ---- */
    
    Sample_interval = fabs(1.0*Fringe_speed / Frame_rate);
    Element_freq = 1. / (N_frames_cut * Sample_interval);
        
    /* Theoretical values */
    fwhm_time_th = 15.0;  /* expressed in microns - TO BE CHECKED */
    pos_freq_th =  (Wavenum_theo)/(Element_freq);
    fwhm_freq_th = (Wavenum_theo)/(6.0*Element_freq); /* the factor 6 is EXPERIMENTAL ! */

    /* fringes_centering_outputs->data[0] not used for now */
    fringes_centering_outputs->data[0] = 0.0;

    /* Cancellation of the fringes_length/2 side parts of the wl transform to avoid edge
    effects in the detection of the peak - pkervell 17-Oct-02 */
    vndrs_cancel_wl_edges(norm_t_wavelets, (long)(fringes_length/2));
    
    /* Computes a fit of the wavelets power peak - segransan Jul-02 - modified pkervell Aug-02 */
    /*if(fringes_centering_outputs->data[0] >= threshold_ratio)*/
    vndrs_wave_select(norm_t_wavelets, buffer_peak);

    /* Where are the fringes ? In principle already at the center of the scan !!
        we check only processed fringes... */
    /*vndrs_where2_max(norm_t_wavelets, pos_max); - modified pkervell 17-Oct-02 */
    /* wavelets_fringes_center = (long)pos_max->data[0]- N_frames_cut/2; in pixels from center of QL */
    wavelets_fringes_center = 0.0;
    
    /* fringes_centering_outputs->data[1] = wavelets_fringes_center; position of the fringes in time domain */
    fringes_centering_outputs->data[2] = buffer_peak->data[0]; /* residual of the gauss2D fit of the peak */
    fringes_centering_outputs->data[3] = buffer_peak->data[1]; /* peak maximum value */
    fringes_centering_outputs->data[4] = buffer_peak->data[2]; /* fwhm of peak in time domain */    
    fringes_centering_outputs->data[5] = buffer_peak->data[3]; /* position in frequency domain */
    fringes_centering_outputs->data[6] = buffer_peak->data[4]; /* fwhm in frequency domain */
    fringes_centering_outputs->data[7] = buffer_peak->data[5]; /* background of fit */

    /* Values from the scan */
    in_to_outband_ratio = fringes_centering_outputs->data[0];
    fringes_center = fringes_centering_outputs->data[1];	  /* position of the fringes in time domain (not used) */
    residual = fringes_centering_outputs->data[2]; 		  /* residual of the gauss2D fit of the peak */
    peak_max = fringes_centering_outputs->data[3];		  /* peak in wl maximum value (ADU) */
    fwhm_time = fringes_centering_outputs->data[4] * 1000000.0 * (Fringe_speed / Frame_rate);
                                                                  /* fwhm of fringe peak in time domain (microns) */    
    pos_freq = fringes_centering_outputs->data[5]; 	  	  /* position in frequency domain (pixels) */
    fwhm_freq = fringes_centering_outputs->data[6]; 		  /* fwhm in frequency domain (pixels) */
    background_fit = fringes_centering_outputs->data[7]; 	  /* background of fit (ADUs) */
    

    /* printf(" pos_freq_th  = %f , pos_freq  = %f \n", pos_freq_th, pos_freq);
    printf(" thres_peak   = %f , peak_max  = %f \n", threshold_peak_max, peak_max);
    printf(" fwhm_time_th = %f , fwhm_time = %f \n", fwhm_time_th, fwhm_time);
    printf(" fwhm_freq_th = %f , fwhm_freq = %f \n", fwhm_freq_th, fwhm_freq);*/
    
    /* Actual tests. There is a hierarchy : if the first is not ok, the others are not performed */
    /* fringe peak too faint ? */
    if(peak_max < threshold_peak_max && fringes_flag == 1) fringes_flag = -2;
    /* correct width for the fringe packet ? */
    if(fwhm_time > (fwhm_time_th*(1 + threshold_fwhm_time/100.0))
     | fwhm_time < (fwhm_time_th*(1 - threshold_fwhm_time/100.0))
     && fringes_flag == 1) fringes_flag = -3;
    /* correct position in the frequency domain ? */
    if(pos_freq > (pos_freq_th*(1 + threshold_pos_freq/100.0))
     | pos_freq < (pos_freq_th*(1 - threshold_pos_freq/100.0))
     && fringes_flag == 1) fringes_flag = -4;
    /* peak in the PSD is the correct width ? */
    if(fwhm_freq > (fwhm_freq_th*(1 + threshold_fwhm_freq/100.0))
     | fwhm_freq < (fwhm_freq_th*(1 - threshold_fwhm_freq/100.0))
     && fringes_flag == 1) fringes_flag = -5;
    /* are residuals of the gauss2d fit too large ? */
    if(residual >= threshold_piston && fringes_flag == 1) fringes_flag = -6;
    
    vndrs_free_vector_double(buffer_peak);
    vndrs_free_vector_double(pos_max);

    return fringes_flag;
}

/*
    Detection of fringes and centering
    added pkervell 01-Sep-02
    This function detects and centers the fringes on one scan, and outputs the
    cut scan and the detection flag.
*/

double vndrs_detect_center_fringes(/* inputs */
                                full_scan *Scan_on_precut, /* output for QL */
                                double Apod_blockout,
                                double Apod_transition,
                                long N_frames_cut,
                                double param_wl,  /* usually 17.0 fringes */
                                double fringes_length,
                                int ionic_in,
                                double Wavenum_min,
                                double Wavenum_max,
                                double Wavenum_min_out,
                                double Wavenum_max_out,
                                double Sample_interval_precut,
                                double Wavenum_interval_precut,
                                /* output */
                                full_scan *Scan_on_cut,
                                double *fringes_off_center
                                )
{
    long N_frames_precut;
    long wavelets_fringes_center;
    vector_double *coi_QL, *wave_time_QL, *pos_max_QL;
    matrix_complex *t_wavelets_QL;
    matrix_double *norm_t_wavelets_QL;
    individual_scan *quicklook_signal, *quicklook_noise;
        
    N_frames_precut = Scan_on_precut->I1->dimension;

    /*  MEMORY ALLOCATION FOR QUICK LOOK WAVELETS */
    coi_QL = vndrs_calloc_vector_double(N_frames_precut);
    wave_time_QL = vndrs_calloc_vector_double(N_frames_precut);
    t_wavelets_QL = vndrs_calloc_matrix_complex(N_frames_precut,N_frames_precut/2);
    pos_max_QL = vndrs_calloc_vector_double(3);
    quicklook_signal = vndrs_calloc_vector_double(N_frames_precut);
    quicklook_noise = vndrs_calloc_vector_double(N_frames_precut);
    
    /* Recomputation of a quicklook signal from the original signals */
    /* This procedure vndrs_quicklook presents no memory leak ! :-) */

    vndrs_quicklook(Scan_on_precut,
                    Apod_blockout,
                    Apod_transition,
                    ionic_in,
                    Wavenum_min,
                    Wavenum_max,
                    Wavenum_min_out,
                    Wavenum_max_out,
                    Sample_interval_precut,
                    quicklook_signal,
                    quicklook_noise);
   
    /*
    vndrs_fprintf_vector_double(Scan_on_precut->QL, "original_quicklook_signal.dat");
    vndrs_fprintf_vector_double(quicklook_signal, "quicklook_signal.dat");
    */
    
    vndrs_wavelet(N_frames_precut, param_wl,
                    0,
                    N_frames_precut/2.0,
                    coi_QL,
                    quicklook_signal,
                    t_wavelets_QL);                

    norm_t_wavelets_QL = vndrs_wavelet_power_spectrum(t_wavelets_QL);
    /* BEWARE : changed to have MEMORY ALLOCATED INSIDE FUNCTION - pkervell - 04-Aug-02 */

    /* Where are the fringes ? At the position of the maximum in the wl spectrum ! */
    vndrs_where2_max(norm_t_wavelets_QL, pos_max_QL);

    wavelets_fringes_center = (long)pos_max_QL->data[0]- N_frames_precut/2 + 1; /*ecart en pixel au centre de QL*/
    
    /* We check here that we can center the fringes, knowing their offset position */
    if((pos_max_QL->data[0]<=N_frames_cut/2) | (pos_max_QL->data[0]>(N_frames_precut - N_frames_cut/2)))
        *fringes_off_center = 1.0; /* fringes cannot be centered properly */
    else
        *fringes_off_center = 0.0;


    /* Now we trim the scan to have the fringes centered, using the fringes position from the wavelets */
    vndrs_trim_scan(Scan_on_precut->I1, N_frames_cut, wavelets_fringes_center,
                    Scan_on_cut->I1);
    vndrs_trim_scan(Scan_on_precut->I2, N_frames_cut, wavelets_fringes_center,
                    Scan_on_cut->I2);
    vndrs_trim_scan(Scan_on_precut->PA, N_frames_cut, wavelets_fringes_center,
                    Scan_on_cut->PA);
    vndrs_trim_scan(Scan_on_precut->PB, N_frames_cut, wavelets_fringes_center,
                    Scan_on_cut->PB);
    vndrs_trim_scan(Scan_on_precut->QL, N_frames_cut, wavelets_fringes_center,
                    Scan_on_cut->QL);


    /* we output the QL signal in the Signal_on_precut quicklook vector */
    vndrs_vector_memcpy(Scan_on_precut->QL, quicklook_signal);
    
    /* Memory release */
    vndrs_free_matrix_double(norm_t_wavelets_QL); /* memory was allocated inside pow spectrum function */
    vndrs_free_matrix_complex(t_wavelets_QL);
    vndrs_free_vector_double(wave_time_QL);
    vndrs_free_vector_double(coi_QL);
    vndrs_free_vector_double(pos_max_QL);
    vndrs_free_vector_double(quicklook_signal);
    vndrs_free_vector_double(quicklook_noise);
    
    return wavelets_fringes_center;

}
  
/*
    Removes the slope of a vector. The N_frames_avg values of the
    vector are considered to evaluate the start and end points.
*/

void vndrs_remove_vector_slope(vector_double *input_vector, /* input vector, REPLACED */
                          long N_frames_avg,
                          vector_double *subtracted_vector) /* the subtracted vector (linear slope) */
{
    long I, N_frames;
    double Start_avg = 0.0, End_avg = 0.0, slope;
    
    N_frames = input_vector->dimension;
    
    for(I=0;I<N_frames_avg;I++)
    {
        Start_avg += input_vector->data[I]/N_frames_avg;
        End_avg += input_vector->data[N_frames - 1 - I]/N_frames_avg;
    }
    
    /* We now have two points to compute the linear vector: (0,Start_avg) and (N_frames,End_avg) */
    slope = (End_avg - Start_avg)/N_frames;
    
    for(I=0;I<N_frames;I++)
    {
        subtracted_vector->data[I] = Start_avg + slope * I;
        input_vector->data[I] -= subtracted_vector->data[I];
    }
}

/*
    added pkervell 01-Sep-02
    Computation of the Quicklook signal used for fringe detection.
    It is slightly different from the quicklook used at the telescope in that there is no
    division by the photometry.
*/
    
void vndrs_quicklook(full_scan *Scan_on_precut, /* input */
                     double Apod_blockout,
                     double Apod_transition,
                     int ionic_in,
                     double Wavenum_min,
                     double Wavenum_max,
                     double Wavenum_min_out,
                     double Wavenum_max_out,
                     double Sample_interval_precut,
                      /* outputs */
                     individual_scan *quicklook_signal,
                     individual_scan *quicklook_noise)
{
    vector_double   *I1, *I2, *PA, *PB,
                    *I1_off, *I2_off, *PA_off, *PB_off,
                    *Apodisation_filter,
                    *QL_filtered, *Noise_filtered, *Filter;
                    
    double a, s1, s2, avg_ql, avg_ql_off, slope,
            Half_bandwidth, Central_wavenumber;
    long N_frames, i;
    
    N_frames = Scan_on_precut->I1->dimension;
    
    /* Memory allocation */
    I1 = vndrs_calloc_vector_double(N_frames);
    I2 = vndrs_calloc_vector_double(N_frames);
    PA = vndrs_calloc_vector_double(N_frames);
    PB = vndrs_calloc_vector_double(N_frames);
    I1_off = vndrs_calloc_vector_double(N_frames);
    I2_off = vndrs_calloc_vector_double(N_frames);
    PA_off = vndrs_calloc_vector_double(N_frames);
    PB_off = vndrs_calloc_vector_double(N_frames);
    
    vndrs_vector_memcpy(I1, Scan_on_precut->I1);
    vndrs_vector_memcpy(I2, Scan_on_precut->I2);
    vndrs_vector_memcpy(PA, Scan_on_precut->PA);
    vndrs_vector_memcpy(PB, Scan_on_precut->PB);

    /* Quicklook = I1 - a * I2 for MONA, or I1 for IONIC - pkervell 07-mar-03 */    
    if(ionic_in == 0) /* added - pkervell - 07-Mar-03 */
    {
        /* computation of 'a' */
        s1 = 0;
        s2 = 0;
        for(i=0; i<N_frames; i++)
        {
            s1 += I1->data[i] * I2->data[i];
            s2 += I2->data[i] * I2->data[i];
        }
        a = fabs(s1/s2);   
        
        /* Noise - not used for now - pkervell - 21-Sep-02 */
        vndrs_vector_scale(I2_off, a);
        vndrs_vector_add(I1_off, I2_off);
    
        /* I1 = I1 - a*I2 */
        vndrs_vector_scale(I2, a);
        vndrs_vector_sub(I1, I2);
    }
    
    /* Removal of the QL signal slope */
    slope = (I1->data[N_frames-1] - I1->data[0])*1.0/(1.0*N_frames);
    for (i=0; i<N_frames; i++)
        I1->data[i] = I1->data[i] - (i*1.0) * slope;

    /* Removal of the QL noise slope */
    slope = (I1_off->data[N_frames-1] - I1_off->data[0])*1.0/(1.0*N_frames);
    for (i=0; i<N_frames; i++)
        I1_off->data[i] = I1_off->data[i] - (i*1.0) * slope;

    
    /* Zeroing of the average of the QL and QL noise */
    avg_ql = vndrs_mean_scan(I1);
    vndrs_vector_add_constant(I1, -avg_ql);
    avg_ql_off = vndrs_mean_scan(I1_off);
    vndrs_vector_add_constant(I1_off, -avg_ql_off);
    
    /* Apodisation of the QL */
    /* Construction of the apodisation filter (MEMORY FOR Apodisation_filter IS ALLOCATED INSIDE THE ROUTINE) */
    Apodisation_filter = vndrs_apodisation_window(N_frames,
                                                  Apod_blockout, 
                                                  Apod_transition);
    vndrs_vector_mul(I1, Apodisation_filter);  
    vndrs_vector_mul(I1_off, Apodisation_filter);  

    /* Frequency filtering */
    /* Prepare the bandpass filter for the interferometric signals */
    Half_bandwidth = (Wavenum_max - Wavenum_min) * 0.5;
    Central_wavenumber = (Wavenum_max + Wavenum_min) * 0.5;
    Filter = vndrs_compute_filter(N_frames,
				  Sample_interval_precut,
				  Central_wavenumber,
				  Half_bandwidth,
				  GENERAL_FILTER_ORDER);
    /* Note that memory for the variable 'Filter' has been allocated inside the routine just called! */
    
    QL_filtered = vndrs_filter(I1, Filter); /* memory allocated inside */
    vndrs_free_vector_double(Filter);

    /* Now we filter for the higher freq bandpass */
    Half_bandwidth = (Wavenum_max_out - Wavenum_min_out) * 0.5;
    Central_wavenumber = (Wavenum_max_out + Wavenum_min_out) * 0.5;
    Filter = vndrs_compute_filter(N_frames,
				  Sample_interval_precut,
				  Central_wavenumber,
				  Half_bandwidth,
				  GENERAL_FILTER_ORDER);
    /* Note that memory for the variable 'Filter' has been allocated inside the routine just called! */    
    Noise_filtered = vndrs_filter(I1, Filter); /* memory allocated inside */
    /* In FLUOR code, the noise is filtered at the same frequency as the signal */
    
    /* Apodisation of the final QL to avoid edge problems transition is 1/10th of the scan */
    vndrs_vector_mul(QL_filtered, Apodisation_filter);  
    vndrs_vector_mul(Noise_filtered, Apodisation_filter);  
    vndrs_free_vector_double(Apodisation_filter);

    /* Final QL and noise */
    vndrs_vector_memcpy(quicklook_signal, QL_filtered);
    vndrs_vector_memcpy(quicklook_noise, Noise_filtered);

    /* Memory release */
    vndrs_free_vector_double(Filter);
    vndrs_free_vector_double(QL_filtered);
    vndrs_free_vector_double(Noise_filtered);
    vndrs_free_vector_double(I1);
    vndrs_free_vector_double(I2);
    vndrs_free_vector_double(PA);
    vndrs_free_vector_double(PB);
    vndrs_free_vector_double(I1_off);
    vndrs_free_vector_double(I2_off);
    vndrs_free_vector_double(PA_off);
    vndrs_free_vector_double(PB_off);
}

/*
  This function is used to select only the scans that have been processed by the
  pipeline. BEWARE: MEMORY ALLOCATED INSIDE FUNCTION !!!!
*/

vector_double *vndrs_extract_nonzero(
				/* inputs*/      
				vector_double *mu2_vector,
				vector_double *flag_processed)
{
  long i, N_scans_on, N_processed = 0;
  long index = 0;
  vector_double *extracted_mu2;

  N_scans_on = mu2_vector->dimension;

  for(i=0;i<N_scans_on;i++)
    if(flag_processed->data[i] > 0)
        N_processed++;
        
  extracted_mu2 = vndrs_calloc_vector_double(N_processed);

  for (i=0;i<N_scans_on;i++)
      if(flag_processed->data[i] > 0)
      {
          extracted_mu2->data[index] = mu2_vector->data[i];
          index++;
      }
  
  return(extracted_mu2);
}


/*
  Cuts a series of scans according to the delay line offsets that have
  been saved together with the scans. pkervell 13-mar-2002
*/

void vndrs_trim_batch (
                           /* --------------- */
                           /* input variables */
                           /* --------------- */
                           batch *Batch_in,
			   /* Input batch to trim */
			   double Fringe_speed,
			   double Frame_rate,
			   int center_fringes,
			   /* boolean to center or not the fringes */
			   long N_cut,
			   /* Number of frames to keep around the fringes */
                           /* series of scans */
                           /* ---------------- */
                           /* output variables */
                           /* ---------------- */
                           batch *Batch_out
			   )
{
  long I, J, min_index, max_index,ind;
  long N_scans, N_frames;
  double index_fringes;

  N_scans = Batch_in->number_of_scans;
  N_frames = Batch_in->fullscans[0]->I1->dimension;
  if(N_cut>N_frames) N_cut = N_frames;

  for (I=0; I<N_scans; I++)
    {
      /* Computes the rank of the center of the fringes in the scan, based on the value of the
	 OPD offset */
      if(center_fringes == 0)
	{
	  min_index = N_frames/2 - N_cut/2;
	  max_index = min_index + N_cut;
	}
      else
	{
	  index_fringes = vndrs_round(Batch_in->scan_data[I].dl_offset / (1000000.*Fringe_speed / Frame_rate));
	  /* the units of the fringe speed is meters/s while the DL offset is in microns*/

	  /* Boundaries of the cut */
	  min_index = (int)((N_frames/2) + index_fringes - (N_cut/2));
	  max_index = (int)(min_index + N_cut);
	  
	  if(min_index < 0.0)
	    {
	      min_index = 0;
	      max_index = N_cut;
	    }
	  if(max_index > N_frames)
	    {
	  max_index = N_frames;
	  min_index = N_frames - N_cut;
	    }
	}

      ind = 0;
      /* Actual creation of the trimmed batch of scans */
      for(J=min_index; J<max_index; J++)
	{
	  Batch_out->fullscans[I]->I1->data[ind] = Batch_in->fullscans[I]->I1->data[J];
	  Batch_out->fullscans[I]->I2->data[ind] = Batch_in->fullscans[I]->I2->data[J];
	  Batch_out->fullscans[I]->PA->data[ind] = Batch_in->fullscans[I]->PA->data[J];
	  Batch_out->fullscans[I]->PB->data[ind] = Batch_in->fullscans[I]->PB->data[J];
	  Batch_out->fullscans[I]->QL->data[ind] = Batch_in->fullscans[I]->QL->data[J];
	  ind = ind + 1;
	}
      /* printf("I:%d, MinIndex = %d, MaxIndex = %d \n",I,min_index, max_index); */
      
      /* Additional parameters saved into output batch structure */
      Batch_out->scan_data[I].dl_offset = Batch_in->scan_data[I].dl_offset;
      Batch_out->scan_data[I].time = Batch_in->scan_data[I].time;
      /* Remark: the time is encoded in seconds since Julian date 2400000.5 */
    }

}

/*
  Cuts an individual scan according to the position of the fringes, given in pixels from
  the left side of the scan. pkervell 18-mar-2002
*/

void vndrs_trim_scan (
                           /* --------------- */
                           /* input variables */
                           /* --------------- */
                           individual_scan *scan,
			   long N_frames_cut,
			   /* Number of frames to keep around the fringes */
			   long fringes_position,
			   /* Fringes position relatively to the center, in PIXELS */
                           /* ---------------- */
                           /* output variables */
                           /* ---------------- */
                           individual_scan *scan_out
			   )
{
  long I, min_index, max_index, ind;
  long N_frames;

  N_frames = scan->dimension;

  /* as defined here max_index is in fact not the last index, but one larger - mschoell */
  min_index = fringes_position + (N_frames/2) - (N_frames_cut/2);
  if (min_index < 0) min_index = 0;
  if (min_index + N_frames_cut  > N_frames) min_index = N_frames - N_frames_cut;
  max_index = min_index + N_frames_cut;
  
  /* Actual creation of the trimmed scan */
  for(I=min_index, ind=0; I<max_index; I++, ind++)
    {
      scan_out->data[ind] = scan->data[I];
    }
}


/*
   Cuts a vector_double from index_min to index_max. The output variable scan_out is NOT
   allocated inside the function. index_init IS included, index_end IS NOT included.
   pkervell 09-Sep-02
*/
void vndrs_trim_scan_segment (
                           /* --------------- */
                           /* input variables */
                           /* --------------- */
                           individual_scan *scan,
			   long index_init,
			   long index_end,
			   /* index of begin and end of the trim, in PIXELS */
                           /* ---------------- */
                           /* output variables */
                           /* ---------------- */
                           individual_scan *scan_out
			   )
{
  long I, ind;
  long N_frames;

  N_frames = scan->dimension;
  if(index_end > N_frames) index_end = N_frames;
  if(index_init < 0) index_end = 0;
  
  /* Actual creation of the trimmed scan */
  for(I=index_init, ind=0; I<index_end; I++, ind++)
      scan_out->data[ind] = scan->data[I];
}


/*
    Computes the transfer function between the photometric and
    interferometric signals of LdV. FLUOR METHOD, modified by pkervell - 06-Sep-02
    Described in [1], section 2.6
 */
void vndrs_compute_matrix (
                           /* --------------- */
                           /* input variables */
                           /* --------------- */
                           batch *Batch_A,
                           /* series of scans obtained with beam A only (PB = 0) */
                           batch *Batch_B,
                           /* series of scans obtained with beam B only (PA = 0) */
                           vector_double *Off_offsets,
                           vector_double *WA_mean,
                           vector_double *WB_mean,
 			   int ionic_in,
                           /* ---------------- */
                           /* output variables */
                           /* ---------------- */
                           /* the next two entries are changed from the original concept,
                              where they were defined as vectors of matrices */
                           matrix_double *Variances,
                           /* Matrix of variances */
                           matrix_double *Matrix,
                           /* The transfer matrix that is computed in this routine (see comments below) */
                           matrix_double *Standard_dev,
                           /* The standard errors of the kappas (2x2 matrix) */
                           vector_double *Relative_MSE,
                           /* Mean squared error of the fit (2 x 1 vector) */
			   int verbose_on
                          )
/* This routines computes the transfer matrix between the photometric and interferometric signals
   (both measured at photometers (see [2], Eq. (19) on p. 383):

               [ I1 ]   [ kappa_1A  kappa_1B ]   [ PA ]
               |    | = |                    | * |    |
               [ I2 ]   [ kappa_2A  kappa_2B ]   [ PB ]

   This matrix can be evaluated DIRECTLY from the data (I1, I2, PA, PB), without requiring
   an a priori knowledge of the individual transmissions and gains in the system. The evaluation
   is performed by adjusting a least square fit of a linear combination of PA and PB to I1 for
   the 2 coefficients 'kappa_1A' and 'kappa_1B' and to I2 for the 2 coefficients 'kappa_2A'
   and 'kappa_2B'.
*/
{
    /* Local variables */
    long N_samples_A, N_samples_B, N_frames, N_scans_mixed;
    double Mean_A_with_B_only, Mean_B_with_A_only, Mean_B_with_B_only, Mean_A_with_A_only;
    double Mean_I1_squared, Mean_I2_squared;
    double Matrix_00, Matrix_01, Matrix_10, Matrix_11;
    /* The four coefficients of the transfer matrix:

               [ Matrix_00  Matrix_01 ]   [ kappa_1A  kappa_1B ]   
               |                      | = |                    | 
               [ Matrix_10  Matrix_11 ]   [ kappa_2A  kappa_2B ]
    */
    double Std_00, Std_01, Std_10, Std_11;
    long I;
    double MSE_of_fit_I1, MSE_of_fit_I2;
    matrix_double *Covariance_I1, *Covariance_I2;
    full_scan *Scan_AB;
    /* The full scan obtained by averaging over the batch 'Scan_AB' (concatenation of batches A and B) */
    individual_scan *Filtered_signals_A_I1, *Filtered_signals_A_I2, *Filtered_signals_A_PA, *Filtered_signals_A_PB;
    individual_scan *Filtered_signals_B_I1, *Filtered_signals_B_I2, *Filtered_signals_B_PA, *Filtered_signals_B_PB;
    individual_scan *Signal_fit_I1, *Signal_fit_I2;
    full_scan *Avg_fullscan_A_only, *Avg_fullscan_B_only;
    vector_double *kappa_vec, *buffer;
    double sigma_kappa, mean_kappa;

    
    /*----*/
    
    /* Get the number of scans in the batches 'Batch_A' and 'Batch_B' --each obtained with
       only one of the 2 beams */
    N_samples_A = Batch_A->number_of_scans;
    N_samples_B = Batch_B->number_of_scans;
    /* We consider the smallest number of the two for N_scans */
    if(N_samples_A < N_samples_B) N_scans_mixed = N_samples_A; else N_scans_mixed = N_samples_B;
    
    /* Get the number of frames in the mixed scans (It is assumed that the number of frames,
       i.e. the dimension of an individual scan, is the same for both batches A and B.*/
    N_frames = Batch_A->fullscans[0]->I1->dimension;

    /* Allocate memory for some vectors */
    Scan_AB = vndrs_calloc_full_scan(N_frames*2);
    Signal_fit_I1 = vndrs_calloc_vector_double(N_frames*2);
    Signal_fit_I2 = vndrs_calloc_vector_double(N_frames*2);
    Covariance_I1 = vndrs_calloc_matrix_double(2, 2);
    Covariance_I2 = vndrs_calloc_matrix_double(2, 2);
    kappa_vec = vndrs_calloc_vector_double(N_frames);
    buffer = vndrs_calloc_vector_double(N_frames);
        
    /* Compute and store the input scans (single-beam batches A and B) */
    Avg_fullscan_A_only = vndrs_average_batch_to_fullscan(Batch_A);
    Avg_fullscan_B_only = vndrs_average_batch_to_fullscan(Batch_B);

    /* Removal of the background offsets */
    vndrs_vector_add_constant(Avg_fullscan_A_only->I1, -Off_offsets->data[0]);
    vndrs_vector_add_constant(Avg_fullscan_A_only->I2, -Off_offsets->data[1]);
    vndrs_vector_add_constant(Avg_fullscan_A_only->PA, -Off_offsets->data[2]);
    vndrs_vector_add_constant(Avg_fullscan_A_only->PB, -Off_offsets->data[3]);
    vndrs_vector_add_constant(Avg_fullscan_B_only->I1, -Off_offsets->data[0]);
    vndrs_vector_add_constant(Avg_fullscan_B_only->I2, -Off_offsets->data[1]);
    vndrs_vector_add_constant(Avg_fullscan_B_only->PA, -Off_offsets->data[2]);
    vndrs_vector_add_constant(Avg_fullscan_B_only->PB, -Off_offsets->data[3]);

    /* There is a strange problem happening sometimes in that the beam A and beam B only are inverted:
    there is no light on A in Beam_A and no light on B in Beam_B. Of course, this does not change the result
    of the fit, but it does change the errors, that tend to be "NaN". Here below we make a check, and then
    invert the two vectors if necessary. This may come from the acquisition software being improperly setup
    when we change the delay lines (?) and therefore which beam is which. */

    Mean_A_with_A_only = vndrs_mean_scan(Avg_fullscan_A_only->PA);
    Mean_A_with_B_only = vndrs_mean_scan(Avg_fullscan_B_only->PA);
    Mean_B_with_A_only = vndrs_mean_scan(Avg_fullscan_A_only->PB);
    Mean_B_with_B_only = vndrs_mean_scan(Avg_fullscan_B_only->PB);
    
    if(Mean_B_with_B_only < Mean_B_with_A_only | Mean_A_with_A_only < Mean_A_with_B_only)
    {
        vndrs_vector_memcpy(buffer, Avg_fullscan_A_only->I1);
        vndrs_vector_memcpy(Avg_fullscan_A_only->I1, Avg_fullscan_B_only->I1);
        vndrs_vector_memcpy(Avg_fullscan_B_only->I1, buffer);
        
        vndrs_vector_memcpy(buffer, Avg_fullscan_A_only->I2);
        vndrs_vector_memcpy(Avg_fullscan_A_only->I2, Avg_fullscan_B_only->I2);
        vndrs_vector_memcpy(Avg_fullscan_B_only->I2, buffer);
        
        vndrs_vector_memcpy(buffer, Avg_fullscan_A_only->PA);
        vndrs_vector_memcpy(Avg_fullscan_A_only->PA, Avg_fullscan_B_only->PA);
        vndrs_vector_memcpy(Avg_fullscan_B_only->PA, buffer);
        
        vndrs_vector_memcpy(buffer, Avg_fullscan_A_only->PB);
        vndrs_vector_memcpy(Avg_fullscan_A_only->PB, Avg_fullscan_B_only->PB);
        vndrs_vector_memcpy(Avg_fullscan_B_only->PB, buffer);
        
        printf("  Beam A only and Beam B only files are inverted on disk, processing unaffected.\n\n");
        sprintf(vndrs_protocol.text,"# Beam A only and Beam B only files contents are inverted on disk. \n");
        vndrs_runtime_protocol();
    }
   
    /* We have here a minor problem in that there is a small ammount of light that is present
    in PA when we are in PB only. The reason for this is not clear, maybe reflections inside the MONA
    box at the extremities of the fibers ? In principle, this should not affect the fitting procedure
    though, as it does not depend on having a true dark for both beams - pkervell - 06-Sep-02 */
    
    /* Frequency filtering */

    Filtered_signals_A_I1 = vndrs_filter(Avg_fullscan_A_only->I1, WA_mean);
    Filtered_signals_A_I2 = vndrs_filter(Avg_fullscan_A_only->I2, WA_mean);
    Filtered_signals_A_PA = vndrs_filter(Avg_fullscan_A_only->PA, WA_mean);
    Filtered_signals_A_PB = vndrs_filter(Avg_fullscan_A_only->PB, WA_mean);
    Filtered_signals_B_I1 = vndrs_filter(Avg_fullscan_B_only->I1, WB_mean);
    Filtered_signals_B_I2 = vndrs_filter(Avg_fullscan_B_only->I2, WB_mean);
    Filtered_signals_B_PA = vndrs_filter(Avg_fullscan_B_only->PA, WB_mean);
    Filtered_signals_B_PB = vndrs_filter(Avg_fullscan_B_only->PB, WB_mean);
    
    /* Concatenation of the two fullscans A and B only filtered */
    for(I=0;I<(N_frames);I++)
    {
        Scan_AB->I1->data[I] = Filtered_signals_A_I1->data[I];
        Scan_AB->I2->data[I] = Filtered_signals_A_I2->data[I];
        Scan_AB->PA->data[I] = Filtered_signals_A_PA->data[I];
        Scan_AB->PB->data[I] = Filtered_signals_A_PB->data[I];
    }
    for(I=0;I<(N_frames);I++)
    {
        Scan_AB->I1->data[I+N_frames] = Filtered_signals_B_I1->data[I];
        Scan_AB->I2->data[I+N_frames] = Filtered_signals_B_I2->data[I];
        Scan_AB->PA->data[I+N_frames] = Filtered_signals_B_PA->data[I];
        Scan_AB->PB->data[I+N_frames] = Filtered_signals_B_PB->data[I];
    }
    
    Mean_I1_squared = vndrs_mean_squared_scan(Scan_AB->I1);
    Mean_I2_squared = vndrs_mean_squared_scan(Scan_AB->I2);
    
    
    /* ------------------------------------------------------------------------ */
    /* COMPUTE THE TRANSFER MATRIX COEFFICIENTS                                 */
    /* ------------------------------------------------------------------------ */ 
    /* The four coefficients of the transfer matrix:

               [ Matrix_00  Matrix_01 ]   [ kappa_1A  kappa_1B ]   
               |                      | = |                    | 
               [ Matrix_10  Matrix_11 ]   [ kappa_2A  kappa_2B ]
       
       The 'kappa'-coefficients are computed by least square fitting the
       interferometric signals (adjusted to zero mean, apodized and filtered)
       to linear combinations of the photometric signals (pre-processed in
       the same way as their interferometric counterparts.) */

    /* (k1A,k1B): between I1 and (PA,PB) */
    vndrs_linfit_2(Scan_AB->I1,
                   Scan_AB->PA,
                   Scan_AB->PB,
                   &Matrix_00,
                   &Matrix_01,
                   Covariance_I1,
                   Signal_fit_I1,
                   &MSE_of_fit_I1);    
 
    /* Set the elements of the transfer matrix 'Matrix' to the 2 computed values: */
    set_element_matrix_double(Matrix, 0, 0, Matrix_00);
    set_element_matrix_double(Matrix, 0, 1, Matrix_01);

    set_element_matrix_double(Variances, 0, 0,
        get_element_matrix_double(Covariance_I1, 0, 0));
    set_element_matrix_double(Variances, 0, 1,
        get_element_matrix_double(Covariance_I1, 1, 1));

    Relative_MSE->data[0] = MSE_of_fit_I1/Mean_I1_squared;
    
    if(ionic_in == 0)
    { /* If we are not using IONIC - pkervell - 13-Feb-03 */
        /* (k2A,k2B): between I2 and (PA,PB) */
        vndrs_linfit_2(Scan_AB->I2,
                    Scan_AB->PA,
                    Scan_AB->PB,
                    &Matrix_10,
                    &Matrix_11,
                    Covariance_I2,
                    Signal_fit_I2,
                    &MSE_of_fit_I2);

        Relative_MSE->data[1] = MSE_of_fit_I2/Mean_I2_squared;

        /* Set the elements of the transfer matrix 'Matrix' to the 2 computed values: */
        set_element_matrix_double(Matrix, 1, 0, Matrix_10);
        set_element_matrix_double(Matrix, 1, 1, Matrix_11);
        set_element_matrix_double(Variances, 1, 0,
            get_element_matrix_double(Covariance_I2, 0, 0));
        set_element_matrix_double(Variances, 1, 1,
            get_element_matrix_double(Covariance_I2, 1, 1));
    }
    else
    {
        /* Set the elements of the transfer matrix 'Matrix' for I2 to zero in IONIC case */
        set_element_matrix_double(Matrix, 1, 0, 0.0);
        set_element_matrix_double(Matrix, 1, 1, 0.0);
        set_element_matrix_double(Variances, 1, 0, 0.0);
        set_element_matrix_double(Variances, 1, 1, 0.0);
     }
    

 
 
    /* Estimation of the standard error bars on each Kappa - pkervell 06-Sep-02
       To compute the standard error bars of the kappas, the principle is that we compute, for example for
       kappa1A, a vector of values: kappa_vec[i] = (I1[i]-kappa1A_mean*PA[i])/PB[i] and then we derive the
       standard deviation from this vector. We do it only on the part of Scan_AB where the corresponding
       photometry is not zero.
       
       BEWARE:
       Though the variable is called Standard_dev, this is the STANDARD ERROR that is stored inside !!
    */
       
    /* Kappa1A */
    for(I=0;I<(N_frames);I++)
        kappa_vec->data[I] = (Scan_AB->I1->data[I] - Matrix_01 * Scan_AB->PB->data[I])/Scan_AB->PA->data[I];
    /* We take the possible bias into account */
    sigma_kappa = vndrs_stddev_scan(kappa_vec);
    mean_kappa = vndrs_mean_scan(kappa_vec);
    Std_00 = sqrt(sigma_kappa*sigma_kappa + (mean_kappa - Matrix_00)*(mean_kappa - Matrix_00));
    set_element_matrix_double(Standard_dev, 0, 0, Std_00);

    /* Kappa1B */
    for(I=(N_frames);I<(2*N_frames);I++)
        kappa_vec->data[I-N_frames] = (Scan_AB->I1->data[I] - Matrix_00 * Scan_AB->PA->data[I])/Scan_AB->PB->data[I];
    sigma_kappa = vndrs_stddev_scan(kappa_vec);
    mean_kappa = vndrs_mean_scan(kappa_vec);
    Std_01 = sqrt(sigma_kappa*sigma_kappa + (mean_kappa - Matrix_01)*(mean_kappa - Matrix_01));
    set_element_matrix_double(Standard_dev, 0, 1, Std_01);
    
    if(ionic_in == 0)
    {	/* If we are not using IONIC - pkervell - 13-Feb-03 */
        /* Kappa2A */
        for(I=0;I<(N_frames);I++)
            kappa_vec->data[I] = (Scan_AB->I2->data[I] - Matrix_11 * Scan_AB->PB->data[I])/Scan_AB->PA->data[I];
        sigma_kappa = vndrs_stddev_scan(kappa_vec);
        mean_kappa = vndrs_mean_scan(kappa_vec);
        Std_10 = sqrt(sigma_kappa*sigma_kappa + (mean_kappa - Matrix_10)*(mean_kappa - Matrix_10));
        set_element_matrix_double(Standard_dev, 1, 0, Std_10);
        
        /* Kappa2B */
        for(I=N_frames;I<(2*N_frames);I++)
            kappa_vec->data[I-N_frames] = (Scan_AB->I2->data[I] - Matrix_10 * Scan_AB->PA->data[I])/Scan_AB->PB->data[I];
        sigma_kappa = vndrs_stddev_scan(kappa_vec);
        mean_kappa = vndrs_mean_scan(kappa_vec);
        Std_11 = sqrt(sigma_kappa*sigma_kappa + (mean_kappa - Matrix_11)*(mean_kappa - Matrix_11));
        set_element_matrix_double(Standard_dev, 1, 1, Std_11);
    }
    else
    {
        set_element_matrix_double(Standard_dev, 1, 0, 0.0);
        set_element_matrix_double(Standard_dev, 1, 1, 0.0);
    }
    
    if(verbose_on == 1)
      {
	vndrs_fprintf_vector_double(Scan_AB->PA, "Scan_AB_PA.dat");
	vndrs_fprintf_vector_double(Scan_AB->PB, "Scan_AB_PB.dat");
	vndrs_fprintf_vector_double(Scan_AB->I1, "Scan_AB_I1.dat");
	vndrs_fprintf_vector_double(Scan_AB->I2, "Scan_AB_I2.dat");
	vndrs_fprintf_matrix_double(Covariance_I1, "Covariance_I1.dat");
	vndrs_fprintf_matrix_double(Covariance_I2, "Covariance_I2.dat");
	vndrs_fprintf_matrix_double(Matrix, "TransferMatrix.dat");
	vndrs_fprintf_matrix_double(Variances, "VarianceMatrix.dat");
      }

    /* Free memory */
    vndrs_free_full_scan(Scan_AB);
    vndrs_free_full_scan(Avg_fullscan_A_only);
    vndrs_free_full_scan(Avg_fullscan_B_only);
    vndrs_free_vector_double(Filtered_signals_A_I1);
    vndrs_free_vector_double(Filtered_signals_A_I2);
    vndrs_free_vector_double(Filtered_signals_A_PA);
    vndrs_free_vector_double(Filtered_signals_A_PB);
    vndrs_free_vector_double(Filtered_signals_B_I1);
    vndrs_free_vector_double(Filtered_signals_B_I2);
    vndrs_free_vector_double(Filtered_signals_B_PA);
    vndrs_free_vector_double(Filtered_signals_B_PB);
    vndrs_free_matrix_double(Covariance_I1);
    vndrs_free_matrix_double(Covariance_I2);
    vndrs_free_vector_double(Signal_fit_I1);
    vndrs_free_vector_double(Signal_fit_I2);
    vndrs_free_vector_double(kappa_vec);
    vndrs_free_vector_double(buffer);


    return;
}





/* The following routine sets the coefficients of the transfer matrix --either by computing
   them (--> fitting procedure) or by setting them to the user-preset values
   modified pkervell 13-mar-2002 */

void vndrs_get_transfer_matrix(/* --------------- */
                               /* input variables */
                               /* --------------- */ 
                               batch *Batch_A,
                               /* series of scans obtained with beam A only */
                               batch *Batch_B,
                               /* series of scans obtained with beam B only */
                               vector_double *Off_offsets,
                               vector_double *WA_mean,
                               vector_double *WB_mean,
 			       int ionic_in,
                               /* -------------------------- */
                               /* input and output variables */
                               /* -------------------------- */ 
                               matrix_double *Matrix,
                               /* The 2 x 2 transfer matrix between the photometric and interferometric signals
                                  (both measured at photometers (see [2], Eq. (19) on p. 383):

                                          [ I1 ]   [ kappa_1A  kappa_1B ]   [ PA ]
                                          |    | = |                    | * |    |
                                          [ I2 ]   [ kappa_2A  kappa_2B ]   [ PB ]
			       */
                               matrix_double *Standard_dev, /* added pkervell 06-Sep-02 */
                               vector_double *Relative_MSE,       /* Residual variance on the fit of I - pkervell 19-apr-02 */
			       int verbose_on)
{
    /* LOCAL VARIABLES */
    /* --------------- */
    matrix_double *Stand_err; /* not used for now */

    Stand_err = vndrs_calloc_matrix_double(2,2);
    
    /* Compute transfer matrix */
    vndrs_compute_matrix(Batch_A,
			 Batch_B,
                         Off_offsets,
                         WA_mean,
                         WB_mean,
 			 ionic_in,
			 Stand_err,
			 Matrix,
                         Standard_dev,
			 Relative_MSE,
			 verbose_on);
    
    vndrs_free_matrix_double(Stand_err);

    return;
}



/*-----------------------------------------------------------------------*/
/* MODIFIED version for computation of the transfer matrix by least      */
/* squares fitting. Also outputs true backgrounds on sky and in lab.     */
/* pkervell 13-Oct-02	 						 */
/*-----------------------------------------------------------------------*/

void vndrs_compute_matrix_least_squares(
                           /* --------------- */
                           /* input variables */
                           /* --------------- */
                           batch *Batch_off,
                           /* series of scans obtained with beam A and beam B off */
                           batch *Batch_A,
                           /* series of scans obtained with beam A only */
                           batch *Batch_B,
                           /* series of scans obtained with beam B only */
                           batch *Batch_on,
                           /* series of scans obtained on source */
			   int ionic_in,
                           /* ---------------- */
                           /* output variables */
                           /* ---------------- */
                           matrix_double *Matrix,
                           /* The transfer matrix that is computed in this routine (see comments below) */
                           matrix_double *Standard_errors,
                           /* The standard errors of the kappas (2x2 matrix) */
                           int *Autotest_case,
                           /* Flag to indicate that the matrix was computed using the Autotest case */
			   int verbose_on)
{
/* The computation of the kappas with the regular FLUOR method is dependent on the level of the
background. This is not nice as we do not know the backgrounds in advance ! The off-source batch
gives us only an approximation of the background level, as we are offsetting the parabolas, and
therefore we are looking into the laboratory, which is warmer than the sky.

We simply make the assumption here that there is not too much VARIABLE flux coming from the
beam that is not "on" (PA or PB) in the interferometric channels. Constant fluxes (BG) are of
no importance. */
    
    long I, N_scans_mixed, N_scans_on, N_scans_off, N_scans_ok, N_frames;
    double kappa1A, kappa1B, kappa2A, kappa2B, dummy;
    vector_double *Flux_vector_I1, *Flux_vector_I2, *Flux_vector_PA, *Flux_vector_PB;
    double ct_I1_1, ct_I2_1, ct_I1_2, ct_I2_2;
    double I1A_raw, I1B_raw, I2A_raw, I2B_raw, PAA_raw, PBB_raw, PAB_raw, PBA_raw;
    vector_double *On_source_Mean, *On_source_Stddev, *Off_source_Mean, *Off_source_Stddev, *dummy_vec;
    double Std_00, Std_01, Std_10, Std_11;
    double RFI_I1, RFI_I2, RFI_PA, RFI_PB;
    vector_double *kappa_vec;
    double sigma_kappa, mean_kappa;
    
    N_scans_mixed = Batch_A->number_of_scans;
    if(N_scans_mixed > Batch_B->number_of_scans) N_scans_mixed = Batch_B->number_of_scans;
    N_scans_off = Batch_off->number_of_scans;
    N_scans_on = Batch_on->number_of_scans;
    N_frames = Batch_on->fullscans[0]->I1->dimension;
    
    On_source_Mean = vndrs_calloc_vector_double(4);
    On_source_Stddev = vndrs_calloc_vector_double(4);
    Off_source_Mean = vndrs_calloc_vector_double(4);
    Off_source_Stddev = vndrs_calloc_vector_double(4);
    dummy_vec = vndrs_calloc_vector_double(4);
    
    /* Detection if we are in the Autotest case, where the fluxes are TOO STABLE to make a good
    least squares fit */
    vndrs_batch_statistics(Batch_on,
                        On_source_Mean,
                        On_source_Stddev,
                        dummy_vec,
                        dummy_vec);

    vndrs_batch_statistics(Batch_off,
                        Off_source_Mean,
                        Off_source_Stddev,
                        dummy_vec,
                        dummy_vec);
    
    /* We define here relative fluctuation indexes */
    RFI_I1 = (On_source_Stddev->data[0]/Off_source_Stddev->data[0])/sqrt(On_source_Mean->data[0]/Off_source_Mean->data[0]);
    RFI_I2 = (On_source_Stddev->data[1]/Off_source_Stddev->data[1])/sqrt(On_source_Mean->data[1]/Off_source_Mean->data[1]);
    RFI_PA = (On_source_Stddev->data[2]/Off_source_Stddev->data[2])/sqrt(On_source_Mean->data[2]/Off_source_Mean->data[2]);
    RFI_PB = (On_source_Stddev->data[3]/Off_source_Stddev->data[3])/sqrt(On_source_Mean->data[3]/Off_source_Mean->data[3]);
    
    
    if(RFI_I1 < 1.5 || RFI_I2 < 1.5 || RFI_PA < 1.5 || RFI_PB < 1.5)
    {
        printf("\n  *** Photometric fluctuations too low to compute kappas by direct least squares.\n");
        sprintf(vndrs_protocol.text,"# *** Photometric fluctuations too low to compute kappas by direct least squares.\n");
        vndrs_runtime_protocol();
        *Autotest_case = 1;

        Flux_vector_I1 = vndrs_calloc_vector_double(N_scans_off+N_scans_mixed);
        Flux_vector_I2 = vndrs_calloc_vector_double(N_scans_off+N_scans_mixed);
        Flux_vector_PA  = vndrs_calloc_vector_double(N_scans_off+N_scans_mixed);
        Flux_vector_PB  = vndrs_calloc_vector_double(N_scans_off+N_scans_mixed);

        /* Off source averages - filling of the first part of the flux vectors with off-source
        flux values. The LS method works well without this, but not in Autotest and when the flux
        fluctuations of the star are not large enough. pkervell 21-Oct-02 */
        for(I=0; I<N_scans_off; I++)
        {
            Flux_vector_I1->data[I] = vndrs_mean_scan(Batch_off->fullscans[I]->I1);
            Flux_vector_I2->data[I] = vndrs_mean_scan(Batch_off->fullscans[I]->I2);
            Flux_vector_PA->data[I] = vndrs_mean_scan(Batch_off->fullscans[I]->PA);
            Flux_vector_PB->data[I] = vndrs_mean_scan(Batch_off->fullscans[I]->PB);
        }

    }
    else
    { /* The size of the flux vectors changes depending if Autotest_case or not */
        Flux_vector_I1 = vndrs_calloc_vector_double(N_scans_mixed);
        Flux_vector_I2 = vndrs_calloc_vector_double(N_scans_mixed);
        Flux_vector_PA = vndrs_calloc_vector_double(N_scans_mixed);
        Flux_vector_PB = vndrs_calloc_vector_double(N_scans_mixed);

        *Autotest_case = 0;
    }
    printf("\n  Relative fluctuations indices:    I1 = %7.2f  I2 = %7.2f  PA = %7.2f  PB = %7.2f\n",
    RFI_I1, RFI_I2, RFI_PA, RFI_PB);
    sprintf(vndrs_protocol.text,"# Relative fluctuations indices:    I1 = %7.2f   I2 = %7.2f   PA = %7.2f   PB = %7.2f\n",
    RFI_I1, RFI_I2, RFI_PA, RFI_PB);
    vndrs_runtime_protocol();
    
    /* Kappa1A and Kappa2A */
    if(*Autotest_case == 1)
    {
        for(I=0; I<N_scans_mixed; I++)
        {
            Flux_vector_I1->data[I+N_scans_off] = vndrs_mean_scan(Batch_A->fullscans[I]->I1);
            Flux_vector_I2->data[I+N_scans_off] = vndrs_mean_scan(Batch_A->fullscans[I]->I2);
            Flux_vector_PA->data[I+N_scans_off] = vndrs_mean_scan(Batch_A->fullscans[I]->PA);
            Flux_vector_PB->data[I+N_scans_off] = vndrs_mean_scan(Batch_A->fullscans[I]->PB);

            I1A_raw += Flux_vector_I1->data[I+N_scans_off]/N_scans_mixed;
            I2A_raw += Flux_vector_I2->data[I+N_scans_off]/N_scans_mixed;
            PAA_raw += Flux_vector_PA->data[I+N_scans_off]/N_scans_mixed;
            PBA_raw += Flux_vector_PB->data[I+N_scans_off]/N_scans_mixed;
        }
    }
    else
    {
        for(I=0; I<N_scans_mixed; I++)
        {
            Flux_vector_I1->data[I] = vndrs_mean_scan(Batch_A->fullscans[I]->I1);
            Flux_vector_I2->data[I] = vndrs_mean_scan(Batch_A->fullscans[I]->I2);
            Flux_vector_PA->data[I] = vndrs_mean_scan(Batch_A->fullscans[I]->PA);
            Flux_vector_PB->data[I] = vndrs_mean_scan(Batch_A->fullscans[I]->PB);
        }
        I1A_raw = vndrs_mean_scan(Flux_vector_I1);
        I2A_raw = vndrs_mean_scan(Flux_vector_I2);
        PAA_raw = vndrs_mean_scan(Flux_vector_PA);
        PBA_raw = vndrs_mean_scan(Flux_vector_PB);
    }

    /* There is a strange problem happening sometimes in that the beam A and beam B only are inverted:
    there is no light on A in Beam_A and no light on B in Beam_B. 
    
    Here below we make a check, and then invert the two fits if necessary. This may come from the
    acquisition software being improperly setup when we change the delay lines (?) and therefore
    which beam is which. Or from a difference between 'TEL' and 'PAR' offsets ??? */
    
    if(PAA_raw > PBA_raw) /* Correct case */
    {
        vndrs_computes_least_squares(Flux_vector_PA,
                                    Flux_vector_I1,
                                    &kappa1A,
                                    &ct_I1_1,
                                    &dummy);
    
        vndrs_computes_least_squares(Flux_vector_PA,
                                    Flux_vector_I2,
                                    &kappa2A,
                                    &ct_I2_1,
                                    &dummy);
    }
    else
    {
        printf("\n  *** Single-beam files are inverted on disk. Processing unaffected. \n");
        sprintf(vndrs_protocol.text, "# Single-beam files are inverted on disk. Processing unaffected.\n");
        vndrs_runtime_protocol();
        
        vndrs_computes_least_squares(Flux_vector_PB,
                                    Flux_vector_I1,
                                    &kappa1B,
                                    &ct_I1_1,
                                    &dummy);
    
        vndrs_computes_least_squares(Flux_vector_PB,
                                    Flux_vector_I2,
                                    &kappa2B,
                                    &ct_I2_1,
                                    &dummy);
    }
                                
    if(verbose_on == 1)
    {
        vndrs_fprintf_vector_double(Flux_vector_I1, "Flux_vector_I1_A.dat");
        vndrs_fprintf_vector_double(Flux_vector_I2, "Flux_vector_I2_A.dat");
        vndrs_fprintf_vector_double(Flux_vector_PA, "Flux_vector_PA_A.dat");
        vndrs_fprintf_vector_double(Flux_vector_PB, "Flux_vector_PB_A.dat");
    }

    /* Kappa1B and Kappa2B */
    if(*Autotest_case == 1)
    {
        I1B_raw = 0.0; I2B_raw = 0.0; PAB_raw = 0.0; PBB_raw = 0.0;
            
        for(I=0; I<N_scans_mixed; I++)
        {
            Flux_vector_I1->data[I+N_scans_off] = vndrs_mean_scan(Batch_B->fullscans[I]->I1);
            Flux_vector_I2->data[I+N_scans_off] = vndrs_mean_scan(Batch_B->fullscans[I]->I2);
            Flux_vector_PA->data[I+N_scans_off] = vndrs_mean_scan(Batch_B->fullscans[I]->PA);
            Flux_vector_PB->data[I+N_scans_off] = vndrs_mean_scan(Batch_B->fullscans[I]->PB);
    
            I1B_raw += Flux_vector_I1->data[I+N_scans_off]/N_scans_mixed;
            I2B_raw += Flux_vector_I2->data[I+N_scans_off]/N_scans_mixed;
            PAB_raw += Flux_vector_PA->data[I+N_scans_off]/N_scans_mixed;
            PBB_raw += Flux_vector_PB->data[I+N_scans_off]/N_scans_mixed;
        }
    }
    else
    {
        for(I=0; I<N_scans_mixed; I++)
        {
            Flux_vector_I1->data[I] = vndrs_mean_scan(Batch_B->fullscans[I]->I1);
            Flux_vector_I2->data[I] = vndrs_mean_scan(Batch_B->fullscans[I]->I2);
            Flux_vector_PA->data[I] = vndrs_mean_scan(Batch_B->fullscans[I]->PA);
            Flux_vector_PB->data[I] = vndrs_mean_scan(Batch_B->fullscans[I]->PB);
        }
        I1B_raw = vndrs_mean_scan(Flux_vector_I1);
        I2B_raw = vndrs_mean_scan(Flux_vector_I2);
        PAB_raw = vndrs_mean_scan(Flux_vector_PA);
        PBB_raw = vndrs_mean_scan(Flux_vector_PB);
    }

    if(PBB_raw > PAB_raw)
    {
        vndrs_computes_least_squares(Flux_vector_PB,
                                    Flux_vector_I1,
                                    &kappa1B,
                                    &ct_I1_2,
                                    &dummy);
    
        vndrs_computes_least_squares(Flux_vector_PB,
                                    Flux_vector_I2,
                                    &kappa2B,
                                    &ct_I2_2,
                                    &dummy);
    }
    else
    {
        vndrs_computes_least_squares(Flux_vector_PA,
                                    Flux_vector_I1,
                                    &kappa1A,
                                    &ct_I1_2,
                                    &dummy);
        vndrs_computes_least_squares(Flux_vector_PA,
                                    Flux_vector_I2,
                                    &kappa2A,
                                    &ct_I2_2,
                                    &dummy);
    }
    

    /* Set the elements of the transfer matrix 'Matrix' to the 2 computed values: */
    set_element_matrix_double(Matrix, 0, 0, kappa1A);
    set_element_matrix_double(Matrix, 0, 1, kappa1B);
    set_element_matrix_double(Matrix, 1, 0, kappa2A);
    set_element_matrix_double(Matrix, 1, 1, kappa2B);

    
    if(verbose_on == 1)
    {
        vndrs_fprintf_vector_double(Flux_vector_I1, "Flux_vector_I1_B.dat");
        vndrs_fprintf_vector_double(Flux_vector_I2, "Flux_vector_I2_B.dat");
        vndrs_fprintf_vector_double(Flux_vector_PA, "Flux_vector_PA_B.dat");
        vndrs_fprintf_vector_double(Flux_vector_PB, "Flux_vector_PB_B.dat");
    }
    
    /* Estimation of the error bars. They are computed by checking the residuals that are present on
    the fit of one mixed beams scans. The average of the residuals is considered. */

    /* For the particular case where we have less scans on-source than for mixed beams */
    if(N_scans_on < N_scans_mixed) N_scans_ok = N_scans_on; else N_scans_ok = N_scans_mixed;

    kappa_vec = vndrs_calloc_vector_double(N_scans_ok);
    
    for(I=0; I<N_scans_ok; I++)
    {  /* We test only on the same number of on-source scans than the mixed beams */
        Flux_vector_I1->data[I] = vndrs_mean_scan(Batch_on->fullscans[I]->I1) - Off_source_Mean->data[0];
        Flux_vector_I2->data[I] = vndrs_mean_scan(Batch_on->fullscans[I]->I2) - Off_source_Mean->data[1];
        Flux_vector_PA->data[I] = vndrs_mean_scan(Batch_on->fullscans[I]->PA) - Off_source_Mean->data[2];
        Flux_vector_PB->data[I] = vndrs_mean_scan(Batch_on->fullscans[I]->PB) - Off_source_Mean->data[3];
    }

    /* Kappa1A */
    for(I=0;I<(N_scans_ok);I++)
        if(Flux_vector_PA->data[I] > 0.0)
                kappa_vec->data[I] = (Flux_vector_I1->data[I] - kappa1B * Flux_vector_PB->data[I])/Flux_vector_PA->data[I];
        else kappa_vec->data[I] = kappa1A;
    /* We take the possible bias into account */
    sigma_kappa = vndrs_stddev_scan(kappa_vec)/sqrt(N_scans_mixed); /* we take into account the number of mixed scans */
    mean_kappa = vndrs_mean_scan(kappa_vec);
    Std_00 = sqrt(sigma_kappa*sigma_kappa + (mean_kappa - kappa1A)*(mean_kappa - kappa1A));
    set_element_matrix_double(Standard_errors, 0, 0, Std_00);

    /* Kappa1B */
    for(I=0;I<(N_scans_ok);I++)
        if(Flux_vector_PB->data[I] > 0.0)
                kappa_vec->data[I] = (Flux_vector_I1->data[I] - kappa1A * Flux_vector_PA->data[I])/Flux_vector_PB->data[I];
        else kappa_vec->data[I] = kappa1B;            
    sigma_kappa = vndrs_stddev_scan(kappa_vec)/sqrt(N_scans_mixed);
    mean_kappa = vndrs_mean_scan(kappa_vec);
    Std_01 = sqrt(sigma_kappa*sigma_kappa + (mean_kappa - kappa1B)*(mean_kappa - kappa1B));
    set_element_matrix_double(Standard_errors, 0, 1, Std_01);
    
    /* Kappa2A */
    for(I=0;I<(N_scans_ok);I++)
        if(Flux_vector_PA->data[I] > 0.0)
                kappa_vec->data[I] = (Flux_vector_I2->data[I] - kappa2B * Flux_vector_PB->data[I])/Flux_vector_PA->data[I];
        else kappa_vec->data[I] = kappa2A;
    sigma_kappa = vndrs_stddev_scan(kappa_vec)/sqrt(N_scans_mixed);
    mean_kappa = vndrs_mean_scan(kappa_vec);
    Std_10 = sqrt(sigma_kappa*sigma_kappa + (mean_kappa - kappa2A)*(mean_kappa - kappa2A));
    set_element_matrix_double(Standard_errors, 1, 0, Std_10);
    
    /* Kappa2B */
    for(I=0;I<(N_scans_ok);I++)
        if(Flux_vector_PB->data[I] > 0.0)
                kappa_vec->data[I] = (Flux_vector_I2->data[I] - kappa2A * Flux_vector_PA->data[I])/Flux_vector_PB->data[I];
        else kappa_vec->data[I] = kappa2B;
    sigma_kappa = vndrs_stddev_scan(kappa_vec)/sqrt(N_scans_mixed);
    mean_kappa = vndrs_mean_scan(kappa_vec);
    Std_11 = sqrt(sigma_kappa*sigma_kappa + (mean_kappa - kappa2B)*(mean_kappa - kappa2B));
    set_element_matrix_double(Standard_errors, 1, 1, Std_11);
 
    vndrs_free_vector_double(Flux_vector_I1);
    vndrs_free_vector_double(Flux_vector_I2);
    vndrs_free_vector_double(Flux_vector_PA);
    vndrs_free_vector_double(Flux_vector_PB);
    vndrs_free_vector_double(On_source_Mean);
    vndrs_free_vector_double(Off_source_Mean);
    vndrs_free_vector_double(On_source_Stddev);
    vndrs_free_vector_double(Off_source_Stddev);
    vndrs_free_vector_double(dummy_vec);
    vndrs_free_vector_double(kappa_vec);
}



/*-----------------------------------------------------------------------*/
/* MODIFIED version for computation of the transfer matrix at each scan. */
/* pkervell 05-mar-2002 						 */
/*-----------------------------------------------------------------------*/

void vndrs_compute_matrix_per_scan (
                           /* --------------- */
                           /* input variables */
                           /* --------------- */
                           full_scan *Scan_On,
                           /* scans obtained on source */
                           double Frame_rate,
                           /* fringe aquisition frequency (unit Hz) */
			   matrix_double *Average_Matrix,   /* added pkervell 08-mar-2002 */
			   /* The average matrix of on-source scans used in case the linfit2 does not converge */
                           /* ---------------- */
                           /* output variables */
                           /* ---------------- */
                           /* the next two entries are changed from the original concept,
                              where they were defined as vectors of matrices */
                           matrix_double *Matrix,
                           /* The transfer matrix that is computed in this routine (see comments below) */
			   int verbose_on
                          )
/* This routines computes the transfer matrix between the photometric and interferometric signals
   (both measured at photometers (see [2], Eq. (19) on p. 383):

               [ I1 ]   [ kappa_1A  kappa_1B ]   [ PA ]
               |    | = |                    | * |    |
               [ I2 ]   [ kappa_2A  kappa_2B ]   [ PB ]

   This matrix can be evaluated DIRECTLY from the data (I1, I2, PA, PB), without requiring
   an a priori knowledge of the individual transmissions and gains in the system. The evaluation
   is performed by adjusting a least square fit of a linear combination of PA and PB to I1 for
   the 2 coefficients 'kappa_1A' and 'kappa_1B' and to I2 for the 2 coefficients 'kappa_2A'
   and 'kappa_2B'.
*/
{
    /* Local variables */
    long N_frames;
    double Mean_I1_squared, Mean_I2_squared, Sample_interval_time;
    double Relative_MSE[2];
    /* Mean squared error of the fit (2 x 1 vector) */
    double Matrix_00, Matrix_01, Matrix_10, Matrix_11;
    /* The four coefficients of the transfer matrix:

               [ Matrix_00  Matrix_01 ]   [ kappa_1A  kappa_1B ]
               |                      | = |                    |
               [ Matrix_10  Matrix_11 ]   [ kappa_2A  kappa_2B ]
    */

    double kappa_1A_average, kappa_1B_average, kappa_2A_average, kappa_2B_average;
    /* the coefficients of the average on source transfer matrix between PA, PB and
       the 2 interferometric signals I1, I2 */

    double MSE_of_fit_I1, MSE_of_fit_I2;
    matrix_double *Covariance_I1, *Covariance_I2;

    individual_scan *Apodisation_filter;
    individual_scan *Calib_signals_I1, *Calib_signals_I2, *Calib_signals_PA, *Calib_signals_PB;
    individual_scan *Apodised_signals_I1, *Apodised_signals_I2, *Apodised_signals_PA, *Apodised_signals_PB;
    individual_scan *Filtered_signals_I1, *Filtered_signals_I2, *Filtered_signals_PA, *Filtered_signals_PB;
    individual_scan *Signal_fit_I1, *Signal_fit_I2;

    power_spectrum *Freq_filter;

    /* Get the number of frames */
    N_frames = Scan_On->I1->dimension;

    /* Get the average on-source matrix coefficients */
    /* added pkervell 08-mar-2002 */
    vndrs_get_transfer_matrix_coefficients(Average_Matrix,
					   &kappa_1A_average,
					   &kappa_1B_average,
					   &kappa_2A_average,
					   &kappa_2B_average);

    /* Allocate memory for some vectors */

    Calib_signals_I1    = vndrs_calloc_vector_double(N_frames);
    Calib_signals_I2    = vndrs_calloc_vector_double(N_frames);
    Calib_signals_PA    = vndrs_calloc_vector_double(N_frames);
    Calib_signals_PB    = vndrs_calloc_vector_double(N_frames);

    /*
    Filtered_signals_I1 = vndrs_calloc_vector_double(N_frames);
    Filtered_signals_I2 = vndrs_calloc_vector_double(N_frames);
    Filtered_signals_PA = vndrs_calloc_vector_double(N_frames);
    Filtered_signals_PB = vndrs_calloc_vector_double(N_frames);
    */

    Apodised_signals_I1 = vndrs_calloc_vector_double(N_frames);
    Apodised_signals_I2 = vndrs_calloc_vector_double(N_frames);
    Apodised_signals_PA = vndrs_calloc_vector_double(N_frames);
    Apodised_signals_PB = vndrs_calloc_vector_double(N_frames);

    Signal_fit_I1 = vndrs_calloc_vector_double(N_frames);
    Signal_fit_I2 = vndrs_calloc_vector_double(N_frames);
    Covariance_I1 = vndrs_calloc_matrix_double(2, 2);
    Covariance_I2 = vndrs_calloc_matrix_double(2, 2);

    /* ------------------------------------------------------------------------ */
    /* GENERATE THE 4 INDIVIDUAL SCANS USED FOR THE LEAST SQUARE FITS           */
    /* ------------------------------------------------------------------------ */

    /* Copy the 4 individual scans of the concatenated batch to the
       4 local vectors 'Calib_signals_I1', 'Calib_signals_I2',
       'Calib_signals_PA' and 'Calib_signals_PB' */
    vndrs_vector_memcpy(Calib_signals_I1, Scan_On->I1);
    vndrs_vector_memcpy(Calib_signals_I2, Scan_On->I2);
    vndrs_vector_memcpy(Calib_signals_PA, Scan_On->PA);
    vndrs_vector_memcpy(Calib_signals_PB, Scan_On->PB);

    /* ------------------------------------------------------------------------ */
    /* MULTIPLY THE ABOVE CREATED SCANS WITH THE APODISATION WINDOW             */
    /* ------------------------------------------------------------------------ */
    /* Generate apodisation window (memory for the vector 'Apodisation_filter'
       is allocated within the following routine): */
    Apodisation_filter = vndrs_apodisation_window(N_frames,
                                                  COMPUTE_MATRIX_BLOCKOUT,
                                                  COMPUTE_MATRIX_TRANSITION);

    vndrs_vector_memcpy(Apodised_signals_I1, Calib_signals_I1);
    vndrs_vector_memcpy(Apodised_signals_I2, Calib_signals_I2);
    vndrs_vector_memcpy(Apodised_signals_PA, Calib_signals_PA);
    vndrs_vector_memcpy(Apodised_signals_PB, Calib_signals_PB);

    vndrs_vector_mul(Apodised_signals_I1, Apodisation_filter);
    vndrs_vector_mul(Apodised_signals_I2, Apodisation_filter);
    vndrs_vector_mul(Apodised_signals_PA, Apodisation_filter);
    vndrs_vector_mul(Apodised_signals_PB, Apodisation_filter);

    Sample_interval_time = 1/Frame_rate;

    /* ------------------------------------------------------------------------ */
    /* PERFORM A FREQUENCY FILTERING WITH THE APODISED SCANS                    */
    /* ------------------------------------------------------------------------ */
    Freq_filter = vndrs_compute_filter(N_frames,
                                       Sample_interval_time,
                                       COMPUTE_MATRIX_CENTRAL_FREQ,
                                       COMPUTE_MATRIX_HALF_BANDWIDTH,
                                       GENERAL_FILTER_ORDER);

    Filtered_signals_I1 = vndrs_filter(Apodised_signals_I1, Freq_filter);
    Filtered_signals_I2 = vndrs_filter(Apodised_signals_I2, Freq_filter);
    Filtered_signals_PA = vndrs_filter(Apodised_signals_PA, Freq_filter);
    Filtered_signals_PB = vndrs_filter(Apodised_signals_PB, Freq_filter);

    if(verbose_on == 1)
      {
	vndrs_fprintf_vector_double(Filtered_signals_I1, "Gain_Filtered_I1.dat");
	vndrs_fprintf_vector_double(Filtered_signals_I2, "Gain_Filtered_I2.dat");
	vndrs_fprintf_vector_double(Filtered_signals_PA, "Gain_Filtered_PA.dat");
	vndrs_fprintf_vector_double(Filtered_signals_PB, "Gain_Filtered_PB.dat");
      }

    Mean_I1_squared = vndrs_mean_squared_scan(Filtered_signals_I1);
    Mean_I2_squared = vndrs_mean_squared_scan(Filtered_signals_I2);

    /* ------------------------------------------------------------------------ */
    /* COMPUTE THE TRANSFER MATRIX COEFFICIENTS                                 */
    /* ------------------------------------------------------------------------ */
    /* The four coefficients of the transfer matrix:

               [ Matrix_00  Matrix_01 ]   [ kappa_1A  kappa_1B ]
               |                      | = |                    |
               [ Matrix_10  Matrix_11 ]   [ kappa_2A  kappa_2B ]

       The 'kappa'-coefficients are computed by least square fitting the
       interferometric signals (adjusted to zero mean, apodized and filtered)
       to linear combinations of the photometric signals (pre-processed in
       the same way as their interferometric counterparts.) */

    /* (k1A,k1B): between I1 and (PA,PB) */
    vndrs_linfit_2(Filtered_signals_I1,
                   Filtered_signals_PA,
                   Filtered_signals_PB,
                   &Matrix_00,
                   &Matrix_01,
                   Covariance_I1,
                   Signal_fit_I1,
                   &MSE_of_fit_I1);


    Relative_MSE[0] = MSE_of_fit_I1/Mean_I1_squared;

    /* (k2A,k2B): between I2 and (PA,PB) */
    vndrs_linfit_2(Filtered_signals_I2,
                   Filtered_signals_PA,
                   Filtered_signals_PB,
                   &Matrix_10,
                   &Matrix_11,
                   Covariance_I2,
                   Signal_fit_I2,
                   &MSE_of_fit_I2);

    /* Check if the matrix coefficients are negative -- if yes, apply the on source average coefs. */
    /* modified pkervell 09-mar-2002 */
    if ((Matrix_10 <= 1e-3) || (Matrix_11 <= 1e-3) || (Matrix_00 <= 1e-3) || (Matrix_01 <= 1e-3))
    {
/*      sprintf(vndrs_protocol.text,"\n# Applying AVERAGE MATRIX coefficients, signal too low. \n");vndrs_runtime_protocol(); */

      Matrix_00 = kappa_1A_average;
      Matrix_01 = kappa_1B_average;
      Matrix_10 = kappa_2A_average;
      Matrix_11 = kappa_2B_average;
    }

    vndrs_free_vector_double(Signal_fit_I1);
    vndrs_free_vector_double(Signal_fit_I2);

    /* Set the elements of the transfer matrix 'Matrix' to the 2 computed values: */
    set_element_matrix_double(Matrix, 0, 0, Matrix_00);
    set_element_matrix_double(Matrix, 0, 1, Matrix_01);

    /* Set the elements of the transfer matrix 'Matrix' to the 2 computed values: */
    set_element_matrix_double(Matrix, 1, 0, Matrix_10);
    set_element_matrix_double(Matrix, 1, 1, Matrix_11);

    Relative_MSE[1] = MSE_of_fit_I2/Mean_I2_squared;

    /* Signal output storage */
    /* unlike described in [1], we are not returning the Signal_fits */

    /* Free memory */

    vndrs_free_vector_double(Calib_signals_I1);
    vndrs_free_vector_double(Calib_signals_I2);
    vndrs_free_vector_double(Calib_signals_PA);
    vndrs_free_vector_double(Calib_signals_PB);
    vndrs_free_vector_double(Apodised_signals_I1);
    vndrs_free_vector_double(Apodised_signals_I2);
    vndrs_free_vector_double(Apodised_signals_PA);
    vndrs_free_vector_double(Apodised_signals_PB);

    vndrs_free_vector_double(Filtered_signals_I1);
    vndrs_free_vector_double(Filtered_signals_I2);
    vndrs_free_vector_double(Filtered_signals_PA);
    vndrs_free_vector_double(Filtered_signals_PB);

    vndrs_free_matrix_double(Covariance_I1);
    vndrs_free_matrix_double(Covariance_I2);

    vndrs_free_vector_double(Apodisation_filter);
    vndrs_free_vector_double(Freq_filter);

    /* Commented out pkervell 08-Mar-2002
    sprintf(vndrs_protocol.text,"--------------------------------------------------\n"); vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text,"| Transfer matrix has been computed. Results:    |\n"); vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text,"|   kappa_1A = %22.14e            |\n",
            get_element_matrix_double(Matrix, 0, 0));                                    vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text,"|   kappa_1B = %22.14e            |\n",
            get_element_matrix_double(Matrix, 0, 1));                                    vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text,"|   kappa_2A = %22.14e            |\n",
            get_element_matrix_double(Matrix, 1, 0));                                    vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text,"|   kappa_2B = %22.14e            |\n",
            get_element_matrix_double(Matrix, 1, 1));                                    vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text,"|                                                |\n"); vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text,"| Fitting errors (MSE = Mean Square Error):      |\n"); vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text,"|   MSE of fit I1 = %22.14e       |\n",
            MSE_of_fit_I1);                                                              vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text,"|   sqrt          = %22.14e       |\n",
            sqrt(MSE_of_fit_I1));                                                        vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text,"|   MSE of fit I2 = %22.14e       |\n",
            MSE_of_fit_I2);                                                              vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text,"|   sqrt          = %22.14e       |\n",
            sqrt(MSE_of_fit_I2));
    sprintf(vndrs_protocol.text,"|   Rel. MSE of fit I1 = %22.14e  |\n",
            Relative_MSE[0]);                                                            vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text,"|   Rel. MSE of fit I2 = %22.14e  |\n",
            Relative_MSE[1]);                                                            vndrs_runtime_protocol();
    sprintf(vndrs_protocol.text,"--------------------------------------------------\n"); vndrs_runtime_protocol();
    */

    return;
}




void vndrs_get_transfer_matrix_coefficients(/* ----------------------------- */
                                            /* input variables               */
                                            /* ----------------------------- */ 
                                            const matrix_double *TransferMatrix,
                                            /* ----------------------------- */
                                            /* output variables              */
                                            /* ----------------------------- */
                                            double *kappa_1A,
                                            double *kappa_1B,
                                            double *kappa_2A,
                                            double *kappa_2B)
/* For a given 2 x 2 transfer matrix, this routine returns the 4 coefficients of
   the matrix */
{
    /* Security check: is dimension of transfer matrix 2 x 2 ? */
    #ifdef SECURITY_MODE
    if((TransferMatrix->dimension_1!=2)||(TransferMatrix->dimension_2!=2))
    {
        fprintf(stderr, "Error when setting the coefficients of the transfer matrix!\n");
        fprintf(stderr, "The transfer matrix does NOT have the correct dimension (2 x 2).\n");
        fprintf(stderr, "Its dimension is %ld x %ld\n", TransferMatrix->dimension_1, TransferMatrix->dimension_2);
					  /* csabet 07-Aug-2001. Format precision corrected */
        vndrs_error_alert();
    }
    #endif

    *kappa_1A = get_element_matrix_double(TransferMatrix, 0,0);
    *kappa_1B = get_element_matrix_double(TransferMatrix, 0,1);
    *kappa_2A = get_element_matrix_double(TransferMatrix, 1,0);
    *kappa_2B = get_element_matrix_double(TransferMatrix, 1,1);

    /* Security check: Are all coefficients positive? */
    #ifdef SECURITY_MODE
    if (*kappa_1A < 0. || *kappa_1B < 0. || *kappa_2A < 0. || *kappa_2B < 0.)
    {
        fprintf(stderr, "The 2 x 2 transfer matrix holds negative values.\n");
        fprintf(stderr, "Its coefficients are:\n");
        fprintf(stderr, "  %22.14e   %22.14e\n", *kappa_1A, *kappa_1B);
        fprintf(stderr, "  %22.14e   %22.14e\n", *kappa_2A, *kappa_2B);
        vndrs_error_alert();
    }
    #endif

    return;
}

/*
    Computes the average matrix and the standard deviation from
    an array of 2x2 matrices.
    Described in [1], section 2.3

    Unlike described in [1], this routine will only average one
    matrix, and thus has to be called twice. We also change the
    name accordingly.
 */
matrix_double *vndrs_average_matrix ( /* returns 2x2 matrix */
                                      /* --------------- */
                                      /* input variables */
                                      /* --------------- */
                                      vector_matrices_double *Matrices  /* vector of 2x2 matrices */
                                    )
{
    long I, Number_of_matrices;
    matrix_double *Average_matrix;

    /* calloc initializes to 0 */
    Average_matrix = vndrs_calloc_matrix_double(2,2);
    Number_of_matrices = Matrices->dimension;

    for (I=0; I<Number_of_matrices; I++)
    {
         Average_matrix->data[0] += Matrices->data[I].data[0];
         Average_matrix->data[1] += Matrices->data[I].data[1];
         Average_matrix->data[2] += Matrices->data[I].data[2];
         Average_matrix->data[3] += Matrices->data[I].data[3];
    }

    Average_matrix->data[0] /= (double)Number_of_matrices;
    Average_matrix->data[1] /= (double)Number_of_matrices;
    Average_matrix->data[2] /= (double)Number_of_matrices;
    Average_matrix->data[3] /= (double)Number_of_matrices;

    return Average_matrix;
}

/*
    Integrates the power spectral density of the fringe
    signal in the processed scan, and normalizes it to
    return the squared coherence factors Mu_squared for
    each interferometric channel.
    VERSION WITH OFF-SOURCE
    Described in [1], section 2.8
 */
double vndrs_integrate_fringe_power (  /* --------------- */ 
                                     /* input variables */ 
                                     /* --------------- */ 
                                     individual_scan *Scan_on, 
                                     individual_scan *Scan_off, 
                                     double Wavenumber_interval, 
                                     double Sigma_min, 
                                     double Sigma_max, 
				     int verbose_on,
                                     /* outputs */
                                     power_spectrum *Sp_power_cal, /* added for checking calib - pkervell 02-Sep-02 */
                                     vector_double *Real_part,
                                     vector_double *Imaginary_part,
                                     vector_double *Phase,
                                     power_spectrum *Sp_power_off_cal)
{
    long N_frames;
    double Offset, Area, Mu_squared;
    individual_scan *Scan_on_cal, *Scan_off_cal;
    power_spectrum *Sp_power_on, *Sp_power_off, *Sp_power_off_smooth;
    vector_complex *fft_scan;
    vector_double *modulus_fft, *re_fft, *im_fft;

    /* Set number of frames */
    N_frames = Scan_on->dimension;
    
    Scan_on_cal = vndrs_calloc_vector_double(N_frames);
    Scan_off_cal = vndrs_calloc_vector_double(N_frames);
    fft_scan = vndrs_calloc_vector_complex(N_frames);
    modulus_fft = vndrs_calloc_vector_double(N_frames);
    re_fft = vndrs_calloc_vector_double(N_frames);
    im_fft = vndrs_calloc_vector_double(N_frames);
    
    Offset = vndrs_mean_scan(Scan_off);

    if(verbose_on == 1)
      {
	vndrs_fprintf_vector_double(Scan_on, "Scan_on.dat");
	vndrs_fprintf_vector_double(Scan_off, "Scan_off.dat");
      }

    vndrs_vector_memcpy(Scan_on_cal, Scan_on);
    vndrs_vector_add_constant(Scan_on_cal, -Offset);
    Sp_power_on = vndrs_power_spectrum(Scan_on_cal);

    
    vndrs_zero_pos_scan_center(Scan_on); /* added pkervell 25-Oct-02 */
    vndrs_fft_real_2_complex(Scan_on, fft_scan);
    vndrs_retrieve_modulus(fft_scan, modulus_fft);
    vndrs_retrieve_phase(fft_scan, Phase); /* added pkervell 25-Oct-02 */
    vndrs_retrieve_re_im(fft_scan, re_fft, im_fft);
    /* Normalization after fftw */
    vndrs_vector_scale(re_fft, sqrt(N_frames));
    vndrs_vector_scale(im_fft, sqrt(N_frames));
    
    /* vndrs_fprintf_vector_double(re_fft, "re_fft.dat");
    vndrs_fprintf_vector_double(im_fft, "im_fft.dat");
    vndrs_fprintf_vector_double(modulus_fft, "modulus_fft.dat");
    vndrs_fprintf_vector_double(Phase, "phase_fft.dat");*/
    vndrs_vector_memcpy(Real_part, re_fft);    
    vndrs_vector_memcpy(Imaginary_part, im_fft);    
   
    vndrs_vector_add_constant(Scan_off_cal, -Offset);
    vndrs_vector_memcpy(Scan_off_cal, Scan_off);
    Sp_power_off = vndrs_power_spectrum(Scan_off_cal);

    /* smoothing of the off-source power spectrum */
    /* printf("Sp_power_off->data[1] before smooth = %f\n",Sp_power_off->data[1]);  */
    Sp_power_off_smooth = vndrs_vector_smooth(Sp_power_off, (long)N_frames/64.0);
    /*printf("Sp_power_off->data[1] after smooth = %f\n",Sp_power_off_smooth->data[1]);*/
    
    /* Sp_power_off_smooth->data[0] = 0.0; zero average */
    
    if(verbose_on == 1)
      {
	vndrs_fprintf_vector_double(Sp_power_on, "Sp_power_on.dat");
	vndrs_fprintf_vector_double(Sp_power_off, "Sp_power_off.dat");
	vndrs_fprintf_vector_double(Sp_power_off_smooth, "Sp_power_off_smooth.dat");
      } 
      
    /* Subtraction of the off-source power spectrum from the on-source */
    vndrs_vector_sub(Sp_power_on, Sp_power_off);

    /* copy to output power spectrum variable */
    vndrs_vector_memcpy(Sp_power_cal, Sp_power_on);    
    vndrs_vector_memcpy(Sp_power_off_cal, Sp_power_off);

    vndrs_vector_scale(Sp_power_cal, Wavenumber_interval); /* Normalization to keep the same as Mu^2 */    
    vndrs_vector_scale(Sp_power_off_cal, Wavenumber_interval);
    
    /* Computation of the power in the fringes */
    Area = vndrs_area_power_dark_no_background(Sp_power_on,
                                 Wavenumber_interval, 
                                 Sigma_min, 
                                 Sigma_max,
				 verbose_on); 

    Mu_squared = Area; /* all normalization is now done outside - 26-Sep-02 */

    vndrs_free_vector_double(Scan_on_cal);
    vndrs_free_vector_double(Scan_off_cal);
    vndrs_free_vector_double(Sp_power_on);
    vndrs_free_vector_double(Sp_power_off);
    vndrs_free_vector_double(Sp_power_off_smooth);
    vndrs_free_vector_complex(fft_scan);
    vndrs_free_vector_double(modulus_fft);
    vndrs_free_vector_double(re_fft);
    vndrs_free_vector_double(im_fft);
    
    return Mu_squared;
}


/*  Wavelet transform based computation of mu^2:
    Integrates the power spectral density of the fringe 
    signal in the processed scan, and normalizes it to 
    return the squared coherence factors Mu_squared for 
    each interferometric channel. 
    
    This version of the procedure subtracts the background found at high
    frequency (mostly photon noise bias).
 */ 

double vndrs_integrate_wavelets_power(matrix_double *norm_t_wavelets,
				      double Wavenum_interval,
				      double Wavenum_limits_low,
				      double Wavenum_limits_high,
				      double fringes_length,
				      int verbose_on)
{  /* created pkervell 18-mar-02 */
  double mu_squared;
  long i, j, size_signal;
  double sum_in;
  long index_freq_low, index_freq_high;
  long index_time_min, index_time_max;

  size_signal = norm_t_wavelets->dimension_1;

  /* ******************
     Integration of the fringe peak power, on a rectangular window
     The limits of the integration in pixels are given by the formula: Wavenum_limit/Wavenum_interval
  ****************/

  /* limits in frequency */
  index_freq_low  = Wavenum_limits_low/Wavenum_interval;
  index_freq_high = Wavenum_limits_high/Wavenum_interval;

  /* limits in the time domain */
  index_time_min = (norm_t_wavelets->dimension_1)/2 - fringes_length/2;
  index_time_max = (norm_t_wavelets->dimension_1)/2 + fringes_length/2;

  sum_in = 0.0;

  for(i=index_time_min;i<index_time_max;i++)
      for(j=index_freq_low;j<index_freq_high;j++)
	  sum_in = sum_in + get_element_matrix_double(norm_t_wavelets, i, j);

  mu_squared = sum_in; /* normalization done outside - pkervell 26-Sep-02 */

  return(mu_squared);
}


/*
    Described in [1], section 2.9
    CASE WITH THE PRESENCE OF OFF-SOURCE - pkervell 04-Aug-02
*/
double vndrs_area_power_dark (
                              /* --------------- */
                              /* input variables */
                              /* --------------- */
                              power_spectrum *Spectral_power,
                              power_spectrum *Dark_spectral_power,
                              double Freq_interval,
                              double Fmin,
                              double Fmax,
			      int verbose_on
                             )
{
    power_spectrum *Spectral_power_cal;
    double Integral_power;

    Spectral_power_cal = vndrs_calloc_vector_double(Spectral_power->dimension);
    
    vndrs_vector_memcpy(Spectral_power_cal, Spectral_power);
    vndrs_vector_sub(Spectral_power_cal, Dark_spectral_power);
    
    if(verbose_on == 1)
      vndrs_fprintf_vector_double(Spectral_power_cal, "Spectral_power_cal.dat");

    Integral_power = vndrs_area(Spectral_power_cal, Freq_interval, Fmin, Fmax);

    vndrs_free_vector_double(Spectral_power_cal);

    return Integral_power;
}

/*
    Described in [1], section 2.9
    CASE WITHOUT OFF-SOURCE - pkervell 04-Aug-02
*/
double vndrs_area_power_dark_no_background (
                              /* --------------- */
                              /* input variables */
                              /* --------------- */
                              power_spectrum *Spectral_power,
                              double Freq_interval,
                              double Fmin,
                              double Fmax,
			      int verbose_on
                             )
{
    double Integral_power;

    if(verbose_on == 1)
      vndrs_fprintf_vector_double(Spectral_power, "Spectral_power_cal.dat");

    Integral_power = vndrs_area(Spectral_power, Freq_interval, Fmin, Fmax);

    return Integral_power;
}



/*
    Computes the area of an array x(T) between Tmin and Tmax.
    The area is computed by the trapeze method between the
    fractional indices corresponding to the boundaries. The
    value of x(t) is linearly interpolated between Tmin and Tmax.
    The abscissa origin is T0=0. By default, dt=1. This function
    works even when Tmax-Tmin<1.
    Described in [1], section 2.10
*/

double vndrs_area (
                   /* --------------- */
                   /* input variables */
                   /* --------------- */
                   vector_double *Array,
                   double Dt,
                   double Tmin,
                   double Tmax
                  )
{
    long I, Kmin, Kmax;
    double Integral, Xkmin, Xkmax, Xmin, Xmax, Kmin_dt, Kmax_dt;
    double Sum_Xk, Sum_Xk_dt;

    if (Dt == 0.)
    {
        fprintf(stderr, "The Sampling interval is 0.\n");
        vndrs_error_alert();
    }
    if (Tmin < 0. || Tmax < 0.
     || Tmin/Dt > (Array->dimension)-1 || Tmax/Dt > (Array->dimension)-1)
    {
        fprintf(stderr, "Try to integrate the area outside of the defined function range.\n");
        fprintf(stderr, "Available range: 0, %ld.\n", (Array->dimension)-1);
        fprintf(stderr, "Requested range: %22.14e, %22.14e.\n", Tmin/Dt, Tmax/Dt);
        vndrs_error_alert();
    }

    /*
       Limits for linear interpolation
       Kmin = index of the sample immediately following Tmin
       Xkmin is the value of that sample
       Kmax = index of the sample immediately preceding Tmax
       Xkmax is the value of that sample
     */

    Kmin = (long)(Tmin/Dt) + 1;
    Kmax = (long)(Tmax/Dt);
    Xkmin = Array->data[Kmin];
    Xkmax = Array->data[Kmax];

    /*
       Xmin is the interpolated value at T=Tmin
     */
    Xmin = vndrs_linear_interp(Array, Tmin/Dt);

    /*
       Xmax is the interpolated value at T=Tmax
     */
    Xmax = vndrs_linear_interp(Array, Tmax/Dt);

    /* Compute the 2 values in units of 'Dt' corresponding to
       the 2 limiting indices Kmin and Kmax */
    Kmin_dt = Kmin * Dt;
    Kmax_dt = Kmax * Dt;

    /*
       Computation of the sum of the Array, between Kmin+1 and Kmax-1
     */
    Sum_Xk = 0.;
    /* Modification, Jan 04/01:
       Instead of looping from Kmin+1 to Kmax-1 we
       loop from Kmin+1 to Kmax. */
    for (I=Kmin+1; I<=Kmax; I++)
    {
        Sum_Xk += Array->data[I];
    }

    Sum_Xk_dt = Sum_Xk * Dt;
    
    Integral = Sum_Xk_dt;

    return Integral;
}

/*
    Described in [1], section 2.11
*/
double vndrs_linear_interp (
                            /* --------------- */
                            /* input variables */
                            /* --------------- */
                            vector_double *Array,
                            double Tinterp
                           )
{
     long X1, X2;

     X1 = (long)Tinterp;
     X2 = X1 + 1;

     /* the formula given in [1] is wrong */
     return ( Array->data[X1] + (Array->data[X2] - Array->data[X1]) * (Tinterp - X1) );
}

/*
    This function computes the proportionality factor between
    two sets of data.
    Described in [1], section 2.18
*/
double vndrs_linfit (
                     /* --------------- */
                     /* input variables */
                     /* --------------- */
                     vector_double *Signal_1,
                     vector_double *Signal_2
                    )
{
    long N_frames;
    double XX, XY;
    vector_double *dummy;

    /* the size of the signals must be the same */
    N_frames = Signal_1->dimension;

    dummy = vndrs_calloc_vector_double(N_frames);

    /* Computation of the differential gains S1/S2: this is done by computing the
       linear regression factor between the two signals. */
    
    vndrs_vector_memcpy(dummy, Signal_1);
    vndrs_vector_mul(dummy, Signal_2);
    XY = vndrs_sum_scan(dummy);

    vndrs_vector_memcpy(dummy, Signal_1);
    vndrs_vector_mul(dummy, Signal_1);
    XX = vndrs_sum_scan(dummy);

    vndrs_free_vector_double(dummy);
    
    return (XY/XX);
}

/*
    This funtion computes the coefficients a1 and a2 which give the
    best fit (least square method) between Signal and a1*Signal1+a2*Signal2.
    Described in [1], section 2.19
 */
void vndrs_linfit_2 (
                     /* --------------- */
                     /* input variables */
                     /* --------------- */
                     individual_scan *Signal,
                     individual_scan *Signal1,
                     individual_scan *Signal2,
                     /* ---------------- */
                     /* output variables */
                     /* ---------------- */
                     double *a1,
                     double *a2,
                     matrix_double *Covariances,    /* 2x2 matrix */
                     individual_scan *Signal_fit,   /* Fitted signal */
                     double *MSE_of_fit
                    )
{
    long N_frames;
    double XX, XY, XZ, YY, YZ;
    double denominator;
    vector_double *dummy;
    
    /* the size of the signals must be the same */
    N_frames = Signal->dimension;

    dummy = vndrs_calloc_vector_double(N_frames);

    /* Computation of the differential gains S1/S2: this is done by computing the
       linear regression factor between the two signals. */

    vndrs_vector_memcpy(dummy, Signal1);
    vndrs_vector_mul(dummy, Signal1);
    XX = vndrs_sum_scan(dummy);
    
    vndrs_vector_memcpy(dummy, Signal1);
    vndrs_vector_mul(dummy, Signal2);
    XY = vndrs_sum_scan(dummy);

    vndrs_vector_memcpy(dummy, Signal1);
    vndrs_vector_mul(dummy, Signal);
    XZ = vndrs_sum_scan(dummy);

    vndrs_vector_memcpy(dummy, Signal2);
    vndrs_vector_mul(dummy, Signal2);
    YY = vndrs_sum_scan(dummy);
    
    vndrs_vector_memcpy(dummy, Signal2);
    vndrs_vector_mul(dummy, Signal);
    YZ = vndrs_sum_scan(dummy);

    denominator = XY*XY - YY*XX;

    if (denominator == 0.)
    {
        fprintf(stderr, "linfit_2 can not find a solution.\n");
        vndrs_error_alert();
    }

    *a1 = (YZ*XY - XZ*YY) / denominator;
    *a2 = (XZ*XY - YZ*XX) / denominator;

    /* Compute the result of the fit */
    /* ============================= */
    vndrs_vector_memcpy(Signal_fit, Signal1); /* copy Signal1 to Signal_fit */
    vndrs_vector_scale(Signal_fit, *a1);      /* multiply it by a1 */
    vndrs_vector_memcpy(dummy, Signal2);      /* copy Signal2 to dummy */
    vndrs_vector_scale(dummy, *a2);           /* multiply it by a2 */    
    vndrs_vector_add(Signal_fit, dummy);      /* add Signal_fit to dummy; result is Signal_fit */
    
    /* Compute MSE of fit */
    vndrs_vector_memcpy(dummy, Signal);             /* copy Signal to dummy */
    vndrs_vector_sub(dummy, Signal_fit);            /* subtract Signal_fit from dummy */
    *MSE_of_fit = vndrs_mean_squared_scan(dummy);   /* take the mean of the squared vector--> M(ean) S(quared) E(rror) */
    /* printf("MSE of fit: %12e \n", *MSE_of_fit); */
    
    vndrs_free_vector_double(dummy);

    return;
}



/* -------------------------------------------------------------- */
/* |                                                            | */
/* |   ADDITIONAL SUPPORT ROUTINES                              | */
/* |                                                            | */
/* -------------------------------------------------------------- */

void vndrs_error_alert(
		      /* input variables */
		      void
		      /* output variables */
		      )
{
    printf("> VINCI Data Reduction Software had to stop because of an error.\n");
	fprintf(stderr, "\n");
    exit(0);
}

		
/* ------------------------------------------------------------- */
/* ALL KINDS OF CONVERSIONS                                      */
/* ------------------------------------------------------------- */

void vndrs_vector_double_2_vector_complex(
                                          /* --------------- */
                                          /* input variables */
                                          /* --------------- */
                                          const vector_double *RealVector,
                                          /* ---------------- */
                                          /* output variables */
                                          /* ---------------- */
                                          vector_complex *ComplexVector
                                         )
/* This routine converts the REAL-valued input vector 'RealVector' into
   a COMPLEX vector 'ComplexVector' whose real part matches the real vector
   and whose imaginary part is zero.

   Make sure that memory for both vectors has been allocated before
   calling this function!! */
{
    long i;

    for(i=0; i<RealVector->dimension; i++)
    {
        ComplexVector->data[i].re = RealVector->data[i];
        ComplexVector->data[i].im = 0.;
    }
    return;
}

void vndrs_vector_complex_2_vector_double(
                                          /* --------------- */
                                          /* input variables */
                                          /* --------------- */
                                          const vector_complex *ComplexVector,
                                          /* ---------------- */
                                          /* output variables */
                                          /* ---------------- */                                          
                                          vector_double *RealVector
                                         )
/* This routine converts the COMPLEX-valued input vector 'ComplexVector' into
   a REAL-valued vector 'RealVector' whose components are the modulus (!!)
   of the components of the COMPLEX-valued vector.
 
   We use the modulus instead of the real-part! One could also use the real-part,
   since --within the VINCI Data Reduction Software-- the complex vector will
   have imaginary part zero. However, taking the modulus is more safe, since,
   in any case, the power of the signal is conserved.

   Make sure that memory for both vectors has been allocated before
   calling this function!! */
{
    long i;

    for(i=0; i<ComplexVector->dimension; i++)
    {
        RealVector->data[i] = vndrs_abs_complex(ComplexVector->data[i]);
    }
    return;
}

void vndrs_vector_complex_2_vector_double_real(
                                          /* --------------- */
                                          /* input variables */
                                          /* --------------- */
                                          const vector_complex *ComplexVector,
                                          /* ---------------- */
                                          /* output variables */
                                          /* ---------------- */                                          
                                          vector_double *RealVector
                                         )
/* This routine converts the COMPLEX-valued input vector 'ComplexVector' into
   a REAL-valued vector 'RealVector' whose components are the real part
   of the COMPLEX-valued vector.
 
   Make sure that memory for both vectors has been allocated before
   calling this function!! */
{
    long i;

    for(i=0; i<ComplexVector->dimension; i++)
    {
        RealVector->data[i] = ComplexVector->data[i].re;
    }
    return;
}

/* Routine that averages over a batch each of the four
   scan types (I1,I2,PA,PB) and returns one fullscan */
full_scan *vndrs_average_batch_to_fullscan(
                                           /* --------------- */
                                           /* input variables */
                                           /* --------------- */ 
                                           batch *Batch
                                          )
/* For a given batch 'Batch', this routine averages over all full scans forming
   the batch. Example: The batch has the size 4 x 256 x 1000 (i.e. 1000 full
   scans each consisting of 4 individ. scans (I1, I2, PA, PB) each having
   a length of 256 data points. The result of the application of this
   routine would be a single full scan of size 4 x 256.

   IMPORTANT: Memory for the full scan (output variable) is allocated
   within this routine! */
{
    /* Local variables */
    /* long i;	*/
    unsigned int i;	/* long is not used here csabet 26.Jun.01 */
    long dimension_of_individual_scans_in_batch;
    full_scan *FullScan;
    double ScalingFactor;
    
    vector_double *Sum_I1, *Sum_I2, *Sum_PA, *Sum_PB;

    /* Get dimension of the 4 individual scans forming a full scan */
    dimension_of_individual_scans_in_batch = Batch->fullscans[0]->I1->dimension;
    
    /* Allocate memory for the resulting fullscan */
    FullScan = vndrs_calloc_full_scan(dimension_of_individual_scans_in_batch);    
    
    /* 'Fill' the full scan (=average over the batch) */
    /* ---------------------------------------------- */
    /* Allocate memory for the 'sum' vectors (and set their components to zero) */
    Sum_I1 = vndrs_calloc_vector_double(dimension_of_individual_scans_in_batch);
    Sum_I2 = vndrs_calloc_vector_double(dimension_of_individual_scans_in_batch);
    Sum_PA = vndrs_calloc_vector_double(dimension_of_individual_scans_in_batch);
    Sum_PB = vndrs_calloc_vector_double(dimension_of_individual_scans_in_batch);

    for(i=0; i<Batch->number_of_scans; i++)
    {
        vndrs_vector_add(Sum_I1, Batch->fullscans[i]->I1);
        vndrs_vector_add(Sum_I2, Batch->fullscans[i]->I2);
        vndrs_vector_add(Sum_PA, Batch->fullscans[i]->PA);
        vndrs_vector_add(Sum_PB, Batch->fullscans[i]->PB);
    }

    /* Now, the average individual scans can be computed by scaling the
       'sum' vectors by (1/Batch->number_of_scans) */
    ScalingFactor = 1./((double)(Batch->number_of_scans));
    vndrs_vector_scale(Sum_I1, ScalingFactor);
    vndrs_vector_scale(Sum_I2, ScalingFactor);
    vndrs_vector_scale(Sum_PA, ScalingFactor);
    vndrs_vector_scale(Sum_PB, ScalingFactor);
    
    vndrs_vector_memcpy(FullScan->I1, Sum_I1);
    vndrs_vector_memcpy(FullScan->I2, Sum_I2);
    vndrs_vector_memcpy(FullScan->PA, Sum_PA);
    vndrs_vector_memcpy(FullScan->PB, Sum_PB);

    /* Free memory for the 'sum' vectors (not needed any more) */
    vndrs_free_vector_double(Sum_I1);
    vndrs_free_vector_double(Sum_I2);
    vndrs_free_vector_double(Sum_PA);
    vndrs_free_vector_double(Sum_PB);

    /* Return the full scan (as the result) */
    return FullScan;
}

/* Routine that concatenates two batches producing a single batch */
batch *vndrs_concatenate_batches_scan_direction(
                                                /* --------------- */
                                                /* input variables */
                                                /* --------------- */ 
                                                const batch *Batch_A,
                                                const batch *Batch_B
                                               )
/* For the 2 batches 'Batch_A' and 'Batch_B' this routine generates a new
   batch that has twice the size. Example: One has 2 batches of size
   4 x 256 x 1000 (i.e. each batch consists of 1000 full scans,
   each full scan consisting of 4 individual scans each having a
   length of 256 data points.
   The 'concatenated' batch (output of this routine) will have the size
   4 x 256 x 2000.

   Memory for the concatenated batch is allocated within this routine!!
*/
{
    /* local variables */
    long i;
    batch *ConcatenatedBatch;
    long number_of_full_scans_in_concatenated_batch;
    long dimension_of_individual_scans_in_concatenated_batch; 
    
    #ifdef SECURITY_MODE
    /* Check if the 2 batches to be concatenated share the same length of their
       full scans */
    if((Batch_A->fullscans[0]->I1->dimension)!=(Batch_B->fullscans[0]->I1->dimension))
    {
        fprintf(stderr, "Error in routine 'vndrs_concatenate_batches_scan_direction'");
        fprintf(stderr, "When concatenate 2 batches\n");
        fprintf(stderr, "the individual scan length must be equal in both batches.\n");
        fprintf(stderr, "You tried to concatenate 2 batches having the following\n");
        fprintf(stderr, "individual scan lengths: %ld, %ld\n",
                Batch_A->fullscans[0]->I1->dimension,
                Batch_B->fullscans[0]->I1->dimension );  /* csabet 07-Aug-2001. Format precision corrected */
        vndrs_error_alert();
    }
    #endif
    
    /* Now, it is made sure that the individual scans of the 2 batches to be concatenated,
       share the same dimension. We store this dimension in a local variable. */
    dimension_of_individual_scans_in_concatenated_batch = Batch_A->fullscans[0]->I1->dimension;
    
    /* Compute number of full scans in concatenated batch (= sum of the number of full
       scans in the 2 batches to be concatenated) */
    number_of_full_scans_in_concatenated_batch = Batch_A->number_of_scans + Batch_B->number_of_scans;

    /* Allocate memory for the concatenated batch */
    ConcatenatedBatch = vndrs_calloc_batch(number_of_full_scans_in_concatenated_batch,
                                           dimension_of_individual_scans_in_concatenated_batch);
    
    /* FILL CONCATENATED BATCH          */
    /* -------------------------------- */
    for(i=0; i<number_of_full_scans_in_concatenated_batch; i++)
    /* Loop; all full scans of the concatenated batch */
    {
        /* Check if we are in the 'first half' of the batch or
           if we are in the second half ... */
        /* if(i<Batch_A->number_of_scans)	*/
        if(i< (long) (Batch_A->number_of_scans))	/* To be assessec. csabet 25.Jun.01 */
	{
	    /* First half of concatenated batch: Copy the contents of 'Batch_A' */
            vndrs_vector_memcpy(ConcatenatedBatch->fullscans[i]->I1, Batch_A->fullscans[i]->I1);
            vndrs_vector_memcpy(ConcatenatedBatch->fullscans[i]->I2, Batch_A->fullscans[i]->I2);
            vndrs_vector_memcpy(ConcatenatedBatch->fullscans[i]->PA, Batch_A->fullscans[i]->PA);
            vndrs_vector_memcpy(ConcatenatedBatch->fullscans[i]->PB, Batch_A->fullscans[i]->PB);
        }
        else
	{
            /* Second half of concatenated batch: Copy the contents of 'Batch_B' */
	    vndrs_vector_memcpy(ConcatenatedBatch->fullscans[i]->I1,
                                Batch_B->fullscans[i-Batch_A->number_of_scans]->I1);
            vndrs_vector_memcpy(ConcatenatedBatch->fullscans[i]->I2,
                                Batch_B->fullscans[i-Batch_A->number_of_scans]->I2);
            vndrs_vector_memcpy(ConcatenatedBatch->fullscans[i]->PA,
                                Batch_B->fullscans[i-Batch_A->number_of_scans]->PA);
            vndrs_vector_memcpy(ConcatenatedBatch->fullscans[i]->PB,
                                Batch_B->fullscans[i-Batch_A->number_of_scans]->PB);
        }
    }

    /* Return concatenated batch as output variable */
    return ConcatenatedBatch;
}

/* Routine that concatenates two batches producing a single batch (scan direction) */
batch *vndrs_concatenate_batches_frame_direction(
                                                 /* --------------- */
                                                 /* input variables */
                                                 /* --------------- */ 
                                                 const batch *Batch_A,
                                                 const batch *Batch_B
                                                )
/* For the 2 batches 'Batch_A' and 'Batch_B' this routine generates a new
   batch whose individual scans have twice the sice.
   Example: One has 2 batches of size 4 x 256 x 1000 (i.e. each batch
   consists of 1000 full scans, each full scan consisting of 4 individual
   scans each having a length of 256 data points.
   The 'concatenated' batch (output of this routine) will have the size
   4 x 512 x 1000.

   Memory for the concatenated batch is allocated within this routine!!
*/
{
    /* local variables */
    long i, j;
    batch *ConcatenatedBatch;
    long number_of_full_scans_in_concatenated_batch;
    long dimension_of_individual_scans_in_concatenated_batch;
    long dimension_of_individual_scans_in_single_batch; 
    
    #ifdef SECURITY_MODE
    /* Check if the 2 batches to be concatenated share the same length of their
       individual scans */
    if((Batch_A->fullscans[0]->I1->dimension)!=(Batch_B->fullscans[0]->I1->dimension))
    {
        fprintf(stderr, "Error in routine 'vndrs_concatenate_batches_frame_direction'.");
        fprintf(stderr, "When concatenate 2 batches\n");
        fprintf(stderr, "the individual scan length must be equal in both batches.\n");
        fprintf(stderr, "You tried to concatenate 2 batches having the following\n");
        fprintf(stderr, "individual scan lengths: %ld, %ld\n",  /* csabet 07-Aug-2001. Format precision corrected */
                Batch_A->fullscans[0]->I1->dimension,
                Batch_B->fullscans[0]->I1->dimension );
        vndrs_error_alert();
    }
    #endif
    
    
    /* Now, it is made sure that the individual scans of the 2 batches to be concatenated,
       share the same dimension. The individual scan dimension of the concatenated batch
       is twice the individual scan dimension of each scan. We store this dimension
       in a local variable. */
    dimension_of_individual_scans_in_single_batch = Batch_A->fullscans[0]->I1->dimension;
    dimension_of_individual_scans_in_concatenated_batch = 2*dimension_of_individual_scans_in_single_batch;
    
    /* The number of full scans in concatenated batch is equal to the number of full
       scans in each of the 2 batches to be concatenated: */
    number_of_full_scans_in_concatenated_batch = Batch_A->number_of_scans;

    /* Allocate memory for the concatenated batch */
    ConcatenatedBatch = vndrs_calloc_batch(number_of_full_scans_in_concatenated_batch,
                                           dimension_of_individual_scans_in_concatenated_batch);
    
    
    /* FILL CONCATENATED BATCH          */
    /* -------------------------------- */
    for(i=0; i<number_of_full_scans_in_concatenated_batch; i++)
    /* Loop; all full scans of the concatenated batch */
    {
        /* Loop over the 4 individual scans of the full scan #i */
        for(j=0; j<dimension_of_individual_scans_in_concatenated_batch; j++)
        {
            /* Check if we are in the 'first half' of the individual scan */
            if(j<dimension_of_individual_scans_in_single_batch)
            {
                /* We are in the first half of the individual scans, --> take Batch A */
                ConcatenatedBatch->fullscans[i]->I1->data[j] = Batch_A->fullscans[i]->I1->data[j];
                ConcatenatedBatch->fullscans[i]->I2->data[j] = Batch_A->fullscans[i]->I2->data[j];
                ConcatenatedBatch->fullscans[i]->PA->data[j] = Batch_A->fullscans[i]->PA->data[j];
                ConcatenatedBatch->fullscans[i]->PB->data[j] = Batch_A->fullscans[i]->PB->data[j];
            }
            else
            {
                /* We are in the first half of the individual scans, --> take Batch B */
                ConcatenatedBatch->fullscans[i]->I1->data[j] = 
                    Batch_B->fullscans[i]->I1->data[j-dimension_of_individual_scans_in_single_batch];
                ConcatenatedBatch->fullscans[i]->I2->data[j] = 
                    Batch_B->fullscans[i]->I2->data[j-dimension_of_individual_scans_in_single_batch];
                ConcatenatedBatch->fullscans[i]->PA->data[j] = 
                    Batch_B->fullscans[i]->PA->data[j-dimension_of_individual_scans_in_single_batch];
                ConcatenatedBatch->fullscans[i]->PB->data[j] = 
                    Batch_B->fullscans[i]->PB->data[j-dimension_of_individual_scans_in_single_batch];
            }
        }  
    }

    /* Return concatenated batch as output variable */
    return ConcatenatedBatch;
}

double vndrs_juliandate_to_local_sidereal(double julian_date, double longitude)
{

    /* Method taken from http://www.seti.net/calculators.htm
       longitude in radians
       local sidereal time in decimal hours */
    double GST, LST, T0, T, UTdate, UTtime;
    double  Pi = 3.14159265358979311599796346854418516;

    /* Julian date for UT 0h GMT for the considered date */
    /* changed trunc to floor for compatibility - pkervell 20may02 */
    if((julian_date-floor(julian_date))>0.5) UTdate = floor(julian_date) + 0.5;
    else UTdate = floor(julian_date)-0.5;

    UTtime = (julian_date - UTdate)*24.0;  /* Greenwich Mean Time (UTC) in decimal hours */

    T = (UTdate - 2451545.0)/36525.0;
    T0 = 6.697374558 + (2400.051336*T) + (0.000025862*T*T) + (UTtime*1.0027379093);
    GST = (T0/24.0 - floor(T0/24.0))*24;
    LST = GST - longitude*(12.0/Pi); /* longitude is in radians, beware of sign convention of longitude
					West is taken >0 ! */
    if(LST<0) LST = LST+24.0;
    LST = (LST/24.0 - floor(LST/24.0))*24;
    /* printf(" GST = %f \n",GST);
       printf(" LST = %f \n",LST); */
    return LST;

}

void vndrs_convert_HA_Dec_to_Alt_Az(double hour_angle,
				    double Dec_object,
				    double latitude,
				    double *altitude,
				    double *azimuth)
{
/* hour_angle is in decimal hours,
   Dec_object is in degrees,
   latitude is in radians,
   Method taken from dltestCoordU1U3.c, written by AWA for delay line testing */

    double sinofalt, decRad, haRad, alt, altD, azD, tazx, tazy, az;
    double  Pi = 3.14159265358979311599796346854418516;

    /* printf("hour_angle = %f, Dec_object = %f, latitude = %f\n",hour_angle,Dec_object,latitude); */
    decRad = Dec_object*Pi/180.0;
    haRad = hour_angle*Pi/12.0;

    sinofalt = (sin(decRad)*sin(latitude))+
               (cos(decRad)*cos(latitude)*cos(haRad));
    alt = asin(sinofalt);

    tazx = cos(decRad)*sin(haRad);
    tazy = (sin(latitude)*cos(decRad)*cos(haRad)-cos(latitude)*sin(decRad));
    az = atan2(tazx,tazy);
    /* same convention as in header of FITS files chosen: S=0, cw */
    az = (az/(2.0*Pi)-floor(az/(2.0*Pi)))*(2*Pi);  /* modulo 2*Pi */
    /* changed trunc to floor for compatibility reasons - pkervell 20may02 */
    if (az<0.0) az += 2*Pi;
    
    /* in degrees and understandable range */
    azD = az*180.0/Pi;
    altD = alt*180.0/Pi;
    if (azD>180.) azD = azD-360.0;
    if (azD<0.) azD = azD+360.0;
    if (altD>180.) altD = altD-360.0;

/*    printf(" altD = %f \n ",altD);
      printf(" azD = %f \n ",azD); */
    *altitude = altD;
    *azimuth = azD;
/*    printf(" Altitude = %f \n ", *altitude);
      printf(" Azimuth  = %f \n ", *azimuth);*/
}

void vndrs_project_baseline(double hour_angle,
			    double Dec_object,
			    double latitude,
			    double t1x, double t1y, double t1z,
			    double t2x, double t2y, double t2z,
                            double *altitude_object,
                            double *azimuth_object,
			    double *projected_baseline,
			    double *azimuth_baseline,
			    double *u,
			    double *v)
{
    double baseline, angle_uv, sidereal_OPD, az;
    double  Pi = 3.14159265358979311599796346854418516;
    double dx, dy, dz, w, cha, sha, cdec, sdec, clat, slat, ihar, decr;
    double buffer_alt_obj, buffer_az_obj;
    
    dx = t2x-t1x;
    dy = t2y-t1y;
    dz = t2z-t1z;
    /*  printf(" DX = %f, DY = %f, DZ = %f \n",dx,dy,dz);*/


    /* 1) Method taken from dltestCoordU1U3.c, written by AWA for delay lines testing */
    /* ------------------------------------------------------------------------------ */

    /* When az in 90 deg from baseline angle we should have sidereal_OPD = 0
       this angle is (az=0 at south) +90degs (to go to positive U axis),
       +angle_uv (which is 38.660+18.984 for UT1-UT3 for example).

       Baseline azimuth is in degrees, projected baseline is in meters */

    vndrs_convert_HA_Dec_to_Alt_Az(hour_angle, Dec_object, latitude, &buffer_alt_obj, &buffer_az_obj);
    *altitude_object = buffer_alt_obj;
    *azimuth_object = buffer_az_obj;

    baseline = sqrt(dx*dx + dy*dy + dz*dz); /* ground baseline */
    angle_uv = -atan2(dy,dx); /* angle of baseline on the ground in radians, south is 0, counted cw */
    /* Azimuth of the object is counted clockwise from the south = 0 */
    sidereal_OPD = baseline * cos(buffer_az_obj*Pi/180.0-((Pi/2.0)+angle_uv)) * cos(buffer_alt_obj*Pi/180.0);

    *projected_baseline = sqrt(baseline*baseline-sidereal_OPD*sidereal_OPD);


    /* 2) Method taken from IDL processing tools, designed originally by mschoell     */
    /* ------------------------------------------------------------------------------ */

    /* compute u,v,w for the current baseline this is double with the previous computation, but this
     allows to check the formulas from two different origins... */
    /* ihar = hour_angle*Pi/12.0; bug corrected - pkervell 25-Oct-02 */
    ihar = hour_angle*Pi/12.0;
    decr = Dec_object*Pi/180.0;
    cha = cos(ihar);
    sha = sin(ihar);
    cdec = cos(decr);
    sdec = sin(decr);
    clat = cos(latitude);
    slat = sin(latitude);

    /* from Markus Schoeller */
    *u = dx*cha - dy*slat*sha + dz*clat*sha;
    *v = dx*sha*sdec + dy*(slat*cha*sdec + clat*cdec) - dz*(clat*cha*sdec - slat*cdec); 
    w = dx*sha*cdec - dy*(slat*cha*cdec - clat*sdec) + dz*(clat*cha*cdec + slat*sdec);

    /* printf(" Projected AWA = %f, Projected MSC = %f \n",*projected_baseline, sqrt((*u)*(*u) + (*v)*(*v)));*/
    /*THE TWO METHODS AGREE PERFECTLY !! INCREDIBLE :-) */
    /* BUT... There may be a bug because the convention of the direction of the azimuth on the ground and
    on sky seem not to agree - pkervell - 25-Oct-02 */

    az = 90 - (atan2(*v,*u) * 180.0/Pi); /* TO BE CHECKED: sign ok ? DIFFERENT FROM IDL !! */
    /* 90-az added - pkervell - 27-Mar-03 */
    if((az < 0 ) && (az >= -180.0)) *azimuth_baseline = az + 180.0;
    else if(az < -180.0) *azimuth_baseline = az + 360.0;
    else if(az >  180.0) *azimuth_baseline = az - 180.0;
    else if((az >=0) && (az <= 180.0)) *azimuth_baseline = az;
    /* additional tests added to have az between 0 and 180 deg - pkervell - 20-Mar-03 */
    /* printf(" Azimuth = %f \n", *azimuth_baseline); */
}

/*
    Conversion of HH MM SS.SS/DD MM SS.SS right ascension and Dec
    to decimal degrees values - pkervell - 18-Oct-02
*/

void vndrs_convert_RA_Dec_to_decimal_degs(double RA_h, double RA_m, double RA_s,
                                          double Dec_d, double Dec_m, double Dec_s,
                                          double *RA, double *Dec) /* out */
{
    double sign;

    if(fabs(Dec_d) > 0.0)
        sign = Dec_d / fabs(Dec_d);
    else sign = 1.0;
    
    *RA  = RA_h * 15.0 + RA_m / 4.0 + RA_s / 240.0;
    *Dec = Dec_d + sign * Dec_m / 60.0 + sign * Dec_s / 3600.0;
}

/*
    Conversion of decimal degrees right ascension and Dec
    to HH MM SS.SS/DD MM SS.SS values - pkervell - 18-Oct-02
*/

void vndrs_convert_decimal_degs_to_RA_Dec(double RA, double Dec, /* in */
                                          double *RA_h, double *RA_m, double *RA_s, /* out */
                                          double *Dec_d, double *Dec_m, double *Dec_s)
{
    double sign;
    
    *RA_h = floor(RA/15.0);
    *RA_m = floor((RA/15.0-*RA_h) * 60.0);
    *RA_s = 240.0 * (RA - *RA_h * 15.0 - *RA_m / 4.0);
    
    if(fabs(Dec) > 0.0)
        sign = Dec/fabs(Dec);
    else
        sign = 1.0;
        
    *Dec_d = sign * floor(fabs(Dec));
    *Dec_m = floor(60.0 * (fabs(Dec) - fabs(*Dec_d)));
    *Dec_s = 60.0 * ((fabs(Dec) - fabs(*Dec_d))*60.0 - *Dec_m);
}


/* ----------- */

