/***********************************************************************

Wavelet transform library
Damien Segransan, Observatoire de Geneve, 2002

modified pkervell, 15-Mar-02
revised  segransan, xx-Jul-02
revised  pkervell, 29-Jul-02
revised  pkervell, 04-Aug-02 Memory management
***********************************************************************/

#include <math.h> 
#include "fftw.h"
#include "stdio.h"
#include "vndrs.h"

#define Pi       3.14159265358979311599796346854418516
#define complex fftw_complex 
#define plan_c  fftw_plan 

/*
n : interferogram dimension. prefer a power of 2. (256, 512, 1024)

y : interferogram

param : this parameter gives the numbers of fringes at the FHWM in the gaussian.
        param = 17 corresponds to 7 fringes at FHWM
        param = (17./7.) * nbfringes
        5 fringes at FHWM =>  param = 12
        9 fringes at FHWM =>  param = 21
wave : is the wavelet transfor of y. This is a N by N/2 array.

We need to make N/2 fourier transform to make the wavelet transform.
It's time consuming but one can speed up the process by doing 
the wavelet transform for the only range of frequencies that are interesting.
X     = (lambda0/5) * i  with i=1..N
sigma = 5/(lambda0*N)*j  with j=-N/2..0..N/2-1
the fringe peak is located at 
sigma0       = +- 1./lambda0  => 5*j0/N=1. <=> j0=N/5.
that correspond to j0 = 51.2 for VINCI

Dsigma0=Dlamdba0/lambda0^2 => 
5/(lambda0*N)*Deltaj0= Dsigma0 <=> Deltaj0 = Dsigma0*(lambda0*N)/5
that correspond to a 9 pixel FHWM for VINCI

We could do the transform from j0-3*Deltaj0 -> j0+3*Deltaj0
which corresponds roughly to a range 24-78 pix => 
54 fourier transforms instead of 128 ! we can round it to 64
*/

/* THERE IS SOME MEMORY NOT RELEASED HERE !! */

/* ------------------------------------------------------------------ */




void vndrs_wavelet(long n,
		   double param,
		   long index_freq_min,
		   long index_freq_max,
                   /* output variables */
		   vector_double *coi,  /* Cone-of-influence [Sec.3g] */
		   vector_double *y,
		   matrix_complex  *wave)
{
/* coi : e-folding factor used for the cone of influence.*/
    long i,j;
    double ymean, d_kwave, fourier_factor;
    
    vector_double *kwave, *scale;
    vector_complex *yfft, *wsave, *daughter;
    
    /*plan_c cp;*/
    fftw_plan cp;
    
    /* Memory allocation */
    kwave     = vndrs_calloc_vector_double(n);
    scale     = vndrs_calloc_vector_double(n);
    yfft      = vndrs_calloc_vector_complex(n); 
    wsave     = vndrs_calloc_vector_complex(n);
    daughter  = vndrs_calloc_vector_complex(n);    

    ymean=0.0;
    for(i=0;i<n;i++) ymean +=  y->data[i]/((double)(n));   
        
    /* ** find the FFT of the time series [Eqn(3)] */
    for (i=0;i<n;i++)
    {
	yfft->data[i].re = y->data[i]-ymean;
	yfft->data[i].im = 0.;
    }

    cp = fftw_create_plan(n,FFTW_FORWARD,FFTW_ESTIMATE);
    fftw_one(cp,yfft->data,wsave->data);
    fftw_destroy_plan(cp);
    
    for (i=0;i<n;i++)
    {
	yfft->data[i].re = wsave->data[i].re/(double)(n);
	yfft->data[i].im = wsave->data[i].im/(double)(n);
    }
    /*** construct the wavenumber array [Eqn(5)] */
    d_kwave = 2.*Pi/((double)(n));
    kwave->data[0] = 0.;
    
    for (i=1 ; i<n/2+1 ; i++) kwave->data[i] = i * d_kwave;
    for (i= n/2+1 ; i<n ; i++) kwave->data[i] = -Pi + (i-n/2+1) * d_kwave;
    
    /* Scaling factor that allow to compare fourier frequency to wavelet's frequency */
    if (param <= 0.) param = 6.;
    
    fourier_factor = Pi*4./(param+sqrt(param*param+2.));
    /*  fprintf(stderr, " fourier factor :  %10.5f   ",fourier_factor ); */

    /* ********* Main wavelet loop *********/
    
    cp= fftw_create_plan(n,FFTW_BACKWARD,FFTW_ESTIMATE);
    
    for (j = (long)index_freq_min; j < (long)index_freq_max; j++)
    {
	scale->data[j] = 1.0 * n/((j+1)*fourier_factor);
	/* CALL WAVE FUNCTION that creates daughter wavelet for the scale j */	
        /* fprintf(stderr, "Iteration %ld; Calls: %ld, Memory: %ld\n", j,
	  vndrs_memory.calloc_vector_complex_calls, vndrs_memory.currently_allocated);*/
		
	vndrs_wave_function(n, param, scale->data[j], kwave, daughter);
        /*coi_vect :  Cone-of-Influence.
	  This is a vector of N points that contains the maximum period 
	  of useful information at that particular time.
	  Frequencies smaller than this are subject to edge effects.
	*/
	coi->data[j]     = sqrt(2) * scale->data[j];
	coi->data[n-j-1] = coi->data[j];

	for (i = 0; i < n; i++)
        {
	    wsave->data[i].re = daughter->data[i].re * yfft->data[i].re - 
		daughter->data[i].im * yfft->data[i].im;
	    wsave->data[i].im = daughter->data[i].re * yfft->data[i].im + 
		daughter->data[i].im * yfft->data[i].re ;
        }
            
	/* **    inverse FFT [Eqn(4)] */
	fftw_one(cp,wsave->data,daughter->data);
	
	/* store the wavelet transform, discard zero-padding at end*/
	for (i = 0; i < n; i++)
	    set_element_matrix_complex( wave, i, j, daughter->data[i]);

	}
  
    fftw_destroy_plan(cp);
    
    /* Memory DE-allocation */
    vndrs_free_vector_complex(daughter);
    vndrs_free_vector_double(kwave);
    vndrs_free_vector_double(scale);
    vndrs_free_vector_complex(yfft);
    vndrs_free_vector_complex(wsave);
    
    return;
    
}    /* END WAVELET INTERFERO*/


/* ------------------------------------------------------------------ */




void vndrs_wave_function(long nk,
			 double param,
			 double scale1,
			 vector_double *kwave,
			 vector_complex *daughter_wave
    )
{
    double expnt,norm,test;
    long  k;
    /*le facteur sqrt(2*Pi/dt) sert a normaliser
      l'energie totale de l'ondelette a N car 
      2*Pi <=> N*dt
      total energy=N   [Eqn(7)]*/
    
    /**************************   Morlet wavelet*/
    if (param < 0.)  param = 6.;
    
    /* total energy=N   [Eqn(7)]*/
    /* Farge Annu review of fluid mechanics, 1992, vol 24,p 395, eq 24*/
    
    norm = sqrt(Pi*2.*scale1) * pow(Pi,-0.25);
    daughter_wave->data[0].re = 0.0; daughter_wave->data[0].im = 0.0;
    test=0.0;

    for (k = 1; k < nk/2+1; k++)
    {
	if(kwave->data[k]>0.)
	    expnt = -0.5 * pow(scale1*kwave->data[k]-param,2);
	else
	    expnt = 0.0;
	
	/* pkervell 17-mar-2002: it is not nice to use the boolean like this: ...*(kwave[k-1]>0.) */
	/* je ne suis pas d'accord avec toi Pierre, au contraire, 
	   c'est tres efficace! D. Segransan 24-mar-2002*/
	
      if((kwave->data[k]>0.) && (expnt>-100.0))
	  daughter_wave->data[k].re = norm*exp(expnt);
      else
	  daughter_wave->data[k].re = 0.0;
      
      daughter_wave->data[k].im = 0.0;
      
      test+= pow(daughter_wave->data[k-1].re,2);      
    }
    
    for (k=nk/2+2;k<=nk;++k)
	{
	daughter_wave->data[k-1].re = 0.; 
	daughter_wave->data[k-1].im = 0.;
	}
    
  
  if ( vndrs_round(test) != nk )
      for (k = 1;k<nk/2+1;k++)
      {
        if(test > 0.0)
            daughter_wave->data[k].re *= sqrt(nk)/sqrt(test);
        else
        {
            daughter_wave->data[k].re *= 0.0;
            printf(" Daughter wave error... \n"); vndrs_error_alert();
        }
      }

  return;
  /* wave_function_interfero_ */
}


matrix_double *vndrs_wavelet_power_spectrum(matrix_complex *wave)
/* MEMORY ALLOCATED INSIDE FUNCTION - pkervell 04-Aug-02 */
{
    long e, f, n;
    double buffer_wavelets;
    matrix_double *wave_power_spectrum;
    
    n = wave->dimension_1;
    
    wave_power_spectrum = vndrs_calloc_matrix_double(n,n/2);   /* added - pkervell -04-Aug-02 */
    
    for(e=0; e<n; e++)
	{
	for(f=0; f<(n/2); f++)
	    {
	    buffer_wavelets = vndrs_abs_complex(get_element_matrix_complex(wave,e,f));
	    set_element_matrix_double(wave_power_spectrum,e,f,buffer_wavelets*buffer_wavelets);
	    }
	}

    return wave_power_spectrum;
    
}

matrix_double *vndrs_wavelet_phase(matrix_complex *wave)
/* MEMORY ALLOCATED INSIDE FUNCTION - pkervell 25-Oct-02 */
/* BEWARE: the position of the center of the scan has to be at the origin of the vector...
   I am not sure how this works with wavelets, and cone of influence, so this is not used for now.
*/
{
    long e, f, n;
    double buffer_wavelets;
    matrix_double *wave_phase;
    
    n = wave->dimension_1;
    
    wave_phase = vndrs_calloc_matrix_double(n,n/2);   /* added - pkervell -04-Aug-02 */
    
    for(e=0; e<n; e++)
	{
	for(f=0; f<(n/2); f++)
	    {
	    buffer_wavelets = vndrs_arg_complex(get_element_matrix_complex(wave,e,f));
	    set_element_matrix_double(wave_phase,e,f,buffer_wavelets);
	    }
	}

    return wave_phase;
    
}

void vndrs_wavelet_power_spectrum_frequency_average(long n,
						    matrix_double *wave_power_spectrum,
						    vector_double  *wave_time)
{
    long e,f;

    for(e=0; e<n; e++)
	{
	wave_time->data[e] =0.0;
	for(f=0; f<(n/2); f++)
	    wave_time->data[e] +=  get_element_matrix_double(wave_power_spectrum,e,f)/((double)(n)/2);
	}
}

void vndrs_wavelet_power_spectrum_time_average(long n,
					       matrix_double *wave_power_spectrum,
					       vector_double  *wave_freq)
{
    long e,f;
    
    for(f=0; f<(n/2); f++)
	{
	wave_freq->data[f] =0.0;
	for(e=0; e<n; e++)
	    wave_freq->data[f] +=  get_element_matrix_double(wave_power_spectrum,e,f)/(double)(n);
	}
}



void vndrs_where2_max(matrix_double *arr, vector_double *posmax)
{
  long i,j,tmp;

  tmp = get_element_matrix_double(arr,0,0);

  for (i = 0; i < arr->dimension_1; i++)
      {
      for (j = 0; j < arr->dimension_2; j++)
	  {
	  if (get_element_matrix_double(arr,i,j)>=tmp)
	      {
	      tmp = get_element_matrix_double(arr,i,j);
	      posmax->data[0]=(double)i;
	      posmax->data[1]=(double)j;
	      posmax->data[2]=get_element_matrix_double(arr,i,j);
	      } 
	  }
      }
}


/*-----------------------------------------------------*/
double FACTORIAL(double m)
{
    double  n;
    n=1.0;
    m = (double) ((long)(m));
    while (m>1) n *= m--;
    return n;
}





